/**********************************************************************/
/*                                                                    */
/*            areas.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "plug.h"

int Band = 1;
int Gui = NOGUI;
/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };
char *Spec =    { "areas" };
char *Comment = { "areas effecter, diffuse the edge of the selection" };
PLUGITEM Item[] = {
{ "width", GETINT,  &Band, 1,    100, "diffuse edge pixel width" },
{ "gui",   SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui", SWITCH,  &Gui,  NOGUI,'N', "not use GUI interface" },
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "width of areas edge", "width=%d" }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/**********************************************************************/

/**********************************************************************/
/*                                                                    */
/*             function get_tmpcolor()                                */
/*                                                                    */
/*             unsigned int *dst : accumlated color <IN/OUT>          */
/*             unsigned int *src : accumlate color  <IN>              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Average the dst+src color and set tmpcolor.  */
/*                                                                    */
/**********************************************************************/
void get_tmpcolor( unsigned int *dst, unsigned int *src )
{
	unsigned int dr, dg, db;
	unsigned int sr, sg, sb;

	dr = GET_RED( *dst );
	dg = GET_GREEN( *dst );
	db = GET_BLUE( *dst );
	sr = GET_RED( *src );
	sg = GET_GREEN( *src );
	sb = GET_BLUE( *src );
	dr += sr; dg += sg; db += sb;
	dr /= 2; dg /= 2; db /= 2; *dst = 0;
	*dst = ((dr << 16) | (dg << 8) | (db));
}


/**********************************************************************/
/*                                                                    */
/*             function set_tmpcolor()                                */
/*                                                                    */
/*             unsigned int *dst : original Image pixel               */
/*             unsigned int *src : tmp accumlated color               */
/*             int rate          : now depth of areas band            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Replace the pixel by the accumlated tmp color.*/
/*                                                                    */
/**********************************************************************/
void set_tmpcolor( unsigned int *dst, unsigned int *src, int rate )
{
	unsigned int dr, dg, db;
	unsigned int sr, sg, sb;

	dr = GET_RED( *dst );
	dg = GET_GREEN( *dst );
	db = GET_BLUE( *dst );
	sr = GET_RED( *src );
	sg = GET_GREEN( *src );
	sb = GET_BLUE( *src );
	dr = ((rate+1) * dr + (Band-rate) * sr) / (Band+1);
	dg = ((rate+1) * dg + (Band-rate) * sg) / (Band+1);
	db = ((rate+1) * db + (Band-rate) * sb) / (Band+1);
	*dst = ( (*dst & MASK) | (dr << 16) | (dg << 8) | (db));
}

/**********************************************************************/
/*                                                                    */
/*             function fat_process()                                 */
/*                                                                    */
/*             char *tmp      : work area                             */
/*             struct Shm *shm: data struct                           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Faten the selected area and exec areas.      */
/*                                                                    */
/**********************************************************************/
void fat_process( char *tmp, struct Shm *shm )
{
	char *p, *q;
	unsigned int *r;
	int i, j, k;
	int count;
	int width, height;
	unsigned int tmpcolor;

	width = shm->width; height = shm->height;

	for( i = 0; i < height * width; i++ ){
	      if( MASKOFF( shm->data[i] ) ){
		      tmp[i] = 1;
	      }
        }
	for( k = 0; k < Band; k++ ){
		p = tmp; q = tmp + width * height; r = shm->data;
		for( i = 0; i < height; i++ ){
			for( j = 0; j < width; j++, p++, q++, r++ ){
				if( *p ){
					*q = *p;
					continue;
				}
				count = 0; tmpcolor = 0;
				if( i > 0 ){
					if( *(p - width) ){
						count += *(p - width);
						get_tmpcolor( &tmpcolor, r - width );
					}
				}
				if( i < height - 1 ){
					if( *(p + width) ){
						count += *(p + width);
						get_tmpcolor( &tmpcolor, r + width );
					}
				}
				if( j > 0 ){
					if( p[-1] ){
						count += p[-1];
						get_tmpcolor( &tmpcolor, &r[-1] );
					}
				}
				if( j < width - 1 ){
					if( p[1] ){
						count += p[1];
						get_tmpcolor( &tmpcolor, &r[1] );
					}
				}
				*q = count;
				if( count ){
					set_tmpcolor( r, &tmpcolor, k );
				}
			}
		}
		memcpy( tmp, tmp + width * height, width * height );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	int width, height;
	char *tmp;

	width = file->buf->width;
	height = file->buf->height;

        if( file->buf->mask_kind == DEPTHMASK ){
		return( EILLMASK );
	}
	if( (tmp = (char *)malloc( width * height * 2 )) == NULL ){
		return( EMALLOC );
	}
	memset( tmp, 0, width * height * 2 );
	fat_process( tmp, file->buf );
	free( tmp );
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
