/**********************************************************************/
/*                                                                    */
/*             band.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>
#include "plug.h"

#define AUTO	0
#define USEGUI	1
#define MANUAL	2
int Gui = AUTO;
unsigned char *GuiStr[] = { "auto", "gui", "manual" };

unsigned int Upper = 255;
unsigned int  Lower = 0;
int Rate = 5;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =	{ "band" };
char *Comment = { "band optimizer, between upper and lower, arrange the color" };

PLUGITEM Item[] = {
{ "gui",   SWITCH,  &Gui, USEGUI,  'G',    "use GUI interface" },
{ "manual", SWITCH, &Gui, MANUAL,  'M',    "not use GUI, set manually" },
{ "auto",  SWITCH,  &Gui, AUTO,  'A',      "not use GUI, set automatically" },
{ "upper", GETINT,  &Upper, 0, 255,      "upper limit of the band optimize" },
{ "lower", GETINT,  &Lower, 0, 255,      "lower limit of the band optimize" },
{ "rate",  GETINT,  &Rate, 1, 100,       "threthold value for auto mode" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ RADIO, 0, "config mode"},
{ ITEM,  0, "GUI(G)", "gui" },
{ ITEM,  1, "Manual(M)", "manual" },
{ ITEM,  2, "Auto(A)", "auto" },
{ RADIOEND, -1 },
{ ENTRY, 5, "percent to flatten B or W", "rate=%d" },
{ ENTRY, 3, "upper threthold", "upper=%d" },
{ ENTRY, 4, "lower threthold", "lower=%d" }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

struct fPosition {	/* for call spline */
	float x;
	float y;
};

struct Packet {		/* packet to send & receive between xband */
	int shmid;	/* IPC shm ID */
	int Band[256];	/* amount of each depth */
} Packet;

unsigned int *Band = Packet.Band;	/* amount of each depth */

struct fPosition *spline_line( struct fPosition *pos, int dim, int num, int pat );

/**********************************************************************/
/*                                                                    */
/*             function change_value()                                */
/*                                                                    */
/*             unsigned int *v : each pixel                           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Change each pixel by dividing between Lower  */
/*                     and Upper.                                     */
/**********************************************************************/
void change_value( unsigned int *v )
{
	unsigned int rgb[3];
	int i;

        rgb[0] = GET_RED(*v);
	rgb[1] = GET_GREEN(*v);
	rgb[2] = GET_BLUE(*v);

	for( i = 0; i < 3; i++ ){
		if( rgb[i] <= Lower ){
			rgb[i] = 0x00000000;
		} else if( rgb[i] >= Upper ){
			rgb[i] = 0x000000ff;
		} else {
			rgb[i] = (rgb[i] - Lower) * 256 / (Upper - Lower);
		}
	}
	*v = ((*v & MASK) | ((rgb[0] << 16) & RED) | 
	      ((rgb[1] << 8) & GREEN) | (rgb[2] & BLUE));
}

/**********************************************************************/
/*                                                                    */
/*             function show_map()                                    */
/*                                                                    */
/*             struct File *file  : Image data                        */
/*                                                                    */
/*             return  int  : 0==success  -1==fail                    */
/*                                                                    */
/*             function: Call X-Interface and set value.              */
/*                                                                    */
/**********************************************************************/
int show_map( struct File *file )
{
	int p1[2], p2[2];
	int sti, sto;
	int ret;
	char buff[256];

	if( pipe( p1 ) == -1 ){
		printf( "cannot open pipe\n" );
		return( -1 );
	}
	if( pipe( p2 ) == -1 ){
		printf( "cannot open pipe\n" );
		return( -1 );
	}
	ret = 0;
	if( fork() ){
		sleep(2);
		sto = dup( 1 );
		close( 1 );
		dup2( p1[1], 1 );
		sti = dup( 0 );
		close( 0 );
		dup2( p2[0], 0 );
		Packet.shmid = file->mes->shmid;
		if( write( 1, &Packet, sizeof(struct Packet) ) <= 0 ){
			close( 1 );
			dup2( sti, 1 );
			fprintf( stderr, "pipe write error\n" );
			return( -1 );
		}
		wait( &ret );
		if( ret == 0 ){
			read( 0, buff, sizeof(Upper) );
			Upper = *(int *)buff;
			read( 0, buff, sizeof(Lower) );
			Lower = *(int *)buff;
		} else {
			ret = EUSRBREAK;
		}
		close( 0 );
		dup2( sti, 0 );
		close( sti );
		close( 1 );
		dup2( sto, 1 );
		close( sto );
		close( p1[0] ); close( p1[1] );
		close( p2[0] ); close( p2[1] );
	} else {
		close( 1 );
		dup2( p2[1], 1 );
		close( 0 );
		dup2( p1[0], 0 );
		memset( buff, 0, 256 );
		sprintf( buff, "%d", Upper );
		sprintf( &buff[30], "%d", Lower );
		sprintf( &buff[50], "%sxband", getenv("GEDO_PATH") );
		execl( &buff[50], "xband", buff, &buff[30], NULL );
		fprintf( stderr, "load program %s\n", &buff[50] );
		perror("");
		exit(1);
	}
	return( ret );
}
	


/**********************************************************************/
/*                                                                    */
/*             function set_band()                                    */
/*                                                                    */
/*             struct File *file  : Image data                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function:  Get amount of depth map.                    */
/*                                                                    */
/**********************************************************************/
void set_band( struct File *file )
{
	int width, height;
	unsigned int *data;
	int i;
	unsigned int val, max;

	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	memset( Band, 0, sizeof(Band) );
	for( i = 0; i < height * width; i++ ){
		val  = GET_RED(data[i]);
		val += GET_GREEN(data[i]);
		val += GET_BLUE(data[i]);
		val /= 3;
		Band[val]++;
	}
	for( max = 0, i = 0; i < 256; i++ ){
		if( Band[i] > max ){
			max = Band[i];
		}
	}
	for( i = 0; i < 256; i++ ){
		Band[i] = Band[i] * 256 / max;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_thread()                                  */
/*                                                                    */
/*             struct fPosition *xy: returns from spline curve        */
/*             int *org            : amount of depth                  */
/*             int onum            : depth number                     */
/*             int xnum            : spline number                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Called from auto_band(), find the nearlest   */
/*                      Upper and Lower.                              */
/**********************************************************************/
void set_thread( struct fPosition *xy, int *org, int onum, int xnum )
{
	int perc;
	int sum;
	int i, flag;
	int j;

	flag = perc = sum = 0;

	for( i = 0; i < onum; i++ ){
		sum += org[i];
	}
	sum = sum * Rate / 100;
	for( j = 0, i = 0; i < xnum; ){
		if( xy[i].y > xy[i+1].y ){
			flag = 1;
			break;
		} else if( xy[i].y < xy[i+1].y ){
			flag = -1;
			break;
		}
		for( i++; j <= xy[i].x; j++ ){
			perc += org[i];
		}
	}
	for( ; i < xnum; i++ ){
		for( ; j <= xy[i].x; j++ ){
			perc += org[i];
		}
		if( flag == 1 ){
			if( xy[i].y < xy[i+1].y || perc > sum  ){
				break;
			}
		} else {
			if( xy[i].y > xy[i+1].y || perc > sum ){
				break;
			}
		}
	}
	Lower = (int)xy[i].x;

	j = onum; perc = 0;
	for( i = xnum - 2; i >= 1; ){
		if( xy[i].y > xy[i-1].y ){
			flag = 1;
			break;
		} else if( xy[i].y < xy[i-1].y ){
			flag = -1;
			break;
		}
		for( i--; j >= xy[i].x; j-- ){
			perc += org[i];
		}
	}
	for( ; i >= 1; i-- ){
		for( ; j >= xy[i].x; j-- ){
			perc += org[i];
		}
		if( flag == 1 ){
			if( xy[i].y < xy[i-1].y || perc > sum ){
				break;
			}
		} else {
			if( xy[i].y > xy[i-1].y || perc > sum ){
				break;
			}
		}
	}
	Upper = (int)xy[i].x;
}


#define PART  8

/**********************************************************************/
/*                                                                    */
/*             function auto_band()                                   */
/*                                                                    */
/*             struct File *file  : Image data                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Find Upper and Lower automatically.          */
/*                                                                    */
/**********************************************************************/
void auto_band( struct File *file )
{
	int i, j;
	unsigned int maxx, maxy, minx, miny;
	struct fPosition xy[64];
	struct fPosition *ret;
	int total;

	for( i = 0; i < 256 / PART; i++ ){
		maxx = maxy = 0;
		minx = miny = 9999;
		for( j = 0; j < PART; j++ ){
			if( Band[i*PART+j] > maxy ){
				maxy = Band[i*PART+j];
				maxx = i*PART+j;
			}
			if( Band[i*PART+j] < miny ){
				miny = Band[i*PART+j];
				minx = i*PART+j;
			}
		}
		if( minx > maxx ){
			xy[i*2].x = (float)maxx;
			xy[i*2].y = (float)maxy;
			xy[i*2+1].x = (float)minx;
			xy[i*2+1].y = (float)miny;
		} else {
			xy[i*2].x = (float)minx;
			xy[i*2].y = (float)miny;
			xy[i*2+1].x = (float)maxx;
			xy[i*2+1].y = (float)maxy;
		}
	}
	if( (ret = spline_line( xy, 3, 256 / PART * 2, 4 )) == NULL ){
		printf( "spline error\n" );
		return;
	}
	total = (256 / PART * 2 - 1) * 4;
	set_thread( ret, Band, 256, total );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	char *str;
	int width, height;
	unsigned int *data;
	int i;

	str = file->mes->u.data;
	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}
	for( i = 0; i < height * width; i++ ){
		if( MASKOFF(data[i]) ){
			change_value( &data[i] );
		}
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	int ret, tmp;
	char *str;

	str = file->mes->u.data;
	ret = 0;
	if( str[0] == '\0' ){
		if( Gui != USEGUI ){
			ret = config_comline( file );
		}
	} else {
		ret = line_config( file );
	}
	if( ret != 0 ){
		return( ret );
	}

	if( Gui == USEGUI ){
		set_band( file );
		auto_band( file );
		if( show_map( file ) != 0 ){
			return( EUSRBREAK );
		}
	} else if( Gui == AUTO ){
		set_band( file );
		auto_band( file );
	}
	if( Lower > Upper ){
		tmp = Lower;
		Lower = Upper;
		Upper = tmp;
	}
	status( CONFIG_ELSE, str );
	return( 0 );
}
