/**********************************************************************/
/*                                                                    */
/*             blush.c  gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "plug.h"

int Boader = 20;
float Va = 0.0f;
float Vb = 1.0f;
float Vc = 0.0f;
#define OBEY	0
#define SIMPLE	1
#define DEPTH	2
int Mode = OBEY;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =    { "blush" };
char *Comment = { "blush selecter" };

PLUGITEM Item[] = {
{ "width", 	GETINT,  &Boader,1, 360,   "blush size" },
{ "a"	, 	GETFLOAT,&Va, 	0, 0, 	   "blush centurity: -1 < 0ff < 0 center < 1" },
{ "b"	, 	GETFLOAT,&Vb, 	0, 0, 	   "blush density:   -0.5 < light < 0 < heavy < 0.1" },
{ "c"	, 	GETFLOAT,&Vc, 	0, 0, 	   "blush solidity:  0 < diffuse < 1 < solid" },
{ "obey", 	SWITCH, &Mode, OBEY, 'O', "obey current mask mode" },
{ "simple", 	SWITCH, &Mode, SIMPLE, 'S', "make simple-mask blush" },
{ "depth", 	SWITCH, &Mode, DEPTH, 'D', "make depth-mask blush"  },
{ "gui",        SWITCH, &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH, &Gui,  NOGUI, 'N', "not use GUI interface" } };
int ItemNum = sizeof(Item) / sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "blush width ", "width=%d" },
{ RADIO,4, "MaskMode " },
{ ITEM, 4, "obey(o)", "obey" },
{ ITEM, 5, "Simple(s)", "simple" },
{ ITEM, 6, "Depth(d)", "depth" },
{ RADIOEND, -1 },
{ BLOCK, -1, "blush parameters" },
{ ENTRY, 1, "Solidity(0 < diffuse < 1 < solid)", "a=%g" },
{ ENTRY, 2, "Density(-0.5 < light < 0 < heavy < 0.1)", "b=%g" },
{ ENTRY, 3, "Centerity(-1 < off < 0 < enforce < 1)", "c=%g" },
{ BLOCKEND, -1, }};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

float Sth;			/* half of circle, for make round figure */
static int Width, Height;	/* width and height of image */

/**********************************************************************/
/*                                                                    */
/*             function randam_dist()                                 */
/*                                                                    */
/*             int val     :  distant from center                     */
/*                                                                    */
/*             return  int :  1==ON  0==OFF                           */
/*                                                                    */
/*             function:  Calc blush point ON/OFF from distant val    */
/*                        for SIMPLEMASK                              */
/**********************************************************************/
int randam_dist( int val )
{
	float tmp;
	float v;

	v = (float)val + Va;
	tmp = (float)rand() / (float)RAND_MAX;
	v  = (Sth - v) / Sth;
	if( tmp > v * Vb + Vc ){
		return( 0 );
	} else {
		return( 1 );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function randam_val()                                  */
/*                                                                    */
/*             int val      : distant from center                     */
/*                                                                    */
/*             return  int  : depth of mask at the point              */
/*                                                                    */
/*             function:  Calc the depth of mask at the point for     */
/*                        DEPTHMASK.                                  */
/**********************************************************************/
int randam_val( int val )
{
	float tmp;
	float v;

	v = (float)val + Va;
	tmp = (float)rand() / (float)RAND_MAX;
	v  = (Sth - v) / Sth;
	v = (v * Vb + Vc) + (tmp - 0.5);
	if( v > 1.0f ){
		return( 0xff );
	} else if( v < 0.0f ){
		return( 0x00 );
	} else {
		return( (int)(v * 256.0f) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function blush()                                       */
/*                                                                    */
/*             int mask            : kind of mask                     */
/*             unsigned int *data  : data adress(top)                 */
/*             struct Pos *st      : selected positions               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: blush selecter effective routine             */
/*                                                                    */
/**********************************************************************/
void blush( int mask, unsigned int *data, struct Pos *st )
{
	int left, up;
	int h, w, b;
	unsigned int *p;
	int i, j;
	int tmp;

	if( Boader == 1 ){
		i = st->x + st->y * Width;
		if( mask == SIMPLEMASK ){
			CLRMASK(data[i]);
		} else {
			data[i] |= MASK;
		}
		return;
	}
	Sth = (Boader / 2 + 1) * (Boader / 2 + 1 );
	if( (w = st->x - Boader / 2) < 0 ){
		left = 0;
		w = Boader + w;
	} else if( (w = st->x + Boader / 2) >= Width ){
		left = st->x - w;
		w = Boader - w + Width - 1;
	} else {
		left = st->x - Boader / 2;
		w = Boader;
	}
	if( (h = st->y - Boader / 2) < 0 ){
		up = 0;
		h = Boader + h;
	} else if( (h = st->y + Boader / 2) >= Height ){
		up = st->y - h;
		h = Boader - h + Height - 1;
	} else {
		up = st->y - Boader / 2;
		h = Boader;
	}
	p = data + up * Width + left;
	b = Boader / 2 + 1;
	for( i = 0; i < h; i++ ){
		for( j = 0; j < w; j++ ){
			tmp = i - b;
			tmp *= tmp;
			tmp += (j-b) * (j-b);
			if( mask == SIMPLEMASK ){
				if( randam_dist( tmp ) ){
					CLRMASK( p[j] );
				}
			} else {
				tmp = randam_val( tmp );
				p[j] &= ~MASK;
				p[j] |= (tmp << 24);
			}
	        }
		p += Width;
       }
}


/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *data, *p;
	int i, rmode;

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	data = file->buf->data;

	check_range( sb, Width, Height, 0, NEGLECT );
	if( sb->defnum < 1 ){
		return( EILLNUMBER );
	}
	if( Mode == SIMPLE ){
		rmode = SIMPLEMASK;
	} else if( Mode == DEPTH ){
		rmode = DEPTHMASK;
	} else {
		rmode = file->buf->mask_kind;
	}
	if( rmode == SIMPLEMASK ){
		for( i = 0, p = data; i < Height * Width; i++, p++ ){
			SETMASK(*p);
		}
	} else {
		for( i = 0, p = data; i < Height * Width; i++, p++ ){
			CLRMASK(*p);
		}
	}

        for( i = 0; i < sb->defnum; i++ ){
		srand( sb->def[i].x + sb->def[i].y );
		blush( rmode, data, &sb->def[i] );
	}
	if( Mode == SIMPLE ){
		file->buf->mask_kind = SIMPLEMASK;
	} else if( Mode == DEPTH ){
		file->buf->mask_kind = DEPTHMASK;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
