/**********************************************************************/
/*                                                                    */
/*              bmp.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*           C1: added grayscale configration                         */
/*           D0: not malloc/free pallet memory.. make static array.   */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include "plug.h"
#include "reduce.h"

#define FULLCOLOR	0
#define PALLET256	1
/* added grayscale */
#define GRAY256		2
#define PALLET16	3
#define PALLET2		4
int WriteMode = FULLCOLOR;

#define WINDOWS		0
#define OS2		1
int OS_flag = WINDOWS;
int Write_OS_flag = WINDOWS;
int Gui = NOGUI;

/********************** configure desciption ******************/
char *Spec = 	{ "bmp" };
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Comment = { "WINDOWS & OS/2 bmp loader/saver" };


PLUGITEM Item[] = {
{ "fullcolor", NOLOAD | SWITCH, &WriteMode, FULLCOLOR, 'F', "fullcolor 24bit save"  },
{ "pallet256", NOLOAD | SWITCH, &WriteMode, PALLET256, 'P', "256 color pallet save" },
/* added grayscale */
{ "gray", 	NOLOAD | SWITCH, &WriteMode, GRAY256, 'G', "grayscale save" },
{ "pallet16",  NOLOAD | SWITCH, &WriteMode, PALLET16, '1', "16 color pallet save" },
{ "bilevel",   NOLOAD | SWITCH, &WriteMode, PALLET2, 'B', "bilevel save" },
{ "windows",   NOLOAD | SWITCH, &Write_OS_flag, WINDOWS, 'W', "WINDOWS format save" },
{ "os2",       NOLOAD | SWITCH, &Write_OS_flag, OS2, 'O', "OS/2 format save" },
{ "gui",        NOLOAD | SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      NOLOAD | SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
 };
int      ItemNum = sizeof(Item) / sizeof(PLUGITEM);

struct Tags Tag[] = {
{ NOLOAD | RADIO, 0, "Color Mode" },
{ NOLOAD | ITEM, 0, "FullColor(F)", "fullcolor" },
{ NOLOAD | ITEM, 1, "256-Color(P)", "pallet256" },
{ NOLOAD | ITEM, 2, "grayscale(G)", "gray" },
{ NOLOAD | ITEM, 3, "16-Color(1)", "pallet16" },
{ NOLOAD | ITEM, 4, "B/W(B)",    "bilevel" },
{ NOLOAD | RADIOEND, -1 },
{ NOLOAD | RADIO, 4, "OS Mode" },
{ NOLOAD | ITEM, 5, "Windos BMP(W)", "windows" },
{ NOLOAD | ITEM, 6, "OS/2 BMP(O)", "os2" },
{ NOLOAD | RADIOEND, -1 },
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);
DEFAULTSTATUS;

/********************** description end ********************/

#define BYTE unsigned char
#define WORD unsigned short

#pragma pack(1)
struct BitMapFileHeader {	/* Windows and OS/2 header */
	BYTE bfType[2];		/* 'B' 'M' signature */
	long bfSize;		/* filesize */
	unsigned short bfReserved1;
	unsigned short bfReserved2;
	long bfOffBits;		/* start of image */
	long bciSize;		/* original info member, info size */
};

#pragma pack(1)
struct BitMapInfoHeader {	/* Windows info */
	long biWidth;		/* width */
	long biHeight;		/* height */
	unsigned short biPlanes;	/* == 1 */
	unsigned short biBitCount;	/* 1==bilevel, 4==pallet16,
					8==pallet256, 24==fullcolor */
	long biCompression;		/* 0==normal, 1==compressed */
	long biSizeImage;		/* not set */
	long biXPlesPerMeter;		/* not set */
	long biYPelsPerMeter;		/* not set */
	long biClrUsed;			/* pallet entry */
	long biClrImportant;		/* pallet entry */
};

#pragma pack(1)
struct RGBQuad {	/* Windows color table */
	BYTE rgbBlue;
	BYTE rgbGreen;
	BYTE rgbRed;
	BYTE rgbReserved;
};

#pragma pack(1)
struct BitMapCoreInfoHeader {		/* OS/2 info */
	unsigned short bcWidth;		/* width */
	unsigned short bcHeight;	/* height */
	unsigned short bcPlanes;	/* == 1 */
	unsigned short bcBitCount;	/* 1==bilevel, 4==pallet16,
					8==pallet256, 24==fullcolor */
};

#pragma pack(1)
struct RGBTriple {	/* OS/2 color table */
	BYTE rgbBlue;
	BYTE rgbGreen;
	BYTE rgbRed;
};

#pragma pack(1)
union RGB {
	struct RGBQuad q;
	long  d;
};

union RGB Rgb[256];	/* Color Table for Windows */
struct RGBTriple Rgbt[256];	/* Color Table for OS/2 */
int BitCount;		/* 1==biLevel, 4==16 color, 8==256 Color, 24==24bit */
int Compression = 0;	/* compression flag */
int ImageSize;		/* Total size of Compressed Image */
unsigned char *Shadow;	/* decode buffer */
unsigned char *Shp;	/* decode buffer pointer */

int LSB_flag;		/* machine is LSB or MSB */

static unsigned int BinMask[8] /* Bit Mask */
= { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

/**********************************************************************/
/*                                                                    */
/*             function wconv()                                       */
/*                                                                    */
/*             unsigned short v : SHORT data                          */
/*                                                                    */
/*             return  unsigned short : SHORT data                    */
/*                                                                    */
/*             function: convert byte order(SHORT) if MSB machine     */
/*                                                                    */
/**********************************************************************/
unsigned short wconv( unsigned short v )
{
	register unsigned short v1, v2;

	if( LSB_flag ){
		return( v );
	}
	v1 = (v & 0x00ff) << 8;
	v2 = (v & 0xff00) >> 8;
	return( v1 | v2 );
}

/**********************************************************************/
/*                                                                    */
/*             function dconv()                                       */
/*                                                                    */
/*             unsigned long v : LONG data                            */
/*                                                                    */
/*             return  unsigned long : LONG data                      */
/*                                                                    */
/*             function: conver byte order(LONG) if MSB machine       */
/*                                                                    */
/**********************************************************************/
unsigned long dconv( unsigned long v )
{
	register unsigned long v1, v2, v3, v4;
	
	if( LSB_flag ){
		return( v );
	}
	v1 = (v & 0x000000ff) << 24;
	v2 = (v & 0x0000ff00) << 8;
	v3 = (v & 0x00ff0000) >> 8;
	v4 = (v & 0xff000000) >> 24;
	return( v1 | v2 | v3 | v4 );
}

/**********************************************************************/
/*                                                                    */
/*            function windows_read()                                 */
/*                                                                    */
/*            args  struct File *file  : file data                    */
/*                  struct BitMapFileHeader *header : read header     */
/*                  int size   : left header size                     */
/*                                                                    */
/*            return int   : 0==success  0!=errorcode                 */
/*                                                                    */
/*            function: Read Windows headers.                         */
/*                                                                    */
/**********************************************************************/
int windows_read( struct File *file, struct BitMapFileHeader *header, int size )

{
	struct BitMapInfoHeader info;

	if( read( file->fd, &info, size-4 ) != size-4 ){
		file->mes->datasize = 0;
		return( EREADFILE );
	}

	info.biBitCount = wconv( info.biBitCount );
	info.biClrUsed  = dconv( info.biClrUsed );
	info.biWidth    = dconv( info.biWidth );
	info.biHeight	= dconv( info.biHeight );
	info.biCompression = dconv( info.biCompression );
	info.biSizeImage = dconv( info.biSizeImage );
	header->bfOffBits = dconv( header->bfOffBits );
	header->bfSize	= dconv( header->bfSize );


	if( info.biClrUsed == 0L && info.biBitCount != 24 ){
		info.biClrUsed = (long)(1 << info.biBitCount);
	}
	if( info.biBitCount != 24 ){
		if( read( file->fd, Rgb, sizeof(struct RGBQuad) * (size_t)info.biClrUsed ) != sizeof(struct RGBQuad) * (size_t)info.biClrUsed ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EREADFILE );
		}
	}

	if( info.biCompression ){
		ImageSize = header->bfSize - header->bfOffBits;
		Compression = 1;
	} else {
		Compression = 0;
	}

	file->buf->width = info.biWidth;
	file->buf->height = info.biHeight;
	BitCount = info.biBitCount;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function os2_read()                                     */
/*                                                                    */
/*            args  struct File *file  : file data                    */
/*                  struct BitMapFileHeader *header  : read header    */
/*                                                                    */
/*            return int   : 0==success  0!=errorcode                 */
/*                                                                    */
/*            function: Read OS/2 headers.                            */
/*                                                                    */
/**********************************************************************/
int os2_read( struct File *file, struct BitMapFileHeader *header )
{
	struct BitMapCoreInfoHeader cinfo;
	int size;

	if( read( file->fd, &cinfo, 8 ) != 8 ){
		file->mes->datasize = 0;
		return( EREADFILE );
	}

	cinfo.bcBitCount = wconv( cinfo.bcBitCount );
	cinfo.bcWidth    = wconv( cinfo.bcWidth );
	cinfo.bcHeight	= wconv( cinfo.bcHeight );

	if( cinfo.bcBitCount != 24 ){
		switch( cinfo.bcBitCount ){
		case 8:
			size = 3 * 256;
			break;
		case 4:
			size = 3 * 16;
			break;
		case 1:
			size = 3 * 2;
			break;
		default:
			return( EILLFORM );
			break;
		}
		if( read( file->fd, Rgbt, size ) != size ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EREADFILE );
		}
	}

	Compression = 0;

	file->buf->width = cinfo.bcWidth;
	file->buf->height = cinfo.bcHeight;
	BitCount = cinfo.bcBitCount;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function read_header()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ HEADER of the SPEC format from file     */
/*          B0  can deal bilevel bmp                                  */
/*          B2  can deal compression                                  */
/**********************************************************************/
int read_header( struct File *file )
{
	struct BitMapFileHeader header;
	int ret;

	if( (file->fd = open( file->name, O_RDONLY )) == -1 ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EOPENFILE );
	}
	if( read( file->fd, &header, sizeof(struct BitMapFileHeader) ) != sizeof(struct BitMapFileHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EREADFILE );
	}
	if( header.bfType[0] != 'B' || header.bfType[1] != 'M' ){
		file->mes->datasize = 0;
		return( EILLFORM );
	}
	if( header.bciSize == 12 ){
		OS_flag = OS2;
		ret = os2_read( file, &header );
	} else {
		OS_flag = WINDOWS;
		ret = windows_read( file, &header, header.bciSize );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function set_colortable()                              */
/*                                                                    */
/*             unsigned int *p : expanded pixel buffer                */
/*             int at :          color table pixel                    */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: Expand color table pixel to RGB color.       */
/*                                                                    */
/**********************************************************************/
static void set_colortable( unsigned int *p, int at )
{
	if( OS_flag == WINDOWS ){
		*p = (((unsigned int)(Rgb[at].q.rgbRed) << 16) & RED );
		*p |= (((unsigned int)(Rgb[at].q.rgbGreen) << 8) & GREEN );
		*p |= ((unsigned int)(Rgb[at].q.rgbBlue) & BLUE );
	} else {
		*p = (((unsigned int)(Rgbt[at].rgbRed) << 16) & RED );
		*p |= (((unsigned int)(Rgbt[at].rgbGreen) << 8) & GREEN );
		*p |= ((unsigned int)(Rgbt[at].rgbBlue) & BLUE );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function init_decomp()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: Get decompress buffer and read all image.    */
/*                                                                    */
/**********************************************************************/
static int init_decomp( struct File *file )
{
	if( (Shadow = malloc( ImageSize )) == NULL ){
		file->mes->datasize = 0;
		return( EMALLOC );
	}
	if( read( file->fd, Shadow, ImageSize ) != ImageSize ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		free( Shadow );
		return( EREADFILE );
	}
	Shp = Shadow;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function next_decomp()                                 */
/*                                                                    */
/*             unsigned char *buff: must be set buffer                */
/*             int size :           must be the size after decompress */
/*                                                                    */
/*             return  int: 0=Success other=Fail(illegal comp data)   */
/*                                                                    */
/*             function: Decompression do.                            */
/*                                                                    */
/**********************************************************************/
static int next_decomp( unsigned char *buff, int size )
{
	int nybble;
	int status;
	int count;
	int i;
	unsigned char *p;

	p = buff;
	if( BitCount == 4 ){
		status = 0;
		while( 1 ){
			if( *Shp == 0 ){
				Shp++;
				if( *Shp == 0 ){
					Shp++;
					break;
				} else if( *Shp == 1 ){
					Shp++;
					free( Shadow );
					break;
				} else {
					count = *Shp++;
					nybble = 0;
					for( i = 0; i < count; i++ ){
						if( nybble == 0 ){
							if( status == 0 ){
								*p = (*Shp & 0xf0);
								status = 1;
							} else {
								*p |= (*Shp >> 4);
								p++;
								status = 0;
							}
							nybble = 1;
						} else {
							if( status == 0 ){
								*p = (*Shp << 4);
								status = 1;
							} else {
								*p |= (*Shp & 0x0f);
								p++;
								status = 0;
							}
							nybble = 0;
							Shp++;
						}
					}
					if( count % 2 ){
						count++;
						Shp++;
					}
					count /= 2;
					if( count % 2 ){
						Shp++;
					}
				}
			} else {
				count = *Shp++;
				nybble = 0;
				for( i = 0; i < count; i++ ){
					if( nybble == 0 ){
						if( status == 0 ){
							*p = (*Shp & 0xf0);
							status = 1;
						} else {
							*p |= (*Shp >> 4);
							status = 0;
							p++;
						}
						nybble = 1;
					} else {
						if( status == 0 ){
							*p = (*Shp << 4);
							status = 1;
						} else {
							*p |= (*Shp & 0x0f);
							status = 0;
							p++;
						}
						nybble = 0;
					}
				}
				Shp++;
			}
			if( p - buff >= size ){
				return( -1 );
			}
		}
	} else if( BitCount == 8 ){
		while( 1 ){
			if( *Shp == 0 ){
				Shp++;
				if( *Shp == 0 ){
					Shp++;
					break;
				} else if( *Shp == 1 ){
					Shp++;
					free( Shadow );
					break;
				} else {
					count = *Shp++;
					for( i = 0; i < count; i++ ){
						*p++ = *Shp++;
					}
					if( count % 2 ){
						Shp++;
					}
				}
			} else {
				count = *Shp++;
				for( i = 0; i < count; i++ ){
					*p++ = *Shp;
				}
				Shp++;
			}
			if( p - buff >= size ){
				return( -1 );
			}
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function read_image()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ IMAGES of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_image( struct File *file )
{
	int size;
	unsigned char *buff;
	unsigned int *data;
	struct Shm *buf;
	int i, k;
	int j;
	unsigned int v;
	int ret;

	buf = file->buf;
	switch( BitCount ){
	case 1:
		size = (buf->width + 7) / 8;
		break;
	case 4:
		size = (buf->width + 1) / 2;
		break;
	case 8:
		size = buf->width;
		break;
	case 24:
		size = buf->width * 3;
		break;
	default:
		size = 0;
		break;
	}
        if( size % 4 ){
		size = size + (4 - size % 4);
	}

	if( Compression ){
		ret = init_decomp( file );
		if( ret != 0 ){
			close( file->fd );
			return( ret );
		}
	}
	if( (buff = malloc( size )) == NULL ){
		file->mes->datasize = 0;
		close( file->fd );
		return( EMALLOC );
	}

        for( k = file->buf->height - 1; k >= 0; k-- ) {
		data = file->buf->data + k * file->buf->width;
		if( Compression ){
			if( next_decomp( buff, size ) != 0 ){
				close( file->fd );
				free( buff );
				return( EREADFILE );
			}
		} else {
			if( read( file->fd, buff, size ) != size ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				free( buff );
				return( EREADFILE );
			}
		}
		if( BitCount == 1 ){
			for( i = 0; i < (buf->width + 7) / 8; i++ ){
				for( j = 0; j < 8; j++, data++ ){
					if( i * 8 + j >= buf->width ){
						break;
					}
					if( BinMask[j] & buff[i] ){
						set_colortable( data, 1 );
					} else {
						set_colortable( data, 0 );
					}
				}
			}
		} else if( BitCount == 4 ){
			for( i = 0; i < (buf->width + 1) / 2; i++ ){
				for( j = 0; j < 2; j++, data++ ){
					if( i * 2 + j >= buf->width ){
						break;
					}
					if( j == 0 ){
						v = (buff[i] >> 4);
					} else {
						v = (buff[i] & 0x0f );
					}
					set_colortable( data, v );
				}
			}
		} else if( BitCount == 8 ){
			for( i = 0; i < file->buf->width; i++, data++ ){
				set_colortable( data, buff[i] );
			}
		} else  {
		        for( i = 0; i < file->buf->width; i++, data++ ) {
				*data = (((unsigned int)(buff[i*3+2]) << 16) & RED );
				*data |= (((unsigned int)(buff[i*3+1]) << 8) & GREEN );
				*data |= ((unsigned int)(buff[i*3]) & BLUE );
			}
		}
	}
	file->buf->mask_kind = SIMPLEMASK;
	close( file->fd );
	free( buff );
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function write_header()                                */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE HEADER of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_header( struct File *file )
{
	struct BitMapFileHeader header;
	struct BitMapInfoHeader info;
	struct BitMapCoreInfoHeader cinfo;
	int size;

	if( (file->fd = open( file->name, O_RDWR | O_TRUNC )) == -1 ){
		if( (file->fd = open( file->name, O_RDWR | O_CREAT, 0664)) == -1 ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
	}
	memset( &info, 0, sizeof(info) );
	memset( &cinfo, 0, sizeof(cinfo) );
	memset( &header, 0, sizeof(header) );


	header.bfType[0] = 'B';
	header.bfType[1] = 'M';
	if( Write_OS_flag == WINDOWS ){
		header.bciSize = 40L;
		info.biWidth = file->buf->width;
		info.biHeight = file->buf->height;
		switch( WriteMode ){
		case FULLCOLOR:
			info.biBitCount = 24;
			size = file->buf->width * 3;
			break;
		case PALLET256:
		case GRAY256:	/* added grayscale */
			info.biBitCount = 8;
			info.biClrUsed = 256;
			info.biClrImportant = 256;
			size = file->buf->width;
			break;
		case PALLET16:
			info.biBitCount = 4;
			info.biClrUsed = 16;
			info.biClrImportant = 16;
			size = (file->buf->width+1) / 2;
			break;
		case PALLET2:
			info.biBitCount = 1;
			info.biClrUsed = 2;
			info.biClrImportant = 2;
			size = (file->buf->width+7) / 2;
			break;
		default:
			return( -1 );
			break;
		}
		header.bfOffBits = (long)(sizeof(header) + sizeof(info));
		header.bfOffBits += info.biClrUsed * sizeof(struct RGBQuad);
		if( size % 4 ){
			size += (4L - size % 4);
		}
		header.bfSize = size * info.biHeight + header.bfOffBits;
		info.biPlanes = 1;

		header.bfSize = dconv(header.bfSize);
		header.bfOffBits = dconv(header.bfOffBits);
		header.bciSize = dconv(header.bciSize);
		info.biWidth = dconv(info.biWidth);
		info.biHeight = dconv(info.biHeight);
		info.biPlanes = wconv(info.biPlanes);
		info.biBitCount = wconv(info.biBitCount);
		info.biClrUsed = dconv(info.biClrUsed);
		info.biClrImportant = dconv(info.biClrImportant);

		if( write( file->fd, &header, sizeof(header) ) != sizeof(header) ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EWRITEFILE );
		}	
		if( write( file->fd, &info, sizeof(info) ) != sizeof(info) ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EWRITEFILE );
		} /* end of WINDOWS */
	} else {	/* OS2 */
		int map;

		header.bciSize = 12L;
		cinfo.bcWidth = file->buf->width;
		cinfo.bcHeight = file->buf->height;
		switch( WriteMode ){
		case FULLCOLOR:
			cinfo.bcBitCount = 24;
			size = file->buf->width * 3;
			map = 0;
			break;
		case PALLET256:
		case GRAY256: /* added grayscale */
			cinfo.bcBitCount = 8;
			size = file->buf->width;
			map = 256;
			break;
		case PALLET16:
			cinfo.bcBitCount = 4;
			size = (file->buf->width+1) / 2;
			map = 16;
			break;
		case PALLET2:
			cinfo.bcBitCount = 1;
			size = (file->buf->width+7) / 2;
			map = 2;
			break;
		default:
			return( -1 );
			break;
		}
		header.bfOffBits = (long)(sizeof(header) + sizeof(cinfo));
		header.bfOffBits += map * sizeof(struct RGBTriple);
		if( size % 4 ){
			size += (4L - size % 4);
		}
		header.bfSize = size * cinfo.bcHeight + header.bfOffBits;
		cinfo.bcPlanes = 1;

		header.bfSize = dconv(header.bfSize);
		header.bfOffBits = dconv(header.bfOffBits);
		header.bciSize = dconv(header.bciSize);
		cinfo.bcWidth = dconv(cinfo.bcWidth);
		cinfo.bcHeight = dconv(cinfo.bcHeight);
		cinfo.bcPlanes = wconv(cinfo.bcPlanes);
		cinfo.bcBitCount = wconv(cinfo.bcBitCount);

		if( write( file->fd, &header, sizeof(header) ) != sizeof(header) ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EWRITEFILE );
		}	
		if( write( file->fd, &cinfo, sizeof(cinfo) ) != sizeof(cinfo) ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EWRITEFILE );
		}
	} /* end of OS2 */

	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE IMAGES of the SPEC format to file       */
/*             D0: use reduce pallet static area(not malloc).         */     
/*             D0: free reduce image buffer                           */
/*                                                                    */
/**********************************************************************/
int write_image( struct File *file )
{
	int size, ret;
	unsigned char *buff;
	struct Shm *buf;
	int i, j, k;
	unsigned int *data;
	unsigned char *bdata;
	struct ReduceInfo ri;
/* D0 add, use reduce pallet static area(not malloc) */     
	static struct Color pallet[256];
/* end D0 */

	buf = file->buf;
	switch( WriteMode ){
	case FULLCOLOR:
		size = file->buf->width * 3;
		break;
	case PALLET256:
	case GRAY256:	/* added grayscale */
		size = file->buf->width;
		break;
	case PALLET16:
		size = (file->buf->width+1) / 2;
		break;
	case PALLET2:
		size = (file->buf->width+7) / 8;
		break;
	default:
		return( -1 );
		break;
	}
	if( size % 4 ){
		size = size + (4 - size % 4);
	} else {
		size = size;
	}

	if( (buff = malloc( size )) == NULL ){
		file->mes->datasize = 0;
		close( file->fd );
		return( EMALLOC );
	}
	ri.submode = 0;
	switch( WriteMode ){
	case PALLET256:
		ri.mode = GETPALLETCOLOR;
		ri.target = 256;
		ri.tmp_colors = 16;
		ri.submode = (COLOR | TOTALNUM | ZAGGY);
		break;
/* added grayscale */
	case GRAY256:
		ri.mode = GETPALLETCOLOR;
		ri.target = 256;
		ri.tmp_colors = 256;
		ri.submode = (BW | INDEPENDENT | FLAT);
		break;
	case PALLET16:
		ri.mode = GETPALLETCOLOR;
		ri.target = 16;
		ri.tmp_colors = 4;
		ri.submode = (COLOR | TOTALNUM | ZAGGY);
		break;
	case PALLET2:
		ri.mode = GETPALLETCOLOR;
		ri.target = 2;
		ri.tmp_colors = 2;
		ri.submode = (BW | TOTALNUM | ZAGGY);
		break;
	}
	if( WriteMode != FULLCOLOR ){
		if( (ri.image = malloc( file->buf->width * file->buf->height )) == NULL ){
			return( EMALLOC );
		}
/* D0 delete, use reduce pallet static area(not malloc)
		if( (ri.pallet = malloc( sizeof(struct Color) * ri.target )) == NULL ){
			return( EMALLOC );
		}*/
/* D0 add */
		ri.pallet = pallet;
/* end D0 */
		if( (ret = get_reduce_image( file->buf, &ri )) != 0 ){
			return( ret );
		}
		if( Write_OS_flag == WINDOWS ){
			for( i = 0; i < ri.target; i++ ){
				Rgb[i].q.rgbRed = ri.pallet[i].R;
				Rgb[i].q.rgbGreen = ri.pallet[i].G;
				Rgb[i].q.rgbBlue = ri.pallet[i].B;
			}
			if( write( file->fd, Rgb, sizeof(struct RGBQuad) * ri.target ) <= 0 ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				free( buff );
				return( EWRITEFILE );
			}
		} else {
			for( i = 0; i < ri.target; i++ ){
				Rgbt[i].rgbRed = ri.pallet[i].R;
				Rgbt[i].rgbGreen = ri.pallet[i].G;
				Rgbt[i].rgbBlue = ri.pallet[i].B;
			}
			if( write( file->fd, Rgbt, sizeof(struct RGBTriple) * ri.target ) <= 0 ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				free( buff );
				return( EWRITEFILE );
			}
		}
/* D0 add, check free image buffer */
	} else {
	     ri.image = NULL;
	     ri.pallet = NULL;
/* end D0 */
	}

	if( WriteMode == FULLCOLOR ){
		for( i = buf->height - 1; i >= 0; i-- ){
			data = buf->data + i * buf->width;
			for( j = 0, k = 0; k < buf->width; k++ ){
				buff[j++] = (unsigned char)GET_BLUE(data[k]);
				buff[j++] = (unsigned char)GET_GREEN(data[k]);
				buff[j++] = (unsigned char)GET_RED(data[k]);
			}
			if( write( file->fd, buff, size ) != size ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				free( buff );
				return( EWRITEFILE );
			}
		}
/* added grayscale */
	} else if( WriteMode == PALLET256 || WriteMode == GRAY256 ){
		for( i = buf->height - 1; i >= 0; i-- ){
			bdata = ri.image + i * buf->width;
			for( j = 0, k = 0; k < buf->width; k++, j++ ){
				buff[j] = bdata[k];
			}
			if( write( file->fd, buff, size ) != size ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				free( buff );
/* D0 add */
				free( ri.image );
/* end D0 */
				return( EWRITEFILE );
			}
		}
	} else if( WriteMode == PALLET16 ){
		for( i = buf->height - 1; i >= 0; i-- ){
			bdata = ri.image + i * buf->width;
			for( j = 0, k = 0; k < buf->width; k++ ){
				if( k % 2 ){
					buff[j++] |= bdata[k];
				} else {
					buff[j] = (bdata[k] << 4);
				}
			}
			if( write( file->fd, buff, size ) != size ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				free( buff );
/* D0 add */
				free( ri.image );
/* end D0 */
				return( EWRITEFILE );
			}
		}
	} else if( WriteMode == PALLET2 ){
		int count;
		for( i = buf->height - 1; i >= 0; i-- ){
			bdata = ri.image + i * buf->width;
			count = 0;
			for( j = 0; j < size; j++ ){
				buff[j] = 0;
				for( k = 0; k < 8; k++ ){
					if( count++ >= buf->width ){
						break;
					}
					if( *bdata++ ){
						buff[j] |= BinMask[k];
					}
				}
			}
			if( write( file->fd, buff, size ) != size ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				free( buff );
/* D0 add */
				free( ri.image );
/* end D0 */
				return( EWRITEFILE );
			}
		}
	}
	close( file->fd );		
	free( buff );
/* D0 add */
	if( ri.image ){
	     free( ri.image );
	}
/* end D0 */
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
