/**********************************************************************/
/*                                                                    */
/*           circle.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "plug.h"

int Border = 2;

#define CIRCLE  0
#define ELLIPSE 1
#define ARC     2

int Mode = CIRCLE;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };
char *Spec = 	{ "circle" };
char *Comment = { "circle/ellipse/arc selecter" };

PLUGITEM Item[] = {
{ "width", 	GETINT,  &Border,1, 360,   "circle line width" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" },
{ "circle", 	SWITCH, &Mode, CIRCLE, 'C', "select circle min(X,Y)"  },
{ "ellipse", 	SWITCH, &Mode, ELLIPSE, 'E', "select ellipse X-Y" },
{ "arc", SWITCH | INITONLY, &Mode, ARC, 'A', "draw pi" },
};
int ItemNum = sizeof(Item) / sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "circle line width  ", "width=%d" },
{ RADIO, 3, "Mode " },
{ ITEM, 3, "circle(c)", "circle" },
{ ITEM, 4, "ellipse(e)", "ellipse" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

void status( int kind, char *buff )
{
	if( Mode == ARC ){
		TagNum = 1;
	} else {
		TagNum = sizeof(Tag) / sizeof(struct Tags);
	}
	get_default_status( kind, buff);
}
/********************** description end ********************/

int Width;		/* image width == file->buf->width */
int Height;		/* image height == file->buf->height */
unsigned int *Data;	/* data top == file->buf->data */

#define GETX2Y2(x,y)	((x-centerx)*(x-centerx)+(y-centery)*(y-centery))

/**********************************************************************/
/*                                                                    */
/*             function do_circle()                                   */
/*                                                                    */
/*             struct Pos *p1 : Point 1                               */
/*             struct Pos *p2 : Point 2                               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Select minimal p1 and p2 full circle.        */
/*                                                                    */
/**********************************************************************/
void do_circle( struct Pos *p1, struct Pos *p2 )
{
	int i, j;
	int minx, maxx, miny, maxy;
	int centerx, centery;
	int radmax, radmin;
	int tmp;
	unsigned int *p;

	minx = min( p1->x, p2->x );
	miny = min( p1->y, p2->y );
	maxx = max( p1->x, p2->x );
	maxy = max( p1->y, p2->y );
	centerx = (maxx + minx) / 2;
	centery = (maxy + miny) / 2;
	radmax = min( maxx-minx, maxy-miny );
	radmax = radmax / 2;
	radmin = radmax - Border;
	radmin = (radmin>0)?(radmin):(0);
	radmax = radmax * radmax;
	radmin = radmin * radmin;
	
	p = &Data[miny * Width];
	for( i = miny; i <= maxy; i++ ){
		for( j = minx; j <= maxx; j++ ){
			tmp = GETX2Y2( j, i );
			if( tmp <= radmax && tmp >= radmin ){
				CLRMASK(p[j]);
			}
		}
		p += Width;
	}
}

#define GETELLI(x,y)	((x-centerx)*(x-centerx)*rady+(y-centery)*(y-centery)*radx)

/**********************************************************************/
/*                                                                    */
/*             function do_ellipse()                                  */
/*                                                                    */
/*             struct Pos *p1 : Point 1                               */
/*             struct Pos *p2 : Point 2                               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Select Ellipse in p1 to p2.                  */
/*                                                                    */
/**********************************************************************/
void do_ellipse( struct Pos *p1, struct Pos *p2 )
{
	int minx, maxx, miny, maxy;
	int radmx, radmy, radxy, radmxy;
	int i, j;
	int centerx, centery;
	int radx, rady;
	int tmp;
	unsigned int *p;

	minx = min( p1->x, p2->x );
	miny = min( p1->y, p2->y );
	maxx = max( p1->x, p2->x );
	maxy = max( p1->y, p2->y );
	centerx = (maxx + minx) / 2;
	centery = (maxy + miny) / 2;
	radx    = (maxx - minx) / 2;
	rady	= (maxy - miny) / 2;
	radmx =  radx - Border;
	radmx =  (radmx>0)?(radmx):(0);
	radmy =  rady - Border;
	radmy =  (radmy>0)?(radmy):(0);
	radx = radx * radx;
	rady = rady * rady;
	radxy = radx * rady;
	radmx = radmx * radmx;
	radmy = radmy * radmy;
	radmxy = radmx * radmy;

	p = &Data[miny * Width];
	for( i = miny; i <= maxy; i++ ){
		for( j = minx; j <= maxx; j++ ){
			tmp = GETELLI( j, i );
			if( tmp <= radxy && tmp >= radmxy ){
				CLRMASK(p[j]);
			}
		}
		p += Width;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_arc()                                      */
/*                                                                    */
/*             struct Pos *p1 : Position 1                            */
/*             struct Pos *p2 : Position 2                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw arc by p1 and p2.                       */
/*                                                                    */
/**********************************************************************/
void do_arc( struct Pos *p1, struct Pos *p2 )
{
	int rad, rrad;
	int centerx, centery;
	int minx, maxx, miny, maxy;
	int radmin;
	int i, j, tmp;
	unsigned int *p;

	rad = (p1->x - p2->x) * (p1->x - p2->x) + (p1->y - p2->y) * (p1->y -p2->y);;
	rrad = (int)sqrt((double)rad);
	centerx = p1->x;
	centery = p1->y;
	radmin = (rrad - Border < 0)?(0):(rrad - Border);
	radmin = radmin * radmin;
	minx = (centerx - rrad > 0)?(0):(centerx - rrad);
	maxx = (centerx + rrad >= Width)?(Width-1):(centerx + rrad);
	miny = (centery - rrad > 0)?(0):(centery - rrad);
	maxy = (centerx + rrad <= Height)?(Height-1):(centerx + rrad);
	
	p = &Data[miny * Width];
	for( i = miny; i <= maxy; i++ ){
		for( j = minx; j <= maxx; j++ ){
			tmp = GETX2Y2( j, i );
			if( tmp <= rad && tmp >= radmin ){
				CLRMASK(p[j]);
			}
		}
		p += Width;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	int i;
	unsigned int *p;

	if( Mode == ARC ){	/* trick the configure */
		ItemNum = 1;
	} else {
		ItemNum = 3;
	}

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	Data = file->buf->data;

	switch( Mode ){
	case CIRCLE:
	case ELLIPSE:
		if( sb->selnum != SKBOX ){
			return( EILLSELECTION );
		}
		break;
	case ARC:
		if( sb->selnum != SKLINES ){
			return( EILLSELECTION );
		}
		break;
	default:
		return( EILLSELECTION );
		break;
	}
	check_range( sb, Width, Height, 2, ADJUST );
	if( sb->defnum < 2 ){
		return( EILLNUMBER );
	}

	for( i = 0, p = Data; i < Height * Width; i++, p++ ){
		SETMASK(*p);
	}
	if( Mode != ARC ){
		if( sb->defnum % 2 == 1 ){
			sb->defnum--;
		}
 		for( i = 0; i < sb->defnum; i += 2 ){
			if( Mode == CIRCLE ){
				do_circle( &sb->def[i], &sb->def[i+1] );
			} else {
				do_ellipse( &sb->def[i], &sb->def[i+1] );
			}
		}
	} else {
		do_arc( &sb->def[0], &sb->def[1] );
	}
	file->buf->mask_kind = SIMPLEMASK;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
		if( Mode == ARC ){
			TagNum = 1;
		}
		if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
		return( 0 );
	} else {
		if( Mode == ARC ){
			ItemNum = 3;
		} else {
			ItemNum = 7;
		}
		return( line_config( file ) );
	}
}
