/**********************************************************************/
/*                                                                    */
/*             color.c  gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "plug.h"


int Allow= 10;
char *SignChr[3] = { "-", "", "+" };
char AllowB[65] = { "10" };	/* "+10"==+10-0, "10"==+5-5, "-10"==-10+0 */
int AKind = 1;			/*     2,             1,           0      */

#define ALL  0
#define AREA 1
int Mode = ALL;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "color" };
char *Comment = { "color selecter" };

PLUGITEM Item[] = {
{ "allow", 	GETSTR,  &AllowB,  8,  0, "allowance of color difference" },
{ "area",	SWITCH,  &Mode, AREA, 'R',   "sequencial area mode"  },
{ "all",	SWITCH,  &Mode, ALL, 'L',    "all-over mode" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "allowance of the color", "allow=%s" },
{ RADIO, 1, "Mode " },
{ ITEM, 2, "all-over select(L)", "all" },
{ ITEM, 1, "area in the point(R)", "area" },
{ RADIOEND, -1 }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* file->buf-> */
unsigned int *Data;		/* data top file->buf->data*/

int Red, Green, Blue;		/* Selected Point Color */

int check_colorrange( unsigned int p );
int (*check_color)(unsigned int) = check_colorrange;

#define data(p,q)  Data[(q)*Width+(p)]
#define abs(p)  (p>0)?(p):((p)*(-1))

/**********************************************************************/
/*                                                                    */
/*             function check_colorrange()                            */
/*                                                                    */
/*             unsigned int p : Color value                           */
/*                                                                    */
/*             return  int : 0==not in range, 1==in range             */
/*                                                                    */
/*             function: Check the point if is within Allowance from  */
/*                      the color of Selected, case AKind==1.         */
/**********************************************************************/
int check_colorrange( unsigned int p )
{
	int tmp;
	int r, g, b;

	tmp = GET_RED(p);
	r = tmp - Red;
	r = abs(r);
	tmp = GET_GREEN(p);
	g = tmp - Green;
	g = abs(g);
	tmp = GET_BLUE(p);
	b = tmp - Blue;
	b = abs(b);
	if( r + g + b < Allow ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function check_lower()                                 */
/*                                                                    */
/*             unsigned int p : Color value                           */
/*                                                                    */
/*             return  int : 0==not in range, 1==in range             */
/*                                                                    */
/*             function: Check the point if is within Allowance from  */
/*                      the color of Selected, case AKind==0.         */
/**********************************************************************/
int check_lower( unsigned int p )
{
int tmp;
int r, g, b;
	tmp = GET_RED(p);
	r = Red - tmp;
	if( r < 0 ){
		return( 0 );
	}
	tmp = GET_GREEN(p);
	g = Green - tmp;
	if( g < 0 ){
		return( 0 );
	}
	tmp = GET_BLUE(p);
	b = Blue - tmp;
	if( b < 0 ){
		return( 0 );
	}
	if( r + g + b < Allow ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function check_upper()                                 */
/*                                                                    */
/*             unsigned int p : Color value                           */
/*                                                                    */
/*             return  int : 0==not in range, 1==in range             */
/*                                                                    */
/*             function: Check the point if is within Allowance from  */
/*                      the color of Selected, case AKind==2.         */
/**********************************************************************/
int check_upper( unsigned int p )
{
int tmp;
int r, g, b;
	tmp = GET_RED(p);
	r = tmp -  Red;
	if( r < 0 ){
		return( 0 );
	}
	tmp = GET_GREEN(p);
	g = tmp - Green;
	if( g < 0 ){
		return( 0 );
	}
	tmp = GET_BLUE(p);
	b = tmp - Blue;
	if( b < 0 ){
		return( 0 );
	}
	if( r + g + b < Allow ){
		return( 1 );
	} else {
		return( 0 );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function set_area()                                    */
/*                                                                    */
/*             int x : position X                                     */
/*             int y : position Y                                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: If AREA case, search the sequential area     */
/*                      from Selected point by recursive call.        */
/*                      Not become so deep recursive, expanded the call.*/
/**********************************************************************/
void set_area( int x, int y )
{
	volatile int i, left, right, up, down, naname;

	if( x >= Width || x < 0 ){
		return;
	}
	if( y >= Height || y < 0 ){
		return;
	}
	if( (*check_color)( data(x, y) ) == 0 ){
		return;
	}
	CLRMASK( data(x, y) );
	for( i = x + 1; i < Width; i++ ){
		if( (*check_color)( data(i, y) ) ){
			if( MASKON( data(i, y) ) ){
				CLRMASK( data(i, y) );
			} else {
				break;
			}
		} else {
			break;
		}
	}
	right = i;
	for( i = x - 1; i >= 0; i-- ){
		if( (*check_color)( data(i, y) ) ){
			if( MASKON( data(i, y) ) ){
				CLRMASK( data(i, y) );
			} else {
				break;
			}
		} else {
			break;
		}
	}
	left = i;
	for( i = y + 1; i < Height; i++ ){
		if( (*check_color)( data(x, i) ) ){
			if( MASKON( data(x, i) ) ){
				CLRMASK( data(x, i) );
			} else {
				break;
			}
		} else {
			break;
		}
	}
	down = i;
	for( i = y - 1; i >= 0; i-- ){
		if( (*check_color)( data(x, i) ) ){
			if( MASKON( data(x, i) ) ){
				CLRMASK( data(x, i) );
			} else {
				break;
			}
		} else {
			break;
		}
	}
	up = i;

	naname = min( Width - x - 1, Height - y - 1);
	for( i = 1; i <= naname; i++ ){
		if( (*check_color)( data(x + i, y + i) ) ){
			if( MASKON( data(x + i, y + i) ) ){
				CLRMASK( data(x+i, y+i) );
			} else {
				break;
			}
		} else {
			break;
		}
	}
	naname = i;
	for( i = 1; i < naname; i++ ){
		if( (*check_color)( data(x + i, y + i) ) ){
			set_area(x + i, y + i );
		}
	}
	naname = min( Width - x - 1, y );
	for( i = 1; i <= naname; i++ ){
		if( (*check_color)( data(x + i, y - i) ) ){
			if( MASKON( data(x + i, y - i) ) ){
				CLRMASK( data(x+i, y-i) );
			} else {
				break;
			}
		} else {
			break;
		}
	}
	naname = i;
	for( i = 1; i < naname; i++ ){
		if( (*check_color)( data(x + i, y - i) ) ){
			set_area(x + i, y - i );
		}
	}
	naname = min( x, Height - y - 1 );
	for( i = 1; i <= naname; i++ ){
		if( (*check_color)( data(x - i, y + i) ) ){
			if( MASKON( data(x - i, y + i) ) ){
				CLRMASK( data(x-i, y+i) );
			} else {
				break;
			}
		} else {
			break;
		}
	}
	naname = i;
	for( i = 1; i < naname; i++ ){
		if( (*check_color)( data(x - i, y + i) ) ){
			set_area(x - i, y + i );
		}
	}
	naname = min( x, y );
	for( i = 1; i <= naname; i++ ){
		if( (*check_color)( data(x - i, y - i) ) ){
			if( MASKON( data(x - i, y - i) ) ){
				CLRMASK( data(x-i, y-i) );
			} else {
				break;
			}
		} else {
			break;
		}
	}
	naname = i;
	for( i = 1; i < naname; i++ ){
		if( (*check_color)( data(x - i, y - i) ) ){
			set_area(x - i, y - i );
		}
	}

	for( i = x + 1; i < right; i++ ){
		if( (*check_color)( data(i + 1, y + 1) ) ){
			if( MASKON( data(i + 1, y + 1) ) ){
				set_area( i+1, y+1 );
			}
		} 
		if( (*check_color)( data(i + 1, y - 1) ) ){
			if( MASKON( data(i + 1, y - 1) ) ){
				set_area( i+1, y-1 );
			}
		}
	}
	for( i = x - 1; i > left; i-- ){
		if( (*check_color)( data(i - 1, y + 1) ) ){
			if( MASKON( data(i - 1, y + 1) ) ){
				set_area( i-1, y+1 );
			}
		}
		if( (*check_color)( data(i - 1, y - 1) ) ){
			if( MASKON( data(i - 1, y - 1) ) ){
				set_area( i-1, y-1 );
			}
		}
	}
	for( i = y + 1; i < down; i++ ){
		if( (*check_color)( data(x + 1, i + 1) ) ){
			if( MASKON( data(x + 1, i + 1) ) ){
				set_area( x + 1, i+1 );
			}
		}
		if( (*check_color)( data(x - 1, i + 1) ) ){
			if( MASKON( data(x - 1, i + 1) ) ){
				set_area( x-1, i+1 );
			}
		}
	}
	for( i = y - 1; i > up; i-- ){
		if( (*check_color)( data(x + 1, i - 1) ) ){
			if( MASKON( data(x + 1, i - 1) ) ){
				set_area( x+1, i-1 );
			}
		}
		if( (*check_color)( data(x - 1, i - 1) ) ){
			if( MASKON( data(x - 1, i - 1) ) ){
				set_area( x-1, i-1 );
			}
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *data, *p;
	int i;
	int pos;

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	data = file->buf->data;

	if( sb->selnum != SKPOINT ){
		return( EILLSELECTION );
	}
	check_range( sb, Width, Height, 1, NEGLECT );
	if( sb->defnum < 1 ){
		return( EILLNUMBER );
	}
	pos = sb->def[0].y * Width + sb->def[0].x;
	Red = (int)GET_RED(data[pos]);
	Green = (int)GET_GREEN(data[pos]);
	Blue = (int)GET_BLUE(data[pos]);

	if( Mode == AREA ){
		Data = data;
		for( i = 0, p = data; i < Height * Width; i++, p++ ){
			SETMASK(*p);
		}
		set_area( sb->def[0].x, sb->def[0].y );
	} else {
		for( i = 0, p = data; i < Height * Width; i++, p++ ){
			if( !(*check_color)( *p ) ){
				SETMASK(*p);
			} else {
				CLRMASK(*p);
			}
		}
	}
	file->buf->mask_kind = SIMPLEMASK;
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function allow()                                       */
/*                                                                    */
/*             char *str : set allow option                           */
/*                                                                    */
/*             return  int : 0==success -1=fail                       */
/*                                                                    */
/*             function: Exchange string to num of the allow option.  */
/*                                                                    */
/**********************************************************************/
int allow( char *str )
{
	int al;

	if( *str == '+' ){
		check_color = check_upper;
		str++;
		AKind = 2;
	} else if( *str == '-' ){
		check_color = check_lower;
		str++;
		AKind = 0;
	} else {
		check_color = check_colorrange;
		AKind = 1;
	}
	al = atoi( str );
	if( al > 0 ){
		Allow = al;
		return( 0 );
	} else {
		return( -1 );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	int ret;
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			ret = config_use_gui( file );
		} else {
			ret = config_comline( file );
		}
	} else {
		ret = line_config( file );
	}
	if( ret == 0 ){
		if( allow( AllowB ) < 0 ){
			return( -1 );
		}
	}
	return( ret );
}

