/**********************************************************************/
/*                                                                    */
/*            matrix.c   gedo PLUG specified routine                  */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "plug.h"
#include "matrix.h"

char MatStr[129] = { "0.1:0.1:0.1:0.1:0.2:0.1:0.1:0.1:0.1" };
/* Matrix value of each in string */
float Matrix[3][3];	/* Matrix value of each in float value */
char SaveName[256];	/* Matrix data file to save */
char LoadName[256];	/* Matrix data file to load */
int  List = 0;		/* list command to show available matrix data */
int Gui = NOGUI;

void set_var( char *work );

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };
char *Spec =    { "colorop" };
char *Comment = { "color operator effecter, tranfar by the filter" };
PLUGITEM Item[] = {
{ "matrix", GETSTR, MatStr, 128, 0,  "transfar matrix 3x3, like 1:1:1:1:1.5:1:1:1:1" },
{ "save",   GETSTR, SaveName, 255, 0,"name the current matrix" }, 
{ "load",   GETSTR, LoadName, 255, 0,"load the named matrix" },
{ "list",   SWITCH, &List,  1, 0,     "search the named matrix" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ NOTCL | ENTRY, 0, "matrix = ", "matrix=%s" },
};

int TagNum = sizeof(Tag) / sizeof(struct Tags);
DEFAULTSTATUS;
/**********************************************************************/


/**********************************************************************/
/*                                                                    */
/*             function filter_process()                              */
/*                                                                    */
/*             struct Shm *shm: data struct                           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: For each pixel, do the filter process.       */
/*                                                                    */
/**********************************************************************/
void filter_process( struct Shm *shm )
{
	unsigned int *p;
	int i;
	int len;
	unsigned char rc, gc, bc;
	float r, g, b;
	float rx, gx, bx;

	len = shm->width * shm->height;
	p = shm->data;

	for( i = 0; i < len; i++, p++ ){
		if( MASKOFF( *p ) ){
			r = (float)GET_RED( *p );
			g = (float)GET_GREEN( *p );
			b = (float)GET_BLUE( *p );
			rx = Matrix[0][0] * r + Matrix[0][1] * g + Matrix[0][2] * b;
			gx = Matrix[1][0] * r + Matrix[1][1] * g + Matrix[1][2] * b;
			bx = Matrix[2][0] * r + Matrix[2][1] * g + Matrix[2][2] * b;
			if( rx < 0.0 ){
				rc = 0;
			} else if( rx >= 255.0 ){
				rc = 255;
			} else {
				rc = (unsigned char)rx;
			}
			if( gx < 0.0 ){
				gc = 0;
			} else if( gx >= 255.0 ){
				gc = 255;
			} else {
				gc = (unsigned char)gx;
			}
			if( bx < 0.0 ){
				bc = 0;
			} else if( bx >= 255.0 ){
				bc = 255;
			} else {
				bc = (unsigned char)bx;
			}
			*p = ((*p & MASK) | ((unsigned int)rc << 16) 
			      | ((unsigned int)gc << 8) | (unsigned int)bc);
		}
        }
}

/**********************************************************************/
/*                                                                    */
/*             function prepare_matrix()                              */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Prepare the matrix to use.                   */
/*                                                                    */
/**********************************************************************/
int prepare_matrix( void )
{	
	int i, j;
	char *p;

	for( i = j = 0, p = MatStr; j < 3 && *p; ){
		if( *p == ':' ){
			if( ++i >= 3 ){
				i = 0; j++;
			}
			p++;
		} else {
			Matrix[j][i] = atof( p );
			for( ; *p && *p != ':'; p++ );
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
        if( file->buf->mask_kind == DEPTHMASK ){
		return( EILLMASK );
	}
	if( prepare_matrix() < 0 ){
		strcpy( file->mes->u.data, "illegal matrix, cannot interpret" );
		return( -1 );
	}
	filter_process( file->buf );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config_gui()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG,     */
/*             if GUI mode only.                                      */
/**********************************************************************/
int config_gui( struct File *file )
{
	FILE *fp;
	FILE *fpp;
	char buff[256];
	char fifo[256];
	char *e;


	if( (fp = create_fifo( fifo )) == NULL ){
		return( -1 );
	}
	if( (e = getenv("GEDO_PATH")) == NULL ){
		e = "./";
	}
	sprintf( buff, "%smatgui <%s &", e, fifo );
	if( (fpp = popen( buff, "r" )) == NULL ){
		return( -1 );
	}
	fprintf( fp, "set name \"%s\"\n"
		"set m1	   %g\nset m2      %g\nset m3     %g\n"
		"set m4    %g\nset m5      %g\nset m6     %g\n"
		"set m7	   %g\nset m8      %g\nset m9     %g\n",
		Spec,
		Matrix[0][0], Matrix[0][1], Matrix[0][2],
		Matrix[1][0], Matrix[1][1], Matrix[1][2],
		Matrix[2][0], Matrix[2][1], Matrix[2][2] );
	fclose( fp );

	fgets( buff, sizeof(buff)-1, fpp );
	pclose( fpp );

	if( strcmp( buff, "abort\n" ) == 0 ){
		return( EUSRBREAK );
	}
	buff[strlen(buff) - 1] = '\0';
	strcpy( file->mes->u.data, buff );
	return( line_config( file ) );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	int ret;
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			prepare_matrix();
			return( config_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		LoadName[0] = SaveName[0] = '\0'; List = 0;
		ret = line_config( file );
		if( ret != 0 ){
			return( ret );
		}
		if( List ){
			List = 0;
			search_matrix();
		}
		if( LoadName[0] ){
			ret = load_matrix( LoadName, Matrix, MatStr );
			if( ret != 0 ){
				return( ret );
			}
		}
		if( SaveName[0] ){
			prepare_matrix();
			ret = save_matrix( SaveName, Matrix );
		}
		return( ret );
	}
}

