/**********************************************************************/
/*                                                                    */
/*          command.c   gedo PLUG additional library                  */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <limits.h>
#include "plug.h"

extern int Init_flag;
static char *Kind_str[5] = { "<int>", "<hex>", "<float>", "<string>", "" };
char *StrServer[]
	 = { "main", "loader", "selecter", "effecter", "viewer" };

struct Radio {			/* RADIO button struct */
	char mess[80];
	int num;
	void *target;
	PLUGITEM *pi[16];
	struct Tags *tag[16];
};

/**********************************************************************/
/*                                                                    */
/*             function axtol()                                       */
/*                                                                    */
/*             char *str : string of HEX value                        */
/*             int max   : interpret to long value until max          */
/*                                                                    */
/*             return  long : HEX value                               */
/*                                                                    */
/*             function: Convert HEX string to long value.            */
/*                                                                    */
/**********************************************************************/
long axtol( char *str, int max )
{
	long ret;
	int i;

	ret = 0L;
	for( i = 0; i < max; i++ ){
		if( isdigit( str[i] ) ){
			ret = ret * 16 + str[i] - '0';
		} else if( isxdigit( str[i] ) ){
			ret = ret * 16 + toupper( str[i] ) - 'A' + 10;
		} else if( str[i] == ' ' ){
			continue;
		} else {
			return( -1L );
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function ltoax()                                       */
/*                                                                    */
/*             long val : <in> long value                             */
/*             char *str: <out>string of HEX to set                   */
/*             int max  : <in> maximum of the length                  */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Convert long value to the HEX string.        */
/*                                                                    */
/**********************************************************************/
void ltoax( long val, char *str, int max )
{
	int i;
	unsigned char work[8];
	unsigned remain;

	for( i = 0; i < max; i++ ){
		remain = val % 16;
		val = val / 16;
		if( remain <= 9 ){
			work[i] = (unsigned char)remain + '0';	
		} else {
			work[i] = (unsigned char)remain - 10 + 'a';
		}
		if( val == 0 ){
			break;
		}
	}
	for( max = max - i - 1; max > 0; max-- ){
		*str++ = ' ';
	}
	for( ; i >= 0; i-- ){
		*str++ = work[i];
	}
}

/**********************************************************************/
/*                                                                    */
/*             function saxtol()                                      */
/*                                                                    */
/*             char *str : string of HEX value                        */
/*                                                                    */
/*             return  long : HEX value                               */
/*                                                                    */
/*             function: Convert HEX string to long value.            */
/*                                                                    */
/**********************************************************************/
long saxtol( char *str )
{
	long ret;
	int i;

	ret = 0L;
	for( i = 0; i < 8; i++ ){
		if( isdigit( str[i] ) ){
			ret = ret * 16 + str[i] - '0';
		} else if( isxdigit( str[i] ) ){
			ret = ret * 16 + toupper( str[i] ) - 'A' + 10;
		} else {
			break;
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function command_line()                                */
/*                                                                    */
/*             char *str : string from gedo config                    */
/*                                                                    */
/*             return  char ** : splited to double pointer            */
/*                                                                    */
/*             function: Handle gedo-config string like argv.         */
/*                                                                    */
/**********************************************************************/
char ** command_line( char *str )
{
	char * p;
	static char combuff[1024];
	static char *ret[20];
	int status;
	int count;
	int i, j, len;

	status = 0; count = 0;
	strcpy( combuff, str );
	for( p = combuff; *p; p++ ){
		if( isspace( *p ) ){
			if( status == 0 ){
				*p = '\0';
			}
		} else if( *p == ',' ){
			status = 0;
			*p = '\0';
			count++;
		} else {
			if( status == 0 ){
				status = 1;
				ret[count] = p;
			}
		}
	}
	ret[++count] = NULL;
	for( i = 0; i < count; i++ ){
		len = strlen( ret[i] );
		for( j = len - 1; j >= 0; j-- ){
			if( isspace( ret[i][j] ) ){
				ret[i][j] = '\0';
			} else {
				break;
			}
		}
	}
	return( ret );
}
		
/**********************************************************************/
/*                                                                    */
/*             function check_range()                                 */
/*                                                                    */
/*             struct Sbuf *sb : sended points from gedo              */
/*             int w           : width of the image                   */
/*             int h           : height of the image                  */
/*             int num         : check number                         */
/*             int kind        : ==NEGLECT delete over-range points   */
/*                               ==ADJUST  make over-range points     */
/*                                              boundary value.       */
/*             return NONE                                            */
/*                                                                    */
/*             function: Check the send points from gedo.             */
/*                                                                    */
/**********************************************************************/
void check_range( struct Sbuf *sb, int w, int h, int num, int kind )
{
	int flag, i;
	struct Sbuf n;

	n.defnum = 0;
	if( kind == NEGLECT ){
		for( i = 0; i < ((num==0)?(sb->defnum):(num)); i++ ){
			flag = 0;
			if( sb->def[i].x >= w ){
				flag = 1;
			} else if( sb->def[i].x < 0 ){
				flag = 1;
			}
			if( sb->def[i].y >= h ){
				flag = 1;
			} else if( sb->def[i].y < 0 ){
				flag = 1;
			}
			if( !flag ){
				n.def[n.defnum].x = sb->def[i].x;
				n.def[n.defnum].y = sb->def[i].y;
				n.defnum++;
			}
		}
		sb->defnum = n.defnum;
		for( i = 0; i < n.defnum; i++ ){
			sb->def[i].x = n.def[i].x;
			sb->def[i].y = n.def[i].y;
		}
	} else if( kind == ADJUST ){
		for( i = 0; i < ((num==0)?(sb->defnum):(num)); i++ ){
			if( sb->def[i].x >= w ){
				sb->def[i].x = w - 1;
			} else if( sb->def[i].x < 0 ){
				sb->def[i].x = 0;
			}
			if( sb->def[i].y >= h ){
				sb->def[i].y = h - 1;
			} else if( sb->def[i].y < 0 ){
				sb->def[i].y = 0;
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function check_item()                                  */
/*                                                                    */
/*             unsigned char *in : each config string                 */
/*             PLUGITEM *pi      : each item of the Plug Item         */
/*                                                                    */
/*             return int  : 0==diffrent 1==handled -1==error         */
/*                                                                    */
/*             function: Process the config string by Plug Item.      */
/*                                                                    */
/**********************************************************************/
int check_item( unsigned char *in, PLUGITEM *pi )
{
	int len;
	unsigned int type;
	long int val;
	float fval;

	if( Init_flag ){
		if( pi->type & NOINIT ){
			return( 0 );
		}
	} else {
		if( pi->type & INITONLY ){
			return( 0 );
		}
	}
	type = (pi->type & COMMANDMASK);
	if( pi->name != NULL ){
		len = strlen(pi->name);
		if( strncmp( in, pi->name, len ) != 0 ){
			return( 0 );
		}
		if( in[len] ){
			if( in[len] != '=' ){
				fprintf( stderr, "illegal =\n" );
				return( -1 );
			}
			len++;
		}
	} else {
		len = 0;
	}

	switch( type ){
	case GETINT:
	case GETHEX:
		if( type == GETINT ){
			val = atoi( &in[len] );
		} else {
			val = saxtol( &in[len] );
		}
		if( pi->omni1 == 0 && pi->omni2 == 0 ){
			*(int *)pi->set_target = val;
		} else if( val >= pi->omni1 && val <= pi->omni2 ){
			*(int *)pi->set_target = val;
		} else {
			fprintf( stderr, "illegal value\n" );
			return( -1 );
		}			
		break;
	case GETFLOAT:
		fval = atof( &in[len] );
		*(float *)pi->set_target = fval;
		break;
	case GETSTR:
		if( strlen( &in[len] ) <= pi->omni1 ){
			strcpy( pi->set_target, &in[len] );
		} else {
			fprintf( stderr, "set string too long\n" );
			return( -1 );
		}
		break;
	case SWITCH:
		*(int *)pi->set_target = pi->omni1;
		break;
	default:
		fprintf( stderr, "unrecognize command\n" );
		return( -1 );
		break;
	}
	return( 1 );
}

/**********************************************************************/
/*                                                                    */
/*             function display_config()                              */
/*                                                                    */
/*             PLUGITEM *pi      : top item of the Plug Item          */
/*             int num           : number of Plug item                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Display the config parameters from help()    */
/*                      and init().                                   */
/**********************************************************************/
void display_config( int kind, PLUGITEM *pi, int num )
{
	int i;
	int type;
	unsigned char buff[64];

	if( num ){
		for( i = 0; i < num; i++ ){
			type = pi[i].type;
			if( kind == CONFIG_INIT ){ /* called from usage() */
				if( type & NOINIT ){
					continue;
				}
			} else {
				if( type & INITONLY ){
					continue;
				}
				if( (kind == CONFIG_LOADER) 
				   && (type & NOLOAD) ){
					continue;
				} else if( (kind == CONFIG_SAVER) 
					  && (type & NOSAVE) ){
					continue;
				}
			}

			type &= COMMANDMASK; 
			switch( type ){
			case GETINT:
			case GETHEX:
			case GETFLOAT:
			case GETSTR:
				if( pi[i].name ){
					sprintf( buff, "%s=%s",
						 pi[i].name, 
						Kind_str[type] );
					printf( "\t%-20s:%s\n", buff, 
						pi[i].comment );
				} else {
					printf( "\t%-20s:%s\n", Kind_str[type], 
						pi[i].comment );
				}
				break;
			case SWITCH:
				printf( "\t%-20s:%s\n", pi[i].name, pi[i].comment );
				break;
			}
		}
	} else {
		printf( "\tno configration parameters\n" );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function check_itemnum()                                */
/*                                                                    */
/*            args  int kind  : configure kind                        */
/*                  PLUGITEM *pi: PLUG each configure item           */
/*                  int num   :  number of PLUGITEM                   */
/*                                                                    */
/*            return int  : number of alive Item                      */
/*                                                                    */
/*            function: Return the PLUGITEM alive number.             */
/*                                                                    */
/**********************************************************************/
int check_itemnum( int kind,  PLUGITEM *pi, int num )
{
	int i, ret;
	unsigned int mask;

	if( kind == CONFIG_ELSE ){
		mask = INITONLY;
	} else if( kind == CONFIG_INIT ){
		mask = NOINIT;
	} else if( kind == CONFIG_LOADER ){
		mask = NOLOAD;
	} else if( kind == CONFIG_SAVER ){
		mask = NOSAVE;
	} else {
		mask = 0xffffffff;
	}
	for( ret = i = 0; i < num; i++ ){
		if( (pi[i].type & mask) == 0 ){
			ret++;
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function get_commandspec()                              */
/*                                                                    */
/*            args  struct File *file : Transaction with GEDO         */
/*                                                                    */
/*            return  int : command index                             */
/*                                                                    */
/*            function: From transaction abstruct configure kind.     */
/*                                                                    */
/**********************************************************************/
int get_commandspec( struct File *file )
{
	switch( file->mes->command ){
	case CLOAD:
		return( 2 );
		break;
	case CSAVE:
		return( 3 );
		break;
	case CSELECT:
	case CEFFECT:
		return( 0 );
		break;
	default:
		return( 1 );
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function proc_radio()                                   */
/*                                                                    */
/*            args  struct Tags *tg  : Tag of configure               */
/*                  struct Radio *rad: Made Radio struct              */
/*                                                                    */
/*            return  int : SKIP command                              */
/*                                                                    */
/*            function: if RADIO Item, make up Radio input.           */
/*                                                                    */
/**********************************************************************/
static int proc_radio( struct Tags *tg, struct Radio *rad )
{
	memset( rad, 0, sizeof(struct Radio) );
	strcpy( rad->mess, tg->prompt );
	strcat( rad->mess, ": " );
	return( SKIP );
}

/**********************************************************************/
/*                                                                    */
/*            function proc_radioitem()                               */
/*                                                                    */
/*            args  struct Tags *tg  : tag of configure               */
/*                  PLUGITEM *pi     : Configure Item                 */
/*                  struct Radio *rad: Made radio struct              */
/*                                                                    */
/*            return  int  : ERROR==error, success==SKIP command      */
/*                                                                    */
/*            function: Radio Item input prepare.                     */
/*                                                                    */
/**********************************************************************/
static int proc_radioitem( struct Tags *tg, PLUGITEM *pi, struct Radio *rad )
{
	if( (pi->type & COMMANDMASK) != SWITCH ){
		return( ERROR );
	}
	if( rad->target == NULL ){
		rad->target = pi->set_target;
	} else {
		if( rad->target != pi->set_target ){
			return( ERROR );
		}
	}
	strcat( rad->mess, tg->prompt );
	strcat( rad->mess, " " );
	rad->pi[rad->num] = pi;
	rad->tag[rad->num] = tg;
	rad->num++;
	return( SKIP );
}

/**********************************************************************/
/*                                                                    */
/*            function exec_radio()                                   */
/*                                                                    */
/*            args  struct Radio *rad : Radio struct Made.            */
/*                                                                    */
/*            return  int : selected==SKIP command else error         */
/*                                                                    */
/*            function: Get a input at the Radio button.              */
/*                                                                    */
/**********************************************************************/
static int exec_radio( struct Radio *rad )
{
	int c;
	int i;

	printf( "\t%s = ", rad->mess );
	c = getchar();
	if( c == EOF ){
		return( USRBREAK );
	} else if( c == '\n' ){
		return( SKIP );
	}
	while( getchar() != '\n' );
	c = toupper(c);
	for( i = 0; i < rad->num; i++ ){
		if( c == toupper( rad->pi[i]->omni2 ) ){
			*(int *)rad->pi[i]->set_target = rad->pi[i]->omni1;
			rad->tag[i]->status = DONE;
			return( SKIP );
		}
	}
	return( ERROR );
}


/**********************************************************************/
/*                                                                    */
/*            function proc_entry()                                   */
/*                                                                    */
/*            args  PLUGITEM *pi : Configure PLUG item.               */
/*                                                                    */
/*            return  int : DONE==selected SKIP==skipped else ERROR   */
/*                                                                    */
/*            function: Input the ENTRY command.                      */
/*                                                                    */
/**********************************************************************/
static int proc_entry( PLUGITEM *pi )
{
	char buff[129];
	char *p, *q;
	int val;
	float fval;

	fflush( stdin );
	if( fgets( buff, 128, stdin ) == NULL ){
		return( USRBREAK );
	}
	for( p = buff; isspace(*p); p++ );
	if( *p == '\0' ){
		return( SKIP );
	}

	switch( pi->type & COMMANDMASK ){
	case GETINT:
	case GETHEX:
		if( (pi->type & COMMANDMASK) == GETINT ){
			val = atoi( p );
		} else {
			val = saxtol( p );
		}
		if( pi->omni1 == 0 && pi->omni2 == 0 ){
			*(int *)pi->set_target = val;
		} else if( val >= pi->omni1 && val <= pi->omni2 ){
			*(int *)pi->set_target = val;
		} else {
			fprintf( stderr, "illegal value\n" );
			return( ERROR );
		}
		break;
	case GETFLOAT:
		fval = atof( p );
		*(float *)pi->set_target = fval;
		break;
	case GETSTR:
		for( q = &p[strlen(p)-1]; isspace(*q); q-- ){
			*q = '\0';
		}
		if( strlen( p ) <= pi->omni1 ){
			strcpy( pi->set_target, p );
		} else {
			fprintf( stderr, "set string too long\n" );
			return( ERROR );
		}
		break;
	default:
		return( ERROR );
		break;
	}
	return( DONE );
}

/**********************************************************************/
/*                                                                    */
/*            function new_config_input()                             */
/*                                                                    */
/*            args  unsigned int mask : Inhibit Mask                  */
/*                  struct Tags *tag  : Configure Tag                 */
/*                  int tagnum        : number of tag                 */
/*                  PLUGITEM *item    : Configure Item                */
/*                  int *num          : number of alive tag           */
/*                                                                    */
/*            return int   : 0==success  else==error                  */
/*                                                                    */
/*            function: Configure input routine.                      */
/*                                                                    */
/**********************************************************************/
int new_config_input( unsigned int mask, struct Tags *tag, int tagnum, PLUGITEM *item, int *num )
{
	int i;
	PLUGITEM *pi;
	int ret;
	int mode;
	struct Radio radio;

	mode = 0; *num = 0;
	for( i = 0; i < tagnum; i++ ){
		tag[i].status = SKIP;
		if( tag[i].kind & mask ){
			continue;
		}
		switch( tag[i].kind & COMMANDMASK ){
		case ENTRY:
			printf( "\t%s = ", tag[i].prompt );
			pi = &item[tag[i].num];
			ret = proc_entry( pi );
			break;
		case RADIO:
			ret = proc_radio( &tag[i], &radio );
			mode = RADIO;
			break;
		case RADIOEND:
			ret = exec_radio( &radio );
			mode = 0;
			break;
		case BLOCK:
			if( tag[i].prompt ){
				printf( "\t%s\n", tag[i].prompt );
			}
			mode = BLOCK;
			ret = SKIP;
			break;
		case BLOCKEND:
			if( tag[i].prompt ){
				printf( "\t%s\n", tag[i].prompt );
			}
			mode = 0;
			ret = SKIP;
			break;
		case ITEM:
			pi = &item[tag[i].num];
			if( mode == RADIO ){
				ret = proc_radioitem( &tag[i], pi, &radio );
			} else {
				ret = ERROR;
			}
			break;
		default:
			ret = SKIP;
			break;
		}
		tag[i].status = ret;
		switch( ret ){
		case USRBREAK:
			return( USRBREAK );
			break;
		case ERROR:
			return( ERROR );
			break;
		}
		(*num)++;
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function output_tclcommand()                            */
/*                                                                    */
/*            args  unsigned int mask : Inhibit Mask                  */
/*                  FILE *fp          : Pipe to wish                  */
/*                  struct Tags *tag  : Configure Tag                 */
/*                  int tagnum        : numberr of Tag                */
/*                  PLUGITEM *item    : Configure Item                */
/*                  PLUGINFO *pif     : Configure Info                */
/*                                                                    */
/*            return NONE   :                                         */
/*                                                                    */
/*            function: Create the Tcl/Tk Command to send wish.       */
/*                                                                    */
/**********************************************************************/
void output_tclcommand( unsigned int mask, FILE *fp, struct Tags *tag, int tagnum, PLUGITEM *item, PLUGINFO *pif )
{
	char buff[512];
	char work[256];
	int frame, entry, radio, label, block, list;
	int width, mode, st;
	int blockframe;
	int var;
	int i, j, ret;
	PLUGITEM *pi;
	char *id;
	int oe_flag;

	frame = entry = radio = label = block = blockframe = var = list = 0;

	fprintf( fp, "frame .frm%d\n", frame );
	fprintf( fp, "label .lab%d -text \"%s %s configure\" -fg blue\n", label, StrServer[pif->kind], pif->plugname );
	fprintf( fp, "pack .lab%d -in .frm%d -side left\n", label++, frame++ );

	mode = 0; st = -1; id = NULL;
	oe_flag = 0;
	memset( buff, 0 , sizeof(buff) );
	for( i = 0; i < tagnum; i++ ){
		if( tag[i].kind & mask ){
			continue;
		}
		switch( tag[i].kind & COMMANDMASK ){
		case ENTRY:
			pi = &item[tag[i].num];
			switch( pi->type & COMMANDMASK ){
			case GETSTR:
				width = pi->omni1;
				break;
			case GETINT:
				width = 6;
				break;
			case GETHEX:
			case GETFLOAT:
				width = 8;
				break;
			default:
				width = 8;
				break;
			}
			if( mode == 0 ){
				fprintf( fp, "frame .frm%d\n", frame );
			} else if( mode == BLOCK ){
				fprintf( fp, "frame .blk%d\n", block );
			}
			fprintf( fp, "label .lab%d -width %d -text \"%s\"\n",
				label, 40, tag[i].prompt );
			switch( pi->type & COMMANDMASK ){
			case GETSTR:
				if( *(char *)pi->set_target ){
					fprintf( fp, "set v%d \"%s\"\n", var, (char *)pi->set_target );
				}
				break;
			case GETINT:
				fprintf( fp, "set v%d %d\n", var, *(int *)pi->set_target );
				break;
			case GETHEX:
				fprintf( fp, "set v%d %x\n", var, *(int *)pi->set_target );
			        break;
			case GETFLOAT:
				fprintf( fp, "set v%d %g\n", var, *(float *)pi->set_target );
				break;
			}
			fprintf( fp, "entry .ent%d -width %d -relief sunken -textvariable v%d\n", entry, width, var );
			sprintf( work, ",%s=$v%d", pi->name, var++ );
			strcat( buff, work );
			if( mode == 0 ){
				fprintf( fp, "pack .lab%d .ent%d -in .frm%d -side left\n", label++, entry++, frame++ );
			} else if( mode == BLOCK ){
				fprintf( fp, "pack .lab%d .ent%d -in .blk%d -side left\n", label++, entry++, block++ );
			}
			break;
		case RADIO:
			fprintf( fp, "frame .frm%d\n", frame );
			mode = RADIO;
			fprintf( fp, "label .lab%d -text \"%s\"\n", label, tag[i].prompt );
			sprintf( work, ",$v%d", var );
			strcat( buff, work );
			id = tag[i].result;
			st = radio;
			break;
		case RADIOEND:
			fprintf( fp, "pack " );
			fprintf( fp, ".lab%d ", label++ );
			for( j = st; j < radio; j++ ){
				fprintf( fp, ".rad%d ", j );
			}
			fprintf( fp, "-in .frm%d -side left -padx 5m\n", 
				frame++ );
			mode = 0; var++;
			break;
		case BLOCK:
			mode = BLOCK;
			st = block;
			fprintf( fp, "frame .frm%d\n", frame );
			blockframe = frame++;
			if( tag[i].prompt ){
				fprintf( fp, "frame .blk%d\n", block );
				fprintf( fp, "label .lab%d -text \"%s\" -fg blue\n", label, tag[i].prompt );
				fprintf( fp, "pack .lab%d -in .blk%d -side left\n", label++, block++ );
			}
			ret = SKIP;
			break;
		case BLOCKEND:
			if( mode != BLOCK ){
				ret = ERROR;
			}
			if( tag[i].prompt ){
				fprintf( fp, "frame .blk%d\n", block );
				fprintf( fp, "label .lab%d -text \"%s\" -fg blue\n", label, tag[i].prompt );
				fprintf( fp, "pack .lab%d -in .blk%d -side left\n", label++, block++ );
			}
			fprintf( fp, "pack " );
			for( j = st; j < block; j++ ){
				fprintf( fp, ".blk%d ", j );
			}
			fprintf( fp, "-in .frm%d -side top -pady 1m\n", 
				blockframe );
			mode = 0;
			ret = SKIP;
			break;
		case ITEM:
			pi = &item[tag[i].num];
			if( mode != RADIO ){
				ret = ERROR;
				break;
			}
			if( *(int *)pi->set_target == pi->omni1 ){
				fprintf( fp, "set v%d %s\n", var, tag[i].result );
			}
			fprintf( fp, "radiobutton .rad%d -text \"%s\" -variable v%d -value %s\n", radio++, tag[i].prompt, var, tag[i].result );
			break;
		case INSTCL:
			fprintf( fp, "frame .frm%d\n", frame );
			fprintf( fp, "%s\n", tag[i].prompt );
			if( tag[i].result ){
				fprintf( fp, "pack %s -in .frm%d -side left\n", tag[i].result, frame++ );
			}
			break;
		case INSFUNC:
			(*(void (*)(char *))tag[i].prompt)( work );
			fprintf( fp, "%s\n", work );
			if( tag[i].result ){
				fprintf( fp, "frame .frm%d\n", frame );
				fprintf( fp, "pack %s -in .frm%d -side left\n", tag[i].result, frame++ );
			}
			break;
		case TCLLIST:
			pi = &item[tag[i].num];
			fprintf( fp, "frame .frm%d\n", frame );
			fprintf( fp, "frame .frmlst%d\n", list );
			if( *(char *)pi->set_target ){
				fprintf( fp, "set v%d \"%s\"\n", var, (char *)pi->set_target );
			}
			fprintf( fp, "label .lab%d -text \"%s\"\n", label, tag[i].prompt );
			fprintf( fp, "listbox .lstb%d -relief raised -borderwidth 2 -yscrollcommand \".scroll%d set\"\n", list, list );
			fprintf( fp, "scrollbar .scroll%d -command \".lstb%d yview\"\n", list, list );
			fprintf( fp, "pack .lstb%d -in .frmlst%d -side left\n" , list, list );
			fprintf( fp, "pack .scroll%d -in .frmlst%d -side right -fill y\n", list, list );
			fprintf( fp, "bind .lstb%d <Double-Button-1> {\n", list );
			fprintf( fp, "set v%d [selection get] }\n", var );
			break;
		case TCLDIR:
			(*(void (*)(FILE *, int, int, char *))tag[i].prompt)( fp, list, tag[i].num, tag[i].result );
			break;
		case TCLLISTENT:
			pi = &item[tag[i].num];
			fprintf( fp, "frame .frment%d\n", list );
			fprintf( fp, "label .lab%d -text \"%s\"\n", label+1, tag[i].prompt );
			fprintf( fp, "entry .listent%d -relief sunken -textvariable v%d\n", list, var );
			fprintf( fp, "pack .lab%d .listent%d -in .frment%d -side left\n", label+1, list, list );
			fprintf( fp, "pack .lab%d .frmlst%d .frment%d -in .frm%d\n", label, list, list, frame );
			list++; label+=2; frame++;
			sprintf( work, ",%s=$v%d", pi->name, var++ );
			strcat( buff, work );
			break;
		case TCLLISTEND:
			pi = &item[tag[i].num];
			fprintf( fp, "frame .frment%d\n", list );
			fprintf( fp, "label .lab%d -text \"%s\"\n", label+1, tag[i].prompt );
			fprintf( fp, "label .listlab%d -textvariable v%d\n", list, var );
			fprintf( fp, "pack .lab%d .listlab%d -in .frment%d -side left\n", label+1, list, list );
			fprintf( fp, "pack .lab%d .frmlst%d .frment%d -in .frm%d\n", label, list, list, frame );
			list++; label+=2; frame++;
			sprintf( work, ",%s=$v%d", pi->name, var++ );
			strcat( buff, work );
			break;
		case OKandESC:
			fprintf( fp, "frame .frm%d\n", frame );
			fprintf( fp, "button .but1 -text OK -command { puts \"%s\"; exit }\n",
				tag[i].prompt );
			if( tag[i].result ){
				fprintf( fp, "button .but2 -text ESC -command { puts \"%s\"; exit}\n", tag[i].result );
			} else {
				fprintf( fp, "button .but2 -text ESC -command { puts \"abort\"; exit}\n" );
			}
			fprintf( fp, "pack .but1 .but2 -in .frm%d -side left -padx 5m\n",
			frame++ );
			oe_flag = 1;
		default:
			ret = ERROR;
			break;
		}
	}
	if( oe_flag == 0 ){
		fprintf( fp, "frame .frm%d\n", frame );
		fprintf( fp, "button .but1 -text OK -command { puts \"%s\"; exit }\n",
			&buff[1] );
		fprintf( fp, "button .but2 -text ESC -command { puts \"abort\"; exit}\n" );
		fprintf( fp, "pack .but1 .but2 -in .frm%d -side left -padx 5m\n",
			frame++ );
	}
	fprintf( fp, "pack " );
	for( i = 0; i < frame; i++ ){
		fprintf( fp, ".frm%d ", i );
	}
	fprintf( fp, "-side top -pady 2m\n" );
}

/**********************************************************************/
/*                                                                    */
/*            function create_fifo()                                  */
/*                                                                    */
/*            args  char *filename : Pipe file name                   */
/*                                                                    */
/*            return FILE * : FILE * of pipe                          */
/*                                                                    */
/*            function: Open or Create pipe to send wish(1).          */
/*                                                                    */
/**********************************************************************/
FILE *create_fifo( char *filename )
{
	char *envstr;
	FILE *fp;
	
	if( (envstr = getenv( "GEDO_LWP" )) == NULL ){
		strcpy( filename, "./" );
	} else {
		strcpy( filename, envstr );
	}
	if( filename[strlen(filename) -1] != '/' ){
		strcat( filename, "/" );
	}
	strcat( filename, "fifo" );
	if( (fp = fopen( filename, "w" )) == NULL ){
		return( NULL );
	}
	return( fp );
}

/**********************************************************************/
/*                                                                    */
/*            function each_default_status()                          */
/*                                                                    */
/*            args  int i  :  now Tag number                          */
/*                  struct Tags *tag: Configure Tag                   */
/*                  PLUGITEM *item  : Configure Item                  */
/*                  char *buff      : status must be set              */
/*                                                                    */
/*            return int   :  1==data set to buff  0==not set         */
/*                                                                    */
/*            function: For each Tag, make status().                  */
/*                                                                    */
/**********************************************************************/
int each_default_status( int i, struct Tags *tag, PLUGITEM *item, char *buff )
{
	PLUGITEM *pi;
 
	switch( tag[i].kind & COMMANDMASK ){
	case ENTRY:
		pi = &item[tag[i].num];
		switch( pi->type & COMMANDMASK ){
		case GETSTR:
			sprintf( buff, tag[i].result, (char *)pi->set_target );
			break;
		case GETINT:
		case GETHEX:
			sprintf( buff, tag[i].result, *(int *)pi->set_target );
			break;
		case GETFLOAT:
			sprintf( buff, tag[i].result, *(float *)pi->set_target );
			break;
		}
		break;
	case ITEM:
		pi = &item[tag[i].num];
		if( *(int *)pi->set_target == pi->omni1 ){
			sprintf( buff, "%s", tag[i].result );
			return( 1 );
		}
		return( 0 );
	default:
		return( 0 );
		break;
	}
	return( 1 );
}


/**********************************************************************/
/*                                                                    */
/*             function check_exp()                                   */
/*                                                                    */
/*             char *file  :  file name in the directory              */
/*             char *exp   :  must be found extention                 */
/*             char *out   :  if fouund, base name must copy area     */
/*                                                                    */
/*             return int  :  found the extention==1                  */
/*                            not found==0                            */
/*             function: Find the extension in the <file>. If found,  */
/*                      copy <out> the basename.                      */
/**********************************************************************/
static int check_exp( char *file, char *exp, char *out )
{
	int i, len;

	len = strlen( file );
	for( i = len - 1; i >= 0; i-- ){
		if( file[i] == '.' ){
			break;
		}
	}
	if( i < 0 ){
		return( 0 );
	}
	if( strcmp( &file[i+1], exp ) == 0 ){
		strncpy( out, file, i );
		out[i] = '\0';
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function search_dir()                                  */
/*                                                                    */
/*             char *dirname : search diretory name                   */
/*             char *exp   :  must be found extention                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Find the extension in the directory and      */
/*                      show the basenames.                           */
/**********************************************************************/
void search_dir( char *dirname, char *ext )
{
	struct dirent *ent;
	DIR *dir;
	char buff[NAME_MAX+1];
	int count;

	if( (dir = opendir( dirname )) == NULL ){
		fprintf( stderr, "cannot open %s\n", dirname );
		exit( 1 );
	}
	count = 0;
	while( (ent = readdir( dir )) ){
		if( check_exp( ent->d_name, ext, buff ) ){
			fprintf( stderr, "%s", buff );
			if( ++count >= 5 ){
				fprintf( stderr, "\n" );
				count = 0;
			} else {
				fprintf( stderr, "\t" );
			}
		}
	}
	fprintf( stderr, "\n" );
	closedir( dir );
}

/**********************************************************************/
/*                                                                    */
/*             function search_tcldir()                               */
/*                                                                    */
/*             FILE *fp    : TCL create file handle                   */
/*             int var     : the ListBox number                       */
/*             char *dirname : search diretory name                   */
/*             char *exp   :  must be found extention                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Find the extension in the directory and      */
/*                      show the basenames.                           */
/**********************************************************************/
void search_tcldir( FILE *fp, int var, char *dirname, char *ext )
{
	struct dirent *ent;
	DIR *dir;
	char buff[NAME_MAX+1];

	if( (dir = opendir( dirname )) == NULL ){
		fprintf( stderr, "cannot open %s\n", dirname );
		exit( 1 );
	}
	while( (ent = readdir( dir )) ){
		if( check_exp( ent->d_name, ext, buff ) ){
			fprintf( fp, ".lstb%d insert end \"%s\"\n", var, buff );
		}
	}
	closedir( dir );
}
