/**********************************************************************/
/*                                                                    */
/*              cut.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "plug.h"

char Name[256] = { "noname" };
int List = 0;
int Gui = NOGUI;
void search_file(FILE *, int, int, char *);

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "cut" };
char *Comment = { "cut efffect to cut-file(*.cut)" };

PLUGITEM Item[] = {
{ "list",	SWITCH,  &List, 1, 0,      "list all cutfile in #lwp" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" },
{ "cut",        GETSTR,  &Name, 64, 0,    "save cut file name" },
{ NULL, 	GETSTR,  &Name, 64, 0,    "save cut file name" } };
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ NOTCL | ENTRY, 3, "object cut file name", "cut=%s" },
{ TCLLIST, 3, "object cut file name", NULL },
{ TCLDIR, 0, (char *)search_file, NULL },
{ TCLDIR, 1, (char *)search_file, NULL },
{ TCLLISTENT, 3, "input file name", "cut=%s" },
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/


int Right, Left, Up, Down;	/* cut max-min position in the original */
int Width;			/* width of the image */
int Mask_kind;			/* kind of the mask */

#define WORD unsigned short
struct RawHeader {		/* cut file header */
	WORD spec;		/* signature */
	WORD width;		/* width of cut */
	WORD height;		/* height of cut */
	WORD mask;		/* mask kind of the cut */
};


/**********************************************************************/
/*                                                                    */
/*             function cut()                                         */
/*                                                                    */
/*             unsigned int *data : the data buffer                   */
/*                                                                    */
/*             return  int : 0==success  -1==fail                     */
/*                                                                    */
/*             function: cut main routine.                            */
/*                                                                    */
/**********************************************************************/
int cut( unsigned int *data )
{
	int i, j;
	int fd;
	struct RawHeader header;
	unsigned int *p;
	int len;
	char filename[256];
	unsigned int *buff;
	char *envstr;

	len = sizeof(unsigned int) * (Right - Left + 1);
	if( (buff = (unsigned int*)malloc(len) ) == NULL ){
		return( EMALLOC );
	}
	if( (envstr = getenv( "GEDO_LWP" )) == NULL ){
		strcpy( filename, "./" );
	} else {
		strcpy( filename, envstr );
	}
	if( filename[strlen(filename) -1] != '/' ){
		strcat( filename, "/" );
	}
	strcat( filename, Name );
	strcat( filename, ".cut" );
	fprintf( stderr, "saving %s...\n", filename );
	if( (fd = open( filename, O_RDWR | O_TRUNC )) == -1 ){
		if( (fd = open( filename, O_RDWR | O_CREAT, 0664)) == -1 ){
			free( buff );
			return( EOPENFILE );
		}
	}
	memset( &header, 0, sizeof(struct RawHeader) );
	header.spec = 1;
	header.width = (WORD)(Right - Left +1);
	header.height = (WORD)(Down - Up +1);
	header.mask  = (WORD)Mask_kind;
	if( write( fd, &header, sizeof(struct RawHeader) ) != sizeof(struct RawHeader) ){
		free( buff );
		return( EWRITEFILE );
	}
	p = data + Width * Up;
	for( j = Up; j <= Down; j++ ){
		if( Mask_kind == SIMPLEMASK ){
			memcpy( buff, &p[Left], len );
			for( i = 0; i < Right - Left + 1; i++ ){
				if( MASKON(buff[i]) ){
					if( Mask_kind == SIMPLEMASK ){
						buff[i] |= MASK;
					}
				}
			}
			if( write( fd, buff, len ) != len ){
				free( buff );
				return( EWRITEFILE );
			}
		} else {
			if( write( fd, &p[Left], len ) != len ){
				free( buff );
				return( EWRITEFILE );
			}
		}
		p += Width;
	}
	close( fd );
	free( buff );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function check_mask()                                  */
/*                                                                    */
/*             unsigned int at : pixel data                           */
/*                                                                    */
/*             return  int     : 1==selected  0==not selected         */
/*                                                                    */
/*             function:  Check the selected area for cut minimum.    */
/*                                                                    */
/**********************************************************************/
int check_mask( unsigned int at )
{
	if( Mask_kind == SIMPLEMASK ){
		return( MASKOFF(at) );
	} else {
		return( at & MASK );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	char *str;
	int height;
	unsigned int *data, *p;
	int i, j, ret;

	str = file->mes->u.data;
	Width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;
	Mask_kind = file->buf->mask_kind;

	Left = Width; Right = 0; Up = height; Down = 0;
	for( i = 0, p = data; i < height; i++ ){
		for( j = 0; j < Width; j++, p++ ){
			if( check_mask(*p) ){
				if( Left > j ){
					Left = j;
				}
				if( Right < j ){
					Right = j;
				}
				if( Up > i ){
					Up = i;
				}
				if( Down < i ){
					Down = i;
				}
			}
		}
	}
	if( Left > Right ){
		return( EILLSELECT );
	}
	if( Up > Down ){
		return( EILLSELECT );
	}
	ret = cut( data );
	if( ret != 0 ){
		strcpy( file->mes->u.data, strerror( errno ) );
		file->mes->datasize = strlen( file->mes->u.data );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;
	int ret;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		ret = line_config( file );
		if( ret == 0 && List ){
			search_dir( getenv( "GEDO_LWP" ), "cut" );
			List = 0;
		}
		return( ret );
	}
	return( 0 );
}

void search_file(FILE *fp, int var, int kind, char * arg)
{
	switch( kind ){
	case 0:
		fprintf( fp, ".lstb%d insert end \"noname\"\n", var );
		break;
	case 1:
		search_tcldir( fp, var, getenv( "GEDO_LWP" ), "cut" );
		break;
	default:
		break;
	}
}
