/**********************************************************************/
/*                                                                    */
/*           depth.c    gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "plug.h"

#define LUMIMASK  0
#define REDMASK	  1
#define GREENMASK 2
#define BLUEMASK  3

int  Abstract = LUMIMASK;

int Gui = NOGUI;
/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "depth" };
char *Comment = { "select the depth mask by the color" };

PLUGITEM Item[] = {
{ "lumi",  SWITCH,  &Abstract, LUMIMASK, 'L', "select by luminance" },
{ "red",   SWITCH,  &Abstract, REDMASK,  'R', "select by red depth" },
{ "green", SWITCH,  &Abstract, GREENMASK,'G', "select by green depth" },
{ "blue",  SWITCH,  &Abstract, BLUEMASK, 'B', "select by blue depth" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ RADIO, 0, "select by " },
{ ITEM, 0, "luminance(l)", "lumi" },
{ ITEM, 1, "red(r)", "red" },
{ ITEM, 2, "green(g)", "green" },
{ ITEM, 3, "blue(b)", "blue" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* image width and height */

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *data, *p;
	int i;
	unsigned int r, g, b;

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	data = file->buf->data;

	if( sb->selnum != SKNONE ){
		return( EILLSELECTION );
	}

	for( i = 0, p = data; i < Height * Width; i++, p++ ){
		switch( Abstract ){
		      case LUMIMASK:
			r = GET_RED(*p);
			g = GET_GREEN(*p);
			b = GET_BLUE(*p);
			r = (unsigned char)((float)r * 0.299 + 
					    (float)g * 0.587 + 
					    (float)b * 0.114);
			*p = ((r << 24) | (*p & ~MASK));
			break;
		      case REDMASK:
			r = GET_RED(*p);
			*p = ((r << 24) | (*p & ~MASK));
			break;
		      case GREENMASK:
			g = GET_GREEN(*p);
			*p = ((g << 24) | (*p & ~MASK));
			break;
		      case BLUEMASK:
			b = GET_BLUE(*p);
			*p = ((b << 24) | (*p & ~MASK));
			break;
		      default:
			return( -1 );
		}
	}
	file->buf->mask_kind = DEPTHMASK;
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
