/**********************************************************************/
/*                                                                    */
/*             edge.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "plug.h"


int Allow= 20;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "edge" };
char *Comment = { "select the depth mask by the color" };

PLUGITEM Item[] = {
{ "allow",  GETINT,  &Allow, 0, 65535,    "select edge torelance" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "threthold of detect edge", "allow=%d" }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

#define abs(p)  (p>0)?(p):((p)*(-1))

/**********************************************************************/
/*                                                                    */
/*             function check_edge()                                  */
/*                                                                    */
/*             unsigned int s : src pixel                             */
/*             unsigned int d : dst pixel                             */
/*                                                                    */
/*             return  int  : 1== is a edge   0== not is a edge       */
/*                                                                    */
/*             function: Find the edge.                               */
/*                                                                    */
/**********************************************************************/
int check_edge( unsigned int s, unsigned int d )
{
	unsigned int srgb[3];
	unsigned int drgb[3];
	int i;
	int tmp;

	srgb[0] = GET_RED(s);
	srgb[1] = GET_GREEN(s);
	srgb[2] = GET_BLUE(s);
	drgb[0] = GET_RED(d);
	drgb[1] = GET_GREEN(d);
	drgb[2] = GET_BLUE(d);

	for( i = 0; i < 3; i++ ){
		tmp = srgb[i] - drgb[i];
		if( tmp < 0 ){
			tmp = tmp * (-1);
		}
		if( tmp > Allow ){
			return( 1 );
		}
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *data, *p;
	int width, height;
	int i, j, flag;

	sb = &file->mes->u.sb;
	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	if( sb->selnum != SKNONE ){
		return( EILLSELECTION );
	}

	for( i = 0; i < height*width; i++ ){
		SETMASK( data[i] );
	}
	for( p = data, i = 0; i < height - 1; i++ ){
		for( j = 0; j < width - 1; j++, p++ ){
			flag = 0;
			if( check_edge( *p, *(p+1) ) ){
				flag = 1;
				CLRMASK( *(p+1) );
			}
			if( check_edge( *p, *(p+width-1) ) ){
				flag = 1;
				CLRMASK( *(p+width-1) );
			}
			if( check_edge( *p, *(p+width) ) ){
				flag = 1;
				CLRMASK( *(p+width) );
			}
			if( check_edge( *p, *(p+width+1) ) ){
				flag = 1;
				CLRMASK( *(p+width+1) );
			}
			if( MASKON(*p) && flag ){
				CLRMASK( *p );
			}
		}
	}
	file->buf->mask_kind = SIMPLEMASK;
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
