/**********************************************************************/
/*                                                                    */
/*             edit.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "plug.h"

#define ON	0
#define OFF	1
#define REV	2
#define CLON	3

int Mode = ON;
int Dot = 1;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };
char *Spec = 	{ "edit" };
char *Comment = { "select freely as dots" };

PLUGITEM Item[] = {
{ "dot",  GETINT,  &Dot, 0, 100,     "dot size" },
{ "on",    SWITCH,  &Mode, ON,  'N', "mask set mode" },
{ "off",   SWITCH,  &Mode, OFF, 'F', "mask off mode" },
{ "rev",   SWITCH,  &Mode, REV, 'R', "reverse the mask" },
{ "clr",   SWITCH,  &Mode, CLON,'C', "clear all mask and set dot" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};

int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "dot size ", "dot=%d" },
{ RADIO, 1, "Mode " },
{ ITEM, 1, "on(N)", "on" },
{ ITEM, 2, "off(F)", "off" },
{ ITEM, 3, "rev(R)", "rev" },
{ ITEM, 4, "clr(C)", "clr" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* image width and height */


/**********************************************************************/
/*                                                                    */
/*             function set_point()                                   */
/*                                                                    */
/*             char *buff : temp buffer                               */
/*             int x      : set center                                */
/*             int y      : set center                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Set the mask on the temp buffer.             */
/*                                                                    */
/**********************************************************************/
void set_point( char *buff, int x, int y )
{
	int a, b;
	int l, r, u, d;
	int i;
	char *p;

	a = (Dot > 1);
	b = Dot - a;

	l = max( x-a, 0 );
	r = min( x+b, Width );
	u = max( y-a, 0 );
	d = min( y+b, Height );

	p = buff + u * Width + l;
	for( i = 0; i < d - u; i++ ){
		memset( p, 1, r - l );
		p += Width;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_area()                                    */
/*                                                                    */
/*             unsigned *data  : image data                           */
/*             struct Sbuf *sb : selected points                      */
/*                                                                    */
/*             return  int  : 0==success   else==errorcode            */
/*                                                                    */
/*             function: Set the mask if Dot>1.                       */
/*                                                                    */
/**********************************************************************/
int set_area( unsigned *data, struct Sbuf *sb )
{
	char *buff;
	int i;

	if( (buff = malloc( Width * Height )) == NULL ){
		return( EMALLOC );
	}
	memset( buff, 0, Width * Height );
	for( i = 0; i < sb->defnum; i++ ){
		if( sb->def[i].x > Width || sb->def[i].x < 0 ){
			continue;
		}
		if( sb->def[i].y > Height || sb->def[i].y < 0 ){
			continue;
		}
		set_point( buff, sb->def[i].x, sb->def[i].y );
	}
	for( i = 0; i < Width * Height; i++ ){
		if( buff[i] ){
			if( Mode == ON || Mode == CLON ){
				CLRMASK(data[i]);
			} else if( Mode == OFF ){
				SETMASK(data[i]);
			} else if( Mode == REV ){
				if( MASKON(data[i]) ){
					CLRMASK(data[i]);
				} else {
					SETMASK(data[i]);
				}
			}
		}
	}
	free( buff );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *data, *p;
	int i, ret;

	sb = &file->mes->u.sb;
	if( sb->selnum != SKFREE ){
		return( EILLSELECTION );
	}
	Width = file->buf->width;
	Height = file->buf->height;
	data = file->buf->data;

	if( sb->defnum < 1 ){
		return( EILLNUMBER );
	}
	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}
	for( i = 0, p = data; i < Height * Width; i++, p++ ){
		SETMASK(*p);
	}

	if( Dot == 1 ){
	        for( i = 0; i < sb->defnum; i++ ){
			if( sb->def[i].x > Width || sb->def[i].x < 0 ){
				continue;
			}
			if( sb->def[i].y > Height || sb->def[i].y < 0 ){
				continue;
			}
			p = data + sb->def[i].x + sb->def[i].y * Width;
			if( Mode == ON || Mode == CLON ){
				CLRMASK(*p);
			} else if( Mode == OFF ){
				SETMASK(*p);
			} else if( Mode == REV ){
				if( MASKON(*p) ){
					CLRMASK(*p);
				} else {
					SETMASK(*p);
				}
			}
		}
	} else {
		if( (ret = set_area( data, sb )) != 0 ){
			return( ret );
		}
	}

	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}

