/**********************************************************************/
/*                                                                    */
/*            gamma.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include "plug.h"

int Gui = NOGUI;

#define M_RGB  0
#define M_YIQ  1
int Mode = M_RGB;

/* Used for RGB mode */
unsigned char Rstr[16] = { "00:40:80:c0:ff" };
unsigned char Gstr[16] = { "00:40:80:c0:ff" };
unsigned char Bstr[16] = { "00:40:80:c0:ff" };

/* used for YIQ mode */
unsigned char Ystr[16] = { "00:40:80:c0:ff" };
unsigned char Crstr[16] = { "00:40:80:c0:ff" };
unsigned char Cbstr[16] = { "00:40:80:c0:ff" };

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =	{ "gamma" };
char *Comment = { "gamma correcter, correct the color" };

PLUGITEM Item[] = {
{ "gui",   SWITCH,  &Gui, GUI,  0,    "use GUI interface" },
{ "nogui", SWITCH, &Gui, NOGUI,  0,    "not use GUI, set manually" },
{ "RGB", SWITCH, &Mode, M_RGB, 'R', "RGB gamma mode" },
{ "YIQ", SWITCH, &Mode, M_YIQ, 'Y', "YIQ gamma mode" },
{ "r", GETSTR,  Rstr, 16, 0,      "Red gamma: like xx:xx:xx:xx:xx" },
{ "g", GETSTR,  Gstr, 16, 0,      "Green    :      00 40 80 c0 ff" },
{ "b", GETSTR,  Bstr, 16, 0,      "Blue     :    for each value" },
{ "y", GETSTR,  Ystr, 16, 0,      "Y gamma: like xx:xx:xx:xx:xx" },
{ "Cr", GETSTR,  Crstr, 16, 0,    "Cr       :      00 40 80 c0 ff" },
{ "Cb", GETSTR,  Cbstr, 16, 0,    "Cb       :    for each value" },
{ "q", GETSTR,  Crstr, 16, 0,    "Cr       :      00 40 80 c0 ff" },
{ "i", GETSTR,  Cbstr, 16, 0,    "Cb       :    for each value" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ RADIO, 2, "select Color Mode" },
{ ITEM,  2,  "RGB(R)", "RGB" },
{ ITEM,  3,  "YIQ(Y)", "YIQ" },
{ RADIOEND, -1 },
{ ENTRY, 4, "Red node points ", "r=%s" },
{ ENTRY, 5, "Green node points ", "g=%s" },
{ ENTRY, 6, "Blue node points ", "b=%s" },
{ ENTRY, 7, "Y node points ", "y=%s" },
{ ENTRY, 8, "Cr node points ", "cr=%s" },
{ ENTRY, 9, "Cb node points ", "cb=%s" }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

struct fPosition {	/* for call spline */
	float x;
	float y;
};

struct fPosition *spline_line( struct fPosition *pos, int dim, int num, int pat );

unsigned char RGB[3][256];	/* translate array */

/*****************   translation macro ********************************/
#define get_y(r,g,b) (int)(0.299*(float)r+0.587*(float)g+0.114*(float)b) 
#define get_i(r,g,b) (int)(0.596*(float)r-0.274*(float)g-0.322*(float)b+128.0)
#define get_q(r,g,b) (int)(0.211*(float)r-0.522*(float)g+0.311*(float)b+128.0)

#define get_red(y,i,q) (int)((float)y+0.955*(float)(i-128)+0.622*(float)(q-128)) 
#define get_green(y,i,q) (int)((float)y-0.271*(float)(i-128)-0.647*(float)(q-128)) 
#define get_blue(y,i,q) (int)((float)y-1.109*(float)(i-128)+1.700*(float)(q-128))

#define ajust(p)	((p<0)?(0):((p>255)?(255):p))

/**********************************************************************/
/*                                                                    */
/*            function calc_yiqcolor()                                */
/*                                                                    */
/*            args  unsigned int *rgb : RGB pixel                     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Translate a pixel by trans array for YIQ Mode.*/
/*                                                                    */
/**********************************************************************/
void calc_yiqcolor( unsigned int *rgb )
{
	int y, i ,q;

	y = get_y( rgb[0], rgb[1], rgb[2] );
	i = get_i( rgb[0], rgb[1], rgb[2] );
	q = get_q( rgb[0], rgb[1], rgb[2] );
	y = RGB[0][ajust(y)];
	i = RGB[1][ajust(i)];
	q = RGB[2][ajust(q)];
	rgb[0] = ajust(get_red(y,i,q));
	rgb[1] = ajust(get_green(y,i,q));
	rgb[2] = ajust(get_blue(y,i,q));
}

/**********************************************************************/
/*                                                                    */
/*            function change_value()                                 */
/*                                                                    */
/*            args  unsigned int *v : each pixel                      */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Translate pixel by the gamma-effect for each  */
/*                     Mode.                                          */
/**********************************************************************/
void change_value( unsigned int *v )
{
	unsigned int rgb[3];
	int i;

        rgb[0] = GET_RED(*v);
	rgb[1] = GET_GREEN(*v);
	rgb[2] = GET_BLUE(*v);

	if( Mode == M_RGB ){
		for( i = 0; i < 3; i++ ){
			rgb[i] = RGB[i][rgb[i]];
		}
	} else {
		calc_yiqcolor( rgb ); 
	}
	*v = ((*v & MASK) | ((rgb[0] << 16) & RED) | 
	      ((rgb[1] << 8) & GREEN) | (rgb[2] & BLUE));
}

/**********************************************************************/
/*                                                                    */
/*            function get_hexval()                                   */
/*                                                                    */
/*            args  char *str  :  argment string                      */
/*                  unsigned char *val : Node selected                */
/*                                                                    */
/*            return int : number of points                           */
/*                                                                    */
/*            function: Translate init string to the node selected.   */
/*                                                                    */
/**********************************************************************/
int get_hexval( char *str, unsigned char *val )
{
	int num;
	int v;

	for( num = v = 0; *str; str++ ){
		if( isdigit(*str) ){
			v = v * 16 + *str - '0';
		} else if( isxdigit(*str) ){
			v = v * 16 + toupper(*str) - 'A' + 10;
		} else {
			*val++ = v;
			v = 0;
			num++;
		}
	}
	*val++ = v;
	return( ++num );
}

/**********************************************************************/
/*                                                                    */
/*            function value2spline()                                 */
/*                                                                    */
/*            args  unsigned char *y  : selected Node                 */
/*                  char *val         : result trans array            */
/*                                                                    */
/*            return int   : 0==success  -1==spline fail              */
/*                                                                    */
/*            function: Create trans array from the selected Node.    */
/*                                                                    */
/**********************************************************************/
int value2spline( char *col, char *val )
{
	static unsigned char x[5] = { 0x00, 0x40, 0x80, 0xc0, 0xff };
	unsigned char y[5] = { 0x00, 0x40, 0x80, 0xc0, 0xff };
	int num;
	struct fPosition xy[64];
	struct fPosition *ret;
	int i, j;
	float a, b;

	num = get_hexval( col, y ); 
	if( num < 5 ){
		fprintf( stderr, "need 5 points\n" );
		return( -1 );
	} else if( num > 5 ){
		fprintf( stderr, "more over 5 points, neglect over\n" );
	}
	for( i = 0; i < 5; i++ ){
		xy[i].x = (float)x[i];
		xy[i].y = (float)y[i];
	}
	if( (ret = spline_line( xy, 3, 5, 8 )) == NULL ){
		printf( "spline error\n" );
		return( -1 );
	}

	a = (ret[1].y - ret[0].y) / (ret[1].x - ret[0].x);
	b = (ret[0].y * ret[1].x - ret[1].y * ret[0].x )
		/ (ret[1].x - ret[0].x);
	for( i = j = 0; i < 256; i++ ){
		if( (float)i >= ret[j+1].x ){
			j++;
			a = (ret[j+1].y - ret[j].y) / (ret[j+1].x - ret[j].x);
			b = (ret[j].y * ret[j+1].x - ret[j+1].y * ret[j].x )
				/ (ret[j+1].x - ret[j].x);
		}
		val[i] = (unsigned char)(a * (float)i + b);
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function string2data()                                  */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return int   : 0==success  -1==spline error             */
/*                                                                    */
/*            function: For each element, translate string to trans-  */
/*                     array.                                         */
/**********************************************************************/
int string2data( void )
{
	if( Mode == M_RGB ){
		if( value2spline( Rstr, RGB[0] ) != 0 ){
			return( -1 );
		}
		if( value2spline( Gstr, RGB[1] ) != 0 ){
			return( -1 );
		}
		if( value2spline( Bstr, RGB[2] ) != 0 ){
			return( -1 );
		}
	} else {
		if( value2spline( Ystr, RGB[0] ) != 0 ){
			return( -1 );
		}
		if( value2spline( Crstr, RGB[1] ) != 0 ){
			return( -1 );
		}
		if( value2spline( Cbstr, RGB[2] ) != 0 ){
			return( -1 );
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function show_map()                                     */
/*                                                                    */
/*            args  struct File *file : transaction buff from gedo    */
/*                                                                    */
/*            return int  : 0==success  -1==fail                      */
/*                                                                    */
/*            function: Call GUI process xgamma.                      */
/*                                                                    */
/**********************************************************************/
int show_map( struct File *file )
{
	FILE *fpp;
	char buff[128];

	if( Mode == M_RGB ){
		sprintf( buff, "xgamma %d RGB %s %s %s", file->mes->shmid, Rstr, Gstr, Bstr );
	} else {
		sprintf( buff, "xgamma %d YIQ %s %s %s", file->mes->shmid, Ystr, Crstr, Cbstr );
	}
	if( (fpp = popen( buff, "r" )) == NULL ){
		fprintf( stderr, "exec xgamma fail\n" );
		return( -1 );
	}
	if( fgets( buff, 127, fpp ) == NULL ){
		fprintf( stderr, "cannot read\n" );
		return( -1 );
	}
	pclose( fpp );
	if( strncmp( buff, "error:", 6 ) == 0 ){
		strcpy( file->mes->u.data, &buff[6] );
		return( EGUIERROR );
	}
	if( strcmp( buff, "abort\n" ) == 0 ){
		return( EUSRBREAK );
	}
	strcpy( file->mes->u.data, buff );
	return( 0 );
}
	
/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	char *str;
	int width, height;
	unsigned int *data;
	int i;

	str = file->mes->u.data;
	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}
	for( i = 0; i < height * width; i++ ){
		if( MASKOFF(data[i]) ){
			change_value( &data[i] );
		}
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	int ret;
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( Gui ){
			if( (ret = show_map( file )) != 0 ){
				return( ret );
			}
			if( (ret = line_config( file )) != 0 ){
				return( ret );
			}
		} else {
			if( (ret = config_comline( file ) ) != 0 ){
				return( ret );
			}
		}
	} else {
		if( (ret = line_config( file )) != 0 ){
			return( ret );
		}
	}
	return( string2data() );
}
