/**********************************************************************/
/*                                                                    */
/*             gauss.c  gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>
#include <math.h>
#include "plug.h"

#ifndef PI
#ifdef M_PI
#define PI  M_PI
#endif
#endif

#define C_REPLACE	0
#define C_ADD		1
#define C_SUB		2
int Mode = 0;		/* replace, add or sub, the process mode */

int Gui = NOGUI;

struct Colors {			/* vector data for each colors */
	unsigned int median;	/* median of the color */
	unsigned int Rband;	/* the band from median to */
	unsigned int Lband;	/* the band from median to */
	unsigned int diff;	/* delta~2 */
	unsigned int max;	/* real maximum of the color */
	unsigned int min;	/* real miniimum of the color */
	int map[128];		/* result of the mapping tranlation table */
};	

int GaussFlag = 0;	/* gauss or flat */
struct Colors R   = { 0x80, 0x79, 0x79, 0x30 };
struct Colors G   = { 0x80, 0x79, 0x79, 0x30 };
struct Colors B   = { 0x80, 0x79, 0x79, 0x30 };

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "randcol" };
char *Comment = { "random color painter" };

PLUGITEM Item[] = {
{ "gui",   SWITCH,  &Gui, GUI,  'G',      "use GUI interface" },
{ "nogui", SWITCH,  &Gui, NOGUI,  'N',    "not use GUI interface" },
{ "rep",   SWITCH,  &Mode, C_REPLACE, 'R',   "replace the area by the color" },
{ "add",   SWITCH,  &Mode, C_ADD,     'A',   "add the color to the area" },
{ "sub",   SWITCH,  &Mode, C_SUB,     'S',   "sub the color to the area" },

{ "gauss", SWITCH,  &GaussFlag, 1,    'G',   "gauss mapping" },
{ "flat",  SWITCH,  &GaussFlag, 0,    'F',   "flat random mapping" },

{ "aveR",   GETINT,  &R.median, 0, 0xff,  "RED:random generate median" },
{ "RbandR",  GETINT,  &R.Rband, 0, 0x80,  "RED:random generate band(R)" },
{ "LbandR",  GETINT,  &R.Lband, 0, 0x80,  "RED:random generate band(L)" },
{ "diffR",   GETINT,  &R.diff,   0, 0x80,  "RED:gauss mapping delta" },

{ "aveG",   GETINT,  &G.median, 0, 0xff,  "GREEN:random generate median" },
{ "RbandG",  GETINT,  &G.Rband, 0, 0x80,  "GREEN:random generate band(R)" },
{ "LbandG",  GETINT,  &G.Lband, 0, 0x80,  "GREEN:random generate band(L)" },
{ "diffG",   GETINT,  &G.diff,   0, 0x80,  "GREEN:gauss mapping delta" },

{ "aveB",   GETINT,  &B.median, 0, 0xff,  "BLUE:random generate median" },
{ "RbandB",  GETINT,  &B.Rband, 0, 0x80,  "BLUE:random generate band(R)" },
{ "LbandB",  GETINT,  &B.Lband, 0, 0x80,  "BLUE:random generate band(L)" },
{ "diffB",   GETINT,  &B.diff,   0, 0x80,  "BLUE:gauss mapping delta" },
};

int ItemNum = sizeof(Item)/sizeof(PLUGITEM);
struct Tags Tag[] = {
  { RADIO, 2, "Paint Mode" },
  { ITEM,  2, "replace(R)", "rep" },
  { ITEM,  3, "add(A)", "add" },
  { ITEM,  4, "sub(S)", "sub" },
  { RADIOEND, -1 },

  { RADIO, 5, "Gauss Mode" },
  { ITEM,  5, "gauss(G)", "gauss" },
  { ITEM,  6, "flat(F)", "flat" },
  { RADIOEND, -1 },

  { ENTRY, 7, "RED:random generate median", "aveR=%d"},
  { ENTRY, 8, "RED:random generate band(R)", "RbandR=%d"},
  { ENTRY, 9, "RED:random generate band(L)", "LbandR=%d"},
  { ENTRY, 10, "RED:fauss mapping delta",     "diffR=%d"},

  { ENTRY, 11, "GREEN:random generate median", "aveG=%d"},
  { ENTRY, 12, "GREEN:random generate band(R)", "RbandG=%d"},
  { ENTRY, 13, "GREEN:random generate band(L)", "LbandG=%d"},
  { ENTRY, 14, "GREEN:fauss mapping delta",     "diffG=%d"},

  { ENTRY, 15, "BLUE:random generate median", "aveB=%d"},
  { ENTRY, 16, "BLUE:random generate band(R)", "RbandB=%d"},
  { ENTRY, 17, "BLUE:random generate band(L)", "LbandB=%d"},
  { ENTRY, 18, "BLUE:gauss mapping delta",     "diffB=%d"},
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);
DEFAULTSTATUS;
/********************** description end ********************/

unsigned int Ar, Ag, Ab;	/* Color divided RGB */

#define DIV	128

double Gauss[DIV*4];	/* gauss mapping original value */
void (*painter)(void);

#define calc2(a)	((a)>0xff)?(0xff):(((a)<0)?(0):(a))

/**********************************************************************/
/*                                                                    */
/*            function get_value()                                    */
/*                                                                    */
/*            args  unsigned int v  : original color(each)            */
/*                  Struct Colors *p : color struct(each)             */
/*                                                                    */
/*            return unsigned int : translated color                  */
/*                                                                    */
/*            function: RANDOM case, calculate color of each RGB.     */
/*                                                                    */
/**********************************************************************/
unsigned int get_value( unsigned int v, struct Colors *p )
{
	if( v >= 0x80 ){
		if( p->Lband == 0 ){
			v = p->median;
		} else {
			v = p->median - v % p->Lband;
		}
	} else {
		if( p->Rband == 0 ){
			v = p->median;
		} else {
			v = p->median + v % p->Rband;
		}
	}
	return( calc2(v) );
}

/**********************************************************************/
/*                                                                    */
/*            function random_color()                                 */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: RANDOM case, calculate new color from old.    */
/*                                                                    */
/**********************************************************************/
void random_color( void )
{
	unsigned int r;

	r = rand();

	Ab = (get_value(    r     & 0xff, &B ) & 0xff);
	Ag = (get_value( (r >> 8) & 0xff, &G ) << 8);
	Ar = (get_value( (r >>16) & 0xff, &R ) << 16);
}

/**********************************************************************/
/*                                                                    */
/*            function init_gauss()                                   */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: calculate gauss mapping to prepare.           */
/*                                                                    */
/**********************************************************************/
void init_gauss( void )
{
	int i;
	double total, at;

	for( i = 0; i < DIV * 4; i++ ){
		Gauss[i] = (double)i / (double)DIV;
		Gauss[i] *= Gauss[i];
		Gauss[i] = 1.0 / sqrt(2.0*PI) * exp( -0.5 * Gauss[i] );
	}
	total = 0.0;
	for( i = 0; i < DIV * 4; i++ ){
		at = Gauss[i] / (double)DIV;
		Gauss[i] = total;
		total += at;
	}
	for( i = 0; i < DIV * 4; i++ ){
		Gauss[i] /= total * 2.0;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function remap_gauss()                                  */
/*                                                                    */
/*            args  int d2  :  delta set                              */
/*                  int *map : result of the color gauss mapping      */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: make gauss mapping for each color by real data.*/
/*                                                                    */
/**********************************************************************/
void remap_gauss( int d2, int *map )
{
	int i, j;
	double dv;
	double at;
	double lmap[128];

	dv = (double)DIV / (double)d2;
	at = 0.0;
	for( i = 0; i < 128; i++ ){
		if( at < (double)(4 * DIV) ){
			lmap[i] = Gauss[(int)at] * DIV * 2.0;
			at += dv;
		} else {
			lmap[i] = 0.50 * DIV * 2.0;
		}
	}
	for( i = j = 0; i < 127; i++ ){
		while( (double)j < lmap[i] ){
			map[j++] = i - 1;
			if( j >= 127 ){
				break;
			}
		}
	}
	map[127] = 128 - lmap[127];
}

/**********************************************************************/
/*                                                                    */
/*            function get_range()                                    */
/*                                                                    */
/*            args  struct Colors *c : each color                     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: make confirm the real range of each color.    */
/*                                                                    */
/**********************************************************************/
void get_range( struct Colors *c )
{
	int i;

	for( i = 0; i < 128; i++ ){
		if( c->map[i] == 128 ){
			break;
		}
	}
	i++;
	if( c->Rband > i ){
		c->max = i;
	} else {
		c->max = c->Rband;
	}
	if( c->Lband > i ){
		c->min = i;
	} else {
		c->min = c->Lband;
	}
	if( c->median + c->max > 255 ){
		c->max = 255 - c->median;
	}
	if( (int)c->median - (int)c->min < 0 ){
		c->min = c->median;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function get_gauss()                                    */
/*                                                                    */
/*            args  unsigned int v  : original color(each)            */
/*                  struct Colors *p : color setting                  */
/*                                                                    */
/*            return unsigned int : new color(each)                   */
/*                                                                    */
/*            function: calculate new color for each of RGB in GAUSS  */
/*                     mode.                                          */
/**********************************************************************/
unsigned int get_gauss( unsigned int v, struct Colors *p )
{
	v %= (p->min + p->max);
	if( (int)v - (int)p->min > 0 ){
		v = v - p->min;
		v = p->median + p->map[v];
	} else {
		v = p->min - v;
		v = p->median - p->map[v];
	}
	return( v );
}

/**********************************************************************/
/*                                                                    */
/*            function gauss_color()                                  */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: calculate new GAUSS color.                    */
/*                                                                    */
/**********************************************************************/
void gauss_color( void )
{
	unsigned int r;

	r = rand();

	Ab = (get_gauss(    r     & 0xff, &B ) & 0xff);
	Ag = (get_gauss( (r >> 8) & 0xff, &G ) << 8);
	Ar = (get_gauss( (r >>16) & 0xff, &R ) << 16);
}

/**********************************************************************/
/*                                                                    */
/*             function add_color()                                   */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: In SIMPLEMASK, and at ADD-Mode,              */
/*                      add <Color> to each MASKOFF pixel.            */
/**********************************************************************/
void add_color( unsigned int *p, int size )
{
	unsigned int r, g, b;
	int i;

	for( i = 0; i < size; i++, p++ ){
		if( MASKOFF(*p) ){
			(*painter)();
			r = (*p & RED);
			g = (*p & GREEN);
			b = (*p & BLUE);
			if( (r = r + Ar) > RED ){
				r = RED;
			}
			if( (g = g + Ag) > GREEN ){
				g = GREEN;
			}
			if( (b = b + Ab) > BLUE ){
				b = BLUE;
			}
			*p = ((*p & MASK) | r | g | b);
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function sub_color()                                   */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: In SIMPLEMASK, and at SUB-Mode,              */
/*                      sub <Color> to each MASKOFF pixel.            */
/*                                                                    */
/**********************************************************************/
void sub_color( unsigned int *p, int size )
{
	unsigned int r, g, b;
	int i;

	for( i = 0; i < size; i++, p++ ){
		if( MASKOFF(*p) ){
			(*painter)();
			r = (*p & RED);
			g = (*p & GREEN);
			b = (*p & BLUE);
			if( r < Ar ){
				r = BLACK;
			} else {
				r = r - Ar;
			}
			if( g < Ag ){
				g = BLACK;
			} else {
				g = g - Ag;
			}
			if( b < Ab ){
				b = BLACK;
			} else {
				b = b - Ab;
			}
			*p = ((*p & MASK) | r | g | b);
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function rep_color()                                   */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: In SIMPLEMASK, and at REPLACE-Mode,          */
/*                  replace <Color> at each MASKOFF pixel.            */
/*                                                                    */
/**********************************************************************/
void rep_color( unsigned int *p, int size )
{
	int i;

	for( i = 0; i < size; i++, p++ ){
		if( MASKOFF(*p) ){
			(*painter)();
			*p = ((*p & MASK) | Ar | Ag | Ab);
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function add_dcolor()                                  */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function:  In DEPTHMASK, and at ADD-Mode,              */
/*                      add <Color> to each pixel by Mask depth.      */
/*                                                                    */
/**********************************************************************/
void add_dcolor( unsigned int *p, int size )
{
	unsigned int r, g, b, rate;
	int i;

	for( i = 0; i < size; i++, p++ ){
		(*painter)();
		rate = (*p & MASK) >> 24;
		r = (*p & RED);
		g = (*p & GREEN);
		b = (*p & BLUE);
/*		if( (r = r * (256-rate) / 256 + Ar * rate / 256) > RED ){*/
		if( (r = r + Ar * rate / 256) > RED ){ 
			r = RED;
		}
/*		if( (g = g * (256-rate) / 256 + Ag * rate / 256) > GREEN ){*/
		if( (g = g + Ag * rate / 256) > GREEN ){ 
			g = GREEN;
		}
/*		if( (b = b * (256-rate) / 256 + Ab * rate / 256) > BLUE ){*/
		if( (b = b + Ab * rate / 256) > BLUE ){ 
			b = BLUE;
		}
		*p = ((*p & MASK) | (r & RED) | (g & GREEN) | (b & BLUE));
	}
}

/**********************************************************************/
/*                                                                    */
/*             function sub_dcolor()                                  */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function:  In DEPTHMASK, and at SUB-Mode,              */
/*                      sub <Color> to each pixel by Mask depth.      */
/*                                                                    */
/**********************************************************************/
void sub_dcolor( unsigned int *p, int size )
{
	unsigned int r, g, b, rate;
	int i;
	unsigned int rr, rg, rb;

	for( i = 0; i < size; i++, p++ ){
		(*painter)();
		rate = (*p & MASK) >> 24;
		r = (*p & RED);
		g = (*p & GREEN);
		b = (*p & BLUE);
		rr = Ar * rate / 256;
		rg = Ag * rate / 256;
		rb = Ab * rate / 256;
		if( r < rr ){
			r = BLACK;
		} else {
			r = r - rr;
		}
		if( g < rg ){
			g = BLACK;
		} else {
			g = g - rg;
		}
		if( b < rb ){
			b = BLACK;
		} else {
			b = b - rb;
		}
		*p = ((*p & MASK) | (r & RED) | (g & GREEN) | (b & BLUE));
	}
}

/**********************************************************************/
/*                                                                    */
/*             function rep_dcolor()                                  */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function:  In DEPTHMASK, and at REPLACE-Mode,          */
/*                  replace <Color> to each pixel by Mask depth.      */
/*                                                                    */
/**********************************************************************/
void rep_dcolor( unsigned int *p, int size )
{
	int i;
	unsigned int r, g, b, rate;

	for( i = 0; i < size; i++, p++ ){
		(*painter)();
		rate = (*p & MASK) >> 24;
		if( (r = Ar * rate / 256) > RED ){
			r = RED;
		}
		if( (g = Ag * rate / 256) > GREEN ){
			g = GREEN;
		}
		if( (b = Ab * rate / 256) > BLUE ){
			b = BLUE;
		}
		*p = ((*p & MASK) | (r & RED) | (g & GREEN) | (b & BLUE));
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	char *str;
	int width, height;
	unsigned int *data;
	static int first = 1;

	str = file->mes->u.data;
	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	if( first ){
		srand( height * width );
		init_gauss();
		first = 0;
	}

	if( GaussFlag ){
		remap_gauss( R.diff, R.map );
		get_range( &R );
		remap_gauss( G.diff, G.map );
		get_range( &G );
		remap_gauss( B.diff, B.map );
		get_range( &B );
		painter = gauss_color;
	} else {
		painter = random_color;
	}
		
	if( file->buf->mask_kind == SIMPLEMASK ){
		if( Mode == C_ADD ){
			add_color( data, height * width );
		} else if( Mode == C_SUB ){
			sub_color( data, height * width );
		} else {
			rep_color( data, height * width );
		}
	} else if( file->buf->mask_kind == DEPTHMASK ){
		if( Mode == C_ADD ){
			add_dcolor( data, height * width );
		} else if( Mode == C_SUB ){
			sub_dcolor( data, height * width );
		} else {
			rep_dcolor( data, height * width );
		}
	} else {
		return( EILLMASK );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config_gui()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config_gui( struct File *file )
{
	FILE *fp;
	FILE *fpp;
	char buff[256];
	char fifo[256];
	char *e;
	char *gstr[2] = { "flat", "gauss" };

	if( (fp = create_fifo( fifo )) == NULL ){
		return( -1 );
	}
	if( (e = getenv("GEDO_PATH")) == NULL ){
		e = "./";
	}
	sprintf( buff, "%sxgauss <%s &", e, fifo );
	if( (fpp = popen( buff, "r" )) == NULL ){
		return( -1 );
	}
	fprintf( fp, "set gf \"%s\"\n"
		"set aveR   %d\nset aveG   %d\nset aveB   %d\n"
		"set RbandR %d\nset RbandG %d\nset RbandB %d\n"
		"set LbandR %d\nset LbandG %d\nset LbandB %d\n"
		"set diffR  %d\nset diffG  %d\nset diffB  %d\n",
		gstr[GaussFlag],
		R.median, G.median, B.median,
		R.Rband, G.Rband, B.Rband,
		R.Lband, G.Lband, B.Lband,
		R.diff, G.diff, B.diff );
	fclose( fp );
	
	fgets( buff, sizeof(buff)-1, fpp );
	pclose( fpp );

	if( strcmp( buff, "abort\n" ) == 0 ){
		return( EUSRBREAK );
	}
	buff[strlen(buff) - 1] = '\0';
	strcpy( file->mes->u.data, buff );
	return( line_config( file ) );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( Gui == 0 ){
			return( config_comline( file ) );
		} else {
			return( config_gui( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
