/**********************************************************************/
/*                                                                    */
/*              gif.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*             D0: not malloc/free pallet memory.. make static array. */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>

#include "plug.h"
#include "reduce.h"
#include "lzw.h"


#define PALLET256  	0
#define PALLET16	1
#define PALLET2		2

int WriteMode = PALLET256;
int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =    { "gif" };
char *Comment = { "Compuserve GIF loader/saver" };

PLUGITEM Item[] = {
{ "pallet256", 	NOLOAD | SWITCH, &WriteMode, PALLET256, 'P', "256 color pallet save" },
{ "pallet16", 	NOLOAD | SWITCH, &WriteMode, PALLET16, '1', "16 color pallet save" },
{ "bilevel",    NOLOAD | SWITCH, &WriteMode, PALLET2, 'B', "bilevel save" },
{ "gui",        NOLOAD | SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      NOLOAD | SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int      ItemNum = sizeof(Item) / sizeof(PLUGITEM);

struct Tags Tag[] = {
{ NOLOAD | RADIO, 0, "Color Mode" },
{ NOLOAD | ITEM, 0, "256-Color(P)", "pallet256" },
{ NOLOAD | ITEM, 1, "16-Color(1)", "pallet16" },
{ NOLOAD | ITEM, 2, "B/W(B)",    "bilevel" },
{ NOLOAD | RADIOEND, -1 },
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

int LSB_flag;			/* machine byte order */
unsigned short ColorRes;	/* color resolution */
unsigned short Pow[8]		/* for calucuration, 2^n */
 = { 2, 4, 8, 16, 32, 64, 128, 256 };
unsigned char LZWmin;		/* LZW start bit */
int Width, Height;		/* image width and height */
struct ColorTable {		/* color table */
	unsigned char R;
	unsigned char G;
	unsigned char B;
};
struct ColorTable GT[256];	/* Global Color Table in file */
struct ColorTable LT[256];	/* Local Color Table in file */

unsigned short GCTnum;		/* entry of Global Color Table */
unsigned short LCTnum;		/* entry of Local Color Table */

int GTflag = 0;			/* Global Color Table is here */
int LTflag = 0;			/* Local Color Table is now here */
unsigned char BGpix;		/* Background Color index */

struct Img {			/* Image descripter for each image*/
	int top;		/* image start(y) */
	int left;		/* image start(x) */
	int width;		/* image width(x) */
	int height;		/* image height(y) */
} Img;
int Inter;			/* the image is interlace image */


unsigned char *Tmp;		/* tmp buffer for decode */
int Tp = 0;			/* Tmp buffer now write */
int NowTmp;			/* Tmp buffer now alloced */

/* GIF header descrition */
#pragma pack(1)
struct GIFHeader {
	unsigned char sig[6];	/* signature */
	unsigned short width;	/* image width */
	unsigned short height;	/* image height */
	unsigned char  flag;	/* additional infomation */
	unsigned char  bgcolor;	/* background color */
	unsigned char  ratio;	/* X-Y ratio real */
} Header;

struct IM {			/* Image descripter description */
	unsigned char sig;	/* signature */
	unsigned short left;	/* image start(left) */
	unsigned short top;	/* image start(top) */
	unsigned short width;	/* image width */
	unsigned short height;	/* image height */
	unsigned char flag;	/* additional infomation */
	unsigned char lzw;	/* lzw start bit */
} IM;

/**********************************************************************/
/*                                                                    */
/*             function wconv()                                       */
/*                                                                    */
/*             unsigned short v : SHORT data                          */
/*                                                                    */
/*             return  unsigned short : SHORT data                    */
/*                                                                    */
/*             function: convert byte order(SHORT) if MSB machine     */
/*                                                                    */
/**********************************************************************/
unsigned short wconv( unsigned short v )
{
	register unsigned short v1, v2;

	if( LSB_flag ){
		return( v );
	}
	v1 = (v & 0x00ff) << 8;
	v2 = (v & 0xff00) >> 8;
	return( v1 | v2 );
}

/**********************************************************************/
/*                                                                    */
/*             function dconv()                                       */
/*                                                                    */
/*             unsigned long v : LONG data                            */
/*                                                                    */
/*             return  unsigned long : LONG data                      */
/*                                                                    */
/*             function: conver byte order(LONG) if MSB machine       */
/*                                                                    */
/**********************************************************************/
unsigned long dconv( unsigned long v )
{
	register unsigned long v1, v2, v3, v4;
	
	if( LSB_flag ){
		return( v );
	}
	v1 = (v & 0x000000ff) << 24;
	v2 = (v & 0x0000ff00) << 8;
	v3 = (v & 0x00ff0000) >> 8;
	v4 = (v & 0xff000000) >> 24;
	return( v1 | v2 | v3 | v4 );
}

/**********************************************************************/
/*                                                                    */
/*            function headers()                                      */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: header handler                                */
/*                                                                    */
/**********************************************************************/
int headers( unsigned char *buff, int *status, int *size )
{
	if( buff[0] == 'G' && buff[1] == 'I' && buff[2] == 'F' ){
	} else {
		return( -1 );
	}
	if( buff[3] == '8' && buff[5] == 'a' ){
		if( buff[4] == '7' ){
			fprintf( stderr, "GIF file 87a version\n" );
		} else if( buff[4] == '9' ){
			fprintf( stderr, "GIF file 89a version\n" );
		} else {
			return( EILLFORM );
		}
	}
	Width = wconv(*(unsigned short *)&buff[6]);
	Height = wconv(*(unsigned short *)&buff[8]);

	if( buff[10] & 0x80 ){
		*status = 1;
		GTflag = 1;
	} else {
		*status = 2;
		GTflag = 0;
	}

	ColorRes = ((buff[10] & 0x70) >> 4) + 1;
	GCTnum = Pow[(buff[10] & 0x07)];
	BGpix = buff[11];
	if( *status == 1 ){
		*size = GCTnum * 3;
	} else {
		*size = 1;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function GCT()                                          */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: Read Global Color Table.                      */
/*                                                                    */
/**********************************************************************/
int GCT( unsigned char *buff, int *status, int *size )
{
	int i;

	for( i = 0; i < GCTnum; i++ ){
		GT[i].R = *buff++;
		GT[i].G = *buff++;
		GT[i].B = *buff++;
	}
	*status = 2; *size = 1;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function Data()                                         */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: Next Block specify.                           */
/*                                                                    */
/**********************************************************************/
int Data( unsigned char *buff, int *status, int *size )
{
	if( buff[0] == 0x2c ){
		*status = 3; *size = 9;
	} else if( buff[0] == 0x21 ){
		*status = 4; *size = 1;
	} else if( buff[0] == 0x3b ){
		*status = -1; *size = 0;
	} else {
		return( EILLFORM );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function ID()                                           */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: Image descripter read.                        */
/*                                                                    */
/**********************************************************************/
int ID( unsigned char *buff, int *status, int *size )
{
	Img.left = wconv(*(unsigned short *)&buff[0]);
	Img.top = wconv(*(unsigned short *)&buff[2]);
	Img.width = wconv(*(unsigned short *)&buff[4]);
	Img.height = wconv(*(unsigned short *)&buff[6]);

	if( Img.top + Img.height > Height ){
		fprintf( stderr, "height=%d  local %d+%d\n", Height, Img.top, Img.height );
		return( EILLFORM );
	}
	if( Img.left + Img.width > Width ){
		fprintf( stderr, "width=%d  local %d+%d\n", Width, Img.left, Img.width );
		return( EILLFORM );
	}
	if( buff[8] & 0x80 ){
		*status = 8;
		LTflag = 1;
	} else {
		*status = 9;
		LTflag = 0;
	}
	if( buff[8] & 0x40 ){
		Inter = 1;
	} else {
		Inter = 0;
	}

	LCTnum = Pow[(buff[8] & 0x07)];
	if( *status == 8 ){
		*size = LCTnum * 3;
	} else {
		*size = 1;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function ImageBlock1()                                  */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: First Image block proc.                       */
/*                      Prepare the decode.                           */
/**********************************************************************/
int ImageBlock1( unsigned char *buff, int *status, int *size )
{
	if( Tmp == NULL ){
		if( (Tmp = malloc( 256 * 64 )) == NULL ){
			return( EMALLOC );
		}
		NowTmp = 256 * 64;
	}
	Tp = 0;
	LZWmin = buff[0] + 1;
	*status = 10; *size = 1;
	return( 0 );	
}

/**********************************************************************/
/*                                                                    */
/*            function ImageBlock2()                                  */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: Second Image proc. Recognize the strip size.  */
/*                                                                    */
/**********************************************************************/
int ImageBlock2( unsigned char *buff, int *status, int *size )
{
	if( buff[0] == 0 ){
		*status = 12; *size = 0;
	} else {
		*status = 11; *size = buff[0];
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function ImageBlock3()                                  */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: Read in the compressed data.                  */
/*                                                                    */
/**********************************************************************/
int ImageBlock3( unsigned char *buff, int *status, int *size )
{
	unsigned char *tmp;

	if( Tp + *size >= NowTmp ){
		if( (tmp = realloc( Tmp, NowTmp + 256 * 16 )) == NULL ){
			return( EMALLOC );
		}
		NowTmp += 256 * 16;
	}
	memcpy( &Tmp[Tp], buff, *size );
	Tp += *size;
	*status = 10; *size = 1;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function LCT()                                          */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: Read in the Local Color Table.                */
/*                                                                    */
/**********************************************************************/
int LCT( unsigned char *buff, int *status, int *size )
{
	int i;

	for( i = 0; i < LCTnum; i++ ){
		LT[i].R = *buff++;
		LT[i].G = *buff++;
		LT[i].B = *buff++;
	}

	*status = 9; *size = 1;
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function Extension()                                    */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                  int *mode   : extension kind                      */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: Externsion proc.                              */
/*                                                                    */
/**********************************************************************/
int Extension( unsigned char *buff, int *status, int *size, int *mode )
{
	if( buff[0] == 0xf9 ){
		*status = 6; *size = 1;
	} else if( buff[0] == 0xfe ){
		*status = 6; *size = 1;
	} else if( buff[0] == 0x01 ){
		*status = 6; *size = 1;
	} else if( buff[0] == 0xff ){
		*status = 6; *size = 1;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function DataBlock()                                    */
/*                                                                    */
/*            args  unsigned char *buff : read content                */
/*                  int *status : now status(set next status)         */
/*                  int *size   : now size(set next size)             */
/*                  int *mode   : extension kind                      */
/*                                                                    */
/*            return int   :  0==success  -1==fail                    */
/*                                                                    */
/*            function: Neglect the extension.                        */
/*                                                                    */
/**********************************************************************/
int DataBlock( unsigned char *buff, int *status, int *size, int mode )
{
	*size = buff[0]; *status = 6;
	switch( mode ){
	case 0xf9:
		fprintf( stderr, "found Graphic Control Block\n" );
		break;
	case 0xfe:
		fprintf( stderr, "found Comment Extension Block\n" );
		break;
	case 0x01:
		fprintf( stderr, "found Plain Text Extension Block\n" );
		break;
	case 0xff:
		fprintf( stderr, "found Application Extension Block\n" );
		break;
	case 0x00:
		*status = 2; *size = 1;
		break;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function normal_proc()                                  */
/*                                                                    */
/*            args  struct Shm *buf       : Data Buffer               */
/*                  unsigned char *buff   : uncompressed data         */
/*                  struct ColorTable *ct : using color table         */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: set normal data to the image buffer.          */
/*                                                                    */
/**********************************************************************/
void normal_proc( struct Shm *buf, unsigned char *buff, struct ColorTable *ct )
{
	unsigned int *p;
	int i, j;

	for( i = 0; i < Img.height; i++ ){
		p = buf->data + (Img.top + i) * Width + Img.left;
		for( j = 0; j < Img.width; j++, buff++ ){
			p[j] = ((unsigned int)ct[*buff].R << 16);
			p[j] |= ((unsigned int)ct[*buff].G << 8);
			p[j] |= ((unsigned int)ct[*buff].B);
		}
	}
}
/**********************************************************************/
/*                                                                    */
/*            function inter_proc()                                   */
/*                                                                    */
/*            args  struct Shm *buf       : Data Buffer               */
/*                  unsigned char *buff   : uncompressed data         */
/*                  struct ColorTable *ct : using color table         */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: set interlace data to the image buffer.       */
/*                                                                    */
/**********************************************************************/
void inter_proc( struct Shm *buf, unsigned char *buff, struct ColorTable *ct )
{
	unsigned int *p;
	int i, j;
	int line, count;
	int phase;

	line = count = phase = 0;
	for( i = 0; i < Img.height; i++ ){
		p = buf->data + (Img.top + line) * Width + Img.left;
		for( j = 0; j < Img.width; j++, buff++ ){
			p[j] = ((unsigned int)ct[*buff].R << 16);
			p[j] |= ((unsigned int)ct[*buff].G << 8);
			p[j] |= ((unsigned int)ct[*buff].B);
		}
		switch( phase ){
		case 0:
		case 1:
			line += 8;
			if( line >= Img.height ){
				if( phase == 0 ){
					line = 4;
					phase = 1;
				} else {
					line = 2;
					phase = 2;
				}
			}
			break;
		case 2:
			line += 4;
			if( line >= Img.height ){
				line = 1;
				phase = 3;
			}
			break;
		case 3:
			line += 2;
			if( line >= Img.height ){
				phase = -1;
			}
			break;
		default:
			return;
			break;
		}
	}
}
/**********************************************************************/
/*                                                                    */
/*            function render_image()                                 */
/*                                                                    */
/*            args  struct Shm *buf : data to set                     */
/*                                                                    */
/*            return int   : 0==success  else==errorcode              */
/*                                                                    */
/*            function: Expand the read data to the buffer.           */
/*                                                                    */
/**********************************************************************/
int render_image( struct Shm *buf )
{
	unsigned char *buff;
	int ret;
	int size;
	struct ColorTable *ct;

	size = Img.width * Img.height;
	if( (buff = malloc( size )) == NULL ){
		free( Tmp ); Tmp = NULL;
		return( EMALLOC );
	}
	ret = uncompress( LZWmin, Tmp, buff, size, GIF );
	if( ret == -1 ){
		fprintf( stderr, "decompress error\n" );
		free( buff ); free( Tmp ); Tmp = NULL;
		return( EILLFORM );
	}
	if( LTflag ){ 
		ct = LT;
	} else if( GTflag ){
		ct = GT;
	} else {
		fprintf( stderr, "color table not defined\n" );
		free( buff ); free( Tmp ); Tmp = NULL;
		return( EILLFORM );
	}
	if( Inter ){
		inter_proc( buf, buff, ct );
	} else {
		normal_proc( buf, buff, ct );
	}
	free( buff ); free( Tmp ); Tmp = NULL;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function set_background()                               */
/*                                                                    */
/*            args  struct Shm *buf : data to set                     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: 1st, set background colot to the buffer.      */
/*                                                                    */
/**********************************************************************/
void set_background( struct Shm *buf )
{
	int i;
	unsigned int col;

	col = ((unsigned int)GT[BGpix].R << 16);
	col |= ((unsigned int)GT[BGpix].G << 8);
	col |= ((unsigned int)GT[BGpix].B);
	for( i = 0; i < buf->width * buf->height; i++ ){
		buf->data[i] = col;
	}
}


/**********************************************************************/
/*                                                                    */
/*             function read_header()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ HEADER of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_header( struct File *file )
{
	unsigned char buff[256 * 3+1];
	FILE *fp;
	int size;
	int status;
	int mode;
	int error, ret;

	error = 0;
	if( (fp = fopen( file->name, "rb" )) == NULL ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EOPENFILE );
	}
	size = 13; status = 0;
/* A kind of FSM(Finite State Machine) */
	while( 1 ){
		if( size > 256 * 3 ){
			status = -2;
			file->mes->datasize = 0;
			error = EILLFORM;
			break;
		}
		if( fread( buff, 1, size, fp ) != size ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			error = EREADFILE;
			break;
		}
		if( status < 0 ){
			break;
		}
		switch( status ){
		case 0:
			if( (ret = headers( buff, &status, &size )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 1:
			if( (ret = GCT( buff, &status, &size )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			set_background( file->buf );
			break;
		case 2:
			if( (ret = Data( buff, &status, &size )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 3:
			if( (ret = ID( buff, &status, &size )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 9:
			if( (ret = ImageBlock1( buff, &status, &size )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 4:
			if( (ret = Extension( buff, &status, &size, &mode )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 6:
			if( (ret = DataBlock( buff, &status, &size, mode )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 8:
			if( (ret = LCT( buff, &status, &size )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 10:
			if( (ret = ImageBlock2( buff, &status, &size )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 11:
			if( (ret = ImageBlock3( buff, &status, &size )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			break;
		case 12:
			if( (ret = render_image( file->buf )) != 0 ){
				status = -2; size = 0; error = ret;
			}
			status = 2; size = 1;
			break;
		default:
			status = -2; size = 0; error = EILLFORM;
			break;
		}
	}

	fclose( fp );
	if( status != -1 ){
		return( error );
	}

	file->buf->width = Width;
	file->buf->height = Height;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function read_image()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ IMAGES of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_image( struct File *file )
{
	/* no operation */
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function write_header()                                */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE HEADER of the SPEC format to file       */
/*          B0: rewrite logic                                         */
/*                                                                    */
/**********************************************************************/
int write_header( struct File *file )
{
	if( (file->fd = open( file->name, O_RDWR | O_TRUNC )) == -1 ){
		if( (file->fd = open( file->name, O_RDWR | O_CREAT, 0664)) == -1 ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
		memset( &Header, 0, sizeof(struct GIFHeader) );
	}
	memcpy( Header.sig, "GIF87a", 6 );
	Header.width = wconv((unsigned short)file->buf->width);
	Header.height = wconv((unsigned short)file->buf->height);
	if( WriteMode == PALLET256 ){
		Header.flag = 0xf7;
	} else if( WriteMode == PALLET16 ){
		Header.flag = 0xf3;
	} else if( WriteMode == PALLET2 ){
		Header.flag = 0xf0;
	}
	Header.bgcolor = 0;
	Header.ratio = 0;

	if( write( file->fd, &Header, sizeof(struct GIFHeader) ) != sizeof(struct GIFHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EWRITEFILE );
	}

	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE IMAGES of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_image( struct File *file )
{
	int i, j;
	int ret;
	struct ReduceInfo ri;
	unsigned char CT[256*3];
	unsigned char *buff;
	unsigned char *p;
/* D0 add, use reduce pallet static area(not malloc) */     
	static struct Color pallet[256];
/* end D0 */

	ri.submode = 0;
	switch( WriteMode ){
	case PALLET256:
		ri.mode = GETPALLETCOLOR;
		ri.target = 256;
		ri.tmp_colors = 16;
		ri.submode = (COLOR | TOTALNUM | ZAGGY);
		IM.lzw = 8;
		break;
	case PALLET16:
		ri.mode = GETPALLETCOLOR;
		ri.target = 16;
		ri.tmp_colors = 4;
		ri.submode = (COLOR | TOTALNUM | ZAGGY);
		IM.lzw = 4;
		break;
	case PALLET2:
		ri.mode = GETPALLETCOLOR;
		ri.target = 2;
		ri.tmp_colors = 2;
		ri.submode = (BW | TOTALNUM | ZAGGY);
		IM.lzw = 2;
		break;
	}
	if( (ri.image = malloc( file->buf->width * file->buf->height )) == NULL ){
		return( EMALLOC );
	}
/* D0 delete, use reduce pallet static area(not malloc)
	if( (ri.pallet = malloc( sizeof(struct Color) * ri.target )) == NULL ){
		free( ri.image );
		return( EMALLOC );
	}*/
/* D0 add */
	ri.pallet = pallet;
/* end D0 */
	if( (ret = get_reduce_image( file->buf, &ri )) != 0 ){
/*		free( ri.image ); free( ri.pallet );*/
		free( ri.image );
		return( ret );
	}

	for( i = j = 0; i < ri.target; i++ ){
		CT[j++] = ri.pallet[i].R;
		CT[j++] = ri.pallet[i].G;
		CT[j++] = ri.pallet[i].B;
	}
	if( write( file->fd, CT, 3 * ri.target ) <= 0 ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		close( file->fd );
/*		free( ri.image ); free( ri.pallet );*/
		free( ri.image );
		return( EWRITEFILE );
	}
	IM.sig = 0x2c;
	IM.top = IM.left = 0;
	IM.width = wconv((unsigned short)file->buf->width);
	IM.height = wconv((unsigned short)file->buf->height);
	IM.flag = 0;
	if( write( file->fd, &IM, sizeof(struct IM) ) <= 0 ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		close( file->fd );
/*		free( ri.image ); free( ri.pallet );*/
		free( ri.image );
		return( EWRITEFILE );
	}
	if( (buff = malloc( file->buf->width * file->buf->height )) == NULL ){
		close( file->fd );
/*		free( ri.image ); free( ri.pallet );*/
		free( ri.image );
		return( EMALLOC );
	}

	if( (ret = compress( IM.lzw+1, ri.image, buff, IM.width * IM.height, GIF )) <= 0 ){
		strcpy( file->mes->u.data, "failed compress" );
		file->mes->datasize = strlen(file->mes->u.data);
		close( file->fd ); free( buff );
/*		free( ri.image ); free( ri.pallet );*/
		free( ri.image );
		return( -1 );
	}
	for( p = buff; ret > 0; ){
		if( ret >= 256 ){
			CT[0] = 0xff;
			memcpy( &CT[1], p, 255 );
			if( write( file->fd, CT, 256 ) <= 0 ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd ); free( buff );
/*				free( ri.image ); free( ri.pallet );*/
				free( ri.image );
				return( EWRITEFILE );
			}
			ret -= 255;
			p += 255;
		} else {
			CT[0] = (unsigned char)ret;
			memcpy( &CT[1], p, ret );
			if( write( file->fd, CT, ret + 1 ) <= 0 ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd ); free( buff );
/*				free( ri.image ); free( ri.pallet );*/
				free( ri.image );
				return( EWRITEFILE );
			}
			CT[0] = 0;
			CT[1] = 0x3b;
			if( write( file->fd, CT, 2 ) <= 0 ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd ); free( buff );
/*				free( ri.image ); free( ri.pallet );*/
				free( ri.image );
				return( EWRITEFILE );
			}
			p += ret;
			ret = 0;
		}
	}
	close( file->fd ); free( buff );
/*	free( ri.image ); free( ri.pallet );*/
	free( ri.image );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
