/**********************************************************************/
/*                                                                    */
/*           inner.c    gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.9  for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "plug.h"

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =	{ "inner" };
char *Comment = { "inner selecter, make thick current mask interior" };

PLUGITEM Item[];
int ItemNum = 0;
struct Tags Tag[];
int TagNum = 0;
DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* image width and height */
static unsigned int *Data;	/* image buffer */

#define MASK2ON(p)	(((p)&0x08000000))
#define MASK2OFF(p)	(((p)&0x08000000) == 0)
#define CLRMASK2(p)	((p)=(p)&0xf7ffffff)
#define SETMASK2(p)	((p)=(p)|0x08000000)

#define xydata(x,y) Data[(y)*Width+(x)]

/**********************************************************************/
/*                                                                    */
/*             function check_outer()                                 */
/*                                                                    */
/*             int x   : point X                                      */
/*             int y   : point Y                                      */
/*                                                                    */
/*             return  int  : 0==out of the mask  1==inner of the mask*/
/*                                                                    */
/*             function: Check the point is inner or outer of the     */
/*                      current mask.                                 */
/**********************************************************************/
int check_outer( int x, int y )
{
	int i;

	if( x == Width - 1 || x == 0 ){
		return( 0 );
	}
	if( y == Height - 1 || y == 0 ){
		return( 0 );
	}
	for( i = x+1; i < Width; i++ ){
		if( MASKOFF( xydata(i,y) ) ){
			break;
		}
	}
	if( i == Width ){
		return( 0 );
	}

	for( i = x-1; i >= 0; i-- ){
		if( MASKOFF( xydata(i,y) ) ){
			break;
		}
	}
	if( i == -1 ){
		return( 0 );
	}

	for( i = y+1; i < Height; i++ ){
		if( MASKOFF( xydata(x,i) ) ){
			break;
		}
	}
	if( i == Height ){
		return( 0 );
	}

	for( i = y-1; i >= 0; i-- ){
		if( MASKOFF( xydata(x,i) ) ){
			break;
		}
	}
	if( i == -1 ){
		return( 0 );
	}
	return( 1 );
}

/**********************************************************************/
/*                                                                    */
/*             function set_outer()                                   */
/*                                                                    */
/*             int x   : start point X                                */
/*             int y   : start point Y                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Recursively the inner point clear the mask   */
/*                      and do the near points.                       */
/**********************************************************************/
void set_outer( int x, int y )
{
	volatile int i, left, right, up, down;

	if( MASKOFF( xydata(x, y) )  ){
		return;
	}
	if( MASK2ON( xydata(x,y) ) ){
		return;
	}
	if( x >= Width || x < 0 ){
		return;
	}
	if( y >= Height || y < 0 ){
		return;
	}
	SETMASK2( xydata(x, y) );
	for( i = x + 1; i < Width; i++ ){
		if( MASKON( xydata(i, y) ) ){
			SETMASK2( xydata(i, y) );
		} else {
			break;
		}
	}
	right = i;
	for( i = x - 1; i >= 0; i-- ){
		if( MASKON( xydata(i, y) ) ){
			SETMASK2( xydata(i, y) );
		} else {
			break;
		}
	}
	left = i;
	for( i = y + 1; i < Height; i++ ){
		if( MASKON( xydata(x, i) ) ){
			SETMASK2( xydata(x, i) );
		} else {
			break;
		}
	}
	down = i;
	for( i = y - 1; i >= 0; i-- ){
		if( MASKON( xydata(x, i) ) ){
			SETMASK2( xydata(x, i) );
		} else {
			break;
		}
	}
	up = i;

	for( i = x + 1; i < right; i++ ){
		set_outer( i, y+1 );
		set_outer( i, y-1 );
	}
	for( i = x - 1; i > left; i-- ){
		set_outer( i, y+1 );
		set_outer( i, y-1 );
	}
	for( i = y + 1; i < down; i++ ){
		set_outer( x + 1, i );
		set_outer( x-1, i );
	}
	for( i = y - 1; i > up; i-- ){
		set_outer( x+1, i );
		set_outer( x-1, i );
	}
}



/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *p;
	int i, j;
	struct Pos *pos;
	int count;

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	Data = file->buf->data;

	if( sb->selnum != SKNONE ){
		return( EILLSELECTION );
	}
	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}

	p = Data;
	if( (pos = (struct Pos *)malloc( sizeof(struct Pos) * Width * Height )) == NULL ){
		return( EMALLOC );
	}
	count = 0;
        printf( "wait for a while....\n" );
        for( j = 0; j < Width; j++, p++ ){
		if( MASKON( *p ) ){
			if( check_outer( j, 0 ) == 0 ){
				pos[count].x = j;
				pos[count++].y = 0;
			}
		}
	}
        for( j = 0; j < Width; j++, p++ ){
		if( MASKON( *p ) ){
			if( check_outer( j, Height - 1 ) == 0 ){
				pos[count].x = j;
				pos[count++].y = Height - 1;
			}
		}
	}
	for( i = 1; i < Height - 1; i++ ){
		if( MASKON( *p ) ){
			if( check_outer( 0, i ) == 0 ){
				pos[count].x = 0;
				pos[count++].y = i;
			}
		}
		if( MASKON( *p ) ){
			if( check_outer( Width - 1, i ) == 0 ){
				pos[count].x = Width - 1;
				pos[count++].y = i;
			}
		}
	}
	for( i = 0; i < count; i++ ){
		set_outer( pos[i].x, pos[i].y );
	}
	for( i = 0, p = Data; i < Height * Width; i++, p++ ){
		if( MASK2ON( *p ) ){
			CLRMASK2( *p );
			SETMASK( *p );
		} else {
			CLRMASK( *p );
		}
	}
	free( pos );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		printf( "inner: no configration parameter\n" );
		str[0] = '\0';
		return( 0 );
	} else {
		return( line_config( file ) );
	}
}
