/**********************************************************************/
/*                                                                    */
/*          kanji.c     gedo PLUG text.c and xsting.c library routine */
/*                   for multi KANJI Code to read                     */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <string.h>
#include "xstring.h"

/* KANJI code mode */
#define JIS	0
#define EUC	1
#define SJIS	2

/* WSTRING is the 2byte unsigned char  to express ASCII and Kanji.
  It's expressions.

typedef unsigned char  WSTRING;
 
ASCII		00 xx
KANJI		xx yy  by JIS Code without ESC, KIN, KOUT, K2
End of line     00 0a
End of string   00 00

Converted by kanji_set(), and used to display at Draw_str()  */

/* symbol for the end detection */
#define EOI	(unsigned short)0xffff

/* if JIS code, ESC KIS K2 means KANJI Start, and ESC KSO K2 means KANJI end */
#define	ESC		0x1b		/* control code : ESCAPE	*/
#define KIS		0x24		/* '$' */
#define KSO		0x28		/* '(' */
#define	K2		0x42		/* 'B' */

#define BYTE unsigned char
#define KANJI 1
#define NOKANJI 0

/* like stdio stream */
#define GETCHAR()	(in[ip])?(in[ip++]):(EOI)
#define PUTCHAR(a)	(out[op++]=(a))
#define UNGETC(a)	(ip--)

/**********************************************************************/
/*                                                                    */
/*            function jc_jmstojis()                                  */
/*                                                                    */
/*            args  unsigned short c   : SJIS kanji code              */
/*                                                                    */
/*            return  unsigned short   : JIS kanji code               */
/*                                                                    */
/*            function: Conver a SJIS code to a JIS code.             */
/*                                                                    */
/**********************************************************************/
static unsigned short jc_jmstojis( unsigned short c )
{
	register unsigned short s0, s1;

	s0 = c & 0x00ff;
	s1 = (c & 0xff00) >> 8;
	if( s1 > 0x9f+0x40 ){
		s1 -= 0x40;
	}
	s1 -= 0x81;
	s1 = s1 << 1;
	s1 += 0x21;
	if( s0 > 0x5f+0x20 ){
		s0 -= 0x20;
	} else {
		s0 -= 0x1f;
	}
	if( s0 >= 0x7f ){
		s1++; s0 -= 0x5e;
	}
	if( s1 < 0x21 || s1 > 0x7e || s0 < 0x21 || s0 > 0x7e ){
		return( 0 );
	}
	return( (s1 << 8) | s0 );
}


/**********************************************************************/
/*                                                                    */
/*            function sjis_kanji()                                   */
/*                                                                    */
/*            args  unsigned char *in  : string include SJIS codes    */
/*                  WSTRING       *out : internal WSTRING code        */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: Convert string include SJIS code to the       */
/*                 internal WSTRING encoding.                         */
/**********************************************************************/
static void sjis_kanji( unsigned char *in, WSTRING *out )
{
	register unsigned short c0, c1;
	register int ip = 0;
	register int op = 0;
	int status;
	unsigned short tmp;

	status = NOKANJI;
	while( (c0 = GETCHAR()) != EOI ){
		if( status == KANJI ){
			if( (c0 >= 0x81 && c0 <= 0x9f) ||
			    (c0 >= 0xe0 && c0 <= 0xfc) ){
				c1 = GETCHAR();
				tmp = jc_jmstojis( (c0 << 8) | c1 );
				if( tmp == 0 ){
					PUTCHAR( 0x22 );
					PUTCHAR( 0x22 );
				} else {
					PUTCHAR( (BYTE)(tmp >> 8) );
					PUTCHAR( (BYTE)(tmp & 0xff) );
				}
			} else if( c0 < 0x20 ){
				PUTCHAR( 0 );
				PUTCHAR( c0 );
			} else {
				status = NOKANJI;
				if( c0 == '\\' ){
					if( (c1 = GETCHAR()) == 'n' ){
						PUTCHAR( 0 );
						PUTCHAR( '\n' );
					} else {
						PUTCHAR( 0 );
						PUTCHAR( c0 );
						UNGETC( c1 );
					}
				} else {
					PUTCHAR( 0 );
					PUTCHAR( c0 );
				}
			}
		} else {
			if( (c0 >= 0x81 && c0 <= 0x9f) ||
			    (c0 >= 0xe0 && c0 <= 0xfc) ){
				status = KANJI;
				c1 = GETCHAR();
				tmp = jc_jmstojis( (c0 << 8) | c1 );
				if( tmp == 0 ){
					PUTCHAR( 0x22 );
					PUTCHAR( 0x22 );
				} else {
					PUTCHAR( (BYTE)(tmp >> 8) );
					PUTCHAR( (BYTE)(tmp & 0xff) );
				}
			} else if( c0 == '\\' ){
				if( (c1 = GETCHAR()) == 'n' ){
					PUTCHAR( 0 );
					PUTCHAR( '\n' );
				} else {
					PUTCHAR( 0 );
					PUTCHAR( c0 );
					UNGETC( c1 );
				}
			} else {
				PUTCHAR( 0 );
				PUTCHAR( c0 );
			}
		}
	}
	PUTCHAR( 0 );
	PUTCHAR( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function jis_kanji()                                    */
/*                                                                    */
/*            args  unsigned char *in  : string include JIS codes     */
/*                  WSTRING       *out : internal WSTRING code        */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: Convert string include JIS code to the        */
/*                 internal WSTRING encoding.                         */
/*                                                                    */
/**********************************************************************/
static void jis_kanji( unsigned char *in, WSTRING *out )
{
	register unsigned short c0, c1, c2;
	register int ip = 0;
	register int op = 0;
	int status;

	status = NOKANJI;
	while( (c0 = GETCHAR()) != EOI ){
		if( c0 == ESC ){
			c1 = GETCHAR();
			if( c1 == KIS ){
				if( status == NOKANJI ){
					c2 = GETCHAR();
					if( c2 == K2 ){
						status = KANJI;
					} else {
						PUTCHAR( c0 );
						PUTCHAR( c1 );
						UNGETC( c2 );
					}
				} else {
					PUTCHAR( c0 );
					PUTCHAR( c1 );
				}
			} else if( c1 == KSO ){
				if( status == KANJI ){
					c2 = GETCHAR();
					if( c2 == K2 ){
						status = NOKANJI;
					} else {
						PUTCHAR( c0 );
						PUTCHAR( c1 );
						UNGETC( c2 );
					}
				} else {
					PUTCHAR( c0 );
					PUTCHAR( c1 );
				}
			} else {
				PUTCHAR( c0 );
				PUTCHAR( c1 );
			}
		} else if( c0 < 0x20 ){
			PUTCHAR( 0 );
			PUTCHAR( c0 );
		} else if( status == KANJI ){
			c1 = GETCHAR();
			PUTCHAR( c0 );
			PUTCHAR( c1 );
		} else if( c0 == '\\' ){
			if( (c1 = GETCHAR()) == 'n' ){
				PUTCHAR( 0 );
				PUTCHAR( '\n' );
			} else {
				PUTCHAR( 0 );
				PUTCHAR( c0 );
				UNGETC( c1 );
			}
		} else {
			PUTCHAR( 0 );
			PUTCHAR( c0 );
		}
	}
	PUTCHAR( 0 );
	PUTCHAR( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function euc_kanji()                                    */
/*                                                                    */
/*            args  unsigned char *in  : string include EUC codes     */
/*                  WSTRING       *out : internal WSTRING code        */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: Convert string include EUC code to the        */
/*                 internal WSTRING encoding.                         */
/**********************************************************************/
static void euc_kanji( unsigned char *in, WSTRING *out )
{
	register unsigned short c0, c1;
	register int ip = 0;
	register int op = 0;
	int status;

	status = NOKANJI;
	while( (c0 = GETCHAR()) != EOI ){
		if( status == KANJI ){
			if( c0 > 0x80 ){
				c1 = GETCHAR();
				PUTCHAR( c0 & 0x7f );
				PUTCHAR( c1 & 0x7f );
			} else {
				status = NOKANJI;
				if( c0 == '\\' ){
					if( (c1 = GETCHAR()) == 'n' ){
						PUTCHAR( 0 );
						PUTCHAR( '\n' );
					} else {
						PUTCHAR( 0 );
						PUTCHAR( c0 );
						UNGETC( c1 );
					}
				} else {
					PUTCHAR( 0 );
					PUTCHAR( c0 );
				}
			}
		} else {
			if( c0 >= 0x80 ){
				status = KANJI;
				c1 = GETCHAR();
				PUTCHAR( c0 & 0x7f );
				PUTCHAR( c1 & 0x7f );
			} else {
				if( c0 == '\\' ){
					if( (c1 = GETCHAR()) == 'n' ){
						PUTCHAR( 0 );
						PUTCHAR( '\n' );
					} else {
						PUTCHAR( 0 );
						PUTCHAR( c0 );
						UNGETC( c1 );
					}
				} else {
					PUTCHAR( 0 );
					PUTCHAR( c0 );
				}
			}
		}
	}
	PUTCHAR( 0 );
	PUTCHAR( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function kanji_set()                                    */
/*                                                                    */
/*            args  unsigned char *text : string include KANJI code   */
/*                  int mode  :  Kanji coding flag(JIS,EUC,SJIS)      */
/*                  WSTRING *out : string of internal KANJI codes     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Convert string include etc. KANJI coding to   */
/*               the internal coding WSTRING.                         */
/**********************************************************************/
void kanji_set( unsigned char *text, int mode, unsigned char *out  )
{
	if( mode == JIS ){
		jis_kanji( text, out );
	} else if( mode == EUC ){
		euc_kanji( text, out );
	} else if( mode == SJIS ){
		sjis_kanji( text, out );
	}
}

