/**********************************************************************/
/*                                                                    */
/*             line.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "plug.h"

int Border = 5;

#define LINE	0
#define SPLINE	1
int Mode = LINE;

#define RAWLINE  0
#define FINELINE 1
int Fine = RAWLINE;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "line" };
char *Comment = { "line/spline selecter" };

PLUGITEM Item[] = {
{ "width",  GETINT,  &Border, 0, 400,    "line width" },
{ "spline", SWITCH,  &Mode, SPLINE, 'S',   "spline mode" },
{ "line",   SWITCH,  &Mode, LINE,   'L',   "line mode" },
{ "fine",   SWITCH,  &Fine, FINELINE, 'F', "fine quality mode" },
{ "coase",  SWITCH,  &Fine, RAWLINE, 'C',  "coase quality mode" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "width of the line", "width=%d" },
{ RADIO, -1, "Mode " },
{ ITEM, 2, "normal line(l)", "line" },
{ ITEM, 1, "spline(s)", "spline" },
{ RADIOEND, -1 },
{ RADIO, -1, "Quality " },
{ ITEM, 3, "fine(f)", "fine" },
{ ITEM, 4, "coase(c)", "coase" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* image width and height */
int HB;				/* a half of the Border */

struct fPosition {	/* for spline call */
	float x;
	float y;
};

struct fPosition *spline_line( struct fPosition *pos, int dim, int num, int pat );

/**********************************************************************/
/*                                                                    */
/*             function paint_hline()                                 */
/*                                                                    */
/*             unsigned int *p : line start data                      */
/*             int pos         : line start X                         */
/*             int line        : line start Y                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Mask off the a little horizonal case.        */
/*                                                                    */
/**********************************************************************/
void paint_hline( unsigned int *p, int pos, int line )
{
	int i;
	int bdr;
	int off;

	if( Border == 1 ){
		CLRMASK(p[pos]);
		return;
	}
	if( (bdr = pos - Border / 2) < 0 ){
		off = 0;
		bdr = Border + bdr;
	} else if( (bdr = pos + Border / 2) >= Width ){
		off = pos - Border / 2;
		bdr = Border - bdr + Width - 1;
	} else {
		off = pos - Border / 2;
		bdr = Border;
	}
	for( i = 0; i < bdr; i++ ){
		CLRMASK(p[off+i]);
	}
}

/**********************************************************************/
/*                                                                    */
/*             function paint_vline()                                 */
/*                                                                    */
/*             unsigned int *p : line start data                      */
/*             int pos         : line start X                         */
/*             int line        : line start Y                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Mask off a little vertical case.             */
/*                                                                    */
/**********************************************************************/
void paint_vline( unsigned int *p, int pos, int line )
{
	int i;
	int bdr;

	if( Border == 1 ){
		CLRMASK(p[pos]);
		return;
	}
	if( (bdr = line - Border / 2) < 0 ){
		p -= Width * line;
		bdr = Border + bdr;
	} else if( (bdr = line + Border / 2) >= Height ){
		p -= (Border / 2) * Width;
		bdr = Border - bdr + Height - 1;
	} else {
		p -= (Border / 2) * Width;
		bdr = Border;
	}
	for( i = 0; i < bdr; i++ ){
		CLRMASK(p[pos]);
		p += Width;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function round()                                       */
/*                                                                    */
/*             double f     : round value                             */
/*                                                                    */
/*             return  int  : rounded value                           */
/*                                                                    */
/*             function: Round the float value to the int value.      */
/*                                                                    */
/**********************************************************************/
int round( double f )
{
	double r;

	r = floor( f );
	if( f - r >= 0.5 ){
		return( (int)r + 1 );
	} else {
		return( (int)r );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_line()                                    */
/*                                                                    */
/*             unsigned int *data  : image data                       */
/*             struct Pos s        : start position                   */
/*             struct Pos e        : end position                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Coase line main for each line.               */
/*                                                                    */
/**********************************************************************/
void set_line( unsigned int *data, struct Pos s, struct Pos e )
{
	int i, j;
	unsigned int *p;
	int min, max, off, d, omax, dn;
	double a;
	int dx, dy;
	void (*paint_line)( unsigned int *, int, int );

	if( s.x == e.x ){
		min = min( s.y, e.y );
		max = max( s.y, e.y );
		for( p = &data[min * Width], i = min; i <= max; i++ ){
			paint_hline( p, s.x, i );
			p += Width;
		}
	} else if( s.y == e.y ){
		min = min( s.x, e.x );
		max = max( s.x, e.x );
		for( p = &data[s.y * Width], i = min; i <= max; i++ ){
			paint_vline( p, i, s.y );
		}
	} else {
		if( (dx = e.x - s.x) > 0 ){
			dx = dx + 1;
		} else {
			dx = dx - 1;
		}
		if( (dy = e.y - s.y) > 0 ){
			dy = dy + 1;
		} else {
			dy = dy - 1;
		}
		a = (double)dx / (double)dy;
		if( a < 1.0 && a > -1.0 ){
			paint_line = paint_hline;
		} else {
			paint_line = paint_vline;
		}

		if( s.y > e.y ){
			min = e.y;
			off = e.x;
			omax = s.x;
			max = s.y;
		} else {
			min = s.y;
			off = s.x;
			omax = e.x;
			max = e.y;
		}
		p = &data[min * Width];
		d = 0;
		for( i = min; i <= max; i++ ){
			dn = round(a * (double)(i - min + 1));
			if( d > dn ){
				for( j = d; j > dn && off-j >= omax; j-- ){
					(*paint_line)( p, off+j, i );
				}
			} if( d < dn ){
				for( j = d; j < dn && off+j <= omax ; j++ ){
					(*paint_line)( p, off+j, i );
				}
			} else {
				(*paint_line)( p, off+d, i );
			}
			d = dn;
			p += Width;
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function paint_fineline()                              */
/*                                                                    */
/*             unsigned int *data : image data top                    */
/*             double x        : center position X                    */
/*             double y        : center position Y                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Paint fine line by drawing a circle.         */
/*                                                                    */
/**********************************************************************/
void paint_fineline( unsigned int *data, double x, double y )
{
	int i, j;
	double dx, dy;

	for( i = max((int)y-HB,0); i <= min((int)y+HB,Height); i++ ){   
		for( j = max((int)x-HB,0); j <= min((int)x+HB,Width); j++ ){
			dx = (double)j - x; dy = (double)i - y;
			if( dx * dx + dy * dy <= (double)(HB * HB) ){
				CLRMASK( data[i*Width+j] );
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_fineline()                                */
/*                                                                    */
/*             unsigned int *data  : image data                       */
/*             struct Pos s        : start position                   */
/*             struct Pos e        : end position                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Fine line main for each line.                */
/*                                                                    */
/**********************************************************************/
void set_fineline( unsigned int *data, struct Pos s, struct Pos e )
{
	int i;
	int min, max, off, omax;
	double dn;
	double a;

	HB = Border / 2;
	if( s.x == e.x ){
		min = min( s.y, e.y );
		max = max( s.y, e.y );
		for( i = min; i <= max; i++ ){
			paint_fineline( data, (double)s.x, i );
		}
	} else if( s.y == e.y ){
		min = min( s.x, e.x );
		max = max( s.x, e.x );
		for( i = min; i <= max; i++ ){
			paint_fineline( data, (double)i, s.y );
		}
	} else {
		a = (double)(e.x - s.x) / (double)(e.y - s.y);

		if( s.y > e.y ){
			min = e.y;
			off = e.x;
			omax = s.x;
			max = s.y;
		} else {
			min = s.y;
			off = s.x;
			omax = e.x;
			max = e.y;
		}
		if( a < 1.0 && a > -1.0 ){
			for( i = min; i <= max; i++ ){
				dn = a * (double)(i - min + 1) + (double)off;
				paint_fineline( data, dn, (double)i );
			}
		} else {
			if( off < omax ){
				for( i = off; i <= omax; i++ ){
					dn = 1.0 / a * (double)(i - off + 1) + (double)min;
					paint_fineline( data, (double)i, dn );
				}
			} else {
				for( i = omax; i <= off; i++ ){
					dn = 1.0 / a * (double)(i - off + 1) + (double)min;
					paint_fineline( data, (double)i, dn );
				}
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *data, *p;
	int i;
	struct fPosition pos[MAXSP];
	struct fPosition *ret;
	int total;
	struct Pos x0, x1;

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	data = file->buf->data;

	if( Mode == LINE  ){
		if( sb->selnum != SKLINES ){
			return( EILLSELECTION );
		}
	} else if( Mode == SPLINE ){
		if( sb->selnum != SKSPLINE ){
			return( EILLSELECTION );
		}
	} else {
		return( EILLSELECTION );
	}
	check_range( sb, Width, Height, 0, ADJUST );
	if( sb->defnum < 2 ){
		return( EILLNUMBER );
	}

	for( i = 0, p = data; i < Height * Width; i++, p++ ){
		SETMASK(*p);
	}
	if( Mode == LINE ){
	        for( i = 0; i < sb->defnum - 1; i++ ){
			if( Fine == FINELINE ){
				set_fineline( data, sb->def[i], sb->def[i+1] );
			} else {
				set_line( data, sb->def[i], sb->def[i+1] );
			}
		}
	} else if( Mode == SPLINE ){
		for( i = 0; i < sb->defnum; i++ ){
			pos[i].x = (float)sb->def[i].x;
			pos[i].y = (float)sb->def[i].y;
		}
		total = (sb->defnum-1) * 10;

        	if( (ret = spline_line( pos, 3, sb->defnum, 10 )) == NULL ){
			return( -1 );
		}
		x0.x = (int)pos[0].x; x0.y = (int)pos[0].y;
        	for( i = 0; i < total; i++ ){
			x1.x = (int)ret[i].x; x1.y = (int)ret[i].y;
			if( Fine == FINELINE ){
				set_fineline( data, x0, x1 );
			} else {
				set_line( data, x0, x1 );
			}
			x0.x = x1.x; x0.y = x1.y;
		}
	}
	file->buf->mask_kind = SIMPLEMASK;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/*             A0 rewrite                                             */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
	return( 0 );
}
