/**********************************************************************/
/*                                                                    */
/*           loader.c   gedo LOADER main routine                      */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <setjmp.h>
#include "plug.h"

extern char *Spec;			/* plug name */
extern char *Version;			/* version of plug */
extern char *Comment;			/* description of plug */

extern PLUGITEM Item[];		        /* plug config for init items */
extern int ItemNum;			/* number of init item */
extern int LSB_flag;			/* check the machine is LSB or MSB */
extern struct Tags Tag[];		/* plug interface items */
extern int TagNum;			/* number of Tag */

PLUGINFO PIf;				/* plug description item */

#define MAXVIEWER	10		/* maximum of viewer(2) process */
struct File File[MAXVIEWER];		/* viewer struct */
int FileNum = 0;			/* number of viewers */
int Init_flag;				/* called from usage() */
jmp_buf jmpbuf;				/* ctrl+C jump buffer */
extern int Gui;

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file : process infomation                 */
/*                                                                    */
/*             return int : 0==success -1==unknown error              */
/*                          0> error code                             */
/*             function: dummy interface                              */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	return( -1 );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file : process infomation                 */
/*                                                                    */
/*             return int : 0==success -1==unknown error              */
/*                          0> error code                             */
/*             function: dummy interface                              */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	return( -1 );
}

/**********************************************************************/
/*                                                                    */
/*             function line_config()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success 1==version or status            */
/*                          else error                                */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int line_config( struct File *file )
{
	char **com;
	int ret, i;
	unsigned char *str;
	PLUGITEM *pi;
	int num;
	unsigned int mask;
	int flag;

	str = file->mes->u.data;
	pi = Item;
	num = ItemNum;
	if( file->mes->command == CLOAD ){
		mask = NOLOAD;
		flag = CONFIG_LOADER;
	} else if( file->mes->command == CSAVE ){
		mask = NOSAVE;
		flag = CONFIG_SAVER;
	} else {
		return( ENOCOMMAND );
	}
	if( check_itemnum( flag, Item, ItemNum ) == 0 ){
		fprintf( stderr, "\tno config parameters\n" );
		sprintf( str, "no config" );
		return( 0 );
	}

	com = command_line( str );
	while( *com ){
		if( strcmp( *com, "version" ) == 0 ){
			strcpy( str, Version );
			printf( "%s\n", str );
			return( GETSTATUS );
		} else if( strcmp( *com, "status" ) == 0 ){
			status( flag, str );	
			printf( "%s\n", str );
			return( GETSTATUS );
		}
		ret = 0;
		for( i = 0; i < num; i++ ){
			if( mask & pi[i].type ){
				continue;
			}
			ret = check_item( *com, &pi[i] );
			if( ret == -1 ){
				fprintf( stderr, "check_item error %s\n", *com );
				return( -1 );
			}
			if( ret ){
				break;
			}
		}
		if( ret == 0 ){
			fprintf( stderr, "WARNING:unknown item %s\n", *com );
		}
		com++;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config_comline()                              */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success 1==version or status            */
/*                          else error                                */
/*             function:CONFIGURE the PLUG at Transaction INTERCTIVE  */
/*                                                                    */
/**********************************************************************/
int config_comline( struct File *file )
{
	char buff[512];
	int i, ret, num;
	int first;
	unsigned int mask;
	int flag;

	switch( file->mes->command ){
	case CLOAD:
		mask = INITONLY | NOLOAD;
		flag = CONFIG_LOADER;
		break;
	case CSAVE:
		mask = INITONLY | NOSAVE;
		flag = CONFIG_SAVER;
		break;
	case CSELECT:
	case CEFFECT:
		flag = CONFIG_ELSE;
		mask = INITONLY;
		break;
	default:
		return( -1 );
	}
	if( check_itemnum( get_commandspec( file ), Item, ItemNum ) == 0 ){
		strcpy( file->mes->u.data , "no config" );
		return( 0 );
	}
	printf( "%s: configer\n", Spec );
	status( flag, buff );
	printf( "\tcurrent : %s\n", buff );
	ret = new_config_input( mask, Tag, TagNum, Item, &num );
	if( ret == USRBREAK ){
		return( EUSRBREAK );
	} else if( ret == ERROR ){
		return( EILLCONFIG );
	}
	first = 1;
	for( i = 0; i < num; i++ ){
		if( Tag[i].status == DONE ){
			switch( Item[Tag[i].num].type & COMMANDMASK ){
			case GETSTR:
				sprintf( buff, Tag[i].result, Item[Tag[i].num].set_target );
				break;
			case GETINT:
			case GETHEX:
				sprintf( buff, Tag[i].result, *(int *)(Item[Tag[i].num].set_target) );
				break;
			case GETFLOAT:
				sprintf( buff, Tag[i].result, *(float *)(Item[Tag[i].num].set_target) );
				break;
			case SWITCH:
				sprintf( buff, Tag[i].result );
				break;
			default:
				continue;
				break;
			}
			if( first ){
				strcpy( file->mes->u.data, buff );
				first = 0;
			} else {
				strcat( file->mes->u.data, "," );
				strcat( file->mes->u.data, buff );
			}
		}
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config_use_gui()                              */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success 1==version or status            */
/*                          else error                                */
/*             function:CONFIGURE the PLUG at Transaction GUI         */
/*                                                                    */
/**********************************************************************/
int config_use_gui( struct File *file )
{
	FILE *fp;
	FILE *fpp;
	char buff[256];
	char fifo[256];
	unsigned int mask;

	switch( file->mes->command ){
	case CLOAD:
		mask = INITONLY | NOLOAD;
		break;
	case CSAVE:
		mask = INITONLY | NOSAVE;
		break;
	case CSELECT:
	case CEFFECT:
		mask = INITONLY;
		break;
	default:
		mask = 0xffffffff;
		break;
	}
	mask |= NOTCL;
	if( check_itemnum( get_commandspec( file ), Item, ItemNum ) == 0 ){
		strcpy( file->mes->u.data , "no config" );
		return( 0 );
	}
	if( (fp = create_fifo( fifo )) == NULL ){
		return( -1 );
	}
	sprintf( buff, "wish <%s &", fifo );
	if( (fpp = popen( buff, "r" )) == NULL ){
		return( -1 );
	}
	output_tclcommand( mask, fp, Tag, TagNum, Item, &PIf );
	fclose( fp );
	
	fgets( buff, sizeof(buff)-1, fpp );
	pclose( fpp );

	if( strcmp( buff, "abort\n" ) == 0 ){
		return( EUSRBREAK );
	}
	buff[strlen(buff) - 1] = '\0';
	strcpy( file->mes->u.data, buff );
	return( line_config( file ) );
}

/**********************************************************************/
/*                                                                    */
/*             function get_default_status()                          */
/*                                                                    */
/*             int kind  : configure kind                             */
/*             char *buff: must set string                            */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: DEFAULT status() calles this.                */
/*                                                                    */
/**********************************************************************/
void get_default_status( int kind, char *buff )
{
	int i;
	unsigned int mask;
	int first;
	char work[256];

	first = 1;
	switch( kind ){
	case CONFIG_LOADER:
		mask = INITONLY | NOLOAD;
		break;
	case CONFIG_SAVER:
		mask = INITONLY | NOSAVE;
		break;
	case CONFIG_ELSE:
		mask = INITONLY;
		break;
	case CONFIG_INIT:
		mask = NOINIT;
		break;
	default:
		return;
		break;
	}
	for( i = 0; i < TagNum; i++ ){
		if( Tag[i].kind & mask ){
			continue;
		}
		if( each_default_status( i, Tag, Item, work ) ){
			if( first ){
				strcpy( buff, work );
				first = 0;
			} else {
				strcat( buff, "," );
				strcat( buff, work );
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function help()                                        */
/*                                                                    */
/*             args NONE                                              */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: HELP Messages at Transaction HELP            */
/*                                                                    */
/**********************************************************************/
void help( void )
{
	printf( "%s\n%s %s \"params,..\": %s\n", PIf.version, StrServer[PIf.kind],
		PIf.plugname, PIf.comment );
	printf( "loader:configration parameters\n" );
	if( check_itemnum( CONFIG_LOADER, Item, ItemNum ) == 0 ){
		display_config( CONFIG_LOADER, NULL, 0 );
	} else {
		display_config( CONFIG_LOADER, Item, ItemNum );
	}
	printf( "saver:configration parameters\n" );
	if( check_itemnum( CONFIG_SAVER, Item, ItemNum ) == 0 ){
		display_config( CONFIG_SAVER, NULL, 0 );
	} else {
		display_config( CONFIG_SAVER, Item, ItemNum );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function usage()                                       */
/*                                                                    */
/*             args NONE                                              */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: usage from command line --help               */
/*                                                                    */
/**********************************************************************/
static void usage( void )
{
	printf( "%s\n%s [options]: %s\n", PIf.version, PIf.filename, PIf.comment );
	if( check_itemnum( 1, Item, ItemNum ) == 0 ){
		display_config( CONFIG_INIT, NULL, 0 );
	} else {
		display_config( CONFIG_INIT, Item, ItemNum );
	}
	exit( 1 );
}

/**********************************************************************/
/*                                                                    */
/*             function init()                                        */
/*                                                                    */
/*             int argc   : argc recieved GEDO at initialize          */
/*             char **argv: argv recieved GEDO at initialize          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: initilize the PLUG                           */
/*                                                                    */
/**********************************************************************/
int init( int argc, char **argv )
{
	int ret, i;
	char *ag;

	unsigned char test[4];
	unsigned short *s;

	test[0] = 0x01; test[1] = 0x00;
	s = (unsigned short *)test;
	if( *s == 0x0001 ){
		LSB_flag = 1;
	} else if( *s == 0x0100 ){
		LSB_flag = 0;
	} else {
		fprintf( stderr, "MIDDLE ENDIAN, how to handle it!\n" );
		exit( 1 );
	}

	++argv; Init_flag = 1;
	while( --argc ){
		if( strncmp( *argv, "--help", 6 ) == 0 ){
			usage();
		}
		ret = 0;
		if( **argv == '-' ){
			ag = *argv + 1;
		} else {
			ag = *argv;
		}
		for( i = 0; i < ItemNum; i++ ){
			if( Item[i].type & NOINIT ){
				continue;
			}
			ret = check_item( ag, &Item[i] );
			if( ret == -1 ){
				fprintf( stderr, "check_item error %s\n", *argv );
			}
			if( ret ){
				break;
			}
		}
		if( ret == 0 ){
			fprintf( stderr, "unknown item %s\n", *argv );
		}
		argv++;
	}
	Init_flag = 0;
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             int argc : argment called from GEDO                    */
/*             char **argv : argment called from GEDO                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Whole main routine of LOADER/SAVER plug.     */
/*                       Defined init(), process(), config(), help()  */
/*                      read_header(), read_image(), write_header()   */
/*                      and write_image() defined at each plug file.  */
/**********************************************************************/
void main( int argc, char **argv )
{
	struct Message mes;
	struct proc proc;
	char *e;
 
	proc.spec = Spec;
	proc.file = &File[FileNum];
	File[FileNum].mes  = &mes;
	proc.pid = getpid();
	proc.kind = SLOADER;

/* add C3 for set all GUI mode */
	if( (e = getenv( "GEDO_GUI" )) && *e ){
		Gui = 1;
	}

	PIf.kind     = SLOADER;
	PIf.filename = argv[0];
	PIf.plugname = Spec;
	PIf.version  = Version;
	PIf.comment  = Comment;
	PIf.item_num = ItemNum;
	PIf.item = Item;

	if( init( argc, argv ) == -1 ){
		printf( "illegal initilization of %s\n", proc.spec );
	}
	process( &proc );
}



