/**********************************************************************/
/*                                                                    */
/*              lzw.c   gedo PLUG lzw library                         */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "lzw.h"

typedef int STRING;
#define TABLEMAX   8192			/* maximum of Table */
#define STRINGMAX (64 * 256)		/* string buffer alloc unit */

struct Table{				/* Binary Tree */
	STRING name;			/* word of dictionary */
	int len;			/* length of name */
	int left;			/* left node */
	int right;			/* right node */
};
static struct Table Table[TABLEMAX];	/* Dictionary table */

#define HASHSIZE 257			/* hash size */
static int Hash[HASHSIZE];		/* hash top */

static int NowTable = 0;		/* used table */
static unsigned char *StringBuff = NULL;	/* string buffer */
static int Stringp = 0;			/* used string buffer */
static int StringMax;			/* now maximum of string buffer */


/* for original lzw.c */
typedef unsigned short LZSTR;
static unsigned char *At;		/* now reading in-buffer */
static unsigned char *Top;		/* top of in-buffer */
static int  Sft;			/* shift count of At */
static int  Unit;			/* LZW now reading unit */

static int StartUnit;			/* LZW start unit */
static int StartTable;			/* start table num, ClrCode+2 */
static int ClrCode;			/* LZW ClrCode, Starts[StartUnit] */
static int EofCode;			/* LZW EOF Code, ClrCode+1 */
static int Starts[15]			/* ClrCode by StartUnit */ 
 = { 0, 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192 }; 
static int TiffGif;			/* TIFF type or GIF type */
/* TIFF type --> read from LSB to MSB, Table full detect at Starts[Unit] */ 
/* GIF type  --> read from MSB to LSB, Table full detect at Starts[Unit]-1 */ 

/**********************************************************************/
/*                                                                    */
/*            function buff_strdup()                                  */
/*                                                                    */
/*            args  unsigned char *name  : word of dictionary         */
/*                  int len   :  length of name                       */
/*                                                                    */
/*            return  STRING   :  word in StringBuff                  */
/*                                                                    */
/*            function: Copy name to the StringBuff.                  */
/*                                                                    */
/**********************************************************************/
static STRING buff_strdup( unsigned char *name, int len )
{
	STRING ret;
	unsigned char *tmp;

	if( Stringp + len >= StringMax ){
		if( (tmp = (unsigned char *)realloc( StringBuff, StringMax + STRINGMAX )) == NULL ){ 
			StringBuff = NULL;
			fprintf( stderr, "string buff overflow\n" );
			return( (STRING)-1 );
		}
		StringBuff = tmp;
		StringMax += STRINGMAX;
	}
	ret = Stringp;
	memcpy( &StringBuff[Stringp], name, len );
	Stringp += len;
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function lzw_init_tree()                                */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return int : 0==success   -1==fail                      */
/*                                                                    */
/*            function: Initilize the Binary Tree.                    */
/*                                                                    */
/**********************************************************************/
static int lzw_init_tree(void)
{
	int i;

	if( StringBuff ){
		free( StringBuff );
	}
	if( (StringBuff = malloc( STRINGMAX )) == NULL ){
		return( -1 );
	}
	StringMax = STRINGMAX;
	Stringp = 0;
	NowTable = 0;
	memset( Table, 0, sizeof(struct Table) * TABLEMAX );
	memset( StringBuff, 0, sizeof(StringBuff) );
	for( i = 0; i < HASHSIZE; i++ ){
		Hash[i] = -1;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function get_hash()                                     */
/*                                                                    */
/*            args  unsigned char *s  : word of dictionary            */
/*                  int len   :  length of name                       */
/*                                                                    */
/*            return  int   : hash value                              */
/*                                                                    */
/*            function: get hash value from s.                        */
/*                                                                    */
/**********************************************************************/
static int get_hash( unsigned char *s, int len )
{
	int ret, i;

	for( ret = 0, i = 0; i < len; i++ ){
		ret += (int)s[i];
	}
	return( (ret+len) % HASHSIZE );
}

/**********************************************************************/
/*                                                                    */
/*            function str_cmp()                                      */
/*                                                                    */
/*            args  STRING name1  : word in the dictionary            */
/*                  unsigned char *name2  : search name               */
/*                  int len   :  length of name2                      */
/*                                                                    */
/*            return  int   : 0==equal  else==different               */
/*                                                                    */
/*            function: Compare name1(in dict) and name2(from in).    */
/*                                                                    */
/**********************************************************************/
static int str_cmp( STRING name1, unsigned char *name2, int len )
{
	return( memcmp( &StringBuff[name1], name2, len ) );
}

/**********************************************************************/
/*                                                                    */
/*            function lzw_search_tree()                              */
/*                                                                    */
/*            args  unsigned char *name  : search name                */
/*                  int len   :  length of name                       */
/*                                                                    */
/*            return int   : -1==not here  else TableNo.              */
/*                                                                    */
/*            function: Search name in the dictionary.                */
/*                                                                    */
/**********************************************************************/
static int lzw_search_tree( unsigned char *name, int len )
{
	int v;
	int at;
	int cmp;

	v = get_hash( name, len );
	if( Hash[v] == -1 ){
		return( -1 );
	}
	for( at = Hash[v]; at != -1; ){
		if( Table[at].len < len ){
			at = Table[at].left;
		} else if( Table[at].len == len ){
			cmp = str_cmp( Table[at].name, name, len ) ;
			if( cmp == 0 ){
				return( at );
			} else if( cmp < 0 ){
				at = Table[at].left;
			} else {
				at = Table[at].right;
			}
		} else {
			at = Table[at].right;
		}
	}
	return( -1 );
}


/**********************************************************************/
/*                                                                    */
/*            function get_newentry()                                 */
/*                                                                    */
/*            args  unsigned char *name  : new word                   */
/*                  int len   :  length of name                       */
/*                                                                    */
/*            return  int   : new entry of the Table                  */
/*                                                                    */
/*            function:  Make new Table entry.                        */
/*                                                                    */
/**********************************************************************/
static int get_newentry( unsigned char *name, int len )
{
	int ret;

	Table[NowTable].name = buff_strdup( name, len );
	Table[NowTable].len = len;
	Table[NowTable].left = -1;
	Table[NowTable].right = -1;
	ret = NowTable;

	if( Table[ret].name == -1 ){
		return( -1 );
	}

	if( ++NowTable >= TABLEMAX ){
		fprintf( stderr, "table overflow\n" );
		return( -1 );
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*            function lzw_add_tree()                                 */
/*                                                                    */
/*            args  unsigned char *name  : word of dictionary         */
/*                  int len   :  length of name                       */
/*                                                                    */
/*            return int   : NowTable(Now Used Table)                 */
/*                           not the added Table!                     */
/*            function:  Add the name to the dictinary.               */
/*                                                                    */
/**********************************************************************/
int lzw_add_tree( unsigned char *name, int len )
{
	int at, prev;
	int v, new;
	int cmp;

	cmp = 0;
	v = get_hash( name, len );
	if( Hash[v] == -1 ){
		new = get_newentry( name, len );
		Hash[v] = new;
		return( NowTable );
	}
	prev = -1;
	for( at = Hash[v]; at != -1; ){
		prev = at;
		if( Table[at].len  < len ){
			at = Table[at].left;
		} else if( Table[at].len == len ){
			cmp = str_cmp( Table[at].name, name, len );
			if( cmp == 0 ){
				return( NowTable );
			} else if( cmp < 0 ){
				at = Table[at].left;
			} else {
				at = Table[at].right;
			}
		} else {
			at = Table[at].right;
		}
	}
	new = get_newentry( name, len );
	if( prev == -1 ){
		at = Hash[v];
	} else {
		at = prev;
	}
	if( Table[at].len == len ){
		if( cmp < 0 ){
			Table[new].left = Table[at].left;
			Table[at].left = new;
		} else {
			Table[new].right = Table[at].right;
			Table[at].right = new;
		}
	} else if( Table[at].len < len ){
		Table[new].left = Table[at].left;
		Table[at].left = new;
	} else {
		Table[new].right = Table[at].right;
		Table[at].right = new;
	}
	return( NowTable );
}


/**********************************************************************/
/*                                                                    */
/*            function lzw_init_table()                               */
/*                                                                    */
/*            args  unsigned char *out  : Output buffer               */
/*                  int start   :  start unit of input                */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: Initilize the encoder.                        */
/*                                                                    */
/**********************************************************************/
static void lzw_init_table( unsigned char *out, int start )
{
	Top = At = out;
	*At = 0;
	Sft = 0;
	Unit = StartUnit = start;
	StartTable = Starts[StartUnit] + 2;
	ClrCode = StartTable - 2;
	EofCode = StartTable - 1;
	lzw_init_tree();
}

/**********************************************************************/
/*                                                                    */
/*            function lzw_write_byte()                               */
/*                                                                    */
/*            args  int v  :  0==bit OFF  1==bit ON                   */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: Flush write the bit by BYTE.                  */
/*                                                                    */
/**********************************************************************/
static void lzw_write_byte( int v )
{
	static unsigned char writemask[8] = 
		{ 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };

	if( v ){
		if( TiffGif == TIFF ){
			*At |= writemask[7 - Sft];
		} else {
			*At |= writemask[Sft];
		}
	}
	if( ++Sft == 8 ){
		Sft = 0;
		*++At = 0;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function lzw_write_code()                               */
/*                                                                    */
/*            args  LZSTR c : Write Code                              */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function:  Write the Code c to output buffer.           */
/*                                                                    */
/**********************************************************************/
static void lzw_write_code( LZSTR c )
{
	int i;
	static LZSTR bitmask[16] = 
            { 0x8000, 0x4000, 0x2000, 0x1000, 0x0800, 0x0400, 0x0200, 0x0100,
              0x0080, 0x0040, 0x0020, 0x0010, 0x0008, 0x0004, 0x0002, 0x0001};

	if( TiffGif == TIFF ){
		for( i = 16 - Unit; i < 16; i++ ){
			lzw_write_byte( c & bitmask[i] );
		}
	} else {
		for( i = 15; i >= 16 - Unit; i-- ){
			lzw_write_byte( c & bitmask[i] );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*            function lzw_search_table()                             */
/*                                                                    */
/*            args  unsigned char *name  : search name                */
/*                  int len   :  length of name                       */
/*                                                                    */
/*            return  LZSTR   : 0xffff==not found                     */
/*                              else==found TableNo.                  */
/*            function: Search the Dictionary by name.                */
/*                                                                    */
/**********************************************************************/
static LZSTR lzw_search_table( unsigned char *name, int len )
{
	int ret;

	if( len == 1 ){
		return( (LZSTR)name[0] );
	} else {
		ret = lzw_search_tree( name, len );
		if( ret == -1 ){
			return( 0xffff );
		} else {
			return( (LZSTR)(ret + StartTable) );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*            function lzw_add_table()                                */
/*                                                                    */
/*            args  unsigned char *name  : new name                   */
/*                  int len   :  length of name                       */
/*                                                                    */
/*            return  int   :  0==success  -1==fail                   */
/*                                                                    */
/*            function: Add new entry of dictionary by name.          */
/*                                                                    */
/**********************************************************************/
static int lzw_add_table( unsigned char *name, int len )
{
	int ret;
	int cmp;

	ret = lzw_add_tree( name, len );
	if( ret == -1 ){
		return( -1 );
	}
	if( TiffGif == TIFF ){
		cmp = ret + StartTable;
	} else {
		cmp = ret + StartTable -1;
	}
	switch( cmp ){
	case 8:
	case 16:
	case 32:
	case 64:
	case 128:
	case 256:
	case 512:
	case 1024:
	case 2048:
		Unit++;
		break;
	case 4096:
		lzw_write_code( ClrCode );
		Unit = StartUnit;
		StartTable = Starts[StartUnit] + 2;
		lzw_init_tree();
		return( 0 );
		break;
	}
	return( 0 );
}


#define INTERMAX  500
/**********************************************************************/
/*                                                                    */
/*            function compress()                                     */
/*                                                                    */
/*            args  int start  : start unit                           */
/*                  unsigned char *in  : input buffer                 */
/*                  unsigned char *out : output buffer                */
/*                  int size   :  size of input                       */
/*                                                                    */
/*            return int   :  0==success  -1==error                   */
/*                                                                    */
/*            function: LZW-Compress <in> and write <out>.            */
/*                                                                    */
/**********************************************************************/
int compress( int start, unsigned char *in, unsigned char *out, int size, int kind )
{
	int i;
	unsigned char tmpstr[INTERMAX];
	unsigned char conc[INTERMAX];
	int concnum, tmpnum;
	unsigned char k;
	int ret;

	if( kind == TIFF ){
		TiffGif = TIFF;
	} else if( kind == GIF ){
		TiffGif = GIF;
	}
	ret = 0;
	lzw_init_table( out, start );
	lzw_write_code( ClrCode );
	concnum = tmpnum = 0;
	for( i = 0; i < size; i++ ){
		if( concnum >= INTERMAX || tmpnum >= INTERMAX ){
			fprintf( stderr, "internal string overflow\n" );
			ret = -1;
			goto ex;
		}
		k = in[i];
		conc[concnum++] =  k;
		switch( lzw_search_table( conc, concnum ) ){
		case 0xfffe:
			ret = -1;
			goto ex;
			break;
		case 0xffff:
			lzw_write_code( lzw_search_table( tmpstr, tmpnum ) );
			if( lzw_add_table( conc, concnum ) != 0 ){
				ret = -1;
				goto ex;
			}
			tmpstr[0] = conc[0] = k;
			tmpnum = concnum = 1;
			break;
		default:
			tmpstr[tmpnum++] = k;
			break;
		}
	}
	lzw_write_code( lzw_search_table( tmpstr, tmpnum ) );
	if( lzw_add_table( conc, concnum ) != 0 ){
		ret = -1;
		goto ex;
	}
	lzw_write_code( EofCode );
ex:	if( StringBuff ){
		free( StringBuff );
	}
	StringBuff = NULL;
	if( ret != 0 ){
		return( ret );
	} else {
		if( Sft == 0 ){
			return( At - Top );
		} else {
			return( At - Top + 1 );
		}
	}
}
