/**********************************************************************/
/*                                                                    */
/*            matfile.c  gedo PLUG effmatrix library routine          */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <memory.h>
#include "plug.h"
#include "matrix.h"

struct FileMatrix {		/* file data format */
	char name[16];		/* name of matrix */
	float data[9];		/* matrix data */
};
extern char *Spec;

/**********************************************************************/
/*                                                                    */
/*            function  open_matfile()                                */
/*                                                                    */
/*            args  char *path  : path of matrix data file            */
/*                  char *flag  : read mode                           */
/*                                                                    */
/*            return  FILE * :  FILE pointer                          */
/*                                                                    */
/*            function: Open the matrix data file in <path>.          */
/*                                                                    */
/**********************************************************************/
static FILE *open_matfile( char *path, char *flag  )
{
	char file[256];
	int len;

	strcpy( file, path );
	len = strlen(file);
	if( file[len-1] != '/' ){
		strcat( file, "/" );
		len++;
	}
	sprintf( &file[len], "%s.dat", Spec );
	return( fopen( file, flag ) );
}

/**********************************************************************/
/*                                                                    */
/*            function search_each_matrix()                           */
/*                                                                    */
/*            args  char *path : path matrix data file in             */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: For each path, search all matrix data in it.  */
/*                                                                    */
/**********************************************************************/
static void search_each_matrix( char *path )
{
	FILE *fp;
	struct FileMatrix buff;
	int count;

	if( (fp = open_matfile( path, "rb" )) ){
		fprintf( stderr, "%s in %s\n", Spec, path ); count = 0;
		while( fread( &buff, sizeof(struct FileMatrix), 1, fp ) ){
			if( count == 5 ){
				fprintf( stderr, "\t%s\n", buff.name );
				count = 0;
			} else {
				fprintf( stderr, "\t%s", buff.name );
			}
		}
		fprintf( stderr, "\n" );
		fclose( fp );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function search_matrix()                                */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Search available name of matrix in all path.  */
/*                                                                    */
/**********************************************************************/
void search_matrix( void )
{
	char *path;

	path =  "./";
	search_each_matrix( path );
	if( (path = getenv( "GEDO_LWP" )) != NULL ){
		search_each_matrix( path );
	}
	if( (path = getenv( "GEDO_GWP" )) != NULL ){
		search_each_matrix( path );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function find_name()                                    */
/*                                                                    */
/*            args  char *path  : path the matrix data file in        */
/*                  char *name  : must be search name                 */
/*                  struct FileMatrix *buf : read buffer              */
/*                                                                    */
/*            return  int   :  0==found  -1==not found                */
/*                                                                    */
/*            function: Search matrix by name in <path>/matrix.dat    */
/*                                                                    */
/**********************************************************************/
static int find_name( char *path, char *name, struct FileMatrix *buf )
{
	FILE *fp;
	int ret;
	struct FileMatrix readb;

	if( (fp = open_matfile( path, "rb" )) == NULL ){
		return( -1 );
	}
	ret = -1;
	while( fread( &readb, sizeof(struct FileMatrix), 1, fp ) ){
		if( strcmp( readb.name, name ) == 0 ){
			ret = 0;
			memcpy( buf, &readb, sizeof(struct FileMatrix) );
		}
	}
	fclose( fp );
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*            function load_matrix()                                  */
/*                                                                    */
/*            args  char *name  : matrix name                         */
/*                  float mat[][]: read buffer of the matrix          */
/*                  char *matstr: read buffer express string of matrix*/
/*                                                                    */
/*            return  int  : 0==found it,  else==not found            */
/*                                                                    */
/*            function: Search and Load the named matrix by the name. */
/*                                                                    */
/**********************************************************************/
int load_matrix( char *name, float mat[3][3], char *matstr )
{
	struct FileMatrix buff;
	int i, j;
	char tmp[256];

	if( find_name( getenv( "GEDO_GWP" ), name, &buff ) == -1 ){
		if( find_name( getenv( "GEDO_LWP" ) , name, &buff ) == -1 ){
			if( find_name( "./", name, &buff ) == -1 ){
				printf( "cannot find %s\n", name );
				return( EOPENFILE );
			}
		}
	}
	for( i = 0; i < 3; i++ ){
		for( j = 0; j < 3; j++ ){
			mat[i][j] = buff.data[i * 3 + j];
		}
	}
	matstr[0] = '\0';
	for( i = 0; i < 8; i++ ){
		sprintf( tmp, "%g:", buff.data[i] );
		strcat( matstr, tmp );
	}
	sprintf( tmp, "%g", buff.data[i] );
	strcat( matstr, tmp );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function save_matrix()                                  */
/*                                                                    */
/*            args  char *name  :  name the matrix must be save       */
/*                  float mat[][]: the matrix                         */
/*                                                                    */
/*            return  int : 0==success   else==fail                   */
/*                                                                    */
/*            function: Save the matrix by name to LWP/matrix.dat.    */
/*                                                                    */
/**********************************************************************/
int save_matrix( char *name, float mat[3][3] )
{
	FILE *fp;
	long pos;
	long ret;
	struct FileMatrix readb;
	int i, j;

	if( (fp = open_matfile( getenv( "GEDO_LWP" ), "r+b" )) == NULL ){
		if( (fp = open_matfile( getenv( "GEDO_LWP" ), "w+b" )) == NULL ){
			return( EOPENFILE );
		}
	}
	ret = -1L;
	while( 1 ){
		pos = ftell( fp );
		if( fread( &readb, sizeof(struct FileMatrix), 1, fp ) <= 0 ){
			break;
		}
		if( strcmp( readb.name, name ) == 0 ){
			ret = pos;
		}
	}
	strcpy( readb.name, name );
	for( i = 0; i < 3; i++ ){
		for( j = 0; j < 3; j++ ){
			readb.data[i * 3 + j] = mat[i][j];
		}
	}
	if( ret != -1L ){
		if( fseek( fp, ret, SEEK_SET ) == -1L ){
			return( ESEEK );
		}
	}
	if( fwrite( &readb, sizeof(struct FileMatrix), 1, fp ) != 1 ){
		return( EWRITEFILE );
	}
	fclose( fp );
	return( 0 );
}
