/**********************************************************************/
/*                                                                    */
/*                matgui.c   gedo PLUG specified routine              */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <stdio.h>


struct FileMatrix {		/* file data format */
	char name[16];		/* name of matrix */
	float data[9];		/* matrix data */
};

struct FM {			/* file matrix data */
	struct FileMatrix M;	/* each matrix */
	int kind;		/* not used */
	struct FM *next;	/* next matrix */
};

int FMnum = 0;		/* not used */
struct FM *top = NULL;	/* top of list */
struct FM *at;		/* now list */
char *Spec;

/**********************************************************************/
/*                                                                    */
/*            function *open_matfile()                                */
/*                                                                    */
/*            args  char *path  : matrix file path                    */
/*                  char *flag  : open mode                           */
/*                                                                    */
/*            return  FILE   :  FILE pointer                          */
/*                                                                    */
/*            function: open the matrix data file.                    */
/*                                                                    */
/**********************************************************************/
static FILE *open_matfile( char *path, char *flag  )
{
	char file[256];
	int len;

	strcpy( file, path );
	len = strlen(file);
	if( file[len-1] != '/' ){
		strcat( file, "/" );
		len++;
	}
	sprintf( &file[len], "%s.dat", Spec );
	return( fopen( file, flag ) );
}

/**********************************************************************/
/*                                                                    */
/*            function find_matrix()                                  */
/*                                                                    */
/*            args  char *file : file pointer of the matrix data file */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function:  read the matrix data.                        */
/*                                                                    */
/**********************************************************************/
void find_matrix( char *file )
{
	FILE *fp;
	struct FileMatrix buff;
	struct FM *fm;

	if( (fp = open_matfile( file, "rb" )) ){
		while( fread( &buff, sizeof(struct FileMatrix), 1, fp ) ){
			if( (fm = malloc(sizeof(struct FM))) == NULL ){
				fprintf( stderr, "cannot malloc\n" );
				return;
			}
			memcpy( &fm->M, &buff, sizeof(struct FileMatrix) ); 
			fm->next = top;
			top = fm;
		}
		fclose( fp );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function findfirst_matrix()                             */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: prepare to search matrix by name.             */
/*                                                                    */
/**********************************************************************/
void findfirst_matrix( void )
{
	char *path;

	if( (path = getenv( "GEDO_GWP" )) != NULL ){
		find_matrix( path );
	}
	if( (path = getenv( "GEDO_LWP" )) != NULL ){
		find_matrix( path );
	}
	path = "./";
	find_matrix( path );
	at = top;
}

/**********************************************************************/
/*                                                                    */
/*            function findnext_matrix()                              */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return char * : matrix name                             */
/*                                                                    */
/*            function: search matrix                                 */
/*                                                                    */
/**********************************************************************/
char * findnext_matrix( void )
{
	char *ret;

	if( at == NULL ){
		return( NULL );
	}
	ret = at->M.name;
	at = at->next;
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function *search_data()                                 */
/*                                                                    */
/*            args  char *name   :  name of matrix                    */
/*                                                                    */
/*            return struct FileMatrix : found matrix data            */
/*                                      if NULL, search fail.         */
/*            function: search the matrix by name.                    */
/*                                                                    */
/**********************************************************************/
struct FileMatrix *search_data( char *name )
{
	struct FM *fm;

	for( fm = top; fm; fm = fm->next ){
		if( strcmp( name, fm->M.name ) == 0 ){
			return( &fm->M );
		}
	}
	return( NULL );
}

/**********************************************************************/
/*                                                                    */
/*            function save_matrix()                                  */
/*                                                                    */
/*            args  struct FileMatrix *fm : save matrix               */
/*                                                                    */
/*            return char * : error string                            */
/*                            NULL==success                           */
/*            function: save the matrix to the data file.             */
/*                                                                    */
/**********************************************************************/
char  * save_matrix( struct FileMatrix *fm )
{
	FILE *fp;
	long pos;
	long ret;
	struct FileMatrix readb;

	if( (fp = open_matfile( getenv( "GEDO_LWP" ), "r+b" )) == NULL ){
		if( (fp = open_matfile( getenv( "GEDO_LWP" ), "w+b" )) == NULL ){
			return( "file open error" );
		}
	}
	ret = -1L;
	while( 1 ){
		pos = ftell( fp );
		if( fread( &readb, sizeof(struct FileMatrix), 1, fp ) <= 0 ){
			break;
		}
		if( strcmp( readb.name, fm->name ) == 0 ){
			ret = pos;
		}
	}
	if( ret != -1L ){
		if( fseek( fp, ret, SEEK_SET ) == -1L ){
			return( "file seek error" );
		}
	}
	if( fwrite( fm, sizeof(struct FileMatrix), 1, fp ) != 1 ){
		return( "file write error" );
	}
	fclose( fp );
	return( NULL );
}


/**********************************************************************/
/*                                                                    */
/*            function Matrix_Proc()                                  */
/*                                                                    */
/*            args  ClientData cd  : call back data, not used.        */
/*                  Tcl_Interp *interp : TclTk interpreter handle     */
/*                  int argc  :  argment number                       */
/*                  char **argv : argments                            */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: matrix command on Tcl/Tk matgui.tcl           */
/*                                                                    */
/**********************************************************************/
int Matrix_Proc( ClientData cd, Tcl_Interp *interp, int argc, char **argv )
{
	char *ret;
	struct FileMatrix *fm;
	struct FileMatrix fmb;
	int i;
	char work1[5], work2[10], *workp;

	if( argc < 2 ){
		return( TCL_ERROR );
	}
	Spec = Tcl_GetVar( interp, "name", 0 );
	if( strcmp( argv[1], "open" ) == 0 ){
		findfirst_matrix();
	} else if( strcmp( argv[1], "read" ) == 0 ){
		ret =  findnext_matrix();
		if( ret ){
			strcpy( interp->result, ret );
		} else {
			interp->result[0] = '\0';
		}
	} else if( strcmp( argv[1], "get" ) == 0 ){
		if( argc < 3 ){
			return( TCL_ERROR );
		}
		if( (fm = search_data( argv[2] )) == NULL ){
			return( TCL_ERROR );
		}
		for( i = 0; i < 9; i++ ){
			sprintf( work1, "m%d", i+1 );
			sprintf( work2, "%g", fm->data[i] );
			Tcl_SetVar( interp, work1, work2, 0 );
		}
	} else if( strcmp( argv[1], "save" ) == 0 ){
		if( argc < 3 ){
			return( TCL_ERROR );
		}
		strcpy( fmb.name, argv[2] );
		for( i = 0; i < 9; i++ ){
			sprintf( work1, "m%d", i+1 );
			workp = Tcl_GetVar( interp, work1, 0 );
			fmb.data[i] = atof(workp);
		}
		ret = save_matrix( &fmb );
		if( ret ){
			strcpy( interp->result, ret );
		} else {
			interp->result[0] = '\0';
		}
	} else {
		return( TCL_ERROR );
	}
	return( TCL_OK );
}


/**********************************************************************/
/*                                                                    */
/*            function Matrix_Init()                                  */
/*                                                                    */
/*            args  Tcl_Interp *interp : TclTk interpreter handle     */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: set the command matrix to the interp.         */
/*                                                                    */
/**********************************************************************/
int Matrix_Init( Tcl_Interp *interp )
{
	Tcl_CreateCommand( interp, "matrix", Matrix_Proc, (ClientData *)NULL,
			  (Tcl_CmdDeleteProc *)NULL );
	return( TCL_OK );
}

/**********************************************************************/
/*                                                                    */
/*            function main()                                         */
/*                                                                    */
/*            args   NONE                                             */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: main routine                                  */
/*                                                                    */
/**********************************************************************/
int main()
{
	Tcl_Interp *interp;
	char buff[256];
	char *e;

	interp = Tcl_CreateInterp();
	if( Tcl_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "%s\n", interp->result );
		exit( 1 );
	}

	if( Tk_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "Tk_Init failed:%s\n", interp->result );
		exit( 1 );
	}

	if( Matrix_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "Dith_Init failed:%s\n", interp->result );
		exit( 1 );
	}

	while( fgets( buff, sizeof(buff)-1, stdin ) ){
		if( Tcl_Eval( interp, buff ) == TCL_ERROR ){
			fprintf( stderr, "line read error\n" );
			exit( 1 );
		}
	}
	if( (e = getenv("GEDO_PATH")) == NULL ){
		e = "./";
	}
	sprintf( buff, "%smatgui.tcl", e );
	if( Tcl_EvalFile( interp, buff ) == TCL_ERROR ){
		fprintf( stderr, "matgui.tcl TCL error\n" );
		exit( 1 );
	}
	Tk_MainLoop();
	return( 0 );
}
