/**********************************************************************/
/*                                                                    */
/*            matrix.c   gedo PLUG specified routine                  */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*    C1:  filter_process(), edges of graphics not set! Fixed         */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "plug.h"
#include "matrix.h"

char MatStr[129] = { "0.1:0.1:0.1:0.1:0.2:0.1:0.1:0.1:0.1" };
float Matrix[3][3];
char SaveName[256];
char LoadName[256];
int  List = 0;
int Gui = NOGUI;

void set_var( char *work );

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };
char *Spec =    { "matrix" };
char *Comment = { "matrix effecter, tranfar by the filter" };
PLUGITEM Item[] = {
{ "matrix", GETSTR, MatStr, 128, 0,  "transfar matrix 3x3, like 1:1:1:1:1.5:1:1:1:1" },
{ "save",   GETSTR, SaveName, 255, 0,"name the current matrix" }, 
{ "load",   GETSTR, LoadName, 255, 0,"load the named matrix" },
{ "list",   SWITCH, &List,  1, 0,     "search the named matrix" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ NOTCL | ENTRY, 0, "matrix = ", "matrix=%s" },
};

int TagNum = sizeof(Tag) / sizeof(struct Tags);
DEFAULTSTATUS;
/**********************************************************************/

float Weight;		/* total of the matrix value */

#define Src(i,j)	src[(i)*w+(j)]
/**********************************************************************/
/*                                                                    */
/*             function filter()                                      */
/*                                                                    */
/*             unsigned int *src : tmp accumlated color top           */
/*             unsigned int *dst : original Image pixel top           */
/*             int x             : now position X                     */
/*             int y             : now position Y                     */
/*             int w             : width of the image                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Replace the pixel by the matrix filtering.   */
/*                                                                    */
/**********************************************************************/
void filter( unsigned int *src, unsigned int *dst, int x, int y, int w )
{
	int i, j, k;
	unsigned int in[3][3];
	int val[3];
	float wt;

	for( k = 0; k < 3; k++ ){
		for( i = 0; i < 3; i++ ){
			for( j = 0; j < 3; j++ ){
				if( k == 0 ){
					in[i][j] = GET_RED( Src(y+i-1,x+j-1) );
				} else if( k == 1 ){
					in[i][j] = GET_GREEN( Src(y+i-1,x+j-1) );
				} else if( k == 2 ){
					in[i][j] = GET_BLUE( Src(y+i-1,x+j-1) );
				}
			}
		}
		wt = 0.0;
		for( i = 0; i < 3; i++ ){
			for( j = 0; j < 3; j++ ){
				wt += Matrix[i][j] * (float)in[i][j];
			}
		}
		wt = wt / Weight + 0.5;
		val[k] = (int)wt;
		if( val[k] > 0xff ){
			val[k] = 0xff;
		} else if( val[k] < 0 ){
			val[k] = 0;
		}
	}
	dst[y*w+x] = ((val[0] << 16) | (val[1] << 8) | val[2]);
}

/**********************************************************************/
/*                                                                    */
/*             function filter_process()                              */
/*                                                                    */
/*             char *tmp      : work area                             */
/*             struct Shm *shm: data struct                           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: For each pixel, do the filter process.       */
/*           C1: Edges cannot set ... Fixed                           */
/**********************************************************************/
void filter_process( unsigned int *tmp, struct Shm *shm )
{
	unsigned int *p, *q;
	int i, j;
	int width, height;

	width = shm->width; height = shm->height;
	p = shm->data; q = tmp;


	memcpy( q, p, shm->width * shm->height * sizeof(unsigned int) );

	for( i = 1; i < height - 1; i++ ){
		for( j = 1; j < width - 1; j++ ){
			filter( p, q, j, i, width );
		}
        }
	p = shm->data; q = tmp;
	for( i = 0; i < width * height; i++, p++, q++ ){
		if( MASKOFF( *p ) ){
			*p = ((*p & MASK ) | *q);
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function prepare_matrix()                              */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Prepare the matrix to use.                   */
/*                                                                    */
/**********************************************************************/
int prepare_matrix( void )
{	
	int i, j;
	char *p;

	for( i = j = 0, p = MatStr; j < 3 && *p; ){
		if( *p == ':' ){
			if( ++i >= 3 ){
				i = 0; j++;
			}
			p++;
		} else {
			Matrix[i][j] = atof( p );
			for( ; *p && *p != ':'; p++ );
		}
	}
		
	Weight = 0.0;
	for( i = 0; i < 3; i++ ){
		for( j = 0; j < 3; j++ ){
			Weight += Matrix[i][j];
		}
	}

	if( Weight < 0.0001 && Weight > -0.0001 ){
		Weight = 1.0;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	int width, height;
	unsigned int *tmp;

	width = file->buf->width;
	height = file->buf->height;

        if( file->buf->mask_kind == DEPTHMASK ){
		return( EILLMASK );
	}
	if( prepare_matrix() < 0 ){
		strcpy( file->mes->u.data, "illegal matrix, cannot interpret" );
		return( -1 );
	}
	if( (tmp = malloc( width * height * 4 )) == NULL ){
		return( EMALLOC );
	}
	filter_process( tmp, file->buf );
	free( tmp );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config_gui()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config_gui( struct File *file )
{
	FILE *fp;
	FILE *fpp;
	char buff[256];
	char fifo[256];
	char *e;


	if( (fp = create_fifo( fifo )) == NULL ){
		return( -1 );
	}
	if( (e = getenv("GEDO_PATH")) == NULL ){
		e = "./";
	}
	sprintf( buff, "%smatgui <%s &", e, fifo );
	if( (fpp = popen( buff, "r" )) == NULL ){
		return( -1 );
	}
	fprintf( fp, "set name \"%s\"\n"
		"set m1	   %g\nset m2      %g\nset m3     %g\n"
		"set m4    %g\nset m5      %g\nset m6     %g\n"
		"set m7	   %g\nset m8      %g\nset m9     %g\n",
		Spec,
		Matrix[0][0], Matrix[0][1], Matrix[0][2],
		Matrix[1][0], Matrix[1][1], Matrix[1][2],
		Matrix[2][0], Matrix[2][1], Matrix[2][2] );
	fclose( fp );
	
	fgets( buff, sizeof(buff)-1, fpp );
	pclose( fpp );

	if( strcmp( buff, "abort\n" ) == 0 ){
		return( EUSRBREAK );
	}
	buff[strlen(buff) - 1] = '\0';
	strcpy( file->mes->u.data, buff );
	return( line_config( file ) );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	int ret;
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			prepare_matrix();
			return( config_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		LoadName[0] = SaveName[0] = '\0'; List = 0;
		ret = line_config( file );
		if( ret != 0 ){
			return( ret );
		}
		if( List ){
			List = 0;
			search_matrix();
		}
		if( LoadName[0] ){
			ret = load_matrix( LoadName, Matrix, MatStr );
			if( ret != 0 ){
				return( ret );
			}
		}
		if( SaveName[0] ){
			prepare_matrix();
			ret = save_matrix( SaveName, Matrix );
		}
		return( ret );
	}
}

