/**********************************************************************/
/*                                                                    */
/*            mosaic.c   gedo PLUG specified routine                  */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "plug.h"

int UnitHeight = 4;
int UnitWidth = 4;

#define AVERAGE		0
#define MAXIMUM		1
#define MINIMUM		2
#define INDEPENDENT	3
int ColorOp = AVERAGE;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };
char *Spec =    { "mosaic" };
char *Comment = { "mosaic effecter, make area mosaics" };
PLUGITEM Item[] = {
{ "width", GETINT, &UnitWidth, 2, 100, "unit size X" },
{ "height", GETINT, &UnitHeight, 2, 100, "unit size Y" },
{ "average", SWITCH, &ColorOp, AVERAGE, 'A', "take average color of the area" },
{ "max", SWITCH, &ColorOp, MAXIMUM, 'X', "take maximum color of the area" },
{ "min", SWITCH, &ColorOp, MINIMUM, 'N', "take minimum color of the area" },
{ "indep", SWITCH, &ColorOp, INDEPENDENT, 'I', "RGB split color of the area" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" },
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "width of the unit ", "width=%d" },
{ ENTRY, 1, "height of the unit", "height=%d" },
{ RADIO, 2, "Mode " },
{ ITEM, 2, "average(a)", "average" },
{ ITEM, 3, "maximum(x)", "max" },
{ ITEM, 4, "minimum(n)", "min" },
{ ITEM, 5, "independent(i)", "indep" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/**********************************************************************/

int Counter;	/* how many pixels in the tail? */
int R, G, B;	/* accumurater for each RGB */

/**********************************************************************/
/*                                                                    */
/*             function clear_color()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: For each tile, clean up the sums.            */
/*                                                                    */
/**********************************************************************/
void clear_color( void )
{
	Counter = 0;
	if( ColorOp == MINIMUM ){
		R = G = B = 0xff;
	} else {
		R = G = B = 0;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function calc_color()                                  */
/*                                                                    */
/*             args   unsigned int v : color pixel                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: For each pixel in a tile, accumlate RGB by   */
/*                   the mode.                                        */
/**********************************************************************/
void calc_color( unsigned int v )
{
	unsigned int tmp;

	Counter++;
	switch( ColorOp ){
	case AVERAGE:
	case INDEPENDENT:
		R += GET_RED(v);
		G += GET_GREEN(v);
		B += GET_BLUE(v);
		break;
	case MAXIMUM:
		if( R < (tmp = GET_RED(v)) ){
			R = tmp;
		}
		if( G < (tmp = GET_GREEN(v)) ){
			G = tmp;
		}
		if( B < (tmp = GET_BLUE(v)) ){
			B = tmp;
		}
		break;
	case MINIMUM:
		if( R > (tmp = GET_RED(v)) ){
			R = tmp;
		}
		if( G > (tmp = GET_GREEN(v)) ){
			G = tmp;
		}
		if( B > (tmp = GET_BLUE(v)) ){
			B = tmp;
		}
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function get_color()                                   */
/*                                                                    */
/*             args    NONE                                           */
/*                                                                    */
/*             return unsigned int : result of calc_color() in a tile */
/*                                                                    */
/*             function: Get result of accumlation for each a tile.   */
/*                                                                    */
/**********************************************************************/
unsigned int get_color( void )
{
	if( ColorOp == MAXIMUM || ColorOp == MINIMUM ){
		return( (R << 16) | (G << 8) | B );
	} else {
		R /= Counter;
		G /= Counter;
		B /= Counter;
		return( (R << 16) | (G << 8) | B );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function get_indep()                                   */
/*                                                                    */
/*             args    int x: position X                              */
/*                     int y: position Y                              */
/*                                                                    */
/*             return unsigned int : RGB value depend to position     */
/*                                                                    */
/*             function: If independent mode, pixel is R or G or B    */
/*                    value average of tail depnd to position.        */
/**********************************************************************/
unsigned int get_indep( int x, int y )
{
	switch( (x + y) % 3 ){
	case 0:
		return( R << 16 );
		break;
	case 1:
		return( G << 8 );
		break;
	case 2:
		return( B );
		break;
	}
	return( 0x00ffffff );
}


/**********************************************************************/
/*                                                                    */
/*             function each_tail()                                   */
/*                                                                    */
/*             args    unsigned int *p: top of the image buff         */
/*                     int x: tile left top position X                */
/*                     int y: tile left top position Y                */
/*                     int w: image width                             */
/*                     int h: image height                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: For each tile, make mosaiced image.          */
/*                                                                    */
/**********************************************************************/
void each_tail( unsigned int *p, int x, int y, int w, int h )
{
	int i, j;
	unsigned int col;
	unsigned int *st;

	st = &p[y * w + x];
	clear_color();
	for( i = 0; i < UnitHeight; i++ ){
		if( y + i >= h ){
			break;
		}
		for( j = 0; j < UnitWidth; j++ ){
			if( x + j >= w ){
				break;
			}
			calc_color( st[j] );
		}
		st += w;
	}
	col = get_color();
	st = &p[y * w + x];
	for( i = 0; i < UnitHeight; i++ ){
		if( y + i >= h ){
			break;
		}
		for( j = 0; j < UnitWidth; j++ ){
			if( x + j >= w ){
				break;
			}
			if( ColorOp == INDEPENDENT ){
				col = get_indep( x+j, y+i );
			}
			st[j] = ((st[j] & MASK) | col);
		}
		st += w;
	}
}


/**********************************************************************/
/*                                                                    */
/*             function tail_proc()                                   */
/*                                                                    */
/*             args    unsigned int *p: top of the image buff         */
/*                     int w: image width                             */
/*                     int h: image height                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Make square tile Image.                      */
/*                                                                    */
/**********************************************************************/
void tail_proc( unsigned int *p, int w, int h )
{
	int i, j;
	int x, y;

	x = (w + UnitWidth - 1) / UnitWidth;
	y = (h + UnitHeight - 1) / UnitHeight;

	for( i = 0; i < y; i++ ){
		for( j = 0; j < x; j++ ){
			each_tail( p, j*UnitWidth, i*UnitHeight, w, h );
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*             function triangle_proc()                               */
/*                                                                    */
/*             args    unsigned int *p: top of the image buff         */
/*                     int w: image width                             */
/*                     int h: image height                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Make triangle tile Image. NOT HERE YET.      */
/*                                                                    */
/**********************************************************************/
void triangle_proc( unsigned int *p, int w, int h )
{
	/* FIXME */
}

/**********************************************************************/
/*                                                                    */
/*             function mosaic()                                      */
/*                                                                    */
/*             args    unsigned int *p: top of the image buff         */
/*                     int w: image width                             */
/*                     int h: image height                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Make mosaic image.                           */
/*                                                                    */
/**********************************************************************/
void mosaic( unsigned int *p, int w, int h )
{
	tail_proc( p, w, h );
/*	if( Mode == TAIL ){
		tail_proc( p, w, h );
	} else if( Mode == TRIANGLE ){
		triangle_proc( p, w, h );
	}*/
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	int width, height;
	int j;
	unsigned int *p, *q;
	unsigned int *shadow;

	width = file->buf->width;
	height = file->buf->height;

        if( file->buf->mask_kind == DEPTHMASK ){
		return( EILLMASK );
	}

	if( (shadow = malloc( width * height * 4 )) == NULL ){
		return( EMALLOC );
	}
	memcpy( shadow, file->buf->data, width * height * 4 );
	mosaic( shadow, width, height );
	p = file->buf->data; q = shadow;
	for( j = 0; j < height * width; j++, p++, q++ ){
		if( MASKOFF(*p) ){
			*p = *q;
		}
	}
	free( shadow );
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
