/**********************************************************************/
/*                                                                    */
/*             move.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "plug.h"

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "move" };
char *Comment = { "move selecter, move current mask" };

PLUGITEM Item[];
int ItemNum = 0;
struct Tags Tag[];
int TagNum = 0;
DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* image width and height */
static unsigned int *Data;	/* image buffer */
static char *Buff;		/* temporary buffer */

/**********************************************************************/
/*                                                                    */
/*             function make_tmpmask()                                */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int  : 0==success  else=errorcode              */
/*                                                                    */
/*             function: Make temporary buffer and save now mask.     */
/*                                                                    */
/**********************************************************************/
int make_tmpmask( void )
{
	int i, j;
	unsigned int *p;
	unsigned char *q;

	if( (Buff = malloc( Width * Height )) == NULL ){
		return( EMALLOC );
	}
	p = Data; q = Buff;

	for( i = 0; i < Height; i++ ){
		for( j = 0; j < Width; j++ ){
			*q++ = (unsigned char)GET_MASK(*p++);
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function restore_mask()                                */
/*                                                                    */
/*             int x   : move X                                       */
/*             int y   : move Y                                       */
/*                                                                    */
/*             return  int : 0==success  else==errorcode              */
/*                                                                    */
/*             function: Re-set the moved mask.                       */
/*                                                                    */
/**********************************************************************/
int restore_mask( int x, int y )
{
	unsigned char *s;
	unsigned int *d;
	int dstx, dsty, srcx, srcy, lenx, leny;
	int di, si, dj, sj;

	if( x >= 0 ){
		dstx = x;
		srcx = 0;
		lenx = Width;
	} else {
		dstx = 0;
		srcx = (-1) * x;
		lenx = Width + x;
	}
	if( y >= 0 ){
		dsty = y;
		srcy = 0;
		leny = Height;
	} else {
		dsty = 0;
		srcy = (-1) * y;
		leny = Height + y;
	}

	for( di = dsty, si = srcy; di < leny; di++, si++ ){
		d = Data + di * Width; s = Buff + si * Width;
		for( dj = dstx, sj = srcx; dj < lenx; dj++, sj++ ){
			d[dj] = ((d[dj] & ~MASK) | ((unsigned int)s[sj] << 24 ));
		}
	}
	free( Buff );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *p;
	int i;
	int x, y;
	int ret;

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	Data = file->buf->data;

	if( sb->selnum != SKBOX ){
		return( EILLSELECTION );
	}
	check_range( sb, Width, Height, 2, NEGLECT );
	if( sb->defnum < 2 ){
		return( EILLNUMBER );
	}

	if( (ret = make_tmpmask()) != 0 ){
		return( ret );
	}
	if( file->buf->mask_kind == SIMPLEMASK ){
		for( i = 0, p = Data; i < Height * Width; i++, p++ ){
			SETMASK(*p);
		}
	} else if( file->buf->mask_kind == DEPTHMASK ){
		for( i = 0, p = Data; i < Height * Width; i++, p++ ){
			*p = (*p & ~MASK);
		}
	}
	x = sb->def[1].x - sb->def[0].x;
	y = sb->def[1].y - sb->def[0].y;

	if( (ret = restore_mask( x, y )) != 0 ){
		return( ret );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		printf( "move: move selecter\n\tNO configration parameter\n" );
		str[0] = '\0';
		return( 0 );
	} else {
		return( line_config( file ) );
	}
}
