/**********************************************************************/
/*                                                                    */
/*           text.c     gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include "plug.h"
#include "xstring.h"

#define NOGUI	0
#define USEGUI  1
int Gui = NOGUI;

static unsigned char Text[TEXTMAX] = { "test string" };
static char JFontname[FONTMAX] = { "kanji16" };
static char EFontname[FONTMAX] = { "fixed" };
static char Fontname[FONTMAX] = {  "" };
static char FileText[TEXTMAX] = { "file.txt" };
static int  InFileText = 0;

static int KMode = JIS;
static char *SKMode[3] = { "jis", "euc", "sjis" };

static int AddW = 0;
static int AddH = 0;
static int MaxW = -1;
static int Scale = 20;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "text" };
char *Comment = { "string selecter, mask of font form" };

PLUGITEM Item[] = {
/*0*/{ "gui",     SWITCH,  &Gui,    USEGUI, 'G',   "exec GUI interface" },   
/*1*/{ "nogui",   SWITCH,  &Gui,    NOGUI,  'N',   "not exec GUI interface" },
/*2*/{ "text",    GETSTR,  &Text,     80,  0,   "display text" },
/*3*/{ "filetext",GETSTR,  &FileText, 80,  0,   "display text in the file" },
/*4*/{ "infiletext",SWITCH,  &InFileText, 1,  'F', "get display text from <filetext>" },
/*5*/{ "linetext",SWITCH,  &InFileText, 0,  'T',   "get display text from commandline" },
/*6*/{ "jfont",    GETSTR,  &JFontname, 80, 0,    "JAPANESE font fullname" },
/*7*/{ "efont",    GETSTR,  &EFontname, 80, 0,    "ASCII font fullname" },
/*8*/{ "font",     GETSTR, &Fontname,   80, 0,    "finding font name" },
/*9*/{ "jis",     SWITCH,  &KMode,    JIS, 'J', "include JIS japanese code" },
/*10*/{ "euc",     SWITCH,  &KMode,    EUC, 'E', "include EUC japanese code" },
/*11*/{ "sjis",    SWITCH,  &KMode,   SJIS, 'S', "include Shift JIS japanese code" },
/*12*/{ "addW",    GETINT,  &AddW, -100,   100, "add width between characters" },
/*13*/{ "addH",    GETINT,  &AddH, -100,   100, "add height between line" },
/*14*/{ "MaxW",    GETINT,  &MaxW, -1,   1000,  "maximum width of the area" },
/*15*/{ "scale",    GETINT,  &Scale, 0,   80,   "scale for SCALABLE font" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);
struct Tags Tag[] = {
/*0*/{ NOTCL | RADIO, 0, "Config Mode" },
/*1*/{ NOTCL | ITEM,  0, "GUI(G)", "gui" },
/*2*/{ NOTCL | ITEM,  1, "NOT gui(N)", "nogui" },
/*3*/{ NOTCL | RADIOEND, -1 },
/*4*/{ NOTCL | RADIO, 4, "Text is" },
/*5*/{ NOTCL | ITEM, 4, "in FILE(F)", "infiletext" },
/*6*/{ NOTCL | ITEM, 5, "this text(T)", "linetext" },
/*7*/{ NOTCL | RADIOEND, -1 },
/*8*/{ NOTCL | ENTRY, 3, "Text is in the file", "filetext=%s" },
/*9*/{ NOTCL | ENTRY, 2, "Text is here", "text=%s" },
/*10*/{ NOTCL | RADIO, 9, "KANJI is" },
/*11*/{ NOTCL | ITEM,  9, "JIS(J)", "jis" },
/*12*/{ NOTCL | ITEM,  10, "EUC(E)", "euc" },
/*13*/{ NOTCL | ITEM, 11, "SJIS(S)", "sjis" },
/*14*/{ NOTCL | RADIOEND, -1 },
/*15*/{ NOTCL | ENTRY, 8, "font name is", "font=%s" },
/*16*/{ NOTCL | ENTRY,12, "add Horizontal", "addW=%d" },
/*17*/{ NOTCL | ENTRY,13, "add Vertical",   "addH=%d" },
/*18*/{ NOTCL | ENTRY,14, "maximum width of area(-1==no limit)", "MaxW=%d" },
/*19*/{ NOTCL | ENTRY,15, "scale for SCALABLE font", "scale=%d" },
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

/**********************************************************************/
/*                                                                    */
/*            function status()                                       */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return int : 0==success  else==errorcode                */
/*                                                                    */
/*            function: status command handler called from command.c  */
/*                                                                    */
/**********************************************************************/
void status( int flag, char *buff )
{
     static int prev_status = 0;
     static char NextBuff[257];
     int len, tlen;
     char *wbuff;
     
     if( prev_status ){
	  strcpy( buff, &NextBuff[1] );
	  prev_status = 0;
	  return;
     }
     if( InFileText ){
/* Trick for not to appear Tag[9] element(Text) in status buffer */
	  Tag[9].kind |= INITONLY;
	  get_default_status( flag, buff );
	  Tag[9].kind &= ~INITONLY;
     } else {
	  Tag[8].kind |= INITONLY;
	  Tag[9].kind |= INITONLY;
	  get_default_status( flag, buff );
	  Tag[8].kind &= ~INITONLY;
	  Tag[9].kind &= ~INITONLY;
     }
     len = strlen(buff); 
     wbuff = buff;

     tlen = strlen(JFontname) + 7;
     if( len + tlen >= 255 ){
	  fprintf( stderr, "status is too long\n"
		   "plese call more once\n" );
	  prev_status = 1;
	  wbuff = NextBuff;
     }
     strcat( wbuff, ",jfont=" );
     strcat( wbuff, JFontname );

     tlen = strlen(EFontname) + 7;
     if( len + tlen >= 255 ){
	  fprintf( stderr, "status is too long\n"
		   "plese call more once\n" );
	  prev_status = 1;
	  wbuff = NextBuff;
     }
     strcat( wbuff, ",efont=" );
     strcat( wbuff, EFontname );

     if( InFileText == 0 ){
	  tlen = strlen(Text) + 6;
	  if( len + tlen >= 255 ){
	       fprintf( stderr, "status is too long\n"
			"plese call more once\n" );
	       prev_status = 1;
	       wbuff = NextBuff;
	  } else {
	       strcat( wbuff, ",text=" );
	       strcat( wbuff, Text );
	  }
     }
}
/********************** description end ********************/


/**********************************************************************/
/*                                                                    */
/*            function get_fontimage()                                */
/*                                                                    */
/*            args  struct fontimage *fi : header buffer to set       */
/*                  unsigned char *buff  : data buffer to set         */
/*                                                                    */
/*            return int   :  0==success  else==errorcode             */
/*                                                                    */
/*            function: Get font image from Pixmap and set to <fi>    */
/*                     and <buff>.                                    */
/**********************************************************************/
struct fontimage *get_fontimage( struct File *file, int *code )
{
     FILE *fpp;
     struct fontimage *fi;
     struct fontimage res;
     char buff[TEXTMAX*2];
     int len;
     int ret;

     if( InFileText ){
	  sprintf( buff, "xstrsvr "
		   "'filetext=%s' "
		   "'jfont=%s' "
		   "'efont=%s' "
		   "'%s' "
		   "'addW=%d' "
		   "'addH=%d' "
		   "'MaxW=%d' "
		   "'mode=do'",
		   FileText, 
		   JFontname, EFontname, SKMode[KMode], AddW, 
		   AddH, MaxW );
     } else {
	  sprintf( buff, "xstrsvr "
		   "'text=%s' "
		   "'jfont=%s' "
		   "'efont=%s' "
		   "'%s' "
		   "'addW=%d' "
		   "'addH=%d' "
		   "'MaxW=%d' "
		   "'mode=do'",
		   Text,
		   JFontname, EFontname, SKMode[KMode], AddW, 
		   AddH, MaxW );
     }
     if( Fontname[0] != '\0' ){
	  char buff2[TEXTMAX];
	  sprintf( buff2, " 'font=%s' 'scale=%d'",Fontname, Scale );
	  strcat( buff, buff2 );
     }

     if( (fpp = popen( buff, "r" )) == NULL ){
	  perror( "pipe open to xstrsvr" );
	  *code = EOPENFILE;
	  return NULL;
     }
     if( fread( &res, sizeof(struct fontimage), 1, fpp ) != 1 ){
	  pclose( fpp );
	  *code = EREADFILE;
	  return NULL;
     }
     if( res.flag != ENOERROR ){
	  strcpy( file->mes->u.data, res.jfont );
	  pclose( fpp );
	  *code =  res.flag;
	  return NULL;
     }

     len = res.width * res.height;
     if( (fi = malloc( sizeof(struct fontimage) + len )) == NULL ){
	  pclose( fpp );
	  *code =  EMALLOC;
	  return NULL;
     }
     memcpy( fi, &res, sizeof(struct fontimage) );
     if( (ret = fread( fi->data, sizeof(char), len, fpp )) != (size_t)len ){
	  pclose( fpp );
	  *code = EREADFILE;
	  return NULL;
     }
     pclose( fpp );
     if( strcmp( res.efont, EFontname ) != 0 ){
	  fprintf( stderr, "efont different %s %s\n", res.efont, EFontname );
	  strcpy( EFontname, res.efont );
	  Fontname[0] = '\0';
     }
     if( strcmp( res.jfont, JFontname ) != 0 ){
	  fprintf( stderr, "jfont different %s %s\n", res.jfont, JFontname );
	  strcpy( JFontname, res.jfont );
	  Fontname[0] = '\0';
     }
     
     *code = ENOERROR;
     return fi;
}	


/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	int width, height;
	unsigned int *data, *p;
	char *q;
	int i, j;
	int minx, maxx, miny, maxy;
	struct fontimage *fi;
	int code;

	if( (fi = get_fontimage( file, &code )) == NULL ){
	     return code;
	}
	sb = &file->mes->u.sb;
	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	if( sb->selnum != SKBOX ){
		return( EILLSELECTION );
	}
	check_range( sb, width, height, 2, ADJUST );
	if( sb->defnum < 2 ){
		return( EILLNUMBER );
	}

	for( i = 0, p = data; i < height * width; i++, p++ ){
		SETMASK(*p);
	}

	minx = min( sb->def[0].x, sb->def[1].x );
	miny = min( sb->def[0].y, sb->def[1].y );
	maxx = max( sb->def[0].x, sb->def[1].x );
	maxy = max( sb->def[0].y, sb->def[1].y );
	if( minx + fi->width >= width ){
		maxx = width - minx;
	} else {
		maxx = fi->width;
	}
	if( miny + fi->height >= height ){
		maxy = height - miny;
	} else {
		maxy = fi->height;
	}

	for( i = 0; i < maxy; i++ ){
		p = data + (miny + i) * width;
		q = fi->data + i * fi->width;
		for( j = 0; j < maxx; j++ ){
			if( q[j] == 0 ){
				CLRMASK(p[minx+j]);
			}
		}
	}
	file->buf->mask_kind = SIMPLEMASK;
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function config_textGUI()                               */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return  int : 0==success  else==errorcode               */
/*                                                                    */
/*            function: Invoke the xstring for GUI selection and get  */
/*                   font image.                                      */
/**********************************************************************/
static int config_textGUI( struct File *file )
{
     FILE *fpp;
     char buff[TEXTMAX*2];
     int len;
     char *p;

     if( InFileText ){
	  sprintf( buff, "xstring.tcl "
		   "'set thetext \"%s\"' "
		   "'set thefile %s' "
		   "'set Tmode 1' "
		   "'set jfont %s' "
		   "'set efont %s' "
		   "'set Kmode %d' "
		   "'set addw %d' "
		   "'set addh %d' "
		   "'set maxw %d' "
		   "'set thescale %d' "
		   "'set thefont %s'",
		   Text, FileText, 
		   JFontname, EFontname, KMode, AddW, 
		   AddH, MaxW, Scale, Fontname );
     } else {
	  sprintf( buff, "xstring.tcl "
		   "'set thetext \"%s\"' "
		   "'set thefile %s' "
		   "'set Tmode 0' "
		   "'set jfont %s' "
		   "'set efont %s' "
		   "'set Kmode %d' "
		   "'set addw %d' "
		   "'set addh %d' "
		   "'set maxw %d' "
		   "'set thescale %d' "
		   "'set thefont %s'",
		   Text, FileText, 
		   JFontname, EFontname, KMode, AddW, 
		   AddH, MaxW, Scale, Fontname );
	}
	if( (fpp = popen( buff, "r" )) == NULL ){
		perror( "pipe open to xstring" );
		return( EOPENFILE );
	}
	fgets( buff, sizeof(buff), fpp );
	pclose( fpp );
	if( strncmp( buff, "abort", 5 ) == 0 ){
	     return EUSRBREAK;
	}

/* trick for NOT returned item font=<select font>....
  Because the REAL font is set at efont or jfont.
  If set, the config string will be too long. 
  This trick must be at ONLY GUI, because command line directly set 
  item font. */ 

	for( p = buff; *p && *p != ','; p++ );
	if( *p == ',' ){
	     if( strncmp( p+1, "font=", 5 ) == 0 ){
		  char *at, *til;
		  at = p + 6;
		  for( p = at; *p && *p != ','; p++ );
		  if( *p == ',' ){
		       *p = '\0';
		       til = p + 1;
		  } else {
		       til = NULL;
		  }
		  strcpy( Fontname, at );
		  at -= 5;
		  if( til ) {
		       while( (*at++ = *til++) );
		  } else {
		       *(at-1) = '\0';
		  }
	     }
	}

/* trick END */
	len = strlen(buff);
	if( len >= 256 ){
	     len = 256;
	     fprintf( stderr, "too long configure items\n" );
	}
	strncpy( file->mes->u.data, buff, len );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
     char *str;
     int ret;
     
     str = file->mes->u.data;

     if( str[0] ){
	  return( line_config( file ) );
     } else {
	  if( Gui ){
	       if( (ret = config_textGUI( file )) != 0 ){
		    return( ret );
	       }
	       ret = line_config( file );
	       return( ret );
	  } else {
	       return( config_comline( file ) );
	  }
     }
}

