/**********************************************************************/
/*                                                                    */
/*             paint.c  gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>
#include "plug.h"

unsigned long Color = 0;

#define C_REPLACE	0
#define C_ADD		1
#define C_SUB		2
int Mode = 0;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "paint" };
char *Comment = { "paint the color into the image" };

PLUGITEM Item[] = {
{ "gui",   SWITCH,  &Gui, GUI,  'G',      "use GUI interface" },
{ "nogui", SWITCH,  &Gui, NOGUI,  'N',    "not use GUI interface" },
{ "color", GETHEX,  &Color, 0, 0x00ffffff, "paint color" },
{ "rep",   SWITCH,  &Mode, C_REPLACE, 'R',   "replace the area by the color" },
{ "add",   SWITCH,  &Mode, C_ADD,     'A',   "add the color to the area" },
{ "sub",   SWITCH,  &Mode, C_SUB,     'S',   "sub the color to the area" },
{ NULL,    GETHEX,  &Color, 0, 0x00ffffff, "paint color == (color=<hex>)" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);
struct Tags Tag[] = {
  { RADIO, 3, "Paint Mode" },
  { ITEM,  3, "replace(R)", "rep" },
  { ITEM,  4, "add(A)", "add" },
  { ITEM,  5, "sub(S)", "sub" },
  { RADIOEND, -1 },
  { ENTRY, 2, "the paint color", "color=%06lx"}
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);
DEFAULTSTATUS;
/********************** description end ********************/

unsigned int Ar, Ag, Ab;	/* Color divided RGB */


/**********************************************************************/
/*                                                                    */
/*             function add_color()                                   */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: In SIMPLEMASK, and at ADD-Mode,              */
/*                      add <Color> to each MASKOFF pixel.            */
/**********************************************************************/
void add_color( unsigned int *p, int size )
{
	unsigned int r, g, b;
	int i;

	for( i = 0; i < size; i++, p++ ){
		if( MASKOFF(*p) ){
			r = (*p & RED);
			g = (*p & GREEN);
			b = (*p & BLUE);
			if( (r = r + Ar) > RED ){
				r = RED;
			}
			if( (g = g + Ag) > GREEN ){
				g = GREEN;
			}
			if( (b = b + Ab) > BLUE ){
				b = BLUE;
			}
			*p = ((*p & MASK) | r | g | b);
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function sub_color()                                   */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: In SIMPLEMASK, and at SUB-Mode,              */
/*                      sub <Color> to each MASKOFF pixel.            */
/*                                                                    */
/**********************************************************************/
void sub_color( unsigned int *p, int size )
{
	unsigned int r, g, b;
	int i;

	for( i = 0; i < size; i++, p++ ){
		if( MASKOFF(*p) ){
			r = (*p & RED);
			g = (*p & GREEN);
			b = (*p & BLUE);
			if( r < Ar ){
				r = BLACK;
			} else {
				r = r - Ar;
			}
			if( g < Ag ){
				g = BLACK;
			} else {
				g = g - Ag;
			}
			if( b < Ab ){
				b = BLACK;
			} else {
				b = b - Ab;
			}
			*p = ((*p & MASK) | r | g | b);
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function rep_color()                                   */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: In SIMPLEMASK, and at REPLACE-Mode,          */
/*                  replace <Color> at each MASKOFF pixel.            */
/*                                                                    */
/**********************************************************************/
void rep_color( unsigned int *p, int size )
{
	int i;

	for( i = 0; i < size; i++, p++ ){
		if( MASKOFF(*p) ){
			*p = ((*p & MASK) | Ar | Ag | Ab);
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function add_dcolor()                                  */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function:  In DEPTHMASK, and at ADD-Mode,              */
/*                      add <Color> to each pixel by Mask depth.      */
/*                                                                    */
/**********************************************************************/
void add_dcolor( unsigned int *p, int size )
{
	unsigned int r, g, b, rate;
	int i;

	for( i = 0; i < size; i++, p++ ){
		rate = (*p & MASK) >> 24;
		r = (*p & RED);
		g = (*p & GREEN);
		b = (*p & BLUE);
/*		if( (r = r * (256-rate) / 256 + Ar * rate / 256) > RED ){*/
		if( (r = r + Ar * rate / 256) > RED ){ 
			r = RED;
		}
/*		if( (g = g * (256-rate) / 256 + Ag * rate / 256) > GREEN ){*/
		if( (g = g + Ag * rate / 256) > GREEN ){ 
			g = GREEN;
		}
/*		if( (b = b * (256-rate) / 256 + Ab * rate / 256) > BLUE ){*/
		if( (b = b + Ab * rate / 256) > BLUE ){ 
			b = BLUE;
		}
		*p = ((*p & MASK) | (r & RED) | (g & GREEN) | (b & BLUE));
	}
}

/**********************************************************************/
/*                                                                    */
/*             function sub_dcolor()                                  */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function:  In DEPTHMASK, and at SUB-Mode,              */
/*                      sub <Color> to each pixel by Mask depth.      */
/*                                                                    */
/**********************************************************************/
void sub_dcolor( unsigned int *p, int size )
{
	unsigned int r, g, b, rate;
	int i;
	unsigned int rr, rg, rb;

	for( i = 0; i < size; i++, p++ ){
		rate = (*p & MASK) >> 24;
		r = (*p & RED);
		g = (*p & GREEN);
		b = (*p & BLUE);
		rr = Ar * rate / 256;
		rg = Ag * rate / 256;
		rb = Ab * rate / 256;
		if( r < rr ){
			r = BLACK;
		} else {
			r = r - rr;
		}
		if( g < rg ){
			g = BLACK;
		} else {
			g = g - rg;
		}
		if( b < rb ){
			b = BLACK;
		} else {
			b = b - rb;
		}
		*p = ((*p & MASK) | (r & RED) | (g & GREEN) | (b & BLUE));
	}
}

/**********************************************************************/
/*                                                                    */
/*             function rep_dcolor()                                  */
/*                                                                    */
/*             unsigned int *p : image top                            */
/*             int size        : image size = width * heigth          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function:  In DEPTHMASK, and at REPLACE-Mode,          */
/*                  replace <Color> to each pixel by Mask depth.      */
/*                                                                    */
/**********************************************************************/
void rep_dcolor( unsigned int *p, int size )
{
	int i;
	unsigned int r, g, b, rate;

	for( i = 0; i < size; i++, p++ ){
		rate = (*p & MASK) >> 24;
		if( (r = Ar * rate / 256) > RED ){
			r = RED;
		}
		if( (g = Ag * rate / 256) > GREEN ){
			g = GREEN;
		}
		if( (b = Ab * rate / 256) > BLUE ){
			b = BLUE;
		}
		*p = ((*p & MASK) | (r & RED) | (g & GREEN) | (b & BLUE));
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	char *str;
	int width, height;
	unsigned int *data;

	str = file->mes->u.data;
	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	Ar = (Color & RED);
	Ag = (Color & GREEN);
	Ab = (Color & BLUE);

	if( file->buf->mask_kind == SIMPLEMASK ){
		if( Mode == C_ADD ){
			add_color( data, height * width );
		} else if( Mode == C_SUB ){
			sub_color( data, height * width );
		} else {
			rep_color( data, height * width );
		}
	} else if( file->buf->mask_kind == DEPTHMASK ){
		if( Mode == C_ADD ){
			add_dcolor( data, height * width );
		} else if( Mode == C_SUB ){
			sub_dcolor( data, height * width );
		} else {
			rep_dcolor( data, height * width );
		}
	} else {
		return( EILLMASK );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function paint_gui()                                    */
/*                                                                    */
/*             args    struct File *file : message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success  -1==fail                       */
/*                                                                    */
/*             function: Execute the X-Interface and get the result.  */
/*                                                                    */
/**********************************************************************/
int paint_gui( struct File *file )
{
	FILE *fpp;
	char buff[128];

	if( (fpp = popen( "xcolor", "r" )) == NULL ){
		fprintf( stderr, "exec xcolor fail\n" );
		return( -1 );
	}
	if( fgets( buff, 127, fpp ) == NULL ){
		fprintf( stderr, "cannot read\n" );
		return( -1 );
	}
	pclose( fpp );
	if( strncmp( buff, "error:", 6 ) == 0 ){
		strcpy( file->mes->u.data, &buff[6] );
		return( EGUIERROR );
	}
	if( strcmp( buff, "abort\n" ) == 0 ){
		return( EUSRBREAK );
	}
	Color = axtol( buff, 6 );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	int ret;
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( Gui == 0 ){
			TagNum = sizeof(Tag) / sizeof(struct Tags);
			return( config_comline( file ) );
		} else {
			if( (ret = paint_gui( file )) != 0 ){
				return( ret );
			}
			TagNum = 5;
			ret = config_comline( file );
			if( ret == 0 ){
				if( *str == '\0' ){
					sprintf( str, "color=%06lx", Color );
				} else {
					sprintf( &str[strlen(str)], ",color=%06lx", Color );
				}
			}
			return( ret );
		}
	} else {
		return( line_config( file ) );
	}
}
