/**********************************************************************/
/*                                                                    */
/*             pass.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "plug.h"

int Gui = NOGUI;

#define LIMLOWER  0
#define LIMLOWEQ  1
#define LIMUPPER  2
#define LIMUPPEQ  3
#define LIMEQUAL  4
static int LimMode = LIMLOWER;

#define CHGCUT    0
#define CHGFILL	  1
#define CHGADD    2
#define CHGSUB    3
static int ChgMode = CHGCUT;
static int AddVal  = 0;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "pass" };
char *Comment = { "low/high pass filter for the depth mask" };

PLUGITEM Item[] = {
{ "LT", SWITCH, &LimMode, LIMLOWER, '1', "select lower value from the selected point" },
{ "LE", SWITCH, &LimMode, LIMLOWEQ, '2', "select low or equal value from the selected point" },
{ "GT", SWITCH, &LimMode, LIMUPPER, '3', "select upper value from the selected point" },
{ "GE", SWITCH, &LimMode, LIMUPPEQ, '4', "select upper or equal value from the selected point" },
{ "EQ", SWITCH, &LimMode, LIMEQUAL, '5', "select equal value from the selected point" },
{ "cut", SWITCH, &ChgMode, CHGCUT, 'c',  "mask value becomes 0" },
{ "fill", SWITCH, &ChgMode, CHGFILL, 'f',  "mask value becomes 0xff" },
{ "add", SWITCH, &ChgMode, CHGADD, 'a',  "mask value added <value>" },
{ "sub", SWITCH, &ChgMode, CHGSUB, 's',  "mask value subed <value>" },
{ "value", GETINT, &AddVal, 0, 255, "if add/sub mode, the value added or subed" },
{ "gui",        SWITCH, &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH, &Gui,  NOGUI, 'N', "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ RADIO, 0, "select" },
{ ITEM, 0, "<(LT:1)", "LT" },
{ ITEM, 1, "<=(LE:2)", "LE" },
{ ITEM, 2, ">(GT:3)", "GT" },
{ ITEM, 3, ">=(GE:4)", "GE" },
{ ITEM, 4, "=(EQ:5)", "EQ" },
{ RADIOEND, -1 },
{ RADIO, 5, "effect" },
{ ITEM, 5, "cut(c)", "cut" },
{ ITEM, 6, "fill(f)", "fill" },
{ ITEM, 7, "add(a)", "add" },
{ ITEM, 8, "sub(s)", "sub" },
{ RADIOEND, -1 },
{ ENTRY, 9, "threthold value", "value=%d" }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* image width and height */
static unsigned int MaskValue;	/* Mask depth at the selected point */

/**********************************************************************/
/*                                                                    */
/*             function lower()                                       */
/*                                                                    */
/*             unsigned int val    :                                  */
/*                                                                    */
/*             return  int                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
int lower( unsigned int val )
{
	if( GET_MASK( val ) < MaskValue ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function lowereq()                                     */
/*                                                                    */
/*             unsigned int val    :                                  */
/*                                                                    */
/*             return  int                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
int lowereq( unsigned int val )
{
	if( GET_MASK( val ) <= MaskValue ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function upper()                                       */
/*                                                                    */
/*             unsigned int val    :                                  */
/*                                                                    */
/*             return  int                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
int upper( unsigned int val )
{
	if( GET_MASK( val ) > MaskValue ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function uppereq()                                     */
/*                                                                    */
/*             unsigned int val    :                                  */
/*                                                                    */
/*             return  int                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
int uppereq( unsigned int val )
{
	if( GET_MASK( val ) >= MaskValue ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function equal()                                       */
/*                                                                    */
/*             unsigned int val    :                                  */
/*                                                                    */
/*             return  int                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
int equal( unsigned int val )
{
	if( GET_MASK( val ) == MaskValue ){
		return( 1 );
	} else {
		return( 0 );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function cut()                                         */
/*                                                                    */
/*             unsigned int *p     :                                  */
/*                                                                    */
/*             return NONE                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
void cut( unsigned int *p )
{
	*p &= ~MASK;
}

/**********************************************************************/
/*                                                                    */
/*             function fill()                                        */
/*                                                                    */
/*             unsigned int *p     :                                  */
/*                                                                    */
/*             return NONE                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
void fill( unsigned int *p )
{
	*p |= MASK;
}

/**********************************************************************/
/*                                                                    */
/*             function add()                                         */
/*                                                                    */
/*             unsigned int *p     :                                  */
/*                                                                    */
/*             return NONE                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
void add( unsigned int *p )
{
	unsigned int v;

	v = GET_MASK(*p);
	v += AddVal;
	if( v > 0xff ){
		v = 0xff;
	}
	*p = ((*p & ~MASK) | (v << 24));
}

/**********************************************************************/
/*                                                                    */
/*             function sub()                                         */
/*                                                                    */
/*             unsigned int *p     :                                  */
/*                                                                    */
/*             return NONE                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
void sub( unsigned int *p )
{
	int v;

	v = (int)GET_MASK(*p);
	v -= AddVal;
	if( v < 0 ){
		v = 0x00;
	}
	*p = ((*p & ~MASK) | (v << 24));
}



/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *data, *p;
	int i;
	int (*check_lim)( unsigned int );
	void (*change_mask)( unsigned int * );

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	data = file->buf->data;

	if( sb->selnum != SKPOINT ){
		return( EILLSELECTION );
	}
	check_range( sb, Width, Height, 1, NEGLECT );
	if( sb->defnum < 1 ){
		return( EILLNUMBER );
	}
	if( file->buf->mask_kind != DEPTHMASK ){
		return( EILLMASK );
	}

	switch( LimMode ){
	case LIMLOWER:
		check_lim = lower;
		break;
	case LIMLOWEQ:
		check_lim = lowereq;
		break;
	case LIMUPPER:
		check_lim = upper;
		break;
	case LIMUPPEQ:
		check_lim = uppereq;
		break;
	case LIMEQUAL:
		check_lim = equal;
		break;
	default:
		return( -1 );
	}
	switch( ChgMode ){
	case CHGCUT:
		change_mask = cut;
		break;
	case CHGFILL:
		change_mask = fill;
		break;
	case CHGADD:
		change_mask = add;
		break;
	case CHGSUB:
		change_mask = sub;
		break;
	default:
		return( -1 );
	}


	p = &data[sb->def[0].y * Width + sb->def[0].x];
	MaskValue = GET_MASK( *p );

	for( i = 0, p = data; i < Height * Width; i++, p++ ){
		if( (*check_lim)( *p ) ){
			(*change_mask)( p );
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}

