/**********************************************************************/
/*                                                                    */
/*            paste.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "plug.h"

char Name[256] = { "noname" };
int Rate = 100;

#define RESIZE	1
#define REAL	0
#define MULTI	2
/* add D2 */
#define XRATE	3
#define YRATE	4
/* end D2 */
int Resize = REAL;

#define NONE		0
#define XMIRROR		1
#define YMIRROR		2
#define XYMIRROR	3
int MirrorFlag = NONE;

int Xajst = 0;
int Yajst = 0;

int Gui = NOGUI;
int List = 0;
void search_file(FILE *, int, int, char *);

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "paste" };
char *Comment = { "paste the cut-file(*.cut)" };

PLUGITEM Item[] = {
{ "rate", 	GETINT,  &Rate,	 1, 100,   "rate of merge" },
{ "resize", 	SWITCH,  &Resize, RESIZE, 'R', "resize the selected area"  },
{ "real", 	SWITCH,  &Resize, REAL, 'L', "not resize and paste real size" },
{ "multi",      SWITCH,  &Resize, MULTI, 'M', "multi paste mode" },
/* add D2 */
{ "xrate",      SWITCH,  &Resize, XRATE, 'X', "resize by X axis rate" },
{ "yrate",      SWITCH,  &Resize, YRATE, 'Y', "resize by Y axis rate" },
/* end D2 */
{ "xmirror",	SWITCH,  &MirrorFlag, XMIRROR, 'X', "multi mode X-mirror" },
{ "ymirror",	SWITCH,  &MirrorFlag, YMIRROR, 'Y', "multi mode Y-mirror" },
{ "xymirror",	SWITCH,  &MirrorFlag, XYMIRROR, 'D', "multi mode XY-mirror" },
{ "nomirror",	SWITCH,  &MirrorFlag, NONE, 'N', "multi mode NO mirror" },
{ "xajst",	GETINT,  &Xajst,  -600, 600, "multi X ajustment" },
{ "yajst",	GETINT,  &Yajst,  -600, 600, "multi Y ajustment" },
{ "list",	SWITCH,  &List, 1, 0,      "list all cutfile in #lwp" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" },
{ "name", 	GETSTR,  &Name, 64, 0,    "target cut file name" },
{ NULL, 	GETSTR,  &Name, 64, 0,    "target cut file name" } };
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "merge rate ", "rate=%d" },
{ NOTCL | ENTRY, 15, "object cut file name", "name=%s" },
{ TCLLIST, 15, "object cut file name", "name=%s" },
{ TCLDIR, 0, (char *)search_file, NULL },
{ TCLLISTEND, 15, "selected file is", "cut=%s" },
{ RADIO, 1, "Mode " },
{ ITEM, 1, "re-size(r)", "resize" },
{ ITEM, 2, "real-size(l)", "real" },
{ ITEM, 3, "multi-mode(m)", "multi" },
{ ITEM, 4, "x-rate(x)", "xrate" },
{ ITEM, 5, "y-rate(y)", "yrate" },
{ RADIOEND, -1 },
{ RADIO, 6, "Multi Mode Mirror Option" },
{ ITEM, 6, "X mirrro(x)", "xmirror" },
{ ITEM, 7, "Y mirror(y)", "ymirror" },
{ ITEM, 8, "XY mirror(d)", "xymirror" },
{ ITEM, 9, "NO mirror(n)", "nomirror" },
{ RADIOEND, -1 },
{ ENTRY, 10, "Multi Mode X ajustment ", "xajst=%d" },
{ ENTRY, 11, "Multi Mode Y ajustment ", "yajst=%d" },
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

int Right, Left, Up, Down;	/* cut max-min position in the original */
int Width;			/* width of the image */
int Height;			/* height of the image */

#define WORD unsigned short
struct RawHeader {
	WORD spec;	/* LSB or MSB resolve flag */
	WORD width;	/* width */
	WORD height;	/* height */
	WORD mask;	/* mask kind */
};

#define Data1(q,p)	data1[(p)*w+(q)]
#define Data2(q,p)	data2[(p)*w+(q)]

/**********************************************************************/
/*                                                                    */
/*             function rev()                                         */
/*                                                                    */
/*             WORD v        : reverse short                          */
/*                                                                    */
/*             return  WORD  : reversed short                         */
/*                                                                    */
/*             function:  Reverse the byte order at 2byte WORD.       */
/*                                                                    */
/**********************************************************************/
WORD rev( WORD v )
{
	WORD ret;

	ret = ((v & 0xff00) >> 8);
	ret |= ((v & 0x00ff) << 8);
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function put_color()                                   */
/*                                                                    */
/*             unsigned int *d : dest color pixel                     */
/*             unsigned int *s : src  color pixel                     */
/*             int rt          : merge rate                           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Put and merge the dest pixel by the src      */
/*                      pixel color.(Used file to image)              */
/**********************************************************************/
void put_color( unsigned int *d, unsigned int *s, int rt )
{
	int sr, sg, sb, dr, dg, db;

	dr = GET_RED(*d);
	sr = GET_RED(*s);
	dg = GET_GREEN(*d);
	sg = GET_GREEN(*s);
	db = GET_BLUE(*d);
	sb = GET_BLUE(*s);
	dr = (dr * (100-rt) + sr * rt) / 100;
	dg = (dg * (100-rt) + sg * rt) / 100;
	db = (db * (100-rt) + sb * rt) / 100;
	*d = ((*d & MASK) | (dr << 16) | (dg << 8) | db);
}


/**********************************************************************/
/*                                                                    */
/*             function realsize_proc()                               */
/*                                                                    */
/*             unsigned int *data : original image buffer             */
/*             unsigned int *buff : read cut file image               */
/*             int fwidth         : read cut image width              */
/*             int fheight        : read cut image height             */
/*             int mask           : mask kind of the original image   */
/*                                                                    */
/*             return int :   0==success  else==errorcode             */
/*                                                                    */
/*             function: Do the real size paste from cut file to the  */
/*                      original image.                               */
/**********************************************************************/
int realsize_proc( unsigned int *data, unsigned int *buff, 
			int fwidth, int fheight, int mask )
{
	unsigned int *s, *d;
	struct box {
		int left;
		int right;
		int up;
		int down;
	} src, dst;
	int dwidth, dheight;
	int di, si, dj, sj;
	unsigned int rt;

	dwidth = Right - Left + 1;
	dheight = Down - Up + 1;
	if( dwidth > fwidth ){
		dst.left = (dwidth - fwidth) / 2 + Left;
		src.left = 0;
		dst.right = dst.left + fwidth;
		src.right = fwidth;
	} else {
		dst.left = Left;
		src.left = (fwidth - dwidth) / 2;
		dst.right = Right + 1;
		src.right = src.left + dwidth;
	}
	if( dheight > fheight ){
		dst.up = (dheight - fheight) / 2 + Up;
		src.up = 0;
		dst.down = dst.up + fheight;
		src.down = fheight;
	} else {
		dst.up = Up;
		src.up = (fheight - dheight) / 2;
		dst.down = Down + 1;
		src.down = src.up + dheight;
	}

	for( di = dst.up, si = src.up; di < dst.down; di++, si++ ){
		d = data + di * Width; s = buff + si * fwidth;
		for( dj = dst.left, sj = src.left; dj < dst.right; dj++, sj++ ){
			if( mask == SIMPLEMASK ){
				if( MASKOFF( d[dj] ) && MASKOFF( s[sj] ) ){
					if( Rate == 100 ){
						d[dj] = ((d[dj] & MASK ) | (s[sj] & ~MASK));
					} else {
						put_color( &d[dj], &s[sj], Rate );
					}
				}
			} else {
				if( MASKOFF( d[dj] ) ){
					rt = GET_MASK( s[sj] ) * (unsigned int)Rate / 256;
					put_color( &d[dj], &s[sj], rt );
				}
			}
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function put_color2()                                  */
/*                                                                    */
/*             unsigned int *d  : dest color pixel                    */
/*             unsigned int *s1 : src1 color pixel                    */
/*             unsigned int *s2 : src2 color pixel                    */
/*             float rate       : merge rate                          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Put the dest pixel by merge the src1 and src2*/
/*                      pixel color.(Used set tmp buff calc)          */
/**********************************************************************/
void put_color2( unsigned int *d, unsigned int *s1, unsigned int *s2,
			 float rate )
{
	unsigned int sr, sg, sb, sm, dr, dg, db, dm;

	dr = GET_RED(*s1);
	sr = GET_RED(*s2);
	dg = GET_GREEN(*s1);
	sg = GET_GREEN(*s2);
	db = GET_BLUE(*s1);
	sb = GET_BLUE(*s2);
	dm = GET_MASK(*s1);
	sm = GET_MASK(*s2);
	dr = (unsigned int)((float)dr * (1.0-rate) + (float)sr * rate);
	dg = (unsigned int)((float)dg * (1.0-rate) + (float)sg * rate);
	db = (unsigned int)((float)db * (1.0-rate) + (float)sb * rate);
	dm = (unsigned int)((float)dm * (1.0-rate) + (float)sm * rate); 
	*d = ((dm << 24) | (dr << 16) | (dg << 8) | db);
}



/**********************************************************************/
/*                                                                    */
/*             function resize_w()                                    */
/*                                                                    */
/*             unsigned int *data1 : dest image buffer(target)        */
/*             unsigned int *data2 : src image buffer(file)           */
/*             int w1              : dest width(target)               */
/*             int w2              : src width(file)                  */
/*             int h               : file(src) height                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Only width, resize the data1 image to the    */
/*                      target width(w1).                             */
/**********************************************************************/
void resize_w( unsigned int *data1, unsigned int *data2, 
			int w1, int w2, int h )
{
	float f, fi, fp;
	int i, j, ip;
	unsigned int *d1, *d2;

	f = (float)w2 / (float)w1;
	d1 = data1; d2 = data2;
	for( j = 0; j < h; j++ ){
		for( i = 0; i < w2; i++ ){
			fi = (float)i / f;
			ip = (int)fi;
			fp = fi - (float)ip;
			if( ip == w1 - 1 ){
				put_color2( &d2[i], &d1[ip], &d1[ip], fp );
			} else {
				put_color2( &d2[i], &d1[ip], &d1[ip+1], fp );
			}
		}
		d1 += w1; d2 += w2;
	}
}


/**********************************************************************/
/*                                                                    */
/*             function resize_h()                                    */
/*                                                                    */
/*             unsigned int *data1 : dest image buffer(target)        */
/*             unsigned int *data2 : src image buffer(after resize_w) */
/*             int h1              : dest height(target)              */
/*             int h2              : src height(file)                 */
/*             int w               : resized width(target width)      */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Only height, resize the data1 image to the   */
/*                      target height(h1)                             */
/**********************************************************************/
void resize_h( unsigned int *data1, unsigned int *data2, 
			int h1, int h2, int w )
{
	float f, fi, fp;
	int i, j, ip;

	f = (float)h2 / (float)h1;
	for( j = 0; j < w; j++ ){
		for( i = 0; i < h2; i++ ){
			fi = (float)i / f;
			ip = (int)fi;
			fp = fi - (float)ip;
			if( ip == h1 - 1 ){
				put_color2( &Data2(j,i), &Data1(j,ip), &Data1(j,ip), fp );
			} else {
				put_color2( &Data2(j,i), &Data1(j,ip), &Data1(j,ip+1), fp );
			}
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*             function xyrate_proc()                                 */
/*                                                                    */
/*             unsigned int *data : original image buffer             */
/*             unsigned int *buff : read cut file image               */
/*             int fwidth         : read cut image width              */
/*             int fheight        : read cut image height             */
/*             int mask           : mask kind of the original image   */
/*                                                                    */
/*             return int :   0==success  else==errorcode             */
/*                                                                    */
/*             function: Do the resize paste, resize cut file image   */
/*                      by selected X or Y size.                      */
/*             D2: create new                                         */
/**********************************************************************/
int xyrate_proc( unsigned int *data, unsigned int *buff, 
			int fwidth, int fheight, int mask )
{
	unsigned int *buff2, *buff3;
	int len;
	int dwidth, dheight, nv;
	int di, si, dj, sj;
	unsigned int *s, *d;
	struct box {
		int left;
		int right;
		int up;
		int down;
	} src, dst;

	dwidth = Right - Left + 1;
	dheight = Down - Up + 1;

	if( Resize == XRATE ){
	     len = dwidth * fheight * sizeof(unsigned int);
	     if( (buff2 = (unsigned int *)malloc( len )) == NULL ){
		  return( EMALLOC );
	     }
	     resize_w( buff, buff2, fwidth, dwidth, fheight );
	     
	     nv = dwidth * fheight / fwidth;
	     len = dwidth * nv * sizeof(unsigned int);
	     if( (buff3 = (unsigned int *)malloc( len )) == NULL ){
		  return( EMALLOC );
	     }
	     resize_h( buff2, buff3, fheight, nv, dwidth );
	     fheight = nv;

	     if( dheight > fheight ){
		  dst.up = (dheight - fheight) / 2 + Up;
		  src.up = 0;
		  dst.down = dst.up + fheight;
		  src.down = fheight;
	     } else {
		  dst.up = Up;
		  src.up = (fheight - dheight) / 2;
		  dst.down = Down + 1;
		  src.down = src.up + dheight;
	     }
	     dst.left = Left; src.left = 0;
	     dst.right = Right + 1; src.right = dwidth;
	     fwidth = dwidth;
	} else if( Resize == YRATE ){
	     nv = dheight * fwidth / fheight;
	     len = nv * fheight * sizeof(unsigned int);
	     if( (buff2 = (unsigned int *)malloc( len )) == NULL ){
		  return( EMALLOC );
	     }
	     resize_w( buff, buff2, fwidth, nv, fheight );
	     fwidth = nv;
	     
	     len = fwidth * dheight * sizeof(unsigned int);
	     if( (buff3 = (unsigned int *)malloc( len )) == NULL ){
		  return( EMALLOC );
	     }
	     resize_h( buff2, buff3, fheight, dheight, fwidth );

	     if( dwidth > fwidth ){
		  dst.left = (dwidth - fwidth) / 2 + Left;
		  src.left = 0;
		  dst.right = dst.left + fwidth;
		  src.right = fwidth;
	     } else {
		  dst.left = Left;
		  src.left = (fwidth - dwidth) / 2;
		  dst.right = Right + 1;
		  src.right = src.left + dwidth;
	     }
	     dst.up = Up; src.up = 0;
	     dst.down = Down + 1; src.down = dheight;
	} else {
	     return -1;
	}

	for( di = dst.up, si = src.up; di < dst.down; di++, si++ ){
		d = data + di * Width; s = buff3 + si * fwidth;
		for( dj = dst.left, sj = src.left; dj < dst.right; dj++, sj++ ){
		     if( MASKOFF( d[dj] ) && MASKOFF( s[sj] ) ){
			  if( Rate == 100 ){
			       d[dj] = ((d[dj] & MASK ) | (s[sj] & ~MASK));
			  } else {
			       put_color( &d[dj], &s[sj], Rate );
			  }
		     }
		}
	}

	free( buff2 );
	free( buff3 );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function resize_proc()                                 */
/*                                                                    */
/*             unsigned int *data : original image buffer             */
/*             unsigned int *buff : read cut file image               */
/*             int fwidth         : read cut image width              */
/*             int fheight        : read cut image height             */
/*             int mask           : mask kind of the original image   */
/*                                                                    */
/*             return int :   0==success  else==errorcode             */
/*                                                                    */
/*             function: Do the resize paste, resize cut file image   */
/*                      by selected size of original and paste.       */
/**********************************************************************/
int resize_proc( unsigned int *data, unsigned int *buff, 
			int fwidth, int fheight, int mask )
{
	unsigned int *buff2, *buff3;
	int len;
	int dwidth, dheight;
	int di, si, dj, sj;
	unsigned int *s, *d;
	unsigned int tmp;

	dwidth = Right - Left + 1;
	dheight = Down - Up + 1;

	len = dwidth * fheight * sizeof(unsigned int);
	if( (buff2 = (unsigned int *)malloc( len )) == NULL ){
		return( EMALLOC );
	}
	resize_w( buff, buff2, fwidth, dwidth, fheight );

	len = dwidth * dheight * sizeof(unsigned int);
	if( (buff3 = (unsigned int *)malloc( len )) == NULL ){
		return( EMALLOC );
	}

	resize_h( buff2, buff3, fheight, dheight, dwidth );

	for( di = Up, si = 0; di <= Down; di++, si++ ){
		d = data + di * Width; s = buff3 + si * dwidth;
		for( dj = Left, sj = 0; dj <= Right; dj++, sj++ ){
			if( MASKOFF( d[dj] ) ){
				tmp = GET_MASK(s[sj]);
				if( tmp < 0x80 ){
					if( Rate == 100 ){
						d[dj] = ((d[dj] & MASK ) | (s[sj] & ~MASK));
					} else {
						put_color( &d[dj], &s[sj], Rate );
					}
				}
			}
		}
	}
	free( buff2 );
	free( buff3 );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function multi_proc()                                  */
/*                                                                    */
/*             unsigned int *data : original image buffer             */
/*             unsigned int *buff : read cut file image               */
/*             int fwidth         : read cut image width              */
/*             int fheight        : read cut image height             */
/*             int mask           : mask kind of the original image   */
/*                                                                    */
/*             return int :   0==success  else==errorcode             */
/*                                                                    */
/*             function: Do the multi paste, repeat image with no     */
/*                      resize.                                       */
/**********************************************************************/
int multi_proc( unsigned int *data, unsigned int *buff, 
			int fwidth, int fheight, int mask )
{
	int i, j;
	unsigned int *p, *q;
	int xs, ys;
	int sj, si;
	int xaj, yaj;

	if( Xajst >= 0 ){
		xaj = Xajst % fwidth;
	} else {
		xaj = fwidth - ((Xajst * (-1)) % fwidth);
	}
	if( Yajst >= 0 ){
		yaj = Yajst % fheight;
	} else {
		yaj = fheight - ((Yajst * (-1)) % fheight);
	}
	p = data;
	ys = 1; si = yaj;
	for( i = 0; i < Height; i++ ){
		q = &buff[si * fwidth];
		xs = 1; sj = xaj;
		for( j = 0; j < Width; j++, p++ ){
			if( MASKOFF(*p) ){
				if( GET_MASK(q[sj]) < 0x80 ){
					*p = q[sj];
				}
			}
			sj += xs;
			if( sj < 0 ){
				sj = 0; xs = 1;
			} else if( sj >= fwidth ){
				if( MirrorFlag & 1 ){
					sj = fwidth - 1; xs = -1;
				} else {
					sj = 0; xs = 1;
				}
			}
		}
		si += ys;
		if( si < 0 ){
			si = 0; ys = 1;
		} else if( si >= fheight ){
			if( MirrorFlag & 2 ){
				si = fheight - 1; ys = -1;
			} else {
				si = 0; ys = 1;
			}
		}
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function paste()                                       */
/*                                                                    */
/*             unsigned int *data  : image buffer                     */
/*                                                                    */
/*             return  int  : 0==success  else==errorcode             */
/*                                                                    */
/*             function: Prepare the paste, read cut file and call    */
/*                      image processor.                              */
/**********************************************************************/
int paste( unsigned int *data )
{
	int fd;
	struct RawHeader header;
	int len;
	int ret;
	int fwidth, fheight;
	unsigned int *buff;
	char filename[256];
	char *envstr;

	ret = -1;
	if( (envstr = getenv( "GEDO_LWP" )) == NULL ){
		strcpy( filename, "./" );
	} else {
		strcpy( filename, envstr );
	}
	if( filename[strlen(filename) -1] != '/' ){
		strcat( filename, "/" );
	}
	strcat( filename, Name );
	strcat( filename, ".cut" );

	fprintf( stderr, "loading %s...\n", filename );
	if( (fd = open( filename, O_RDONLY )) == -1 ){
		return( EOPENFILE );
	}
	if( read( fd, &header, sizeof(struct RawHeader) ) != sizeof(struct RawHeader) ){
		return( EREADFILE );
	}
	if( header.spec == 1 ){
		fwidth = (int)header.width;
		fheight = (int)header.height;
	} else if( header.spec == 0x0100 ){
		fwidth = (int)rev( header.width );
		fheight = (int)rev( header.height );
	} else {
		return( EILLFORM );
	}
	len = fwidth * fheight * sizeof(unsigned int);
	if( (buff = (unsigned int *)malloc( len )) == NULL ){
		return( EMALLOC );
	}
	if( read( fd, buff, len ) <= 0 /*!= len*/ ){
		return( EREADFILE );
	}
	close( fd );

	switch( Resize ){
	case RESIZE:
		if( header.mask == SIMPLEMASK ){
			ret = resize_proc( data, buff, fwidth, fheight, header.mask );
		} else {
			ret = EILLMASK;
		}
		break;
	case REAL:
		ret = realsize_proc( data, buff, fwidth, fheight, header.mask );
		break;
	case MULTI:
		ret = multi_proc( data, buff, fwidth, fheight, header.mask );
		break;
/* D2 add */
	case XRATE:
	case YRATE:
		if( header.mask == SIMPLEMASK ){
		     ret = xyrate_proc( data, buff, fwidth, fheight, header.mask );
		} else {
		     ret = EILLMASK;
		}
		break;
/* D2 end */
	}
	free( buff );
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	char *str;
	unsigned int *data, *p;
	int i, j, ret;

	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}

	str = file->mes->u.data;
	Width = file->buf->width;
        Height = file->buf->height;
	data = file->buf->data;

	Left = Width; Right = 0; Up =Height; Down = 0;
	for( i = 0, p = data; i < Height; i++ ){
		for( j = 0; j < Width; j++, p++ ){
			if( MASKOFF( *p ) ){
				if( Left > j ){
					Left = j;
				}
				if( Right < j ){
					Right = j;
				}
				if( Up > i ){
					Up = i;
				}
				if( Down < i ){
					Down = i;
				}
			}
		}
	}
	if( Left > Right ){
		return( EILLSELECT );
	}
	if( Up > Down ){
		return( EILLSELECT );
	}
	ret = paste( data );
	if( ret != 0 ){
		strcpy( file->mes->u.data, strerror( errno ) );
		file->mes->datasize = strlen( file->mes->u.data );
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;
	int ret;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		ret = line_config( file );
		if( ret == 0 && List ){
			search_dir( getenv( "GEDO_LWP" ), "cut" );
			List = 0;
		}
		return( ret );
	}
	return( 0 );
}

void search_file(FILE *fp, int var, int kind, char * arg)
{
	search_tcldir( fp, var, getenv( "GEDO_LWP" ), "cut" );
}
