/**********************************************************************/
/*                                                                    */
/*              pic.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>

#include "plug.h"
int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "pic" };
char *Comment = { "Personal Links pic format(*.pic) loader/saver" };

PLUGITEM Item[];	/* no configuration */
int      ItemNum = 0;

struct Tags Tag[];
int TagNum = 0;

DEFAULTSTATUS;
/********************** description end ********************/

#define BYTE unsigned char
#define WORD unsigned short

struct PicHeader {
	long OS_flag;		/* OS UNIX==0 */
	BYTE format;		/* data save format == 0 */
	BYTE n_component;	/* byte per pixel =='3' */
	BYTE version;		/* version=='00' */
	BYTE extra_flag;	/* LINKS flag only  =='01' */
	BYTE block_size[4];	/* I/O block size =='    ' */
	BYTE x_offset[4];	/* start position X =='   0' */
	BYTE y_offset[4];	/* start position Y =='   0' */
	BYTE x_size[4];		/* image width */
	BYTE y_size[4];		/* image height */
	BYTE component[16];	/* image order =='rgb\0' */
	BYTE reserve[212];
	BYTE filename[0x80];	/* filename */
	BYTE date[0x80];	/* create date by UNIX format */
	BYTE reserve2[256*2];
};

int LSB_flag;		/* not used */

/**********************************************************************/
/*                                                                    */
/*             function read_header()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ HEADER of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_header( struct File *file )
{
	struct PicHeader header;

	if( (file->fd = open( file->name, O_RDONLY )) == -1 ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EOPENFILE );
	}
	if( read( file->fd, &header, sizeof(struct PicHeader) ) != sizeof(struct PicHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EREADFILE );
	}
	if( header.format != 0 ){
		file->mes->datasize = 0;
		return( EILLFORM );
	}
	if( header.n_component != '3' ){
		file->mes->datasize = 0;
		return( EILLFORM );
	}
/*	if( header.extra_flag & ~1 == 0 ){
		file->mes->datasize = 0;
		return( EILLFORM );
	}*/
	if( strcmp( (char *)header.component, "rgb" ) != 0 ){
		file->mes->datasize = 0;
		return( EILLFORM );
	}

	file->buf->width = axtol( header.x_size, 4 );
	file->buf->height = axtol( header.y_size, 4 );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function read_image()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ IMAGES of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_image( struct File *file )
{
	long size;
	unsigned char *buff;
	struct Shm *buf;
	int i, j;

	buf = file->buf;
	size = buf->width * buf->height * 3;
	if( (buff = malloc( size )) == NULL ){
		file->mes->datasize = 0;
		close( file->fd );
		return( EMALLOC );
	}
	if( read( file->fd, buff, size ) != size ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		close( file->fd );
		return( EREADFILE );
	}
	close( file->fd );
	for( i = 0, j = 0; i < buf->width * buf->height; i++ ){
		buf->data[i] = (((unsigned int)(buff[j++]) << 16) & 0x00ff0000 ); 
		buf->data[i] |= (((unsigned int)(buff[j++]) << 8) & 0x0000ff00 );
		buf->data[i] |= (((unsigned int)buff[j++]) & 0x000000ff);
	}
	buf->mask_kind = SIMPLEMASK;
	free( buff );
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function write_header()                                */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE HEADER of the SPEC format to file       */
/*          B0: rewrite logic                                         */
/*                                                                    */
/**********************************************************************/
int write_header( struct File *file )
{
	struct PicHeader header;
	time_t t;
	char *p;

	if( (file->fd = open( file->name, O_RDWR | O_TRUNC )) == -1 ){
		if( (file->fd = open( file->name, O_RDWR | O_CREAT, 0664)) == -1 ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
		memset( &header, 0, sizeof(struct PicHeader) );
	} else {
		if( read( file->fd, &header, sizeof(struct PicHeader) ) != sizeof(struct PicHeader) ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EREADFILE );
		}
		if( lseek( file->fd, 0L, 0 ) == -1L ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( ESEEK );
		}			
	}

	header.n_component = '3';
	header.extra_flag = 1;
	memcpy( header.block_size, "   0", 4 );
	memcpy( header.x_offset, "   0", 4 );
	memcpy( header.y_offset, "   0", 4 );
	ltoax( file->buf->width, header.x_size, 4 );
	ltoax( file->buf->height, header.y_size, 4 );
	memcpy( header.component, "rgb", 3 );
	strcpy( (char *)&header.filename, file->name );

	time( &t );
	if( (p = ctime( &t )) ){
		strcpy( (char *)&header.date, p );
	}

	if( write( file->fd, &header, sizeof(struct PicHeader) ) != sizeof(struct PicHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EWRITEFILE );
	}

	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE IMAGES of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_image( struct File *file )
{
	long size;
	unsigned char *buff;
	struct Shm *buf;
	int i, j;

	buf = file->buf;
	size = buf->width * buf->height * 3;
	if( (buff = malloc( size )) == NULL ){
		file->mes->datasize = 0;
		close( file->fd );
		return( EMALLOC );
	}
	for( i = 0, j = 0; i < buf->width * buf->height; i++ ){
		buff[j++] = (unsigned char)GET_RED(buf->data[i]);
		buff[j++] = (unsigned char)GET_GREEN(buf->data[i]);
		buff[j++] = (unsigned char)GET_BLUE(buf->data[i]);
	}
	if( write( file->fd, buff, size ) != size ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		close( file->fd );
		return( EWRITEFILE );
	}
	close( file->fd );
	free( buff );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( file->mes->command == CLOAD ){
		if( str[0] == '\0' ){
			printf( "loader pic: no configration parameter\n" );
			file->mes->u.data[0] = '\0';
		} else {
			return( line_config( file ) );
		}
	} else if( file->mes->command == CSAVE ){
		if( str[0] == '\0' ){
			printf( "saver pic: no configration parameter\n" );
			file->mes->u.data[0] = '\0';
		} else {
			return( line_config( file ) );
		}
	}
	return( 0 );
}
