/**********************************************************************/
/*                                                                    */
/*             plug.c   gedo PLUG transaction routine                 */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*             11.21,1996 A0 add sleep in end_proc                    */
/*              4.23,1997 B2 change for multiple viewer               */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <memory.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>

#include "plug.h"

static char *WarningForm	/* warning string format */
= " %s %s warning -- %s\n";
static char *NKind[4] 		/* kind name string */
= { "", "loader", "selecter", "effecter" };
static int CtrlCFlag = 0;	/* ^C detection */
extern jmp_buf jmpbuf;

/* B2 add */
#define MAXVIEWER	10		/* maximum of viewer(2) process */
extern struct File File[MAXVIEWER];	/* viewer database */
extern int FileNum;			/* number of viewers */

/**********************************************************************/
/*                                                                    */
/*             function ctrl_c()                                      */
/*                                                                    */
/*             int signo  : signal No.                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: ^C signal handler.                           */
/*                                                                    */
/**********************************************************************/
void ctrl_c( int signo )
{
	signal( SIGINT, ctrl_c );
	CtrlCFlag = 1;
}

void ctrl_c_break( int signo )
{
	signal( SIGINT, ctrl_c_break );
	longjmp( jmpbuf, 1 );
}

/* add C1 */
void usr1( int signo )
{
	signal( SIGUSR1, usr1 );
}

/**********************************************************************/
/*                                                                    */
/*             function set_shm()                                     */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int : 0==success  errorcode=fail               */
/*                                                                    */
/*             function: Set the viewer infomation by the IPC-ID      */
/*                      from GEDO Message.                            */
/*           B2  rewrite for multi window system                      */
/*                                                                    */
/**********************************************************************/
int set_shm( struct proc *proc )
{
	int i;
	struct shmid_ds dat;
	struct Message *mes;

	mes = proc->file->mes;
	if( proc->file->shmid == mes->shmid ){
		return( 0 );
	}
	for( i = 0; i < FileNum; i++ ){
		if( mes->shmid == File[i].shmid ){
			File[i].mes = mes;
			proc->file = &File[i];
			return( 0 );
		}
	}
	if( i == MAXVIEWER ){
		return( EVIEWERMAX );
	}
	File[i].mes = mes;
	File[i].shmid = mes->shmid;
	proc->file = &File[i];
	FileNum++;

	if( shmctl( proc->file->mes->shmid, IPC_STAT, &dat ) == -1 ){
		return( ESHMAT );
	}
	proc->shm_max = (dat.shm_segsz - sizeof(struct Shm)) / sizeof(unsigned int);
	proc->file->buf = (struct Shm *)shmat( proc->file->mes->shmid, NULL, 0 );
	if( proc->file->buf == (struct Shm *)-1 ){
		return( ESHMAT );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function end_proc()                                    */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*             int errno         : error code occured and send        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: End the viewers and send end message to GEDO.*/
/*                                                                    */
/**********************************************************************/
void end_proc( struct proc *proc, int errno )
{
	struct Message *mes;
	int i;

	mes = proc->file->mes;
	if( errno ){
		sprintf( mes->u.data, "fatal error: %s suicided", proc->spec );
		mes->status  = STFATAL;
	} else {
		sprintf( mes->u.data, "%s killed", proc->spec );
		mes->status  = STDONE;
	}

	for( i = 0; i < FileNum; i++ ){
		if( File[i].buf ){
			shmdt( (char *)File[i].buf );
		}
	}

	mes->server = proc->kind;
	mes->errorcode  = errno;
	mes->datasize = strlen( mes->u.data );
	mes->mtype = (long)mes->pid[SMAIN];
	msgsnd( proc->mesid, (struct msgbuf *)mes, sizeof(struct Message) - sizeof(long), 0 );
/* A0 add Heavy GEDO msg snd/rcv timing cannot work */
	sleep( 1 );
/* A0 end */
	exit( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function reply()                                       */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*             int dest          : Destination of the reply           */
/*                                                                    */
/*             return  int : 0==success  -1==fail                     */
/*                                                                    */
/*             function: Reply the result to <dest> process.          */
/*                                                                    */
/**********************************************************************/
int reply( struct proc *proc, int dest )
{
	struct Message *mes;

	if( CtrlCFlag ){
		CtrlCFlag = 0;
		return( 0 );
	}
	mes = proc->file->mes;
	mes->server = proc->kind;
	if( mes->pid[dest] <= 1 ){
		return( -1 );
	}
	mes->mtype = (long)mes->pid[dest];
	if( msgsnd( proc->mesid, (struct msgbuf *)mes, sizeof(struct Message) - sizeof(long), 0 ) == -1 ){
		end_proc( proc, EPIPEWRITE );
	}
#if  DEBUG
	printf( "%s:reply:kill dest=%d pid=%d\n", proc->spec, dest, mes->pid[dest] );
#endif
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function load_command()                                */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int : 0==success                               */
/*                                                                    */
/*             function: load command process.                        */
/*                                                                    */
/**********************************************************************/
int load_command( struct proc *proc )
{
	struct Message *mes;
	int ret;

	mes = proc->file->mes;
	mes->receiver = SMAIN;
	switch( mes->subcommand ){
	case SCEXEC: 
		if( (ret = set_shm( proc )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			break;
		}	
		strcpy( proc->file->name, mes->u.data );
		if( (ret = read_header( proc->file )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			if( mes->datasize == 0 ){
				sprintf( mes->u.data, WarningForm, NKind[proc->kind], proc->spec, "read_header error" );
			}
			break;
		}
		if( proc->file->buf->width * proc->file->buf->height >= proc->shm_max ){
			mes->status = STWARNING;
			mes->errorcode = ETOOLARGE;
			sprintf( mes->u.data, WarningForm, NKind[proc->kind], proc->spec, "read_header: too large" );
			break;
		}
		if( (ret = read_image( proc->file )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			if( mes->datasize == 0 ){
				sprintf( mes->u.data, WarningForm, NKind[proc->kind], proc->spec, "read_image error" );
			}
			break;
		}
		if( mes->comatr & ADISP ){
			mes->receiver = SVIEWER;
		}
		mes->status = STDONE;
		mes->errorcode = ENOERROR;
		break;
	case SCCONFIG:
		if( (ret = set_shm( proc )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			break;
		}
		ret = config( proc->file );
		if( ret < 0 ){
			mes->status = STWARNING;
			mes->errorcode = EILLCONFIG;
			mes->u.data[0] = '\0';
		} else if( ret == 0 || ret == GETSTATUS ){
			mes->status = STDONE;
			mes->errorcode = ENOERROR;
		} else {
			mes->status = STWARNING;
			mes->errorcode = ret;
		}
		break;
	    case SCHELP:
		help();
		mes->status = STDONE;
		mes->errorcode = ENOERROR;
		mes->u.data[0] = '\0';
		break;
	    default:
		mes->status = STWARNING;
		mes->errorcode = ENOSUBCOMMAND;
		sprintf( mes->u.data, "%s:subcommad %d is not here\n", proc->spec, mes->subcommand );
		break;
	}
	mes->datasize = strlen( mes->u.data ); 
	reply( proc, mes->receiver );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function save_command()                                */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int : 0==success                               */
/*                                                                    */
/*             function: save command process.                        */
/*                                                                    */
/**********************************************************************/
int save_command( struct proc *proc )
{
	struct Message *mes;
	int ret;

	mes = proc->file->mes;
	switch( mes->subcommand ){
	case SCEXEC: 
		if( (ret = set_shm( proc )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			break;
		}
		strcpy( proc->file->name, mes->u.data );
		if( (ret = write_header( proc->file )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			if( mes->datasize == 0 ){
				sprintf( mes->u.data, WarningForm, NKind[proc->kind], proc->spec, "write_header error" );
			}
			break;
		}
		if( (ret = write_image( proc->file )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			if( mes->datasize == 0 ){
				sprintf( mes->u.data, WarningForm, NKind[proc->kind], proc->spec, "write_image error" );
			}
			break;
		}
		mes->status = STDONE;
		mes->errorcode = ENOERROR;
		sprintf( mes->u.data, "%s:save %s done\n", proc->spec, proc->file->name );
		break;
	    case SCCONFIG:
		if( (ret = set_shm( proc )) != 0 ){ 
			mes->status = STWARNING;
			mes->errorcode = ret;
			break;
		}
/* B2 for config status and version */
		ret = config( proc->file );
		if( ret < 0 ){
			mes->status = STWARNING;
			mes->errorcode = EILLCONFIG;
			mes->u.data[0] = '\0';
		} else if( ret == 0 || ret == GETSTATUS ){
			mes->status = STDONE;
			mes->errorcode = ENOERROR;
		} else {
			mes->status = STWARNING;
			mes->errorcode = ret;
		}
		break;
	    case SCHELP:
		help();
		mes->status = STDONE;
		mes->errorcode = ENOERROR;
		mes->u.data[0] = '\0';
		break;
	    default:
		mes->status = STWARNING;
		mes->errorcode = ENOSUBCOMMAND;
		sprintf( mes->u.data, "%s:subcommad %d is not here\n", proc->spec, mes->subcommand );
		break;
	}
	mes->datasize = strlen( mes->u.data ); 
	reply( proc, SMAIN );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function  read_pos()                                   */
/*                                                                    */
/*             struct Message *mes : Message from GEDO(SKFILE)        */
/*             int *status         : error code                       */
/*                                                                    */
/*             return  struct Message * : Message developed POSLIST   */
/*                                                                    */
/*             function: Convert SKFILE Message to SKPOS Message.     */
/*                                                                    */
/**********************************************************************/
struct Message *read_pos( struct Message *mes, int *status )
{
	struct Shead sh;
	FILE *fp;
	struct Message *ret;
	int size;

	if( (fp = fopen( mes->u.sf.file, "r" )) == NULL ){
		*status = EOPENFILE;
		return( NULL );
	}
	if( fread( &sh, sizeof(struct Shead), 1, fp ) != 1 ){
		*status = EREADFILE;
		return( NULL );
	}
	if( sh.defnum <= 0 ){
		*status = EUSRBREAK;
		return( NULL );
	}
	mes->u.sf.defnum = sh.defnum;
	size = sizeof(struct Message) + sizeof(struct Pos) * (sh.defnum - 30);
	if( (ret = malloc( size )) == NULL ){
		*status = EMALLOC;
		return( NULL );
	}
	memcpy( ret, mes, sizeof(struct Message) );
	if( fread( ret->u.sb.def, sizeof(struct Pos), (size_t)sh.defnum, fp ) != (size_t)sh.defnum ){
		*status = EREADFILE;
		return( NULL );
	}
	fclose( fp );
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function select_command()                              */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int : 0==success                               */
/*                                                                    */
/*             function: select command process                       */
/*                                                                    */
/**********************************************************************/
int select_command( struct proc *proc )
{
	struct Message *mes, *tmp;
	int ret, st;

	mes = proc->file->mes;

	switch( mes->subcommand ){
	    case SCEXEC: 
		if( (ret = set_shm( proc )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			sprintf( mes->u.data, WarningForm, NKind[proc->kind], proc->spec, "shm cannot get" );
			break;
		}
		if( mes->u.sb.skind == SKFILE ){
			if( (mes = read_pos( mes, &st )) == NULL ){
				mes = proc->file->mes;
				mes->status = STWARNING;
				mes->errorcode = st;
				break;
			}
			tmp = proc->file->mes;
			proc->file->mes = mes;
		} else {
			tmp = NULL;
		}
		if( (ret = gedo_select( proc->file )) != 0 ){
			mes->status = STWARNING;
			if( ret < 0 ){
				mes->errorcode = EILLSELECT;
				mes->u.data[0] = '\0';
			} else {
				mes->errorcode = ret;
			}
		} else {
			mes->status = STDONE;
			mes->errorcode = ENOERROR;
		}
		if( mes->u.sb.skind == SKFILE ){
			tmp->status = mes->status;
			tmp->errorcode = mes->errorcode;
			free( mes );
			proc->file->mes = tmp;
			mes = tmp;
		}
		if( mes->comatr & ADISP ){
			mes->receiver = SVIEWER;
		} else {
			mes->receiver = SMAIN;
		}
		break;
	    case SCCONFIG:
		if( (ret = set_shm( proc )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			break;
		}
/* B2 for config status and version */
		ret = config( proc->file );
		if( ret < 0 ){
			mes->status = STWARNING;
			mes->errorcode = EILLCONFIG;
			mes->u.data[0] = '\0';
		} else if( ret == 0 || ret == GETSTATUS ){
			mes->status = STDONE;
			mes->errorcode = ENOERROR;
		} else {
			mes->status = STWARNING;
			mes->errorcode = ret;
		}
		mes->receiver = SMAIN;
		break;
	    case SCHELP:
		help();
		mes->status = STDONE;
		mes->errorcode = ENOERROR;
		mes->u.data[0] = '\0';
		mes->receiver = SMAIN;
		break;
	    default:
		mes->status = STWARNING;
		mes->errorcode = ENOSUBCOMMAND;
		sprintf( mes->u.data, "%s:subcommad %d is not here\n", proc->spec, mes->subcommand );
		mes->receiver = SMAIN;
		break;
	}
	mes->datasize = strlen( mes->u.data ); 
	reply( proc, mes->receiver );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function effect_command()                              */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int : 0==success                               */
/*                                                                    */
/*             function: effect command process.                      */
/*                                                                    */
/**********************************************************************/
int effect_command( struct proc *proc )
{
	struct Message *mes;
	int ret;

	mes = proc->file->mes;
	switch( mes->subcommand ){
	    case SCEXEC: 
		if( (ret = set_shm( proc )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			sprintf( mes->u.data, WarningForm, NKind[proc->kind], proc->spec, "shm cannot get" );
			break;
		}
		if( (ret = gedo_effect( proc->file )) != 0 ){
			mes->status = STWARNING;
			if( ret < 0 ){
				mes->errorcode = EILLEFFECT;
				mes->u.data[0] = '\0';
			} else {
				mes->errorcode = ret;
			}
		} else {
			mes->status = STDONE;
			mes->errorcode = ENOERROR;
		}
		if( mes->comatr & ADISP ){
			mes->receiver = SVIEWER;
		} else {
			mes->receiver = SMAIN;
		}
		break;
	    case SCCONFIG:
		if( (ret = set_shm( proc )) != 0 ){
			mes->status = STWARNING;
			mes->errorcode = ret;
			break;
		}
/* B2 for config status and version */
		ret = config( proc->file );
		if( ret < 0 ){
			mes->status = STWARNING;
			mes->errorcode = EILLCONFIG;
			mes->u.data[0] = '\0';
		} else if( ret == 0 || ret == GETSTATUS ){
			mes->status = STDONE;
			mes->errorcode = ENOERROR;
		} else {
			mes->status = STWARNING;
			mes->errorcode = ret;
		}
		mes->receiver = SMAIN;
		break;
	    case SCHELP:
		help();
		mes->status = STDONE;
		mes->errorcode = ENOERROR;
		mes->u.data[0] = '\0';
		mes->receiver = SMAIN;
		break;
	    default:
		mes->status = STWARNING;
		mes->errorcode = ENOSUBCOMMAND;
		sprintf( mes->u.data, "%s:subcommad %d is not here\n", proc->spec, mes->subcommand );
		mes->receiver = SMAIN;
		break;
	}
	mes->datasize = strlen( mes->u.data ); 
	reply( proc, mes->receiver );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function get_MesID()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : IPC message ID                           */
/*                                                                    */
/*             function: Get IPC message ID from env or default.      */
/*                                                                    */
/**********************************************************************/
int get_MesID( void )
{
	char *env;
	int id, ret;

	if( (env = getenv( "GEDO_IPCMES" )) == NULL ){
		id = IPCMSGID;
	} else {
		id = atoi( env );
		if( id <= 0 ){
			id = IPCMSGID;
		}
	}
	if( (ret = msgget( id, IPC_CREAT | 0666 )) == -1 ){
		return( -1 );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function check_gedo()                                  */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : 0==called from GEDO  -1==called from sh  */
/*                                                                    */
/*             function: Check called from and inhibit called from    */
/*                      shell.                                        */
/**********************************************************************/
int check_gedo( void )
{
	if( getenv( "GEDO_GWP" ) == NULL ){
		if( getenv( "GEDO_LWP" ) == NULL ){
			fprintf( stderr, "Plugs executed from command line.\n"
				"Plugs called only from gedo.\n"
				"If you want options, add --help\n" );
			return( -1 );
		}
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function process()                                     */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int : 0==success  -1==fatal error              */
/*                                                                    */
/*             function: Event loop called from main() of each        */
/*                      in loader.c, selecter.c or effecter.c         */
/**********************************************************************/
int process( struct proc *proc )
{
	struct Message *mes;

	if( check_gedo() == -1 ){
		return( -1 );
	}
        if( (proc->mesid = get_MesID()) == -1 ){
		printf( "plug:cannot make IPC message\n" );
		return( -1 );
	}
	mes = proc->file->mes;
	kill( getppid(), SIGUSR1 ); 
	signal( SIGUSR1, usr1 );	/* add C1 */
	while( 1 ){
		signal( SIGINT, ctrl_c );
		CtrlCFlag = 0; mes->status = -1;
		if( msgrcv( proc->mesid, (struct msgbuf *)mes, sizeof(struct Message) - sizeof(long), proc->pid, 0 ) == -1 ){
			mes->errorcode = EPIPEREAD;
			mes->status = STFATAL;
			mes->u.data[0] = '\0';
			reply( proc, SMAIN );
		}
		if( mes->status ){
			CtrlCFlag = 0;
			continue;
		}
		signal( SIGINT, ctrl_c_break );
		if( setjmp( jmpbuf ) ){
			mes->errorcode = EUSRBREAK;
			mes->status = STWARNING;
			mes->u.data[0] = '\0';
			reply( proc, SMAIN );
			continue;
		}
		switch( mes->command ){
		      case CLOAD:
			if( proc->kind == SLOADER ){
				load_command( proc );
			} else {
				mes->errorcode = ENOCOMMAND;
				mes->status = STWARNING;
				reply( proc, SMAIN );
			}
		        break;
		      case CSAVE:
			if( proc->kind == SLOADER ){
				save_command( proc );
			} else {
				mes->errorcode = ENOCOMMAND;
				mes->status = STWARNING;
				reply( proc, SMAIN );
			}
		        break;
		      case CSELECT:
			if( proc->kind == SSELECTER ){
				select_command( proc );
			} else {
				mes->errorcode = ENOCOMMAND;
				mes->status = STWARNING;
				reply( proc, SMAIN );
			}
			break;
		      case CEFFECT:
			if( proc->kind == SEFFECTER ){
				effect_command( proc );
			} else {
				mes->errorcode = ENOCOMMAND;
				mes->status = STWARNING;
				reply( proc, SMAIN );
			}
			break;
	              case CKILL:
			end_proc( proc, 0 );
			return( 0 );
		        break;
		      default:
			mes->status = STWARNING;
			mes->errorcode = ENOCOMMAND;
			reply( proc, SMAIN );
		        break;
		}
	}
	return( 0 );
}

#if 0
/*each files*/
void main( int argc, char **argv )
{
	Proc.spec = "pic";
	Proc.file = &File;
	File.mes  = &Mes;
	signal( SIGINT, SIG_IGN );
	Proc.pid = getpid();
	Proc.kind = SLOADER;
	if( init( argc, argv ) != 0 ){
		printf( "illegal initilization of %s\n", proc.spec );
	}
	process( &Proc );
}
#endif
