/****************************************************************************/
/*                                                                          */
/*                Graphic EDitor Omniotent(GEDO)                            */
/*                plug.h     --  plug-in common defintion                   */
/*                                  by K.Sugiura ver 0.92                   */
/*         B2: add CKRAWSTRING to get_config_input()                        */
/*                                                                          */
/****************************************************************************/

#include "../common.h"
#include "../loader.h"

struct File {			/* data infomation */
	int fd;			/* file descripter if loader/saver */
	int shmid;		/* IPC shm ID of the window */
	struct Shm *buf;	/* Data struct */
	struct Message *mes;	/* Transaction struct */
	char name[128];		/* file name */
};

struct Config {			/* user interface of config() */
	char *mes;		/* prompt message, if NULL eof */
	void *val;		/* value must be set */
	int kind;		/* must be set value kind */
	int status;		/* input has done? */
};

struct proc {			/* process infomation */
	int  mesid;		/* IPC mes ID */
	int  pid;		/* self process ID */
	int  kind;		/* LOADER/SAVER, SELECTER or EFFECTER */
	char *spec;		/* name of the plug */
	int shm_max;		/* IPC shm Maximum pixels */
	struct File *file;	/* Data struct */
};

/* symbol used in struct Config kind member  get_config_input() */
#define CKSTRING 0	/* input string(only one word) */
#define CKINT    1	/* input interger value */
#define CKFLOAT  2	/* input float value */
#define CKCHAR   3	/* input char value(make toupper) */
#define CKRAWSTRING 4	/* input strings(string until \n) */

/* symbol used in check_range() */
#define NEGLECT  0	/* delete the over-range points */
#define ADJUST   1	/* make the over-range points to the boundary value */

/* used for DataBase of the plug */
struct PLUG_ITEM {		/* each configuration */
	unsigned char *name;	/* distinct name */
	int type;		/* type of configuration */
	void *set_target;	/* set value */
	int omni1;		/* used by type */
	int omni2;		/* used by type */
	char *comment;		/* comment for name */
};

struct PLUG_INFO {		/* configuration of the plug */
	unsigned char *plugname;	/* plug name (Spec) */
	unsigned char *filename;	/* file name(argv[0]) */
	unsigned char *version;		/* gedo version */
	int kind;			/* SLOADER/SSELECTER/SEFFECTER */
	int item_num;			/* number of PLUGITEM */
	struct PLUG_ITEM *item;		/* items */
	char *comment;			/* total comment */
};

typedef struct PLUG_INFO PLUGINFO;
typedef struct PLUG_ITEM PLUGITEM;

/* plugitem.type symbol */
#define GETINT	 0	/* The data follow name= is a int value. */
#define GETHEX   1	/* The data follow name= is a HEX-int value. */
#define GETFLOAT 2	/* The data follow name= is a float value. */
#define GETSTR   3	/* The data follow name= is a string. */
#define SWITCH   4	/* Name is a switch(ON/OFF) */
#define NOINIT   0x8000	/* not configure at Init */
#define INITONLY 0x4000   /* not configure at commandline */
#define NOLOAD   0x2000   /* loader load command NOT avail */
#define NOSAVE   0x1000   /* loader save command NOT avail */
#define NOTCL	 0x0800	  /* TCL GUI NOT avail */
#define COMMANDMASK 0x00ff

#define GUI   1
#define NOGUI 0

/* kind of command, used Tag[].kind */
#define ENTRY  1	/* the Tag is entry */
#define RADIO  2	/* start of radiobutton */
#define RADIOEND 3	/* end of radiobutton */
#define BLOCK   4	/* start of block(narrow distance) */
#define BLOCKEND 5	/* end of block */
#define ITEM   6	/* item of radiobutton */

/* after, Tcl/Tk only command */
#define INSTCL 7	/* insert Tcl/Tk command */
#define INSFUNC 8	/* insert function generate Tcl/Tk command */
#define OKandESC 9	/* user defined OK/ESC button */
#define TCLLIST 10	/* make listbox(start) */
#define TCLDIR 11	/* make listbox item */
#define TCLLISTEND 12	/* end listbox with label selection displayed */
#define TCLLISTENT 13	/* end listbox with entry selection */

struct Tags {
	short kind;	/* kind of command */
	short num;	/* number refferd PLUGITEM Item */
	char *prompt;	/* prompt for the showing */
	char *result;	/* result format */
	int status;	/* how to processed the Tag? */
};

/* value used in Tag[].status */
#define ERROR  -1	/* error occured */
#define DONE    0	/* selected and success */
#define SKIP    1	/* not selected */
#define USRBREAK 2	/* user aborted the configuration */


/* prototypes */
/* command.c utilities */
long axtol( char *str, int max );
void ltoax( long val, char *str, int max );
long saxtol( char *str );
char ** command_line( char *str );
void check_range( struct Sbuf *sb, int w, int h, int num, int kind );
int check_item( unsigned char *in, PLUGITEM *pi );
int check_itemnum( int kind,  PLUGITEM *pi, int num );
int get_commandspec( struct File *file );
int new_config_input( unsigned int mask, struct Tags *tag, int tagnum, PLUGITEM *item, int *num );
void output_tclcommand( unsigned int mask, FILE *fp, struct Tags *tag, int tagnum, PLUGITEM *item, PLUGINFO *pif );
FILE *create_fifo( char *filename );
int each_default_status( int i, struct Tags *tag, PLUGITEM *item, char *buff );
void display_config( int kind, PLUGITEM *pi, int num );
void search_dir( char *dirname, char *ext );
void search_tcldir( FILE *fp, int var, char *dirname, char *ext );

/* each loader.c selecter.c effecter.c */
int  read_header(struct  File *file);
int  read_image(struct  File *file);
int  write_header(struct  File *file);
int  write_image(struct  File *file);
int  gedo_select( struct File *);
int gedo_effect( struct File *);
int  config( struct File * );
int  line_config( struct File * );
int  line_config2( char * );
int config_use_gui( struct File *file );
int config_comline( struct File *file );
void help( void );
int init( int, char ** );
int process(struct proc * proc );
void status( int kind, char *buff );
void loadstatus( int kind, char *buff );
void savestatus( int kind, char *buff );
void get_default_status( int kind, char *buff );

#define STATUS(fmt)	void status(int kind, char *buff){sprintf(buff,fmt,
#define STATUSEND	);}
#define DEFAULTSTATUS  void status(int kind,char *buff){get_default_status(kind,buff);}

#define CONFIG_ELSE    0
#define CONFIG_INIT    1
#define CONFIG_LOADER  2
#define CONFIG_SAVER   3
#define CONFIG_SELECTER 0
#define CONFIG_EFFECTER 0
