/**********************************************************************/
/*                                                                    */
/*              pnm.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/* D2 : not check pnm format 3rd line value..                         */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include "plug.h"

#define PPM		0
#define PGM		1
#define PBM		2
int WriteMode = PPM;

int ReadMode = PPM;
unsigned char LoadExec[256];
unsigned char SaveExec[256];

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =    { "pnm" };
char *Comment = { "pnm(J.Poskanzer\'s Portable BitMap) loader/saver" };


PLUGITEM Item[] = {
{ "fullcolor",  NOLOAD | SWITCH, &WriteMode, PPM, 'F', "fullcolor save(ppm)"  },
{ "gray", 	NOLOAD | SWITCH, &WriteMode, PGM, 'G', "grayscale save(pgm)" },
{ "bilevel",    NOLOAD | SWITCH, &WriteMode, PBM, 'B', "bilevel save(pbm)" },
{ "loadexec", 	NOSAVE | GETSTR, &LoadExec, 64, 0,  "pre process by sh(1)" },
{ "saveexec", 	NOLOAD | GETSTR, &SaveExec, 64, 0,  "post process by sh(1)" },
{ "gui",        NOLOAD | SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      NOLOAD | SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
 };
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ NOLOAD | RADIO, 0, "Color Mode" },
{ NOLOAD | ITEM, 0, "fullolor(F)", "fullcolor" },
{ NOLOAD | ITEM, 1, "grayscale(G)", "gray" },
{ NOLOAD | ITEM, 2, "B/W(B)",    "bilevel" },
{ NOLOAD | RADIOEND, -1 },
{ NOLOAD | ENTRY, 4, "execute command after save", "saveexec=%s" },
{ NOSAVE | ENTRY, 3, "execute command before load", "loadexec=%s" },
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

unsigned char Mask[8]	/* bit mask */
 = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
#define get_y(r,g,b) (unsigned int)(0.299*(float)r+0.587*(float)g+0.114*(float)b) 
int LSB_flag;		/* dummy, not used */

/**********************************************************************/
/*                                                                    */
/*            function make_commandline()                             */
/*                                                                    */
/*            args  unsigned char *out  : <OUT> output command-line   */
/*                  unsigned char *exec : <IN>  exec string from conf */
/*                  unsigned char *name : <IN>  file name to handle   */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Make Exec command line complete for loader.   */
/*                                                                    */
/**********************************************************************/
void make_commandline( unsigned char *out, unsigned char *exec, unsigned char *name )
{
	if( strchr( exec, '%' ) ){
		sprintf( out, exec, name );
	} else {
		sprintf( out, "cat %s | ", name );
		strcat( out, exec );
	}
}
/**********************************************************************/
/*                                                                    */
/*            function make_savecommandline()                         */
/*                                                                    */
/*            args  unsigned char *out  : <OUT> output command-line   */
/*                  unsigned char *exec : <IN>  exec string from conf */
/*                  unsigned char *name : <IN>  file name to handle   */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Make Exec command line complete for saver.    */
/*                                                                    */
/**********************************************************************/
void make_savecommandline( unsigned char *out, unsigned char *exec, unsigned char *name )
{
	unsigned char buff[1024];

	if( WriteMode == PBM ){
		if( exec[0] == '\0' ){
			sprintf( out, "pgmtopbm > %s", name );
		} else {
			if( strchr( exec, '%' ) ){
				strcpy( out, "pgmtopbm | " );
				sprintf( buff, exec, name );
				strcat( out, buff );
			} else {
				sprintf( out, "pgmtopbm | %s >%s", exec, name );
			}
		}
	} else {
		if( strchr( exec, '%' ) ){
			sprintf( out, exec, name );
		} else {
			sprintf( out, "%s >%s", exec, name );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*            function reset_exec()                                   */
/*                                                                    */
/*            args  unsigned char *exec : exec string from config     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: If config Exec is NULL, clear the exec string.*/
/*                                                                    */
/**********************************************************************/
void reset_exec( unsigned char *exec )
{
	unsigned char *p;

	for( p = exec; *p && isspace(*p); p++ );
	if( *p == '\0' ){
		exec[0] = '\0';
	}
	if( p != exec ){
		memmove( exec, p, strlen(p) + 1 );
	}
}
/**********************************************************************/
/*                                                                    */
/*             function read_header()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ HEADER of the SPEC format from file     */
/*          B0  can deal bilevel pnm                                  */
/*          B2  can deal compression                                  */
/**********************************************************************/
int read_header( struct File *file )
{
	unsigned char buff[10240];
	unsigned char *p;
	unsigned int *q;
	unsigned char max;
	int status;
	int count;
	int i, j, k;
	FILE *fpp = NULL;
	int ret = 0;

	reset_exec( LoadExec );
	if( LoadExec[0] == '\0' ){
		if( (file->fd = open( file->name, O_RDONLY )) == -1 ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
	} else {
		make_commandline( buff, LoadExec, file->name );
		if( (fpp = popen( buff, "r" )) == NULL ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
		file->fd = fileno(fpp);
	}

	status = 0;
	q = file->buf->data;
	i = j = k = 0;
	while( 1 ){
		if( (count = read( file->fd, buff, sizeof(buff) )) < 0 ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			ret = EREADFILE;
			goto ex;
		}
		if( count == 0 ){
			break;
		}
		p = buff;
		switch( status ){
		case 0:
			if( p[0] != 'P' ){
				file->mes->datasize = 0;
				ret = EILLFORM;
				goto ex;
			}
			switch( p[1] ){
			case '4':
				ReadMode = PBM;
				break;
			case '5':
				ReadMode = PGM;
				break;
			case '6':
				ReadMode = PPM;
				break;
			case '1':
			case '2':
			case '3':
				strcpy( file->mes->u.data, "ASCII format" );
				file->mes->datasize = strlen(file->mes->u.data);
				ret = ENOTFEATURE;
				goto ex;
				break;
			default:
				ret = EILLFORM;
				goto ex;
				break;
			}
			p += 2;
			count -= 2;
			if( count && isspace(*p) ){
				p++, count--;
			} else {
				ret = EILLFORM;
				goto ex;
				break;
			}
			file->buf->width = atoi( p );
			for( ; count && isdigit(*p); p++, count-- );
			if( count && isspace(*p) ){
				p++, count--;
			} else {
				ret = EILLFORM;
				goto ex;
				break;
			}
			if( !count ){
				break;
			}
			file->buf->height = atoi( p );
			for( ; count && isdigit(*p); p++, count-- );
			if( count && isspace(*p) ){
				p++, count--;
			} else {
				ret = EILLFORM;
				goto ex;
				break;
			}
			if( !count ){
				break;
			}
			if( ReadMode != PBM ){
				max = atoi( p );
				for( ; count && isdigit(*p); p++, count-- );
				if( count && isspace(*p) ){
					p++, count--;
				} else {
					ret = EILLFORM;
					goto ex;
					break;
				}
				if( !count ){
					break;
				}
/* D2 : not check 3rd line value.. 1999.11.6 */
#if 0
				if( max != 255 ){
					ret = EILLFORM;
					goto ex;
				}
#endif
			}
			if( file->buf->width * file->buf->height 
			   >= MAXWIDTH * MAXHEIGHT ){
				ret = 0; /* ETOOLARGE */
				goto ex;
			}
			status  = 1;
			/* FALLTHRU */
		case 1:
			q = file->buf->data;
			i = j = k = 0;
			status = 2;
			/* FALLTHRU */
		case 2:
			while( count > 0 ){
				switch( ReadMode ){
				case PPM:
					if( k == 0 ){
						*q = (*p << 16);
						k = 1;
					} else if( k == 1 ){
						*q |= (*p << 8);
						k = 2;
					} else if( k == 2 ){
						*q |= *p;
						k = 0;
						q++;
						if( ++i >= file->buf->width ){
							if( ++j >= file->buf->height ){
								ret = 0;
								goto ex;
							}
							i = 0;
						}
					}
					break;
				case PGM:
					*q = ((*p << 16) | (*p << 8) | *p);
					q++;
					if( ++i >= file->buf->width ){
						if( ++j >= file->buf->height ){
							ret = 0;
							goto ex;
						}
						i = 0;
					}
					break;
				case PBM:
					for( k = 0; k < 8; k++ ){
						if( Mask[k] & *p ){
							*q++ = 0;
						} else {
							*q++ = 0x00ffffff;
						}
						if( ++i >= file->buf->width ){
							if( ++j >= file->buf->height ){
								ret = 0;
								goto ex;
							}
							i = 0;
							break;
						}
					}
					break;
				}
				count--; p++;
			}
			break;
		default:
			ret = EILLFORM;
			goto ex;
			break;
		}
	}
ex:	if( fpp ){
		pclose( fpp );
	} else {
		close( file->fd );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function read_image()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ IMAGES of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_image( struct File *file )
{
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function write_header()                                */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE HEADER of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_header( struct File *file )
{
	reset_exec( SaveExec );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE IMAGES of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_image( struct File *file )
{
	int i, j;
	unsigned char buff[256*3];
	unsigned int *p;

	FILE *fpp = NULL;
	int ret = 0;

	if( SaveExec[0] == '\0' && WriteMode != PBM ){
		if( (file->fd = open( file->name, O_RDWR | O_TRUNC )) == -1 ){
			if( (file->fd = open( file->name, O_RDWR | O_CREAT, 0664)) == -1 ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				return( EOPENFILE );
			}
		}
	} else {
		make_savecommandline( buff, SaveExec, file->name );
		if( (fpp = popen( buff, "w" )) == NULL ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
		file->fd = fileno(fpp);
	}
	if( WriteMode == PPM ){
		sprintf( buff, "P6\n%d %d\n255\n", file->buf->width, file->buf->height );
	} else {
		sprintf( buff, "P5\n%d %d\n255\n", file->buf->width, file->buf->height );
	}
	if( write( file->fd, buff, strlen(buff) ) != strlen(buff) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		ret = EWRITEFILE;
		goto ex;
	}

	p = file->buf->data;
	for( i = j = 0; i < file->buf->width * file->buf->height; i++, p++ ){
		if( WriteMode == PPM ){
			buff[j++] = GET_RED(*p);
			buff[j++] = GET_GREEN(*p);
			buff[j++] = GET_BLUE(*p);
		} else {
			buff[j++] = get_y( GET_RED(*p), GET_GREEN(*p), GET_BLUE(*p) );
		}
		if( j >= 256 * 3 ){
			if( write( file->fd, buff, 256 * 3 ) != 256 * 3 ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				ret = EWRITEFILE;
				goto ex;
			}
			j = 0;
		}
	}
	if( write( file->fd, buff, j ) != j ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		ret = EWRITEFILE;
		goto ex;
	}

ex:	if( fpp ){
		pclose( fpp );
	} else {
		close( file->fd );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
