/**********************************************************************/
/*                                                                    */
/*             poly.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "plug.h"

#define LINE	0
#define SPLINE	1

int Mode = LINE;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "poly" };
char *Comment = { "poly/spline poly selecter" };

PLUGITEM Item[] = {
{ "spline", SWITCH,  &Mode, SPLINE, 'S',   "spline mode" },
{ "line",   SWITCH,  &Mode, LINE,   'L',   "line mode" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ RADIO, 0, "Mode " },
{ ITEM, 0, "normal line(l)", "line" },
{ ITEM, 1, "spline(s)", "spline" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* image width and height */
unsigned int *Data;		/* image top */

/* matrix to inc/dec position(strategy) */
static int Matx[8] = { -1,  0,  1, -1,  1, -1,  0,  1 };
static int Maty[8] = { -1, -1, -1,  0,  0,  1,  1,  1 };

#define xydata(x,y) Data[(y)*Width+(x)]

struct fPosition {	/* for spline call */
	float x;
	float y;
};

struct fPosition *spline_poly( struct fPosition *pos, int dim, int num, int pat );

/**********************************************************************/
/*                                                                    */
/*             function check_line_inner()                            */
/*                                                                    */
/*             int x    : X position                                  */
/*             int y    : Y position                                  */
/*                                                                    */
/*             return  int  : 0==out of the area  8==inside           */
/*                                                                    */
/*             function: Check the position is inside or outside      */
/*                      of the selected area.                         */
/**********************************************************************/
int check_line_inner( int x, int y )
{
	int i;
	int status, count;

	status = 0; count = 0;
	for( i = x+1; i < Width; i++ ){
		if( status == 0 ){
			if( MASKOFF( xydata(i,y) ) ){
				status = 1;
				count++;
			}
		} else {
			if( MASKON( xydata(i,y) ) ){
				status = 0;
			}
		}
	}
	if( count == 0 || (count % 2 == 0) ){
		return( 0 );
	}

	status = 0; count = 0;
	for( i = x-1; i >= 0; i-- ){
		if( status == 0 ){
			if( MASKOFF( xydata(i,y) ) ){
				status = 1;
				count++;
			}
		} else {
			if( MASKON( xydata(i,y) ) ){
				status = 0;
			}
		}
	}
	if( count == 0 || (count % 2 == 0) ){
		return( 0 );
	}

	status = 0; count = 0;
	for( i = y+1; i < Height; i++ ){
		if( status == 0 ){
			if( MASKOFF( xydata(x,i) ) ){
				status = 1;
				count++;
			}
		} else {
			if( MASKON( xydata(x,i) ) ){
				status = 0;
			}
		}
	}
	if( count == 0 || (count % 2 == 0) ){
		return( 0 );
	}

	status = 0; count = 0;
	for( i = y-1; i >= 0; i-- ){
		if( status == 0 ){
			if( MASKOFF( xydata(x,i) ) ){
				status = 1;
				count++;
			}
		} else {
			if( MASKON( xydata(x,i) ) ){
				status = 0;
			}
		}
	}
	if( count == 0 || (count % 2 == 0) ){
		return( 0 );
	}
	return( 8 );
}

/**********************************************************************/
/*                                                                    */
/*             function set_inner()                                   */
/*                                                                    */
/*             int x   : inner position X(start)                      */
/*             int y   : inner position Y(start)                      */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: MaskOff the inner area from start position.  */
/*                                                                    */
/**********************************************************************/
void set_inner( int x, int y )
{
	volatile int i, left, right, up, down;

	if( MASKOFF( xydata(x, y) )  ){
		return;
	}
	CLRMASK( xydata(x, y) );
	for( i = x + 1; i < Width; i++ ){
		if( MASKON( xydata(i, y) ) ){
			CLRMASK( xydata(i, y) );
		} else {
			break;
		}
	}
	right = i;
	for( i = x - 1; i >= 0; i-- ){
		if( MASKON( xydata(i, y) ) ){
			CLRMASK( xydata(i, y) );
		} else {
			break;
		}
	}
	left = i;
	for( i = y + 1; i < Height; i++ ){
		if( MASKON( xydata(x, i) ) ){
			CLRMASK( xydata(x, i) );
		} else {
			break;
		}
	}
	down = i;
	for( i = y - 1; i >= 0; i-- ){
		if( MASKON( xydata(x, i) ) ){
			CLRMASK( xydata(x, i) );
		} else {
			break;
		}
	}
	up = i;

	for( i = x + 1; i < right; i++ ){
		set_inner( i, y+1 );
		set_inner( i, y-1 );
	}
	for( i = x - 1; i > left; i-- ){
		set_inner( i, y+1 );
		set_inner( i, y-1 );
	}
	for( i = y + 1; i < down; i++ ){
		set_inner( x + 1, i );
		set_inner( x-1, i );
	}
	for( i = y - 1; i > up; i-- ){
		set_inner( x+1, i );
		set_inner( x-1, i );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function set_area()                                    */
/*                                                                    */
/*             struct Sbuf *sb  : selected points                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: From near position of each selected line,    */
/*                      check it is inner and if inner area fill start.*/
/**********************************************************************/
void set_area( struct Sbuf *sb )
{
	int i, j;
	int dx, dy;

	for( i = 0; i < sb->defnum; i++ ){
		for( j = 0; j < 8; j++ ){
			dx = sb->def[i].x + Matx[j];
			dy = sb->def[i].y + Maty[j];
			if( dx < 0 || dx >= Width ){
				continue;
			}
			if( dy < 0 || dy >= Height ){
				continue;
			}
			if( MASKOFF( xydata( dx, dy ) ) ){
				continue;
			}
			if( check_line_inner( dx, dy ) ){
				set_inner( dx, dy );
				return;
			}
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*             function set_sparea()                                  */
/*                                                                    */
/*             struct fPosition *at: spline create points             */
/*             int total           : spline create points num         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: From near position of each selected line,    */
/*                      check it is inner and if inner area fill start.*/
/*                      This is spline case.                          */
/**********************************************************************/
void set_sparea( struct fPosition *at, int total )
{
	int i, j;
	int dx, dy;

	for( i = 0; i < total; i++ ){
		for( j = 0; j < 8; j++ ){
			dx = (int)at[i].x + Matx[j];
			dy = (int)at[i].y + Maty[j];
			if( dx < 0 || dx >= Width ){
				continue;
			}
			if( dy < 0 || dx >= Height ){
				continue;
			}
			if( MASKOFF( xydata( dx, dy ) ) ){
				continue;
			}
			if( check_line_inner( dx, dy ) ){
				set_inner( dx, dy );
				return;
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function round()                                       */
/*                                                                    */
/*             double f      : float value                            */
/*                                                                    */
/*             return  int   : rounded value(int converted)           */
/*                                                                    */
/*             function: Round the float value.                       */
/*                                                                    */
/**********************************************************************/
int round( double f )
{
	double r;

	r = floor( f );
	if( f - r >= 0.5 ){
		return( (int)r + 1 );
	} else {
		return( (int)r );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function set_line()                                    */
/*                                                                    */
/*             unsigned int *data  : image top                        */
/*             struct Pos s        : line start                       */
/*             struct Pos e        : line end                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: 1st, line the selected positions and 2nd     */
/*                      fill inside.                                  */
/**********************************************************************/
void set_line( unsigned int *data, struct Pos s, struct Pos e )
{
	int i, j;
	unsigned int *p;
	int min, max, off, d, omax, dn;
	double a;
	int dx, dy;

	if( s.x == e.x ){
		min = min( s.y, e.y );
		max = max( s.y, e.y );
		for( p = &data[min * Width + s.x], i = min; i <= max; i++ ){
			CLRMASK(*p);
			p += Width;
		}
	} else if( s.y == e.y ){
		min = min( s.x, e.x );
		max = max( s.x, e.x );
		for( p = &data[s.y * Width + min], i = min; i <= max; i++ ){
			CLRMASK(*p);
			p++;
		}
	} else {
		if( (dx = e.x - s.x) > 0 ){
			dx = dx + 1;
		} else {
			dx = dx - 1;
		}
		if( (dy = e.y - s.y) > 0 ){
			dy = dy + 1;
		} else {
			dy = dy - 1;
		}
		a = (double)dx / (double)dy;
		if( s.y > e.y ){
			min = e.y;
			off = e.x;
			omax = s.x;
			max = s.y;
		} else {
			min = s.y;
			off = s.x;
			omax = e.x;
			max = e.y;
		}
		p = &data[min * Width];
		d = 0;
		for( i = min; i <= max; i++ ){
			dn = round(a * (double)(i - min + 1));
			if( d > dn ){
				for( j = d; j > dn && off-j >= omax; j-- ){
					CLRMASK(p[off+j]);
				}
			} if( d < dn ){
				for( j = d; j < dn && off+j <= omax ; j++ ){
					CLRMASK(p[off+j]);
				}
			} else {
				CLRMASK(p[off+d]);
			}
			d = round(a * (double)(i - min + 1));
			if( off >= Width ){
				off = Width;
			} else if( off <= Width * (-1) ){
				off = Width * (-1);
			}
			p += Width;
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	unsigned int *p;
	int i;
	struct fPosition pos[MAXSP];
	struct fPosition *ret;
	int total;
	struct Pos x0, x1;

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	Data = file->buf->data;

	if( Mode == LINE ){
		if( sb->selnum != SKPOLY ){
			return( EILLSELECTION );
		}
	} else if( Mode == SPLINE ){
		if( sb->selnum != SKSPPOLY ){
			return( EILLSELECTION );
		}
	} else {
		return( EILLSELECTION );
	}
	check_range( sb, Width, Height, 0, ADJUST );
	if( sb->defnum < 3 ){
		return( EILLNUMBER );
	}


	for( i = 0, p = Data; i < Height * Width; i++, p++ ){
		SETMASK(*p);
	}
	if( Mode == LINE ){
		for( i = 0; i < sb->defnum - 1; i++ ){
			set_line( Data, sb->def[i], sb->def[i+1] );		}
		set_line( Data, sb->def[i], sb->def[0] );
		set_area( sb );
	} else if( Mode == SPLINE ){
		for( i = 0; i < sb->defnum; i++ ){
			pos[i].x = (float)sb->def[i].x;
			pos[i].y = (float)sb->def[i].y;
		}
		total = sb->defnum * 10;

	        if( (ret = spline_poly( pos, 3, sb->defnum, 10 )) == NULL ){
			return( -1 );
		}
		x0.x = (int)ret[0].x; x0.y = (int)ret[0].y;
	        for( i = 0; i < total; i++ ){
			x1.x = (int)ret[i].x; x1.y = (int)ret[i].y;
			set_line( Data, x0, x1 );
			x0.x = x1.x; x0.y = x1.y;
		}
		x1.x = (int)ret[0].x; x1.y = (int)ret[0].y;
		set_line( Data, x0, x1 );
		set_sparea( ret, total );
	}
	file->buf->mask_kind = SIMPLEMASK;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else { 
		return( line_config( file ) );
	}	
}
