/**********************************************************************/
/*                                                                    */
/*              psinterp.c   gedo PLUG specified routine              */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define STACKMAX	50
static double Stack[STACKMAX];		/* PostScript calculate stack */
static int Sp;				/* Stack Pointer of Stack */
static double Val;			/* nubmber handling */

/* Postscript command to nymonic */
#define NUM	1
#define DUP	2
#define POP	3
#define ADD	4
#define SUB	5
#define MUL	6
#define DIV	7
#define IDIV	8
#define MOD	9
#define CLEAR	10
#define EXCH	11
#define COPY	12
#define INDEX	13
#define ROLL	14
#define PSTACK	15

/* lexer() command */
#define INIT	0
#define GET	1

/* pair of Postscript command name and nymonic */
struct OP {
	char *name;
	int index;
};

struct OP Op[] = {
{"dup",DUP},
{"pop",POP},
{"add",ADD},
{"sub",SUB},
{"mul",MUL},
{"div",DIV},
{"idiv",IDIV},
{"mod",MOD},
{"clear",CLEAR},
{"exch",EXCH},
{"copy",COPY},
{"index",INDEX},
{"roll", ROLL},
{"pstack", PSTACK},
};

#define OPMAX (sizeof(Op)/sizeof(struct OP))

extern int Width, Height;

/**********************************************************************/
/*                                                                    */
/*            function operator()                                     */
/*                                                                    */
/*            args  char *s : operator name                           */
/*                  int len : length of s                             */
/*                                                                    */
/*            return int : nymonic code                               */
/*                                                                    */
/*            function: find the nymonic code from the command name.  */
/*                                                                    */
/**********************************************************************/
static int operator( char *s, int len )
{
	int i;

	for( i = 0; i < OPMAX; i++ ){
		if( strncmp( s, Op[i].name, len ) == 0 ){
			return( Op[i].index );
		}
	}
	return( -1 );
}

/**********************************************************************/
/*                                                                    */
/*            function lexer()                                        */
/*                                                                    */
/*            args  int com   : command to lexer(INIT or GET)         */
/*                  char *buff: PostScript command string             */
/*                                                                    */
/*            return  int : nymonic of PostScript atom.               */
/*                                                                    */
/*            function: read the string and split the PostScript      */
/*                     nymonics.                                      */
/**********************************************************************/
static int lexer( int com, char *buff )
{
	static char *p;
	static int state;

	if( com == INIT ){
		state = 0;
		p = buff;
		return( 0 );
	}

	for( ;; p++ ){
		if( state ){
			if( *p == '\0' || isspace(*p) ){
				int s;

				s = state;
				state = 0;
				if( *p == '\0' ){
					*p = ' ';
					*(p+1) = '\0';
				}
				return( operator( &p[-s], s ) );
			} else if( isalpha(*p) || isdigit(*p) ){
				state++;
			}
			
		} else {
			if( isdigit(*p) || *p == '-' || *p == '.' ){
				Val = atof( p );
				for( ; *p; p++ ){
					if( isdigit(*p) || *p == '-' || *p == '.' 
					   || *p == 'E' || *p == 'e' || *p == '+' ){
					} else {
						break;
					}
				}
				return( NUM );
			} else if( isalpha(*p) ){
				state = 1;
			} else if( isspace(*p) ){
			} else if( *p == '\0' ){
				state = 0;
				return( 0 );
			} else if( *p == '%' ){
				if( *(p+1) == 'x' ){
					Val = (double)Width;
					p += 2;
					return( NUM );
				} else if( *(p+1) == 'y' ){
					Val = (double)Height;
					p += 2;
					return( NUM );
				} else {
					fprintf( stderr, "undefined %%%c\n", *(p+1) );
					return( -1 );
				}
			}
		}
	}
}	
			
/**********************************************************************/
/*                                                                    */
/*            function roll()                                         */
/*                                                                    */
/*            args  int num : roll depth                              */
/*                  int ct  : roll count                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: PostScript roll command                       */
/*                                                                    */
/**********************************************************************/
static void roll( int num, int ct )
{
	double tmp;
	int i, j;

	if( Sp - num < 0 ){
		fprintf( stderr, "stack underflow\n"  );
		return;
	}
	if( ct > 0 ){
		for( i = 0; i < ct; i++ ){
			tmp = Stack[Sp-1];
			for( j = Sp-1; j > Sp-num; j-- ){
				Stack[j] = Stack[j-1];
			}
			Stack[j] = tmp;
		}
	} else {
		ct *= -1;
		for( i = 0; i < ct; i++ ){
			tmp = Stack[Sp-num];
			for( j = Sp-num + 1; j < Sp; j++ ){
				Stack[j-1] = Stack[j];
			}
			Stack[Sp-1] = tmp;
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*            function parse_ps()                                     */
/*                                                                    */
/*            args  char *buf : PostScript command string             */
/*                                                                    */
/*            return int  : 0==success  -1==error                     */
/*                                                                    */
/*            function:  execute PostScript command.                  */
/*                                                                    */
/**********************************************************************/
int parse_ps( char *buf )
{
	int lex;
	int i;

	lexer( INIT, buf );
	while( (lex = lexer( GET, NULL )) ){
		switch( lex ){
		case NUM:
			if( Sp+1 >= STACKMAX ){
				fprintf( stderr, "stack overflow\n" );
				return( -1 );
			}
			Stack[Sp++] = Val;
			break;
		case DUP:
			if( Sp+1 >= STACKMAX ){
				fprintf( stderr, "stack overflow\n" );
				return( -1 );
			}
			Stack[Sp] = Stack[Sp-1];
			Sp++;
			break;
		case POP:
			if( Sp - 1 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			Sp--;
			break;
		case ADD:
			if( Sp - 2 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			Val = Stack[Sp-1] + Stack[Sp-2];
			Stack[Sp-2] = Val;
			Sp--;
			break;
		case SUB:
			if( Sp - 2 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			Val = Stack[Sp-2] - Stack[Sp-1];
			Stack[Sp-2] = Val;
			Sp--;
			break;
		case MUL:
			if( Sp - 2 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			Val = Stack[Sp-1] * Stack[Sp-2];
			Stack[Sp-2] = Val;
			Sp--;
			break;
		case DIV:
			if( Sp - 2 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			if( Stack[Sp-1] == 0.0 ){
				fprintf( stderr, "division by 0\n" );
				return( -1 );
			}
			Val = Stack[Sp-2] / Stack[Sp-1];
			Stack[Sp-2] = Val;
			Sp--;
			break;
		case IDIV:
			if( Sp - 2 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			if( Stack[Sp-1] == 0.0 ){
				fprintf( stderr, "division by 0\n" );
				return( -1 );
			}
			Val = (double)((int)Stack[Sp-2] / (int)Stack[Sp-1]);
			Stack[Sp-2] = Val;
			Sp--;
			break;
		case MOD:
			if( Sp - 2 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			if( Stack[Sp-1] == 0.0 ){
				fprintf( stderr, "division by 0\n" );
				return( -1 );
			}
			Val = (double)((int)Stack[Sp-2] % (int)Stack[Sp-1]);
			Stack[Sp-2] = Val;
			Sp--;
			break;
		case CLEAR:
			Sp = 0;
			break;
		case EXCH:
			if( Sp - 2 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			Val = Stack[Sp-1];
			Stack[Sp-1] = Stack[Sp-2];
			Stack[Sp-2] = Val;
			break;
		case COPY:
			if( Sp - 1 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			Val = Stack[--Sp];
			if( Sp - Val < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			for( i = 0; i < Val; i++ ){
				Stack[Sp+i] = Stack[Sp - (int)Val + i];
				if( Sp+i >= STACKMAX ){
					fprintf( stderr, "stack overflow\n" );
					return( -1 );
				}
			}
			Sp = Sp + i;
			break;
		case INDEX:
			if( Sp - 1 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			Val = Stack[--Sp];
			Stack[Sp] = Stack[Sp - (int)Val];
			Sp++;
			break;
		case ROLL:
			if( Sp - 2 < 0 ){
				fprintf( stderr, "stack underflow\n" );
				return( -1 );
			}
			Sp -= 2;
			roll( (int)Stack[Sp], (int)Stack[Sp+1] );
			break;
		case PSTACK:
			for( i = 0; i < Sp; i++ ){
				fprintf( stderr, "stack %d: %g\n", i, Stack[Sp-i - 1] );
			}
			break;
		default:
			fprintf( stderr, "undefined OPcode %d\n", lex );
			return( -1 );
			break;
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function pop()                                          */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return float : Stack top value                          */
/*                                                                    */
/*            function: the POP command from the Stack                */
/*                                                                    */
/**********************************************************************/
float pop( void )
{
	if( Sp - 1 < 0 ){
		fprintf( stderr, "stack underflow\n" );
		return( -1 );
	}
	Sp--;
	return( (float)Stack[Sp] );
}
