/**********************************************************************/
/*                                                                    */
/*              raw.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "plug.h"

#define MERGE  0
#define INTERA 1
#define INTERB 2

int Mode = MERGE;
int Gui = NOGUI;

/********************** configure desciption ******************/
char *Spec =	{ "raw" };
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Comment = { "sugiura RAW format(*.cut) loader/saver" };

PLUGITEM Item[];	/* no configuration */
int      ItemNum = 0;

struct Tags Tag[];
int TagNum = 0;

DEFAULTSTATUS;
/********************** description end ********************/


#define BYTE unsigned char
#define WORD unsigned short

struct RawHeader {
	WORD spec;	/* LSB or MSB resolve flag */
	WORD width;	/* width */
	WORD height;	/* height */
	WORD mask;	/* mask kind */
};

int LSB_flag;		/* not used */

/**********************************************************************/
/*                                                                    */
/*             function rev()                                         */
/*                                                                    */
/*             WORD v        : reverse short                          */
/*                                                                    */
/*             return  WORD  : reversed short                         */
/*                                                                    */
/*             function:  Reverse the byte order at 2byte WORD.       */
/*                                                                    */
/**********************************************************************/
WORD rev( WORD v )
{
	WORD ret;

	ret = ((v & 0xff00) >> 8);
	ret |= ((v & 0x00ff) << 8);
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function read_header()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ HEADER of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_header( struct File *file )
{
	struct RawHeader header;

	if( (file->fd = open( file->name, O_RDONLY )) == -1 ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EOPENFILE );
	}
	if( read( file->fd, &header, sizeof(struct RawHeader) ) != sizeof(struct RawHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EREADFILE );
	}
	if( header.spec == 1 ){
		file->buf->width = (int)header.width;
		file->buf->height = (int)header.height;
		file->buf->mask_kind = (int)header.mask;
	} else if( header.spec == 0x0100 ){
		file->buf->width = (int)rev( header.width );
		file->buf->height = (int)rev( header.height );
		file->buf->mask_kind = (int)rev( header.mask );
	} else {
		file->mes->datasize = 0;
		return( EILLFORM );
	}
       	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function read_image()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ IMAGES of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_image( struct File *file )
{
	long size;
	struct Shm *buf;

	buf = file->buf;
	size = buf->width * buf->height * 4;
	if( read( file->fd, buf->data, size ) != size ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		close( file->fd );
		return( EREADFILE );
	}
	close( file->fd );
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function write_header()                                */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE HEADER of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_header( struct File *file )
{
	struct RawHeader header;

	if( (file->fd = open( file->name, O_RDWR | O_TRUNC )) == -1 ){
		if( (file->fd = open( file->name, O_RDWR | O_CREAT, 0664)) == -1 ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
	}
	memset( &header, 0, sizeof(struct RawHeader) );
	header.spec = 1;
	header.width = (WORD)file->buf->width;
	header.height = (WORD)file->buf->height;
	header.mask = (WORD)file->buf->mask_kind;
	if( write( file->fd, &header, sizeof(struct RawHeader) ) != sizeof(struct RawHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EWRITEFILE );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE IMAGES of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_image( struct File *file )
{
	long size;
	struct Shm *buf;

	buf = file->buf;
	size = buf->width * buf->height * 4;
	if( write( file->fd, buf->data, size ) != size ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		close( file->fd );
		return( EREADFILE );
	}
	close( file->fd );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( file->mes->command == CLOAD ){
		if( str[0] == '\0' ){
			printf( "loader bmp: no configration parameter\n" );
			file->mes->u.data[0] = '\0';
		} else {
			return( line_config( file ) );
		}
	} else if( file->mes->command == CSAVE ){
		if( str[0] == '\0' ){
			printf( "saver bmp: no configration parameter\n" );
			file->mes->u.data[0] = '\0';
		} else {
			return( line_config( file ) );
		}
	}
	return( 0 );
}
