/**********************************************************************/
/*                                                                    */
/*             redtree.c gedo PLUG utility. for libreduce subroutine. */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*         D2: Re-write reduce(). Messages, DEBUG, loop condition.    */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <unistd.h>

/* #define DEBUG  1 */
#define TABLEMAX   (16 * 16 * 16)	/* Maximum of the Table */

struct Table{			/* Binary TREE for regist used color */
	unsigned char hash;	/* hash value, calced from index color */
	unsigned char name;	/* node value, calced from index color */
	int count;		/* use count */
	int index;		/* index after merged */
	int left;		/* left node */
	int right;		/* right node */
};
static struct Table Table[TABLEMAX];

struct Color {			/* Color table requested get_pallet() */
	int R;
	int G;
	int B;
};
static struct Color *Color;

#define HASHSIZE 64		/* hash size */
static int Hash[HASHSIZE];	/* hash top */

static int NowTable = 0;	/* used table */
static int Colors;		/* now total colors */
static int Index;		/* merged index */

/* each, calc R,G,B from hash and name */
#define get_R(h,n)	(((h & 0x30) >> 2) | ((n & 0x30) >> 4))
#define get_G(h,n)	(((h & 0x0c)) | ((n & 0x0c) >> 2))
#define get_B(h,n)	(((h & 0x03) << 2) | ((n & 0x03)))

/**********************************************************************/
/*                                                                    */
/*            function get_newentry()                                 */
/*                                                                    */
/*            args  unsigned char name : lower color index            */
/*                  unsigned char hash : upper color index            */
/*                                                                    */
/*            return  int : registed table No.                        */
/*                                                                    */
/*            function: Make new entry of Table.                      */
/*                                                                    */
/**********************************************************************/
static int get_newentry( unsigned char name, unsigned char hash )
{
	int ret;

	Table[NowTable].name = name;
	Table[NowTable].hash = hash;
	Table[NowTable].index = 0;
	Table[NowTable].count = 1;
	Table[NowTable].left = -1;
	Table[NowTable].right = -1;
	ret = NowTable++;
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function init_tree()                                    */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Initialize Binary Tree.                       */
/*                                                                    */
/**********************************************************************/
void init_tree( void )
{
	int i;

	NowTable = 0;
	for( i = 0; i < HASHSIZE; i++ ){
		Hash[i] = -1;
	}
}


/**********************************************************************/
/*                                                                    */
/*            function search_tree()                                  */
/*                                                                    */
/*            args  unsigned char hash : upper color index            */
/*                  unsigned char value: lower color index            */
/*                                                                    */
/*            return int : -1==not found  else==found node            */
/*                                                                    */
/*            function: Search color is in the Tree?                  */
/*                                                                    */
/**********************************************************************/
int search_tree( unsigned char hash, unsigned char value )
{
	unsigned char cmp;
	int at;

	if( Hash[hash] == -1 ){
		return( -1 );
	}
	for( at = Hash[hash]; at != -1; ){
		cmp = Table[at].name;
		if( value == cmp ){
			return( at );
		} else if( value < cmp ){
			at = Table[at].right;
		} else {
			at = Table[at].left;
		}
	}
	return( -1 );
}

/**********************************************************************/
/*                                                                    */
/*            function add_tree()                                     */
/*                                                                    */
/*            args  unsigned char hash : upper color index            */
/*                  unsigned char name : lower color index            */
/*                                                                    */
/*            return int : Now used table(NowTable)                   */
/*                                                                    */
/*            function: Search and add new color to the Tree.         */
/*                                                                    */
/**********************************************************************/
int add_tree( unsigned char hash, unsigned char name )
{
	int at, prev;
	int new;
	int cmp;

	if( Hash[hash] == -1 ){
		new = get_newentry( name, hash );
		Hash[hash] = new;
		return( NowTable );
	}
	prev = -1;
	for( at = Hash[hash]; at != -1; ){
		cmp = Table[at].name;
		prev = at;
		if( cmp == name ){
			Table[at].count++;
			return( NowTable );
		} else if( cmp < name ){
			at = Table[at].left;
		} else {
			at = Table[at].right;
		}
	}
	new = get_newentry( name, hash );
	if( prev == -1 ){
		at = Hash[hash];
	} else {
		at = prev;
	}
	if( cmp < name ){
		Table[new].left = Table[at].left;
		Table[at].left = new;
	} else {
		Table[new].right = Table[at].right;
		Table[at].right = new;
	}
	return( NowTable );
}

/**********************************************************************/
/*                                                                    */
/*            function new_name()                                     */
/*                                                                    */
/*            args  unsigned char name : lower color index            */
/*                  int rd  :  R differencial                         */
/*                  int gd  :  G differencial                         */
/*                  int bd  :  B differencial                         */
/*                                                                    */
/*            return int : -1==not in the hash  else==must be search  */
/*                                                        color       */
/*            function: Get each color expected in the hash.          */
/*                                                                    */
/**********************************************************************/
int new_name( unsigned char name, int rd, int gd, int bd )
{
	int r, g, b;
	unsigned char ret;

	r = ((name & 0x30) >> 4);
	g = ((name & 0x0c) >> 2);
	b = ((name & 0x03));
	r += rd; g += gd; b += bd;
	if( r < 0 || r > 0x03 ){
		return( -1 );
	}
	if( g < 0 || g > 0x03 ){
		return( -1 );
	}
	if( b < 0 || b > 0x03 ){
		return( -1 );
	}
	ret = ((r << 4) | (g << 2) | b);
	return( (int)ret );
}

/**********************************************************************/
/*                                                                    */
/*            function new_hashname()                                 */
/*                                                                    */
/*            args  struct Table *t : base color                      */
/*                  int *hash : new hash                              */
/*                  int *name : new name                              */
/*                  int rd  :  R differencial                         */
/*                  int gd  :  G differencial                         */
/*                  int bd  :  B differencial                         */
/*                                                                    */
/*            return int : -1==not tree value   else==must be search  */
/*                                                        color       */
/*                                                                    */
/*            function: Get each color expected all over the tree.    */
/*                                                                    */
/**********************************************************************/
int new_hashname( struct Table *t, int *hash, int *name, int rd, int gd, int bd )
{
	int r, g, b;

	r = (((t->hash & 0x30) >> 2) | ((t->name & 0x30) >> 4));
	g = (((t->hash & 0x0c)) | ((t->name & 0x0c) >> 2));
	b = (((t->hash & 0x03) << 2) | ((t->name & 0x03)));

	r += rd; g += gd; b += bd;

	if( r < 0 || r > 0x0f ){
		return( -1 );
	}
	if( g < 0 || g > 0x0f ){
		return( -1 );
	}
	if( b < 0 || b > 0x0f ){
		return( -1 );
	}

	*name = (((r & 0x03) << 4) | ((g & 0x03) << 2) | (b & 0x03));
	*hash = ((((r >> 2) & 0x03)<< 4) | (((g >> 2) & 0x03) << 2) | ((b >> 2) & 0x03));
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function renum_index()                                  */
/*                                                                    */
/*            args  int old : old index value                         */
/*                  int new : new index value                         */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Change old index to the serial new index.     */
/*                                                                    */
/**********************************************************************/
void renum_index( int old, int new )
{
	int i;

	for( i = 0; i < NowTable; i++ ){
		if( Table[i].index == old ){
			Table[i].index = new;
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*            function make_color()                                   */
/*                                                                    */
/*            args  unsigned char *ind : real color infomation        */
/*                                      from ZIndex[]                 */
/*            return int : 0==success  -1==fail                       */
/*                                                                    */
/*            function: Re-calc new real color from index and ZIndex. */
/*                                                                    */
/**********************************************************************/
int make_color( unsigned char *ind, int scale )
{
	int i, j, k, flag;
	int r, g, b, c;
	int R, G, B;

	for( i = 0; i < NowTable; i++ ){
		if( Table[i].index == 0 ){
			Table[i].index = Index++;
		}
	}
	for( i = k = 0; i < Index; i++ ){
		flag = 0;
		for( j = 0; j < NowTable; j++ ){
			if( Table[j].index == i ){
				Table[j].index = k;
				flag = 1;
			}
		}
		if( flag ){
			k++;
		}
	}
	Index = k;
	if( (Color = malloc( Index * sizeof(struct Color))) == NULL ){
		fprintf( stderr, "fatal:cannot malloc\n" );
		return( -1 );
	}
	for( i = 0; i < Index; i++ ){
		r = g = b = c = 0;
		for( j = 0; j < NowTable; j++ ){
			if( Table[j].index == i ){
				R = get_R( Table[j].hash, Table[j].name );
				G = get_G( Table[j].hash, Table[j].name );
				B = get_B( Table[j].hash, Table[j].name );
				r += (ind[R*scale] * Table[j].count);
				g += (ind[G*scale] * Table[j].count);
				b += (ind[B*scale] * Table[j].count);
				c += Table[j].count;
			}
		}
		if( c == 0 ){
			continue;
		}
		Color[i].R = r / c;
		Color[i].G = g / c;
		Color[i].B = b / c;
		if( Color[i].R > 0xff ){
			printf( "%d:R not regal %d\n", i, Color[i].R );
			Color[i].R = 0xff;
		}
		if( Color[i].R < 0 ){
			printf( "%d:R not regal %d\n", i, Color[i].R );
			Color[i].R = 0;
		}
		if( Color[i].G > 0xff ){
			printf( "%d:G not regal %d\n", i, Color[i].G );
			Color[i].G = 0xff;
		}
		if( Color[i].G < 0 ){
			printf( "%d:G not regal %d\n", i, Color[i].G );
			Color[i].G = 0;
		}
		if( Color[i].B > 0xff ){
			printf( "%d:B not regal %d\n", i, Color[i].B );
			Color[i].B = 0xff;
		}
		if( Color[i].B < 0 ){
			printf( "%d:B not regal %d\n", i, Color[i].B );
			Color[i].B = 0;
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function get_pallet()                                   */
/*                                                                    */
/*            args  struct Color *col : result color pallet           */
/*                                                                    */
/*            return int : total color number                         */
/*                                                                    */
/*            function: Return result color pallet.                   */
/*                                                                    */
/**********************************************************************/
int get_pallet( struct Color *col )
{
	int i;

	for( i = 0; i < Index; i++ ){
		col[i].R = Color[i].R;
		col[i].G = Color[i].G;
		col[i].B = Color[i].B;
	}
	return( Index );
}

/**********************************************************************/
/*                                                                    */
/*            function end_tree()                                     */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Free the TREE.                                */
/*                                                                    */
/**********************************************************************/
void end_tree( void )
{
	free( Color );
}

/**********************************************************************/
/*                                                                    */
/*            function reduce_proc()                                  */
/*                                                                    */
/*            args  int thr : use count threthold                     */
/*                  struct Table *t : color TREE                      */
/*                  int r  : R differencial                           */
/*                  int g  : G differencial                           */
/*                  int b  : B differencial                           */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Search the nearest node.                      */
/*                                                                    */
/**********************************************************************/
void reduce_proc( int thr, struct Table *t, int r, int g, int b )
{
	int at, found, name;

	if( t->count > thr ){
		return;
	}
	if( (name = new_name( t->name, r, g, b )) != -1 ){
		if( (found = search_tree( t->hash, name )) != -1 ){
			if( t->count + Table[found].count > thr ){
				return;
			}
			if( t->index == 0 && Table[found].index == 0 ){
				Colors--;
				t->index = Table[found].index = Index++;
			} else if( t->index == 0 ){
				Colors--;
				t->index = Table[found].index;
			} else if( Table[found].index == 0 ){
				Colors--;
				Table[found].index = t->index;
			}
		}
	} else if( new_hashname( t, &at, &name, r, g, b ) == 0 ){
		if( (found = search_tree( at, name )) != -1 ){
			if( t->count + Table[found].count > thr ){
				return;
			}
			if( t->index == 0 && Table[found].index == 0 ){
				Colors--;
				t->index = Table[found].index = Index++;
			} else if( t->index == 0 ){
				Colors--;
				t->index = Table[found].index;
			}
		}
	}
}


#define T3   10000
#define T2   10000
#define T1   1

static struct RTable {	/* define the search order */
	int mag;	/* priority */
	int R;		/* R differencial */
	int G;		/* G differencial */
	int B;		/* B differencial */
} RT[] = {
	{ T3, 2, 0, 0 },
	{ T3, -2, 0, 0 },
	{ T3, 0, 2, 0 },
	{ T3, 0, -2, 0 },
	{ T3, 0, 0, 2 },
	{ T3, 0, 0, -2 },
	{ T2, 1, 1, 0 },
	{ T2, 1, -1, 0 },
	{ T2, -1, 1, 0 },
	{ T2, -1, -1, 0 },
	{ T2, 0, 1, 1 },
	{ T2, 0, 1, -1 },
	{ T2, 0, -1, 1 },
	{ T2, 0, -1, -1 },
	{ T2, 1, 0, 1 },
	{ T2, 1, 0, -1 },
	{ T2, -1, 0, 1 },
	{ T2, -1, 0, -1 },
	{ T1, 1, 0, 0 },
	{ T1, -1, 0, 0 },
	{ T1, 0, 1, 0 },
	{ T1, 0, -1, 0 },
	{ T1, 0, 0, 1 },
	{ T1, 0, 0, -1 },
};

#define RNUM  (sizeof(RT) / sizeof(struct RTable))

/**********************************************************************/
/*                                                                    */
/*            function reduce()                                       */
/*                                                                    */
/*            args  int target : total colors must be                 */
/*                                                                    */
/*            return int : 0==success  -1==fail                       */
/*                                                                    */
/*            function: reduce interface, merge the TREE.             */
/*            D2: Re-write. messages, DEBUG, loop condition.          */
/*                                                                    */
/**********************************************************************/
int reduce( int target )
{
	int i, t, k;
	int prev;

	Colors = NowTable; Index = 1;
	fprintf(stderr, "tries Colors=%d into %d ..", Colors, target );
	prev = Colors;
	for( t = 100; t < 20000; t += 100 ){
		for( i = 0; i < NowTable; i++ ){
			for( k = 0; k < RNUM; k++ ){
				reduce_proc( t / RT[k].mag, &Table[i], RT[k].R, RT[k].G, RT[k].B );
				if( Colors <= target ){
					goto ex;
				}
			}
		}
#ifdef DEBUG
		if( prev != Colors ){
		     fprintf(stderr, "%d:%d ", t,Colors );
		     prev = Colors;
		}
#endif
	}
ex:	if( Colors > target ){
		fprintf( stderr, "failed until %d colors\n", Colors );
		return( -1 );
	} else {
		fprintf( stderr, "success at %d pixel used\n",  t );
		return( 0 );
	}
}

/* D2: OLD version.
int reduce( int target )
{
	int i, t, k;

	Colors = NowTable; Index = 1;
fprintf(stderr, "Colors=%d ... ", Colors );
	for( t = 100; t < 100000; t += 100 ){
		for( i = 0; i < NowTable; i++ ){
			for( k = 0; k < RNUM; k++ ){
				reduce_proc( t / RT[k].mag, &Table[i], RT[k].R, RT[k].G, RT[k].B );
				if( Colors <= target ){
					goto ex;
				}
			}
		}
		fprintf(stderr, "%d", Colors );
	}
ex:	if( Colors > target ){
		fprintf( stderr, "Total Colors %d reduced to %d, reduce failed\n", NowTable, Colors );
		return( -1 );
	} else {
		fprintf( stderr, "Reduce %d colors to %d, until %d pixel used\n", NowTable, Colors, t );
		return( 0 );
	}
}
  D2 OLD version end */

/**********************************************************************/
/*                                                                    */
/*            function add_table()                                    */
/*                                                                    */
/*            args  unsigned char *s : color index R,G,B              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Add new entry to the TREE.                    */
/*                                                                    */
/**********************************************************************/
void add_table( unsigned char *s )
{
	unsigned char hash, name;
	unsigned char r, g, b;

	if( s[0] > 0x0f ){
		fprintf( stderr, "illegal value R:%d\n", s[0] );
		return;
	}
	if( s[1] > 0x0f ){
		fprintf( stderr, "illegal value R:%d\n", s[1] );
		return;
	}
	if( s[2] > 0x0f ){
		fprintf( stderr, "illegal value R:%d\n", s[2] );
		return;
	}
	r = (s[0] & 0x03);
	g = (s[1] & 0x03);
	b = (s[2] & 0x03);
	name = ((r << 4) | (g << 2) | b);
	r = ((s[0] >> 2) & 0x03);
	g = ((s[1] >> 2) & 0x03);
	b = ((s[2] >> 2) & 0x03);
	hash = ((r << 4) | (g << 2) | b);

	add_tree( hash, name );
}


/**********************************************************************/
/*                                                                    */
/*            function search_table()                                 */
/*                                                                    */
/*            args  unsigned char *s   : original color index(RGB)    */
/*                  unsigned char *ret : reduced real color(RGB)      */
/*                                                                    */
/*            return int : 0==success  -1==fail                       */
/*                                                                    */
/*            function: Get new reduced real color by original index. */
/*                                                                    */
/**********************************************************************/
int search_table( unsigned char *s, unsigned char *ret )
{
	unsigned char hash, name;
	unsigned char r, g, b;
	int at;

	r = (s[0] & 0x03);
	g = (s[1] & 0x03);
	b = (s[2] & 0x03);
	name = ((r << 4) | (g << 2) | b);
	r = ((s[0] >> 2) & 0x03);
	g = ((s[1] >> 2) & 0x03);
	b = ((s[2] >> 2) & 0x03);
	hash = ((r << 4) | (g << 2) | b);

	at = search_tree( hash, name );
	if( at == -1 ){
		fprintf( stderr, "cannot find %x:%x\n", hash, name );
		return( -1 );
	}
	ret[0] = (unsigned char)Color[Table[at].index].R;
	ret[1] = (unsigned char)Color[Table[at].index].G;
	ret[2] = (unsigned char)Color[Table[at].index].B;
	ret[3] = Table[at].index;
	return( 0 );
}
