/**********************************************************************/
/*                                                                    */
/*             reduce.c  gedo PLUG specified routine                  */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>
#include "plug.h"
#include "reduce.h"

int Mode = FLAT;
int PalletMode;

int Colors = 256;
int ColorFlag = 1;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec    = { "reduce" };
char *Comment = { "reduce/pallet color efffect" };

PLUGITEM Item[] = {
{ "colors", 	GETINT,  &Colors,	2, 65535,   "reduce target color" },
{ "mode=color",  SWITCH, &ColorFlag, 1, 'C', "color mode"  },
{ "mode=b/w", SWITCH, &ColorFlag, 0, 'B', "b/w mode, discard color" },
{ "jaggy",    SWITCH, &Mode,  ZAGGY, 'J',  "dizzar mode(pallet mode neglect)" },
{ "flat",     SWITCH, &Mode,  FLAT, 'F',   "flat mode(pallet mode neglect)" },
{ "pallet",SWITCH | INITONLY, &PalletMode, 1, 'P', "pallet mode" },
{ "reduce",SWITCH | INITONLY, &PalletMode, 0, 'R', "reduce mode" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};

struct Tags Tag[] = {
{ ENTRY, 0, "reduce target color ", "colors=%d" },
{ RADIO, 1, "Color Mode " },
{ ITEM, 1, "color(c)", "mode=color" },
{ ITEM, 2, "b/w(b)", "mode=b/w" },
{ RADIOEND, -1 },
{ RADIO, 3, "Dizzer Mode " },
{ ITEM, 3, "dizzer(z)", "zaggy" },
{ ITEM, 4, "flat(f)", "flat" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

int ItemNum = sizeof(Item)/sizeof(PLUGITEM);
DEFAULTSTATUS;
/********************** description end ********************/

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	struct ReduceInfo ri;
	int ret;

	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}

	ri.mode = TRANSONLY;
	ri.image = NULL;
	ri.pallet = NULL;

	ri.submode = 0;
	if( ColorFlag ){
		ri.submode |= COLOR;
	} else {
		ri.submode |= BW;
	}

	ri.target = Colors;
	if( PalletMode ){
		if( ColorFlag ){
			ri.submode |= TOTALNUM;
			if( Colors >= 150 ){
				ri.tmp_colors = 16;
			} else if( Colors >= 40 ){
				ri.tmp_colors = 8;
			} else if( Colors >= 8 ){
				ri.tmp_colors = 4;
			} else {
				ri.tmp_colors = 2;
			}
			ri.submode |= ZAGGY;
		} else {
			ri.target = ri.tmp_colors = Colors;
			ri.submode |= INDEPENDENT;
			if( Colors < 8 ){
				ri.submode |= ZAGGY;
			} else {
				ri.submode |= FLAT;
			}
		}
	} else {
		ri.target = ri.tmp_colors = Colors;
		ri.submode |= INDEPENDENT;
		if( Mode == ZAGGY ){
			ri.submode |= ZAGGY;
		} else {
			ri.submode |= FLAT;
		}
	}
	ret = get_reduce_image( file->buf, &ri );
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( PalletMode ){
			TagNum = 5;
		} else {
			TagNum = 9;
		}
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}

