/**********************************************************************/
/*                                                                    */
/*            reducelib.c  gedo PLUG specified routine                */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*           C1: improve bilevel logic                                */
/*           D1: under 256 colors, not reduce color num               */
/*           D2: To repeat reduce proc, ri->tmp_color decreased and   */
/*              wrap get_reduce_image(), ret value -2 represents      */
/*              reduce failure at pallet_colors(), get_reduce_image2() */
/*           D2: caress mistake fixed at init_scale(), check_colornum() */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>
#include "plug.h"
#include "reduce.h"

struct Scale {			/* Transfar real color to divided color */
	unsigned char col;	/* real color becomes <col> */
	unsigned char ind;	/* <col> index */
};

static struct Scale  Scale[256];  /* Transfar Array */
static unsigned char Zindex[256]; /* Reverse index of Scale */

static int Zaggy1 = 4;		 /* 4*4 tail become a unit */
static int Zaggy2 = 4;		 /* a unit represent <Zaggy2> gradation */

static unsigned char ZMask4[4][4][4] = {	/* Zaggy2 == 4 Trans Mask */
	{{1,1,1,1},{1,1,1,1},{1,1,1,1},{1,1,1,1}},
	{{1,1,1,1},{1,0,0,1},{1,0,0,1},{1,1,1,1}},
	{{0,1,0,1},{1,0,1,0},{0,1,0,1},{1,0,1,0}},
	{{0,0,0,0},{0,1,1,0},{0,1,1,0},{0,0,0,0}} };

static unsigned char ZMask8[8][4][4] = {	/* Zaggy2 == 8 Trans Mask */
	{{1,1,1,1},{1,1,1,1},{1,1,1,1},{1,1,1,1}},
	{{1,1,1,1},{1,0,1,1},{1,1,0,1},{1,1,1,1}},
	{{1,1,1,1},{1,0,0,1},{1,0,0,1},{1,1,1,1}},
	{{1,1,0,1},{1,0,1,0},{0,1,0,1},{1,0,1,1}},
	{{0,1,0,1},{1,0,1,0},{0,1,0,1},{1,0,1,0}},
	{{0,1,0,1},{1,0,0,0},{0,0,0,1},{1,0,1,0}},
	{{1,0,0,0},{0,0,0,1},{0,1,0,0},{0,0,1,0}},
	{{0,0,0,0},{0,0,0,1},{0,1,0,0},{0,0,0,0}} };

#define get_y(r,g,b) (unsigned int)(0.299*(float)r+0.587*(float)g+0.114*(float)b) 

static unsigned char (*Zaggy_color)( unsigned char, int, int);

void init_tree(void);
int reduce(int target);
void add_table( unsigned char *s );
int search_table( unsigned char *s, unsigned char *ret );
int make_color( unsigned char *ind, int scale );
int get_pallet( struct Color *col );
void end_tree(void);

/**********************************************************************/
/*                                                                    */
/*            function pallet_colors()                                */
/*                                                                    */
/*            args  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                  struct Shm *buf   : Image data                    */
/*                                                                    */
/*            return  int : 0==suucess  -1==error  -2==reduce fail    */
/*                                                                    */
/*            function: Reduce executive routine.                     */
/*                                                                    */
/**********************************************************************/
static int pallet_colors( struct ReduceInfo *ri, struct Shm *buf )
{
	int w, h;
	unsigned int *p;
	unsigned char *q;
	unsigned char buff[3];
	unsigned char ret[4];
	int i;

	w = buf->width;
	h = buf->height;
	p = buf->data;
	init_tree();

/* regist all pixel to the tree */
	for( i = 0; i < h * w; i++, p++ ){
		buff[0] = GET_RED(*p);
		buff[1] = GET_GREEN(*p);
		buff[2] = GET_BLUE(*p);
		add_table( buff );
	}

/* merge tree to ri->target */
	if( reduce( ri->target ) != 0 ){
/* D2: To repeat failed reduce, report reduce fail as -2 */
		return( -2 );
/* D2 del      	return( -1 );*/
	}

/* send real color infomation to the tree */
	if( make_color( Zindex, Zaggy2 ) != 0 ){
		fprintf( stderr, "make_color error\n" );
		return( -1 );
	}

/* translate real color to the reduced */
	p = buf->data;
	if( ri->mode == GETPALLETCOLOR ){
		q = ri->image;
	} else {
		q = NULL;
	}
	for( i = 0; i < h * w; i++, p++ ){
		buff[0] = GET_RED(*p);
		buff[1] = GET_GREEN(*p);
		buff[2] = GET_BLUE(*p);
		if( search_table( buff, ret ) == -1 ){
			return( -1 );
		}
		if( ri->mode == GETPALLETCOLOR ){
			*q++ = ret[3];
		}
		*p = ((*p & MASK) | (ret[0] << 16) | (ret[1] << 8) | ret[2]); 
	}

/* if need, get pallet */
	if( ri->mode == GETPALLETCOLOR ){
		if( get_pallet( ri->pallet ) <= 0 ){
			return( -1 );
		}
	}

	end_tree();
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function init_scale()                                   */
/*                                                                    */
/*            args  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: By ri, set internal data to prepare.          */
/*                                                                    */
/**********************************************************************/
static void init_scale( struct ReduceInfo *ri )
{
	int i, j;
	float t;
	unsigned char thr[256];
	int col;

	if( (ri->submode & TOTALNUM) && (ri->submode & ZAGGY) ){
		if( ri->submode & COLOR ){
			Zaggy1 = Zaggy2 = 4;
		} else {
  			if( ri->tmp_colors <= 8 ){
				Zaggy1 = 4; Zaggy2 = 8;
			} else {
				Zaggy1 = Zaggy2 = 4;
			}
		}
	}

	if( (ri->submode & TOTALNUM) || (ri->submode & ZAGGY) ){
		col = (ri->tmp_colors - 1) * Zaggy2 + 1;
	} else {
		col = ri->tmp_colors;
	}
	if( col < 256 ){
/* D2: careless mistake, not create good Scale */
		t = 256.0 / (float)col;
/* D2 del	t = 256 / col; */
		for( i = 0; i < col - 1; i++ ){
			thr[i] = (unsigned char)(t * (float)(i+1));
		}
		for( i = j = 0; i < 256; i++ ){
			if( j < col - 1 && i >= thr[j] ){
				j++;
			}
			Scale[i].ind = j;
			if( j == 0 ){
				Scale[i].col = 0;
			} else if( j >= col - 1 ){
				Scale[i].col = 0xff;
			} else {
				Scale[i].col = ((int)thr[j-1] + (int)thr[j]) / 2;
			}
			Zindex[j] = Scale[i].col;
		}
	} else {
		fprintf( stderr, "target color > 256, ignore init_scale()\n" );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function change_proc()                                  */
/*                                                                    */
/*            args  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                  unsigned int v   : real color                     */
/*                                                                    */
/*            return  unsigned int   : scaled color                   */
/*                                                                    */
/*            function: 1st step of FLAT, scale the real color.       */
/*                                                                    */
/**********************************************************************/
static unsigned int change_proc( struct ReduceInfo *ri, unsigned int v )
{
	unsigned int r, g, b;
	unsigned int ret;

	r = GET_RED(v);
	g = GET_GREEN(v);
	b = GET_BLUE(v);
	if( ri->submode & COLOR ){
		if( ri->tmp_colors < 256 ){
			r = Scale[r].col;
			g = Scale[g].col;
			b = Scale[b].col;
		}
		ret = ((v & MASK) | (r << 16) | (g << 8) | b);
	} else {
		g = get_y( r, g, b );
		if( ri->tmp_colors < 256 ){
			g = Scale[g].col;
		}
		ret = ((v & MASK) | (g << 16) | (g << 8) | g);
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*            function zaggy_palletcolor()                            */
/*                                                                    */
/*            args  unsigned char v : color index each R, G, B        */
/*                  int x  : X position of pixel % 4                  */
/*                  int y  : Y position of pixel % 4                  */
/*                                                                    */
/*            return  unsigned char : Zaggied color index each R,G,B  */
/*                                                                    */
/*            function: If get pallet, RGB is an index of scaled,     */
/*                     and must be dizzar process.                    */
/**********************************************************************/
static unsigned char zaggy_palletcolor( unsigned char v, int x, int y )
{
	unsigned char ret;
	int ind, ind2;

	ind = Scale[v].ind % Zaggy2;
	ind2 = Scale[v].ind / Zaggy2;
	if( Zaggy2 == 4 ){
		if( ZMask4[ind][x][y] ){
			ret = ind2;
		} else {
			ret = ind2 + 1;
		}
	} else {
		if( ZMask8[ind][x][y] ){
			ret = ind2;
		} else {
			ret = ind2 + 1;
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function zaggy_color()                                  */
/*                                                                    */
/*            args  unsigned char v : real color  each R, G, B        */
/*                  int x  : X position of pixel % 4                  */
/*                  int y  : Y position of pixel % 4                  */
/*                                                                    */
/*            return  unsigned char : Zaggied real color  each R,G,B  */
/*                                                                    */
/*            function: If trans only, RGB is a real color,           */
/*                     and must be dizzar process.                    */
/*                                                                    */
/**********************************************************************/
static unsigned char zaggy_color( unsigned char v, int x, int y )
{
	unsigned char ret;
	int ind, ind2;

	ind = Scale[v].ind % Zaggy2;
	ind2 = Scale[v].ind / Zaggy2;
	if( Zaggy2 == 4 ){
		if( ZMask4[ind][x][y] ){
			ret = Zindex[ind2 * Zaggy2];
		} else {
			ret = Zindex[(ind2 + 1) * Zaggy2];
		}
	} else {
		if( ZMask8[ind][x][y] ){
			ret = Zindex[ind2 * Zaggy2];
		} else {
			ret = Zindex[(ind2 + 1) * Zaggy2];
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function zaggy_coloreach()                              */
/*                                                                    */
/*            args  unsigned char v : color pixel, index or real      */
/*                  int x  : X position of pixel % 4                  */
/*                  int y  : Y position of pixel % 4                  */
/*                                                                    */
/*            return  unsigned char : Zaggied pixel, index or real    */
/*                                                                    */
/*            function: If COLOR MODE, each RGB must be dizzared.     */
/*                                                                    */
/**********************************************************************/
static unsigned int zaggy_coloreach( unsigned int v, int x, int y )
{
	unsigned int r, g, b;
	unsigned int ret;

	r = GET_RED(v);
	r = (*Zaggy_color)( r, x, y );
	g = GET_GREEN(v);
	g = (*Zaggy_color)( g, x, y );
	b = GET_BLUE(v);
	b = (*Zaggy_color)( b, x, y );
	ret = ((v & MASK) | (r << 16) | (g << 8) | b);
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function zaggy_bweach()                                 */
/*                                                                    */
/*            args  unsigned char v : B/W pixel, index or real        */
/*                  int x  : X position of pixel % 4                  */
/*                  int y  : Y position of pixel % 4                  */
/*                                                                    */
/*            return  unsigned char : Zaggied pixel, index or real    */
/*                                                                    */
/*            function: If B/W MODE, calc Y and dizzar.               */
/*                                                                    */
/**********************************************************************/
static unsigned int zaggy_bweach( unsigned int v, int x, int y )
{
	unsigned int g;
	unsigned int ret;

	g = GET_GREEN(v);
	g = (*Zaggy_color)( g, x, y );
	ret = ((v & MASK) | (g << 16) | (g << 8) | g);
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function zaggy_mode()                                   */
/*                                                                    */
/*            args  struct Shm *buf   : Image data                    */
/*                  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: Do the Dizzar process for each pixel.         */
/*                                                                    */
/**********************************************************************/
static void zaggy_mode( struct Shm *buf, struct ReduceInfo *ri )
{
	int width, height;
	unsigned int *p;
	int i, j;
	unsigned int (*Zaggy_each)( unsigned int, int, int);

	width = buf->width;
	height = buf->height;
	p = buf->data;

	if( ri->submode & COLOR ){
		Zaggy_each = zaggy_coloreach;
	} else {
		Zaggy_each = zaggy_bweach;
	}
	if( ri->submode & TOTALNUM ){
		Zaggy_color = zaggy_palletcolor;
		for( i = 0; i < height; i++ ){
			for( j = 0; j < width; j++, p++ ){
				*p = (*Zaggy_each)( *p, i % Zaggy1, j % Zaggy1 );
			}
		}
	} else {
		Zaggy_color = zaggy_color;
		for( i = 0; i < height; i++ ){
			for( j = 0; j < width; j++, p++ ){
				if( MASKOFF(*p) ){
					*p = (*Zaggy_each)( *p, i % Zaggy1, j % Zaggy1 );
				}
			}
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*            function gray_scale()                                   */
/*                                                                    */
/*            args  struct Shm *buf   : Image data                    */
/*                  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: ONLY case (BW | INDEPENDENT | FLAT), gray     */
/*                     scale conversion only.                         */
/**********************************************************************/
void gray_scale( struct Shm *buf, struct ReduceInfo *ri )
{
	int width, height;
	int i;
	unsigned int *p;
	unsigned char *q;
	unsigned int r, g, b;

	width = buf->width;
	height = buf->height;
	p = buf->data;
	q = ri->image;
	for( i = 0; i < width * height; i++, p++ ){
		r = GET_RED(*p);
		g = GET_GREEN(*p);
		b = GET_BLUE(*p);
		*q++ = get_y( r, g, b );
	}
/* add C1 */
	if( ri->pallet == NULL ){
		return;
	}
	for( i = 0; i < 256; i++ ){
		ri->pallet[i].R = ri->pallet[i].G = ri->pallet[i].B = (unsigned char)i;
	}
}

/* add C1 -- for bilevel logic improved */
unsigned char Biarray[][6][6] = {
/*0*/	{ {0,0,0,0,0,0}, {0,0,0,0,0,0}, {0,0,0,0,0,0}, 
		  {0,0,0,0,0,0}, {0,0,0,0,0,0}, {0,0,0,0,0,0} },
/*1*/	{ {0,0,0,0,0,0}, {0,0,0,0,0,0}, {0,0,1,0,0,0}, 
		  {0,0,0,0,0,0}, {0,0,0,0,0,0}, {0,0,0,0,0,0} },
/*2*/	{ {0,0,0,0,0,0}, {0,1,0,0,0,0}, {0,0,0,0,0,0}, 
		  {0,0,0,0,0,0}, {0,0,0,0,1,0}, {0,0,0,0,0,0} },
/*3*/	{ {1,0,0,0,0,0}, {0,0,0,0,0,0}, {0,0,0,0,1,0}, 
		  {0,0,0,0,0,0}, {0,0,1,0,0,0}, {0,0,0,0,0,0} },
/*4*/	{ {0,1,0,0,0,0}, {0,0,0,0,1,0}, {0,0,0,0,0,0}, 
		  {0,0,1,0,0,0}, {1,0,0,0,0,0}, {0,0,0,0,0,0} },
/*5*/	{ {1,0,0,0,0,0}, {0,0,0,0,1,0}, {0,0,0,0,0,0}, 
		  {0,0,1,0,0,0}, {0,0,0,0,0,1}, {0,1,0,0,0,0} },
/*6*/	{ {0,1,0,0,0,0}, {0,0,0,0,0,1}, {0,0,1,0,0,0}, 
		  {1,0,0,0,0,0}, {0,1,0,0,0,0}, {0,0,0,1,0,0} },
/*7*/	{ {1,0,0,0,0,0}, {0,0,1,0,0,0}, {1,0,0,0,0,0}, 
		  {0,0,0,1,0,1}, {0,1,0,0,0,0}, {0,0,0,0,1,0} },
/*8*/	{ {1,0,0,0,0,0}, {0,0,1,0,1,0}, {1,0,0,0,0,0}, 
		  {0,0,0,1,0,1}, {0,1,0,0,0,0}, {0,0,0,0,1,0} },
/*9*/	{ {1,0,0,0,0,0}, {0,0,1,0,1,0}, {1,0,0,0,0,0}, 
		  {0,0,0,1,0,1}, {0,1,0,0,0,0}, {0,0,1,0,1,0} },
/*10*/	{ {0,0,1,0,0,0}, {1,0,0,1,0,0}, {0,1,0,0,1,0}, 
		  {0,0,0,1,0,0}, {0,0,1,0,0,1}, {0,1,0,0,1,0} },
/*11*/	{ {0,0,0,0,0,1}, {1,0,0,1,0,0}, {0,1,0,0,1,0}, 
		  {1,0,0,1,0,0}, {0,0,1,0,0,1}, {0,1,0,0,1,0} },
/*12*/	{ {0,0,1,0,0,1}, {1,0,0,1,0,0}, {0,1,0,0,1,0}, 
		  {1,0,0,1,0,0}, {0,0,1,0,0,1}, {0,1,0,0,1,0} },
/*13*/	{ {0,0,0,1,0,1}, {1,0,0,0,1,0}, {0,1,0,0,0,1}, 
		  {0,0,1,0,1,0}, {0,1,0,1,0,1}, {1,0,1,0,0,0} },
/*14*/	{ {0,1,0,1,0,1}, {1,0,0,0,1,0}, {0,1,0,0,0,1}, 
		  {0,0,1,0,1,0}, {0,1,0,1,0,1}, {1,0,1,0,0,0} },
/*15*/	{ {0,1,0,1,0,1}, {1,0,0,0,1,0}, {0,1,0,1,0,1}, 
		  {0,0,1,0,1,0}, {0,1,0,1,0,1}, {1,0,1,0,0,0} },
/*16*/	{ {0,1,0,1,0,1}, {1,0,0,0,1,0}, {0,1,0,1,0,1}, 
		  {0,0,1,0,1,0}, {0,1,0,1,0,1}, {1,0,1,0,1,0} },
/*17*/	{ {0,1,0,1,0,1}, {1,0,0,0,1,0}, {0,1,0,1,0,1}, 
		  {1,0,1,0,1,0}, {0,1,0,1,0,1}, {1,0,1,0,1,0} },
/*18*/	{ {0,1,0,1,0,1}, {1,0,1,0,1,0}, {0,1,0,1,0,1}, 
		  {1,0,1,0,1,0}, {0,1,0,1,0,1}, {1,0,1,0,1,0} },

/*17*/	{ {1,0,1,0,1,0}, {0,1,1,1,0,1}, {1,0,1,0,1,0}, 
		  {0,1,0,1,0,1}, {1,0,1,0,1,0}, {0,1,0,1,0,1} },
/*16*/	{ {1,0,1,0,1,0}, {0,1,1,1,0,1}, {1,0,1,0,1,0}, 
		  {1,1,0,1,0,1}, {1,0,1,0,1,0}, {0,1,0,1,0,1} },
/*15*/	{ {1,0,1,0,1,0}, {0,1,1,1,0,1}, {1,0,1,0,1,0}, 
		  {1,1,0,1,0,1}, {1,0,1,0,1,0}, {0,1,0,1,1,1} },
/*14*/	{ {1,0,1,0,1,0}, {0,1,1,1,0,1}, {1,0,1,1,1,0}, 
		  {1,1,0,1,0,1}, {1,0,1,0,1,0}, {0,1,0,1,1,1} },
/*13*/	{ {1,1,1,0,1,0}, {0,1,1,1,0,1}, {1,0,1,1,1,0}, 
		  {1,1,0,1,0,1}, {1,0,1,0,1,0}, {0,1,0,1,1,1} },
/*12*/	{ {1,1,0,1,1,0}, {0,1,1,0,1,1}, {1,0,1,1,0,1}, 
		  {0,1,1,0,1,1}, {1,1,0,1,1,0}, {1,0,1,1,0,1} },
/*11*/	{ {1,1,1,1,1,0}, {0,1,1,0,1,1}, {1,0,1,1,0,1}, 
		  {0,1,1,0,1,1}, {1,1,0,1,1,0}, {1,0,1,1,0,1} },
/*10*/	{ {1,1,0,1,1,1}, {0,1,1,0,1,1}, {1,0,1,1,0,1}, 
		  {1,1,1,0,1,1}, {1,1,0,1,1,0}, {1,0,1,1,0,1} },
/*9*/	{ {0,1,1,1,1,1}, {1,1,0,1,0,1}, {0,1,1,1,1,1}, 
		  {1,1,1,0,1,0}, {1,0,1,1,1,1}, {1,1,0,1,0,1} },
/*8*/	{ {0,1,1,1,1,1}, {1,1,0,1,0,1}, {0,1,1,1,1,1}, 
		  {1,1,1,0,1,0}, {1,0,1,1,1,1}, {1,1,1,1,0,1} },
/*7*/	{ {0,1,1,1,1,1}, {1,1,0,1,1,1}, {0,1,1,1,1,1}, 
		  {1,1,1,0,1,0}, {1,0,1,1,1,1}, {1,1,1,1,0,1} },
/*6*/	{ {1,0,1,1,1,1}, {1,1,1,1,1,0}, {1,1,0,1,1,1}, 
		  {0,1,1,1,1,1}, {1,0,1,1,1,1}, {1,1,1,0,1,1} },
/*5*/	{ {0,1,1,1,1,1}, {1,1,1,1,0,1}, {1,1,1,1,1,1}, 
		  {1,1,0,1,1,1}, {1,1,1,1,1,0}, {1,0,1,1,1,1} },
/*4*/	{ {1,0,1,1,1,1}, {1,1,1,1,0,1}, {1,1,1,1,1,1}, 
		  {1,1,0,1,1,1}, {0,1,1,1,1,1}, {1,1,1,1,1,1} },
/*3*/	{ {0,1,1,1,1,1}, {1,1,1,1,1,1}, {1,1,1,1,0,1}, 
		  {1,1,1,1,1,1}, {1,1,0,1,1,1}, {1,1,1,1,1,1} },
/*2*/	{ {1,1,1,1,1,1}, {1,0,1,1,1,1}, {1,1,1,1,1,1}, 
		  {1,1,1,1,1,1}, {1,1,1,1,0,1}, {1,1,1,1,1,1} },
/*1*/	{ {1,1,1,1,1,1}, {1,1,1,1,1,1}, {1,1,0,1,1,1}, 
		  {1,1,1,1,1,1}, {1,1,1,1,1,1}, {1,1,1,1,1,1} },
/*0*/	{ {1,1,1,1,1,1}, {1,1,1,1,1,1}, {1,1,1,1,1,1}, 
		  {1,1,1,1,1,1}, {1,1,1,1,1,1}, {1,1,1,1,1,1} },
};


/**********************************************************************/
/*                                                                    */
/*            function gray_biarray()                                 */
/*                                                                    */
/*            args  int col:  Y value of the pixel(blightness)        */
/*                  int   x:  X potion of the pixel                   */
/*                  int   y:  Y potion of the pixel                   */
/*                                                                    */
/*            return  Binary value of the blightness(from matrix)     */
/*                                                                    */
/*            function: ONLY case (BW | TOTALNUM | ZAGGY), BILEVEL    */
/*                     B/W conversion only.(Encode bilevel)           */
/**********************************************************************/
unsigned char get_biarray( int col, int x, int y )
{
	int num;
	int x1, y1;

	if( col < 16 ){
		num = col / 8;
	} else if( col < 248 ){
		num = (col - 16) / 7 + 2;
	} else {
		num = 35;
	}
	x1 = x % 6; y1 = y % 6;
	return( Biarray[num][x1][y1] );
}


/**********************************************************************/
/*                                                                    */
/*            function make_bilevel()                                 */
/*                                                                    */
/*            args  struct Shm *buf   : Image data                    */
/*                  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: ONLY case (BW | TOTALNUM | ZAGGY), BILEVEL    */
/*                     B/W conversion only.                           */
/**********************************************************************/
int make_bilevel( struct Shm *buf, struct ReduceInfo *ri )
{
	int width, height;
	int i, j;
	unsigned int *p;
	unsigned char *q;
	unsigned int r, g, b, y;
	int flag;
	int count;
	double ratio;

	width = buf->width;
	height = buf->height;
	p = buf->data;
	q = ri->image;

	flag = 0; count = 0;
	for( i = 0; i < width * height; i++, p++ ){
	     if( (*p & ~MASK) == ~MASK 
		 || (*p & ~MASK) == 0 ){
	     } else {
		  flag = 1;
		  r = GET_RED(*p);
		  g = GET_GREEN(*p);
		  b = GET_BLUE(*p);
		  y = get_y( r, g, b );
		  if( y > 0x30 && y < 0xd0 ){
		       count++;
		  }
	     }
	}

	p = buf->data;
	ratio = (double)count;
	ratio /= (double)(width*height);
	if( flag == 0 ){
	     for( i = 0; i < width * height; i++, p++, q++ ){
		  if( (*p & ~MASK) == ~MASK ){
		       *q = 1;
		  } else if( (*p & ~MASK) == 0 ){
		       *q = 0;
		  } else {
		       fprintf( stderr, "cannot happen\n" );
		  }
	     }
	} else if( ratio < 0.05 ){
	     for( i = 0; i < height * width; i++, p++, q++ ){
		  r = GET_RED(*p);
		  g = GET_GREEN(*p);
		  b = GET_BLUE(*p);
		  y = get_y( r, g, b );
		  if( y <= 0x80 ){
		       *q = 0;
		  } else {
		       *q = 1;
		  }
	     }
	} else {
	     for( i = 0; i < height; i++ ){
		  for( j = 0; j < width; j++, p++ ){
		       r = GET_RED(*p);
		       g = GET_GREEN(*p);
		       b = GET_BLUE(*p);
		       *q++ = get_biarray( get_y( r, g, b ), j, i );
		  }
	     }
	}
	ri->pallet[0].R = ri->pallet[0].G = ri->pallet[0].B = 0;
	ri->pallet[1].R = ri->pallet[1].G = ri->pallet[1].B = 0xff;
	return( 0 );
}

/* added C1 */

/* add D1, under 256 not reduce colors */
struct Color *Parse = NULL;	/* purse of the colors */
int ParseNum = 0;		/* alloced Parse num */
int ParseNow = 0;		/* used Parse num */

/**********************************************************************/
/*                                                                    */
/*            function clear_parse()                                  */
/*                                                                    */
/*            args  int max  : maximum of the pallet colors           */
/*                                                                    */
/*            return  int  : success==0  realloc error==-1            */
/*                                                                    */
/*            function: Initialize the purse of colors.               */
/*                                                                    */
/**********************************************************************/
int clear_parse( int max )
{
	void *tmp;

	if( ParseNum != max ){
		if( (tmp = realloc( Parse, max * sizeof(struct Color) ))
			 == NULL ){
			fprintf( stderr, "realloc error\n" );
			return( -1 );
		}
		ParseNum = max;
		Parse = (struct Color *)tmp;
	}
	ParseNow = 0;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function  parse_256color()                              */
/*                                                                    */
/*            args  unsigned int col  : the pixel                     */
/*                                                                    */
/*            return  int : the number of the registrated colors now  */
/*                                                                    */
/*            function: Registrate the pallet from the color at the   */
/*                     pixel.                                         */
/**********************************************************************/
int parse_256color( unsigned int col )
{
	int i;
	unsigned char r, g, b;

	r = GET_RED(col);
	g = GET_GREEN(col);
	b = GET_BLUE(col);
	for( i = 0; i < ParseNow; i++ ){
		if( r != Parse[i].R ){
			continue;
		}
		if( g != Parse[i].G ){
			continue;
		}
		if( b != Parse[i].B ){
			continue;
		}
		return( ParseNow );
	}
	Parse[i].R = r;
	Parse[i].G = g;
	Parse[i].B = b;
	return( ++ParseNow );
}

/**********************************************************************/
/*                                                                    */
/*            function search_256color()                              */
/*                                                                    */
/*            args  unsigned int col  : the pixel                     */
/*                                                                    */
/*            return  unsigned char : the index of the pallet         */
/*                                                                    */
/*            function: Get the index of the color pallet from the    */
/*                     color at the pixel.                            */
/**********************************************************************/
unsigned char search_256color( unsigned int col )
{
	int i;
	unsigned char r, g, b;

	r = GET_RED(col);
	g = GET_GREEN(col);
	b = GET_BLUE(col);
	for( i = 0; i < ParseNow; i++ ){
		if( r != Parse[i].R ){
			continue;
		}
		if( g != Parse[i].G ){
			continue;
		}
		if( b != Parse[i].B ){
			continue;
		}
		return( i );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function get_256pallet()                                */
/*                                                                    */
/*            args  struct Colors *col: ri->pallet, the pallet        */
/*                  int num  : used pallet color num                  */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: Get the registrated color pallet.             */
/*                                                                    */
/**********************************************************************/
void get_256pallet( struct Color *col, int num )
{
	int i;

	for( i = 0; i < num; i++ ){
		col[i].R = Parse[i].R;
		col[i].G = Parse[i].G;
		col[i].B = Parse[i].B;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function check_colornum()                               */
/*                                                                    */
/*            args  struct Shm *buf   : Image data                    */
/*                  int max  : maximum of the pallet colors           */
/*                                                                    */
/*            return  int: resistrated color num                      */
/*                                                                    */
/*            function: Chekc the total color num, and registrate the */
/*                     colors.                                        */
/**********************************************************************/
int check_colornum( struct Shm *buf, int max )
{
	int i;
	unsigned int *p;
	int ret;

	if( clear_parse( max ) == -1 ){
		return( -1 );
	}

	p = buf->data; ret = -1;
	for( i = 0; i < buf->width * buf->height; i++, p++ ){
/* D2: Smash arena, FIXed */
		if( (ret = parse_256color( *p )) >= max ){
/* D2 del	if( (ret = parse_256color( *p )) > max ){ */
			return( ret );
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function get_256color()                                 */
/*                                                                    */
/*            args  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                  struct Shm *buf   : Image data                    */
/*                  int max  : registrated color num                  */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: ONLY case ri->target < (total color num).     */
/*                     Not convert the colors, only get pallets.      */
/**********************************************************************/
void get_256color( struct ReduceInfo *ri, struct Shm *buf, int max )
{
	int i;
	unsigned int *p;
	unsigned char *q;

	if( ri->mode != GETPALLETCOLOR ){
		return;
	}
	q = ri->image;
	p = buf->data;
	for( i = 0; i < buf->width * buf->height; i++, p++ ){
		*q++ = search_256color( *p );
	}

/* get pallet */
	get_256pallet( ri->pallet, max );
}
/* added D1 */

/**********************************************************************/
/*                                                                    */
/*            function get_reduce_image2()                            */
/*                                                                    */
/*            args  struct Shm *buf   : Image data                    */
/*                  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                                                                    */
/*            return int   : 0==success  -1==error  -2==reduce failed */
/*                                                                    */
/*            function: Do the reduce/pallet process, main routine.   */
/*                                                                    */
/**********************************************************************/
/* D2: wrap get_reduce_image(), and rename .._image() to .._image2()  */
int get_reduce_image2( struct Shm *buf, struct ReduceInfo *ri )
/* D2 int get_reduce_image( struct Shm *buf, struct ReduceInfo *ri )*/
{
	int width, height;
	unsigned int *p;
	int i;
	int ret;
	struct Shm *shadow;
	int size;

	width = buf->width;
	height = buf->height;

	init_scale( ri );
	if( ri->mode == GETPALLETCOLOR ){
		size = sizeof(unsigned int) * (width * height + 4);
		if( (shadow = malloc( size )) == NULL ){
			return( EMALLOC );
		}
		memcpy( shadow, buf, size );
		p = shadow->data;
		if( ri->submode == (BW | INDEPENDENT | FLAT) ){
			gray_scale( shadow, ri );
/* add C1 for improve bilevel logic */
		} else if( ri->submode == (BW | TOTALNUM | ZAGGY)
			  && ri->target == 2 && ri->tmp_colors ==2 ){
			make_bilevel( shadow, ri );
/* add D1, under 256 colors, not reduce colors */
		} else if( (ret = check_colornum( buf, ri->target )) < ri->target ){
			if( ret == -1 ){
				free( shadow );
				return( -1 );
			}
			printf( "under 256 colors, not reduce colors(%d)\n", ret );
			get_256color( ri, shadow, ret );
/* add end */
		} else {
			printf( "over 256 colors, do reduce colors(%d)\n", ret );
			for( i = 0; i < height * width; i++, p++ ){
				*p = change_proc( ri, *p );
			}
			zaggy_mode( shadow, ri );
			if( ri->submode & TOTALNUM ){
				if( (ret = pallet_colors( ri, shadow )) != 0 ){
					free( shadow );
					return( ret );
				}
			}
		}
		free( shadow );
	} else {
		p = buf->data;
		for( i = 0; i < height * width; i++, p++ ){
			if( MASKOFF(*p) ){
				*p = change_proc( ri, *p );
			}
		}
		if( (ri->submode & ZAGGY) || (ri->submode & TOTALNUM) ){
			zaggy_mode( buf, ri );
		}
		if( ri->submode & TOTALNUM ){
			if( (ret = pallet_colors( ri, buf )) != 0 ){
				return( ret );
			}
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function get_reduce_image()                             */
/*                                                                    */
/*            args  struct Shm *buf   : Image data                    */
/*                  struct ReduceInfo *ri : Reduce Infomation struct  */
/*                                                                    */
/*            return int   : 0==success  else==fail                   */
/*                                                                    */
/*            function: Do the reduce/pallet process, main routine.   */
/*            D2: create new, to repeat reduce process.               */
/*                                                                    */
/**********************************************************************/
int get_reduce_image( struct Shm *buf, struct ReduceInfo *ri )
{
     int org, ret;

     org = ri->tmp_colors;
     if( (ret = get_reduce_image2( buf, ri )) == -2 ){
	  ri->tmp_colors--;
	  while( ri->tmp_colors >= org / 2 ){  
	         fprintf( stderr, "retry %d division\n", ri->tmp_colors );
		 ret = get_reduce_image2( buf, ri );
		 if( ret != -2 ){
		      return ret;
		 }
		 ri->tmp_colors--;
	  }
	  fprintf( stderr, "cannot more retry... give up!\n" );
	  ret = -1;
     }
     return ret;
}
