/**********************************************************************/
/*                                                                    */
/*           rotate.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/* D1 : rotate 90,180,270,0 is a special case!                        */
/* D2 : show filename now proc.                                       */
/* D2 : get rad= option as INT to FLOAT.                              */
/* D2 : created newly to cut-off real rotated area.                   */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>

#include "plug.h"

/*             This is a virtual plug, not affect the Shm data.       */

float Rad;
char Name[256] = { "noname" };
char NewName[256] = { "rotate" };

#define ROTATE  0
#define MIRROR  1

#ifndef PI
#ifdef M_PI
#define PI  M_PI
#endif
#endif

/* add D2 */
int Comp;
/* end D2 */

int Mode = ROTATE;
int Gui = NOGUI;
int List = 0;
void search_file(FILE *, int, int, char *);

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =    { "rotate" };
char *Comment = { "rotate/mirror efffect of cut-file(*.cut)" };

PLUGITEM Item[] = {
{ "rad", 	GETFLOAT,  &Rad,	 0, 360,   "degree of rotation" },
{ "new", 	GETSTR,  &NewName, 64, 0, "save cut file name" },
{ "mode=mirror", SWITCH, &Mode, MIRROR, 'M', "mirror mode"  },
{ "mode=rotate", SWITCH, &Mode, ROTATE, 'R', "rotate mode" },
/* D2 add */
{ "comp=yes",   SWITCH,  &Comp, 1, 'Y', "comprement between" },
{ "comp=no",    SWITCH,  &Comp, 0, 'N', "NO comprement between" },
/* end D2 */
{ "list",	SWITCH,  &List, 1, 0,      "list all cutfile in #lwp" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" },
{ "name", 	GETSTR,  &Name, 64, 0,    "target cut file name" },
{ NULL, 	GETSTR,  &Name, 64, 0,    "target cut file name" } };
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "degree of rotation ", "rad=%g" },
{ NOTCL | ENTRY, 9, "object cut file name", "name=%s" },
{ TCLLIST, 9, "object cut file name", NULL },
{ TCLDIR, 0, (char *)search_file, NULL },
{ TCLDIR, 1, (char *)search_file, NULL },
{ TCLLISTEND, 9, "input file name", "name=%s" },
{ NOTCL | ENTRY, 1, "create cut file name", "new=%s" },
{ TCLLIST, 1, "create cut file name", "new=%s" },
{ TCLDIR, 1, (char *)search_file, NULL },
{ TCLLISTENT, 1, "create file is", "new=%s" },
{ RADIO, 2, "Mode " },
{ ITEM, 2, "mirror(m)", "mode=mirror" },
{ ITEM, 3, "rotate(r)", "mode=rotate" },
{ RADIOEND, -1 },
{ RADIO, 4, "Comprement " },
{ ITEM, 4, "yes(y)", "comp=yes" },
{ ITEM, 5, "no(n)", "comp=no" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

#define WORD unsigned short
struct RawHeader {
	WORD spec;	/* LSB or MSB resolve flag */
	WORD width;	/* width */
	WORD height;	/* height */
	WORD mask;	/* mask kind */
};

/* D2: get maximun/minimun of cut area */
struct pos4 {
     int minx;
     int maxx;
     int miny;
     int maxy;
};

/**********************************************************************/
/*                                                                    */
/*             function rev()                                         */
/*                                                                    */
/*             WORD v        : reverse short                          */
/*                                                                    */
/*             return  WORD  : reversed short                         */
/*                                                                    */
/*             function:  Reverse the byte order at 2byte WORD.       */
/*                                                                    */
/**********************************************************************/
WORD rev( WORD v )
{
	WORD ret;

	ret = ((v & 0xff00) >> 8);
	ret |= ((v & 0x00ff) << 8);
	return( ret );
}

unsigned int get_comp( unsigned int *s, double x, double y, int w, int h )
{
     double r, g, b;
     int x0, y0;
     double px0, py0;
     int at;
     int flag;
     unsigned int ir, ig, ib, im;

     x0 = (int)x; y0 = (int)y;
     at = y0 * w + x0;
     px0 = x - (double)x0; py0 = y - (double)y0;
     flag = 1;
     if( x - 1 < w ){
	  r = (double)GET_RED(s[at]) * px0 + (double)GET_RED(s[at+1]) * (1.0 - px0);
	  g = (double)GET_GREEN(s[at]) * px0 + (double)GET_GREEN(s[at+1]) * (1.0 - px0);
	  b = (double)GET_BLUE(s[at]) * px0 + (double)GET_BLUE(s[at+1]) * (1.0 - px0);
     } else {
	  r = (double)GET_RED(s[at]);
	  g = (double)GET_GREEN(s[at]);
	  b = (double)GET_BLUE(s[at]);
	  flag = 0;
     }
     at += w;
     if( y - 1 < h ){
	  r = r * py0 + (double)GET_RED(s[at]) * (1.0 - py0);
	  g = g * py0 + (double)GET_GREEN(s[at]) * (1.0 - py0);
	  b = b * py0 + (double)GET_BLUE(s[at]) * (1.0 - py0);
     } else {
	  flag = 0;
     }
     at += 1;
     if( flag ){
	  px0 = px0 * py0;
	  r = r * px0 + (double)GET_RED(s[at]) * (1.0 - px0);
	  g = g * px0 + (double)GET_GREEN(s[at]) * (1.0 - px0);
	  b = b * px0 + (double)GET_BLUE(s[at]) * (1.0 - px0);
     }
     at = at - w - 1;
     im = GET_MASK(s[at]);
     ir = (unsigned int)r; ig = (unsigned int)g; ib = (unsigned int)b;
     if( ir < 0 ) ir = 0;
     if( ig < 0 ) ig = 0;
     if( ib < 0 ) ib = 0;
     if( ir > 255 ) ir = 255;
     if( ig > 255 ) ig = 255;
     if( ib > 255 ) ib = 255;
     return ((im << 24) | (ir << 16) | (ig << 8) | ib);
}

/**********************************************************************/
/*                                                                    */
/*             function rotate_proc()                                 */
/*                                                                    */
/*             unsigned int *src   : original cut data                */
/*             unsigned int *dest  : rotated cut data                 */
/*             int fwidth          : original width                   */
/*             int fheight         : original height                  */
/*             int dwidth          : maximum width(&height) of rotate */
/*                                   result (original sqrt(h*h+w*w))  */
/*                                                                    */
/*             return  int : 0==success 0!=errorcode                  */
/*                                                                    */
/*             function: rotated cut file data.                       */
/*                                                                    */
/* D2 : get rad= option as INT to FLOAT.                              */
/* D2 : created newly to cut-off real rotated area.                   */
/**********************************************************************/
int rotate_proc2( unsigned int *src, unsigned int *dest, int fwidth, int fheight, int dwidth )
{
     unsigned int *p;
     double vcos, vsin;
/* D2: comprement Mode */
     double di, dj;
/* end D2 */
     int i, j, x, y;
     double ci, cj, cx, cy;
     double dx, dy;

/* D1: add 90, 180, 270, 0 is a special case */
/* D2 : get rad= option as INT to FLOAT. */
     if( fmod( Rad, 360 ) == 0.0 ){
	  vcos = 1.0;
	  vsin = 0.0;
     } else if( fmod( Rad, 360 ) == 90.0 ){
	  vcos = 0.0;
	  vsin = 1.0;
     } else if( fmod( Rad, 360 ) == 180.0 ){
	  vcos = -1.0;
	  vsin = 0.0;
     } else if( fmod( Rad, 360 ) == 270.0 ){
	  vcos = 0.0;
	  vsin = -1.0;
     } else {
/* D1 end add 90, 180, 270, 0 ia a special case */
	  vcos = cos( (double)Rad * PI / 180.0 );
	  vsin = sin( (double)Rad * PI / 180.0 );
     }
     ci = (double)fheight / 2.0; cj = (double)fwidth / 2.0;
     cx = (double)dwidth / 2.0, cy = (double)dwidth / 2.0;
     p = dest;
     for( y = 0; y < dwidth; y++ ){
	  for( x = 0; x < dwidth; x++, p++ ){
	       dx = (double)x; dy = (double)y;
	       if( Mode == ROTATE ){
		    dj = (dx-cx) * vcos - (dy-cy) * vsin + cj;
		    di = (dx-cx) * vsin + (dy-cy) * vcos + ci;
	       } else if( Mode == MIRROR ){
		    dj = (dx-cx) * vcos * vcos + (dy-cy) * vsin * vcos;
		    di = (dx-cx) * vsin * vcos + (dy-cy) * vsin * vsin;
		    dj = dj + (dj - (x-cx)) + cj;
		    di = di + (di - (y-cy)) + ci;
	       } else {
		    return( -1 );
	       }
	       i = (int)di; j = (int)dj;
	       if( j < 0 || j >=fwidth || i < 0 || i >= fheight ){
		    *p = (MASK | 0);
	       } else {
		    if( Comp == 1 ){
			 *p = get_comp( src, dj, di, fwidth, fheight );
		    } else {
			 *p = src[i*fwidth+j];
		    }
	       }
	  }
     }
     return 0;
}

/**********************************************************************/
/*                                                                    */
/*             function get_minmax()                                  */
/*                                                                    */
/*             struct pos4 *pos    : <OUT> set min/max of real data   */ 
/*             unsigned int *d     : rotate result data               */
/*             int w               : result data width                */
/*             int h               : result data height               */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: get min/max of result data at x and y.       */
/* D2: created newly to cut-off real rotated area.                    */
/**********************************************************************/
void get_minmax( struct pos4 *pos, unsigned int *d, int w, int h )
{
     int i, j;

     pos->minx = pos->miny = 999999;
     pos->maxx = pos->maxy = -1;
     for( i = 0; i < h; i++ ){
	  for( j = 0; j < w; j++, d++ ){
	       if( MASKOFF(*d) ){
		    if( pos->minx > j ){
			 pos->minx = j;
		    }
		    if( pos->maxx < j ){
			 pos->maxx = j;
		    }
		    if( pos->miny > i ){
			 pos->miny = i;
		    }
		    if( pos->maxy < i ){
			 pos->maxy = i;
		    }
	       }
	  }
     }
}

/**********************************************************************/
/*                                                                    */
/*             function write_cutfile()                               */
/*                                                                    */
/*             int mask            : original mask kind               */
/*             unsigned int *dest  : result cut data                  */
/*             struct pos4 *pos    : <OUT> set min/max of real data   */ 
/*                                   calculated at get_minmax().      */
/*             int dwidth          : maximum width(&height) of rotate */
/*                                   result (original sqrt(h*h+w*w))  */
/*                                                                    */
/*             return  int : 0==success 0!=errorcode                  */
/*                                                                    */
/*             function: write the rotated cut file                   */
/* D2: created newly to cut-off real rotated area.                    */
/**********************************************************************/
int write_cutfile( int mask, unsigned int *dest, struct pos4 *pos, int dwidth )
{
	int fd;
	struct RawHeader header;
	char filename[256];
	char *envstr;
	int i, len;

	if( (envstr = getenv( "GEDO_LWP" )) == NULL ){
		strcpy( filename, "./" );
	} else {
		strcpy( filename, envstr );
	}
	if( filename[strlen(filename) -1] != '/' ){
		strcat( filename, "/" );
	}
	strcat( filename, NewName );
	strcat( filename, ".cut" );
	if( (fd = open( filename, O_RDWR | O_TRUNC )) == -1 ){
		if( (fd = open( filename, O_RDWR | O_CREAT, 0664)) == -1 ){
			free( dest );
			return( EOPENFILE );
		}
	}
	len = (pos->maxx - pos->minx + 1) * sizeof(int);
	header.spec = 1;
	header.width = pos->maxx - pos->minx + 1;
	header.height = pos->maxy - pos->miny + 1;
	header.mask = mask;
	if( write( fd, &header, sizeof(struct RawHeader) ) != sizeof(struct RawHeader) ){
		return( EREADFILE );
	}
	dest += dwidth * pos->miny;
	for( i = pos->miny; i <= pos->maxy; i++ ){
	     if( write( fd, &dest[pos->minx], len  ) != len ){
		  return( EREADFILE );
	     }
	     dest += dwidth;
	}
	close( fd );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function rotate_proc()                                 */
/*                                                                    */
/*             unsigned int *buff  : original cut data                */
/*             int fwidth          : original width                   */
/*             int fheight         : original height                  */
/*             int mask            : original mask kind               */
/*                                                                    */
/*             return  int : 0==success 0!=errorcode                  */
/*                                                                    */
/*             function: write the rotated cut file                   */
/* D2: re-write to cut-off real rotated area.                         */
/**********************************************************************/
int rotate_proc( unsigned int *buff, int fwidth, int fheight, int mask )
{
     int dwidth;
     int len, ret;
     unsigned int *buff2;
     struct pos4 pos;

     dwidth = (int)sqrt( (double)(fwidth*fwidth + fheight*fheight) );
     len = dwidth * dwidth * sizeof(unsigned int);
     if( (buff2 = (unsigned int *)malloc( len )) == NULL ){
	  return( EMALLOC );
     }
     memset( buff2, 0, len );
     if( (ret = rotate_proc2( buff, buff2, fwidth, fheight, dwidth )) != 0 ){
	  free( buff2 );
	  return ret;
     }
     get_minmax( &pos, buff2, dwidth, dwidth );
     ret = write_cutfile( mask, buff2, &pos, dwidth );
     free( buff2 );
     return ret;
}

/**********************************************************************/
/*                                                                    */
/*             function proc()                                        */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : 0==success 0!=errorcode                  */
/*                                                                    */
/*             function: Read cut file and rotate cut.                */
/*                                                                    */
/**********************************************************************/
int proc( void )
{
	int fd;
	struct RawHeader header;
	int len;
	int ret;
	int fwidth, fheight;
	unsigned int *buff;
	char filename[256];
	char *envstr;

	if( (envstr = getenv( "GEDO_LWP" )) == NULL ){
		strcpy( filename, "./" );
	} else {
		strcpy( filename, envstr );
	}
	if( filename[strlen(filename) -1] != '/' ){
		strcat( filename, "/" );
	}
	strcat( filename, Name );
	strcat( filename, ".cut" );

	if( (fd = open( filename, O_RDONLY )) == -1 ){
		return( EOPENFILE );
	}
	if( read( fd, &header, sizeof(struct RawHeader) ) != sizeof(struct RawHeader) ){
		return( EREADFILE );
	}
	if( header.spec == 1 ){
		fwidth = (int)header.width;
		fheight = (int)header.height;
	} else if( header.spec == 0x0100 ){
		fwidth = (int)rev( header.width );
		fheight = (int)rev( header.height );
	} else {
		return( EILLFORM );
	}
	len = fwidth * fheight * sizeof(unsigned int);
	if( (buff = (unsigned int *)malloc( len )) == NULL ){
		return( EMALLOC );
	}
	if( read( fd, buff, len ) != len ){
		return( EREADFILE );
	}
	close( fd );

	ret = rotate_proc( buff, fwidth, fheight, header.mask );
	free( buff );
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*             This is a virtual plug, not affect the Shm data.       */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	int ret;

/* D2: add show filename to proc */
	fprintf( stderr, "now rotate %s.cut to %s.cut\n", Name, NewName );
/* D2 end */

	ret= proc();
	if( ret != 0 ){
		strcpy( file->mes->u.data, strerror( errno ) );
		file->mes->datasize = strlen( file->mes->u.data );
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;
	int ret;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		ret = line_config( file );
		if( ret == 0 && List ){
			search_dir( getenv( "GEDO_LWP" ), "cut" );
			List = 0;
		}
		return( ret );
	}
}

void search_file(FILE *fp, int var, int kind, char * arg)
{
	switch( kind ){
	case 0:
		fprintf( fp, ".lstb%d insert end \"noname\"\n", var );
		break;
	case 1:
		search_tcldir( fp, var, getenv( "GEDO_LWP" ), "cut" );
		break;
	default:
		break;
	}
}
