/**********************************************************************/
/*                                                                    */
/*          sharpen.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "plug.h"

int Allow = 20;
int Value = 5;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =	{ "sharp" };
char *Comment = { "sharpen the image" };

PLUGITEM Item[] = {
{ "allow", 	GETINT,  &Allow, 1, 65535, "edge detection threshold" },
{ "value", 	GETINT,  &Value, 1, 64,    "add value to emphasis edge" },
{ "gui",        SWITCH, &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH, &Gui,  NOGUI, 'N', "not use GUI interface" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "threthold to detect edge", "allow=%d" },
{ ENTRY, 1, "emphasis value to the edge", "value=%d" }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

#define abs(p)  (p>0)?(p):((p)*(-1))

/**********************************************************************/
/*                                                                    */
/*             function check_edge()                                  */
/*                                                                    */
/*             unsigned int *s  : the pixel                           */
/*             unsigned int *d  : compare pixel(neighbor)             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: If <s> is the edge agaist <d>, sharpen       */
/*                      ,add <s> and sub <d> by Value.                */
/**********************************************************************/
void check_edge( unsigned int *s, unsigned int *d )
{
	unsigned int srgb[3];
	unsigned int drgb[3];
	int i;
	int tmp;
	int ts, td;
	int flag;

	srgb[0] = GET_RED(*s);
	srgb[1] = GET_GREEN(*s);
	srgb[2] = GET_BLUE(*s);
	drgb[0] = GET_RED(*d);
	drgb[1] = GET_GREEN(*d);
	drgb[2] = GET_BLUE(*d);

	flag = 0;
	for( i = 0; i < 3; i++ ){
		ts = (int)srgb[i]; td = (int)drgb[i];
		tmp = ts - td;
		if( tmp > Allow ){
			ts += Value;
			td -= Value;
			flag = 1;
		} else if( tmp < Allow * (-1) ){
			ts -= Value;
			td += Value;
			flag = 1;
		}
		if( ts > 0xff ){
			srgb[i] = 0xff;
		} else if( ts < 0 ){
			srgb[i] = 0;
		} else {
			srgb[i] = (unsigned int)ts;
		}
		if( td > 0xff ){
			drgb[i] = 0xff;
		} else if( td < 0 ){
			drgb[i] = 0;
		} else {
			drgb[i] = (unsigned int)td;
		}
	}
	if( flag ){
		if( MASKOFF(*s) ){
			*s = ((*s & MASK) | ((srgb[0] << 16) & RED) | 
			      ((srgb[1] << 8) & GREEN) | (srgb[2] & BLUE));
		}
		if( MASKOFF(*d) ){
			*d = ((*d & MASK) | ((drgb[0] << 16) & RED) | 
			      ((drgb[1] << 8) & GREEN) | (drgb[2] & BLUE));
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	unsigned int *data, *p;
	int width, height;
	int i, j;

	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}
	for( i = 1; i < height - 1; i++ ){
		for( j = 1; j < width - 1; j++ ){
			p = &data[i * width + j];
			check_edge( p, p+width+1 );
			check_edge( p, p+width );
			check_edge( p, p+width-1 );
			check_edge( p, p+1 );
			check_edge( p, p-1 );
			check_edge( p, p-width+1 );
			check_edge( p, p-width );
			check_edge( p, p-width-1 );
		}
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
