/**********************************************************************/
/*                                                                    */
/*           tcolor.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "plug.h"


#define AVERAGE	0
#define MAXIMUM	1
#define MINIMUM	2

int Mode = AVERAGE;

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "tcolor" };
char *Comment = { "report colors in the area, select, after config get $config" };

PLUGITEM Item[] = {
{ "average", SWITCH,  &Mode, AVERAGE, 'A',   "report average RGB in the area" },
{ "maximum", SWITCH,  &Mode, MAXIMUM, 'X',   "report maximum RGB in the area" },
{ "minimum", SWITCH,  &Mode, MINIMUM, 'N',   "report minimum RGB in the area" },
{ "gui",        SWITCH, &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH, &Gui,  NOGUI, 'N', "not use GUI interface" }
 };
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ RADIO, 0, "Query Mode" },
{ ITEM, 0, "average(a)", "average" },
{ ITEM, 1, "maximum(x)", "maximum" },
{ ITEM, 2, "minimum(n)", "minimum" },
{ RADIOEND, -1 }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

unsigned int Maximum;	/* maximum value */
unsigned int Minimum;	/* minimum value */
unsigned int Average;	/* average value */
int Mask_kind;		/* kind of mask, if DEPTHMASK >= 0x80 check.*/

/**********************************************************************/
/*                                                                    */
/*             function check_mask()                                  */
/*                                                                    */
/*             unsigned int v : color pixel                           */
/*                                                                    */
/*             return  int    : 1==must get sum  0==ignore            */
/*                                                                    */
/*             function: Check each mask and determin to accumurate.  */
/*                                                                    */
/**********************************************************************/
int check_mask( unsigned int v )
{
	if( Mask_kind == SIMPLEMASK ){
		return( MASKOFF(v) );
	} else if( Mask_kind == DEPTHMASK ){
		if( v >= 0x80000000 ){
			return( 1 );
		} else {
			return( 0 );
		}
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	int width, height;
	unsigned int *data, *p;
	int i, j;
	int minx, maxx, miny, maxy;
	unsigned long sumr, sumg, sumb;
	unsigned char maxr, maxg, maxb;
	unsigned char minr, ming, minb;
	int count;
	unsigned char r, g, b;

	sb = &file->mes->u.sb;
	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;
	Mask_kind = file->buf->mask_kind;

	if( sb->selnum != SKBOX ){
		return( -1 );
	}
	check_range( sb, width, height, 2, ADJUST );
	if( sb->defnum < 2 ){
		return( -1 );
	}

	minx = min( sb->def[0].x, sb->def[1].x );
	miny = min( sb->def[0].y, sb->def[1].y );
	maxx = max( sb->def[0].x, sb->def[1].x );
	maxy = max( sb->def[0].y, sb->def[1].y );

	p = &data[miny * width];
	sumr = sumg = sumb = count = 0;
	maxr = maxg = maxb = 0;
	minr = ming = minb = 0xff;
	
	for( i = miny; i <= maxy; i++ ){
		for( j = minx; j <= maxx; j++ ){
			if( check_mask( p[j] ) ){
				r = (unsigned char)GET_RED(p[j]);
				g = (unsigned char)GET_GREEN(p[j]);
				b = (unsigned char)GET_BLUE(p[j]);
				sumr += (long)r;
				sumg += (long)g;
				sumb += (long)b;
				if( maxr < r ){
					maxr = r;
				}
				if( maxg < g ){
					maxg = g;
				}
				if( maxb < b ){
					maxb = b;
				}
				if( minr > r ){
					minr = r;
				}
				if( ming > g ){
					ming = g;
				}
				if( minb > b ){
					minb = b;
				}
				count++;
			}
		}
		p += width;
	}
	if( count == 0 ){
		return( -1 );
	}
	sumr /= count; sumg /= count; sumb /= count;
	printf( "points=%d avarage R:G:B= %02x:%02x:%02x\n", count, (unsigned char)sumr, (unsigned char)sumg, (unsigned char)sumb );
	Average = ((sumr << 16) | (sumg << 8 ) | sumb);
	printf( "          max= %02x:%02x:%02x  min= %02x:%02x:%02x\n", maxr, maxg, maxb, minr, ming, minb );
	Maximum = ((maxr << 16) | (maxg << 8 ) | maxb);
	Minimum = ((minr << 16) | (ming << 8 ) | minb);
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;
	int ret;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			ret = config_use_gui( file );
		} else {
			ret = config_comline( file );
		}
	} else {
		ret = line_config( file );
	}
	if( ret == 0 ){
		if( Mode == AVERAGE ){
			sprintf( str, "%06x", Average );
		} else if( Mode == MAXIMUM ){
			sprintf( str, "%06x", Maximum );
		} else if( Mode == MINIMUM ){
			sprintf( str, "%06x", Minimum );
		} else {
			return( -1 );
		}
	}
	return( ret );
}
