/**********************************************************************/
/*                                                                    */
/*           text.c     gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include "plug.h"
#include "xstring.h"

#define NOGUI	0
#define USEGUI  1
int Gui = NOGUI;

static unsigned char Text[TEXTMAX] = { "test string" };
static char Fontpattern[FONTMAX] = { "*" };
static char Fontname[FONTMAX] = { "fixed" };
static char FileText[TEXTMAX] = { "file.txt" };
static int  InFileText = 0;

static int KMode = JIS;
static char *SKMode[3] = { "jis", "euc", "sjis" };

static int AddW = 0;
static int AddH = 0;
static int MaxW = -1;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "text" };
char *Comment = { "string selecter, mask of font form" };

PLUGITEM Item[] = {
{ "gui",     SWITCH,  &Gui,    USEGUI, 'G',   "exec GUI interface" },   
{ "nogui",   SWITCH,  &Gui,    NOGUI,  'N',   "not exec GUI interface" },   
{ "text",    GETSTR,  &Text,     80,  0,   "display text" },
{ "filetext",GETSTR,  &FileText, 80,  0,   "display text in the file" },
{ "infiletext",SWITCH,  &InFileText, 1,  'F', "get display text from <filetext>" },
{ "linetext",SWITCH,  &InFileText, 0,  'T',   "get display text from commandline" },
{ "pattern", GETSTR,  &Fontpattern, 80, 0, "font search pattern" },
{ "font",    GETSTR,  &Fontname, 80, 0,    "font fullname" },
{ "jis",     SWITCH,  &KMode,    JIS, 'J',    "include JIS japanese code" },
{ "euc",     SWITCH,  &KMode,    EUC, 'E',    "include EUC japanese code" },
{ "sjis",    SWITCH,  &KMode,   SJIS, 'S',    "include Shift JIS japanese code" },
{ "addW",    GETINT,  &AddW, -100,   100,   "add width between characters" },
{ "addH",    GETINT,  &AddH, -100,   100,   "add height between line" },
{ "MaxW",    GETINT,  &MaxW, -1,   1000,    "maximum width of the area" }
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);
struct Tags Tag[] = {
{ RADIO, 0, "Config Mode" },
{ ITEM,  0, "GUI(G)", "gui" },
{ ITEM,  1, "NOT gui(N)", "nogui" },
{ RADIOEND, -1 },
{ RADIO, 0, "Text is" },
{ ITEM, 4, "in FILE(F)", "infiletext" },
{ ITEM, 5, "this text(T)", "linetext" },
{ RADIOEND, -1 },
{ ENTRY, 3, "Text is in the file", "filetext=%s" },
{ ENTRY, 2, "Text is here", "text=%s" },
{ RADIO, 8, "KANJI is" },
{ ITEM,  8, "JIS(J)", "jis" },
{ ITEM,  9, "EUC(E)", "euc" },
{ ITEM, 10, "SJIS(S)", "sjis" },
{ RADIOEND, -1 },
{ ENTRY, 6, "Search font patten", "pattern=%s" },
{ ENTRY, 7, "font name is", "font=%s" },
{ ENTRY,11, "add Horizontal", "addW=%d" },
{ ENTRY,12, "add Vertical",   "addH=%d" },
{ ENTRY,13, "maximum width of area(-1==no limit)", "MaxW=%d" }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

/**********************************************************************/
/*                                                                    */
/*            function status()                                       */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return int : 0==success  else==errorcode                */
/*                                                                    */
/*            function: status command handler called from command.c  */
/*                                                                    */
/**********************************************************************/
void status( int flag, char *buff )
{
	static int prev_status = 0;

	if( InFileText ){
		Tag[9].kind |= INITONLY;
		get_default_status( flag, buff );
		Tag[9].kind &= ~INITONLY;
	} else {
		if( prev_status ){
			sprintf( buff, "text=%s", Text );
			prev_status = 0;
			return;
		}
		Tag[8].kind |= INITONLY;
		Tag[9].kind |= INITONLY;
		get_default_status( flag, buff );
		if( strlen(buff) + strlen(Text) + 6 >= 255 ){
			fprintf( stderr, "status is too long\n"
				"plese call more once\n" );
			prev_status = 1;
		} else {
			strcat( buff, ",text=" );
			strcat( buff, Text );
		}
		Tag[8].kind &= ~INITONLY;
		Tag[9].kind &= ~INITONLY;
	}
}
/********************** description end ********************/

static struct fontimage Image;	/* get font image header */
static char *Buff;		/* font image buffer */
static int NowAlloc = 10240;	/* Buff alloc size */

Display *D;		/* X11 Display */
Pixmap P;		/* Pixmap to write font image */
XFontStruct *Kanji;	/* use font for display KANJI */
XFontStruct *Ascii;	/* use font for display ASCII */

int exec_configure( struct File *file );

/**********************************************************************/
/*                                                                    */
/*            function first()                                        */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return int : 0==success  else==errorcode                */
/*                                                                    */
/*            function: Total initialize proc called select() and     */
/*                  config().                                         */
/**********************************************************************/
int first( int flag, struct File *file )
{
	if( Buff ){
		return( 0 );
	}	
	if( (Buff = malloc( NowAlloc )) == NULL ){
		return( EMALLOC );
	}
	D = XOpenDisplay( "" );
	if( D == NULL ){
	 	return( ENOAVAILX );
	}
	if( (Ascii = XLoadQueryFont( D, "8x16" )) == NULL ){
		XCloseDisplay( D );
		return( EILLSELECTION );
	}
	if( (Kanji = XLoadQueryFont( D, "kanji18" )) == NULL ){
		XFreeFont( D, Ascii );
		XCloseDisplay( D );
		return( EILLSELECTION );
	}
	if( flag ){
		return( exec_configure( file ) );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	int width, height;
	unsigned int *data, *p;
	char *q;
	int i, j;
	int minx, maxx, miny, maxy;

	sb = &file->mes->u.sb;
	width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	first( 1, file );
	if( sb->selnum != SKBOX ){
		return( EILLSELECTION );
	}
	check_range( sb, width, height, 2, ADJUST );
	if( sb->defnum < 2 ){
		return( EILLNUMBER );
	}

	for( i = 0, p = data; i < height * width; i++, p++ ){
		SETMASK(*p);
	}

	minx = min( sb->def[0].x, sb->def[1].x );
	miny = min( sb->def[0].y, sb->def[1].y );
	maxx = max( sb->def[0].x, sb->def[1].x );
	maxy = max( sb->def[0].y, sb->def[1].y );
	if( minx + Image.width >= width ){
		maxx = width - minx;
	} else {
		maxx = Image.width;
	}
	if( miny + Image.height >= height ){
		maxy = height - miny;
	} else {
		maxy = Image.height;
	}

	for( i = 0; i < maxy; i++ ){
		p = data + (miny + i) * width;
		q = Buff + i * Image.width;
		for( j = 0; j < maxx; j++ ){
			if( q[j] == 0 ){
				CLRMASK(p[minx+j]);
			}
		}
	}
	file->buf->mask_kind = SIMPLEMASK;
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function do_xconfig()                                   */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return  int : 0==success  else==errorcode               */
/*                                                                    */
/*            function: Invoke the xstring for GUI selection and get  */
/*                   font image.                                      */
/**********************************************************************/
static int do_xconfig( int max_w )
{
	FILE *fpp;
	char buff[TEXTMAX*2];

	if( InFileText ){
		sprintf( buff, "xstring filetext=%s pattern=%s %s addH=%d addW=%d MaxW=%d",
			FileText, Fontpattern, SKMode[KMode], AddH, AddW, max_w );
	} else {
		sprintf( buff, "xstring text=\'%s\' pattern=%s %s addH=%d addW=%d MaxW=%d",
			Text, Fontpattern, SKMode[KMode], AddH, AddW, max_w );
	}
	if( (fpp = popen( buff, "r" )) == NULL ){
		perror( "pipe open to xstring" );
		return( EOPENFILE );
	}
	if( fread( &Image, sizeof(struct fontimage), 1, fpp ) != 1 ){
		pclose( fpp );
		return( EREADFILE );
	}
	if( Image.flag == EUSRBREAK ){
		pclose( fpp );
		return( EUSRBREAK );
	} else if( Image.flag != ENOERROR ){
		pclose( fpp );
		return( Image.flag );
	}
	strcpy( Fontname, Image.fontname );
	if( NowAlloc < Image.width * Image.height ){
		if( (Buff = realloc( Buff, Image.width * Image.height )) == NULL ){
			pclose( fpp );
			return( EMALLOC );
		}
		NowAlloc = Image.width * Image.height;
	}
	if( fread( Buff, sizeof(char), Image.width * Image.height, fpp ) != (size_t)(Image.width * Image.height) ){
		pclose( fpp );
		return( EREADFILE );
	}
	pclose( fpp );
	return( 0 );
}	

/**********************************************************************/
/*                                                                    */
/*            function write_fontimage()                              */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return int : 0==success  else==errorcode                */
/*                                                                    */
/*            function: Get font image from Pixmap by calling get_font*/
/*                     image().                                       */
/**********************************************************************/
int write_fontimage( void )
{
	unsigned char *tmp;
	int size;

	size = Image.width * Image.height;
	if( NowAlloc < size ){
		if( (tmp = realloc( Buff, size )) == NULL ){
			return( EMALLOC );
		}
		Buff = tmp;
		NowAlloc = size;
	}
	memset( Buff, 0, size );
	return( get_fontimage( &Image, Buff ) );
}

/**********************************************************************/
/*                                                                    */
/*            function make_window()                                  */
/*                                                                    */
/*            args  WSTRING *txt : display string, see xstring.h      */
/*                                                                    */
/*            return  int : 0==success  else==errorcode               */
/*                                                                    */
/*            function: At background, make X11 Pixmap and get the    */
/*               font image by 2byte string txt.                      */
/**********************************************************************/
static int make_window( WSTRING *txt, int max_w )
{
	XFontStruct *font;
	unsigned long white, black;
	GC gcf;
	int fheight;
	int ret;
	int line;

	ret = 0;
	if( (font = XLoadQueryFont( D, Fontname )) == NULL ){
		return( EILLSELECTION );
	}
	if( font->min_byte1 ){
		XUnloadFont( D, Kanji->fid );
		Kanji = font;
	} else {
		XUnloadFont( D, Ascii->fid );
		Ascii = font;
	}

	fheight = max(Kanji->ascent, Ascii->ascent);

	Image.width = calc_maxwidth( txt, &line, AddW, max_w );
	Image.height = (fheight + MERGIN + AddH) * (line+1) + MERGIN;

	white = WhitePixel( D, 0 );
	black = BlackPixel( D, 0 );
	P = XCreatePixmap( D, RootWindow( D, 0 ), 
		Image.width, Image.height, DefaultDepth(D,0) );

	gcf = XCreateGC( D, P, 0, 0 );
	XSetForeground( D, gcf, white );
	XFillRectangle( D, P, gcf, 0, 0, Image.width, Image.height );
	XSetForeground( D, gcf, black );
	Draw_Str( gcf, txt, fheight + MERGIN + AddH, AddW, max_w );
	XFlush(D);
	ret = write_fontimage();
	XFreeGC( D, gcf );
	XFreePixmap( D, P );
	return( ret );
}



/**********************************************************************/
/*                                                                    */
/*             function exec_configure()                              */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE using the GUI result.               */
/*                                                                    */
/**********************************************************************/
int exec_configure( struct File *file )
{
	WSTRING *out;
	WSTRING OutText[TEXTMAX*2];
	struct stat statbuf;
	int allocsize;
	unsigned char *in;
	int max_w;
	int ret;

	in = out = NULL;

	if( InFileText ){
		if( stat( FileText, &statbuf ) == -1 ){
		     fprintf( stderr, "cannot open %s\n", FileText );
		     return( EOPENFILE );
		}
		allocsize = (statbuf.st_size + 1) * 3;
		if( (in = realloc( in, allocsize )) == NULL ){
			return( EMALLOC );
		}
		out = in + statbuf.st_size + 1;
		memset( in, 0, allocsize );
		if( (ret = read_textfile( in, FileText, statbuf.st_size )) != 0 ){
			free( in );
			return( ret );
		}
	}
	if( MaxW == -1 ){
		if( file->buf == NULL ){
			max_w = 200;
		} else {
			if( file->buf->width == 0 ){
				max_w = 200;
			} else {
				max_w = file->buf->width;
			}
		}
	} else {
		max_w = MaxW;
	}
	if( Gui == USEGUI ){
		if( (ret = do_xconfig( max_w )) != 0 ){
			if( InFileText ){
				free( in );
			}
			return( ret );
		}
	} else {
		if( InFileText ){
			kanji_set( in, KMode, out );
			if( (ret = make_window( out, max_w )) != 0 ){
				free( in );
				return( ret );
			}
			free( in );
		} else {
			kanji_set( Text, KMode, OutText );
			if( (ret = make_window( OutText, max_w )) != 0 ){
				return( ret );
			}
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;
	int ret;

	first( 0, NULL );
	str = file->mes->u.data;

	if( str[0] ){
		ret = line_config( file );
	} else {
	  	if( Gui ){
			ret = config_use_gui( file );
		} else {
			ret = config_comline( file );
		}
	}
	if( ret != 0 ){
		return( ret );
	}
	if( (ret = exec_configure( file )) != 0 ){
		return( ret );
	}
	status( CONFIG_ELSE, str );
	return( 0 );
}

