/**********************************************************************/
/*                                                                    */
/*             tiff.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*             D0: tell() is not system call..                        */
/*             D0: write SAMPLESPERPIXEL Tag for full color image.    */
/*             D0: if StripNum is ONLY 1(small image), use Tag field. */
/*             D0: not malloc/free pallet memory.. make static array. */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "plug.h"
#include "tiff.h"
#include "reduce.h"
#include "lzw.h"

#define BYTE unsigned char
#define WORD unsigned short

#define NONE	0
#define LZW	1
int Compress  = NONE;		/* The file is compressed (read file) */
int WriteComp = NONE;		/* Compression at write file(NO PRESENT) */
static char *WriteCompStr[2]
 = { "none", "lzw" };

#define K_WHITE   0
#define K_BLACK	1
#define K_RGB	2
#define K_COLORMAP 3
#define K_GRAY   4

int WriteMode = K_RGB;
int Gui = NOGUI;

/********************** configure desciption ******************/
char *Spec =    { "tiff" };
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Comment = { "Aldus TIFF loader/saver" };

PLUGITEM Item[] = {
{ "fullcolor",  NOLOAD | SWITCH, &WriteMode, K_RGB, 'F', "fullcolor 24bit save"  },
{ "pallet256", 	NOLOAD | SWITCH, &WriteMode, K_COLORMAP, 'P', "256 color pallet save" },
{ "gray", 	NOLOAD | SWITCH, &WriteMode, K_GRAY, 'G', "grayscale save" },
{ "bilevel",    NOLOAD | SWITCH, &WriteMode, K_WHITE, 'B', "bilevel save" },
{ "gui",        NOLOAD | SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      NOLOAD | SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" }
};
int      ItemNum = sizeof(Item) / sizeof(PLUGITEM);

struct Tags Tag[] = {
{ NOLOAD | RADIO, 0, "Color Mode" },
{ NOLOAD | ITEM, 0, "FullColor(F)", "fullcolor" },
{ NOLOAD | ITEM, 1, "256-Color(P)", "pallet256" },
{ NOLOAD | ITEM, 2, "grayscale(G)", "gray" },
{ NOLOAD | ITEM, 3, "B/W(B)",    "bilevel" },
{ NOLOAD | RADIOEND, -1 }
};
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

#define get_y(r,g,b) (int)(0.299*(float)r+0.587*(float)g+0.114*(float)b) 

struct ImageOff *Ioff;		/* Image Offset(StripOffset) Array */
struct CMap Cmap[256];		/* ColorMap Array */

long OffNum;			/* count of StripOffset */
long LenNum;			/* count of StripByteCount */
int Mode = 0;			/* Current read file Mode, K_RGB or etc */
int StripPerRow = 0;		/* How many Rows in a Strip? (read file)*/

int File_LSB_flag = 0;		/* This File wrote LSB */
int LSB_flag;			/* System is LSB */

struct ReduceInfo RI;		/* call get_reduce_image() interface */

int RpS;			/* StripPerRow  (write file) */
int StripNum;			/* Number of Strip (write file) */
int StripSize;			/* Bytes of a Strip (write file) */

static int BitMask[8]		/* BitMask for BiLevel Mode */
 = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80}; 


/**********************************************************************/
/*                                                                    */
/*            function wconv()                                        */
/*                                                                    */
/*            args  unsigned short v : short value                    */
/*                                                                    */
/*            return unsigned short  : machine byte order value       */
/*                                                                    */
/*            function: Obeying file and machine byte order,          */
/*                     convert short value correctly.                 */
/**********************************************************************/
unsigned short wconv( unsigned short v )
{
	register unsigned short v1, v2;

	if( File_LSB_flag != LSB_flag ){
		v1 = (v & 0x00ff) << 8;
		v2 = (v & 0xff00) >> 8;
		return( v1 | v2 );
	} else {
		return( v );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function dconv()                                        */
/*                                                                    */
/*            args  unsigned short v : long  value                    */
/*                                                                    */
/*            return unsigned short  : machine byte order value       */
/*                                                                    */
/*            function: Obeying file and machine byte order,          */
/*                     convert long  value correctly.                 */
/*                                                                    */
/**********************************************************************/
unsigned long dconv( unsigned long v )
{
	register unsigned long v1, v2, v3, v4;

	if( File_LSB_flag != LSB_flag ){
		v1 = (v & 0x000000ff) << 24;
		v2 = (v & 0x0000ff00) << 8;
		v3 = (v & 0x00ff0000) >> 8;
		v4 = (v & 0xff000000) >> 24;
		return( v1 | v2 | v3 | v4 );
	} else {
		return( v );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function copy_data()                                    */
/*                                                                    */
/*            args  int at  : 0==StripOffsets 1==StripByteCounts      */
/*                  struct ImageOff *work : copy source               */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Copy Strip offset and count to the Ioff[].    */
/*                                                                    */
/**********************************************************************/
void copy_data( int at, struct ImageOff *work )
{
	struct ImageOff *io;
	short *ps;
	long  *pl;
	int i;

	if( TiffIFD.field == V_SHORT ){
		ps = (short *)work; io = Ioff;
		for( i = 0; i < TiffIFD.length; i++ ){
			io->d[at] = (long)wconv(*ps);
			io++; ps++;
		}
	} else if( TiffIFD.field == V_LONG ){
		pl = (long *)work; io = Ioff;
		for( i = 0; i < TiffIFD.length; i++ ){
			io->d[at] = dconv(*pl);
			io++; pl++;
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*            function set_colortable()                               */
/*                                                                    */
/*            args  unsigned short *org : original color table in file*/
/*                  struct CMap *cm  : internal colormap struct       */
/*                  int len   :  entry of color table                 */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: copy and convert ColorMap to internal struct. */
/*                                                                    */
/**********************************************************************/
void set_colortable( unsigned short *org, struct CMap *cm, int len )
{
	int i;

	for( i = 0; i < len; i++, org++, cm++ ){
		cm->R = (unsigned char)(wconv(*org) >> 8);
		cm->G = (unsigned char)(wconv(*(org + len)) >> 8);
		cm->B = (unsigned char)(wconv(*(org + len + len)) >> 8);
	}
}

/**********************************************************************/
/*                                                                    */
/*            function show_data()                                    */
/*                                                                    */
/*            args  struct File *file : data struct                   */
/*                  int fd   :   file descripter                      */
/*                                                                    */
/*            return int   :  0==success  else==errorcode             */
/*                                                                    */
/*            function:  Read each TAG and set data.                  */
/*                                                                    */
/**********************************************************************/
/* D0 tell() is not SYSTEM CALL */
long tell( int fd )
{
     return( lseek( fd, 0, SEEK_CUR ) );
}
/* D0 end */

int show_data( struct File *file, int fd )
{
	long nowpos;
	size_t len;
	union Data *p;
	int flag;
	unsigned short *bf;
	static int inhibit = 0;
	struct Shm *buf;
/* D0	long tell(int fd);*/

	buf = file->buf;
	if( inhibit ){
		return( 0 );
	}
	switch( TiffIFD.tag ){
	case S_NEWSUBFILETYPE:
		if( TiffIFD.length == 1 && TiffIFD.field == V_LONG ){
			p = (union Data *)&TiffIFD.offset;
			if( dconv(p->l) == 0L ){
				inhibit = 0;
			} else {
				inhibit = 1;
			}
		} else {
			return( -1 );
		}
		break;
	case S_SUBFILETYPE:
		if( TiffIFD.length == 1 && TiffIFD.field == V_SHORT ){
			p = (union Data *)&TiffIFD.offset;
			if( wconv(p->i[0]) == 1 ){
				inhibit = 0;
			} else {
				inhibit = 1;
			}
		} else {
			return( -1 );
		}
	case S_IMAGEWIDTH:
		if( TiffIFD.length == 1 ){
			p = (union Data *)&TiffIFD.offset;
			if( TiffIFD.field == V_SHORT ){
				buf->width = (int)wconv(p->i[0]);
			} else if( TiffIFD.field == V_LONG ){
				buf->width = (int)dconv(p->l);
			} else {
				return( -1 );
			}
		} else {
			return( -1 );
		}
		break;
	case S_IMAGELENGTH:
		if( TiffIFD.length == 1 ){
			p = (union Data *)&TiffIFD.offset;
			if( TiffIFD.field == V_SHORT ){
				buf->height = (int)wconv(p->i[0]);
			} else if( TiffIFD.field == V_LONG ){
				buf->height = (int)dconv(p->l);
			} else {
				return( -1 );
			}
		} else {
			return( -1 );
		}
		break;
	case S_BITSPERSAMPLE:
		if(TiffIFD.field == V_SHORT ){
			p = (union Data *)&TiffIFD.offset;
			if( TiffIFD.length == 3 ){
				Mode = K_RGB;
			} else if( TiffIFD.length == 1 ){
				p->i[0] = wconv(p->i[0]);
				if( p->i[0] == 1 ){
					Mode = K_BLACK;
				} else if( p->i[0] == 8 ){
					Mode = K_COLORMAP;
				} else {
					strcpy( file->mes->u.data, "BitsPerSample unknown data" );
					file->mes->datasize = strlen(file->mes->u.data);
					return( ENOTFEATURE );
				}
			} else {
				strcpy( file->mes->u.data, "BitsPerSample unknown data" );
				file->mes->datasize = strlen(file->mes->u.data);
				return( ENOTFEATURE );
			}
		} else {
			return( -1 );
		}
		break;
	case S_COMPRESSION:
		if( TiffIFD.length == 1 && TiffIFD.field == V_SHORT ){
			p = (union Data *)&TiffIFD.offset;
			if( wconv(p->i[0]) == 1 ){
				Compress = 0;
			} else if( wconv(p->i[0]) == 5 ){
				Compress = 1;
			} else {
				strcpy( file->mes->u.data, "not support compression" );
				file->mes->datasize = strlen(file->mes->u.data);
				return( ENOTFEATURE );
			}
		} else {
			return( -1 );
		}
		break;
	case S_PHOTOMETRICINTERPRETATION:
		if( TiffIFD.length == 1 && TiffIFD.field == V_SHORT ){
			p = (union Data *)&TiffIFD.offset;
			switch( wconv(p->i[0]) ){
			case K_RGB:
				if( Mode != K_RGB ){
					return( -1 );
				}
				break;
			case K_COLORMAP:
				if( Mode != K_COLORMAP ){
					return( -1 );
				}
				break;
			case K_BLACK:
				if( Mode == K_BLACK ){
				} else if( Mode == K_COLORMAP ){
					Mode = K_GRAY;
				} else {
					return( -1 );
				}
				break;
			default:
				strcpy( file->mes->u.data, "PhotometricInterpretaion unknow color class" );
				file->mes->datasize = strlen(file->mes->u.data);
				return( ENOTFEATURE );
				break;
			}
		} else {
			return( -1 );
		}
		break;
	case S_ROWSPERSTRIP:
		if( TiffIFD.length == 1 ){
			p = (union Data *)&TiffIFD.offset;
			if( TiffIFD.field == V_SHORT ){
				StripPerRow = (int)wconv(p->i[0]);
			} else if( TiffIFD.field == V_LONG ){
				StripPerRow = (int)dconv(p->l);
			} else {
				return( -1 );
			}
		} else {
			return( -1 );
		}
		break;
	case S_STRIPOFFSETS:
	case S_STRIPBYTECOUNTS:
		if( TiffIFD.tag == S_STRIPOFFSETS ){
			OffNum = TiffIFD.length;
			flag = 0;
		} else if( TiffIFD.tag == S_STRIPBYTECOUNTS ){
			LenNum = TiffIFD.length;
			flag = 1;
		} else {
			return( -1 );
		}
		if( Ioff == NULL ){
			len = (size_t)TiffIFD.length * sizeof(struct ImageOff) * 2; 
			Ioff = (struct ImageOff *)malloc( len );
			if( Ioff == NULL ){
				return( EMALLOC );
			}
		}
		if( (long)Kspace[TiffIFD.field] * TiffIFD.length > 4L ){
			nowpos = tell(fd);
			if( lseek( fd, dconv(TiffIFD.offset), 0 ) == -1L ){
				return( ESEEK );
			}
			if( read( fd, &Ioff[TiffIFD.length], Kspace[TiffIFD.field] * TiffIFD.length ) != Kspace[TiffIFD.field] * TiffIFD.length ){
				return( EREADFILE );
			}
			copy_data( flag, &Ioff[TiffIFD.length] );
			if( lseek( fd, nowpos, 0 ) == -1L ){
				return( ESEEK );
			}
		} else {
			copy_data( flag, (struct ImageOff *)&TiffIFD.offset );
		}
		break;
	case S_PLANARCONFIGURATION:
		if( TiffIFD.length == 1 && TiffIFD.field == V_SHORT ){
			p = (union Data *)&TiffIFD.offset;
			if( wconv(p->i[0]) != 1 ){
				strcpy( file->mes->u.data, "PlanarConfiguration unknown" );
				file->mes->datasize = strlen(file->mes->u.data);
				return( ENOTFEATURE );
			}
		} else {
			return( -1 );
		}
		break;
	case S_COLORMAP:
		len = (size_t)TiffIFD.length;
		bf = (unsigned short *)malloc( len * 2 );
		if( bf == NULL ){
			return( EMALLOC );
		}
		if( (long)Kspace[TiffIFD.field] * TiffIFD.length > 4L ){
			int ret;
			nowpos = tell(fd);
			if( lseek( fd, dconv(TiffIFD.offset), 0 ) == -1L ){
				free( bf );
				return( ESEEK );
			}
			if( (ret = read( fd, bf, len * 2 )) != len * 2 ){
				free( bf );
				return( EREADFILE );
			}
			set_colortable( bf, Cmap, TiffIFD.length / 3 );
			if( lseek( fd, nowpos, 0 ) == -1L ){
				free( bf );
				return( ESEEK );
			}
		} else {
			return( -1 );
		}
		free( bf );
		break;
	default:
		break;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function read_header()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ HEADER of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_header( struct File *file )
{
	int i, j;
	short num;
	long  next;
	int   st;

	if( (file->fd = open( file->name, O_RDONLY )) == -1 ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EOPENFILE );
	}

	if( read( file->fd, &TiffHeader, sizeof(TiffHeader) ) != sizeof(TiffHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EREADFILE );
	}
	if( TiffHeader.sig == 0x4949 ){
		File_LSB_flag = 1;
	} else if( TiffHeader.sig == 0x4d4d ){
		File_LSB_flag = 0;
	} else {
		file->mes->datasize = 0;
		return( EILLFORM );
	}
	TiffHeader.ver = wconv( TiffHeader.ver );
	TiffHeader.next = dconv( TiffHeader.next );
	if( TiffHeader.ver != 0x002a
	   || TiffHeader.next == 0L ){
		file->mes->datasize = 0;
		return( EILLFORM );
	}
	if( lseek( file->fd, TiffHeader.next, 0 ) == -1L ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( ESEEK );
	}
	for( i = 1; ;  ){
		if( read( file->fd, &num, sizeof(short) ) != sizeof(short) ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EREADFILE );
		}
		for( j = 0; j < wconv(num); j++ ){
			if( read( file->fd, &TiffIFD, sizeof(TiffIFD) ) != sizeof(TiffIFD) ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				return( EREADFILE );
			}
			TiffIFD.tag = wconv(TiffIFD.tag);
			TiffIFD.field = wconv(TiffIFD.field);
			TiffIFD.length = dconv(TiffIFD.length);
			st = show_data( file, file->fd );
			if( st == -1 ){
				file->mes->datasize = 0;
				return( EILLFORM );
			} else if( st == -2 ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				return( EREADFILE );
			} else if( st != 0 ){
				return( st );
			}
			i++;
		}
		if( read( file->fd, &next, sizeof(long) ) != sizeof(long) ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EREADFILE );
		}
		next = dconv(next);
		if( next == 0L ){
			return( 0 );
		}
		if( lseek( file->fd, next, 0 ) == -1L ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( ESEEK );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*            function set_shmbuff()                                  */
/*                                                                    */
/*            args  unsigned int *data  : image data(dest)            */
/*                  unsigned char *buff : image data(src)             */
/*                  int width           : width of image              */
/*                  unsigned char *tmp  : decode work buffer          */
/*                                        if need not, NULL           */
/*            return NONE                                             */
/*                                                                    */
/*            function: Actually copy data to the data buffer.        */
/*            D0: re-write                                            */
/*                                                                    */
/**********************************************************************/
void set_shmbuff( unsigned int *data, unsigned char *buff, int width, unsigned char *tmp )
{
	int i, j, k;
	int linepix;
	int ret;
	unsigned char *dt;

/* D0 change */
	if( Mode == K_RGB ){
		linepix = width * StripPerRow * 3;
	} else if( Mode == K_COLORMAP 
		   || Mode == K_GRAY ){
		linepix = width * StripPerRow;
	} else {
	        linepix = (width + 7) / 8 * StripPerRow;
	}
/* end D0 */
	if( Compress ){
		if( (ret = uncompress( 9, buff, tmp, linepix, TIFF )) == -1 ){
			fprintf( stderr, "uncompress error\n" );
			return;
		}
		dt = tmp;
	} else {
		dt = buff;
	}
	if( Mode == K_RGB ){
		for( i = j = 0; j < linepix; i++ ){
			data[i] = (((unsigned int)(dt[j++]) << 16) & 0x00ff0000 ); 
			data[i] |= (((unsigned int)(dt[j++]) << 8) & 0x0000ff00 );
			data[i] |= (((unsigned int)dt[j++]) & 0x000000ff);
		}
	} else if( Mode == K_COLORMAP ){
		for( i = 0; i < linepix; i++ ){
			data[i] = (((unsigned int)(Cmap[dt[i]].R) << 16) & 0x00ff0000 ); 
			data[i] |= (((unsigned int)(Cmap[dt[i]].G) << 8)& 0x0000ff00 );
			data[i] |= ((unsigned int)(Cmap[dt[i]].B) & 0x000000ff);
		}
	} else if( Mode == K_GRAY ){
		for( i = 0; i < linepix; i++ ){
			data[i] = (((unsigned int)dt[i] << 16) & 0x00ff0000 ); 
			data[i] |= (((unsigned int)dt[i] << 8) & 0x0000ff00 );
			data[i] |= ((unsigned int)dt[i] & 0x000000ff);
		}
	} else if( Mode == K_BLACK ){
/* D0 change */
	     for( i = 0; i < StripPerRow; i++ ){
		  for( j = 0, k = 7; j < width; j++, data++ ){
		       if( *dt & BitMask[k] ){
			    *data = 0x00ffffff;
		       } else {
			    *data = 0x00000000;
		       }
		       if( --k < 0 ){
			    k = 7; dt++;
		       }
		  }
		  if( k != 7 ){
		       dt++;
		  }
	     }
	}
/* D0 delete		row = 0;
		for( i = j = 0; i < linepix; j++ ){
			for( k = 7; k >= 0; k--, i++ ){
				if( dt[j] & BitMask[k] ){
				if( row++ >= width ){
					row = 0;
					break;
				}
			}
		}
	}*/
/* end D0 */
}

/**********************************************************************/
/*                                                                    */
/*             function read_image()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ IMAGES of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_image( struct File *file )
{
	unsigned char *buff;
	struct Shm *buf;
	unsigned int *p;
	int i;
	size_t size;
	struct ImageOff *io;
	unsigned char *tmp;
	int linepix;

	if( OffNum != LenNum ){
		return( -1 );
	}
	buf = file->buf;
	size = 0; io = Ioff;
	for( i = 0; i < LenNum; i++ ){
		if( size < io->d[1] ){
			size = io->d[1];
		}
		io++;
	}
	if( (buff = malloc( size )) == NULL ){
		file->mes->datasize = 0;
		close( file->fd );
		return( EMALLOC );
	}
	if( Compress ){
		if( Mode == K_RGB ){
			linepix = file->buf->width * StripPerRow * 3;
		} else {
			linepix = file->buf->width * StripPerRow;
		}
		if( (tmp = malloc( linepix )) == NULL ){
			file->mes->datasize = 0;
			close( file->fd );
			return( EMALLOC );
		}
	} else {
		tmp = NULL;
	}
	p = buf->data;
	for( io = Ioff, i = 0; i < LenNum; i++, io++ ){
		if( lseek( file->fd, io->d[0], 0 ) == -1L ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			close( file->fd );
			return( ESEEK );
		}
		if( read( file->fd, buff, io->d[1] ) != io->d[1] ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			close( file->fd );
			return( EREADFILE );
		}
		set_shmbuff( p, buff, buf->width, tmp );
		p += (file->buf->width * StripPerRow);
	}
	close( file->fd );
	buf->mask_kind = SIMPLEMASK;
	free( buff );
	if( tmp ){
		free( tmp );
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function get_newpallet()                                */
/*                                                                    */
/*            args  struct Shm *buf  : data buffer                    */
/*                  struct ReduceInfo *ri : reduce call interface     */
/*                                                                    */
/*            return int : 0==success  else=errorcode                 */
/*                                                                    */
/*            function: If color reduce necessary, reduce do.         */
/*            D0: not malloc/free pallet memory.. make static array   */
/*                                                                    */
/**********************************************************************/
int get_newpallet( struct Shm *buf, struct ReduceInfo *ri )
{
	int ret;
/* D0 add */
	static struct Color pallet[256];
/* D0 end */

	if( (ri->image = malloc( buf->width * buf->height )) == NULL ){
		return( EMALLOC );
	}
/* D0 delete */
/*	if( (ri->pallet = malloc( sizeof(struct Color) * ri->target )) == NULL ){
		return( EMALLOC );
	}*/
/* D0 add */
	ri->pallet = pallet;
/* D0 end */
	if( (ret = get_reduce_image( buf, ri )) != 0 ){
		return( ret );
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function write_header()                                */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE HEADER of the SPEC format to file       */
/*             D0: for 1 StripNum(small image), re-define RpS.        */
/*                                                                    */
/**********************************************************************/
int write_header( struct File *file )
{
	unsigned char *buff;
	struct TiffHeader *header;
	short *s;
	long *l;
	short  
	int i, j;
	struct TiffIFD	 *ifd;
	int ifds, sc, size, scp;

	if( (file->fd = open( file->name, O_RDWR | O_TRUNC )) == -1 ){
		if( (file->fd = open( file->name, O_RDWR | O_CREAT, 0664)) == -1 ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
	}

	switch( WriteMode ){
	case K_RGB:
		RpS = 8000 / file->buf->width / 3;
/* D0 add: for 1 StripNum(small image), re-define RpS. */
		if( RpS > file->buf->height ){
		     RpS = file->buf->height;
		}
/* end D0 */
		StripSize = RpS * file->buf->width * 3;
		break;
	case K_COLORMAP:
	case K_GRAY:
		RpS = 8000 / file->buf->width;
/* D0 add: for 1 StripNum(small image), re-define RpS. */
		if( RpS > file->buf->height ){
		     RpS = file->buf->height;
		}
/* end D0 */
		StripSize = RpS * file->buf->width;
		break;
	case K_BLACK:
	case K_WHITE:
		RpS = 8000 / ((file->buf->width + 7) / 8);
/* D0 add: for 1 StripNum(small image), re-define RpS. */
		if( RpS > file->buf->height ){
		     RpS = file->buf->height;
		}
/* end D0 */
		StripSize = RpS * ((file->buf->width + 7) / 8);
		break;
	default:
		return( -1 );
		break;
	}
	StripNum = (file->buf->height + RpS - 1) / RpS;
	switch( WriteMode ){
	case K_RGB:
/* D0: add SAMPLESPERPIXEL tag */
/*		ifds = 12;*/
		ifds = 13;
/* D0 end */
		sc =   8 + 2 + ifds * 12 + 4;  /* 158 + 12 */
		size = sc + 6 + 16;  /* 168 */
		break;
	case K_COLORMAP:
/* D0: add SAMPLESPERPIXEL tag */
/*		ifds = 13;*/
		ifds = 14;
/* D0 end */
		sc =   8 + 2 + ifds * 12 + 4;  /* 170 + 12 */
		size = sc + 16 + 6 * 256;
		break;
	case K_GRAY:
	case K_BLACK:
	case K_WHITE:
/* D0: add SAMPLESPERPIXEL tag */
/*		ifds = 12;*/
		ifds = 13;
/* D0 end */
		sc =   8 + 2 + ifds * 12 + 4;  /* 158 + 12 */
		size = sc + 16;
		break;
	default:
		return( -1 );
		break;
	}
/* D0: if StripNum==1, not add */
/* 8 is StripOffsets and StripByteCounts, each 4(V_LONG) */
	if( StripNum > 1 ){
	     size += 8 * StripNum;
	}
/*	size += 8 * StripNum;*/
/* end D0 */
	if( (buff = malloc( size )) == NULL ){
		return( EMALLOC );
	}
	header = (struct TiffHeader *)&buff[0];
	if( LSB_flag ){
		header->sig = 0x4949;
	} else {
		header->sig = 0x4d4d;
	}
	header->ver = 0x002a;
	header->next = 8L;
	
	s = (short *)&buff[8];
	*s = ifds;
	l = (long *)&buff[sc-4];
	*l = 0L;
	scp = sc;

	if( WriteMode != K_RGB ){
		switch( WriteMode ){
		case K_COLORMAP:
			RI.mode = GETPALLETCOLOR;
			RI.target = 256;
			RI.tmp_colors = 16;
			RI.submode = (COLOR | TOTALNUM | ZAGGY);
			break;
		case K_GRAY:
			RI.mode = GETPALLETCOLOR;
			RI.target = 256;
			RI.tmp_colors = 256;
			RI.submode = (BW | INDEPENDENT | FLAT);
			break;
		case K_BLACK:
		case K_WHITE:
			RI.mode = GETPALLETCOLOR;
			RI.target = 2;
			RI.tmp_colors = 2;
			RI.submode = (BW | TOTALNUM | ZAGGY);
			break;
		default:
			fprintf( stderr, "unrec WriteMode %d\n", WriteMode );
			return( -1 );
		}
		if( get_newpallet( file->buf, &RI ) != 0 ){
			fprintf( stderr, "get pallet error\n" );
			return( -1 );
		}
	}

	for( i = 0; i < ifds; i++ ){
		ifd = (struct TiffIFD *)&buff[10 + i * 12];
		switch( i ){
		case 0:
			ifd->tag = S_NEWSUBFILETYPE;
			ifd->length = 1;
			ifd->field = V_LONG;
			ifd->offset = 0L;
			break;
		case 1:
			ifd->tag = S_IMAGEWIDTH;
			ifd->length = 1;
			ifd->field = V_LONG;
			ifd->offset = (long)file->buf->width;
			break;
		case 2:
			ifd->tag = S_IMAGELENGTH;
			ifd->length = 1;
			ifd->field = V_LONG;
			ifd->offset = (long)file->buf->height;
			break;
		case 3:
			ifd->tag = S_BITSPERSAMPLE;
			ifd->field = V_SHORT;
			if( WriteMode == K_RGB ){
				ifd->length = 3;
				ifd->offset = (long)scp; /*158L;*/
				s = (short *)&buff[scp]; /* 158];*/
				*s++ = 8;
				*s++ = 8;
				*s = 8;
				scp += 6;
			} else if( WriteMode == K_COLORMAP 
				  || WriteMode == K_GRAY){
				ifd->length = 1;
				ifd->offset = 8;
			} else if( WriteMode == K_WHITE 
				  || WriteMode == K_BLACK){
				ifd->length = 1;
				ifd->offset = 1;
			} else {
				return( -1 );
			}
			break;
		case 4:
			ifd->tag = S_COMPRESSION;
			ifd->length = 1;
			ifd->field = V_SHORT;
			ifd->offset = 1;
			break;
		case 5:
			ifd->tag = S_PHOTOMETRICINTERPRETATION;
			ifd->length = 1;
			ifd->field = V_SHORT;
			if( WriteMode == K_RGB ){
				ifd->offset = 2;
			} else if( WriteMode == K_COLORMAP ){
				ifd->offset = 3;
			} else if( WriteMode == K_GRAY 
				  || WriteMode == K_WHITE ){
				ifd->offset = 1;
			} else if( WriteMode == K_BLACK ){
				ifd->offset = 0;
			} else {
				return( -1 );
			}
			break;
		case 6:
			ifd->tag = S_STRIPOFFSETS;
			ifd->length = StripNum;
			ifd->field = V_LONG;
/* D0: if StripNum==1, not use scp */
			if( StripNum > 1 ){
/* D0 add end */
			     ifd->offset = (long)scp;  /*180L*/;
			     l = (long *)&buff[scp];
			     for( j = 0; j < StripNum; j++ ){
				  *l++ = (long)(size + StripSize * j);
				  scp += 4;
			     }
/* D0: if StripNum==1, use offset field */
			} else {
			     ifd->offset = (long)size;
			}
/* D0 add end */
			break;
/* D0 add.. to write SAMPLESPERPIXEL */
		case 7:
			ifd->tag = S_SAMPLESPERPIXEL;
			ifd->length = 1;
			ifd->field = V_SHORT;
			if( WriteMode == K_RGB ){
				ifd->offset = 3;
			} else if( WriteMode == K_COLORMAP
				   || WriteMode == K_GRAY 
				   || WriteMode == K_WHITE
				   || WriteMode == K_BLACK ){
				ifd->offset = 1;
			} else {
				return( -1 );
			}
			break;
/* D0 add end, case tag will increase after */
		case 8:
			ifd->tag = S_ROWSPERSTRIP;
			ifd->length = 1;
			ifd->field = V_LONG;
			ifd->offset = (long)RpS;
			break;
		case 9:
			ifd->tag = S_STRIPBYTECOUNTS;
			ifd->length = StripNum;
			ifd->field = V_LONG;
/* D0: if StripNum==1, not use scp */
			if( StripNum > 1 ){
/* end D0 */
			     ifd->offset = (long)scp; /*(180 + StripNum * 4);*/
			     l = (long *)&buff[scp]; /*180 + StripNum * 4];*/
			     for( j = 0; j < StripNum - 1; j++ ){
				  *l++ = StripSize;
				  scp += 4;
			     }
			     if( WriteMode == K_RGB ){
				  *l++ = (long)((file->buf->height - (StripNum - 1) * RpS)
						* file->buf->width * 3);
			     } else if( WriteMode == K_COLORMAP 
					|| WriteMode == K_GRAY ){
				  *l++ = (long)((file->buf->height - (StripNum - 1) * RpS)
						* file->buf->width);
			     } else if( WriteMode == K_BLACK
					|| WriteMode == K_WHITE ){
				  *l++ = (long)((file->buf->height - (StripNum - 1) * RpS)
						* ((file->buf->width +7) / 8));
			     } else {
				  return( -1 );
			     }
			     scp += 4;
/* D0: if StripNum==1, use offset field */
			} else {
			     if( WriteMode == K_RGB ){
				  ifd->offset = (long)(file->buf->height 
						       * file->buf->width * 3);
			     } else if( WriteMode == K_COLORMAP 
					|| WriteMode == K_GRAY ){
				  ifd->offset = (long)(file->buf->height 
						       * file->buf->width);
			     } else if( WriteMode == K_BLACK
					|| WriteMode == K_WHITE ){
				  ifd->offset = (long)(file->buf->height 
						       * ((file->buf->width + 7) / 8));
			     } else {
				  return( -1 );
			     }
			}
/* D0 add end */
			break;
		case 10:
			ifd->tag = S_XRESOLUTION;
			ifd->length = 1;
			ifd->field = V_RATIONAL;
			ifd->offset = (long)scp /*164L*/;
			l = (long *)&buff[scp]; /*164];*/
			*l++ = 150L;
			*l = 1L;
			scp += 8;
			break;
		case 11:
			ifd->tag = S_YRESOLUTION;
			ifd->length = 1;
			ifd->field = V_RATIONAL;
			ifd->offset = (long)scp; /*172L;*/
			l = (long *)&buff[scp]; /*172];*/
			*l++ = 150L;
			*l = 1L;
			scp += 8;
			break;
		case 12:
			ifd->tag = S_RESOLUTIONUNIT;
			ifd->length = 1;
			ifd->field = V_SHORT;
			ifd->offset = 2;
			break;
		case 13:
			if( WriteMode != K_COLORMAP ){
				break;
			}
			ifd->tag = S_COLORMAP;
			ifd->length = 256 * 3;
			ifd->field = V_SHORT;
			ifd->offset = (long)scp;
			s = (unsigned short *)&buff[scp];

			for( i = 0; i < RI.target; i++ ){
				*s    = RI.pallet[i].R;
				*s++ |= ((unsigned short)RI.pallet[i].R << 8);
				scp += 2;
			}
			for( i = 0; i < RI.target; i++ ){
				*s     = RI.pallet[i].G;
				*s++ |= ((unsigned short)RI.pallet[i].G << 8);
				scp += 2;
			}
			for( i = 0; i < RI.target; i++ ){
				*s    = RI.pallet[i].B;
				*s++ |= ((unsigned short)RI.pallet[i].B << 8);
				scp += 2;
			}
			break;
		default:
			break;
		}
	}

	if( scp != size ){
		fprintf( stderr, "illegal operation of IFD\n" );
		return( -1 );
	}
	if( write( file->fd, buff, size ) != size ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EWRITEFILE );
	}
	free( buff );

	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE IMAGES of the SPEC format to file       */
/*             D0: BUG.. cannot write fullcolor tiff image            */
/*                                                                    */
/**********************************************************************/
int write_image( struct File *file )
{
	unsigned char *buff;
	struct Shm *buf;
	unsigned char *p;
	int i, j, k, l, r;
	int rest;

	buf = file->buf;
	if( (buff = malloc( StripSize )) == NULL ){
		file->mes->datasize = 0;
		close( file->fd );
		return( EMALLOC );
	}
	if( WriteMode == K_RGB ){
		for( i = 0, k = 0; k < StripNum - 1; k++ ){
			for( j = 0; j < StripSize; i++ ){
				buff[j++] = (unsigned char)GET_RED(buf->data[i]);
				buff[j++] = (unsigned char)GET_GREEN(buf->data[i]);
				buff[j++] = (unsigned char)GET_BLUE(buf->data[i]);
			}
			if( write( file->fd, buff, StripSize ) != StripSize ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				return( EWRITEFILE );
			}
		}
		rest = (buf->height - (StripNum - 1) * RpS) * file->buf->width;
/* D0 cannot write tiff fullcolor image....*/
/* D0 add */
		k = i;
		for( i = 0, j = 0; i < rest; i++ ){
			buff[j++] = (unsigned char)GET_RED(buf->data[k+i]);
			buff[j++] = (unsigned char)GET_GREEN(buf->data[k+i]);
			buff[j++] = (unsigned char)GET_BLUE(buf->data[k+i]);
/* D0 delete	for( i = 0, j = 0; i < rest; i++ ){
			buff[j++] = (unsigned char)GET_GREEN(buf->data[i]);
			buff[j++] = (unsigned char)GET_BLUE(buf->data[i]);
			buff[j++] = (unsigned char)GET_BLUE(buf->data[i]);
  D0 end */
		}
		rest = rest * 3;
		if( rest % 2 ){
			rest++;
		}
		if( write( file->fd, buff, rest ) != rest ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			close( file->fd );
			return( EWRITEFILE );
		}
	} else if( WriteMode == K_GRAY 
		  || WriteMode == K_COLORMAP ){
		p = RI.image;
		for( k = 0; k < StripNum - 1; k++ ){
			memcpy( buff, p, StripSize );
			if( write( file->fd, buff, StripSize ) != StripSize ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				free( buff );
				free( RI.image );
/* D0 delete			free( RI.pallet );*/
				close( file->fd );
				return( EWRITEFILE );
			}
			p += StripSize;
		}

		rest = (buf->height - (StripNum - 1) * RpS) * file->buf->width;
		memcpy( buff, p, rest );
		if( rest % 2 ){
			rest++;
		}
		if( write( file->fd, buff, rest ) != rest ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			close( file->fd );
			free( buff );
			free( RI.image );
/* D0 delete		free( RI.pallet ); */
			return( EWRITEFILE );
		}
		free( RI.image );
/* D0 delete	free( RI.pallet ); */
	} else if( WriteMode == K_WHITE 
		  || WriteMode == K_BLACK ){
		p = RI.image;
		for( k = 0; k < StripNum - 1; k++ ){
			l = 0; r = 0;
			for( i = 0; i < StripSize; i++ ){
				if( r >= RpS ){
					break;
				}
				buff[i] = 0;
				for( j = 7; j >= 0; j--, p++ ){
					if( WriteMode == K_BLACK && (!*p) ){
						buff[i] |= BitMask[j];
					} else if( WriteMode == K_WHITE && *p ){
						buff[i] |= BitMask[j];
					}
					if( l++ >= file->buf->width ){
						l = 0; r++;
						break;
					}
				}
			}
			if( write( file->fd, buff, StripSize ) != StripSize ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				free( buff );
				free( RI.image );
/* D0 delete			free( RI.pallet ); */
				return( EWRITEFILE );
			}
		}

/* D0 correct calculate rest byte */
		rest = (buf->height - (StripNum - 1) * RpS) * ((file->buf->width + 7) / 8);
/*		rest = (buf->height - (StripNum - 1) * RpS) * ((file->buf->width + 7) / 8);*/
/* D0 end */
		l = 0;
		for( i = 0; i < rest; i++ ){
			buff[i] = 0;
			for( j = 7; j >= 0; j--, p++ ){
				if( WriteMode == K_BLACK && (!*p) ){
					buff[i] |= BitMask[j];
				} else if( WriteMode == K_WHITE && *p ){
					buff[i] |= BitMask[j];
				}
				if( l++ >= file->buf->width ){
					l = 0;
					break;
				}
			}
		}
		if( rest % 2 ){
			rest++;
		}
		if( write( file->fd, buff, rest ) != rest ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			close( file->fd );
			free( buff );
			free( RI.image );
/* D0 delete		free( RI.image );*/
			return( EWRITEFILE );
		}
		free( RI.image );
/* D0 delete	free( RI.image );*/
	}
	close( file->fd );
	free( buff );

	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
		if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
