/**********************************************************************/
/*                                                                    */
/*           undistort.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>

#include "plug.h"

/*             This is a virtual plug, not affect the Shm data.       */

char Name[256] = { "noname" };
char NewName[256] = { "dist" };
int Comp = 0;

int Gui = NOGUI;
int List = 0;
void search_file(FILE *, int, int, char *);

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec =    { "undistort" };
char *Comment = { "distotion adjustment efffect of cut-file(*.cut)" };

PLUGITEM Item[] = {
{ "new", 	GETSTR,  &NewName, 64, 0, "save cut file name" },
{ "comp=yes",   SWITCH,  &Comp, 1, 'Y', "comprement between" },
{ "comp=no",    SWITCH,  &Comp, 0, 'N', "NO comprement between" },
{ "list",	SWITCH,  &List, 1, 0,      "list all cutfile in #lwp" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" },
{ "name", 	GETSTR,  &Name, 64, 0,    "target cut file name" },
{ NULL, 	GETSTR,  &Name, 64, 0,    "target cut file name" } };
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ NOTCL | ENTRY, 6, "object cut file name", "name=%s" },
{ TCLLIST, 6, "object cut file name", NULL },
{ TCLDIR, 0, (char *)search_file, NULL },
{ TCLDIR, 1, (char *)search_file, NULL },
{ TCLLISTEND, 6, "input file name", "name=%s" },
{ NOTCL | ENTRY, 0, "create cut file name", "new=%s" },
{ TCLLIST, 0, "create cut file name", "new=%s" },
{ TCLDIR, 0, (char *)search_file, NULL },
{ TCLLISTENT, 0, "create file is", "new=%s" },
{ RADIO, 1, "Comprement " },
{ ITEM, 1, "yes(y)", "comp=yes" },
{ ITEM, 2, "no(n)", "comp=no" },
{ RADIOEND, -1 }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

#define WORD unsigned short
struct RawHeader {
	WORD spec;	/* LSB or MSB resolve flag */
	WORD width;	/* width */
	WORD height;	/* height */
	WORD mask;	/* mask kind */
};

struct FPOS {
     double x;
     double y;
};

struct FPOS Edge[4];

#define DISTANT(a,b)	((a)*(a)+(b)*(b))

double Dd;
struct FPOS Dest[4], Src[4];
double Od[4] = { 3.9269908170, 5.4977871438, 0.7853981634, 2.3561944902 };
struct FPOS Dc, Sc;
struct param {
     double radr;
     double rada;
     double distr;
     double dista;
     double from;
     double toat;
     double midr;
     double dds;
};
struct param Param[4];

unsigned int *Block;


/**********************************************************************/
/*                                                                    */
/*             function rev()                                         */
/*                                                                    */
/*             WORD v        : reverse short                          */
/*                                                                    */
/*             return  WORD  : reversed short                         */
/*                                                                    */
/*             function:  Reverse the byte order at 2byte WORD.       */
/*                                                                    */
/**********************************************************************/
WORD rev( WORD v )
{
	WORD ret;

	ret = ((v & 0xff00) >> 8);
	ret |= ((v & 0x00ff) << 8);
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function round()                                       */
/*                                                                    */
/*             double f      : double value                            */
/*                                                                    */
/*             return  int   : rounded value(int converted)           */
/*                                                                    */
/*             function: Round the double value.                       */
/*                                                                    */
/**********************************************************************/
int round( double f )
{
	double r;

	r = floor( f );
	if( f - r >= 0.5 ){
		return( (int)r + 1 );
	} else {
		return( (int)r );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function undistort_proc()                              */
/*                                                                    */
/*             unsigned int *buff  : original cut data                */
/*             int fwidth          : original width                   */
/*             int fheight         : original height                  */
/*             int mask            : original mask kind               */
/*                                                                    */
/*             return  int : 0==success 0!=errorcode                  */
/*                                                                    */
/*             function: write the rotated cut file                   */
/*                                                                    */
/**********************************************************************/
void set_param( int i )
{
     double r1, r2, d1, d2;
     double R1, R2, D1, D2;
     int n;
     double sinv, cosv;

     if( i >= 3 ){
	  n = 0;
     } else {
	  n = i + 1;
     }
     d1 = sqrt( DISTANT( Src[i].x - Sc.x, Src[i].y - Sc.y ) );
     d2 = sqrt( DISTANT( Src[n].x - Sc.x, Src[n].y - Sc.y ) );
     sinv = (Src[i].y - Sc.y) / d1;
     cosv = (Src[i].x - Sc.x) / d1;
     if( sinv >= 0.0 ){
	  r1 = acos( cosv );
     } else {
	  r1 = 2 * M_PI - acos( cosv );
     }
     sinv = (Src[n].y - Sc.y) / d2;
     cosv = (Src[n].x - Sc.x) / d2;
     if( sinv >= 0.0 ){
	  r2 = acos( cosv );
     } else {
	  r2 = 2 * M_PI - acos( cosv );
     }
     D1 = D2 = Dd;
     sinv = (Dest[i].y - Dc.y) / D1;
     cosv = (Dest[i].x - Dc.x) / D1;
     if( sinv >= 0.0 ){
	  R1 = acos( cosv );
     } else {
	  R1 = 2 * M_PI - acos( cosv );
     }
     sinv = (Dest[n].y - Dc.y) / D2;
     cosv = (Dest[n].x - Dc.x) / D2;
     if( sinv >= 0.0 ){
	  R2 = acos( cosv );
     } else {
	  R2 = 2 * M_PI - acos( cosv );
     }
     if( R1 > R2 ){
	  R2 += 2 * M_PI;
     }
     if( r1 > r2 ){
	  r2 += 2 * M_PI;
     }
     Param[i].radr = (r2 - r1) / (R2 - R1);
     Param[i].rada = r2 - Param[i].radr * R2;
     Param[i].from = R1;
     Param[i].toat = R2;
     Param[i].distr = (d2 - d1) / (R2 - R1);
     Param[i].dista = d2 - Param[i].distr * R2;
     Param[i].midr = (r1+r2)/2;
     Param[i].dds = cos((r2-r1)/2) / cos((R2-R1)/2);
}

struct FPOS get_pixel( int x, int y )
{
     double r, d;
     int ind;
     double rs, ds;
     struct FPOS ret;
     double sinv, cosv;
     double dd, mid;

     d = sqrt( DISTANT( (double)x - Dc.x, (double)y - Dc.y ) );
     if( d == 0.0 ){
	  r = 0.0;
	  sinv = 0.0;
	  cosv = 1.0;
     } else {
	  sinv = ((double)y - Dc.y) / d;
	  cosv = ((double)x - Dc.x) / d;
     }
     if( sinv >= 0.0 ){
	  r = acos( cosv );
     } else {
	  r = 2 * M_PI - acos( cosv );
     }
     for( ind = 0; ind < 4; ind++ ){
	  if( r >= Param[ind].from && r < Param[ind].toat ){
	       break;
	  }
	  if( r + M_PI * 2.0 >= Param[ind].from && r + M_PI * 2.0 < Param[ind].toat ){
	       r += M_PI * 2.0;
	       break;
	  }
     }
     mid = (Param[ind].from + Param[ind].toat) / 2;
     rs = Param[ind].radr * r + Param[ind].rada;
     if( r > mid ){
	  dd  = cos(r - mid) / cos(rs - Param[ind].midr);
     } else {
	  dd  = cos(mid - r) / cos(Param[ind].midr - rs);
     }
     ds = d * (Param[ind].distr * r + Param[ind].dista) / Dd * dd * Param[ind].dds;
     ret.x = ds * cos(rs) + Sc.x;
     ret.y = ds * sin(rs) + Sc.y;
     return ret;
}

#if 0	       
unsigned int get_value( unsigned int *buff, int fwidth, int fheight, struct FPOS at )
{
     int x1, x2, y1, y2, x, y;
     double x1p, x2p, y1p, y2p;
     double r,g,b;
     int ret;
     unsigned int ir, ig, ib, m;

     x1 = (int)at.x; x2 = x1 + 1;
     y1 = (int)at.y; y2 = y1 + 1;
     x1p = at.x - (double)x1; x2p = 1.0 - x1p;
     y1p = at.y - (double)y1; y2p = 1.0 - y1p;
     x = round(at.x); y = round(at.y);
     if( x1 < 0 ){
	  x1 = x2 = 0;
     } else if( x2 >= fwidth ){
	  x1 = x2 = fwidth - 1;
     }
     if( y1 < 0 ){
	  y1 = y2 = 0;
     } else if( y2 >= fheight ){
	  y1 = y2 = fheight - 1;
     }
     r = x1p * (double)GET_RED(buff[y1*fwidth+x1]) + x2p * (double)GET_RED(buff[y1*fwidth+x2]);
     r = y1p * r + y2p * (double)GET_RED(buff[y2*fwidth+x1]);
     ir = (unsigned int)r;
     g = x1p * (double)GET_GREEN(buff[y1*fwidth+x1]) + x2p * (double)GET_GREEN(buff[y1*fwidth+x2]);
     g = y1p * g + y2p * (double)GET_GREEN(buff[y2*fwidth+x1]);
     ig = (unsigned int)g;
     b = x1p * (double)GET_BLUE(buff[y1*fwidth+x1]) + x2p * (double)GET_BLUE(buff[y1*fwidth+x2]);
     b = y1p * b + y2p * (double)GET_BLUE(buff[y2*fwidth+x1]);
     ib = (unsigned int)b;
     m = GET_MASK(buff[y*fwidth+x]);
     ret = ((m << 24) | (ir << 16) | (ig << 8) | ib);
     return ret;
}
#endif

unsigned int get_comp( unsigned int *s, int w, int h, struct FPOS p )
{
     double r, g, b;
     int x0, y0;
     double px0, py0;
     int at;
     int flag;
     unsigned int ir, ig, ib, im;

     x0 = round(p.x); y0 = round(p.y);
     at = y0 * w + x0;
     px0 = p.x - (double)x0; py0 = p.y - (double)y0;
     flag = 1;
     if( p.x - 1 < w ){
	  r = (double)GET_RED(s[at]) * px0 + (double)GET_RED(s[at+1]) * (1.0 - px0);
	  g = (double)GET_GREEN(s[at]) * px0 + (double)GET_GREEN(s[at+1]) * (1.0 - px0);
	  b = (double)GET_BLUE(s[at]) * px0 + (double)GET_BLUE(s[at+1]) * (1.0 - px0);
     } else {
	  r = (double)GET_RED(s[at]);
	  g = (double)GET_GREEN(s[at]);
	  b = (double)GET_BLUE(s[at]);
	  flag = 0;
     }
     at += w;
     if( p.y - 1 < h ){
	  r = r * py0 + (double)GET_RED(s[at]) * (1.0 - py0);
	  g = g * py0 + (double)GET_GREEN(s[at]) * (1.0 - py0);
	  b = b * py0 + (double)GET_BLUE(s[at]) * (1.0 - py0);
     } else {
	  flag = 0;
     }
     at += 1;
     if( flag ){
	  px0 = px0 * py0;
	  r = r * px0 + (double)GET_RED(s[at]) * (1.0 - px0);
	  g = g * px0 + (double)GET_GREEN(s[at]) * (1.0 - px0);
	  b = b * px0 + (double)GET_BLUE(s[at]) * (1.0 - px0);
     }
     at = at - w - 1;
     im = GET_MASK(s[at]);
     ir = (unsigned int)r; ig = (unsigned int)g; ib = (unsigned int)b;
     if( ir < 0 ) ir = 0;
     if( ig < 0 ) ig = 0;
     if( ib < 0 ) ib = 0;
     if( ir > 255 ) ir = 255;
     if( ig > 255 ) ig = 255;
     if( ib > 255 ) ib = 255;
     return ((im << 24) | (ir << 16) | (ig << 8) | ib);
}

unsigned int get_value( unsigned int *buff, int fwidth, int fheight, struct FPOS at )
{
     int x, y;

     x = round(at.x); y = round(at.y);
     if( y < 0 || y >= fheight || x < 0 || x >= fwidth ){
	  return 0;
     }
     return buff[y*fwidth+x];
}

int undistort_proc( unsigned int *buff, int fwidth, int fheight, int mask )
{
     int i, j;
     int x, y;
     int count;
     double min, d;
     int mind;
     double minx, miny, maxx, maxy;
     double Dx, Dy;
     int w, h, len;
     int fd;
     struct RawHeader header;
     char filename[256];
     char *envstr;
     struct FPOS ret;

     for( count = x = 0; x < fwidth; x++ ){
	  for( i = x, j = 0; i >= 0; i--, j++ ){
	       if( MASKOFF(buff[j*fwidth+i]) ){
		    count++;
		    Edge[0].x = (double)i; Edge[0].y = (double)j;
	       }
	  }
	  if( count > 0 ){
	       break;
	  }
     }
     if( count != 1 ){
	  return( EILLSELECT );
     }

     for( count = 0, x = fwidth - 1; x >= 0; x-- ){
	  for( i = x, j = 0; i < fwidth; i++, j++ ){
	       if( MASKOFF(buff[j*fwidth+i]) ){
		    count++;
		    Edge[1].x = (double)i; Edge[1].y = (double)j;
	       }
	  }
	  if( count > 0 ){
	       break;
	  }
     }
     if( count != 1 ){
	  return( EILLSELECT );
     }
     for( count = 0, y = fheight - 1; y >= 0; y-- ){
	  for( i = fwidth - 1, j = y; j < fheight; i--, j++ ){
	       if( MASKOFF(buff[j*fwidth+i]) ){
		    count++;
		    Edge[2].x = (double)i; Edge[2].y = (double)j;
	       }
	  }
	  if( count > 0 ){
	       break;
	  }
     }
     if( count != 1 ){
	  return( EILLSELECT );
     }

     for( count = 0, y = fheight - 1; y >= 0; y-- ){
	  for( i = 0, j = y; j < fheight; i++, j++ ){
	       if( MASKOFF(buff[j*fwidth+i]) ){
		    count++;
		    Edge[3].x = (double)i; Edge[3].y = (double)j;
	       }
	  }
	  if( count > 0 ){
	       break;
	  }
     }
     if( count != 1 ){
	  return( EILLSELECT );
     }

     minx = miny = 999999.9;
     maxx = maxy = -999999.9;
     for( i = 0; i < 4; i++ ){
	  if( minx > Edge[i].x ){
	       minx = Edge[i].x;
	  }
	  if( miny > Edge[i].y ){
	       miny = Edge[i].y;
	  }
	  if( maxx < Edge[i].x ){
	       maxx = Edge[i].x;
	  }
	  if( maxy < Edge[i].y ){
	       maxy = Edge[i].y;
	  }
     }
     Dx = maxx - minx;
     Dy = maxy - miny;
     Dest[0].x = 0; Dest[0].y = 0;
     Dest[1].x = Dx; Dest[1].y = 0;
     Dest[2].x = Dx; Dest[2].y = Dy;
     Dest[3].x = 0; Dest[3].y = Dy;
     Dc.x = Dx / 2.0; Dc.y = Dy / 2.0;
     Dd = sqrt( DISTANT( Dc.x, Dc.y ) );
     mind = -1;

     for( i = 0; i < 4; i++ ){
	  min = 99999999.0;
	  for( j = 0; j < 4; j++ ){
	       d = DISTANT( Edge[j].x - Dest[i].x, Edge[j].y - Dest[i].y );
	       if( min > d ){
		    min = d;
		    mind = j;
	       }
	  }
	  Src[i].x = Edge[mind].x;
	  Src[i].y = Edge[mind].y;
     }
     Sc.x = (Src[0].x + Src[1].x + Src[2].x + Src[3].x) / 4.0;
     Sc.y = (Src[0].y + Src[1].y + Src[2].y + Src[3].y) / 4.0;
     for( i = 0; i < 4; i++ ){
	  set_param(i);
     }
     w = (int)Dx + 1;
     h = (int)Dy + 1;
     len = w * h * sizeof(unsigned int);
     if( (Block = malloc( len )) == NULL ){
	  fprintf( stderr, "cannot malloc\n" );
	  return( EMALLOC );
     }
     for( j = 0; j < h; j++ ){
	  for( i = 0; i < w; i++ ){
	       ret = get_pixel( i, j );
	       if( Comp == 0 ){
		    Block[j * w + i] = get_value( buff, fwidth, fheight, ret );
	       } else {
		    Block[j * w + i] = get_comp( buff, fwidth, fheight, ret );
	       }
	  }
     }
     if( (envstr = getenv( "GEDO_LWP" )) == NULL ){
	  strcpy( filename, "./" );
     } else {
	  strcpy( filename, envstr );
     }
     if( filename[strlen(filename) -1] != '/' ){
	  strcat( filename, "/" );
     }
     strcat( filename, NewName );
     strcat( filename, ".cut" );
     if( (fd = open( filename, O_RDWR | O_TRUNC )) == -1 ){
	  if( (fd = open( filename, O_RDWR | O_CREAT, 0664)) == -1 ){
	       free( Block );
	       return( EOPENFILE );
	  }
     }
     header.spec = 1;
     header.width = w;
     header.height = h;
     header.mask = SIMPLEMASK;
     if( write( fd, &header, sizeof(struct RawHeader) ) != sizeof(struct RawHeader) ){
	  return( EREADFILE );
     }
     if( write( fd, Block, len ) != len ){
	  return( EREADFILE );
     }
     close( fd );
     free( Block );

     return 0;
}

/**********************************************************************/
/*                                                                    */
/*             function proc()                                        */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : 0==success 0!=errorcode                  */
/*                                                                    */
/*             function: Read cut file and rotate cut.                */
/*                                                                    */
/**********************************************************************/
int proc( void )
{
	int fd;
	struct RawHeader header;
	int len;
	int ret;
	int fwidth, fheight, fmask;
	unsigned int *buff;
	char filename[256];
	char *envstr;

	if( (envstr = getenv( "GEDO_LWP" )) == NULL ){
		strcpy( filename, "./" );
	} else {
		strcpy( filename, envstr );
	}
	if( filename[strlen(filename) -1] != '/' ){
		strcat( filename, "/" );
	}
	strcat( filename, Name );
	strcat( filename, ".cut" );

	if( (fd = open( filename, O_RDONLY )) == -1 ){
		return( EOPENFILE );
	}
	if( read( fd, &header, sizeof(struct RawHeader) ) != sizeof(struct RawHeader) ){
		return( EREADFILE );
	}
	if( header.spec == 1 ){
		fwidth = (int)header.width;
		fheight = (int)header.height;
		fmask = (int)header.mask;
	} else if( header.spec == 0x0100 ){
		fwidth = (int)rev( header.width );
		fheight = (int)rev( header.height );
		fmask = (int)rev( header.mask );
	} else {
		return( EILLFORM );
	}
	if( fmask != SIMPLEMASK ){
		return( EILLMASK );
	}
	len = fwidth * fheight * sizeof(unsigned int);
	if( (buff = (unsigned int *)malloc( len )) == NULL ){
		return( EMALLOC );
	}
	if( read( fd, buff, len ) != len ){
		return( EREADFILE );
	}
	close( fd );

	ret = undistort_proc( buff, fwidth, fheight, fmask );
	free( buff );
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*             This is a virtual plug, not affect the Shm data.       */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	int ret;

	fprintf( stderr, "now undistort %s.cut to %s.cut\n", Name, NewName );
	ret= proc();
	if( ret != 0 ){
		strcpy( file->mes->u.data, strerror( errno ) );
		file->mes->datasize = strlen( file->mes->u.data );
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;
	int ret;

	str = file->mes->u.data;
	if( str[0] == '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		ret = line_config( file );
		if( ret == 0 && List ){
			search_dir( getenv( "GEDO_LWP" ), "cut" );
			List = 0;
		}
		return( ret );
	}
}

void search_file(FILE *fp, int var, int kind, char * arg)
{
	switch( kind ){
	case 0:
		fprintf( fp, ".lstb%d insert end \"noname\"\n", var );
		break;
	case 1:
		search_tcldir( fp, var, getenv( "GEDO_LWP" ), "cut" );
		break;
	default:
		break;
	}
}
