/**********************************************************************/
/*                                                                    */
/*             unit.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include "plug.h"


#define UNITWIDTH 25
int Unit = UNITWIDTH;
char Name[256] = { "dot" };

int Gui = NOGUI;
int List = 0;
void search_file(FILE *, int, int, char *);

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "unit" };
char *Comment = { "unit transfar effect of etc/*.unt" };

PLUGITEM Item[] = {
{ "unit", 	GETINT,  &Unit,	 1, UNITWIDTH,   "unit size" },
{ "list",	SWITCH,  &List, 1, 'L',   "list all cutfile in #gwp" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" },
{ "name", 	GETSTR,  &Name, 64, 0,    "unit file name" },
{ NULL, 	GETSTR,  &Name, 64, 0,    "unit file name" } };
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "unit size ", "unit=%d" },
{ NOTCL | ENTRY, 4, "unit pattern  name", "name=%s" },
{ TCLLIST, 4, "object cut file name", NULL },
{ TCLDIR, 0, (char *)search_file, NULL },
{ TCLDIR, 1, (char *)search_file, NULL },
{ TCLDIR, 2, (char *)search_file, NULL },
{ TCLLISTEND, 4, "patten is", "name=%s" },
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

int UnitNum = 13;	    /* Unit size in a unit file(1-25,step2) */
char Data[13][UNITWIDTH][UNITWIDTH];	/* Unit read array */
int Width;				/* Image width */

#define get_y(r,g,b) (unsigned int)(0.299*(float)r+0.587*(float)g+0.114*(float)b) 

/**********************************************************************/
/*                                                                    */
/*             function load_file()                                   */
/*                                                                    */
/*             char *str   : load pattern name                        */
/*             char *path  : directory unit file may be in            */
/*                                                                    */
/*             return  int : -1==open error(probibary not here)       */
/*                           else==file descripter                    */
/*             function: Try to open the unit file in <path>.         */
/*                                                                    */
/**********************************************************************/
int load_file( char *str, char *path )
{
	char file[256];

	strcpy( file, path );
	if( file[strlen(file)-1] != '/' ){
		strcat( file, "/" );
	}
	strcat( file, str );
	strcat( file, ".unt" );
	return( open( file, O_RDONLY ) );
}

/**********************************************************************/
/*                                                                    */
/*             function load_pattern()                                */
/*                                                                    */
/*             char *str   : load pattern name                        */
/*                                                                    */
/*             return  int : 0==success   else==errorcode             */
/*                                                                    */
/*             function: Read in the pattern file and set Data[][][]. */
/*                                                                    */
/**********************************************************************/
int load_pattern( char *str )
{
	int fd;
	int size;

	if( (fd = load_file( str, "." )) == -1 ){
		if( (fd = load_file( str, getenv( "GEDO_LWP" ) )) == -1 ){
			if( (fd = load_file( str, getenv( "GEDO_GWP" ) )) == -1 ){
				printf( "cannot find %s\n", str );
				return( EOPENFILE );
			}
		}
	}
			
	size = 13 * 25 * 25 * sizeof(char);
	if( read( fd, Data, size ) != size ){
		return( EREADFILE );
	}
	close( fd );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function set_work()                                    */
/*                                                                    */
/*             unsigned char *buff : temporary work area              */
/*             int ind             : the unit used by color depth     */
/*             int x               : position X                       */
/*             int y               : position Y                       */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Set work area the patten.                    */
/*                                                                    */
/**********************************************************************/
void set_work( unsigned char *buff, int ind, int x, int y )
{
	int i, j;
	int st, ed;
	unsigned char *p;

	st = (UNITWIDTH - Unit) / 2;
	ed = st + Unit;
	p = &buff[y * Width + x];
	for( i = st; i < ed; i++ ){
		for( j = 0; j < Unit; j++ ){
			p[j] = Data[ind][i][st+j];
		}
		p += Width;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function index()                                       */
/*                                                                    */
/*             unsigned int val    : the color depth at the pixel     */
/*                                                                    */
/*             return  unsigned int : index of the unit               */
/*                                                                    */
/*             function: Calc the index of the unit from color depth. */
/*                                                                    */
/**********************************************************************/
unsigned int get_index( unsigned int val )
{
	return( val  * UnitNum / 256 );
}

#include <signal.h>
void test( int signo )
{
	fprintf( stderr, "SEGV occured\n" );
	exit(1);
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_effect()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the effecter                            */
/*                                                                    */
/**********************************************************************/
int gedo_effect( struct File *file )
{
	int height;
	unsigned int *data, *p, *r;
	int i, j, k;
	unsigned char *work, *q;
	unsigned int *abs;
	int absw, absh;
	unsigned int R, G, B;
	int ret;

	k = 1; /* trick -W */
	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}
	Width = file->buf->width;
	height = file->buf->height;
	data = file->buf->data;

	if( (ret = load_pattern( Name )) != 0 ){
		strcpy( file->mes->u.data, strerror( errno ) );
		file->mes->datasize = strlen( file->mes->u.data );
		return( ret );
	}
	if( Unit % 2 == 0 ){
		Unit = Unit - 1;
	}

	UnitNum = (Unit - 1) / 2 + 1;
	if( (work = (unsigned char *)malloc( Width * height * 5)) == NULL ){
		return( EMALLOC );
	}
	abs = (unsigned int *)(work + Width * height );

	for( i = 0, p = data, q = work; i < height * Width; i++ ){
		R = GET_RED(*p); G = GET_GREEN(*p); B = GET_BLUE(*p);
		*q++ = (unsigned char)get_y( R, G, B );
		p++;
	}
	if( height % Unit ){
		absh = height / Unit + 1;
	} else {
		absh = height /Unit;
	}
	if( Width % Unit ){
		absw = Width / Unit + 1;
	} else {
		absw = Width /Unit;
	}

	memset( abs, 0, Width * height * sizeof(unsigned int) );
	p = abs; q = work;
	for( i = 0; i < absh; i++ ){
		for( j = 0; j < Unit; j++ ){
			if( i * Unit + j >= height ){
				break;
			}
			for( k = 0; k < Width; k++ ){
				p[k] += *q++;
			}
		}
                if( j != 0 ){
			for( k = 0; k < Width; k++, p++ ){
                                *p = *p / j;
                        }
                }
	}

	p = r = abs;
	for( i = 0; i < absh; i++ ){
		for( j = 0; j < absw; j++ ){
			for( k = 0; k < Unit; k++, r++ ){
				if( j * Unit + k >= Width ){
					goto ex2;
				}
				p[j] += *r;
				*r = 0;
			}
		}
ex2:		for( j = 0; j < absw - 1; j++, p++ ){
			*p = *p / Unit;
		}
		if( k != 0 ){
			*p++ = *p / k;
		}
	}
	p = abs;
	for( i = 0; i < absh; i++ ){
		for( j = 0; j < absw; j++, p++ ){
			set_work( work, get_index(*p), j * Unit, i * Unit );
		}
	}
	for( i = 0, p = data, q = work; i < height * Width; i++, p++, q++ ){
		if( MASKOFF( *p ) ){
			if( *q ){
				*p = ((*p & MASK) | ~MASK);
			} else {
				*p = ((*p & MASK));
			}
		}
	}

	free( work );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;
	int ret;
	char *env;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		ret = line_config( file );
		if( ret == 0 && List ){
			fprintf( stderr, "in ./\n" );
			search_dir( ".", "unt" );
			env = getenv( "GEDO_LWP" );
			fprintf( stderr, "in %s\n", env );
			search_dir( env, "unt" );
			env = getenv( "GEDO_GWP" );
			fprintf( stderr, "in %s\n", env );
			search_dir( env, "unt" );
			List = 0;
		}
		return( ret );
	}
	return( 0 );
}

void search_file(FILE *fp, int var, int kind, char * arg)
{
	switch( kind ){
	case 0:
		search_tcldir( fp, var, "./", "unt" );
		break;
	case 1:
		search_tcldir( fp, var, getenv( "GEDO_LWP" ), "unt" );
		break;
	case 2:
		search_tcldir( fp, var, getenv( "GEDO_GWP" ), "unt" );
		break;
	default:
		break;
	}
}
