/**********************************************************************/
/*                                                                    */
/*            width.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "plug.h"

int Fat = 1;
int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "width" };
char *Comment = { "make fat/thin now selection" };

PLUGITEM Item[] = {
{ "width", 	GETINT,  &Fat,	 -256, 256,   "fat distance >0 fat, <0 thin" },
{ "gui",        SWITCH,  &Gui,  GUI, 'G',  "use GUI interface" },
{ "nogui",      SWITCH,  &Gui,  NOGUI, 'N',  "not use GUI interface" },
};
int ItemNum = sizeof(Item)/sizeof(PLUGITEM);

struct Tags Tag[] = {
{ ENTRY, 0, "faten or thinen width ", "width=%d" }
 };
int TagNum = sizeof(Tag) / sizeof(struct Tags);

DEFAULTSTATUS;
/********************** description end ********************/

static int Width, Height;	/* image width and height */
unsigned int *Data;		/* image data top */

/**********************************************************************/
/*                                                                    */
/*             function fat_process()                                 */
/*                                                                    */
/*             int fat    : fat amount > 0                            */
/*             char *tmp  : temporary buffer                          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Make fat the selected mask.                  */
/*                                                                    */
/**********************************************************************/
void fat_process( int fat, char *tmp )
{
	unsigned int *p;
	char *q;
	int i, j, k;
	int count;

	for( k = 0; k < Fat; k++ ){
		p = Data; q = tmp;
		memset( tmp, 0, Width * Height );
		for( i = 0; i < Height; i++ ){
			for( j = 0; j < Width; j++, p++, q++ ){
				count = 0;
				if( i > 0 ){
					if( MASKOFF( *(p - Width) ) ){
						count++;
					}
				}
				if( i < Height - 1 ){
					if( MASKOFF( *(p + Width) ) ){
						count++;
					}
				}
				if( j > 0 ){
					if( MASKOFF( p[-1] ) ){
						count++;
					}
				}
				if( j < Width - 1 ){
					if( MASKOFF( p[1] ) ){
						count++;
					}
				}
				if( i > 0 && j > 0 ){
					if( MASKOFF( *(p - Width - 1) ) ){
						count++;
					}
				}
				if( i < Height - 1 && j < Width - 1 ){
					if( MASKOFF( *(p + Width + 1) ) ){
						count++;
					}
				}
				if( i > 0 && j < Width - 1 ){
					if( MASKOFF( *(p - Width + 1) ) ){
						count++;
					}
				}
				if( i < Height - 1 && j > 0 ){
					if( MASKOFF( *(p + Width - 1) ) ){
						count++;
					}
				}
				if( count ){
					*q = 1;
				}
			}
		}
		for( i = 0; i < Height * Width; i++ ){
			if( tmp[i] ){
				CLRMASK( Data[i] );
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function thin_process()                                */
/*                                                                    */
/*             int fat     : thin amount > 0                          */
/*             char *tmp   : temporary buffer                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Make thin the selected mask.                 */
/*                                                                    */
/**********************************************************************/
void thin_process( int fat, char *tmp )
{
	unsigned int *p;
	char *q;
	int i, j, k;
	int count;

	for( k = fat; k < 0; k++ ){
		p = Data; q = tmp;
		memset( tmp, 0, Width * Height );
		for( i = 0; i < Height; i++ ){
			for( j = 0; j < Width; j++, p++, q++ ){
				count = 0;
				if( MASKON( *p ) ){
					continue;
				}
				if( i > 0 ){
					if( MASKON( *(p - Width) ) ){
						count++;
					}
				}
				if( i < Height - 1 ){
					if( MASKON( *(p + Width) ) ){
						count++;
					}
				}
				if( j > 0 ){
					if( MASKON( p[-1] ) ){
						count++;
					}
				}
				if( j < Width - 1 ){
					if( MASKON( p[1] ) ){
						count++;
					}
				}
				if( count ){
					*q = 1;
				}
			}
		}
		for( i = 0; i < Height * Width; i++ ){
			if( tmp[i] ){
				SETMASK( Data[i] );
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_select()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: EXEC the selection                           */
/*                                                                    */
/**********************************************************************/
int gedo_select( struct File *file )
{
	struct Sbuf *sb;
	char *tmp;

	sb = &file->mes->u.sb;
	Width = file->buf->width;
	Height = file->buf->height;
	Data = file->buf->data;

	if( sb->selnum != SKNONE ){
		return( EILLSELECTION );
	}
	if( file->buf->mask_kind != SIMPLEMASK ){
		return( EILLMASK );
	}

	if( (tmp = (char *)malloc( Width * Height )) == NULL ){
		return( EMALLOC );
	}
	if( Fat > 0 ){
		fat_process( Fat, tmp );
	} else if( Fat < 0 ){
		thin_process( Fat, tmp );
	}
	free( tmp );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( str[0] ==  '\0' ){
	  	if( Gui ){
			return( config_use_gui( file ) );
		} else {
			return( config_comline( file ) );
		}
	} else {
		return( line_config( file ) );
	}
}
