/**********************************************************************/
/*                                                                    */
/*              wzl.c   gedo PLUG lzw library                         */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include "lzw.h"

static int StartUnit;			/* LZW start unit */
static int ClrCode;			/* LZW ClrCode, Starts[StartUnit] */
static int LZEof;			/* LZW EOF Code, ClrCode+1 */
static int  Unit;			/* LZW now reading unit */

typedef int STRING;
typedef unsigned short LZSTR;

struct LZ {				/* Table of Code */
	STRING s;	                /* word of dictionary */
	int len;			/* length of s */
};

#define TABLEMAX  8192			/* maximum of Table */
#define STRINGMAX (64 * 256)		/* string buffer alloc unit */
static struct LZ Table[TABLEMAX];	/* Dictionary table */
static int NowTable = 0;		/* used table */
static unsigned char *StringBuff = NULL;	/* string buffer */
static int StringMax;			/* now maximum of string buffer */
static int Stringp = 0;			/* used string buffer */
static int StartString;			/* Start position of 2byte longer string */
static int CC[]				/* ClrCode by StartUnit */ 
 = { 0, 0, 0, 4, 8, 16, 32, 64, 128, 256 };

static unsigned char *In_char;		/* Top of in-buffer */
static unsigned char *Out_char;		/* Top of out-buffer */
static int Wp;				/* now writing out-buffer */
static int Rp;				/* now reading in-buffer */
static int Out_size;			/* output maximum */
static int Quit = 0;			/* for error Quit */
static unsigned char At;		/* now reading in-buffer */
static int Sft = 0;			/* shift count of At */
static int TiffGif;			/* TIFF type or GIF type */
/* TIFF type --> write from LSB to MSB, Table full detect at CC[Unit] */ 
/* GIF type  --> write from MSB to LSB, Table full detect at CC[Unit]-1 */ 


/**********************************************************************/
/*                                                                    */
/*            function wzl_buff_strdup()                              */
/*                                                                    */
/*            args  unsigned char *name  : word of dictionary         */
/*                  int len   :  length of name                       */
/*                                                                    */
/*            return  STRING   :  word in StringBuff                  */
/*                                                                    */
/*            function: Copy name to the StringBuff.                  */
/*                                                                    */
/**********************************************************************/
static STRING wzl_buff_strdup( unsigned char *name, int len )
{
	STRING ret;
	unsigned char *tmp;

	if( Stringp + len >= StringMax ){
		if( (tmp = (unsigned char *)realloc( StringBuff, StringMax + STRINGMAX )) == NULL ){ 
			StringBuff = NULL;
			fprintf( stderr, "string buff overflow\n" );
			return( -1 );
		}
		StringBuff = tmp;
		StringMax += STRINGMAX;
	}
	ret = Stringp;
	memcpy( &StringBuff[Stringp], name, len );
	Stringp += len;
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*            function wzl_init_table()                               */
/*                                                                    */
/*            args  int start  :  start unit of input                 */
/*                  unsigned char *in  : Input buffer                 */
/*                  unsigned char *out : Output buffer                */
/*                  int size   :  Output maximum                      */
/*                                                                    */
/*            return  int   :  0==success  -1==fail                   */
/*                                                                    */
/*            function: Initilize the decoder.                        */
/*                                                                    */
/**********************************************************************/
static int wzl_init_table( int start, unsigned char *in, unsigned char *out, int size )
{
	unsigned char b[2];

	Unit = StartUnit = start;
	ClrCode = CC[start];
	LZEof = ClrCode + 1;
	In_char = in;
	Out_char = out;
	Wp = Rp = 0;
	Out_size = size;
	Quit = 0;
	At = Sft = 0;

	
	if( StringBuff ){
		free( StringBuff );
	}
	if( (StringBuff = malloc( STRINGMAX )) == NULL ){
		return( -1 );
	}
	StringMax = STRINGMAX;
	Stringp = 0;
	for( NowTable = 0; NowTable < LZEof+1; NowTable++ ){
		b[0] = (unsigned char)NowTable;
		Table[NowTable].s = wzl_buff_strdup( b, 1 );
		Table[NowTable].len = 1;
		if( Table[NowTable].s == -1 ){
			return( -1 );
		}
	}
	StartString = Stringp;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function *wzl_search_table()                            */
/*                                                                    */
/*            args  LZSTR c   : search code                           */
/*                                                                    */
/*            return  struct LZ   : NULL==not found                   */
/*                                  else==found TableNo.              */
/*            function: Search the Table by code.                     */
/*                                                                    */
/**********************************************************************/
static struct LZ *wzl_search_table( LZSTR c )
{
	if( c >= NowTable ){
		return( NULL );
	} else {
		return( &Table[c] );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function wzl_add_table()                                */
/*                                                                    */
/*            args  struct LZ *c : add code                           */
/*                                                                    */
/*            return  int  : 0==success  -1==fail                     */
/*                                                                    */
/*            function: Add new code to the dictionary.               */
/*                                                                    */
/**********************************************************************/
static int wzl_add_table( struct LZ *c )
{
	int cmp;

	if( NowTable >= TABLEMAX ){
		NowTable++;
		return( 0 );
	}
	Table[NowTable].s = wzl_buff_strdup( &StringBuff[c->s], c->len );
	if( Table[NowTable].s == -1 ){
		return( -1 );
	}
	Table[NowTable++].len = c->len;
	if( TiffGif == TIFF ){
		cmp = NowTable + 1;
	} else {
		cmp = NowTable;
	}
	switch( cmp ){
/*	switch( NowTable ){ GIF*/
/*	switch( NowTable + 1 ){ TIFF*/
	case   8:
	case  16:
	case  32:
	case  64:
	case 128:
	case 256:
	case 512:
	case 1024:
	case 2048:
		Unit++;
		break;
	case 4096:
		return( 0 );
		break;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function wzl_read_byte()                                */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return  int : read bit                                  */
/*                                                                    */
/*            function: Read from input the bit.                      */
/*                                                                    */
/**********************************************************************/
static int wzl_read_byte( void )
{
	static unsigned char writemask[8] = 
		{ 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 }; /* GIF */
/*		{ 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 }; TIFF*/
	int ret;

	if( Sft == 0 ){
		At = In_char[Rp++];
	}
	if( TiffGif == TIFF ){
		ret = (At & writemask[7 - Sft]);
	} else {
		ret = (At & writemask[Sft]);
	}
	if( ++Sft == 8 ){
		Sft = 0;
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function wzl_read_code()                                */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return  LZSTR : Read code from input                    */
/*                                                                    */
/*            function: Read code from input buffer.                  */
/*                                                                    */
/**********************************************************************/
static LZSTR wzl_read_code( void )
{
	int i;
	LZSTR ret;
	static LZSTR bitmask[16] = 
            { 0x8000, 0x4000, 0x2000, 0x1000, 0x0800, 0x0400, 0x0200, 0x0100,
              0x0080, 0x0040, 0x0020, 0x0010, 0x0008, 0x0004, 0x0002, 0x0001};

	ret = 0;
	if( TiffGif == TIFF ){ /* TIFF */
		for( i = 16 - Unit; i < 16; i++ ){
			if( wzl_read_byte() ){
				ret |= bitmask[i];
			}
		}
	} else { /* GIF */
		for( i = 15; i >= 16 - Unit; i-- ){
			if( wzl_read_byte() ){
				ret |= bitmask[i];
			}
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function wzl_write_str()                                */
/*                                                                    */
/*            args  struct LZ *in : write entry                       */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: Write the code translate the real BYTEs.      */
/*                                                                    */
/**********************************************************************/
static void wzl_write_str( struct LZ *in )
{
	int i;

	for( i = 0; i < in->len; i++ ){
		Out_char[Wp++] = StringBuff[in->s + i];
		if( Wp >= Out_size ){
			Quit = 1;
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*            function *wzl_firstadd()                                */
/*                                                                    */
/*            args  LZSTR code1  : oldcode                            */
/*                  LZSTR code2  : newcode                            */
/*                                                                    */
/*            return  struct LZ  :  write code                        */
/*                                                                    */
/*            function: Check the pre-read code.                      */
/*                                                                    */
/**********************************************************************/
static struct LZ *wzl_firstadd( LZSTR code1, LZSTR code2 )
{
	static struct LZ tmp;
	struct LZ *ret;

	ret = wzl_search_table( code1 );
	tmp.s = wzl_buff_strdup( &StringBuff[ret->s], ret->len+1 );
	tmp.len = ret->len;
	if( code1 == code2 ){
		StringBuff[tmp.s + tmp.len] = StringBuff[tmp.s];
	} else {
		ret = wzl_search_table( code2 );
		StringBuff[tmp.s + tmp.len] = StringBuff[ret->s];
	}
	tmp.len++;
	return( &tmp );
}

/**********************************************************************/
/*                                                                    */
/*            function uncompress()                                   */
/*                                                                    */
/*            args  int start  : start unit                           */
/*                  unsigned char *in  : input buffer                 */
/*                  unsigned char *out : output buffer                */
/*                  int size   :  size of input                       */
/*                                                                    */
/*            return int   :  0==success  -1==error                   */
/*                                                                    */
/*            function: LZW-Decode <in> and write <out>.              */
/*                                                                    */
/**********************************************************************/
int uncompress( int start, unsigned char *in, unsigned char *out, int size, int kind )
{
	LZSTR  code, oldcode;
	struct LZ *os;
	int first;
	int ret;

	if( kind == TIFF ){
		TiffGif = TIFF;
	} else if( kind == GIF ){
		TiffGif = GIF;
	}
	if( wzl_init_table( start, in, out, size ) != 0 ){
		fprintf( stderr, "init error\n" );
		ret = -1;
		goto ex;
	}
	ret = 0; first = 1; oldcode = 0;
	while( (code = wzl_read_code()) != LZEof ){
		if( Quit ){
			fprintf( stderr, "output over-run\n" );
			ret = -1;
			goto ex;
		}
		if( first == 1 ){
			if( code != ClrCode ){
				fprintf( stderr, "not compressed %d\n", code );
				ret = -1;
				goto ex;
			}
			first = 0;
		}
		if( code == ClrCode ){
			NowTable = LZEof+1;
			Stringp = StartString;
			Unit = StartUnit;
			code = wzl_read_code();
			if( code == LZEof ){
				fprintf( stderr, "found EOF\n" );
				ret = 0;
				goto ex;
			}
			os = wzl_search_table( code );
			if( os ){
				wzl_write_str( os );
			} else {
				fprintf( stderr, "cannot happen\n" );
				ret = -1;
				goto ex;
				return( -1 );
			}
			oldcode = code;
		} else {
			if( (os = wzl_search_table( code )) ){
				wzl_write_str( os );
				if( wzl_add_table( wzl_firstadd( oldcode, code ) ) != 0 ){
					fprintf( stderr, "add table error\n" );
					ret = -1;
					goto ex;
				}
				oldcode = code;
			} else {
				os = wzl_firstadd( oldcode, oldcode );
				wzl_write_str( os );
				if( wzl_add_table( os ) != 0 ){
					fprintf( stderr, "add table error\n" );
					ret = -1;
					goto ex;
				}
				oldcode = code;
			}
		}
	}
ex:	if( ret == 0 ){
		ret = Wp;
	}
	if( StringBuff ){
		free( StringBuff );
	}
	StringBuff = NULL;
	return( ret );
}
