/**********************************************************************/
/*                                                                    */
/*             xband.c  gedo PLUG X interface program of effband      */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "plug.h"

unsigned int Upper;	 /* Upper limit to find */
unsigned int Lower;	 /* Lower limit to find */
unsigned int *Work; 	 /* work area for previewer */

struct Packet {		/* packet to send & receive between effband */
	int shmid;	/* IPC shm ID */
	int Band[256];	/* amount of each depth */
} Packet;

char *Version = { "GEDO Xplugs by K.Sugiura  ver 1.0\n" };
unsigned int *Band = Packet.Band;	/* amount of each depth */
Visual  *Vis;	/* default Visual from loaderx.c */

static int 	 Rshift, 	/* In 2byte Color, shift count for Red pixel */
		 Gshift, 	/* Shift count for Green pixel */
		 Bshift;	/* Shift count for Blue pixel */

/**********************************************************************/
/*                                                                    */
/*            function set_maskshift()                                */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: From Visual, get each Shift-Count of RGB for  */
/*                                     2 byte color class.            */
/**********************************************************************/
static void set_maskshift( void )
{
	int i;
	signed int tmp;

        tmp = (signed int)Vis->red_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Rshift = i - 8;
	tmp = (signed int)Vis->green_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Gshift = i - 16;
	tmp = (signed int)Vis->blue_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Bshift = i - 24;
}

/**********************************************************************/
/*                                                                    */
/*             function change_value()                                */
/*                                                                    */
/*             unsigned int *v : each pixel                           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Change each pixel by dividing between Lower  */
/*                     and Upper.                                     */
/**********************************************************************/
void change_value( unsigned int *v )
{
	unsigned int rgb[3];
	int i;

        rgb[0] = GET_RED(*v);
	rgb[1] = GET_GREEN(*v);
	rgb[2] = GET_BLUE(*v);

	for( i = 0; i < 3; i++ ){
		if( rgb[i] <= Lower ){
			rgb[i] = 0x00000000;
		} else if( rgb[i] >= Upper ){
			rgb[i] = 0x000000ff;
		} else {
			rgb[i] = (rgb[i] - Lower) * 256 / (Upper - Lower);
		}
	}
	*v = ((*v & MASK) | ((rgb[0] << (16 - Rshift)) & Vis->red_mask) 
	      | ((rgb[1] << (8 - Gshift)) & Vis->green_mask ) 
	      | ((rgb[2] >> Bshift) & Vis->blue_mask) );
}

/**********************************************************************/
/*                                                                    */
/*             function create_image()                                */
/*                                                                    */
/*             Display *D       : X11 Display                         */
/*             Window W         : previewer window                    */
/*             GC gc            : previewer GC                        */
/*             struct Shm *shm  : Image data                          */
/*                                                                    */
/*             return  XImage * : created XImage                      */
/*                                                                    */
/*             function: Create Image for the previewer.              */
/*                                                                    */
/**********************************************************************/
XImage * create_image( Display *D, Window W, GC gc, struct Shm *shm )
{
	XImage  *Image;
	Visual  *Vis;
	int      Dep;
	int 	 size;

	Vis = DefaultVisual( D, 0 );
	Dep = DefaultDepth( D, 0 );
	size = shm->width * shm->height * sizeof(unsigned int);
	if( (Work = (unsigned int *)malloc( size )) == NULL ){
		return( NULL );
	}
	memcpy( Work, shm->data, size );
	Image = XCreateImage( D, Vis, Dep, ZPixmap, 0, 0,
			     0, 0, 32, 0 );
	Image->data    = (char *)Work;
	Image->xoffset = 0;
	Image->depth   = Dep;
	Image->byte_order = LSBFirst;
	Image->bitmap_unit = 32;
	Image->bitmap_bit_order = LSBFirst;
	Image->bytes_per_line = shm->width * 4;
	Image->bits_per_pixel = 32;
	Image->red_mask = Vis->red_mask;
	Image->green_mask = Vis->green_mask;
	Image->blue_mask = Vis->blue_mask;
	Image->width = shm->width;
	Image->height = shm->height;
	XPutImage( D, W, gc, Image, 0, 0, 0, 0, Image->width, Image->height );
	XFlush( D );
	return( Image );
}

/**********************************************************************/
/*                                                                    */
/*             function change_image()                                */
/*                                                                    */
/*             Display *D       : X11 Display                         */
/*             Window W         : previewer window                    */
/*             GC gc            : previewer GC                        */
/*             XImage *image    : previewer Image                     */
/*             struct Shm *shm  : Image data                          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Refresh the previewer image.                 */
/*                                                                    */
/**********************************************************************/
void change_image( Display *D, Window W, GC gc, XImage *image, struct Shm *shm )
{
	int i;

	
	memcpy( Work, shm->data, image->width * image->height * sizeof(unsigned int) );
	for( i = 0; i < image->height * image->width; i++ ){
		if( MASKOFF(Work[i]) ){
			change_value( &Work[i] );
		}
	}
	XPutImage( D, W, gc, image, 0, 0, 0, 0, image->width, image->height );
	XFlush( D );
}

/**********************************************************************/
/*                                                                    */
/*             function  set_shm()                                    */
/*                                                                    */
/*             int shmid         : IPC shm ID                         */
/*                                                                    */
/*             return  struct Shm * : got Image buffer                */
/*                                                                    */
/*             function: Get the image buffer from shmid.             */
/*                                                                    */
/**********************************************************************/
struct Shm *set_shm( int shmid )
{
	struct Shm *ret;

	ret = (struct Shm *)shmat( shmid, NULL, 0 );
	if( ret == (struct Shm *)-1 ){
		return( NULL );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             int argc   : number of argments                        */
/*             char **argv: argments                                  */
/*                                                                    */
/*             return  0=Success 1=Fail                               */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
int main( int argc, char **argv )
{
	Display *D;
	Window  W;
	Window  Wmap;
	Window  Wok, Wesc;
	Window  Wimage;
	Pixmap  P;
	GC      gc;
	GC      gcy;
	GC      gcs;
	GC	gci;
	Colormap cmap;
	XColor col, c;
	XEvent event;
	int i;
	unsigned int now;
	XImage *Image;
	struct Shm *shm;


	if( read( 0, &Packet, sizeof(struct Packet) ) <= 0 ){
		fprintf( stderr, "pipe read error\n" );
		perror("" );
		return( 1 );
	}
	if( (shm = set_shm( Packet.shmid )) == NULL ){
		fprintf( stderr, "get Shm error\n" );
		perror("" );
		return( 1 );
	}
	
	D = XOpenDisplay( "" );
	Vis = DefaultVisual( D, 0 );
	set_maskshift();
	W = XCreateSimpleWindow( D, RootWindow( D, 0 ), 100, 100,
		256, 256+30, 0, BlackPixel(D,0), WhitePixel(D,0) );
	Wimage = XCreateSimpleWindow( D, RootWindow( D, 0 ), 150, 150,
		shm->width, shm->height, 0, BlackPixel(D,0), 
		WhitePixel(D,0) );
	Wmap = XCreateSimpleWindow( D, W, 0, 30,
		256, 256, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wok = XCreateSimpleWindow( D, W, 55, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wesc = XCreateSimpleWindow( D, W, 105, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	P = XCreatePixmap( D, RootWindow( D, 0 ), 
		256, 256, DefaultDepth(D,0) );
	gc = XCreateGC( D, Wmap, 0, 0 );
	gcy = XCreateGC( D, Wmap, 0, 0 );
	gcs = XCreateGC( D, W, 0, 0 );

	gci = XCreateGC( D, Wimage, 0, 0 );
	XMapWindow( D, Wimage );
	if( (Image = create_image( D, Wimage, gc, shm )) == NULL ){
		return( 1 );
	}

	XSetForeground( D, gc, WhitePixel(D,0) );
	XSetForeground( D, gcs, BlackPixel(D,0) );
	XFillRectangle( D, P, gc, 0, 0, 256, 256 );
	cmap = DefaultColormap( D, 0 );
	if( XAllocNamedColor( D, cmap, "red", &col, &c ) == 0 ){
		printf( "color:blue is not here\n" );
		c.pixel = BlackPixel(D,0);
	}
	XSetForeground( D, gc, c.pixel );
	if( XAllocNamedColor( D, cmap, "yellow", &col, &c ) == 0 ){
		printf( "color:blue is not here\n" );
		c.pixel = BlackPixel(D,0);
	}
	XSetForeground( D, gcy, c.pixel );
	XSetFunction( D, gcy, GXxor );
	for( i = 0; i < 256; i++ ){
		XDrawLine( D, P, gc, i, 256-Band[i], i, 256 );
	}
	XSelectInput( D, Wmap, ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonMotionMask | ButtonReleaseMask );
	XSelectInput( D, Wok, ExposureMask | ButtonPressMask );
	XSelectInput( D, Wesc, ExposureMask | ButtonPressMask );
	XMapSubwindows( D, W );
	XMapWindow( D, W );

	XCopyArea( D, P, Wmap, gc, 0, 0, 256, 256, 0, 0 );
	XDrawLine( D, Wmap, gcy, 0, 0, 0, 256 );
	XDrawLine( D, Wmap, gcy, 256, 0, 256, 256 );
	Upper = atoi(argv[1]); Lower = atoi(argv[2]); now = 0;

	while( 1 ){
		XNextEvent( D, &event );
		switch( event.type ){
		case Expose:
			XDrawString( D, Wok, gcs, 2, 15, " OK", 3 );
			XDrawString( D, Wesc, gcs, 2, 15, " ESC", 4 );
			XCopyArea( D, P, Wmap, gc, 0, 0, 256, 256, 0, 0 );
			XDrawLine( D, Wmap, gcy, Lower, 0, Lower, 256 );
			XDrawLine( D, Wmap, gcy, Upper, 0, Upper, 256 );
			change_image( D, Wimage, gci, Image, shm );
			break;
		case KeyPress:
			write( 1, &Upper, sizeof(Upper) );
			write( 1, &Lower, sizeof(Lower) );
			XDestroyWindow( D, W );
			XDestroyWindow( D, Wimage );
			XFreeGC( D, gc );
			XCloseDisplay( D );
			return( 0 );
			break;
		case ButtonPress:
			if( event.xbutton.window == Wok ){
				write( 1, &Upper, sizeof(Upper) );
				write( 1, &Lower, sizeof(Lower) );
				XDestroyWindow( D, W );
				XFreeGC( D, gc );
				XCloseDisplay( D );
				return( 0 );
			} else if( event.xbutton.window == Wesc ){
				XDestroyWindow( D, W );
				XFreeGC( D, gc );
				XCloseDisplay( D );
				return( -1 );
			}
			if( abs(event.xbutton.x - Upper) > abs(event.xbutton.x - Lower) ){
				XDrawLine( D, Wmap, gcy, Lower, 0, Lower, 256 );
				Lower = event.xbutton.x;
				XDrawLine( D, Wmap, gcy, Lower, 0, Lower, 256 );
				now = 0;
			} else {
				XDrawLine( D, Wmap, gcy, Upper, 0, Upper, 256 );
				Upper = event.xbutton.x;
				XDrawLine( D, Wmap, gcy, Upper, 0, Upper, 256 );
				now = 1;
			}
			break;
		case ButtonRelease:
		case MotionNotify:
			if( now == 0 ){
				XDrawLine( D, Wmap, gcy, Lower, 0, Lower, 256 );
				Lower = event.xbutton.x;
				XDrawLine( D, Wmap, gcy, Lower, 0, Lower, 256 );
			} else {
				XDrawLine( D, Wmap, gcy, Upper, 0, Upper, 256 );
				Upper = event.xbutton.x;
				XDrawLine( D, Wmap, gcy, Upper, 0, Upper, 256 );
			}
			if( event.type == ButtonRelease ){
				change_image( D, Wimage, gci, Image, shm );
			}
			break;
		}
	}
}

