/**********************************************************************/
/*                                                                    */
/*           xcolor.c   gedo PLUG X interface program                 */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*     D2: sample window shown inner of the main.                     */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <unistd.h>
#include "plug.h"

char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };

/* Display Modes */
#define MODERG  0	/* R-G axis B independent Mode */
#define MODEGB  1	/* G-B axis R independent Mode */
#define MODEBR  2	/* B-R axis G independent Mode */
#define MODEYI  3	/* Y-I axis Q independent Mode */
#define MODEIQ  4	/* I-Q axis Y independent Mode */
#define MODEQY  5	/* Q-Y axis I independent Mode */
#define MAXMODE 6	/* total number of Mode */
int Mode = MODERG;
char *ModeStr[MAXMODE] = { "RG:B", "GB:R", "BR:G", "YI:Q", "IQ:Y", "QY:I" };

/* Current Data */
unsigned int Red  = 0;	/* Now R value */
unsigned int Green = 0;	/* Now G value */
unsigned int Blue  = 0;	/* Now B value */
unsigned int yiqY  = 0;	/* Now Y value */
signed int yiqI  = 0;	/* Now I value */
signed int yiqQ  = 0;	/* Now Q value */

/* Tranfar funcs  RGB<->YIQ */
#define get_red(y,i,q) (unsigned int)((float)y+0.955*(float)(i-128)+0.622*(float)(q-128)) 
#define get_green(y,i,q) (unsigned int)((float)y-0.271*(float)(i-128)-0.647*(float)(q-128)) 
#define get_blue(y,i,q) (unsigned int)((float)y-1.109*(float)(i-128)+1.700*(float)(q-128))
#define check_rgb(r)	(((int)r<0)?(0):(r>0xff)?(0xff):(r))

#define get_y(r,g,b) (unsigned int)(0.299*(float)r+0.587*(float)g+0.114*(float)b) 
#define get_i(r,g,b) (unsigned int)(0.596*(float)r-0.274*(float)g-0.322*(float)b+128.0)
#define get_q(r,g,b) (unsigned int)(0.211*(float)r-0.522*(float)g+0.311*(float)b+128.0)

Display *D;	/* X11 Display */
Visual  *Vis;	/* default Visual from loaderx.c */

static int 	 Rshift, 	/* In 2byte Color, shift count for Red pixel */
		 Gshift, 	/* Shift count for Green pixel */
		 Bshift;	/* Shift count for Blue pixel */


/**********************************************************************/
/*                                                                    */
/*            function set_maskshift()                                */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: From Visual, get each Shift-Count of RGB for  */
/*                                     2 byte color class.            */
/**********************************************************************/
static void set_maskshift( void )
{
	int i;
	signed int tmp;

        tmp = (signed int)Vis->red_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Rshift = i - 8;
	tmp = (signed int)Vis->green_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Gshift = i - 16;
	tmp = (signed int)Vis->blue_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Bshift = i - 24;
}

/**********************************************************************/
/*                                                                    */
/*             function set_xyvalue()                                 */
/*                                                                    */
/*             XEvent *event  : Mouse selection  in X-Y area          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: From Mouse button position, calc Now selected*/
/*                      Color value.                                  */
/**********************************************************************/
void set_xyvalue( XEvent *event )
{
	switch( Mode ){
	case MODERG:
		Red = event->xbutton.y;
		Green = event->xbutton.x;
		break;
	case MODEGB:
		Green = event->xbutton.y;
		Blue = event->xbutton.x;
		break;
	case MODEBR:
		Blue = event->xbutton.y;
		Red = event->xbutton.x;
		break;
	case MODEYI:
		yiqY = event->xbutton.y;
		yiqI = event->xbutton.x;
		Red = get_red( yiqY, yiqI, yiqQ );
		Green = get_green( yiqY, yiqI, yiqQ );
		Blue = get_blue( yiqY, yiqI, yiqQ );
		Red = check_rgb( Red );
		Green = check_rgb( Green );
		Blue = check_rgb( Blue );
		break;
	case MODEIQ:
		yiqI = event->xbutton.y;
		yiqQ = event->xbutton.x;
		Red = get_red( yiqY, yiqI, yiqQ );
		Green = get_green( yiqY, yiqI, yiqQ );
		Blue = get_blue( yiqY, yiqI, yiqQ );
		Red = check_rgb( Red );
		Green = check_rgb( Green );
		Blue = check_rgb( Blue );
		break;
	case MODEQY:
		yiqQ = event->xbutton.y;
		yiqY = event->xbutton.x;
		Red = get_red( yiqY, yiqI, yiqQ );
		Green = get_green( yiqY, yiqI, yiqQ );
		Blue = get_blue( yiqY, yiqI, yiqQ );
		Red = check_rgb( Red );
		Green = check_rgb( Green );
		Blue = check_rgb( Blue );
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_zvalue()                                  */
/*                                                                    */
/*             XEvent *event  : Mouse selection  in Z area            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: From Mouse button position, calc Now selected*/
/*                      Color value.                                  */
/**********************************************************************/
void set_zvalue( XEvent *event )
{
	switch( Mode ){
	case MODERG:
		Blue = event->xbutton.y;
		break;
	case MODEGB:
		Red = event->xbutton.y;
		break;
	case MODEBR:
		Green = event->xbutton.y;
		break;
	case MODEYI:
		yiqQ = event->xbutton.y;
		Red = get_red( yiqY, yiqI, yiqQ );
		Green = get_green( yiqY, yiqI, yiqQ );
		Blue = get_blue( yiqY, yiqI, yiqQ );
		Red = check_rgb( Red );
		Green = check_rgb( Green );
		Blue = check_rgb( Blue );
		break;
	case MODEIQ:
		yiqY = event->xbutton.y;
		Red = get_red( yiqY, yiqI, yiqQ );
		Green = get_green( yiqY, yiqI, yiqQ );
		Blue = get_blue( yiqY, yiqI, yiqQ );
		Red = check_rgb( Red );
		Green = check_rgb( Green );
		Blue = check_rgb( Blue );
		break;
	case MODEQY:
		yiqI = event->xbutton.y;
		Red = get_red( yiqY, yiqI, yiqQ );
		Green = get_green( yiqY, yiqI, yiqQ );
		Blue = get_blue( yiqY, yiqI, yiqQ );
		Red = check_rgb( Red );
		Green = check_rgb( Green );
		Blue = check_rgb( Blue );
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_image()                                   */
/*                                                                    */
/*             XImage *i1  : Image of X-Y area                        */
/*             XImage *i2  : Image of Z area                          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Accoding to  Mode, set X-Y and Z area image. */
/*                                                                    */
/**********************************************************************/
void set_image( XImage *i1, XImage *i2 )
{
	int i, j;
	unsigned int val;
	unsigned int *p;
	unsigned int red, green, blue;

	switch( Mode ){
	case MODERG:
		p = (unsigned int *)i1->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 256; j++, p++ ){
				val = ((i << 16) | (j << 8) | Blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		p = (unsigned int *)i2->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 60; j++, p++ ){
				val = ((Red << 16) | (Green << 8) | i);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		break;
	case MODEGB:
		p = (unsigned int *)i1->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 256; j++, p++ ){
				val = ((Red << 16) | (i << 8) | j);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		p = (unsigned int *)i2->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 60; j++, p++ ){
				val = ((i << 16) | (Green << 8) | Blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		break;
	case MODEBR:
		p = (unsigned int *)i1->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 256; j++, p++ ){
				val = ((j << 16) | (Green << 8) | i);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		p = (unsigned int *)i2->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 60; j++, p++ ){
				val = ((Red << 16) | (i << 8) | Blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		break;
	case MODEYI:
		p = (unsigned int *)i1->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 256; j++, p++ ){
				red = get_red( i, j, yiqQ );
				green = get_green( i, j, yiqQ );
				blue = get_blue( i, j, yiqQ );
				red = check_rgb( red );
				green = check_rgb( green );
				blue = check_rgb( blue );
				val = ((red << 16) | (green << 8) | blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		p = (unsigned int *)i2->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 60; j++, p++ ){
				red = get_red( yiqY, yiqI, i );
				green = get_green( yiqY, yiqI, i );
				blue = get_blue( yiqY, yiqI, i );
				red = check_rgb( red );
				green = check_rgb( green );
				blue = check_rgb( blue );
				val = ((red << 16) | (green << 8) | blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		break;
	case MODEIQ:
		p = (unsigned int *)i1->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 256; j++, p++ ){
				red = get_red( yiqY, i, j );
				green = get_green( yiqY, i, j );
				blue = get_blue( yiqY, i, j );
				red = check_rgb( red );
				green = check_rgb( green );
				blue = check_rgb( blue );
				val = ((red << 16) | (green << 8) | blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		p = (unsigned int *)i2->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 60; j++, p++ ){
				red = get_red( i, yiqI, yiqQ );
				green = get_green( i, yiqI, yiqQ );
				blue = get_blue( i, yiqI, yiqQ );
				red = check_rgb( red );
				green = check_rgb( green );
				blue = check_rgb( blue );
				val = ((red << 16) | (green << 8) | blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		break;
	case MODEQY:
		p = (unsigned int *)i1->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 256; j++, p++ ){
				red = get_red( j, yiqI, i );
				green = get_green( j, yiqI, i );
				blue = get_blue( j, yiqI, i );
				red = check_rgb( red );
				green = check_rgb( green );
				blue = check_rgb( blue );
				val = ((red << 16) | (green << 8) | blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		p = (unsigned int *)i2->data;
		for( i = 0; i < 256; i++ ){
			for( j = 0; j < 60; j++, p++ ){
				red = get_red( yiqY, i, yiqQ );
				green = get_green( yiqY, i, yiqQ );
				blue = get_blue( yiqY, i, yiqQ );
				red = check_rgb( red );
				green = check_rgb( green );
				blue = check_rgb( blue );
				val = ((red << 16) | (green << 8) | blue);
				*p = (((val >> Rshift) & Vis->red_mask )
				      | ((val >> Gshift) & Vis->green_mask )
				      | ((val >> Bshift) & Vis->blue_mask ));
			}
		}
		break;
	}
}
	
/**********************************************************************/
/*                                                                    */
/*             function set_sampleimage()                             */
/*                                                                    */
/*             XImage *im  : Image of sample tile                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: From RGB, set now selected color to the      */
/*                      sample tile.                                  */
/**********************************************************************/
void set_sampleimage( XImage *im )
{
	int i, j;
	unsigned int val;
	unsigned int *p;

	p = (unsigned int *)im->data;
	val = ((Red << 16) | (Green << 8) | Blue);
	for( i = 0; i < 100; i++ ){
	        for( j = 0; j < 100; j++, p++ ){
		        *p =  (((val >> Rshift) & Vis->red_mask )
			      | ((val >> Gshift) & Vis->green_mask )
			      | ((val >> Bshift) & Vis->blue_mask ));
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*             function  get_ximage()                                 */
/*                                                                    */
/*             int x  : width                                         */
/*             int y  : height                                        */
/*                                                                    */
/*             return  XImage * : get XImage                          */
/*                                                                    */
/*             function: From x and y, create  XImage.                */
/*                                                                    */
/**********************************************************************/
XImage *get_ximage( int x, int y )
{
	XImage *image;
	unsigned int *work;
	Visual  *Vis;
	int      Dep;
	int     size;

	Vis = DefaultVisual( D, 0 );
	Dep = DefaultDepth( D, 0 );
	size = x * y * sizeof(unsigned int);
	if( (work = (unsigned int *)malloc( size )) == NULL ){
	       return( NULL );
	}
	image = XCreateImage( D, Vis, Dep, ZPixmap, 0, 0,
			     0, 0, 32, 0 );
	image->data    = (char *)work;
	image->xoffset = 0;
	image->depth   = Dep;
	image->byte_order = LSBFirst;
	image->bitmap_unit = 32;
	image->bitmap_bit_order = LSBFirst;
	image->bytes_per_line = x * 4;
	image->bits_per_pixel = 32;
	image->red_mask = Vis->red_mask;
	image->green_mask = Vis->green_mask;
	image->blue_mask = Vis->blue_mask;
	image->width = x;
	image->height = y;
	return( image );
}

/**********************************************************************/
/*                                                                    */
/*             function change_prev()                                 */
/*                                                                    */
/*             int *x  :  prev X position->new X position             */
/*             int *y  :  prev Y position->new Y position             */
/*             int *z  :  prev Z position->new Z position             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: If Mode button pressed, Mode changes current */
/*                      XYZ axis to new Mode.                         */
/**********************************************************************/
void change_prev( int *x, int *y, int *z )
{
	int r, g, b;

	switch( Mode ){
	case MODERG:
		r = get_red( *x, *z, *y );
		g = get_green( *x, *z, *y );
		b = get_blue( *x, *z, *y );
		Red = check_rgb( r );
	        Green = check_rgb( g );
		Blue = check_rgb( b );
		*x = Green; *y = Red; *z = Blue;
		break;
	case MODEYI:
		r = get_y( *x, *z, *y );
		g = get_i( *x, *z, *y );
		b = get_q( *x, *z, *y );
		yiqY = check_rgb( r );
		yiqI = check_rgb( g );
		yiqQ = check_rgb( b );
		*x = yiqI; *y = yiqY; *z = yiqQ;
		break;
	default:
		r = *y;
		*y = *x;
		*x = *z;
		*z = r;
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function move_sample()                                 */
/*                                                                    */
/*             args   Window w     : main window                      */
/*                    Window sample: sample window                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: move the sample window to the inner area of  */
/*                      the main window of xcolor.                    */
/*             added D2                                               */
/**********************************************************************/
void move_sample( Window w, Window sample )
{
     int rx, ry;
     Window junkwin;
			     
     (void) XTranslateCoordinates ( D, w, RootWindow( D, 0 ), 
				    0, 0,
				    &rx, &ry, &junkwin);
     XMoveWindow( D, sample, rx + 200, ry + 20 );
     (void) XTranslateCoordinates ( D, sample, RootWindow( D, 0 ), 
				    0, 0,
				    &rx, &ry, &junkwin);
}

/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return int : 0==success  2==aborted                    */
/*                                                                    */
/*             function: X-Interface to paint plug, select paint      */
/*                      color visually.                               */
/**********************************************************************/
int main()
{
	Window  W;
	Window  W1, W2;
	Window  Wok, Wesc, Wmode, Wlabel;
	Window  Wsample;
	XImage  *Image1, *Image2, *Images;
	GC      gc;
	GC	gcs;
	GC      gcy;
	Colormap cmap;
	XColor col, c;
	XEvent event;
	KeySym keysym;
	XComposeStatus compose;
	char buf[10];
	int  shift, flag;
	XTextProperty win_name;	
	XClassHint  class;
	char *argv[1];
	int prevx, prevy, prevz;
	int tmpx, tmpy, tmpz;
/* add D2 */
	int first = 1;

	D = XOpenDisplay( "" );
	if( D == NULL ){
		printf( "error:X11 cannot avail" );
		return( 1 );
	}
	Vis = DefaultVisual( D, 0 );
	set_maskshift();
	W = XCreateSimpleWindow( D, RootWindow( D, 0 ), 100, 100,
		10+256+10+60+10, 30+256+10, 0, BlackPixel(D,0), WhitePixel(D,0) );
	W1 = XCreateSimpleWindow( D, W, 10, 30,
		256, 256, 1, BlackPixel(D,0), WhitePixel(D,0) );
	W2 = XCreateSimpleWindow( D, W, 10+256+10, 30,
		60, 256, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wok = XCreateSimpleWindow( D, W, 10, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wesc = XCreateSimpleWindow( D, W, 60, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wmode = XCreateSimpleWindow( D, W, 110, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wlabel = XCreateSimpleWindow( D, W, 160, 5,
		60, 20, 0, BlackPixel(D,0), WhitePixel(D,0) );
	Wsample = XCreateSimpleWindow( D, RootWindow( D, 0 ), 150, 150,
		100, 100, 1, BlackPixel(D,0), WhitePixel(D,0) );
	gc = XCreateGC( D, W, 0, 0 );
	gcy = XCreateGC( D, W, 0, 0 );
	gcs = XCreateGC( D, Wsample, 0, 0 );

	XSetForeground( D, gc, BlackPixel(D,0) );
	XSetBackground( D, gc, WhitePixel(D,0) );
	cmap = DefaultColormap( D, 0 );
	if( XAllocNamedColor( D, cmap, "yellow", &col, &c ) == 0 ){
		fprintf( stderr, "color:yellow is not here\n" );
		c.pixel = WhitePixel(D,0);
	}
	XSetForeground( D, gcy, c.pixel );
	XSetFunction( D, gcy, GXxor );

	Image1 = get_ximage( 256, 256 ); 
	Image2 = get_ximage( 60, 256 );
	Images = get_ximage( 100, 100 );

	XSelectInput( D, W, ExposureMask );
	XSelectInput( D, Wsample, ExposureMask );
	XSelectInput( D, W1, ButtonPressMask | ButtonMotionMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask );
	XSelectInput( D, W2, ButtonPressMask | ButtonMotionMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask );
	XSelectInput( D, Wok, ButtonPressMask );
	XSelectInput( D, Wesc, ButtonPressMask );
	XSelectInput( D, Wmode, ButtonPressMask );

	win_name.value = (unsigned char *)"xcolor";
	win_name.encoding = XA_STRING;
	win_name.format = 8;
	win_name.nitems = strlen( win_name.value );
	class.res_name = NULL;
	class.res_class = "Xcolor";
	argv[0] = NULL;
	XSetWMProperties( D, W, &win_name, NULL, argv, 0, 
			 NULL, NULL, &class );
	XSetWMProperties( D, Wsample, &win_name, NULL, argv, 0, 
			 NULL, NULL, &class );

	XMapSubwindows( D, W );
	XMapWindow( D, W );
	XMapWindow( D, Wsample );

	set_image( Image1, Image2 );
	XPutImage( D, W1, gc, Image1, 0, 0, 0, 0, 256, 256 );
	XFlush( D );
	XPutImage( D, W2, gc, Image2, 0, 0, 0, 0, 60, 256 );
	XFlush( D );
	set_sampleimage( Images );
	XPutImage( D, Wsample, gcs, Images, 0, 0, 0, 0, 100, 100 );
	XFlush( D );

	prevx = prevy = prevz = 0;
	XDrawLine( D, W1, gcy, prevx, 0, prevx, 256 );
	XDrawLine( D, W1, gcy, 0, prevy, 256, prevy );
	XDrawLine( D, W2, gcy, 0, prevz, 256, prevz );
	XFlush( D );

	shift = tmpx = tmpy = tmpz = 0;

	while( 1 ){
		XNextEvent( D, &event );
		switch( event.type ){
		case Expose:
			if( event.xexpose.count != 0 ){
				break;
			}
			if( event.xexpose.window == Wsample ){
				XPutImage( D, Wsample, gcs, Images, 0, 0, 0, 0, 100, 100 );
				XFlush( D );
				break;
			}
/* add D2 */
			if( first ){
			     first = 0;
			     move_sample( W, Wsample );
			}
/* add D2 end */
		expose:
			XDrawString( D, Wok, gc, 2, 15, " OK", 3 );
			XDrawString( D, Wesc, gc, 2, 15, " ESC", 4 );
			XDrawString( D, Wmode, gc, 2, 15, "MODE", 4 );
			XDrawImageString( D, Wlabel, gc, 2, 15, ModeStr[Mode], 4 );
			XPutImage( D, W1, gc, Image1, 0, 0, 0, 0, 256, 256 );
			XFlush( D );
			XPutImage( D, W2, gc, Image2, 0, 0, 0, 0, 60, 256 );
			XFlush( D );
			XDrawLine( D, W1, gcy, prevx, 0, prevx, 256 );
			XDrawLine( D, W1, gcy, 0, prevy, 256, prevy );
			XDrawLine( D, W2, gcy, 0, prevz, 256, prevz );
			XPutImage( D, Wsample, gcs, Images, 0, 0, 0, 0, 100, 100 );
			XFlush( D );
			break;
		case ButtonPress:
		case ButtonRelease:
		case MotionNotify:
			if( event.xbutton.window == Wok ){
				if( event.type != ButtonPress ){
					break;
				}
				XDestroyWindow( D, W );
				XFreeGC( D, gc );
				XCloseDisplay( D );
				printf( "%02x%02x%02x\n", Red, Green, Blue );
				return( 0 );
			} else if( event.xbutton.window == Wesc ){
				if( event.type != ButtonPress ){
					break;
				}
				XDestroyWindow( D, W );
				XFreeGC( D, gc );
				XCloseDisplay( D );
				printf( "abort\n" );
				return( 2 );
			} else if( event.xbutton.window == Wmode ){
				if( event.type != ButtonPress ){
					break;
				}
				if( ++Mode >= MAXMODE ){
					Mode = 0;
				}
				change_prev( &prevx, &prevy, &prevz );
				set_image( Image1, Image2 );
				goto expose;
			} else if( event.xbutton.window == W1 ){
				if( event.xbutton.x < 0 ){
					event.xbutton.x = 0;
				}
				if( event.xbutton.y < 0 ){
					event.xbutton.y = 0;
				}
				if( event.xbutton.x > 0xff ){
					event.xbutton.x = 0xff;
				}
				if( event.xbutton.y > 0xff ){
					event.xbutton.y = 0xff;
				}
				XDrawLine( D, W1, gcy, prevx, 0, prevx, 256 );
				XDrawLine( D, W1, gcy, 0, prevy, 256, prevy );
				set_xyvalue( &event );
				XDrawLine( D, W1, gcy, event.xbutton.x, 0, event.xbutton.x, 256 );
				XDrawLine( D, W1, gcy, 0, event.xbutton.y, 256, event.xbutton.y );
				prevx = event.xbutton.x;
				prevy = event.xbutton.y;
				if( event.type == ButtonRelease ){
					set_sampleimage( Images );
					XPutImage( D, Wsample, gcs, Images, 0, 0, 0, 0, 100, 100 );
					XFlush( D );
				}
			} else if( event.xbutton.window == W2 ){
				if( event.xbutton.x < 0 ){
					event.xbutton.x = 0;
				}
				if( event.xbutton.y < 0 ){
					event.xbutton.y = 0;
				}
				if( event.xbutton.x > 0xff ){
					event.xbutton.x = 0xff;
				}
				if( event.xbutton.y > 0xff ){
					event.xbutton.y = 0xff;
				}
				XDrawLine( D, W2, gcy, 0, prevz, 256, prevz );
				set_zvalue( &event );
				XDrawLine( D, W2, gcy, 0, event.xbutton.y, 256, event.xbutton.y );
				prevz = event.xbutton.y;
				if( event.type == ButtonRelease ){
					set_image( Image1, Image2 );
					set_sampleimage( Images );
					goto expose;
				}
			}
			break;
		case KeyPress:
		case KeyRelease:
			flag = 0;
			XLookupString( (XKeyEvent *)&event, buf, 10, &keysym, &compose );
			if( event.xbutton.window == W1 ){
				switch( keysym ){
				case XK_Shift_L:
				case XK_Shift_R:
					if( event.type == KeyPress ){
						shift = 1;
					} else {
						shift = 0;
					}
					break;
				case XK_Left:
				        XMoveWindow( D, W, 300, 300 );
					if( shift || (tmpx = prevx - 1) < 0 ){
						tmpx = 0;
					}
					flag = 1;
					break;
				case XK_Right:
				        XMoveWindow( D, W, 100, 300 );
					if( shift || (tmpx = prevx + 1) > 0xff ){
						tmpx = 0xff;
					}
					flag = 1;
					break;
				case XK_Up:
					if( shift || (tmpy = prevy - 1) < 0 ){
						tmpy = 0;
					}
					flag = 1;
					break;
				case XK_Down:
					if( shift || (tmpy = prevy + 1) > 0xff ){
						tmpy = 0xff;
					}
					flag = 1;
					break;
				}
				if( flag == 0 ){
					break;
				}
				XDrawLine( D, W1, gcy, prevx, 0, prevx, 256 );
				XDrawLine( D, W1, gcy, 0, prevy, 256, prevy );
				prevx = event.xbutton.x = tmpx;
				prevy = event.xbutton.y = tmpy;
				set_xyvalue( &event );
				XDrawLine( D, W1, gcy, prevx, 0, event.xbutton.x, 256 );
				XDrawLine( D, W1, gcy, 0, prevy, 256, event.xbutton.y );
				if( event.type == KeyRelease ){
					set_sampleimage( Images );
					XPutImage( D, Wsample, gcs, Images, 0, 0, 0, 0, 100, 100 );
					XFlush( D );
				}
			} else if( event.xbutton.window == W2 ){
				switch( keysym ){
				case XK_Shift_L:
				case XK_Shift_R:
					if( event.type == KeyPress ){
						shift = 1;
					} else {
						shift = 0;
					}
					break;
				case XK_Up:
					if( shift || (tmpz = prevz - 1) < 0 ){
						tmpz = 0;
					}
					break;
					flag = 1;
				case XK_Down:
					if( shift || (tmpz = prevz + 1) > 0xff ){
						tmpz = 0xff;
					}
					flag = 1;
					break;
				}
				if( flag == 0 ){
					break;
				}
				XDrawLine( D, W2, gcy, 0, prevz, 256, prevz );
				prevz = event.xbutton.y = tmpz;
				set_zvalue( &event );
				XDrawLine( D, W2, gcy, 0, prevz, 256, prevz );
				if( event.type == KeyRelease ){
					set_image( Image1, Image2 );
					set_sampleimage( Images );
					goto expose;
				}
			}
			break;
		}
	}
}
