/**********************************************************************/
/*                                                                    */
/*             xgamma.c gedo PLUG X interface program of effband      */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "plug.h"

#define M_RGB  0
#define M_YIQ  1
int Mode = M_RGB;			/* Mode for RGB or YCbCr */


Window  Wred, Wgreen, Wblue;		/* Windows for color button */
Display *D;				/* X11 Display */
GC      gcy;				/* GC for draw spline */

struct fPosition {	/* for call spline */
	float x;
	float y;
};

struct fPosition *spline_line( struct fPosition *pos, int dim, int num, int pat );

unsigned int *Work; 	 /* work area for previewer */

static unsigned char Xpos[5]		/* X axis point 5 */
 = { 0x00, 0x40, 0x80, 0xc0, 0xff };

char *Version
 = { "GEDO Xplugs by K.Sugiura  ver 1.0\n" };
Visual  *Vis;			/* default Visual  */

static int 	 Rshift, 	/* In 2byte Color, shift count for Red pixel */
		 Gshift, 	/* Shift count for Green pixel */
		 Bshift;	/* Shift count for Blue pixel */

static unsigned char RGB[3][256];   /* Translate array for each RGB or YIQ */
static unsigned char Node[3][5];    /* selected point XY */

char *RGBStr[2][3] = {		    /* show string on button for each Mode */
{ " RED ", "GREEN", "BLUE " },
{ "  Y  ", " Cr  ", " Cb  " } };

/*****************   translation macro ********************************/
#define get_y(r,g,b) (int)(0.299*(float)r+0.587*(float)g+0.114*(float)b) 
#define get_i(r,g,b) (int)(0.596*(float)r-0.274*(float)g-0.322*(float)b+128.0)
#define get_q(r,g,b) (int)(0.211*(float)r-0.522*(float)g+0.311*(float)b+128.0)

#define get_red(y,i,q) (int)((float)y+0.955*(float)(i-128)+0.622*(float)(q-128)) 
#define get_green(y,i,q) (int)((float)y-0.271*(float)(i-128)-0.647*(float)(q-128)) 
#define get_blue(y,i,q) (int)((float)y-1.109*(float)(i-128)+1.700*(float)(q-128))

#define ajust(p)	((p<0)?(0):((p>255)?(255):p))

/**********************************************************************/
/*                                                                    */
/*            function flush_result()                                 */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Flush out the result string to stdout,        */
/*                     recieve effgamma by pipe.                      */
/**********************************************************************/
void flush_result( void )
{
	if( Mode == M_RGB ){
		printf( "r=%02x:%02x:%02x:%02x:%02x,",  Node[0][0], Node[0][1], Node[0][2], Node[0][3], Node[0][4] );
		printf( "g=%02x:%02x:%02x:%02x:%02x,",  Node[1][0], Node[1][1], Node[1][2], Node[1][3], Node[1][4] );
		printf( "b=%02x:%02x:%02x:%02x:%02x\n", Node[2][0], Node[2][1], Node[2][2], Node[2][3], Node[2][4] );
	} else {
		printf( "y=%02x:%02x:%02x:%02x:%02x,",  Node[0][0], Node[0][1], Node[0][2], Node[0][3], Node[0][4] );
		printf( "Cr=%02x:%02x:%02x:%02x:%02x,",  Node[1][0], Node[1][1], Node[1][2], Node[1][3], Node[1][4] );
		printf( "Cb=%02x:%02x:%02x:%02x:%02x\n", Node[2][0], Node[2][1], Node[2][2], Node[2][3], Node[2][4] );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function set_maskshift()                                */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: From Visual, get each Shift-Count of RGB for  */
/*                                     2 byte color class.            */
/**********************************************************************/
static void set_maskshift( void )
{
	int i;
	signed int tmp;

        tmp = (signed int)Vis->red_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Rshift = i - 8;
	tmp = (signed int)Vis->green_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Gshift = i - 16;
	tmp = (signed int)Vis->blue_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Bshift = i - 24;
}


/**********************************************************************/
/*                                                                    */
/*            function calc_yiqcolor()                                */
/*                                                                    */
/*            args  unsigned int *rgb : RGB pixel                     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Translate a pixel by trans array for YIQ Mode.*/
/*                                                                    */
/**********************************************************************/
void calc_yiqcolor( unsigned int *rgb )
{
	int y, i ,q;

	y = get_y( rgb[0], rgb[1], rgb[2] );
	i = get_i( rgb[0], rgb[1], rgb[2] );
	q = get_q( rgb[0], rgb[1], rgb[2] );
	y = RGB[0][ajust(y)];
	i = RGB[1][ajust(i)];
	q = RGB[2][ajust(q)];
	rgb[0] = ajust(get_red(y,i,q));
	rgb[1] = ajust(get_green(y,i,q));
	rgb[2] = ajust(get_blue(y,i,q));
}

/**********************************************************************/
/*                                                                    */
/*            function change_value()                                 */
/*                                                                    */
/*            args  unsigned int *v : each pixel                      */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Translate pixel by the gamma-effect for each  */
/*                     Mode.                                          */
/**********************************************************************/
void change_value( unsigned int *v )
{
	unsigned int rgb[3];
	int i;

        rgb[0] = GET_RED(*v);
	rgb[1] = GET_GREEN(*v);
	rgb[2] = GET_BLUE(*v);

	if( Mode == M_RGB ){
		for( i = 0; i < 3; i++ ){
			rgb[i] = RGB[i][rgb[i]];
		}
	} else {
		calc_yiqcolor( rgb ); 
	}
	*v = ((*v & MASK) | ((rgb[0] << (16 - Rshift)) & Vis->red_mask) 
	      | ((rgb[1] << (8 - Gshift)) & Vis->green_mask ) 
	      | ((rgb[2] >> Bshift) & Vis->blue_mask) );
}

/**********************************************************************/
/*                                                                    */
/*             function create_image()                                */
/*                                                                    */
/*             Window W         : previewer window                    */
/*             GC gc            : previewer GC                        */
/*             struct Shm *shm  : Image data                          */
/*                                                                    */
/*             return  XImage * : created XImage                      */
/*                                                                    */
/*             function: Create Image for the previewer.              */
/*                                                                    */
/**********************************************************************/
XImage * create_image( Window W, GC gc, struct Shm *shm )
{
	XImage  *Image;
	Visual  *Vis;
	int      Dep;
	int 	 size;

	Vis = DefaultVisual( D, 0 );
	Dep = DefaultDepth( D, 0 );
	size = shm->width * shm->height * sizeof(unsigned int);
	if( (Work = (unsigned int *)malloc( size )) == NULL ){
		return( NULL );
	}
	memcpy( Work, shm->data, size );
	Image = XCreateImage( D, Vis, Dep, ZPixmap, 0, 0,
			     0, 0, 32, 0 );
	Image->data    = (char *)Work;
	Image->xoffset = 0;
	Image->depth   = Dep;
	Image->byte_order = LSBFirst;
	Image->bitmap_unit = 32;
	Image->bitmap_bit_order = LSBFirst;
	Image->bytes_per_line = shm->width * 4;
	Image->bits_per_pixel = 32;
	Image->red_mask = Vis->red_mask;
	Image->green_mask = Vis->green_mask;
	Image->blue_mask = Vis->blue_mask;
	Image->width = shm->width;
	Image->height = shm->height;
	XPutImage( D, W, gc, Image, 0, 0, 0, 0, Image->width, Image->height );
	XFlush( D );
	return( Image );
}


/**********************************************************************/
/*                                                                    */
/*            function value2spline()                                 */
/*                                                                    */
/*            args  unsigned char *y  : selected Node                 */
/*                  char *val         : result trans array            */
/*                                                                    */
/*            return int   : 0==success  -1==spline fail              */
/*                                                                    */
/*            function: Create trans array from the selected Node.    */
/*                                                                    */
/**********************************************************************/
int value2spline( unsigned char *y, char *val )
{
	struct fPosition xy[64];
	struct fPosition *ret;
	int i, j;
	float a, b;

	for( i = 0; i < 5; i++ ){
		xy[i].x = (float)Xpos[i];
		xy[i].y = (float)y[i];
	}
	if( (ret = spline_line( xy, 3, 5, 8 )) == NULL ){
		printf( "error:spline error" );
		return( -1 );
	}

	a = (ret[1].y - ret[0].y) / (ret[1].x - ret[0].x);
	b = (ret[0].y * ret[1].x - ret[1].y * ret[0].x )
		/ (ret[1].x - ret[0].x);
	for( i = j = 0; i < 256; i++ ){
		if( (float)i >= ret[j+1].x ){
			j++;
			a = (ret[j+1].y - ret[j].y) / (ret[j+1].x - ret[j].x);
			b = (ret[j].y * ret[j+1].x - ret[j+1].y * ret[j].x )
				/ (ret[j+1].x - ret[j].x);
		}
		val[i] = (unsigned char)(a * (float)i + b);
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function change_image()                                */
/*                                                                    */
/*             Display *D       : X11 Display                         */
/*             Window W         : previewer window                    */
/*             GC gc            : previewer GC                        */
/*             XImage *image    : previewer Image                     */
/*             struct Shm *shm  : Image data                          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Refresh the previewer image.                 */
/*                                                                    */
/**********************************************************************/
void change_image( Display *D, Window W, GC gc, XImage *image, struct Shm *shm )
{
	int i;

	
	memcpy( Work, shm->data, image->width * image->height * sizeof(unsigned int) );
	for( i = 0; i < image->height * image->width; i++ ){
		if( MASKOFF(Work[i]) ){
			change_value( &Work[i] );
		}
	}
	XPutImage( D, W, gc, image, 0, 0, 0, 0, image->width, image->height );
	XFlush( D );
}

/**********************************************************************/
/*                                                                    */
/*             function  set_shm()                                    */
/*                                                                    */
/*             int shmid         : IPC shm ID                         */
/*                                                                    */
/*             return  struct Shm * : got Image buffer                */
/*                                                                    */
/*             function: Get the image buffer from shmid.             */
/*                                                                    */
/**********************************************************************/
struct Shm *set_shm( int shmid )
{
	struct Shm *ret;

	ret = (struct Shm *)shmat( shmid, NULL, 0 );
	if( ret == (struct Shm *)-1 ){
		return( NULL );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function get_hexval()                                   */
/*                                                                    */
/*            args  char *str  :  argment string                      */
/*                  unsigned char *val : Node selected                */
/*                                                                    */
/*            return int : number of points                           */
/*                                                                    */
/*            function: Translate init string to the node selected.   */
/*                                                                    */
/**********************************************************************/
int get_hexval( char *str, unsigned char *val )
{
	int num;
	int v;

	for( num = v = 0; *str; str++ ){
		if( isdigit(*str) ){
			v = v * 16 + *str - '0';
		} else if( isxdigit(*str) ){
			v = v * 16 + toupper(*str) - 'A' + 10;
		} else {
			*val++ = v;
			v = 0;
			num++;
		}
	}
	*val = v;
	return( ++num );
}

/**********************************************************************/
/*                                                                    */
/*            function first_draw_spline()                            */
/*                                                                    */
/*            args  Window w  : Window of the spline map              */
/*                  int key   : selected element of the color         */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: From current Node, draw the spline of gamma   */
/*                     and create trans array.(not vanish prev)       */
/**********************************************************************/
void first_draw_spline( Window w, int key )
{
	int i;

	if( value2spline( Node[key], RGB[key] ) != 0 ){
		return;
	}
	for( i = 0; i < 256; i++ ){
		XDrawPoint( D, w, gcy, i, 255 - RGB[key][i] );
	}
	for( i = 0; i < 5; i++ ){
		XDrawArc( D, w, gcy, Xpos[i] - 5, 255 - Node[key][i] - 5, 10, 10, 0, 64 * 360 );
	}
	XFlush(D);
}

/**********************************************************************/
/*                                                                    */
/*            function draw_spline()                                  */
/*                                                                    */
/*            args  Window w  : Window of the spline map              */
/*                  int key   : selected element of the color         */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: From current Node, draw the spline of gamma   */
/*                     and create trans array.(vanish prev)           */
/**********************************************************************/
void draw_spline( Window w, int key )
{
	int i;

	for( i = 0; i < 256; i++ ){
		XDrawPoint( D, w, gcy, i, 255 - RGB[key][i] );
	}
	if( value2spline( Node[key], RGB[key] ) != 0 ){
		return;
	}
	for( i = 0; i < 256; i++ ){
		XDrawPoint( D, w, gcy, i, 255 - RGB[key][i] );
	}
	XFlush(D);
}


/**********************************************************************/
/*                                                                    */
/*            function draw_colorname()                               */
/*                                                                    */
/*            args  int key : now color correct                       */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Mark-up current button represent color.       */
/*                                                                    */
/**********************************************************************/
void draw_colorname( int key )
{
	switch( key ){
	case 0:
		XDrawString( D, Wred, gcy, 2, 15, RGBStr[Mode][key], 5 );
		break;
	case 1:
		XDrawString( D, Wgreen, gcy, 2, 15, RGBStr[Mode][key], 5 );
		break;
	case 2:
		XDrawString( D, Wblue, gcy, 2, 15, RGBStr[Mode][key], 5 );
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function prepare_map()                                  */
/*                                                                    */
/*            args  Window w  : Window to show gamma spline           */
/*                  GC white  : draw white GC                         */
/*                  GC red    : draw red GC                           */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Prepare the canvas to draw spline.            */
/*                                                                    */
/**********************************************************************/
void prepare_map( Window w, GC white, GC red )
{
	int i;

	XFillRectangle( D, w, white, 0, 0, 256, 256 );
	XDrawLine( D, w, red, 0, 255, 255, 0 );
	for( i = 1; i <= 3; i++ ){
		XDrawLine( D, w, red, i * 0x40, 0, i * 0x40, 255 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             int argc   : number of argments                        */
/*             char **argv: argments                                  */
/*                                                                    */
/*             return  0=Success 1=Fail                               */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
int main( int argc, char **argv )
{
	Window  W;		/* parent Window */
	Window  Wmap;		/* Window to draw spline */
	Window  Wok, Wesc;	/* [OK] and [ESC] button */
	Window  Wimage;		/* Previewer window */

	GC      gc;		/* GC for draw default */
	GC      gcs;		/* GC for draw button */
	GC	gci;		/* GC for draw previewer Image */
	GC	gcr;		/* GC for draw red line */

	Colormap cmap;		/* X11 colormap */
	XColor col, c;		/* tmporary color pixel */

	XEvent event;		/* evnet */

	int shmid;		/* IPC shm ID from effgamma */
	struct Shm *shm;	/* create from shmid */

	XImage *Image;		/* previewer image */
	int rgbkey;		/* now correct RED/GREEN/BLUE or Y/Cr/Cb */
	int nownode;		/* selected node */

        /* args evaluate */

	if( argc != 6 ){
		printf( "error:not enough argments\n" );
		return( 1 );
	}

	if( strcmp( argv[2], "RGB" ) == 0 ){
		Mode = M_RGB;
	} else if( strcmp( argv[2], "YIQ" ) == 0 ){
		Mode = M_YIQ;
	} else {
		printf( "error:unknown mode %s\n", argv[2] );
		return( 1 );
	}
	if( get_hexval( argv[3], Node[0] ) != 5 ){
		printf( "error:need 5 points for red\n" );
		return( 1 );
	}
	if( value2spline( Node[0], RGB[0] ) == -1 ){
		return( 1 );
	}
	if( get_hexval( argv[4], Node[1] ) != 5 ){
		printf( "error:need 5 points for green\n" );
		return( -1 );
	}
	if( value2spline( Node[1], RGB[1] ) == -1 ){
		return( 1 );
	}
	if( get_hexval( argv[5], Node[2] ) != 5 ){
		printf( "error:need 5 points for blue\n" );
		return( 1 );
	}
	if( value2spline( Node[2], RGB[2] ) == -1 ){
		return( 1 );
	}
	shmid = atoi( argv[1] );
	if( (shm = set_shm( shmid )) == NULL ){
		printf( "error:get Shm error\n" );
		return( 1 );
	}

	if( shm->height == 0 && shm->width == 0 ){
		printf( "error:image is not loaded yet" );
		return( 1 );
	}
	/* prepare windows */

	D = XOpenDisplay( "" );
	if( D == NULL ){
		printf( "error:X11 cannot start\n" );
		return( 1 );
	}
	Vis = DefaultVisual( D, 0 );
	set_maskshift();
	W = XCreateSimpleWindow( D, RootWindow( D, 0 ), 100, 100,
		256+2, 256+30+2, 0, BlackPixel(D,0), WhitePixel(D,0) );
	Wimage = XCreateSimpleWindow( D, RootWindow( D, 0 ), 150, 150,
		shm->width, shm->height, 0, BlackPixel(D,0), 
		WhitePixel(D,0) );
	Wmap = XCreateSimpleWindow( D, W, 0, 30,
		256, 256, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wok = XCreateSimpleWindow( D, W, 5, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wesc = XCreateSimpleWindow( D, W, 55, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wred = XCreateSimpleWindow( D, W, 105, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wgreen = XCreateSimpleWindow( D, W, 155, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wblue = XCreateSimpleWindow( D, W, 205, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	gc = XCreateGC( D, Wmap, 0, 0 );
	gcy = XCreateGC( D, Wmap, 0, 0 );
	gcr = XCreateGC( D, Wmap, 0, 0 );
	gcs = XCreateGC( D, W, 0, 0 );

	gci = XCreateGC( D, Wimage, 0, 0 );
	XMapWindow( D, Wimage );
	if( (Image = create_image( Wimage, gc, shm )) == NULL ){
		return( EMALLOC );
	}

	XSetForeground( D, gc, WhitePixel(D,0) );
	XSetForeground( D, gcs, BlackPixel(D,0) );
	cmap = DefaultColormap( D, 0 );
	if( XAllocNamedColor( D, cmap, "red", &col, &c ) == 0 ){
		printf( "color:red is not here\n" );
		c.pixel = BlackPixel(D,0);
	}
	XSetForeground( D, gcr, c.pixel );
	XSetLineAttributes( D, gcr, 1, LineOnOffDash, CapButt, JoinMiter );
	if( XAllocNamedColor( D, cmap, "yellow", &col, &c ) == 0 ){
		printf( "color:yellow is not here\n" );
		c.pixel = BlackPixel(D,0);
	}
	XSetForeground( D, gcy, c.pixel );
	XSetFunction( D, gcy, GXxor );
	XSelectInput( D, Wmap, ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonMotionMask | ButtonReleaseMask );
	XSelectInput( D, Wok, ExposureMask | ButtonPressMask );
	XSelectInput( D, Wesc, ExposureMask | ButtonPressMask );
	XSelectInput( D, Wred, ExposureMask | ButtonPressMask );
	XSelectInput( D, Wgreen, ExposureMask | ButtonPressMask );
	XSelectInput( D, Wblue, ExposureMask | ButtonPressMask );
	XMapSubwindows( D, W );
	XMapWindow( D, W );

	/* event loop */

	rgbkey = nownode = 0;

	while( 1 ){
		XNextEvent( D, &event );
		switch( event.type ){
		case Expose:
			XDrawString( D, Wok, gcs, 2, 15, " OK", 3 );
			XDrawString( D, Wesc, gcs, 2, 15, " ESC", 4 );
			XDrawString( D, Wred, gcs, 2, 15, RGBStr[Mode][0], 5 );
			XDrawString( D, Wgreen, gcs, 2, 15, RGBStr[Mode][1], 5 );
			XDrawString( D, Wblue, gcs, 2, 15, RGBStr[Mode][2], 5 );
			draw_colorname( rgbkey );
			prepare_map( Wmap, gc, gcr );
			first_draw_spline( Wmap, rgbkey );
			change_image( D, Wimage, gci, Image, shm );
			break;
		case KeyPress:
			flush_result();
			XDestroyWindow( D, W );
			XDestroyWindow( D, Wimage );
			XFreeGC( D, gc );
			XCloseDisplay( D );
			return( 0 );
			break;
		case ButtonPress:
			if( event.xbutton.window == Wok ){
				flush_result();
				XDestroyWindow( D, W );
				XFreeGC( D, gc );
				XCloseDisplay( D );
				return( 0 );
			} else if( event.xbutton.window == Wesc ){
				printf( "abort\n" );
				XDestroyWindow( D, W );
				XFreeGC( D, gc );
				XCloseDisplay( D );
				return( 0 );
			} else if( event.xbutton.window == Wred ){
				draw_colorname( rgbkey );
				rgbkey = 0;
				draw_colorname( rgbkey );
				prepare_map( Wmap, gc, gcr );
				first_draw_spline( Wmap, rgbkey );
			} else if( event.xbutton.window == Wgreen ){
				draw_colorname( rgbkey );
				rgbkey = 1;
				draw_colorname( rgbkey );
				prepare_map( Wmap, gc, gcr );
				first_draw_spline( Wmap, rgbkey );
			} else if( event.xbutton.window == Wblue ){
				draw_colorname( rgbkey );
				rgbkey = 2;
				draw_colorname( rgbkey );
				prepare_map( Wmap, gc, gcr );
				first_draw_spline( Wmap, rgbkey );
			} else {
				nownode = (event.xbutton.x / 0x20 + 1) / 2;
				XDrawArc( D, Wmap, gcy, Xpos[nownode] - 5, 255 - Node[rgbkey][nownode] - 5, 10, 10, 0, 64 * 360 );
				Node[rgbkey][nownode] = 255 - event.xbutton.y;
				draw_spline( Wmap, rgbkey );
				XDrawArc( D, Wmap, gcy, Xpos[nownode] - 5, 255 - Node[rgbkey][nownode] - 5, 10, 10, 0, 64 * 360 );
			}
			break;
		case ButtonRelease:
		case MotionNotify:
			XDrawArc( D, Wmap, gcy, Xpos[nownode] - 5, 255 - Node[rgbkey][nownode] - 5, 10, 10, 0, 64 * 360 );
			Node[rgbkey][nownode] = 255 - event.xbutton.y;
			draw_spline( Wmap, rgbkey );
			XDrawArc( D, Wmap, gcy, Xpos[nownode] - 5, 255 - Node[rgbkey][nownode] - 5, 10, 10, 0, 64 * 360 );
			if( event.type == ButtonRelease ){
				change_image( D, Wimage, gci, Image, shm );
			}
			break;
		}
	}
}

