/**********************************************************************/
/*                                                                    */
/*              xgauss.c  gedo PLUG specified routine                 */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <stdio.h>
#ifndef PI
#ifdef M_PI
#define PI  M_PI
#endif
#endif


struct Colors {			/* vector data for each colors */
	unsigned int median;	/* median of the color */
	unsigned int Rband;	/* the band from median to */
	unsigned int Lband;	/* the band from median to */
	unsigned int diff;	/* delta~2 */
	unsigned int max;	/* real maximum of the color */
	unsigned int min;	/* real miniimum of the color */
	int map[128];		/* result of the mapping tranlation table */
};	

int GaussFlag = 0;	/* gauss or flat */
struct Colors RGB[3];
int Color;
char *RGBname[3] = { "red", "green", "blue" };
Tk_PhotoHandle phd;	/* Tk image handle */
Tk_PhotoImageBlock pblk;	/* Tk image block */

#define DIV	128

double Gauss[DIV*4];	/* gauss mapping original value */
unsigned int Ab, Ag, Ar;	/* Color divided RGB */

/**********************************************************************/
/*                                                                    */
/*            function init_gauss()                                   */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: calculate gauss mapping to prepare.           */
/*                                                                    */
/**********************************************************************/
void init_gauss( void )
{
	int i;
	double total, at;

	for( i = 0; i < DIV * 4; i++ ){
		Gauss[i] = (double)i / (double)DIV;
		Gauss[i] *= Gauss[i];
		Gauss[i] = 1.0 / sqrt(2.0*PI) * exp( -0.5 * Gauss[i] );
	}
	total = 0.0;
	for( i = 0; i < DIV * 4; i++ ){
		at = Gauss[i] / (double)DIV;
		Gauss[i] = total;
		total += at;
	}
	for( i = 0; i < DIV * 4; i++ ){
		Gauss[i] /= total * 2.0;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function remap_gauss()                                  */
/*                                                                    */
/*            args  int d2  :  delta set                              */
/*                  int *map : result of the color gauss mapping      */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: make gauss mapping for each color by real data.*/
/*                                                                    */
/**********************************************************************/
void remap_gauss( int d2, int *map )
{
	int i, j;
	double dv;
	double at;
	double lmap[128];

	dv = (double)DIV / (double)d2;
	at = 0.0;
	for( i = 0; i < 128; i++ ){
		if( at < (double)(4 * DIV) ){
			lmap[i] = Gauss[(int)at] * DIV * 2.0;
			at += dv;
		} else {
			lmap[i] = 0.50 * DIV * 2.0;
		}
	}
	for( i = j = 0; i < 127; i++ ){
		while( (double)j < lmap[i] ){
			map[j++] = i - 1;
			if( j >= 127 ){
				break;
			}
		}
	}
	map[127] = 128 - lmap[127];
}

/**********************************************************************/
/*                                                                    */
/*            function get_range()                                    */
/*                                                                    */
/*            args  struct Colors *c : each color                     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: make confirm the real range of each color.    */
/*                                                                    */
/**********************************************************************/
void get_range( struct Colors *c )
{
	int i;

	for( i = 0; i < 128; i++ ){
		if( c->map[i] == 128 ){
			break;
		}
	}
	i++;
	if( c->Rband > i ){
		c->max = i;
	} else {
		c->max = c->Rband;
	}
	if( c->Lband > i ){
		c->min = i;
	} else {
		c->min = c->Lband;
	}
	if( c->median + c->max > 255 ){
		c->max = 255 - c->median;
	}
	if( (int)c->median - (int)c->min < 0 ){
		c->min = c->median;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function get_gauss()                                    */
/*                                                                    */
/*            args  unsigned int v  : original color(each)            */
/*                  struct Colors *p : color setting                  */
/*                                                                    */
/*            return unsigned int : new color(each)                   */
/*                                                                    */
/*            function: calculate new color for each of RGB in GAUSS  */
/*                     mode.                                          */
/**********************************************************************/
unsigned int get_gauss( unsigned int v, struct Colors *p )
{
	v %= (p->min + p->max);
	if( (int)v - (int)p->min > 0 ){
		v = v - p->min;
		v = p->median + p->map[v];
	} else {
		v = p->min - v;
		v = p->median - p->map[v];
	}
	return( v );
}

/**********************************************************************/
/*                                                                    */
/*            function gauss_color()                                  */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: calculate new GAUSS color.                    */
/*                                                                    */
/**********************************************************************/
void gauss_color( void )
{
	unsigned int r;

	r = rand();

	Ab = (get_gauss(    r     & 0xff, &RGB[2] ) & 0xff);
	Ag = (get_gauss( (r >> 8) & 0xff, &RGB[1] ) << 8);
	Ar = (get_gauss( (r >>16) & 0xff, &RGB[0] ) << 16);
}

#define calc2(a)	((a)>0xff)?(0xff):(((a)<0)?(0):(a))

/**********************************************************************/
/*                                                                    */
/*            function get_value()                                    */
/*                                                                    */
/*            args  unsigned int v  : original color(each)            */
/*                  Struct Colors *p : color struct(each)             */
/*                                                                    */
/*            return unsigned int : translated color                  */
/*                                                                    */
/*            function: RANDOM case, calculate color of each RGB.     */
/*                                                                    */
/**********************************************************************/
unsigned int get_value( unsigned int v, struct Colors *p )
{
	if( v >= 0x80 ){
		if( p->Lband == 0 ){
			v = p->median;
		} else {
			v = p->median - v % p->Lband;
		}
	} else {
		if( p->Rband == 0 ){
			v = p->median;
		} else {
			v = p->median + v % p->Rband;
		}
	}
	return( calc2(v) );
}

/**********************************************************************/
/*                                                                    */
/*            function random_color()                                 */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: RANDOM case, calculate new color from old.    */
/*                                                                    */
/**********************************************************************/
void random_color( void )
{
	unsigned int r;

	r = rand();

	Ab = (get_value(    r     & 0xff, &RGB[2] ) & 0xff);
	Ag = (get_value( (r >> 8) & 0xff, &RGB[1] ) << 8);
	Ar = (get_value( (r >>16) & 0xff, &RGB[0] ) << 16);
}



/**********************************************************************/
/*                                                                    */
/*            function plot()                                         */
/*                                                                    */
/*            args  Tcl_Interp *interp : TclTk interpreter handle     */
/*                  struct Colors *c   : Now color struct             */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: display the graph of the gauss translation    */
/*                     of each RGB.                                   */
/**********************************************************************/
void plot( Tcl_Interp *interp, struct Colors *c )
{
	int i;
	int map[256];
	char buff[256];
	char *col;
	int flag;
	char *rgb;

	col = Tcl_GetVar( interp, "gf", 0 );
	if( strcmp( col, "gauss" ) == 0 ){
		flag = 1;
	} else if( strcmp( col, "flat" ) == 0 ){
		flag = 0;
	} else {
		return;
	}
	rgb = RGBname[Color];
		
	memset( map, 0, sizeof(map) );
	for( i = 0; i < 128; i++ ){
		if( c->median + c->map[i] < 256 ){
			map[c->median + c->map[i]]++;
		}
		if( c->map[i] && c->median - c->map[i] >= 0 ){
			map[c->median - c->map[i]]++;
		}
	}
	for( i = 0; i < 255; i++ ){
		if( i < c->median - c->Lband 
		   || i >= c->median + c->Rband ){
			col = "black";
		} else {
			col = rgb;
		}
		if( flag == 1 ){
			sprintf( buff, ".c create line %gc %gc %gc %gc "
				"-width 0.005c -tags lines -fill %s",
				(double)i / 20.0 + 0.1, 
				10.1 - (double)map[i] / 5.0, 
				(double)(i+1) / 20.0 + 0.1,
				10.1 - (double)map[i+1] / 5.0,
				col );
		} else {
			sprintf( buff, ".c create line %gc %gc %gc %gc "
				"-width 0.005c -tags lines -fill %s",
				(double)i / 20.0 + 0.1, 
				10.1 - 8.0, 
				(double)(i+1) / 20.0 + 0.1,
				10.1 - 8.0,
				col );
		}
		Tcl_Eval( interp, buff );
	}
}


/**********************************************************************/
/*                                                                    */
/*            function init_sample()                                  */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: create the sample image.                      */
/*                                                                    */
/**********************************************************************/
void init_sample( Tcl_Interp *interp )
{
#if TK_PHOTO_ARG == 2
	phd = Tk_FindPhoto( interp, "sampleimg" );
#elif TK_PHOTO_ARG == 1
	phd = Tk_FindPhoto( "sampleimg" );
#else
	#error "I do NOT know how many args of Tk_FindPhoto() in tk.h,"
	#error "Give -DTK_PHOTO_ARG=[1|2], please."
#endif
	Tk_PhotoGetImage( phd, &pblk );
}

/**********************************************************************/
/*                                                                    */
/*            function sample_draw()                                  */
/*                                                                    */
/*            args  Tcl_Interp *interp : TclTk interpreter handle     */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: draw the sample image by the color.           */
/*                                                                    */
/**********************************************************************/
int sample_draw( Tcl_Interp *interp )
{
	int i, j;
	unsigned char *p, *q;
	void (*do_color)(void);

	p = Tcl_GetVar( interp, "gf", 0 );
	if( strcmp( p, "gauss" ) == 0 ){
		do_color = gauss_color;
	} else if( strcmp( p, "flat" ) == 0 ){
		do_color = random_color;
	} else {
		return( TCL_ERROR );
	}	
	p = pblk.pixelPtr;
	for( i = 0; i < pblk.height; i++ ){
		q = p;
		for( j = 0; j < pblk.width; j++ ){
			(*do_color)();
			q[pblk.offset[0]] = (Ar >> 16);
			q[pblk.offset[1]] = (Ag >> 8);
			q[pblk.offset[2]] = Ab;
			q += pblk.pixelSize;
		}
		p += pblk.pitch;
	}
	p = pblk.pixelPtr; 
	Tk_PhotoPutBlock( phd, &pblk, 0, 0, pblk.width, pblk.height );
	return( TCL_OK );
}

/**********************************************************************/
/*                                                                    */
/*            function Gauss_Proc()                                   */
/*                                                                    */
/*            args  ClientData cd  : call back data, not used.        */
/*                  Tcl_Interp *interp : TclTk interpreter handle     */
/*                  int argc  :  argment number                       */
/*                  char **argv : argments                            */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: gauss command on the Tcl/Tk gauss.tcl         */
/*                                                                    */
/**********************************************************************/
int Gauss_Proc( ClientData cd, Tcl_Interp *interp, int argc, char **argv )
{
	int ave, lband, rband, diff;

	if( argc < 2 ){
		return( TCL_ERROR );
	}
	if( strcmp( argv[1], "plot" ) == 0 ){
	} else if( strcmp( argv[1], "sample" ) == 0 ){
		return( sample_draw( interp ) );
	} else {
		return( TCL_ERROR );
	}
	if( strcmp( argv[2], "red" ) == 0 ){
		Color = 0;
	} else if( strcmp( argv[2], "green" ) == 0 ){
		Color = 1;
	} else if( strcmp( argv[2], "blue" ) == 0 ){
		Color = 2;
	} else {
		return( TCL_ERROR );
	}
	if( argc <= 6 ){
		return( TCL_ERROR );
	}
	ave = atoi(argv[3]);
	lband = atoi(argv[4]);
	rband = atoi(argv[5]);
	diff = atoi(argv[6]);
	if( ave < 0 || ave >= 256 
	   || lband < 0 || lband >= 256
	   || rband < 0 || rband >= 256
	   || diff < 0 || diff >= 128 ){
		return( TCL_ERROR );
	}

	RGB[Color].median = ave;
	RGB[Color].Lband = lband;
	RGB[Color].Rband = rband;
	RGB[Color].diff = diff;
	remap_gauss( RGB[Color].diff, RGB[Color].map );
	get_range( &RGB[Color] );
	plot( interp, &RGB[Color] );
	return( TCL_OK );
}

/**********************************************************************/
/*                                                                    */
/*            function Gauss_Init()                                   */
/*                                                                    */
/*            args  Tcl_Interp *interp : TclTk interpreter handle     */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: set the command gauss to the interp           */
/*                                                                    */
/**********************************************************************/
int Gauss_Init( Tcl_Interp *interp )
{
	Tcl_CreateCommand( interp, "gauss", Gauss_Proc, (ClientData *)NULL,
			  (Tcl_CmdDeleteProc *)NULL );
	return( TCL_OK );
}

/**********************************************************************/
/*                                                                    */
/*            function set_init_value()                               */
/*                                                                    */
/*            args  Tcl_Interp *interp : TclTk interpreter handle     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: set initial value from gedo to the GUI.       */
/*                                                                    */
/**********************************************************************/
void set_init_value( Tcl_Interp *interp )
{
	char *tmp;

	tmp = Tcl_GetVar( interp, "aveR", 0 );
	RGB[0].median = atoi(tmp);
	tmp = Tcl_GetVar( interp, "aveG", 0 );
	RGB[1].median = atoi(tmp);
	tmp = Tcl_GetVar( interp, "aveB", 0 );
	RGB[2].median = atoi(tmp);

	tmp = Tcl_GetVar( interp, "RbandR", 0 );
	RGB[0].Rband = atoi(tmp);
	tmp = Tcl_GetVar( interp, "RbandG", 0 );
	RGB[1].Rband = atoi(tmp);
	tmp = Tcl_GetVar( interp, "RbandB", 0 );
	RGB[2].Rband = atoi(tmp);

	tmp = Tcl_GetVar( interp, "LbandR", 0 );
	RGB[0].Lband = atoi(tmp);
	tmp = Tcl_GetVar( interp, "LbandG", 0 );
	RGB[1].Lband = atoi(tmp);
	tmp = Tcl_GetVar( interp, "LbandB", 0 );
	RGB[2].Lband = atoi(tmp);

	tmp = Tcl_GetVar( interp, "diffR", 0 );
	RGB[0].diff = atoi(tmp);
	tmp = Tcl_GetVar( interp, "diffG", 0 );
	RGB[1].diff = atoi(tmp);
	tmp = Tcl_GetVar( interp, "diffB", 0 );
	RGB[2].diff = atoi(tmp);

	remap_gauss( RGB[0].diff, RGB[0].map );
	get_range( &RGB[0] );
	remap_gauss( RGB[1].diff, RGB[1].map );
	get_range( &RGB[1] );
	remap_gauss( RGB[2].diff, RGB[2].map );
	get_range( &RGB[2] );
}

/**********************************************************************/
/*                                                                    */
/*            function main()                                         */
/*                                                                    */
/*            args   NONE                                             */
/*                                                                    */
/*            return int : 0==success  1==error                       */
/*                                                                    */
/*            function: main routine                                  */
/*                                                                    */
/**********************************************************************/
int main()
{
	Tcl_Interp *interp;
	char buff[256];
	char *e;

	init_gauss();
	interp = Tcl_CreateInterp();
	if( Tcl_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "%s\n", interp->result );
		exit( 1 );
	}

	if( Tk_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "Tk_Init failed:%s\n", interp->result );
		exit( 1 );
	}

	if( Gauss_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "Gauss_Init failed:%s\n", interp->result );
		exit( 1 );
	}

	while( fgets( buff, sizeof(buff)-1, stdin ) ){
		if( Tcl_Eval( interp, buff ) == TCL_ERROR ){
			fprintf( stderr, "line read error\n" );
			exit( 1 );
		}
	}
	set_init_value( interp );
	if( (e = getenv("GEDO_PATH")) == NULL ){
		e = "./";
	}
/*	sprintf( buff, "%sgauss.tcl", e );*/
	sprintf( buff, "/home/gedo/c/gedo2/plug/gauss.tcl" );
	if( Tcl_EvalFile( interp, buff ) == TCL_ERROR ){
		fprintf( stderr, "gaussgui.tcl TCL error\n" );
		fprintf( stderr, "%s\n", interp->result );
		exit( 1 );
	}
	init_sample(interp);
	Tk_MainLoop();
	return( 0 );
}
