/**********************************************************************/
/*                                                                    */
/*          xstring.c   gedo PLUG X interface program                 */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Shell.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "../common.h"
#include "../loader.h"
#include "xstring.h"

char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };
/* fallback resource */
void Search( Widget w, XtPointer cd, XtPointer call );
/* prototype */
XFontStruct * show_image( void );
void save_textimage( void );

#ifndef max
#define max(p,q)	((p<q)?(q):(p))
#endif

Display  *D;	/* X11 Display */
Window   W;	/* Window to display sub-window(image) */
Pixmap P;	/* Pixmap to get window image */
GC gc;		/* GC to draw buttons */

unsigned char *ObjStr;		/* draw text */
WSTRING *Out;			/* ObjStr converted to internal code */
static char FileText[TEXTMAX];	/* Filename in text */
char SelectedFont[FONTMAX];	/* font name selected */
char SearchFont[FONTMAX]	/* font name selection start */
 = { "*" };
char *notfound[2]		/* not found message */
 = { "** not found **", NULL };

/* fallback resource to Widget <text> */
String translationsTable
 = "<Key>Return: Search()";
XtActionsRec actions[]
 = {{ "Search", (XtActionProc)Search } };

/* Toolkit Widgets */
Widget sheet;		/* all-over base formWidgetClass */
Widget text;		/* input select fontname Widget, asciiTextWidgetClass*/
Widget point;		/* input point of font at scalable font,
				 asciiTextWidgetClass*/
Widget list;		/* show list of fontname, listWidgetClass */

char PointSize[20] = { "40" };	/* point for scalable font */
int KMode = JIS;		/* KANJI Coding */
XFontStruct *Kanji;		/* font to draw KANJI text */
XFontStruct *Ascii;		/* font to draw ASCII text */
XFontStruct *Def;		/* font to draw fontname and buttons */
int AddW = 0;			/* distance between characters */
int AddH = 0;			/* distance between lines */
struct fontimage Image;		/* result image header to transact */
int MaxW = 0;			/* maximum of the font image width */

/**********************************************************************/
/*                                                                    */
/*             function fatal()                                       */
/*                                                                    */
/*             int error : error code                                 */
/*             char *str : additional message                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Write error message to seltext process.      */
/*                                                                    */
/**********************************************************************/
void fatal( int error, char *str )
{
	struct fontimage fi;

	memset( &fi, 0, sizeof(struct fontimage) );
	strcpy( fi.fontname, str );
	fi.flag = error;
	fwrite( &fi, sizeof(struct fontimage), 1, stdout );
	free( ObjStr );
	exit( 1 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Write the image of string to selstr process. */
/*                                                                    */
/**********************************************************************/
void write_image( void )
{
	char *buff;
	int ret;
	int size;

	size = Image.width * Image.height;
	if( (buff = malloc( size )) == NULL ){
		fatal( EMALLOC, "malloc error" );
	}
	Image.flag = ENOERROR;
	strcpy( Image.fontname, SelectedFont );
	if( fwrite( &Image, sizeof(struct fontimage), 1, stdout ) != 1 ){
		free( buff );
		fatal( EWRITEFILE, "" );
	}
	if( (ret = get_fontimage( &Image, buff ) ) != 0 ){
		free( buff );
		fatal( ret, "" );
	}
	if( fwrite( buff, sizeof(char), size, stdout ) != size ){
		free( buff );
		fatal( EWRITEFILE, "");
	}
	free( buff );
}


/**********************************************************************/
/*                                                                    */
/*             function check_scalable()                              */
/*                                                                    */
/*             char *str  : font name                                 */
/*                                                                    */
/*             return  int : 1==scalable font  0==not scalable font   */
/*                                                                    */
/*             function: Return the font name is either scalable or   */
/*                     not.                                           */
/**********************************************************************/
int check_scalable( char *str )
{

	for( ; *str; str++ ){
		if( str[0] == '-' && str[1] == '-' ){
			break;
		}
	}
	if( str[2] == '0' && str[4] == '0' ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function make_fontname()                               */
/*                                                                    */
/*             char *dst  :<OUT>  the fontname confirmed points       */
/*             char *src  :<IN>   scalable fontname                   */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Make the scalable fontname to confirmed point*/
/*                  fontname.                                         */
/**********************************************************************/
void make_fontname( char *dst, char *src )
{
	char *tmp;
	int scale;

	XtVaGetValues( point, XtNstring, &tmp, NULL );
	scale = atoi( tmp );
	for( ; (*dst = *src); dst++, src++ ){
		if( src[0] == '-' && src[1] == '-' ){
			break;
		}
	}
	sprintf( dst, "--%d", scale );
	src += 3;
	dst += strlen( dst );
	for( ; (*dst = *src); dst++, src++ ){
		if( src[0] == '-' && src[1] == '0' ){
			dst[1] = '*';
			src++, dst++;
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function Print()                                       */
/*                                                                    */
/*             Widget w        : Widget call this routine             */
/*             XtPointer cd    : additional data                      */
/*             XtPointer call  : selected list(fontname)              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Callback function of list Widget.            */
/*                      Do the preview of Text image, sub-window main.*/
/**********************************************************************/
void Print( Widget w, XtPointer cd, XtPointer call )
{
	XFontStruct *font;
	unsigned long white, black;
	XawListReturnStruct *st;
	Window  quit, esc, cancel;
	GC gcf, gcc;
	int dfheight, fheight, height;
	int namewidth, buttonwidth, width;
	int button1width, button2width, button3width;
	int h, wd;
	XEvent event;
	int line;

/* get selected fontname and set the font to GC gcf */
	st = call;
	if( strcmp( notfound[0], st->string ) == 0 ){
		return;
	}
	if( check_scalable( st->string ) ){
		make_fontname( SelectedFont, st->string ); 
	} else {
		strcpy( SelectedFont, st->string );
	}
	if( (font = XLoadQueryFont( D, SelectedFont )) == NULL ){
		fatal( EILLSELECTION, "font is not here" );
	}
	if( font->min_byte1 ){
		XUnloadFont( D, Kanji->fid );
		Kanji = font;
	} else {
		XUnloadFont( D, Ascii->fid );
		Ascii = font;
	}

/* prepare to make sub-window to preview */
	fheight = max(Kanji->ascent, Ascii->ascent);
	dfheight = Def ->ascent + Def->descent;
	namewidth = MERGIN + XTextWidth( Def, SelectedFont, strlen(SelectedFont) ) + MERGIN;
	button1width = XTextWidth( Def, "OK", 2 );
	button2width = XTextWidth( Def, "ESC", 3 );
	button3width = XTextWidth( Def, "CANCEL", 6 );
	buttonwidth = MERGIN + button1width + MERGIN + button2width + MERGIN + button3width + MERGIN;

/* get image size enough to show preview */
	Image.width = calc_maxwidth( Out, &line, AddW, MaxW );
	height = dfheight * 2 + fheight * line  + MERGIN * 4 + (MERGIN + AddH) * (line - 1);
	Image.height = (fheight + MERGIN + AddH) * (line+1) + MERGIN;
	width = (namewidth > Image.width)?(namewidth):(Image.width);
	width = (width > buttonwidth)?(width):(buttonwidth);

/* create sub-window */
	white = WhitePixel( D, 0 );
	black = BlackPixel( D, 0 );
	W = XCreateSimpleWindow( D, RootWindow( D, 0 ), 100, 100,
		width, height, 0, black, white );

	gc = XCreateGC( D, W, 0, 0 );

	h = MERGIN * 3 + (line - 1) * (AddH + MERGIN)+ dfheight + fheight * line;
	wd = MERGIN;
	quit = XCreateSimpleWindow( D, W, wd, h, button1width, dfheight, 1, black, white ); 
	wd += (MERGIN + button1width);
	esc = XCreateSimpleWindow( D, W, wd, h, button2width, dfheight, 1, black, white ); 
	wd += (MERGIN + button2width);
	cancel = XCreateSimpleWindow( D, W, wd, h, button3width, dfheight, 1, black, white ); 
	
	XSelectInput( D, W, ExposureMask | ButtonPressMask );

/* prepare to draw the text */
	P = XCreatePixmap( D, RootWindow( D, 0 ), 
		Image.width, Image.height, DefaultDepth(D,0) );
	gcf = XCreateGC( D, P, 0, 0 );
	gcc = XCreateGC( D, P, 0, 0 );
	XSetForeground( D, gcc, WhitePixel(D,0) );
	XSetFont( D, gcf, font->fid );
	XFillRectangle( D, P, gcc, 0, 0, Image.width, Image.height );
/* draw the font image */
	Draw_Str( gcf, Out, fheight + MERGIN + AddH, AddW, MaxW );
	XMapWindow( D, W );
	XMapSubwindows( D, W );

/* event loop */
	while( 1 ){
		XNextEvent( D, &event );
		switch( event.type ){
		case Expose:
			XDrawString( D, W, gc, MERGIN, MERGIN+Def->ascent, SelectedFont, strlen(SelectedFont) );
			XCopyArea( D, P, W, gc, 0, 0, Image.width, Image.height, 0, MERGIN + dfheight );
			XDrawString( D, quit, gc, 0, Def->ascent, "OK", 2 );
			XDrawString( D, esc, gc, 0, Def->ascent, "ESC", 3 );
			XDrawString( D, cancel, gc, 0, Def->ascent, "CANCEL", 6 );
			XFlush( D );
			break;
		case ButtonPress:
			if( event.xbutton.subwindow == quit ){
				write_image();
				free( ObjStr );
				exit( 0 );
				goto ex;
			} else if( event.xbutton.subwindow == esc ){
				fatal( EUSRBREAK, "" );
				goto ex;
			} else if( event.xbutton.subwindow == cancel ){
				goto ex;
			}
			break;
		}
	}
/* end proc */
ex:	XFreeGC( D, gcf );
	XFreeGC( D, gc );
	XFreePixmap( D, P );
	XDestroyWindow( D, W );
}


/**********************************************************************/
/*                                                                    */
/*             function Search()                                      */
/*                                                                    */
/*             Widget w        : Widget call this routine             */
/*             XtPointer cd    : additional data                      */
/*             XtPointer call  : input text(for search pattern)       */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Callback function of text Widget.            */
/*                      Get the search pattern from input, and re-    */
/*                      search the new font name pattern.             */
/**********************************************************************/
void Search( Widget w, XtPointer cd, XtPointer call )
{
	char **fontnames;
	int count;
	String tmp;

	XtVaGetValues( text, XtNstring, &tmp, NULL );
	strcpy( SearchFont, tmp );
	if( (fontnames = XListFonts( D, SearchFont, 1024, &count )) == NULL ){
		fontnames = notfound; count = 1;
	}
	XawListChange( list, fontnames, count, 0, True );
}

/**********************************************************************/
/*                                                                    */
/*             function init_arg()                                    */
/*                                                                    */
/*             int argc     : number of argments                      */
/*             char **argv  : argments                                */
/*                                                                    */
/*             return int   : 0==success  else==errorcode             */
/*                                                                    */
/*             function: Initialize data from the invoking argments.  */
/*                                                                    */
/**********************************************************************/
int init_arg( int argc, char **argv )
{
	int ret;
	struct stat statbuf;

	argv++;
	while( --argc > 0 && *argv ){
		if( strncmp( *argv, "text=", 5 ) == 0 ){
			if( (ObjStr = malloc( 256 * 3 )) == NULL ){
				return( EMALLOC );
			}
			Out = ObjStr + 256;
			strcpy( ObjStr, *argv+5 );
		} else if( strncmp( *argv, "filetext=", 9 ) == 0 ){
			strcpy( FileText, *argv + 9 );
			if( stat( FileText, &statbuf ) == -1 ){
				return( EOPENFILE );
			}
			if( (ObjStr = malloc( (statbuf.st_size+1) * 3 )) == NULL ){
				return( EMALLOC );
			}
			memset( ObjStr, 0, (statbuf.st_size+1) * 3 );
			Out = ObjStr + statbuf.st_size + 1;
			if( (ret = read_textfile( ObjStr, FileText, statbuf.st_size )) != 0 ){
				return( ret );
			}
		} else if( strncmp( *argv, "pattern=", 8 ) == 0 ){
			strcpy( SearchFont, *argv+8 );
		} else if( strncmp( *argv, "jis", 3 ) == 0 ){
			KMode = JIS;
		} else if( strncmp( *argv, "euc", 3 ) == 0 ){
			KMode = EUC;
		} else if( strncmp( *argv, "sjis", 4 ) == 0 ){
			KMode = SJIS;
		} else if( strncmp( *argv, "addH=", 5 ) == 0 ){
			AddH = atoi( *argv+5 );
		} else if( strncmp( *argv, "addW=", 5 ) == 0 ){
			AddW = atoi( *argv+5 );
		} else if( strncmp( *argv, "MaxW=", 5 ) == 0 ){
			MaxW = atoi( *argv+5 );
		}
		argv++;
	}
	kanji_set( ObjStr, KMode, Out );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function search_ascii()                                */
/*                                                                    */
/*             char **fl    : font list                               */
/*             int num      : number of font list                     */
/*                                                                    */
/*             return char * : default ascii font name                */
/*                                                                    */
/*             function: search the list and find the ascii default   */
/*                      font name like 8x16 or...                     */
/**********************************************************************/
char *search_ascii( char **fl, int num )
{
     char *ret;
     int i;

     ret = NULL;
     for( i = 0; i < num; i++ ){
	  if( strcmp( fl[i], "8x16" ) == 0 ){
	       return( fl[i] );
	  }
     }
     for( i = 0; i < num; i++ ){
	  if( strncmp( fl[i], "8x1", 3 ) == 0 ){
	       if( strstr( fl[i], "kana" ) != NULL ){
		    continue;
	       }
	       if( strstr( fl[i], "roman" ) != NULL ){
		    continue;
	       }
	       return( fl[i] );
	  }
     }
     for( i = 0; i < num; i++ ){
	  if( strncmp( fl[i], "9x1", 3 ) == 0 ){
	       if( strstr( fl[i], "kana" ) != NULL ){
		    continue;
	       }
	       if( strstr( fl[i], "roman" ) != NULL ){
		    continue;
	       }
	       return( fl[i] );
	  }
     }
     return( NULL );
}

/**********************************************************************/
/*                                                                    */
/*             function search_kanji()                                */
/*                                                                    */
/*             char **fl    : font list                               */
/*             int num      : number of font list                     */
/*                                                                    */
/*             return char * : default kanji font name                */
/*                                                                    */
/*             function: search the list and find the kanji default   */
/*                      font name like kanji16 or kanji18...          */
/**********************************************************************/
char *search_kanji( char **fl, int num )
{
     char *ret;
     int i;

     ret = NULL;
     for( i = 0; i < num; i++ ){
	  if( strcmp( fl[i], "kanji16" ) == 0 ){
	       return( fl[i] );
	  } else if( strcmp( fl[i], "kanji18" ) == 0 ){
	       return( fl[i] );
	  }
     }
     for( i = 0; i < num; i++ ){
	  if( strncmp( fl[i], "kanji", 5 ) == 0 ){
	       return( fl[i] );
	  }
     }
     return( NULL );
}

/**********************************************************************/
/*                                                                    */
/*             function set_default_fonts()                           */
/*                                                                    */
/*             char **fl    : font list                               */
/*             int num      : number of font list                     */
/*                                                                    */
/*             return char * : default ascii font name                */
/*                                                                    */
/*             function: search the list and find and load the default*/
/*                      font.                                         */
/**********************************************************************/
void set_default_fonts( char **fl, int num )
{
     char *asc;
     char *kan;

     if( (Def = XLoadQueryFont( D, "fixed" )) == NULL ){
	  fatal( EILLSELECTION, "fixed font is not here" );
     }
     if( (asc = search_ascii( fl, num )) != NULL ){
	  if( (Ascii = XLoadQueryFont( D, asc )) == NULL ){
	       fatal( EILLSELECTION, "ascii font cannot load" );
	  }
     } else {
	  fprintf(stderr,"ascii font cannot find\n" );
     }
     if( (kan = search_kanji( fl, num )) != NULL ){
	  if( (Kanji = XLoadQueryFont( D, kan )) == NULL ){
	       fatal( EILLSELECTION, "kanji font cannot load" );
	  }
     } else {
	  fprintf(stderr,"kanji font cannot find\n" );
     }
}


/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             int argc     : number of argments                      */
/*             char **argv  : argments                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: main routine of xstring.                     */
/*                                                                    */
/**********************************************************************/
int main( int argc, char **argv )
{
	XtAppContext  app_con;
	Widget toplevel;
	Widget viewp;
	char **fontnames;
	int count, ret;
	XtTranslations trans;
	XawTextBlock txt, pnt;
	static XawTextPosition st1 = 0, ed1 = 0;
	static XawTextPosition st2 = 0, ed2 = 0;

	if( (ret = init_arg( argc, argv )) != 0 ){
		fatal( ret, "not found file" );
	}

/* create widits */
	toplevel = XtVaAppInitialize( &app_con, "FONTS", NULL, 0, 
				     &argc, argv, NULL, NULL );

	XtAppAddActions( app_con, actions, XtNumber(actions) );
	sheet = XtVaCreateManagedWidget( "sheet", formWidgetClass, 
					   toplevel, NULL );
	text = XtVaCreateManagedWidget( "text", asciiTextWidgetClass, sheet, 
					XtNwidth, 200,
				       XtNeditType, XawtextEdit, NULL );
	point = XtVaCreateManagedWidget( "point", asciiTextWidgetClass, sheet, 
					XtNwidth, 30,
					XtNfromHoriz, text,
				       XtNeditType, XawtextEdit, NULL );

/* regist translation to text Widget to be able to input search pattern */
	trans = XtParseTranslationTable( translationsTable );
	XtOverrideTranslations( text, trans );
	txt.firstPos = 0;
	txt.length = strlen( SearchFont );
	txt.ptr = SearchFont;
	txt.format = FMT8BIT;
	ed1 = strlen( SearchFont );
	XawTextReplace( text, st1, ed1, &txt );
	XawTextSetInsertionPoint( text, ed1 );

/* regist translation to point Widget to be able to input point of scalable font */
	pnt.firstPos = 0;
	pnt.length = strlen( PointSize );
	pnt.ptr = PointSize;
	pnt.format = FMT8BIT;
	ed2 = strlen( PointSize );
	XawTextReplace( point, st2, ed2, &pnt );
	XawTextSetInsertionPoint( point, ed2 );

/* create Widget for font-name list */
	viewp = XtVaCreateManagedWidget( "viewport", viewportWidgetClass, 
					sheet,
					XtNfromVert, text,
					XtNallowVert, True,
					XtNheight, 480,
					NULL );
	list = XtVaCreateManagedWidget( "list", listWidgetClass, viewp, 
				       XtNdefaultColumns, 1,
				       XtNforceColumns, True,
				       NULL );

/* default font set */
	D = XtDisplay( sheet );

/* get font names and set to the list Widget */
	if( (fontnames = XListFonts( D, SearchFont, 1024, &count )) == NULL ){
		fontnames = notfound; count = 1;
	}
	XawListChange( list, fontnames, count, 0, True );
	XtAddCallback( list, XtNcallback, Print, NULL );

	set_default_fonts( fontnames, count );

/* realize Widgets and wait events */
	XtRealizeWidget( toplevel );
	XtAppMainLoop( app_con );
	return( 0 );
}



