/**********************************************************************/
/*                                                                    */
/*          xstrsvr.c   gedo PLUG X interface program                 */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#define MAIN   1
#include "../common.h"
#include "../loader.h"
#include "xstring.h"

char *Version = { "GEDO plugs by K.Sugiura  ver 1.0\n" };

Display  *D;	/* X11 Display */
Window   W;	/* Window to display sub-window(image) */
Pixmap P;	/* Pixmap to get window image */
GC gc;		/* GC to draw buttons */

/************ SET FROM ARGMENT ******************/
unsigned char *ObjStr;		/* draw text */
static char FileText[TEXTMAX];	/* Filename in text */
int Mode = 0;			/* if 1, create image. if 0, show only */
char *SearchFont; 	     	/* font name of unchecked Font */
char *JFont;			/* font name of Japanese Font */
char *EFont;			/* font name of ASCII Font */
int ScalablePoint = 40;		/* if SearchFont is scalable font, use this */
int KMode = JIS;		/* KANJI Coding */
int AddW = 0;			/* distance between characters */
int AddH = 0;			/* distance between lines */
int MaxW = 0;			/* maximum of the font image width */
char *Geo;			/* geometry to create window */

XFontStruct *Kanji;		/* font to draw KANJI text */
XFontStruct *Ascii;		/* font to draw ASCII text */
XFontStruct *Def;		/* font to draw messages */
struct fontimage Image;	/* result image header to transact */

WSTRING *Out;			/* ObjStr converted to internal code */

void fatal( int error, char *str );

/**********************************************************************/
/*                                                                    */
/*            function Draw_Str()                                     */
/*                                                                    */
/*            args  GC gcf  : GC for display font                     */
/*                  WSTRING *p : text to display, see xsting.h        */
/*                  int h  : line distance                            */
/*                  int w  : char distance                            */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Write to the Pixmap the strings by defined font.*/
/*                     MultiLine, defined distance.                   */
/**********************************************************************/
void Draw_Str( GC gcf, WSTRING *p, int h, int w, int maxw )
{
	int x, y;
	int tmp;
	int asc;

	asc = Kanji->ascent * 12 / 10;
	if( h < MERGIN ){
		h = MERGIN;
	}
	x = y = MERGIN; 
	for( ; ; p += 2 ){
		if( p[0] == 0 ){
			if( p[1] == 0 ){
				break;
			} else if( p[1] == '\n' ){
				x = MERGIN; y += h;
			} else {
				tmp = XTextWidth( Ascii, (char *)&p[1], 1 ) + w;
				if( x + tmp > maxw ){
				  	x = MERGIN; y += h;
				}
				XSetFont( D, gcf, Ascii->fid );
				XDrawString( D, P, gcf, x, y+Kanji->ascent, (char *)&p[1], 1 );
				x += tmp;
			}
		} else {
			tmp = XTextWidth16( Kanji, (XChar2b *)p, 1 ) + w;
			if( x + tmp > maxw ){
			   	x = MERGIN; y += h;
			}
			XSetFont( D, gcf, Kanji->fid );
			XDrawString16( D, P, gcf, x, y+Kanji->ascent, (XChar2b *)p, 1 );
			x += tmp;
		}
	}
	XFlush( D );
}

/**********************************************************************/
/*                                                                    */
/*            function calc_maxwidth()                                */
/*                                                                    */
/*            args  WSTRING *txt  : text measured width, see xstring.h*/
/*                  int asciimax  : ASCII font maximum width          */
/*                  int kanjimax  : KANJI font maximum width          */
/*                  int *line : <OUT> How many lines in <txt>?        */
/*                                                                    */
/*            return int : Width(pixel) need to display <txt>         */
/*                                                                    */
/*            function: Nearly get the width of <txt> to display.     */
/*                                                                    */
/**********************************************************************/
int calc_maxwidth( WSTRING *txt, int *line, int w, int maxw )
{
	int width, max, tmp;
	WSTRING *p;

	max = width = MERGIN * 2;
	*line = 0;
	for( p = txt; ; p += 2 ){
		if( p[0] == 0 && p[1] == 0 ){
			if( max < width ){
				max = width;
			}
			break;
		} else if( p[0] == 0 ){
			if( p[1] == '\n' ){			
				(*line)++;
				if( max < width ){
					max = width;
				}
				width = MERGIN * 2;
			} else {
				tmp = XTextWidth( Ascii, (char *)&p[1], 1 ) + w;
				if( width + tmp >= maxw ){
				  	(*line)++;
					if( max < width ){
					  	max = width;
					}
					width = MERGIN * 2;
				} else {
				  	width += tmp;
				}
			}
		} else {
		  	tmp = XTextWidth16( Kanji, (XChar2b *)p, 1 ) + w;
			if( width + tmp >= maxw ){
			  	(*line)++;
				if( max < width ){
				  max = width;
				}
				width = MERGIN * 2;
			} else {
			  	width += tmp;
			}
		}
	}
	(*line)++;
	return( max );
}

/**********************************************************************/
/*                                                                    */
/*            function read_textfile()                                */
/*                                                                    */
/*            args  char *out  : read string int <file>               */
/*                  char *file : filename the display Text is in      */
/*                                                                    */
/*            return int   :   0==success   else==errorcode           */
/*                                                                    */
/*            function: If infiletext option is ON, read string from  */
/*                  <FileText>, and set it to <Text>.                 */
/**********************************************************************/
int read_textfile( char *out, char *file, int size )
{
	FILE *fp;

	if( file[0] == '\0' ){
		fatal( EOPENFILE, "need filetext=<filename>\n" );
	}
	if( (fp = fopen( file, "r" )) == NULL ){
		return( EOPENFILE );
	}
	if( fread( out, 1, size, fp ) <= 0 ){
		fclose( fp );
		return( EREADFILE );
	}
	out[size] = '\0';
	fclose( fp );
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function get_fontimage()                                */
/*                                                                    */
/*            args  struct fontimage *fi : header buffer to set       */
/*                  unsigned char *buff  : data buffer to set         */
/*                                                                    */
/*            return int   :  0==success  else==errorcode             */
/*                                                                    */
/*            function: Get font image from Pixmap and set to <fi>    */
/*                     and <buff>.                                    */
/**********************************************************************/
int get_fontimage( struct fontimage *fi, unsigned char *buff )
{
	XImage  *image;
	unsigned int *dt;
	unsigned short *ds;
	int scanline;
	int i, j;

	if( (image = XGetImage( D, P, 0, 0, fi->width, fi->height, 
			AllPlanes, ZPixmap )) == NULL ){
	        
	        fatal( EILLSELECT, "too large text or ?\n" );
	}
	XFlush( D );
	scanline = image->bytes_per_line;

        if( image->bits_per_pixel == 32 ){
		for( i = 0; i < image->height; i++ ){
			dt = (unsigned int *)((unsigned char *)image->data 
						+ i * scanline);
			for( j = 0; j < image->width; j++ ){
				if( dt[j] == 0 ){
					buff[i * fi->width + j] = 0;
				} else {
					buff[i * fi->width + j] = 1;
				}
			}
		}
	} else if( image->bits_per_pixel == 16 ){
		for( i = 0; i < image->height; i++ ){
			ds = (unsigned short *)((unsigned char *)image->data
						+ i * scanline);
			for( j = 0; j < image->width; j++ ){
				if( ds[j] == 0 ){
					buff[i * fi->width + j] = 0;
				} else {
					buff[i * fi->width + j] = 1;
				}
			}
		}
	}
	XDestroyImage( image );
	return( 0 );
}



/**********************************************************************/
/*                                                                    */
/*             function fatal()                                       */
/*                                                                    */
/*             int error : error code                                 */
/*             char *str : additional message                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Write error message to seltext process.      */
/*                                                                    */
/**********************************************************************/
void fatal( int error, char *str )
{
	struct fontimage fi;

	if( Mode == 1 ) {
	     memset( &fi, 0, sizeof(struct fontimage) );
	     strcpy( fi.jfont, str );
	     fi.flag = error;
	     fwrite( &fi, sizeof(struct fontimage), 1, stdout );
	} else {
	     printf( "set error \"%s: %s\"\n", StrErr[error], str );
	}
	free( ObjStr );
	exit( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Write the image of string to selstr process. */
/*                                                                    */
/**********************************************************************/
void write_image( void )
{
	char *buff;
	int ret;
	int size;

	if( Mode == 0 ){
	     printf( "set efont \"%s\"\n", EFont );
	     printf( "set jfont \"%s\"\n", JFont );
	     printf( "set error \"\"\n" );
	     return;
	}
	size = Image.width * Image.height;
	if( (buff = malloc( size )) == NULL ){
		fatal( EMALLOC, "malloc error" );
	}
	Image.flag = ENOERROR;
	strcpy( Image.jfont, JFont );
	strcpy( Image.efont, EFont );
	if( fwrite( &Image, sizeof(struct fontimage), 1, stdout ) != 1 ){
		free( buff );
		fatal( EWRITEFILE, "" );
	}
	if( (ret = get_fontimage( &Image, buff ) ) != 0 ){
		free( buff );
		fatal( ret, "" );
	}
	if( fwrite( buff, sizeof(char), size, stdout ) != size ){
		free( buff );
		fatal( EWRITEFILE, "");
	}
	free( buff );
}



/**********************************************************************/
/*                                                                    */
/*             function init_arg()                                    */
/*                                                                    */
/*             int argc     : number of argments                      */
/*             char **argv  : argments                                */
/*                                                                    */
/*             return int   : 0==success  else==errorcode             */
/*                                                                    */
/*             function: Initialize data from the invoking argments.  */
/*                                                                    */
/**********************************************************************/
int init_arg( int argc, char **argv )
{
	int ret;
	struct stat statbuf;

	argv++;
	while( --argc > 0 && *argv ){
		if( strncmp( *argv, "text=", 5 ) == 0 ){
			if( (ObjStr = malloc( 256 * 3 )) == NULL ){
				return( EMALLOC );
			}
			Out = ObjStr + 256;
			strcpy( ObjStr, *argv+5 );
		} else if( strncmp( *argv, "filetext=", 9 ) == 0 ){
			strcpy( FileText, *argv + 9 );
			if( stat( FileText, &statbuf ) == -1 ){
				return( EOPENFILE );
			}
			if( (ObjStr = malloc( (statbuf.st_size+1) * 3 )) == NULL ){
				return( EMALLOC );
			}
			memset( ObjStr, 0, (statbuf.st_size+1) * 3 );
			Out = ObjStr + statbuf.st_size + 1;
			if( (ret = read_textfile( ObjStr, FileText, statbuf.st_size )) != 0 ){
				return( ret );
			}
		} else if( strcmp( *argv, "mode=do" ) == 0 ){
		        Mode = 1;
		} else if( strncmp( *argv, "font=", 5 ) == 0 ){
			SearchFont = *argv+5;
		} else if( strncmp( *argv, "jfont=", 6 ) == 0 ){
			JFont = *argv+6;
		} else if( strncmp( *argv, "efont=", 6 ) == 0 ){
			EFont = *argv+6;
		} else if( strncmp( *argv, "scale=", 6 ) == 0 ){
			ScalablePoint = atoi( *argv+6 );
		} else if( strncmp( *argv, "jis", 3 ) == 0 ){
			KMode = JIS;
		} else if( strncmp( *argv, "euc", 3 ) == 0 ){
			KMode = EUC;
		} else if( strncmp( *argv, "sjis", 4 ) == 0 ){
			KMode = SJIS;
		} else if( strncmp( *argv, "addH=", 5 ) == 0 ){
			AddH = atoi( *argv+5 );
		} else if( strncmp( *argv, "addW=", 5 ) == 0 ){
			AddW = atoi( *argv+5 );
		} else if( strncmp( *argv, "MaxW=", 5 ) == 0 ){
			MaxW = atoi( *argv+5 );
		} else if( strncmp( *argv, "geo=", 4 ) == 0 ){
		        Geo = *argv+4;
		}
		argv++;
	}
	if( ObjStr == NULL ){
	     fatal( EILLSELECTION, "-text or -filetext is not set" );
	}
	kanji_set( ObjStr, KMode, Out );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function check_scalable()                              */
/*                                                                    */
/*             char *str  : font name                                 */
/*                                                                    */
/*             return  int : 1==scalable font  0==not scalable font   */
/*                                                                    */
/*             function: Return the font name is either scalable or   */
/*                     not.                                           */
/**********************************************************************/
int check_scalable( char *str )
{

	for( ; *str; str++ ){
		if( str[0] == '-' && str[1] == '-' ){
			break;
		}
	}
	if( str[2] == '0' && str[4] == '0' ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function make_fontname()                               */
/*                                                                    */
/*             char *dst  :<OUT>  the fontname confirmed points       */
/*             char *src  :<IN>   scalable fontname                   */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Make the scalable fontname to confirmed point*/
/*                  fontname.                                         */
/**********************************************************************/
void make_fontname( char *dst, char *src )
{
	for( ; (*dst = *src); dst++, src++ ){
		if( src[0] == '-' && src[1] == '-' ){
			break;
		}
	}
	sprintf( dst, "--%d", ScalablePoint );
	src += 3;
	dst += strlen( dst );
	for( ; (*dst = *src); dst++, src++ ){
		if( src[0] == '-' && src[1] == '0' ){
			dst[1] = '*';
			src++, dst++;
		}
	}
}

char *KanjiFontList[] = { "kanji16", "kanji18", NULL };
char *ASCIIFontList[] = { "fixed", "8x16", "8x13", "9x15", "7x14", "7x13", "10x20", NULL };

XFontStruct *set_fontlist( char **name, char **list )
{
     XFontStruct *ret;

     while( *list ){
	  if( (ret = XLoadQueryFont( D, *list )) != NULL ){
	       if( name ){
		    *name = *list;
	       }
	       return ret;
	  }
	  list++;
     }
     return NULL;  
}

void set_winname( void )
{
	XTextProperty win_name;	
	XClassHint  class;
	char *argv[1];

	win_name.value = (unsigned char *)"xstring processor";
	win_name.encoding = XA_STRING;
	win_name.format = 8;
	win_name.nitems = strlen( win_name.value );
	class.res_name = NULL;
	class.res_class = "Gedo";
	argv[0] = NULL;
	XSetWMProperties( D, W, &win_name, NULL, argv, 0, 
			 NULL, NULL, &class );
}

/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             int argc     : number of argments                      */
/*             char **argv  : argments                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: main routine of xstring.                     */
/*                                                                    */
/**********************************************************************/
int main( int argc, char **argv )
{
     unsigned long white, black;
     Window  quit, esc, Wmap, Wename, Wjname;
     GC gcf, gcc;
     int dfheight, fheight, height;
     int namewidth, buttonwidth, width;
     int jnamewidth, enamewidth;
     int button1width, button2width;
     int h, wd;
     XEvent event;
     int line;
     int ret;
     char jnamebuf[FONTMAX+16] = { "Japanese font : " };
     char enamebuf[FONTMAX+16] = { "ASCII font    : " };
     int geox = 100, geoy = 100;
     int geoflag = 0;

     if( (ret = init_arg( argc, argv )) != 0 ){
	  fatal( ret, "failed at parsing argments" );
     }
     
     D = XOpenDisplay( "" );
     if( D == NULL ){
	  fatal( EILLSELECTION, "error:X11 cannot start" );
     }


     if( MaxW <= 0 ) {
	  MaxW = 99999;
     }
     if( SearchFont ){
	  XFontStruct *font;
	  if( check_scalable( SearchFont ) ){
	       char nfont[FONTMAX];
	       make_fontname( nfont, SearchFont ); 
	       SearchFont = nfont;
	  }
	  if( (font = XLoadQueryFont( D, SearchFont )) == NULL ){
	       fatal( EILLSELECTION, "font is not here" );
	  }
	  if( font->min_byte1 ){
	       Kanji = font;
	       JFont = SearchFont;
	  } else {
	       Ascii = font;
	       EFont = SearchFont;
	  }
     }
     if( Kanji == NULL ){
	  if( JFont == NULL ){
	       if( (Kanji = set_fontlist( &JFont, KanjiFontList )) == NULL ){
		    fatal( EILLSELECTION, "FATAL:Japanese default font is not here" );
	       }
	  } else {
	       if( (Kanji = XLoadQueryFont( D, JFont )) == NULL ){
		    fatal( EILLSELECTION, "jfont is not here" );
	       }
	  }
     }
     if( Ascii == NULL ){
	  if( EFont == NULL ){
	       if( (Ascii = set_fontlist( &EFont, ASCIIFontList )) == NULL ){
		    fatal( EILLSELECTION, "FATAL:ASCII default font is not here" );
	       }
	  } else {
	       if( (Ascii = XLoadQueryFont( D, EFont )) == NULL ){
		    fatal( EILLSELECTION, "efont is not here" );
	       }
	  }
     }
     if( (Def = set_fontlist( NULL, ASCIIFontList )) == NULL ){
	  fatal( EILLSELECTION, "FATAL:default font is not here" );
     }
     strcat( jnamebuf, JFont );
     strcat( enamebuf, EFont );

     if( Geo ){
	  int c1, c2;
	  if( sscanf( Geo, "%dx%d+%d+%d", &c1, &c2, &geox, &geoy ) == 4 ){
	       geox += 50; geoy += 50;
	       geoflag = 1;
	  }
     }
     

/* prepare to make sub-window to preview */
	fheight = max(Kanji->ascent, Ascii->ascent);
	dfheight = Def ->ascent + Def->descent;
	height = MERGIN;
	jnamewidth = MERGIN + XTextWidth( Def, jnamebuf, strlen(jnamebuf) ) + MERGIN;
	height += dfheight;
	enamewidth = MERGIN + XTextWidth( Def, enamebuf, strlen(enamebuf) ) + MERGIN;
	height += (dfheight + MERGIN);
	namewidth = (jnamewidth > enamewidth)?(jnamewidth):(enamewidth);
	button1width = XTextWidth( Def, "OK", 2 );
	button2width = XTextWidth( Def, "ESC", 3 );
	buttonwidth = MERGIN + button1width + MERGIN + button2width + MERGIN;
	height += (dfheight + MERGIN);

/* get image size enough to show preview */
	Image.width = calc_maxwidth( Out, &line, AddW, MaxW );
	Image.height = (fheight + MERGIN + AddH) * (line+1) + MERGIN;
	width = (namewidth > Image.width+MERGIN*2)?(namewidth):(Image.width+MERGIN*2);
	width = (width > buttonwidth)?(width):(buttonwidth);
	height += (Image.height + MERGIN);

/* create sub-window */
	white = WhitePixel( D, 0 );
	black = BlackPixel( D, 0 );

	if( Mode == 1 ) {
	     P = XCreatePixmap( D, RootWindow( D, 0 ), 
				Image.width, Image.height, DefaultDepth(D,0) );
	     gcf = XCreateGC( D, P, 0, 0 );
	     gcc = XCreateGC( D, P, 0, 0 );
	     XSetForeground( D, gcc, WhitePixel(D,0) );
	     XFillRectangle( D, P, gcc, 0, 0, Image.width, Image.height );
/* draw the font image */
	     Draw_Str( gcf, Out, fheight + MERGIN + AddH, AddW, MaxW );
	     write_image();
	     free( ObjStr );
	     XFreeGC( D, gcf );
	     XFreeGC( D, gcc );
	     XFreePixmap( D, P );
	     exit( 0 );
	}
	W = XCreateSimpleWindow( D, RootWindow( D, 0 ), geox, geoy,
		width, height, 0, black, white );
	set_winname();

	gc = XCreateGC( D, W, 0, 0 );
	XSetFont( D, gc, Def->fid );

	h = MERGIN;
	Wjname = XCreateSimpleWindow( D, W, MERGIN, h,
		jnamewidth, dfheight, 0, black, white );
	h += dfheight;
	Wename = XCreateSimpleWindow( D, W, MERGIN, h,
		enamewidth, dfheight, 0, black, white );
	h += (dfheight + MERGIN);
	Wmap = XCreateSimpleWindow( D, W, MERGIN, h,
		Image.width, Image.height, 1, black, white );

	h += (Image.height + MERGIN);
	wd = MERGIN;
	quit = XCreateSimpleWindow( D, W, wd, h, button1width, dfheight, 1, black, white ); 
	wd += (MERGIN + button1width);
	esc = XCreateSimpleWindow( D, W, wd, h, button2width, dfheight, 1, black, white ); 
	
	XSelectInput( D, W, ExposureMask | ButtonPressMask );

/* prepare to draw the text */
	P = XCreatePixmap( D, RootWindow( D, 0 ), 
		Image.width, Image.height, DefaultDepth(D,0) );
	gcf = XCreateGC( D, P, 0, 0 );
	gcc = XCreateGC( D, P, 0, 0 );
	XSetForeground( D, gcc, WhitePixel(D,0) );
	XFillRectangle( D, P, gcc, 0, 0, Image.width, Image.height );
/* draw the font image */
	Draw_Str( gcf, Out, fheight + MERGIN + AddH, AddW, MaxW );
	XMapWindow( D, W );
	XMapSubwindows( D, W );
	if( geoflag ) {
	     XMoveWindow( D, W, geox, geoy );
	}

/* event loop */
	while( 1 ){
		XNextEvent( D, &event );
		switch( event.type ){
		case Expose:
			XDrawString( D, Wjname, gc, MERGIN, Def->ascent, jnamebuf, strlen(jnamebuf) );
			XDrawString( D, Wename, gc, MERGIN, Def->ascent, enamebuf, strlen(enamebuf) );
			XCopyArea( D, P, Wmap, gc, 0, 0, Image.width, Image.height, 0, 0 );
			XDrawString( D, quit, gc, 0, Def->ascent, "OK", 2 );
			XDrawString( D, esc, gc, 0, Def->ascent, "ESC", 3 );
			XFlush( D );
			break;
		case ButtonPress:
			if( event.xbutton.subwindow == quit ){
				write_image();
				free( ObjStr );
				exit( 0 );
				goto ex;
			} else if( event.xbutton.subwindow == esc ){
				fatal( EUSRBREAK, "" );
				goto ex;
			}
			break;
		}
	}
/* end */
ex:	XFreeGC( D, gcf );
	XFreeGC( D, gc );
	XFreePixmap( D, P );
	XDestroyWindow( D, W );
	return 0;
}
