/**********************************************************************/
/*                                                                    */
/*              xwd.c   gedo PLUG specified routine                   */
/*                                                                    */
/*                by K.sugiura   Ver 0.95 for GEDO of LINUX           */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>

#include "plug.h"

int Gui = NOGUI;

/********************** configure desciption ******************/
char *Version = { "GEDO plugs by K.Sugiura  ver 1.0" };
char *Spec = 	{ "xwd" };
char *Comment = { "X-WINDOW dump(*.xwd) loader/saver" };

PLUGITEM Item[];	/* no configuration */
int      ItemNum = 0;

struct Tags Tag[];
int TagNum = 0;

DEFAULTSTATUS;
/********************** description end ********************/

#define BYTE unsigned char
#define WORD unsigned short

int LSB_flag;		/* Machine byte order */
int File_LSB_flag;	/* image byte order, but file header is MSB fix */

typedef unsigned long CARD32;
typedef unsigned short CARD16;
typedef unsigned char CARD8 ;

typedef struct _xwd_file_header {
	/* header_size = SIZEOF(XWDheader) + length of null-terminated
	 * window name. */
	CARD32 header_size ;		

	CARD32 file_version ;	/* = XWD_FILE_VERSION above */
	CARD32 pixmap_format ;	/* ZPixmap or XYPixmap */
	CARD32 pixmap_depth ;	/* Pixmap depth */
	CARD32 pixmap_width ;	/* Pixmap width */
	CARD32 pixmap_height ;	/* Pixmap height */
	CARD32 xoffset ;		/* Bitmap x offset, normally 0 */
	CARD32 byte_order ;		/* of image data: MSBFirst, LSBFirst */

	/* bitmap_unit applies to bitmaps (depth 1 format XY) only.
	 * It is the number of bits that each scanline is padded to. */
	CARD32 bitmap_unit ;		

	CARD32 bitmap_bit_order ;	/* bitmaps only: MSBFirst, LSBFirst */

	/* bitmap_pad applies to pixmaps (non-bitmaps) only.
	 * It is the number of bits that each scanline is padded to. */
	CARD32 bitmap_pad ;		

	CARD32 bits_per_pixel ;	/* Bits per pixel */

	/* bytes_per_line is pixmap_width padded to bitmap_unit (bitmaps)
	 * or bitmap_pad (pixmaps).  It is the delta (in bytes) to get
	 * to the same x position on an adjacent row. */
	CARD32 bytes_per_line ;
	CARD32 visual_class ;	/* Class of colormap */
	CARD32 red_mask ;		/* Z red mask */
	CARD32 green_mask ;		/* Z green mask */
	CARD32 blue_mask ;		/* Z blue mask */
	CARD32 bits_per_rgb ;	/* Log2 of distinct color values */
	CARD32 colormap_entries ;	/* Number of entries in colormap; not used? */
	CARD32 ncolors ;		/* Number of XWDColor structures */
	CARD32 window_width ;	/* Window width */
	CARD32 window_height ;	/* Window height */
	CARD32 window_x ;		/* Window upper left X coordinate */
	CARD32 window_y ;		/* Window upper left Y coordinate */
	CARD32 window_bdrwidth ;	/* Window border width */
} XWDFileHeader;

typedef struct {		/* color table of .xwd */
        CARD32	pixel;		/* index(not serial) */
        CARD16	red;
	CARD16	green;
	CARD16	blue;
        CARD8	flags;		/* ==7 */
        CARD8	pad;		/* ==0 */
} XWDColor;

typedef struct {
	unsigned char red;
	unsigned char green;
	unsigned char blue;
} ColorTable;

ColorTable Ctable[0x800];	/* private color table */

#define RGB32  0
#define RGB16  1
#define PSEUDO 2
int Mode;			/* read file mode RGB32, RGB16 or PSEUDO */

unsigned long RMask, GMask, BMask;	/* mask for RGB16 */
int Rshift, Gshift, Bshift;		/* shift calc for RGB16 */
int LineByte;				/* byte at a scan line */

/**********************************************************************/
/*                                                                    */
/*             function dconv()                                       */
/*                                                                    */
/*             unsigned long  v : LONG  data                          */
/*                                                                    */
/*             return  unsigned long  : LONG  data                    */
/*                                                                    */
/*             function: convert byte order(LONG) allways.            */
/*                                                                    */
/**********************************************************************/
unsigned long dconv( unsigned long v )
{
	unsigned long v1,v2,v3, v4;

	v1 = (v & 0xff000000) >> 24;
	v2 = (v & 0x00ff0000) >> 8;
	v3 = (v & 0x0000ff00) << 8;
	v4 = (v & 0x000000ff) << 24;
	return( v1 | v2 | v3 | v4 );
} 

/**********************************************************************/
/*                                                                    */
/*             function wconv()                                       */
/*                                                                    */
/*             unsigned short v : SHORT data                          */
/*                                                                    */
/*             return  unsigned short : SHORT data                    */
/*                                                                    */
/*             function: convert byte order(SHORT) allways.           */
/*                                                                    */
/**********************************************************************/
unsigned short wconv( unsigned  short v )
{
	unsigned short v1, v2;

	v1 = (v & 0xff00) >> 8;
	v2 = (v & 0x00ff) << 8;
	return( v1 | v2  );
} 

/**********************************************************************/
/*                                                                    */
/*             function Dconv()                                       */
/*                                                                    */
/*             unsigned long  v : LONG  data                          */
/*                                                                    */
/*             return  unsigned long  : LONG  data                    */
/*                                                                    */
/*             function: convert byte order(LONG)  if LSB machine     */
/*                                                                    */
/**********************************************************************/
unsigned long Dconv( unsigned long v )
{
	if( LSB_flag ){
		return( dconv( v ) );
	} else {
		return( v );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function Wconv()                                       */
/*                                                                    */
/*             unsigned short v : SHORT data                          */
/*                                                                    */
/*             return  unsigned short : SHORT data                    */
/*                                                                    */
/*             function: convert byte order(SHORT) if LSB machine     */
/*                                                                    */
/**********************************************************************/
unsigned short Wconv( unsigned short v )
{
	if( LSB_flag ){
		return( wconv( v ) );
	} else {
		return( v );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function D_conv()                                      */
/*                                                                    */
/*             unsigned long  v : LONG  data                          */
/*                                                                    */
/*             return  unsigned long  : LONG  data                    */
/*                                                                    */
/*             function: convert byte order(LONG) if Machine and file */
/*                      byte order is different.                      */
/**********************************************************************/
unsigned long D_conv( unsigned long v )
{
	if( LSB_flag == File_LSB_flag ){
		return( v );
	} else {
		return( dconv( v ) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function W_conv()                                      */
/*                                                                    */
/*             unsigned short v : SHORT data                          */
/*                                                                    */
/*             return  unsigned short : SHORT data                    */
/*                                                                    */
/*             function: convert byte order(SHORT) if Machine and file*/
/*                      byte order is different.                      */
/**********************************************************************/
unsigned short W_conv( unsigned short v )
{
	if( LSB_flag == File_LSB_flag ){
		return( v );
	} else {
		return( wconv( v ) );
	}
}



/**********************************************************************/
/*                                                                    */
/*             function read_header()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ HEADER of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_header( struct File *file )
{
	XWDFileHeader header;
	char filename[256];
	unsigned long size;
	unsigned long colornum;
	int i;
	XWDColor ctbuff;
	int index;

	if( (file->fd = open( file->name, O_RDONLY )) == -1 ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EOPENFILE );
	}
	if( read( file->fd, &header, sizeof(XWDFileHeader) ) != sizeof(XWDFileHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EREADFILE );
	}
	if( Dconv(header.pixmap_format) != 2 ){
		return( ENOTFEATURE );
	}
	if( Dconv(header.visual_class) < 2 ){
		return( ENOTFEATURE );
	}
	size = Dconv(header.header_size);
	size -= sizeof(XWDFileHeader);
	if( read( file->fd, filename, size ) != size ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EREADFILE );
	}
	LineByte = Dconv( header.bytes_per_line );

	if( Dconv( header.byte_order ) == 1 ){
		File_LSB_flag = 0;
	} else {
		File_LSB_flag = 1;
	}

	switch( Dconv( header.bits_per_pixel ) ){
	case 0x20:
		Mode = RGB32;
		break;
	case 0x10:
		Mode = RGB16;
		break;
	case 0x08:
		Mode = PSEUDO;
		break;
	default:
		file->mes->datasize = 0;
		return( EILLFORM );
		break;
	}

	colornum = Dconv(header.ncolors);
	size = sizeof(XWDColor);
	for( i = 0; i < colornum; i++ ){
		if( read( file->fd, &ctbuff, size ) != size ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EREADFILE );
		}
		if( Mode != PSEUDO ){
			continue;
		}
		index = Dconv(ctbuff.pixel);
		if( index >= 0x800 ){
			return( ENOTFEATURE );
		}
		if( ctbuff.flags != 7 ){
			return( EILLFORM );
		}
		Ctable[index].red = ((Wconv(ctbuff.red) & 0xff00) >> 8);
		Ctable[index].green = ((Wconv(ctbuff.green) & 0xff00) >> 8);
		Ctable[index].blue = ((Wconv(ctbuff.blue) & 0xff00) >> 8);
	}

	RMask = Dconv( header.red_mask );
	GMask = Dconv( header.green_mask );
	BMask = Dconv( header.blue_mask );
	for( i = 0; i < 32; i++ ){
		if( (RMask << i) >= 0x00800000L ){
			break;
		}
	}
	Rshift = i;
	for( i = 0; i < 32; i++ ){
		if( (GMask << i) >= 0x00008000L ){
			break;
		}
	}
	Gshift = i;
	for( i = 0; i < 32; i++ ){
		if( (BMask << i) >= 0x00000080L ){
			break;
		}
	}
	Bshift = i;
	
	file->buf->width = Dconv( header.window_width );
	file->buf->height = Dconv( header.window_height );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function color16to32()                                 */
/*                                                                    */
/*             unsigned long in : Color pixel of 16 bit color         */
/*                                                                    */
/*             return  unsigned long: Color pixel of 32 bit color     */
/*                                                                    */
/*             function: Convert 16bit color to 32bit color, by       */
/*                      RGBmask and RGBshift at file.                 */
/**********************************************************************/
unsigned long color16to32( unsigned long in )
{
	unsigned long ret;
	unsigned long tmp;

	tmp = (in & RMask);
	ret = (tmp << Rshift);
	tmp = (in & GMask);
	ret |= (tmp << Gshift);
	tmp = (in & BMask);
	ret |= (tmp << Bshift);
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function colorpto32()                                  */
/*                                                                    */
/*             unsigned long in : Color pixel of pseudo color         */
/*                                                                    */
/*             return  unsigned long: Color pixel of 32 bit color     */
/*                                                                    */
/*             function: Convert pseudocolor to 32bit color.          */
/*                                                                    */
/**********************************************************************/
unsigned long colorpto32( int in )
{
	unsigned long ret;
	unsigned short tmp;

	tmp = Ctable[in].red;
	ret = ((unsigned long)tmp << 16);
	tmp = Ctable[in].green;
	ret |= ((unsigned long)tmp << 8);
	tmp = Ctable[in].blue;
	ret |= ((unsigned long)tmp);
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function read_image()                                  */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function: READ IMAGES of the SPEC format from file     */
/*                                                                    */
/**********************************************************************/
int read_image( struct File *file )
{
	long size;
	unsigned short *tmp;
	unsigned char *tmpc;
	struct Shm *buf;
	int i, j;
	unsigned int *p;

	buf = file->buf;
	size = LineByte;
	p = buf->data;
	switch( Mode ){
	case RGB32:
		for( i = 0; i < buf->height; i++ ){
			if( read( file->fd, p, size ) != size ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				return( EREADFILE );
			}
			for( j = 0; j < file->buf->width; j++ ){
				p[j] = D_conv(p[j]);
			}
			p += buf->width;
		}
		break;
	case RGB16:
		if( (tmp = (unsigned short *)malloc( size )) == NULL ){
			return( EMALLOC );
		}
		for( i = 0; i < buf->height; i++ ){
			if( read( file->fd, tmp, size ) != size ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				return( EREADFILE );
			}
			for( j = 0; j < buf->width; j++ ){
				p[j] = color16to32( (unsigned long)W_conv(tmp[j]) );
			}
			p += buf->width;
		}
		free( tmp );
		break;
	case PSEUDO:
		if( (tmpc = (unsigned char *)malloc( size )) == NULL ){
			return( EMALLOC );
		}
		for( i = 0; i < buf->height; i++ ){
			if( read( file->fd, tmpc, size ) != size ){
				strcpy( file->mes->u.data, strerror(errno) );
				file->mes->datasize = strlen(file->mes->u.data);
				close( file->fd );
				return( EREADFILE );
			}
			for( j = 0; j < buf->width; j++ ){
				p[j] = colorpto32( (int)tmpc[j] );
			}
			p += buf->width;
		}
		free( tmpc );
		break;
	}
	close( file->fd );
	buf->mask_kind = SIMPLEMASK;
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function write_header()                                */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE HEADER of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_header( struct File *file )
{
	XWDFileHeader header;
	int len;

	if( (file->fd = open( file->name, O_RDWR | O_TRUNC )) == -1 ){
		if( (file->fd = open( file->name, O_RDWR | O_CREAT, 0664)) == -1 ){
			strcpy( file->mes->u.data, strerror(errno) );
			file->mes->datasize = strlen(file->mes->u.data);
			return( EOPENFILE );
		}
		memset( &header, 0, sizeof(XWDFileHeader) );
	}
	header.header_size = Dconv( sizeof(XWDFileHeader) + strlen(file->name) + 1);
	header.file_version = Dconv( 7 );
	header.pixmap_format = Dconv( 2 );
	header.pixmap_depth = Dconv( 0x20 );
	header.bits_per_pixel = Dconv( 0x20 );
	header.bytes_per_line = Dconv( (file->buf->width + 3) / 4 * 4 * 4 );
	header.red_mask = Dconv( 0x00ff0000 );
	header.green_mask = Dconv( 0x0000ff00 );
	header.blue_mask = Dconv( 0x000000ff );
	header.bits_per_rgb = Dconv( 8 );
	header.pixmap_width = Dconv( file->buf->width );
	header.pixmap_height = Dconv( file->buf->height );
	header.xoffset = Dconv( 0 );
	header.byte_order = Dconv( (LSB_flag)?(0):(1) );
	header.bitmap_unit = Dconv( 0x20 );
	header.bitmap_bit_order = Dconv( (LSB_flag)?(0):(1) );
	header.bitmap_pad = Dconv( 0x20 );
	header.visual_class = Dconv( 4 );
	header.colormap_entries = Dconv( 0x0 );
	header.ncolors = Dconv( 0x0 );
	header.window_width = Dconv( file->buf->width );
	header.window_height = Dconv( file->buf->height );
	header.window_x = Dconv( 100 );
	header.window_y = Dconv( 100 );
	header.window_bdrwidth = Dconv( 0 );

	if( write( file->fd, &header, sizeof(XWDFileHeader) ) != sizeof(XWDFileHeader) ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EWRITEFILE );
	}
	len = strlen(file->name) + 1;
	if( write( file->fd, file->name, len ) != len ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		return( EWRITEFILE );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function write_image()                                 */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:WRITE IMAGES of the SPEC format to file       */
/*                                                                    */
/**********************************************************************/
int write_image( struct File *file )
{
	long size1, size2;
	unsigned int *buffl, *pl;
	struct Shm *buf;
	int i, j;

	buf = file->buf;
	size1= (file->buf->width + 3) / 4 * 4 * 4;
	size2 = buf->height * size1;
	if( (buffl = malloc( size2 )) == NULL ){
		file->mes->datasize = 0;
		close( file->fd );
		return( EMALLOC );
	}
	pl = buffl;
	for( i = 0; i < buf->height; i++ ){
		for( j = 0; j < buf->width; j++ ){
			pl[j] = (buf->data[i * buf->width + j] & 0x00ffffff);
		}
		pl = (unsigned int *)((unsigned char *)pl + size1);
	}
	if( write( file->fd, buffl, size2 ) != size2 ){
		strcpy( file->mes->u.data, strerror(errno) );
		file->mes->datasize = strlen(file->mes->u.data);
		close( file->fd );
			return( EWRITEFILE );
	}
	free( buffl );
	close( file->fd );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function config()                                      */
/*                                                                    */
/*             struct File *file: Recieved Message from GEDO          */
/*                                                                    */
/*             return  int: 0=Success other=Fail(ErrorNO)             */
/*                                                                    */
/*             function:CONFIGURE the PLUG at Transaction CONFIG      */
/*                                                                    */
/**********************************************************************/
int config( struct File *file )
{
	char *str;

	str = file->mes->u.data;
	if( file->mes->command == CLOAD ){
		if( str[0] == '\0' ){
			printf( "loader xwd: no configration parameter\n" );
			file->mes->u.data[0] = '\0';
		} else {
			return( line_config( file ) );
		}
	} else if( file->mes->command == CSAVE ){
		if( str[0] == '\0' ){
			printf( "saver xwd : no configration parameter\n" );
			file->mes->u.data[0] = '\0';
		} else {
			return( line_config( file ) );
		}
	}
	return( 0 );
}
