/**********************************************************************/
/*                                                                    */
/*        builtin1.c -- gedo builtin function routine                 */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*        rev A0:  1996.11.19: add readposf, writeposf                */
/*        rev A0:  1996.11.21: add searchpath in readposf, writeposf  */
/*        rev A0:  1996.11.21: add func_sscanf                        */
/*        rev A0:  1996.11.30: change va_list in sprintf_args         */
/*        rev B0:  1997. 1.16: add func_mcar,func_mcdr                */
/*        rev B0:  1997. 1.16: 1 arg function add KLIST check         */
/*        rev B0:  1997. 1.17: func_funcs add function                */
/*        rev B2:  1997. 5.13: change GLOBAL var masked by TMP var    */
/*        rev D0:  1999.12.11: floating value use #floatunit          */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"
/* add A0 */
#include "../loader.h"
/* end A0 */

#ifdef DOS
#define PI 3.141592653
#endif
#ifndef PI
#define PI M_PI
#endif
/**********************************************************************/
/*                                                                    */
/*             function sprintf_args()                                */
/*                                                                    */
/*             Lisp com: Argments of sprintf                          */
/*             unsigned *arg: Real body of va_list                    */
/*             int max: Maximum of arg                                */
/*                                                                    */
/*             return  va_list: Argments of vsprintf                  */
/*                                                                    */
/*             function: Convert Argments(arg) to va_list to sprintf. */
/*       A0: rewrite.                                                 */
/*                                                                    */
/**********************************************************************/
va_list sprintf_args( Lisp com, unsigned *arg, int max )
{
	va_list argp;
	Lisp at, val;

	argp = (va_list)&arg[0] + sizeof(arg[0]);

	if( List[com].kind == KLIST ){
		at = com;
		while( at ){
			val = List[at].v;
			if( List[val].kind == KNUMBER ){
				argp += sizeof(int);
				((int *)argp)[-1] = List[val].v;
			} else if( List[val].kind == KLITERAL ){
				argp += sizeof(char *);
				((char **)argp)[-1] = tsv(val);
			} else {
				return( NULL );
			}
			at = List[at].next;
			if( (unsigned *)argp - &arg[1] >= max ){
				warning( W_MANYARG );
				break;
			}
		}
	} else {
		val = com;
		if( List[val].kind == KNUMBER ){
			argp += sizeof(int);
			((int *)argp)[-1] = List[val].v;
		} else if( List[val].kind == KLITERAL ){
			argp += sizeof(char *);
			((char **)argp)[-1] = tsv(val);
		} else {
			return( NULL );
		}
	}

	return( (va_list)&arg[1] );
}


/**********************************************************************/
/*                                                                    */
/*             function func_sprintf()                                */
/*                                                                    */
/*             Lisp com: LITERAL, OMNI,...                            */
/*                                                                    */
/*             return  Lisp: LITERAL                                  */
/*                                                                    */
/*             function: sprintf function                             */
/*                                                                    */
/**********************************************************************/
Lisp func_sprintf( Lisp com )
{
	Lisp form, val;
	va_list argp;
	unsigned arg[20];

	val = eval_exp( com );
	if( List[val].kind == KLIST ){
		form = List[val].v;
		argp = sprintf_args( List[val].next, arg, 18 );
	} else {
		form = val;
		argp = NULL;
	}
	if( !form || List[form].kind != KLITERAL ){
		return( NIL );
	}
	if( argp ){
		vsprintf( Work, tsv(form), argp );
	} else {
		strcpy( Work, tsv(form) );
	}
	return( cons( tmp_strdup( Work ), KLITERAL, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_car()                                    */
/*                                                                    */
/*             Lisp com: LIST, POSITION or LITERAL                    */
/*                                                                    */
/*             return  Lisp: OMNI                                     */
/*                                                                    */
/*             function: car function                                 */
/*                                                                    */
/**********************************************************************/
Lisp func_car( Lisp com )
{
	Lisp val;
	char work[2];

	val = eval_exp( com );
/* B0 add */
	if( List[val].kind != KLIST ){
		return( NIL );
	}
	val = List[val].v;
/* B0 end */
	if( (Kind_spec[List[val].kind] & 0xf0) == KSLIST ){
		return( List[val].v );
	} else if( List[val].kind == KPOSITION ){
		return( List[val].v );
	} else if( List[val].kind == KLITERAL ){
		work[0] = TmpString[List[val].v];
		work[1] = '\0';
		return( cons( tmp_strdup(work), KLITERAL, NIL ) ); 
	} else {
		return( val );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_cdr()                                    */
/*                                                                    */
/*             Lisp com: LIST, POSITION or LITERAL                    */
/*                                                                    */
/*             return  Lisp: OMNI                                     */
/*                                                                    */
/*             function: cdr function                                 */
/*                                                                    */
/**********************************************************************/
Lisp func_cdr( Lisp com )
{
	Lisp val;

	val = eval_exp( com );
/* B0 add */
	if( List[val].kind != KLIST ){
		return( NIL );
	}
	val = List[val].v;
/* B0 end */
	if( (Kind_spec[List[val].kind] & 0xf0) == KSLIST ){
		if( List[val].next == NIL ){
			return( cons( NIL, KNIL, NIL ) );
		} else {
			return( List[val].next );
		}
	} else if( List[val].kind == KPOSITION ){
		return( List[val].next );
	} else if( List[val].kind == KLITERAL ){
		strcpy( Work, &TmpString[List[val].v+1] );
		return( cons( tmp_strdup(Work), KLITERAL, NIL ) ); 
	} else {
		return( cons( NIL, KNIL, NIL ) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_mcar()                                    */
/*                                                                    */
/*             Lisp com: LIST, POSITION or LITERAL                    */
/*                                                                    */
/*             return  Lisp: OMNI                                     */
/*                                                                    */
/*             function: Macro car not evaluate.                      */
/*             B0 added                                               */
/**********************************************************************/
Lisp func_mcar( Lisp com )
{
	Lisp val;

	if( List[com].kind != KLIST ){
		return( NIL );
	}
	com = List[com].v;
	if( List[com].kind == KSHARP ){
		val = expand_macro( List[com].v );
	} if( List[com].kind != KLIST ){
		val = eval_exp( com );
	} else {
		val = com;
	}
	if( List[val].kind == KLIST ){
		return( List[val].v );
	} else if( List[val].kind == KPOSLIST ){
		return( List[val].v );
        } else {
		return( NIL );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_mcdr()                                    */
/*                                                                    */
/*             Lisp com: LIST, POSITION or LITERAL                    */
/*                                                                    */
/*             return  Lisp: OMNI                                     */
/*                                                                    */
/*             function: Macro car not evaluate.                      */
/*                                                                    */
/**********************************************************************/
Lisp func_mcdr( Lisp com )
{
	Lisp val;
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	com = List[com].v;
	if( List[com].kind == KSHARP ){
		val = expand_macro( List[com].v );
	} if( List[com].kind != KLIST ){
		val = eval_exp( com );
	} else {
		val = com; 
	}
	if(List[val].kind == KLIST ){
		if( List[val].next == NIL ){
			return( cons( NIL, KNIL, NIL ) );
		} else {
			return( List[val].next );
		}
        } else if( List[val].kind == KPOSLIST ){
		return( List[val].next );
	} else if( List[val].kind == KPOSITION ){
		return( val );
	} else {
		return( NIL );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_cons()                                   */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                       LIST                                         */
/*             return  Lisp: LIST                                     */
/*                                                                    */
/*             function: cons function                                */
/*                                                                    */
/**********************************************************************/
Lisp func_cons( Lisp com )
{
	Lisp at;
	Lisp cdr;
	Lisp car;

	at = eval_exp(com);
	if( List[at].kind != KLIST ){
		return( cons( NIL, KNIL, NIL ) );
	}
	car = List[at].v;
	cdr = List[List[at].next].v;
	if( List[cdr].kind == KNIL ){
		return( cons( car, KLIST, NIL ) );
	} else if( List[cdr].kind == KPOSLIST || List[cdr].kind == KPOSITION ){
		if( List[car].kind == KPOSITION ){
			return( cons( car, KPOSLIST, cdr ) );
		} else {
			return( cons( NIL, KNIL, NIL ) );
		}
	} else if( List[cdr].kind != KLIST ){
		return( cons( car, KLIST, cons( cdr, KLIST, NIL ) ) );
	} else {
		return( cons( car, KLIST, cdr ) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_getenv()                                 */
/*                                                                    */
/*             Lisp com: LITERAL                                      */
/*                                                                    */
/*             return  Lisp: LITERAL                                  */
/*                                                                    */
/*             function: return enviroment-variable                   */
/*                                                                    */
/**********************************************************************/
Lisp func_getenv( Lisp com )
{
	Lisp at;
	char *s;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind != KLITERAL ){
		return( cons( NIL, KNIL, NIL ) );
	}
	if( (s = getenv( tsv(at) )) == NULL ){ 
		return( cons( NIL, KNIL, NIL ) );
	} else {
		return( cons( tmp_strdup(s), KLITERAL, NIL ) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_setenv()                                 */
/*                                                                    */
/*             Lisp ls: LITERAL(env-name)                             */
/*                      LITERAL(value)                                */
/*             return  Lisp: LIST{LITERAL(envname),LITREAL(value)}    */
/*                                                                    */
/*             function: Set the enviroment-variable value.           */
/*                                                                    */
/**********************************************************************/
Lisp func_setenv( Lisp ls )
{
	Lisp at;
	Lisp env, str;
	char *s;

	at = eval_exp(ls);
	if( List[at].kind != KLIST ){
		return( cons( NIL, KNIL, NIL ) );
	}
	env = List[at].v; str = List[List[at].next].v;
	if( List[env].kind != KLITERAL ){
		return( cons( NIL, KNIL, NIL ) );
	}
	if( List[str].kind == KLITERAL ){
		if( List[str].v ){
			sprintf( Work, "%s=%s", tsv(env), tsv(str) );
		} else {
			sprintf( Work, "%s=", tsv(env) );
		}
	}
	s = (char *)strdup( Work );
	putenv( s );
	return( ls );
}

/**********************************************************************/
/*                                                                    */
/*             function func_commandp()                               */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: INT 1 or NIL                             */
/*                                                                    */
/*             function: command?(list)                               */
/*                                                                    */
/**********************************************************************/
Lisp func_commandp(Lisp com)
{
	Lisp at;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KCOM ){
		return( cons(1, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_stringp()                                */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: INT 1 or NIL                             */
/*                                                                    */
/*             function: string?()                                    */
/*                                                                    */
/**********************************************************************/
Lisp func_stringp(Lisp com)
{
	Lisp at;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KLITERAL ){
		return( cons(1, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_listp()                                  */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: INT 1 or NIL                             */
/*                                                                    */
/*             function: list?()                                      */
/*                                                                    */
/**********************************************************************/
Lisp func_listp(Lisp com)
{
	Lisp at;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( (Kind_spec[List[at].kind] & 0xf0) == KSLIST ){
		return( cons(1, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_atomp()                                  */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: INT 1 or NIL                             */
/*                                                                    */
/*             function: atom?()                                      */
/*                                                                    */
/**********************************************************************/
Lisp func_atomp(Lisp com)
{
	Lisp at;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( (Kind_spec[List[at].kind] & 0xf0) == KSLIST ){
		return( cons(NIL, KNIL, NIL) );
	} else {
		return( cons(1, KNUMBER, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_numberp()                                */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: INT 1 or NIL                             */
/*                                                                    */
/*             function: number?()                                    */
/*                                                                    */
/**********************************************************************/
Lisp func_numberp(Lisp com)
{
	Lisp at;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KNUMBER ){
		return( cons(1, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_positionp()                              */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: INT 1 or NIL                             */
/*                                                                    */
/*             function: position?()                                  */
/*                                                                    */
/**********************************************************************/
Lisp func_positionp(Lisp com)
{
	Lisp at;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KPOSITION ){
		return( cons(1, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_maskp()                                  */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: INT 1 or NIL                             */
/*                                                                    */
/*             function: mask?()                                      */
/*                                                                    */
/**********************************************************************/
Lisp func_maskp(Lisp com)
{
	Lisp at;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KMASK ){
		return( cons(1, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_sin()                                    */
/*                                                                    */
/*             Lisp com: INT(degree)                                  */
/*                                                                    */
/*             return  Lisp: INT  (sin(com)*100)                      */
/*                                                                    */
/*             function: sin()                                        */
/*                                                                    */
/**********************************************************************/
Lisp func_sin(Lisp com)
{
	Lisp at;
	double v;
	int val;
/* D0 add */
	double unit;

	if( (at = search_global( "#floatunit" )) == NIL ){
	     unit = 1.0;
	} else {
	     if( List[at].kind == KNUMBER ){
		  unit = (double)List[at].v;
	     } else {
		  unit = 1.0;
	     }
	}
/* D0 end */
	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KNUMBER ){
		v = (double)List[at].v / unit;
		val = (int)(sin( v * PI / 180 ) * unit * 10.0);
		if( val % 10 < 5 ){
			val = val / 10;
		} else {
			val = val / 10 + 1;
		} 
		return( cons(val, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_cos()                                    */
/*                                                                    */
/*             Lisp com: INT(degree)                                  */
/*                                                                    */
/*             return  Lisp: INT (cos(com)*100)                       */
/*                                                                    */
/*             function: cos()                                        */
/*                                                                    */
/**********************************************************************/
Lisp func_cos(Lisp com)
{
	Lisp at;
	double v;
	int val;
/* D0 add */
	double unit;

	if( (at = search_global( "#floatunit" )) == NIL ){
	     unit = 1.0;
	} else {
	     if( List[at].kind == KNUMBER ){
		  unit = (double)List[at].v;
	     } else {
		  unit = 1.0;
	     }
	}
/* D0 end */

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KNUMBER ){
		v = (double)List[at].v / unit;
		val = (int)(cos( v * PI / 180 ) * unit * 10.0);
		if( val % 10 < 5 ){
			val = val / 10;
		} else {
			val = val / 10 + 1;
		} 
		return( cons(val, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_tan()                                    */
/*                                                                    */
/*             Lisp com: INT(degree)                                  */
/*                                                                    */
/*             return  Lisp: INT (tan(com)*100)                       */
/*                                                                    */
/*             function: tan()                                        */
/*                                                                    */
/**********************************************************************/
Lisp func_tan(Lisp com)
{
	Lisp at;
	double v;
	int val;
/* D0 add */
	double unit;

	if( (at = search_global( "#floatunit" )) == NIL ){
	     unit = 1.0;
	} else {
	     if( List[at].kind == KNUMBER ){
		  unit = (double)List[at].v;
	     } else {
		  unit = 1.0;
	     }
	}
/* D0 end */

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KNUMBER ){
		v = (double)List[at].v / unit;
		val = (int)(tan( v * PI / 180 ) * unit * 10.0);
		if( val % 10 < 5 ){
			val = val / 10;
		} else {
			val = val / 10 + 1;
		} 
		return( cons(val, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_asin()                                   */
/*                                                                    */
/*             Lisp com: INT(value*100)                               */
/*                                                                    */
/*             return  Lisp: INT(degree)                              */
/*                                                                    */
/*             function: asin()                                       */
/*                                                                    */
/**********************************************************************/
Lisp func_asin(Lisp com)
{
	Lisp at;
	double v;
	int val;
/* D0 add */
	double unit;

	if( (at = search_global( "#floatunit" )) == NIL ){
	     unit = 1.0;
	} else {
	     if( List[at].kind == KNUMBER ){
		  unit = (double)List[at].v;
	     } else {
		  unit = 1.0;
	     }
	}
/* D0 end */

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KNUMBER ){
		v = (double)List[at].v / unit;
		val = (int)(asin( v ) / PI * 180.0 * 10.0 * unit);
		if( val % 10 < 5 ){
			val = val / 10;
		} else {
			val = val / 10 + 1;
		} 
		return( cons(val, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_acos()                                   */
/*                                                                    */
/*             Lisp com: INT(value*100)                               */
/*                                                                    */
/*             return  Lisp: INT(degree)                              */
/*                                                                    */
/*             function: acos()                                       */
/*                                                                    */
/**********************************************************************/
Lisp func_acos(Lisp com)
{
	Lisp at;
	double v;
	int val;
/* D0 add */
	double unit;

	if( (at = search_global( "#floatunit" )) == NIL ){
	     unit = 1.0;
	} else {
	     if( List[at].kind == KNUMBER ){
		  unit = (double)List[at].v;
	     } else {
		  unit = 1.0;
	     }
	}
/* D0 end */

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KNUMBER ){
		v = (double)List[at].v / unit;
		val = (int)(acos( v ) / PI * 180.0 * unit * 10.0);
		if( val % 10 < 5 ){
			val = val / 10;
		} else {
			val = val / 10 + 1;
		} 
		return( cons(val, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_atan()                                   */
/*                                                                    */
/*             Lisp com: INT(value*100)                               */
/*                                                                    */
/*             return  Lisp: INT(degree)                              */
/*                                                                    */
/*             function: atan()                                       */
/*                                                                    */
/**********************************************************************/
Lisp func_atan(Lisp com)
{
	Lisp at;
	double v;
	int val;
/* D0 add */
	double unit;

	if( (at = search_global( "#floatunit" )) == NIL ){
	     unit = 1.0;
	} else {
	     if( List[at].kind == KNUMBER ){
		  unit = (double)List[at].v;
	     } else {
		  unit = 1.0;
	     }
	}
/* D0 end */

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	if( List[at].kind == KNUMBER ){
		v = (double)List[at].v / unit;
		val = (int)(atan( v ) / PI * 180.0 * 10.0 * unit);
		if( val % 10 < 5 ){
			val = val / 10;
		} else {
			val = val / 10 + 1;
		} 
		return( cons(val, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_atan2()                                  */
/*                                                                    */
/*             Lisp com: INT                                          */
/*                       INT                                          */
/*             return  Lisp: INT(degree)                              */
/*                                                                    */
/*             function: atan(com1/com2)                              */
/*                                                                    */
/**********************************************************************/
Lisp func_atan2(Lisp com)
{
	Lisp at;
	Lisp at1, at2;
	double v;
	int val;
/* D0 add */
	double unit;

	if( (at = search_global( "#floatunit" )) == NIL ){
	     unit = 1.0;
	} else {
	     if( List[at].kind == KNUMBER ){
		  unit = (double)List[at].v;
	     } else {
		  unit = 1.0;
	     }
	}
/* D0 end */

	at = eval_exp(com);
	if( List[at].kind != KLIST ){
		return( cons(NIL, KNIL, NIL) );
	} 
	at1 = List[at].v;
	if( List[List[at].next].kind != KLIST ){
		return( cons(NIL, KNIL, NIL) );
	}
	at2 = List[List[at].next].v;		
	if( List[at1].kind == KNUMBER && List[at2].kind == KNUMBER ){
		v = (double)List[at1].v / (double)List[at2].v;
		val = (int)(atan( v ) / PI * 180.0 * 10.0 * unit);
		if( val % 10 < 5 ){
			val = val / 10;
		} else {
			val = val / 10 + 1;
		} 
		return( cons(val, KNUMBER, NIL) );
	} else {
		return( cons(NIL, KNIL, NIL) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_val()                                    */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: OMNI                                     */
/*                                                                    */
/*             function: Re-execution of com, val().                  */
/*                                                                    */
/**********************************************************************/
Lisp func_val(Lisp com)
{
	Lisp at;

	at = eval_exp(com);
/* B0 add */
	if( List[at].kind != KLIST ){
		return( NIL );
	}
	at = List[at].v;
/* B0 end */
	at = exec_exp(at);
	return( at );
}


/**********************************************************************/
/*                                                                    */
/*             function func_mask()                                   */
/*                                                                    */
/*             Lisp mask: MASK, NIL                                   */
/*                                                                    */
/*             return  Lisp: MASK                                     */
/*                                                                    */
/*             function: if mask=NIL then return current mask,        */
/*                      else set MASK and return mask.                */
/**********************************************************************/
Lisp func_mask( Lisp mask )
{
	int id, num;
/* B0 add */
	mask = eval_exp( mask );
	if( mask == NIL ){
		id = 0;
	} else {
		if( List[mask].kind != KLIST ){
			return( NIL );
		}
		mask = List[mask].v;
		if( List[mask].kind == KMASK ){
			id = List[mask].v;
			set_area( mask );
		} else {
			return( cons( NIL, KNIL, NIL ) );
		}
	}
/* B0 end */
/*	if( mask == NIL ){
		id = 0;
	} else if( List[mask].kind == KMASK ){
		id = List[mask].v;
		set_area( mask );
	} else {
		return( cons( NIL, KNIL, NIL ) );
	}*/
	num = make_tmpmask( id );
	return( cons( num, KMASK, NIL ) );
}


/**********************************************************************/
/*                                                                    */
/*             function axtoi()                                       */
/*                                                                    */
/*             char *str: [0-9a-fA-F]+                                */
/*             int max:   length of str                               */
/*                                                                    */
/*             return  int: converted value                           */
/*                                                                    */
/*             function: Heximal string to integer.                   */
/*                                                                    */
/**********************************************************************/
int axtoi( char *str, int max )
{
	int ret;
	int i;

	ret = 0;
	for( i = 0; str[i] && i < max; i++ ){
		if( isdigit( str[i] ) ){
			ret = ret * 16 + str[i] - '0';
		} else if( isxdigit( str[i] ) ){
			ret = ret * 16 + toupper( str[i] ) - 'A' + 10;
		} else if( str[i] == ' ' ){
			continue;
		} else {
			return( -1 );
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function func_axtoi()                                  */
/*                                                                    */
/*             Lisp ls: LITERAL                                       */
/*                                                                    */
/*             return  Lisp: INT                                      */
/*                                                                    */
/*             function: axtoi()                                      */
/*                                                                    */
/**********************************************************************/
Lisp func_axtoi( Lisp ls )
{
	int num;

	ls = eval_exp(ls);
/* B0 add */
	if( List[ls].kind != KLIST ){
		return( NIL );
	}
	ls = List[ls].v;
/* B0 end */
	if( List[ls].kind == KLITERAL ){
		if( (num = axtoi( tsv(ls), 8)) == -1 ){
			return( cons( NIL, KNIL, NIL ) );
		} else {
			return( cons( num, KNUMBER, NIL ) );
		}
	} else {
		return( cons( NIL, KNIL, NIL ) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_funcs()                                  */
/*                                                                    */
/*             Lisp com: INT 1 : usr function only                    */
/*                           2 : builtin function only                */
/*                           3 : macro only                           */
/*                       NIL or INT other: all  function              */
/*             return  Lisp: LIST{LITERAL, ...}                       */
/*                                                                    */
/*             function: return the LIST of registrated function names.*/
/*                                                                    */
/**********************************************************************/
Lisp func_funcs( Lisp com )
{
	Lisp at;
/* add B0 */
	int kind;

	com = eval_exp(com);
	if( com == NIL ){
		kind = 0;
	} else {
		if( List[com].kind != KLIST ){
			return( NIL );
		} else if( List[List[com].v].kind != KNUMBER ){
			kind = 0;
		} else if( List[List[com].v].v == 1 ){
			kind = 3;
		} else if( List[List[com].v].v == 2 ){
			kind = 4;
		} else if( List[List[com].v].v == 3 ){
			kind = 5;
		} else {
			kind = 0;
		}
	}

	at = search_vars( kind );
/* B0	at = search_vars( 0 );*/
	if( at == NIL ){
		display( M_NOREGST );
	}
	return( at );
}

/**********************************************************************/
/*                                                                    */
/*             function func_vars()                                   */
/*                                                                    */
/*             Lisp com: not used                                     */
/*                                                                    */
/*             return  Lisp: LIST{LITERAL, ...}                       */
/*                                                                    */
/*             function: return the LIST of registrated variables.    */
/*                                                                    */
/**********************************************************************/
Lisp func_vars( Lisp com )
{
	Lisp at;

	at = search_vars( 1 );
	if( at == NIL ){
		display( M_NOREGST );
	}
	return( at );
}

/**********************************************************************/
/*                                                                    */
/*             function func_sharps()                                 */
/*                                                                    */
/*             Lisp com: not used                                     */
/*                                                                    */
/*             return  Lisp: LIST{LITERAL, ... }                      */
/*                                                                    */
/*             function: return the LIST of registrated #.            */
/*                                                                    */
/**********************************************************************/
Lisp func_sharps( Lisp com )
{
	Lisp at;

	at = search_vars( 2 );
	if( at == NIL ){
		display( M_NOREGST );
	}
	return( at );
}

/**********************************************************************/
/*                                                                    */
/*             function func_boundp()                                 */
/*                                                                    */
/*             Lisp ls: LITERAL                                       */
/*                                                                    */
/*             return  Lisp: INT 1 or NIL                             */
/*                                                                    */
/*             function: LITERAL symbol is registrated or not.        */
/*                                                                    */
/**********************************************************************/
Lisp func_boundp( Lisp ls )
{
	Lisp at;

/* B0 add */
	if( List[ls].kind != KLIST ){
		return( NIL );
	}
	ls = List[ls].v;
/* B0 end */
	if( List[ls].kind != KSYMBOL ){
		if( (ls = eval_exp( ls )) == NIL ){
			return( cons( NIL, KNIL, NIL ) );
		}
	}
	if( List[ls].kind == KSYMBOL || List[ls].kind == KLITERAL ){
		if( (at = get_var( List[ls].v )) ){
			return( cons( List[at].kind, KNUMBER, NIL ) );
		} else if( (at = get_mac( List[ls].v )) ){
			return( cons( List[at].kind, KNUMBER, NIL ) );
		} else {
			return( cons( NIL, KNIL, NIL ) );
		}			
	} else {
		return( cons( NIL, KNIL, NIL ) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_mkpos()                                  */
/*                                                                    */
/*             Lisp com: LIST{INT,INT}                                */
/*                                                                    */
/*             return  Lisp: POSITION                                 */
/*                                                                    */
/*             function: Make Position.                               */
/*                                                                    */
/**********************************************************************/
Lisp func_mkpos( Lisp com )
{
	Lisp at;
	Lisp cdr;
	Lisp car;

	at = eval_exp(com);
	if( List[at].kind != KLIST ){
		return( cons( NIL, KNIL, NIL ) );
	}
	car = List[at].v;
	cdr = List[List[at].next].v;
	if( List[car].kind != KNUMBER ){
		return( cons( NIL, KNIL, NIL ) );
	}
	if( List[cdr].kind != KNUMBER ){
		return( cons( NIL, KNIL, NIL ) );
	}
	return( cons( car, KPOSITION, cdr ) );
}
	
/**********************************************************************/
/*                                                                    */
/*             function func_gets()                                   */
/*                                                                    */
/*             Lisp com: not used                                     */
/*                                                                    */
/*             return  Lisp: LITERAL                                  */
/*                                                                    */
/*             function: Read stdin and return the keyin string.      */
/*                                                                    */
/**********************************************************************/
Lisp func_gets( Lisp com )
{
	char buff[256];

	if( check_value( GLOBAL, "#noint" ) ){
		return( cons( NIL, KNIL, NIL ) );
	}
	end_raw_tty();
	fgets( buff, 255, stdin );
	reset_raw_tty();
	buff[strlen(buff)-1] = '\0';
	return( cons( tmp_strdup(buff), KLITERAL, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_echo()                                   */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: OMNI(com)                                */
/*                                                                    */
/*             function: echo()                                       */
/*                                                                    */
/**********************************************************************/
Lisp func_echo( Lisp com )
{
/* B2	Lisp org;*/

/* B0 add */
	com = eval_exp( com );
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	com = List[com].v;
/* B0 end */
/*B2	org = search_hash( GLOBAL, "#silent" );
	set_hash( GLOBAL, "#silent", g_cons( 0, KNUMBER, NIL ) );*/
	set_hash( TMP, "#silent", cons(0, KNUMBER, NIL) );
	display_command( com );
/*B2	if( org ){
		set_hash( GLOBAL, "#silent", org );
	}*/
	return( com );
}

/**********************************************************************/
/*                                                                    */
/*             function func_system()                                 */
/*                                                                    */
/*             Lisp com: LITERAL                                      */
/*                                                                    */
/*             return  Lisp: INT(ret value of system())               */
/*                                                                    */
/*             function: system()                                     */
/*                                                                    */
/**********************************************************************/
Lisp func_system( Lisp com )
{
	int ret;
	char *exec;
	Lisp at;

/* B0 add */
	com = eval_exp( com );
/* B0	if( com != NIL && List[com].kind != KLITERAL ){
		return( cons( NIL, KNIL, NIL ) );
	}
*/
	if( com == NIL ){
/* B2		if( (at = search_hash( GLOBAL, "#shell" )) != NIL ){
			exec = gsv(at);*/
/* B2 created search_global() */
		if( (at = search_global( "#shell" )) != NIL ){
			exec = tsv(at);
		} else {
			return( cons( NIL, KNIL, NIL ) );	
		}
	} else {
/* B0 add */
		if( List[com].kind != KLIST ){
			return( NIL );
		}
		com = List[com].v;
		if( List[com].kind != KLITERAL ){
			return( cons( NIL, KNIL, NIL ) );
		}
/* B0 end */
		exec = tsv(com);
	}
	end_raw_tty();
	flush_operation( 1 );
	ret = system( exec );
	reset_raw_tty();
	return( cons( ret, KNUMBER, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_error()                                  */
/*                                                                    */
/*             Lisp com: LITERAL or NIL                               */
/*                                                                    */
/*             return  Lisp: NIL (not used)                           */
/*                                                                    */
/*             function: If NIL, echo "USERBREAK" and stop.           */
/*                       Else echo com and stop.                      */
/**********************************************************************/
Lisp func_error( Lisp com )
{
	com = eval_exp(com);
	if( com == NIL ){
		warning( W_USR );
/* add B0 */
	}
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	com = List[com].v;
	if( List[com].kind == KLITERAL ){
/* end B0 */
/*B0	} else if( List[com].kind == KLITERAL ){*/
		warning( W_OMNI, tsv(com) );
	} else {
		warning( W_USR );
	}
	if( com == NIL ) {
		warning( W_USR );
	}
	return( NIL );
}

/**********************************************************************/
/*                                                                    */
/*             function p2l()                                         */
/*                                                                    */
/*             Lisp at: POSITION or POSLIST                           */
/*                                                                    */
/*             return  Lisp: LIST                                     */
/*                                                                    */
/*             function: called from func_pos2list(), recursive.      */
/*                                                                    */
/**********************************************************************/
static Lisp p2l( Lisp at )
{
	if( List[at].kind == KPOSITION ){
		return( at );
	} else if( List[at].kind == KPOSLIST ){
		return( cons( List[at].v, KLIST, p2l( List[at].next ) ) );
	} else {
		return( NIL );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function func_pos2list()                               */
/*                                                                    */
/*             Lisp com: POSITION or POSLIST                          */
/*                                                                    */
/*             return  Lisp: LIST{ {INT,INT},...}                     */
/*                                                                    */
/*             function: POSITION/POSLIST to LIST                     */
/*                                                                    */
/**********************************************************************/
Lisp func_pos2list( Lisp com )
{
	com = eval_exp(com);
/* add B0 */
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	com = List[com].v;
/* end B0 */
	if( List[com].kind != KPOSLIST ){
		return( NIL );
	}
	return( p2l( com ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_readposfile()                            */
/*                                                                    */
/*             Lisp com: LITERAL(filename)                            */
/*                                                                    */
/*             return  Lisp: POSLIST                                  */
/*                                                                    */
/*             function: read the posfile and return POSLIST          */
/*             add A0                                                 */
/**********************************************************************/

/* add A0 */
Lisp func_readposfile( Lisp com )
{
	struct Shead sh;
	FILE *fp;
	Lisp ret;
	int i;
	struct Pos pos;
	Lisp x, y;
	Lisp *p;

	com = eval_exp( com );
/* add B0 */
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	com = List[com].v;
/* end B0 */
	if( List[com].kind != KLITERAL ){
		return( NIL );
	}
	if( search_path( tsv(com), MSELECT, Work ) == -1 ){
		warning( W_NOFILE );
	}
	if( (fp = fopen( Work, "r" )) == NULL ){
		return( NIL );
	}
	if( fread( &sh, sizeof(struct Shead), 1, fp ) != 1 ){
		fclose( fp );
		return( NIL );
	}
	if( (p = (Lisp *)malloc( sh.defnum * sizeof(Lisp) )) == NULL ){
		fclose( fp );
		warning( W_MALLOC );
	}
	for( i = 0; i < sh.defnum; i++ ){
		if( fread( &pos, sizeof(struct Pos), 1, fp ) != 1 ){
			fclose( fp );
			free( p );
			return( NIL );
		}
		x = cons( pos.x, KNUMBER, NIL );
		y = cons( pos.y, KNUMBER, NIL );
		p[i] = cons( x, KPOSITION, y );
	}
	ret = p[sh.defnum - 1];
	for( i = sh.defnum - 2; i >= 0; i-- ){
		ret = cons( p[i], KPOSLIST, ret );
	}
	fclose( fp );
	free( p );
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function func_writeposfile()                           */
/*                                                                    */
/*             Lisp com: LITERAL(filename)                            */
/*                       POSLIST                                      */
/*             return  Lisp: INT 1(success), NIL(fail)                */
/*                                                                    */
/*             function: write POSLIST to filename                    */
/*                                                                    */
/**********************************************************************/
Lisp func_writeposfile( Lisp com )
{
	struct Shead sh;
	FILE *fp;
	Lisp file, posl;
	int i;
	struct Pos pos;

	com = eval_exp( com );
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	file = List[com].v; posl = List[List[com].next].v;
	if( List[file].kind != KLITERAL ){
		return( NIL );
	}
	if( List[posl].kind != KPOSLIST ){
		return( NIL );
	}
	search_path( tsv(com), MSELECT, Work );
	if( (fp = fopen( Work, "w" )) == NULL ){
		return( NIL );
	}
	sh.skind = SKFILE;
	if( fwrite( &sh, sizeof(struct Shead), 1, fp ) != 1 ){
		fclose( fp );
		return( NIL );
	}
	for( i = 0; ; i++ ){
		if( List[posl].kind == KPOSITION ){
			pos.x = List[List[posl].v].v;
			pos.y = List[List[posl].next].v;
			if( fwrite( &pos, sizeof(struct Pos), 1, fp ) != 1 ){
				fclose( fp );
				return( NIL );
			}
			break;
		} else if( List[posl].kind == KPOSLIST ){
			pos.x = List[List[List[posl].v].v].v;
			pos.y = List[List[List[posl].v].next].v;
			if( fwrite( &pos, sizeof(struct Pos), 1, fp ) != 1 ){
				fclose( fp );
				return( NIL );
			}
			posl = List[posl].next;
		} else if( List[posl].kind == KLIST ){
			if( List[List[posl].v].kind == KPOSITION ){
				pos.x = List[List[List[posl].v].v].v;
				pos.y = List[List[List[posl].v].next].v;
				if( fwrite( &pos, sizeof(struct Pos), 1, fp ) != 1 ){
					fclose( fp );
					return( NIL );
				}
			} else {
				printf( "not POSLIST No %d:ignore\n", i-- );
			}
			posl = List[posl].next;
		} else {
			fclose( fp );
			return( NIL );
		}
	}
	sh.defnum = i;
	rewind( fp );
	if( fwrite( &sh, sizeof(struct Shead), 1, fp ) != 1 ){
		fclose( fp );
		return( NIL );
	}
	fclose( fp );
	return( cons(1,KNUMBER,NIL) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_sscanf()                                 */
/*                                                                    */
/*             Lisp com: LITERAL(original)                            */
/*                       LITERAL(format)                              */
/*             return  Lisp: LIST{ OMNI, ...}                         */
/*                                                                    */
/*             function: sscanf the original by format, return LIST.  */
/*                                                                    */
/**********************************************************************/
Lisp func_sscanf( Lisp com )
{
	Lisp buf, form, ret;
	int num, i;
	struct args {
		int kind;
		void *pt;
	};
	struct args args[10];
	char argbuff[512];
	char *argp, *p;
	int status;

/* add B0 */
	com = eval_exp( com );
/* end B0 */
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	buf = List[com].v; form = List[List[com].next].v;
	if( List[buf].kind != KLITERAL || List[form].kind != KLITERAL ){
		return( NIL );
	}
	num = 0; argp = argbuff; status = 0;
	for( p = tsv(form); *p; p++ ){
		if( status == 1 ){
			switch( *p ){
			case 's':
				args[num].kind = 1;
				args[num].pt = argp;
				argp += 50;
				num++;
				break;
			case 'c':
				args[num].kind = 2;
				args[num].pt = argp;
				argp += 2;
				num++;
				break;
			case 'd':
			case 'i':
			case 'o':
			case 'x':
			case 'u':
				args[num].kind = 0;
				args[num].pt = argp;
				argp += sizeof(int);
				num++;
				break;
			default:
				return( NIL );
				break;
			}
			status = 0;
		} else if( status == 0 && *p == '%' ){
			if( num > 10 ){
				return( NIL );
			}
			if( argp >= argbuff + sizeof(argbuff) ){
				return( NIL );
			}
			status = 1;
		}
	}
	num = sscanf( tsv(buf), tsv(form), args[0].pt, args[1].pt, args[2].pt, 
	       args[3].pt, args[4].pt, args[5].pt, args[6].pt, args[7].pt, 
	       args[8].pt, args[9].pt );
	ret = NIL;
	for( i = num - 1; i >= 0; i-- ){
		if( args[i].kind == 0 ){
			ret = cons( cons( *(int *)args[i].pt, KNUMBER, NIL ), KLIST, ret );
		} else if( args[i].kind == 1 ){
			ret = cons( cons( tmp_strdup((char *)args[i].pt), KLITERAL, NIL ), KLIST, ret );
		} else {
			*((char *)args[i].pt+1) = '\0';
			ret = cons( cons( tmp_strdup((char *)args[i].pt), KLITERAL, NIL ), KLIST, ret );
		}
	}
	return( ret );
}
/* end A0 */
