/**********************************************************************/
/*                                                                    */
/*        builtin.c -- gedo builtin function routine                  */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*        rev B2:  1997. 6. 5  create new, for addtional funcs at B2. */
/*        rev D2:  1999.12.11: add func_read                          */
/*            D2               add to kill all child process          */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <signal.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

/**********************************************************************/
/*                                                                    */
/*             function func_let()                                    */
/*                                                                    */
/*             Lisp com: LIST, POSITION or LITERAL                    */
/*                                                                    */
/*             return  Lisp: OMNI                                     */
/*                                                                    */
/*             function: Macro car not evaluate.                      */
/*             B2 added                                               */
/**********************************************************************/
Lisp func_let( Lisp com )
{
	Lisp val;
	
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	val = List[com].next;
	com = List[com].v;
	if( val != NIL ){
		val = eval_exp( List[val].v );
	}

	set_hash( TMP, tsv(com), val );
	return( val );
}

/**********************************************************************/
/*                                                                    */
/*             function func_cell()                                   */
/*                                                                    */
/*             Lisp com: LIST of SYMBOL or NUMBER                     */
/*                                                                    */
/*             return  Lisp: COMMAND                                  */
/*                                                                    */
/*             function: Make command cells for expert.               */
/*             B2 added                                               */
/**********************************************************************/
Lisp func_cell( Lisp com )
{
	Lisp v, kind, next, ret;

	if( List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[com].v ){
		v = List[com].v;
	} else {
		return( NIL );
	}
	if( List[com].next ){
		if( List[List[com].next].v ){
			kind = List[List[com].next].v;
			kind = eval_exp(kind);
		} else {
			return( NIL );
		}
	} else {
		return( NIL );
	}
	if( List[v].kind == KNUMBER ){
		v = List[v].v;
	} else if( List[v].kind == KLITERAL ){
		v = List[v].v;
	}
	if( List[kind].kind != KNUMBER ){
		return( NIL );
	}
	next = List[List[List[com].next].next].v;
	if( next ){
		if( List[next].kind != KSYMBOL ){
			next = eval_exp( next );
		}
		if( List[next].kind == KNIL ){
			next = NIL;
		} else if( List[next].kind == KNUMBER && List[next].v == 0 ){
			next = NIL;
		}
	}
	ret = cons( v, List[kind].v, next );
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function func_dot()                                    */
/*                                                                    */
/*             Lisp com: LIST of SYMBOL or NUMBER                     */
/*                                                                    */
/*             return  Lisp: COMMAND                                  */
/*                                                                    */
/*             function: Make command doted cells for expert.         */
/*             B2 added                                               */
/**********************************************************************/
Lisp func_dot( Lisp com )
{
	Lisp fst, snd;

	if( List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[com].v ){
		fst = List[com].v;
		if( List[fst].kind != KSYMBOL ){
			fst = eval_exp( fst );
		}
	} else {
		return( NIL );
	}
	if( List[com].next ){
		if( List[List[com].next].v ){
			snd = List[List[com].next].v;
		} else {
			return( NIL );
		}
	} else {
		return( NIL );
	}
	return( cons( List[fst].v, List[fst].kind, cons( List[snd].v, List[snd].kind, List[snd].next ) ) );
}
/**********************************************************************/
/*                                                                    */
/*             function func_relax()                                  */
/*                                                                    */
/*             Lisp com: OMNI                                         */
/*                                                                    */
/*             return  Lisp: LIST                                     */
/*                                                                    */
/*             function: Make argments list.                          */
/*             B2 added                                               */
/**********************************************************************/
Lisp func_relax( Lisp com )
{
	return( com );
}

/**********************************************************************/
/*                                                                    */
/*             function func_symbol()                                 */
/*                                                                    */
/*             Lisp com: LITERAL                                      */
/*                                                                    */
/*             return  Lisp: SYMBOL                                   */
/*                                                                    */
/*             function: Convert LITERAL to SYMBOL.                   */
/*             B2 added                                               */
/**********************************************************************/
Lisp func_symbol( Lisp com )
{
	Lisp v, ret;

	v = eval_exp(com);
	if( v ){
		v = List[v].v;
		if( List[v].kind == KLITERAL ){
			ret = cons( tmp_strdup(tsv(v)), KSYMBOL, NIL );
			return( ret );
		}
	}
	return( v );
}

/**********************************************************************/
/*                                                                    */
/*             function match()                                       */
/*                                                                    */
/*             Lisp from : match orginal                              */
/*             Lisp toat : match destination                          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : do_match() call this for each match process.*/
/*                                                                    */
/**********************************************************************/
static void match( Lisp from, Lisp toat )
{
	Lisp first,second;
	Lisp obj1, obj2;

	switch( List[toat].kind ){
	case KSYMBOL:
		if( from == NIL ){
			obj1 = cons( NIL, KNIL, NIL );
		} else {
			obj1 = from;
		}
		set_clever_name( tsv(toat), obj1 );
		return;
		break;
	case KLIST:
		second = List[List[toat].next].v;
		break;
	case KPOSITION:
		second = List[toat].next;
		break;
	default:
		return;
		break;
	}
	first = List[toat].v;
	if( from == NIL ){
		obj1 = cons( NIL, KNIL, NIL );
		obj2 = cons( NIL, KNIL, NIL );
	} else {
		obj1 = eval_exp(List[from].v);
		obj2 = eval_exp(List[from].next);
	}
	if( List[first].kind == KSYMBOL ){
		set_clever_name( tsv(first), obj1 );
	}
	if( List[second].kind == KSYMBOL ){
		set_clever_name( tsv(second), obj2 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_match()                                    */
/*                                                                    */
/*             Lisp com : argment of match command                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to match command.                 */
/*                                                                    */
/**********************************************************************/
Lisp func_match( Lisp com )
{
	Lisp from, toat;
	Lisp at1, at2;

	if( com == NIL ){
		return( NIL );
	}
	at2 = List[com].v; 
	if( List[com].next == NIL ){
		return( NIL );
	}
	at1 = List[List[com].next].v;
	if( at1 == NIL || at2 == NIL ){
		return( NIL );
	}

	if( List[at1].kind == KSHARP ){
		from = expand_macro( List[at1].v );
	} else if( List[at1].kind == KSYMBOL ){
		from = get_variable( List[at1].v );
	} else {
		from = eval_exp(at1);
	}
	if( List[at2].kind == KSHARP ){
		toat = expand_macro( List[at2].v );
	} else {
		toat = at2;
	}

	while( 1 ){
		if( List[from].kind == KPOSLIST || List[from].kind == KLIST ){
			if( List[toat].kind == KLIST ){
				if( List[toat].next == NIL ){
					if( List[List[toat].v].kind == KSYMBOL ){
						set_clever_name( tsvv(toat), from );
					} else {
						match( List[from].v, List[toat].v );
					}
					break;
				} else {
					match( List[from].v, List[toat].v );
				}
				if( (from = List[from].next) == NIL ){
					break;
				}
				toat = List[toat].next;
			} else {
				warning( W_ILLKIND, List[toat].kind, "match" );
			}
		} else if( List[from].kind == KPOSITION ){
			if( List[toat].kind == KLIST ){
				match( from, List[toat].v );
				toat = List[toat].next;
				while( toat != NIL ){
					if( List[List[toat].v].kind == KSYMBOL ){
						match( NIL, List[toat].v );
						break;
					} else if( List[List[toat].v].kind == KPOSITION ){
						match( NIL, List[toat].v );
						break;
					}
					toat = List[toat].next;
				}
			} else {
				warning( W_ILLKIND,  List[toat].kind, "match" );
			}
			break;
		} else {
			warning( W_ILLKIND, List[from].kind, "match" );
		}
	}
	return( at1 );
}

/**********************************************************************/
/*                                                                    */
/*             function do_set()                                      */
/*                                                                    */
/*             Lisp com : argment of set command                      */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to set command.                   */
/*                                                                    */
/**********************************************************************/
Lisp func_set( Lisp com )
{
	Lisp at;
	Lisp sp;

	if( com == NIL ){
		return( NIL );
	}
	sp = List[com].v;
	if( List[com].next == NIL ){
		return( NIL );
	}
	at = List[List[com].next].v;

	at = eval_exp(at);
	if( List[sp].kind != KSHARP && List[sp].kind != KSYMBOL ){
		return( NIL );
	}
	set_hash( GLOBAL, tsv(sp), dup_t2gLisp(at) );
	return( at );
}

/**********************************************************************/
/*                                                                    */
/*             function func_map()                                    */
/*                                                                    */
/*             Lisp com : argment of set command                      */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to set command.                   */
/*                                                                    */
/**********************************************************************/
Lisp func_map( Lisp com )
{
	Lisp at;
	Lisp fun;
	Lisp ls;
	Lisp ret, lp, lp2;

	if( com == NIL ){
		return( NIL );
	}
	fun = List[com].v;
	if( List[fun].kind != KSYMBOL ){
		fun = eval_exp( fun );
		if( fun == NIL || List[fun].kind != KLITERAL ){
			return( NIL );
		}
		fun = cons( List[fun].v, KSYMBOL, NIL );
	}

	if( List[com].next == NIL ){
		return( NIL );
	}
	at = eval_exp( List[List[com].next].v );

	ret = cons( NIL, KLIST, NIL );
	lp = ret;
	for( ; at != NIL; at = List[at].next ){
		lp2 = lp;
		lp = cons( NIL, KLIST, NIL );
		List[lp2].next = lp;
	  	ls = cons( fun, KFUNC, cons( List[at].v, KLIST, NIL ) );
		List[lp].v = eval_exp( ls );
	}
	return( List[ret].next );
}
/**********************************************************************/
/*                                                                    */
/*             function func_apply()                                   */
/*                                                                    */
/*             Lisp com : argment of set command                      */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to set command.                   */
/*                                                                    */
/**********************************************************************/
Lisp func_apply( Lisp com )
{
	Lisp fun;

	if( com == NIL ){
		return( NIL );
	}
	fun = List[com].v;
	if( List[fun].kind != KSYMBOL ){
		fun = eval_exp( fun );
		if( fun == NIL || List[fun].kind != KLITERAL ){
			return( NIL );
		}
		fun = cons( List[fun].v, KSYMBOL, NIL );
	}
	return( eval_exp( cons( fun, KFUNC, List[com].next ) ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_plugs()                                  */
/*                                                                    */
/*             Lisp com: INT 1 : loaders                              */
/*                           2 : selecters                            */
/*                           3 : effecters                            */
/*                       NIL or INT other: all plugs                  */
/*             return  Lisp: LIST{LITERAL, ...}                       */
/*                                                                    */
/*             function: return the LIST of registrated plugs        .*/
/*                                                                    */
/**********************************************************************/
Lisp func_plugs( Lisp com )
{
	int kind;

	com = eval_exp(com);
	if( com == NIL ){
		kind = 0;
	} else {
		if( List[com].kind != KLIST ){
			return( NIL );
		} else if( List[List[com].v].kind != KNUMBER ){
			kind = 0;
		} else if( List[List[com].v].v == 1 ){
			kind = 1;
		} else if( List[List[com].v].v == 2 ){
			kind = 2;
		} else if( List[List[com].v].v == 3 ){
			kind = 3;
		} else {
			kind = 0;
		}
	}

	return( search_plugs( kind, 0 ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_loadedplugs()                            */
/*                                                                    */
/*             Lisp com: INT 1 : loaders                              */
/*                           2 : selecters                            */
/*                           3 : effecters                            */
/*                       NIL or INT other: all plugs                  */
/*             return  Lisp: LIST{LITERAL, ...}                       */
/*                                                                    */
/*             function: return the LIST of now loaded plugs          */
/*                                                                    */
/**********************************************************************/
Lisp func_loadedplugs( Lisp com )
{
	int kind;

	com = eval_exp(com);
	if( com == NIL ){
		kind = 0;
	} else {
		if( List[com].kind != KLIST ){
			return( NIL );
		} else if( List[List[com].v].kind != KNUMBER ){
			kind = 0;
		} else if( List[List[com].v].v == 1 ){
			kind = 1;
		} else if( List[List[com].v].v == 2 ){
			kind = 2;
		} else if( List[List[com].v].v == 3 ){
			kind = 3;
		} else {
			kind = 0;
		}
	}

	return( search_plugs( kind, 1 ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_split()                                  */
/*                                                                    */
/*             Lisp com: LITERAL src string                           */
/*                       LITERAL terminator                           */
/*             return  Lisp: LIST{LITERAL, ...}                       */
/*                                                                    */
/*             function: make LIST arg1 LITERAL splitted by arg2      */
/*                                                                    */
/**********************************************************************/
Lisp func_split( Lisp com )
{
	Lisp src, sp;
	Lisp ret, ls;
	char *key, *p, *prev;
	char buf[2];
	char work[256];

	com = eval_exp( com );
	if( List[com].kind != KLIST ){
		return( NIL );
	}
	if( (src = List[com].v) == NIL ){
		return( NIL );
	}
	if( (sp = List[List[com].next].v) == NIL ){
		return( cons( src, KLIST, NIL ) );
	}
	if( List[src].kind != KLITERAL ){
		return( NIL );
	}
	if( List[sp].kind == KLITERAL ){
		key = tsv(sp);
	} else if( List[sp].kind == KNUMBER ){
		buf[0] = List[sp].v;
		buf[1] = '\0';
		key = buf;
	} else {
		return( NIL );
	}
	ret = cons( NIL, KLIST, NIL ); ls = ret;
	prev = NULL;
	for( p = tsv(src); *p; p++ ){
		if( prev ){
			if( strchr( key, *p ) ){
				strncpy( work, prev, p - prev );
				work[p-prev] = '\0';
				prev = NULL;
				List[ls].next = cons( cons( tmp_strdup( work ), KLITERAL, NIL ), KLIST, NIL );
				ls = List[ls].next;
			}
		} else {
			if( !strchr( key, *p ) ){
				prev = p;
			}
		}
	}
	if( prev ){
		strncpy( work, prev, p - prev + 1);
		List[ls].next = cons( cons( tmp_strdup( work ), KLITERAL, NIL ), KLIST, NIL );
		ls = List[ls].next;
	}
		
	return( List[ret].next );
}


/**********************************************************************/
/*                                                                    */
/*             function func_strlen()                                 */
/*                                                                    */
/*             Lisp com: LITERAL src string                           */
/*                                                                    */
/*             return  Lisp: INTEGER                                  */
/*                                                                    */
/*             function: return the length of the string.             */
/*                                                                    */
/**********************************************************************/
Lisp func_strlen( Lisp com )
{
	int len;

	com = eval_exp( com );
	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[com].v == NIL 
	   || List[List[com].v].kind != KLITERAL ){
		return( NIL );
	}
	len = strlen( tsvv( com ) );
	return( cons( len, KNUMBER, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_strcut()                                 */
/*                                                                    */
/*             Lisp com: LITERAL src string                           */
/*                       INTEGER cutting position                     */
/*             return  Lisp: LIST{ LITERAL,LITERAL }                  */
/*                                                                    */
/*             function: split the string at the position.            */
/*                                                                    */
/**********************************************************************/
Lisp func_strcut( Lisp com )
{
	Lisp s;
	int len, num;
	char work[256];
	Lisp s1, s2;

	com = eval_exp( com );
	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[com].v == NIL 
	   || List[List[com].v].kind != KLITERAL ){
		return( NIL );
	}
	s = List[com].v;
	len = strlen( tsv(s) );
	if( List[com].next == NIL || List[List[com].next].v == NIL 
	   || List[List[List[com].next].v].kind != KNUMBER ){
		return( NIL );
	}
	num = List[List[List[com].next].v].v;
	if( num > len ){
		s1 = cons( tmp_strdup(tsv(s)), KLITERAL, NIL );
		s2 = cons( NIL, KNIL, NIL );
	} else if( num <= 0 ){
		s1 = cons( NIL, KNIL, NIL );
		s2 = cons( tmp_strdup(tsv(s)), KLITERAL, NIL );
	} else {
		strncpy( work, tsv(s), num );
		work[num] = '\0';
		s1 = cons( tmp_strdup(work), KLITERAL, NIL );
		strcpy( work, tsv(s) + num );
		s2 = cons( tmp_strdup(work), KLITERAL, NIL );
	}
	return( cons( s1, KLIST, cons( s2, KLIST, NIL ) ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_strncmp()                                */
/*                                                                    */
/*             Lisp com: LITERAL string1                              */
/*                       LITERAL string2                              */
/*             return  Lisp: INTEGER                                  */
/*                                                                    */
/*             function: return strncmp result between 2 strings.     */
/*                                                                    */
/**********************************************************************/
Lisp func_strncmp( Lisp com )
{
	Lisp s1, s2, len;
	int ret;

	com = eval_exp( com );

	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[com].v == NIL 
	   || List[List[com].v].kind != KLITERAL ){
		return( NIL );
	}
	s1 = List[com].v;
	com = List[com].next;

	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[com].v == NIL 
	   || List[List[com].v].kind != KLITERAL ){
		return( NIL );
	}
	s2 = List[com].v;
	com = List[com].next;

	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[com].v == NIL 
	   || List[List[com].v].kind != KNUMBER ){
		return( NIL );
	}
	len = List[com].v;

	ret = strncmp( tsv(s1), tsv(s2), List[len].v );
	return( cons( ret, KNUMBER, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_atoi()                                   */
/*                                                                    */
/*             Lisp com: LITERAL src string                           */
/*                                                                    */
/*             return  Lisp: INTEGER                                  */
/*                                                                    */
/*             function: make LITERAL number to INTEGER.              */
/*                                                                    */
/**********************************************************************/
Lisp func_atoi( Lisp com )
{
	int num;

	com = eval_exp( com );
	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[com].v == NIL 
	   || List[List[com].v].kind != KLITERAL ){
		return( NIL );
	}
	num = atoi( tsvv( com ) );
	return( cons( num, KNUMBER, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_sqrt()                                   */
/*                                                                    */
/*             Lisp com: INTEGER number                               */
/*                                                                    */
/*             return  Lisp: INTEGER  number                          */
/*                                                                    */
/*             function: calculate sqrt using #floatunit.             */
/*                                                                    */
/**********************************************************************/
Lisp func_sqrt( Lisp com )
{
	double num;
	int val;
/* D0 add */
	double unit;
	Lisp at;

	if( (at = search_global( "#floatunit" )) == NIL ){
	     unit = 1.0;
	} else {
	     if( List[at].kind == KNUMBER ){
		  unit = (double)List[at].v;
	     } else {
		  unit = 1.0;
	     }
	}
/* D0 end */

	com = eval_exp( com );

	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	com = List[com].v;
	if( List[com].kind != KNUMBER || List[com].v < 0 ){
		return( NIL );
	}

	num = (double)List[com].v / unit;
	val = (int)(sqrt(num) * unit * 10.0);
	if( val % 10 < 5 ){
	     val = val / 10;
	} else {
	     val = val / 10 + 1;
	}
	return( cons( val, KNUMBER, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_exec()                                   */
/*                                                                    */
/*             Lisp com: LITERAL argments                             */
/*                                                                    */
/*             return  Lisp: INTEGER processID of the BG process      */
/*                                                                    */
/*             function: invoke the background process.               */
/*                                                                    */
/**********************************************************************/
Lisp func_exec( Lisp com )
{
	int pid;
	char *prog;
	int i;
	char *ex[20];

	com = eval_exp( com );

	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	if( List[List[com].v].kind != KLITERAL ){
		warning( W_FGUI );
		return( NIL );
	}
	ex[0] = prog = tsvv(com);
	com = List[com].next;
	for( i = 1; i < 20; i++ ){
		if( com == NIL ){
			ex[i] = NULL;
			break;
		} else if( List[com].kind == KLIST ){
			if( List[List[com].v].kind == KLITERAL ){
				ex[i] = tsvv(com);
				com = List[com].next;
			} else {
				break;
			}
		} else {
			break;
		}
	}
	if( (pid = fork()) == 0 ){
		execv( prog, ex );
		perror( "" );
		fprintf( stderr, "child process %s fail\n", prog );
		exit( 1 );
	}
/* D2 delete, NOT need 
	signal( SIGUSR1, SIG_IGN );*/
/* D2 add to kill all child process */
	check_childPID( pid );
/* end D2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display( M_CHDOK, prog );
	}
	return( cons( pid, KNUMBER, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function func_read()                                   */
/*                                                                    */
/*             Lisp com: LITERAL read filename                        */
/*                                                                    */
/*             return  Lisp: LIST{ LITERAL, .... }                    */
/*                                                                    */
/*             function: return the content of the file by LIST.      */
/*                                                                    */
/**********************************************************************/
Lisp func_read( Lisp com )
{
	char *file;
	FILE *fp;
	Lisp v, ret;

	com = eval_exp( com );

	if( com == NIL || List[com].kind != KLIST ){
		return( NIL );
	}
	com = List[com].v;
	if( List[com].kind != KLITERAL ){
		return( NIL );
	}
	file = tsv(com);
	if( (fp = fopen( file, "r" )) == NULL ){
	     return( NIL );
	}
	ret = NIL;
	while( fgets( Work, sizeof(Work) - 1, fp ) ){
	     if( Work[strlen(Work)-1] == '\n' ){
		  Work[strlen(Work)-1] = '\0';
	     }
	     v = cons( tmp_strdup(Work), KLITERAL, NIL );
	     ret = cons( v, KLIST, ret );
	}
	fclose( fp );
	return( ret );
}
