/**********************************************************************/
/*                                                                    */
/*        childs.c -- gedo upper level transaction request            */
/*                                                                    */
/*                by K.sugiura   Ver 1.0  for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*        11.21,1996 A0 add search_path in do_load, set_posfile       */
/*         5.13,1997 B2 change GLOBAL var masked by TMP var           */
/*         6. 3,1997 B2 add global var #loader, #saver, #selecter and */
/*                      effecter in set_NowChild() and do_configer()  */
/*                   C1 if com_rcv() fail, check the plug is alive?   */
/*                   D2 check maxsize can create at do_create()       */
/*                   D2 add to kill all child process                 */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include "gedo.h"
#include "../loader.h"
#include "prot.h"
#include "error.h"

#define DEBUG         0

#define MAXPOSITION  30		/* maximum positions at selection */
#define CHILDSMAX   100		/* plug allocate unit */
#define IDVIEWER     -1		/* virtual index of viewer child No */

static struct Childs Viewer;	/* Child struct of viewer */
static int ChildsNum;		/* number of plugs */
static int ChildsMax;		/* now alloced Child number */
int Gui;			/* gui process No. */
static int NowSection;		/* now kind of section initilize at gedo.ini */
Child NowChild[4];		/* current plug at each loader, saver, */
				/* selecter, effecter */
static char *StrSection[3]	/* kind string */
 = { "loader", "selecter", "effecter" };
static char *StrLoaded[2]
= {  "here", "absent" };


/**********************************************************************/
/*                                                                    */
/*             function init_Childs()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : 0==success -1=alloc fail                 */
/*                                                                    */
/*             function: Alloc Childs struct.                         */
/*                                                                    */
/**********************************************************************/
int init_Childs( void )
{
	if( (Childs = (struct Childs *)malloc(sizeof(struct Childs) * CHILDSMAX)) == NULL ){
		fatal( F_MALLOC, "Childs" );
		return( -1 );
	}
	ChildsMax = CHILDSMAX;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function free_Childs()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Free Childs struct.                         */
/*                                                                    */
/**********************************************************************/
void free_Childs( void )
{
	free( Childs );
}

/**********************************************************************/
/*                                                                    */
/*             function new_Child()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  Child : new Child struct                       */
/*                                                                    */
/*             function : Get new Childs struct if not alloc.         */
/*                                                                    */
/**********************************************************************/
Child new_Child( void )
{
	struct Childs *tmp;

	if( ChildsNum + 1 >= ChildsMax ){
		if( (tmp = (struct Childs *)realloc( Childs, sizeof(struct Childs) * (ChildsMax + CHILDSMAX) )) == NULL ){
			warning( W_CHCOSM );
			return( NIL );
		}
		Childs = tmp;
		ChildsMax += CHILDSMAX;
#if DEBUG
	printf( "Childs realloc occured nowsize=%d\n", ChildsMax );
#endif
	}
	ChildsNum++;
	return( ChildsNum - 1 );
}

/**********************************************************************/
/*                                                                    */
/*             function get_childsnum()                               */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : total number of registerd plug           */
/*                                                                    */
/*             function : Get total number of registerd plug.         */
/*                                                                    */
/**********************************************************************/
int get_childsnum( void )
{
	return( ChildsNum );
}

/**********************************************************************/
/*                                                                    */
/*             function set_section()                                 */
/*                                                                    */
/*             int func : Section kind                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Set current section if read "Section ***"   */
/*                        at gedo.ini.                                */
/**********************************************************************/
void set_section( int func )
{
	if( Init_flag == 0 ){
		warning( W_NOCOMLN );
		return;
	}
	if( func == MLOADER ){
		NowSection = 0;
	} else if( func == MSELECTER ){
		NowSection = 1;
	} else if( func == MEFFECTER ){
		NowSection = 2;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_selkind()                                 */
/*                                                                    */
/*             Child ch : Childs struct at now                        */
/*             Lisp op  : initialize command at gedo.ini              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Initialize Childs struct at gedo.ini.       */
/*                        2nd stage, especially Section selecter.     */
/**********************************************************************/
void set_selkind( Child ch, Lisp op )
{
	int i;
	char *name;
	int len;

	name = lsv(op);
	for( i = 0; i < SK_MAX; i++ ){
		len = strlen(StrSelKind[i]);
		if( strncmp( name, StrSelKind[i], len ) == 0 ){
			Childs[ch].selecter_kind  = i;
			if( name[len] ){
				Childs[ch].maxnum = atoi( &name[len] );
			} else {
				Childs[ch].maxnum = 0;
			}
			break;
		}

	}
	if( i == SK_MAX ){
		Childs[ch].help = global_strdup(name);
	}
	if( (op = LocalList[op].next) ){
		Childs[ch].help = global_strdup( lsv(op) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_child()                                   */
/*                                                                    */
/*             Lisp at : initilize command in gedo.ini                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Initialize plug infomation from gedo.ini.   */
/*                                                                    */
/**********************************************************************/
void set_child( Lisp at )
{
	Lisp op;
	Child ch;
	Lisp found;
	char *name;

	name = lsv(at);
	if( name[0] == '!' ){
		if( (found = search_hash( GLOBAL, &name[1] )) == NIL ){
			display( M_NOHASH, &name[1] );
			return;
		}
		ch = GlobalList[found].v;
		Childs[ch].used = NOUSE;
		return;
	} else {
		if( (found = search_hash( GLOBAL, name )) == NIL ){
			ch = new_Child();
		} else {
			if( GlobalList[found].kind == KPLUG ){
				ch = GlobalList[found].v;
			} else {
				display( M_ILLHASH, name );
				return;
			}
		}
	}
	Childs[ch].name = global_strdup( lsv(at) );
	Childs[ch].kind = NowSection;
	if( (op = LocalList[at].next ) ){
		Childs[ch].path = global_strdup( lsv(op) );
		if( (op = LocalList[op].next) ){
			Childs[ch].option = global_strdup( lsv(op) );
			if( (op = LocalList[op].next) ){
				set_selkind( ch, op );
			}
		}
	} else {
		warning( W_FAILCHD, lsv(at) );
	}
	display( M_INSTALL, StrSection[NowSection], gs(Childs[ch].name), gs(Childs[ch].path) );
	Childs[ch].used = USED;
	if( found == NIL ){
		found = g_cons( ch, KPLUG, NIL );
	}
	set_hash( GLOBAL, lsv(at), found ); 
}


/**********************************************************************/
/*                                                                    */
/*             function search_type()                                 */
/*                                                                    */
/*             int kind : kind of plug                                */
/*             char *type : plug name                                 */
/*                                                                    */
/*             return  Child : plug infomation struct searched by name */
/*                                                                    */
/*             function : search the plug infomation struct by name.  */
/*                                                                    */
/**********************************************************************/
Child search_type( int kind, char *type )
{
	Lisp found;
	Child ret;

	if( (found = search_hash( GLOBAL, type )) == NIL ){
		return( -1 );
	}
	if( GlobalList[found].kind != KPLUG ){
		return( -1 );
	}
	ret = GlobalList[found].v;
	if( Childs[ret].used == USED && Childs[ret].kind == kind ){
		return( ret );
	} else {
		return( -1 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_NowChild()                                */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Set current plug at each kind.              */
/*          B2 : add global var #loader, #saver, #selecter and #effecter*/
/*                                                                    */
/**********************************************************************/
void set_NowChild( void )
{
	int i;

	NowChild[0] = NowChild[1] = NowChild[2] = NowChild[3] = -1;
	for( i = ChildsNum - 1; i >= 0; i-- ){
		if( Childs[i].used == USED ){
			switch( Childs[i].kind ){
			case 0:
				NowChild[0] = i;
				NowChild[1] = i;
				break;
			case 1:
				NowChild[2] = i;
				break;
			case 2:
				NowChild[3] = i;
				break;
			default:
				warning( W_ILLCHD );
				return;
			}
		}
	}
	if( NowChild[0] == -1 ){
		fatal( F_ILLPLUG, "LOADER" );
	} else if( NowChild[1] == -1 ){
		fatal( F_ILLPLUG, "SAVER" );
	} else if( NowChild[2] == -1 ){
		fatal( F_ILLPLUG, "SELECTER" );
	} else if( NowChild[3] == -1 ){
		fatal( F_ILLPLUG, "EFFECTER" );
	}
/* add B2 */
	smart_set_global_literal( "#loader", gs(Childs[NowChild[0]].name) );
	smart_set_global_literal( "#saver", gs(Childs[NowChild[1]].name) );
	smart_set_global_literal( "#selecter", gs(Childs[NowChild[2]].name) );
	smart_set_global_literal( "#effecter", gs(Childs[NowChild[3]].name) );
/* add B2 end */
}

/**********************************************************************/
/*                                                                    */
/*             function child_help()                                  */
/*                                                                    */
/*             int kind : kind of plug                                */
/*                                                                    */
/*             return  Child : selected plug                          */
/*                                                                    */
/*             function : Show all plug at the kind, and set current  */
/*                       plug by selection.                           */
/**********************************************************************/
Child child_help( int kind )
{
	int i;
	char buff[24];
	int num;
	int count;
	Child *chd;

	if( (chd = (Child *)malloc( sizeof(Child) * ChildsNum )) == NULL ){
		warning( W_MALLOC );
		return( NIL );
	}
	for( count = 0, i = 0; i < ChildsNum; i++ ){
		if( Childs[i].kind == kind && Childs[i].used == USED ){
			chd[count++] = i;
			display( M_HELP, count, gs(Childs[i].name), gs(Childs[i].help) );
		}
	}
	display( M_SELECT, count );
	end_raw_tty();
	buff[0] = '\0';
	fgets( buff, 23, stdin );
	reset_raw_tty();
	if( (num = atoi(buff)) == 0 ){
		warning( W_ABORT );
		return( -1 );
	}
	if( num < 0 || num > count ){
		warning( W_ESEL );
		return( -1 );
	}
	num = chd[num-1];
	free( chd );
	return( num );
}

/**********************************************************************/
/*                                                                    */
/*             function check_message()                               */
/*                                                                    */
/*             struct Message *send : sent message to plug            */
/*             struct Message *rcv  : returned result from plug       */
/*                                                                    */
/*             return  int : 0==successed session  -1=failed session  */
/*                                                                    */
/*             function : Return if the result from plug is normal    */
/*                       or not.                                      */
/**********************************************************************/
static int check_message( struct Message *send, struct Message *rcv )
{
	if( send->command != rcv->command ){
		return( -1 );
	}
	if( send->subcommand != rcv->subcommand ){
		return( -1 );
	}
	if( rcv->status == STWARNING ){
		return( -1 );
	}
	if( rcv->status == STFATAL ){
		return( -1 );
	}
	memcpy( send, rcv, sizeof(struct Message) );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function trans_errmes()                                */
/*                                                                    */
/*             int kind : -2==cannot start the plug                   */
/*                        -1==failed at the send proc of GEDO         */
/*                         0==failed at the plug                      */
/*                         1==failed at the receive of GEDO           */
/*             struct Message *mes : returned message from plug       */
/*             struct Childs *child : the plug                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Show the error message occured at the       */
/*                       transaction with plug.                       */
/*   B2: global #notranserror added                                   */
/**********************************************************************/
static void trans_errmes( int kind, struct Message *mes, struct Childs *child )
{
	char *name;

	name = gs(child->name);
	switch( kind ){
	case -2:
		warning( W_ESTRT, StrCommand[mes->command], name );
		break;
	case -1:
		warning( W_ESND, StrCommand[mes->command], name );
		break;
	case 0:
		display( M_TRNSFAIL, StrCommand[mes->command], name, 
			StrSubCom[mes->subcommand], StrErr[mes->errorcode], 
			StrStatus[mes->status] );
		if( mes->command == CSELECT && mes->subcommand == SCEXEC ){
			if( mes->u.sb.skind == SKPOS ){
				print_sbuf( &mes->u.sb );
			} else {
				display( M_FORM, "\t%s\n", mes->u.sf.file );
			}
		} else if( mes->datasize ){
			display( M_FORM, "\t%s\n", mes->u.data );
		}
/* B2 add */
		if( check_value( GLOBAL, "#notranserror" ) ){
			printf( "#notranserror TRUE\n" );
			if( mes->status == STWARNING ){
				printf( "trans error is warning\n" );
				if( mes->errorcode != EUSRBREAK ){
					printf( "it is not user break\n" );
					break;
				}
			}
		}
/* B2 end */
		warning( W_NONE );
		break;
	case 1:

		warning( W_ERCV, StrCommand[mes->command], name );
		break;
	default:
		warning( W_NONE );
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function send_message()                                */
/*                                                                    */
/*             struct Message *mes : the message now send to plug     */
/*             Child child  : infomation of now sending plug          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Send the message to plug from GEDO.         */
/*                                                                    */
/**********************************************************************/
static void send_message( struct Message *mes, Child child )
{
	struct Message rep;
	int ret;
	struct Childs *ch;

	if( child == IDVIEWER ){
		ch = &Viewer;
	} else {
		ch = &Childs[child];
	}
	memset( &rep, 0, sizeof(struct Message) );
	mes->server = SMAIN;
	mes->shmid = NowShmid;
	if( ch->pid == 0 ){
		if( start_child( ch ) == -1 ){
			warning( W_NONE );
		}
	}
	mes->pid[SMAIN] = ProcessId;
	mes->pid[SVIEWER]   = Viewer.pid;
	if( Viewer.pid == 0 || getpgid( Viewer.pid ) == -1 ){ /* C3 change */
		init_viewer();
                mes->pid[SVIEWER] = Viewer.pid;
	}
	switch( mes->command ){
	case CLOAD:
	case CSAVE:
		mes->receiver = SLOADER;
		mes->pid[SLOADER]   = ch->pid;
		break;
	case CSELECT:
		if( Level != 2 && mes->subcommand == SCEXEC ){
			mes->receiver = SVIEWER;
		} else {
			mes->receiver = SSELECTER;
		}
		mes->pid[SSELECTER] = ch->pid;
		break;
	case CEFFECT:
		mes->receiver = SEFFECTER;
		mes->pid[SEFFECTER] = ch->pid;
		break;
	case CSHOW:
	case CPURGE:
		mes->receiver = SVIEWER;
		break;
	default:
		return;
	}
	ret = com_send( mes );
        if( ret == -2 ){
		if( Viewer.pid && getpgid( Viewer.pid ) == -1 ){
			init_viewer();
                        mes->pid[SVIEWER] = Viewer.pid;
		}
		kill( ch->pid, SIGKILL );
		ch->pid = 0;
		if( start_child( ch ) == -1 ){
			warning( W_NONE );
		}
		ret = com_send( mes );
		if( ret < 0 ){
			trans_errmes( ret, mes, ch );
		}
	} else if( ret == -1 ){
		trans_errmes( ret, mes, ch );
	}
	if( com_rcv( &rep ) == -1 ){
		if( errno == EINTR ){	 /* add C1, plug or viewer1 is broken. */
			if( !Error_flag ){	/* Ctrl+C not pressed */
				if( getpgid( Viewer.pid ) == -1 ){
					Viewer.pid = 0;
				} else {
					ch->pid = 0;
				}
			} else {	/* Ctrl+C pressed */
				return;
			}
		}
		trans_errmes( 1, mes, ch );
		return;
	}
	if( check_message( mes, &rep ) ){
		if( rep.status == STFATAL && rep.errorcode != ENOTCREATE ){
			ch->pid = 0;
		}
		trans_errmes( 0, &rep, ch );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_configer()                                 */
/*                                                                    */
/*             Lisp com : config sequence                             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : All configuration of plug.                  */
/*          B2 : add global var #loader, #saver, #selecter and #effecter*/
/**********************************************************************/
void do_configer( Lisp com )
{
	int spec;
	Child at;
	struct Message mes;
	Lisp lp;
	Lisp local_at, conf;

	local_at = NIL;
	memset( &mes, 0, sizeof(struct Message) );
	switch( List[com].v ){
	case MLOADER:
		mes.command = CLOAD;
		spec = 0;
		break;
	case MSAVER:
		mes.command = CSAVE;
		spec = 0;
		break;
	case MSELECTER:
		mes.command = CSELECT;
		spec = 1;
		break;
	case MEFFECTER:
		mes.command = CEFFECT;
		spec = 2;
		break;
	default:
		warning( W_ILLCOM );
		return;
	}
	if( List[com].next == NIL ){
		at = child_help( spec );
		List[com].next = cons( tmp_strdup( gs(Childs[at].name) ), KSYMBOL, NIL );
	} else if( List[List[com].next].kind == KFN ){
		at = child_help( spec );
		local_at = List[List[com].next].v;
		lp = tmp_strdup( gs(Childs[at].name) );
		List[List[com].next].v =  lp;
		List[List[com].next].kind = KSYMBOL;
		lp = local_strdup( gs(Childs[at].name) );
		LocalList[local_at].next = l_cons( lp, KSYMBOL, NIL );
	} else if( List[List[com].next].kind == KSYMBOL ){
		if( (at = search_type( spec, tsnv(com) )) == -1 ){
			warning( W_NONAME, StrSection[spec], tsnv(com) );
		}
	} else {
		warning( W_TRANS, ComStr[List[com].v-1] );
		return;
	}
	switch( mes.command ){
	case CLOAD:
		NowChild[0] = at;
/* B2 */	smart_set_global_literal( "#loader", gs(Childs[at].name) );
		mes.pid[SLOADER] = Childs[at].pid;
		break;
	case CSAVE:
		NowChild[1] = at;
		mes.pid[SLOADER] = Childs[at].pid;
/* B2 */	smart_set_global_literal( "#saver", gs(Childs[at].name) );
		strcpy( SaveType, tsnv(com) );
		break;
	case CSELECT:
		NowChild[2] = at;
/* B2 */	smart_set_global_literal( "#selecter", gs(Childs[at].name) );
		mes.pid[SSELECTER] = Childs[at].pid;
		break;
	case CEFFECT:
		NowChild[3] = at;
/* B2 */	smart_set_global_literal( "#effecter", gs(Childs[at].name) );
		mes.pid[SEFFECTER] = Childs[at].pid;
		break;
	default:
		return;
		break;
	}
	lp = List[List[com].next].next;
	if( lp == NIL ){
/* add #silentplug B2 */
		if( check_value( GLOBAL, "#silentplug" ) == 0 ){
			display_command( com );
			display( M_DONE );
		}
		return;
	}
	switch( List[lp].kind ){
	case KFN:
		mes.subcommand = SCHELP;
		break;
	case KVOID:
		local_at = List[lp].v;
		mes.subcommand = SCCONFIG;
		mes.datasize = 0;
		mes.u.data[0] = '\0';
		break;
	case KLITERAL:
		mes.subcommand = SCCONFIG;
		strcpy( mes.u.data, tsv(lp) );
		mes.datasize = strlen(mes.u.data);
		break;
	default:
		warning( W_ILLLST );
		break;
	}

	if( Level != 1 ){
		send_message( &mes, at );
		if( mes.status == STDONE ){
			conf = l_cons( local_strdup(mes.u.data), KLITERAL, NIL );
			set_hash( LOCAL, "$config", conf );
			if( List[lp].kind == KVOID ){
				List[lp].kind = KLITERAL;
				List[lp].v = tmp_strdup( mes.u.data );
				List[lp].next = NIL;
				if( local_at ){
					LocalList[local_at].next = conf;
				}
			}
			
		}
/* add #silentplug B2 */
		if( check_value( GLOBAL, "#silentplug" ) == 0 ){
			display_command( com );
			display( M_DONE );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function make_pathname()                               */
/*                                                                    */
/*             char *out  : <OUT> pathname                            */
/*             char *in1  : <IN> path                                 */
/*             char *in2  : <IN> pathname or filename                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Append <in1> and <in2> as a file path.      */
/*                                                                    */
/**********************************************************************/
static void make_pathname( char *out, char *in1, char *in2 )
{
	strcpy( out, in1 );
	if( in1[strlen(in1)-1] != '/' ){
		strcat( out, "/" );
	}
	strcat( out, in2 );
}

/**********************************************************************/
/*                                                                    */
/*             function search_path()                                 */
/*                                                                    */
/*             char *name : file name                                 */
/*             int kind   : kind of plug                              */
/*             char *buff : completed path if found                   */
/*                                                                    */
/*             return int : 0==here  -1==not here                     */
/*                                                                    */
/*             function : Search the filename if is or not.           */
/*                                                                    */
/**********************************************************************/
/* A0 add searchpath */
int search_path( char *name, int kind, char *buff )
{
	Lisp lpath;
	char *path;

	if( access( name, R_OK ) == 0 ){
		strcpy( buff, name );
		return( 0 );
	}
	switch( kind ){
	case MLOAD:
/* B2 search_global() created */
/*		lpath = search_hash( GLOBAL, "#path" );
		if( lpath == NIL ){
			return( -1 );
		}
		if( GlobalList[lpath].kind != KLITERAL
		   && GlobalList[lpath].v == NIL ){
			return( -1 );
		}
		strcpy( buff, gsv(lpath) );*/
		lpath = search_global( "#path" );
		if( lpath == NIL ){
			return( -1 );
		}
		if( List[lpath].kind != KLITERAL
		   && List[lpath].v == NIL ){
			return( -1 );
		}
		make_pathname( buff, tsv(lpath), name );
		break;
	case MSELECT:
		path = getenv( "GEDO_LWP" );
		if( path == NULL ){
			return( -1 );
		}
		make_pathname( buff, path, name );
		break;
	case MINCLUDE:
		lpath = search_global( "#gedopath" );
		if( lpath == NIL ){
			return( -1 );
		}
		if( List[lpath].kind != KLITERAL
		   && List[lpath].v == NIL ){
			return( -1 );
		}
		make_pathname( buff, tsv(lpath), "lisp/" );
		strcat( buff, name );
		break;
	default:
		return( -1 );
		break;
	}
	if( access( buff, R_OK ) == 0 ){
		return( 0 );
	} else {
		return( -1 );
	}
}
/* A0 add searchpath end */

/**********************************************************************/
/*                                                                    */
/*             function do_load()                                     */
/*                                                                    */
/*             Lisp com : COMMAND List                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to load command.                  */
/*                                                                    */
/**********************************************************************/
void do_load( Lisp com )
{
	struct Message mes;
	char *name;
/* B2 add */
	Lisp g;
	int len;

	Work[0] = '\0';
/* A0 searchpath
	if( (path = search_hash( GLOBAL, "#path" )) != NIL ){
		if( GlobalList[path].kind == KLITERAL && GlobalList[path].v != NIL ){
			strcpy( Work, gsv(path) );
		}
	}
*/
	if( List[com].next == NIL ||	
	    List[List[com].next].kind == KVOID ){
		if( LoadFileForm[0] == '\0' ){
			warning( W_NOFNAME, "load" );
		}
		name = gs(FileName[FileNum]);
	} else if( List[List[com].next].kind == KLITERAL ){
		name = tsnv(com);
	} else {
		warning( W_LOAD );
		return;
	}
/* A0 add searchpath*/
	if( search_path( name, MLOAD, Work ) == -1 ){
		warning( W_NOFILE );
	}
/*	strcat( Work, name );*/
/* B2 add */
	len = strlen( Work ) + 1;
/* B2 add search_global */
/*	if( (g = search_hash( GLOBAL, "#viewer")) != NIL ){
		if( GlobalList[g].kind == KLITERAL ){
			strcpy( &Work[len], gsv(g) );
			len += strlen(gsv(g));
			len++;
		}
	}*/
	if( (g = search_global( "#viewer" )) != NIL ){
		if( List[g].kind == KLITERAL ){
			strcpy( &Work[len], tsv(g) );
			len += strlen(tsv(g));
			len++;
		}
	}
/* B2 end */
/* add #silentplug B2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display( M_LOADING, Work );
	}

	switch( Level ){
	case 0:
	case 2:
		memset( &mes, 0, sizeof(struct Message) );
		mes.command = CLOAD;
		mes.subcommand = SCEXEC;
		if( Level == 0 ){
			mes.comatr  =  ADISP;
		} else {
			mes.comatr  =  ANODISP;
		}
/* B2		strcpy( mes.u.data, Work );
		mes.datasize = strlen( mes.u.data ); */
		if( len >= 256 ){
			dwarning( "Message is too long\n" );
		}
		memcpy( mes.u.data, Work, len );
		mes.datasize = len;
/* B2 end */
		send_message( &mes, NowChild[0] );
		break;
	case 1:
	default:
		break;
	}
	if( List[List[com].next].kind == KVOID ){
		LocalList[List[List[com].next].v].next = l_cons( local_strdup( name ), KLITERAL, NIL );
		List[List[com].next].kind = KLITERAL;
		List[List[com].next].v = tmp_strdup( name );
		List[List[com].next].next = NIL;
	}
	smart_set_literal( "$loaded", name );
	set_wh();
/* add #silentplug B2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display_command( com );
		display( M_DONE );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_save()                                     */
/*                                                                    */
/*             Lisp com : COMMAND list                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to save command.                  */
/*                                                                    */
/**********************************************************************/
void do_save( Lisp com )
{
	struct Message mes;
	char buff[FILENAME_MAX];
	char *name;
	Lisp path;
/* D2 add for correct save file path */
	int flag;
/* end D2 */

	Work[0] = '\0';
/* B2 add search_global() */
/*	if( (path = search_hash( GLOBAL, "#path" )) != NIL ){
		if( GlobalList[path].kind == KLITERAL && GlobalList[path].v != NIL ){
			strcpy( Work, gsv(path) );
		}
	}*/
	if( (path = search_global( "#path" )) != NIL ){
		if( List[path].kind == KLITERAL && List[path].v != NIL ){
			strcpy( Work, tsv(path) );
		}
	}

	if( List[com].next == NIL ||	
	    List[List[com].next].kind == KVOID ){
		if( SaveFileForm[0] == '\0' ){
			warning( W_NOFNAME, "save" );
		}
		sprintf( buff, SaveFileForm, SerialSave++, SaveType );
		name = buff;
	} else if( List[List[com].next].kind == KLITERAL ){
		name = tsnv(com);
	} else {
		warning( W_SAVE );
		return;
	}
/* D2 add for correct save file path */
	flag = 0;
	if( name[0] == '/' ){
	     flag = 1;
	} else if( name[0] == '.' ){
	     if( name[1] == '/' ){
		  flag = 1;
	     } else if( name[1] == '.' ){
		  if( name[2] == '/' ){
		       flag = 1;
		  }
	     }
	} else if( name[0] == '~' ){
	     flag = 1;
	}
	if( flag == 1 ){
	     strcpy( Work, name );
	} else {
	     strcat( Work, name );
	}
/*	strcat( Work, name );*/
/* end D2 */
/* add #silentplug B2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display( M_SAVING, Work );
	}

	if( Level != 1 ){
		memset( &mes, 0, sizeof(struct Message) );
		mes.command = CSAVE;
		mes.subcommand = SCEXEC;
		strcpy( mes.u.data, Work );
		mes.datasize = strlen( mes.u.data );
		send_message( &mes, NowChild[1] );
	}
	if( List[List[com].next].kind == KVOID ){
		LocalList[List[List[com].next].v].next = l_cons( local_strdup( name ), KLITERAL, NIL );
		List[List[com].next].kind = KLITERAL;
		List[List[com].next].v = tmp_strdup( name );
		List[List[com].next].next = NIL;
	}
	smart_set_literal( "$saved", name );
/* add #silentplug B2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display_command( com );
		display( M_DONE );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_posdata()                                 */
/*                                                                    */
/*             struct Sbuf *sb : transaction buffer                   */
/*             Lisp at         : argment of select command            */
/*                                                                    */
/*             return  int : 0==success -1==fail                      */
/*                                                                    */
/*             function : Set the argment of select command(POSLIST)  */
/*                       to the transaction buffer.                   */
/**********************************************************************/
int set_posdata( struct Sbuf *sb, Lisp at )
{
	int count;
	Lisp lp;
	Child ch;

	ch = NowChild[SSELECTER];
	sb->selnum = Childs[ch].selecter_kind;
	sb->maxnum = Childs[ch].maxnum;
	sb->skind = SKPOS;
	if( at == NIL || List[at].v == NIL ){
		sb->defnum = 0;
		return( 0 );
	}
	count = 0;
	while( at ){
		if( List[at].kind == KPOSITION ){
			lp = List[at].v;
			if( List[lp].kind != KNUMBER ){
				warning( W_SELECT );
				return( -1 );
			}
			sb->def[count].x = List[lp].v;
			lp = List[at].next;
			if( List[lp].kind != KNUMBER ){
				warning( W_SELECT );
				return( -1 );
			}
			sb->def[count].y = List[lp].v;
			count++;
			break;
		} else if( List[at].kind == KPOSLIST ){ 
			lp = List[at].v;
			if( List[lp].kind != KPOSITION ){
				warning( W_SELECT );
				return( -1 );
			}
			if( List[List[lp].v].kind != KNUMBER ){
				warning( W_SELECT );
				return( -1 );
			}
			sb->def[count].x = List[List[lp].v].v;
			if( List[List[lp].next].kind != KNUMBER ){
				warning( W_SELECT );
				return( -1 );
			}
			sb->def[count].y = List[List[lp].next].v;
			count++;
			at = List[at].next;
			if( at == NIL ){
				warning( W_SELECT );
				break;
			}
		} else {
			warning( W_SELECT );
		}
	}
	sb->defnum = count;
	return( count );
}

/**********************************************************************/
/*                                                                    */
/*             function get_posdata()                                 */
/*                                                                    */
/*             struct Sbuf *sb : transaction buffer                   */
/*                                                                    */
/*             return  Lisp : result from selecter                    */
/*                                                                    */
/*             function : Set the result from selecter to the List    */
/*                       of Position.                                 */
/**********************************************************************/
Lisp get_posdata( struct Sbuf *sb )
{
	Lisp lp;
	int i;
	Lisp x, y;

	if( sb->defnum == 0 ){
		return( NIL );
	} else if( sb->defnum >= MAXPOSITION ){
		sb->defnum = MAXPOSITION;
	}
	i = sb->defnum - 1;
	x = cons( sb->def[i].x, KNUMBER, NIL );
	y = cons( sb->def[i].y, KNUMBER, NIL );
	lp = cons( x, KPOSITION, y );
	for( i = sb->defnum - 2; i >= 0; i-- ){
		x = cons( sb->def[i].x, KNUMBER, NIL );
		y = cons( sb->def[i].y, KNUMBER, NIL );
		lp = cons( cons( x, KPOSITION, y ), KPOSLIST, lp );
	}
	return( lp );
}

/**********************************************************************/
/*                                                                    */
/*             function set_posfile()                                 */
/*                                                                    */
/*             struct Sfile *sf : transaction buffer of filename      */
/*             Lisp ls          : argment of select command           */
/*                                                                    */
/*             return  int : number of position in the file           */
/*                                                                    */
/*             function : Set the filename at the argment of select   */
/*                       to the transaction buffer.                   */
/**********************************************************************/
int set_posfile( struct Sfile *sf, Lisp ls )
{
	Child ch;
	char *file;
	FILE *fp;
	struct Shead sh;

	ch = NowChild[SSELECTER];
	sf->selnum = Childs[ch].selecter_kind;
	sf->maxnum = Childs[ch].maxnum;
	sf->skind = SKFILE;

	file = tsv(ls);
	if( search_path( file, MSELECT, Work ) == -1 ){
		sf->defnum = 0;
	} else {
		if( (fp = fopen( Work, "r" )) == NULL ){
			sf->defnum = 0;
		}
		if( fread( &sh, sizeof(struct Shead), 1, fp ) != 1 ){
			warning( W_SELFILE, file );
		}
		sf->defnum = sh.defnum;
		fclose( fp );
	}
	strcpy( sf->file, Work );
/* A0 end searchpath */
	return( sf->defnum );
}

/**********************************************************************/
/*                                                                    */
/*             function get_posfile()                                 */
/*                                                                    */
/*             struct Sfile *sf : transaction buffer                  */
/*             Lisp ls          : argment of select                   */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : set the result from selecter to argment.    */
/*                                                                    */
/**********************************************************************/
void get_posfile( struct Sfile *sf, Lisp ls )
{
	if( sf->defnum == 0 ){
		List[ls].next = NIL;
		return;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_select()                                   */
/*                                                                    */
/*             Lisp com : COMMAND list                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to select command.                */
/*                                                                    */
/**********************************************************************/
void do_select( Lisp com )
{
	struct Message mes;
	int num;
	int flag;
	Lisp pos, selected, local_at;

	switch( Level ){
	case 0:
		memset( &mes, 0, sizeof(struct Message) );
		mes.command = CSELECT;
		mes.subcommand = SCEXEC;
		if( File_flag ){
			mes.comatr  = ADISP | ANOINT;
		} else {
			if( check_value( GLOBAL, "#noint" ) == 0 ){
				mes.comatr  = ADISP | AINT;
			} else if( In_loop || Func_count ){
				mes.comatr  = ANODISP | ANOINT;
			} else {
				mes.comatr  = ADISP | ANOINT;
			}
		}
		break;
	case 1:
		return;
		break;
	case 2:
		memset( &mes, 0, sizeof(struct Message) );
		mes.command = CSELECT;
		mes.subcommand = SCEXEC;
		mes.comatr  = ANODISP | ANOINT;
		break;
	default:
		return;
		break;
	}

	local_at = NIL;
	if( List[com].next == NIL ){
		flag = 1; pos = NIL;
	} else if( List[List[com].next].kind == KVOID ){
		pos = List[List[com].next].next;
		local_at = List[List[com].next].v;
		if( pos ){
			if( List[pos].kind == KLITERAL ){
				LocalList[local_at].next = 
				      LocalList[LocalList[local_at].next].next;
				flag = 0; 
				List[com].next = List[List[com].next].next;
			} else {
				flag = 2; pos = List[List[com].next].next;
			}
		} else {
			flag = 2; pos = NIL;
		}
	} else if( List[List[com].next].kind == KLITERAL ){
		flag = 0; pos = List[com].next;
	} else {
		flag = 1; pos = List[com].next;
	}

	if( flag == 0 ){
		set_posfile( &mes.u.sf, pos );
		send_message( &mes, NowChild[2] );
		if( mes.status == STDONE ){
			get_posfile( &mes.u.sf, com );
		}
	} else {
		num = set_posdata( &mes.u.sb, pos );
		send_message( &mes, NowChild[2] );
		if( mes.status == STDONE ){
			pos = get_posdata( &mes.u.sb );
			selected = dup_t2lLisp( pos );
			set_hash( LOCAL,"$selected", selected );
			if( flag == 2 ){
				LocalList[local_at].next = selected;
				List[com].next = pos;
			}
		}
	}
/* add #silentplug B2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display_command( com );
		display( M_DONE );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function do_effect()                                   */
/*                                                                    */
/*             Lisp com : COMMAND list                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to effect command.                */
/*                                                                    */
/**********************************************************************/
void do_effect( Lisp com )
{
	struct Message mes;

	if( Level == 1 ){
		return;
	}
	memset( &mes, 0, sizeof(struct Message) );
	if( List[com].next ){
		if( List[List[com].next].kind == KMASK ){
			set_area( List[com].next );
		}
	}
        mes.u.data[0] = '\0';
	mes.command = CEFFECT;
	mes.subcommand = SCEXEC;
	if( Level == 0 ){
		if( check_value( GLOBAL, "#noint" ) == 0 ){
			mes.comatr  = ADISP | AINT;
		} else {
			if( In_loop || Func_count ){
				mes.comatr  = ADISP | ANOINT;
			} else {
				mes.comatr = ADISP | AINT;
			}
		}
	} else {
		mes.comatr = ANODISP;
	}
	mes.datasize = 0;
	send_message( &mes, NowChild[3] );
/* add #silentplug B2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display_command( com );
		display( M_DONE );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function kill_message()                                */
/*                                                                    */
/*             struct Message *mes : message sent to plug for kill plug */
/*                                                                    */
/*             return  int : 0==killed  -1==transaction fail          */
/*                                                                    */
/*             function : Kill the plug process.                      */
/*                                                                    */
/**********************************************************************/
static int kill_message( struct Message *mes )
{
	struct Message rep;

	mes->server = SMAIN;
	mes->shmid = NowShmid;
	if( com_send( mes ) < 0 ){
		return( -1 );
	}
	if( com_rcv( &rep ) < 0 ){
		return( -1 );
	}
        return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function kill_child()                                  */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Kill all plug process.                      */
/*                                                                    */
/**********************************************************************/
void kill_child( void )
{
	int i;
	struct Message mes;

	memset( &mes, 0, sizeof(struct Message) );
	mes.command = CKILL;
	mes.subcommand = SCEXEC;
	mes.pid[SMAIN] = ProcessId;

	if( Viewer.pid ){
		mes.receiver = SVIEWER;
		mes.pid[SVIEWER] = Viewer.pid;
		if( kill_message( &mes ) == -1 ){
			kill( Viewer.pid, SIGKILL );
		}
		Viewer.pid = 0;	/* add C1 */
	}
	if( Gui ){
		kill( Gui, SIGKILL );
	}
	for( i = 0; i < ChildsNum; i++ ){
		mes.receiver = Childs[i].kind;
		mes.pid[mes.receiver] = Childs[i].pid;
		if( Childs[i].pid ){
			if( kill_message( &mes ) == -1 ){
				kill( Childs[i].pid, SIGKILL );
			}
		}
		Childs[i].pid = 0; /* add C1 */
	}
}

/**********************************************************************/
/*                                                                    */
/*             function purge_viewer()                                */
/*                                                                    */
/*             Lisp com : purge viewer                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Purge the viewer.                           */
/*                                                                    */
/**********************************************************************/
void purge_viewer( Lisp com )
{
	struct Message mes;
	int id;
/* D2 add : purging correctly */
	Lisp g;
	int flag, ret;
/* D2 end */

        if( Level != 0 ){
		return;
	}

/* D2 add : purging correctly */
	g = search_global( "#viewer" );
	flag = 1;
	if( g ){
	     if( List[g].kind == KLITERAL ){
		  if( List[g].v ){
		       if( (*tsv(g)) != '\0' ){
			    flag = strcmp( tsv(g), tsv(com) );
		       }
		  }
	     }
	}
/* D2 end */

	id = get_shmid_by_name( com );
	if( id == -1 ){
		fprintf( stderr, "not registed viewer %s\n", tsv(com) );
		return;
	}
	memset( &mes, 0, sizeof(struct Message) );
	mes.command = CPURGE;
	mes.subcommand = SCEXEC;
	*(int *)mes.u.data = id;
	mes.datasize = sizeof(int);
	send_message( &mes, IDVIEWER );
	delete_Mem_by_name( com );
/* D2 purging now viewer.. reset default viewer */
	if( flag == 0 ){ 
	     if( (ret = reset_shmid( NIL )) != 0 ){
		  warning( ret );
	     }
	     if( reconfig_viewer( 1 ) != 0 ){
		  warning( W_OMNI, "cannot up new viewer" );
	     }
	     g = g_cons( NIL, KLITERAL, NIL );
	     set_hash( GLOBAL, "#viewer", g );
	}
/* end D2 */
}

/**********************************************************************/
/*                                                                    */
/*             function do_purge()                                    */
/*                                                                    */
/*             Lisp com : PLUG list                                   */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to purge command.                 */
/*                                                                    */
/**********************************************************************/
void do_purge( Lisp com )
{
	int i;
	struct Message mes;
	Child at;
	char *name;

	if( Level == 1 ){
		return;
	}
	if( com == NIL ){
		warning( W_PURGE );
	}
	if( List[com].kind == KLITERAL ){
		purge_viewer( com );
		return;
	} else if( List[com].kind != KSYMBOL ){
		warning( W_PURGE );
	}

	memset( &mes, 0, sizeof(struct Message) );
	mes.command = CKILL;
	mes.subcommand = SCEXEC;
	mes.pid[SMAIN] = ProcessId;
	name = tsv(com);

	for( i = 0; i < 3; i++ ){
		if( (at = search_type( i, name )) != -1 ){
			break;
		}
	}
/*	if( at == NIL ){ B2*/
	if( at == -1 ){
		warning( W_PNONAME, name );
	}
	if( Childs[at].pid == 0 ){
		warning( W_NOREADY, name );
	}
	mes.receiver = i + 1;
	mes.pid[mes.receiver] = Childs[at].pid;
	if( kill_message( &mes ) == -1 ){
		kill( Childs[at].pid, SIGKILL );
	}
	Childs[at].pid = 0;
/* add #silentplug B2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display( M_DONE );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function show_mask()                                   */
/*                                                                    */
/*             int mask : Mask index                                  */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : show mask command to transaction.           */
/*                                                                    */
/**********************************************************************/
void show_mask( int mask )
{
	struct Message mes;

        if( Level != 0 ){
		return;
	}
	memset( &mes, 0, sizeof(struct Message) );
	mes.command = CSHOW;
	mes.subcommand = SCEXEC;
	mes.u.data[0] = (char)mask + '0';
	mes.u.data[1] = '\0';
	mes.datasize = 1;
	send_message( &mes, IDVIEWER );
}

/**********************************************************************/
/*                                                                    */
/*             function viewer_config()                               */
/*                                                                    */
/*             unsigned char *str : configure string                  */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Configure viewer by the str at transaction. */
/*                                                                    */
/**********************************************************************/
void viewer_config( unsigned char *str )
{
	struct Message mes;

        if( Level != 0 ){
		return;
	}
	memset( &mes, 0, sizeof(struct Message) );
	mes.command = CSHOW;
	mes.subcommand = SCCONFIG;
	strcpy( mes.u.data, str );
	mes.datasize = strlen(mes.u.data);
	send_message( &mes, IDVIEWER );
}

/**********************************************************************/
/*                                                                    */
/*             function reconfig_viewer()                             */
/*                                                                    */
/*             int com : 0 == DOWN  1== UP                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : send to viewer the UP & DOWN message.       */
/*             D2 : create new                                        */
/**********************************************************************/
/* D2 multi viewer selection showing */
int reconfig_viewer( int com )
{
	struct Message mes;

        if( Level != 0 ){
		return 0;
	}
	memset( &mes, 0, sizeof(struct Message) );
	mes.command = CSHOW;
	if( com == 0 ){
	     mes.subcommand = SCDOWN;
	} else {
	     mes.subcommand = SCUP;
	}
	mes.datasize = 0;
	send_message( &mes, IDVIEWER );
	return 0;
}

/**********************************************************************/
/*                                                                    */
/*             function init_viewer()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Invoke the viewer process.                  */
/*                                                                    */
/**********************************************************************/
void init_viewer( void )
{
	char *path;
	Lisp ret;

	if( Level == 2 ){
		return;
	}
	if( (ret = search_hash( GLOBAL, "#gedopath" )) == NIL ){
		path = VIEWERPATH;
	} else if( GlobalList[ret].kind == KLITERAL ){
		strcpy( Work, gsv(ret) );
		strcat( Work, "viewer" );
		path = Work;
	} else {
		path = VIEWERPATH;
	}
	if( (Viewer.pid = fork()) == 0 ){
		execl( path, "viewer", NULL);
		fprintf( stderr, "path=%s", path );
		fatal( F_FVIEWER );
		return;
	}
/* add #silentplug B2 */
/* D2 add to kill all child process */
	check_childPID( Viewer.pid );
/* end D2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display( M_CHDOK, "viewer" );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function init_gui()                                    */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Invoke the gui process.                     */
/*                                                                    */
/**********************************************************************/
/* D2 for new GUI(gedo command launcher)*/
void init_gui( void )
{
	char *path;
	Lisp ret;
	char *ini[4];
	char buff[16];
	int cret;
	char *prog;

	if( Level == 2 ){
		return;
	}
	prog = NULL;
	if( (ret = search_hash( GLOBAL, "#guiprog" )) != NIL ){
	     if( GlobalList[ret].kind == KLITERAL ){
		  prog = gsv(ret);
		  strcpy( Work, prog );
	     }
	}
	if( prog == NULL ) {
	     prog = "gui";
	     path = GEDOPATH;
	     if( (ret = search_hash( GLOBAL, "#gedopath" )) != NIL ){
		  if( GlobalList[ret].kind == KLITERAL ){
		       path = gsv(ret);
		  }
	     }
	     strcpy( Work, path );
	     strcat( Work, prog );
	}

	if( (cret = fork()) == 0 ){
	        int at;
		at = 0;
		ini[at++] = "gui";
		if( Gui != 0 ){
		     sprintf( buff, "gui=%d", Gui );
		     ini[at++] = buff;
		}
/* argment gui=<num> makes trick... all GUI launcher returns message type of
  first gui ProcessID(if first gui process killed). The first Gui PID is the
  ID for ALL GUI LAUNCHERS of IPC message. It is represented static variable
  'Gui' at GEDO main process.*/
		if( (ret = search_hash( GLOBAL, "#path" )) != NIL
		     && GlobalList[ret].kind == KLITERAL ){
		     ini[at++] = gsv(ret);
		}
		ini[at] = NULL;
		execvp( Work, ini  );
		warning( W_FGUI );
	}
	signal( SIGUSR1, SIG_IGN );
	check_childPID( cret );
	if( Gui == 0 ){
	     sprintf( buff, "gui=%d", cret );
	     smart_set_global_literal( "#gui", buff );
	     Gui = cret;
	}
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display( M_CHDOK, "gui" );
	}
}

#if 0
/* old version of GUI... use newer.*/
void init_gui( void )
{
	char *path;
	Lisp ret;
	char buff[16];

	if( Level == 2 ){
		return;
	}
	if( (ret = search_hash( GLOBAL, "#gedopath" )) == NIL ){
		path = GUIPATH;
	} else if( GlobalList[ret].kind == KLITERAL ){
		strcpy( Work, gsv(ret) );
		strcat( Work, "gui" );
		path = Work;
	} else {
		path = GUIPATH;
	}
	if( (Gui = fork()) == 0 ){
		InitFile[0] = "gui";
		execvp( path, InitFile  );
		warning( W_FGUI );
	}
	signal( SIGUSR1, SIG_IGN );
/* add #gui for additional GUI process  C1*/
	sprintf( buff, "gui=%d", Gui );
	smart_set_global_literal( "#gui", buff );

/* add #silentplug B2 */
	if( check_value( GLOBAL, "#silentplug" ) == 0 ){
		display( M_CHDOK, "gui" );
	}
}
#endif


/**********************************************************************/
/*                                                                    */
/*             function do_create()                                   */
/*                                                                    */
/*             Lisp pos : POSITION list                               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to create command.                */
/*            B2 : careless misstake                                  */
/*            D2 : check maxsize can create                           */
/**********************************************************************/
void do_create( Lisp pos )
{
	int w, h;
	Lisp at;
	struct Message mes;
/* add D2 */
	Lisp max;
	Lisp g;
	int len;
/* end D2 */

	if( pos == NIL ){
		warning( W_CREATE );
		return;
	}
	if( List[pos].kind == KPOSITION ){
		if( List[List[pos].v].kind != KNUMBER ){
			warning( W_CREATE );
			return;
		}
		w = List[List[pos].v].v;
		if( List[List[pos].next].kind != KNUMBER ){
			warning( W_CREATE );
			return;
		}
		h = List[List[pos].next].v;
	} else if( List[pos].kind == KPOSLIST ){
		at = List[pos].v;
		if( List[at].kind != KPOSITION ){
			warning( W_CREATE );
			return;
		}
		if( List[List[at].v].kind != KNUMBER ){
			warning( W_CREATE );
			return;
		}
		w = List[List[at].v].v;
		if( List[List[at].next].kind != KNUMBER ){
			warning( W_CREATE );
			return;
		}
		h = List[List[at].next].v;
	} else {
		warning( W_CREATE );
		return;
	}

/* add D2: check maxsize can create */
        if( (max = search_global( "#maxpixel" )) == NIL ){
	     warning( W_OMNI, "global variable #maxpixel cannot find... cannot happen" );
	     return;
	}
	if( List[max].kind != KNUMBER ){
	     warning( W_OMNI, "global variable #maxpixel cannot find... cannot happen" );
	     return;
	}
	if( w * h > List[max].v ){
	     warning( W_OMNI, "create size exceed maxpixel size" );
	     return;
	}
/* end D2 */
	create_shm( w, h );

        if( Level != 0 ){
		return;
	}

	memset( &mes, 0, sizeof(struct Message) );
/* add D2 for make windowname of created image */
	Work[0] = '\0'; len = 1;
	if( (g = search_global( "#viewer" )) != NIL ){
		if( List[g].kind == KLITERAL ){
			strcpy( &Work[1], tsv(g) );
			len = strlen(tsv(g)) + 1;
		}
	}
	if( len >= 256 ){
	     dwarning( "Message is too long\n" );
	}
	memcpy( mes.u.data, Work, len );
/* end D2 */
	mes.command = CLOAD;
	mes.subcommand = SCEXEC;
/* add D2 for make windowname of created image */
/* 	mes.datasize = 0;*/
	mes.datasize = len;
/* end D2 */
	send_message( &mes, IDVIEWER );

/*B2	smart_set_literal( "Loaded", gs(FileName[FileNum]) );*/
	smart_set_literal( "$loaded", gs(FileName[FileNum]) );
	set_wh();
}

/**********************************************************************/
/*                                                                    */
/*             function search_plugs()                                */
/*                                                                    */
/*             Lisp pos : POSITION list                               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to create command.                */
/*            B2 : careless misstake                                  */
/**********************************************************************/
Lisp search_plugs( int kind, int flag )
{
	Lisp at, lst, ret;
	int i;

	ret = lst = cons( NIL, KLIST, NIL );
	for( i = 0; i < ChildsNum; i++ ){
		if( flag && Childs[i].pid == 0 ){ /* not invoked */
			continue;
		}
		if( kind == 0 ){
			at = cons( tmp_strdup( gs(Childs[i].name) ), KLITERAL, NIL );
			List[lst].next = cons( at, KLIST, NIL );
			lst = List[lst].next;
		} else {
			if( Childs[i].kind + 1 == kind ){
				at = cons( tmp_strdup( gs(Childs[i].name) ), KLITERAL, NIL );
				List[lst].next = cons( at, KLIST, NIL );
				lst = List[lst].next;
			}
		}
	}
	if( List[ret].next == NIL ){
		List[ret].kind = KNIL;
		return( ret );
	} else {
		return( List[ret].next );	
	}
}

/**********************************************************************/
/*                                                                    */
/*             function plug_info()                                   */
/*                                                                    */
/*             Lisp pos : POSITION list                               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to create command.                */
/*            B2 : careless misstake                                  */
/**********************************************************************/
char *plug_info( int num )
{
	struct Childs *ch;
	char buff[16];

	ch = &Childs[num];

	if( ch->kind == 1 ){
		if( ch->maxnum ){
			sprintf( buff, "%s%d", StrSelKind[ch->selecter_kind],
				 ch->maxnum );
		} else {
			sprintf( buff, "%s", StrSelKind[ch->selecter_kind] );
		}			
		sprintf( Work, "%s %s: %s %s path=%s option=\"%s\"\n\t%s", 
			 StrSection[ch->kind], gs(ch->name), 
			 StrLoaded[(ch->pid)?(0):(1)], buff,
			 gs(ch->path), gs(ch->option), gs(ch->help) );
	} else {
		sprintf( Work, "%s %s: %s path=%s option=\"%s\"\n\t%s",
			 StrSection[ch->kind], gs(ch->name), 
			 StrLoaded[(ch->pid)?(0):(1)], 
			 gs(ch->path), gs(ch->option), gs(ch->help) );
	}
	return( Work );
}

/* add C1 */
void do_reset( void )
{
	kill_child();
	init_viewer();
	sleep(1);
	do_show( NIL );
}
