/**********************************************************************/
/*                                                                    */
/*         com.c -- gedo low level transaction routine                */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*        B2:  '" can handle correct in start_child()                 */
/*        D2:  change wait signal from wait(2) to pause(2)            */
/*        D2:  add to kill all child process                          */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <ctype.h>

#include <sys/ipc.h>
#include <sys/msg.h>
#include <unistd.h>
#define O_BINARY 0

#include "gedo.h"
#include "prot.h"
#include "../loader.h"
#include "error.h"

#define DEBUG 0


/**********************************************************************/
/*                                                                    */
/*             function usr_catch()                                   */
/*                                                                    */
/*             int signo : signal No.                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Signal handler, not abort at receive SIGUSR1*/
/*                                                                    */
/**********************************************************************/
static void usr_catch( int signo )
{
	signal( SIGUSR1, SIG_IGN );
}

/**********************************************************************/
/*                                                                    */
/*             function start_child()                                 */
/*                                                                    */
/*             struct Childs *child : plug infomation                 */
/*                                                                    */
/*             return  int : 0==success  -1==fail                     */
/*                                                                    */
/*             function : fork() and start plug process from gedo.    */
/*          B2 '" can handle correct                                  */
/*          D0:   changed waiting signal from wait to pause...        */
/*                                                                    */
/**********************************************************************/
int start_child( struct Childs *child )
{
	char *options[INITFILEMAX], *p, *q;
	int num;
	int status;

	options[0] = gs(child->name);
	num = 1; status = 0;
	p = gs(child->option);
	for( q = Work; *p; p++ ){
		if( status == 0 ){
			if( isspace( *p ) ){
			} else {
				*q = *p;
				status = -1;
				options[num++] = q++;
				if( num >= 19 ){
					display( M_MANYARG, options[0] );
					break;
				}
			}
		} else if( status == -1 ){
			if( isspace( *p ) ){
				*q++ = '\0';
				status = 0;
			} else if( *p == '\'' || *p == '\"' ){
				for( ++p; *p; p++ ){
					if( *p == '\'' || *p == '\"' ){
						break;
					} else {
						*q++ = *p;
					}
				}
			} else {
				*q++ = *p;
			}
		}
	}
	*q = '\0';
	options[num] = NULL;

	signal( SIGUSR1, usr_catch );
	if( (child->pid = fork()) == 0 ){
		execvp( gs(child->path), options );
		printf( "fatal:child %s cannot raize\n", gs(child->name) );
		perror( "" );
		exit( 1 );
		return( -1 );
	}
/* D2 changed waiting signal from wait to pause... */
	pause();
	signal( SIGUSR1, SIG_IGN );
	check_childPID( child->pid );
	return( 0 );
/* D2 DELETED
	if( wait( &st ) == -1 ){
		signal( SIGUSR1, SIG_IGN );
		return( 0 );
	} else {
		child->pid = 0;
		signal( SIGUSR1, SIG_IGN );
		return( -1 );
	}*/
#if DEBUG
	printf( "child %s of %s process start OK\n", gs(child->name), gs(child->path) );
#endif
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function com_send()                                    */
/*                                                                    */
/*             struct Message *mes : transaction message              */
/*                                                                    */
/*             return  int : 0==success  -1==transaction error        */
/*                          -2==not here the plug process             */
/*             function : Transaction(SEND) execute low level func.   */
/*                                                                    */
/**********************************************************************/
int com_send( struct Message *mes )
{
	struct Message dum;

        if( mes->pid[mes->receiver] <= 1 ){
		return( -2 );
	}
        if( getpgid( mes->pid[mes->receiver] ) == -1 ){
		return( -2 );
	}
	mes->mtype = (long)mes->pid[mes->receiver];
#if DEBUG
	printf( "send:%s:%s:%s:%s:%s:%s\n", StrServer[mes->server], 
       		StrCommand[mes->command], StrSubCom[mes->subcommand],
       		StrStatus[mes->status], StrErr[mes->errorcode], 
       		mes->u.data );
#endif
	
	while( msgrcv( MesId, (struct msgbuf *)&dum, sizeof(struct Message) - sizeof(long), ProcessId, IPC_NOWAIT ) != -1 ){
	}
	if( msgsnd( MesId, (struct msgbuf *)mes, sizeof(struct Message) - sizeof(long), 0 ) == -1 ){
		return( -1 );
	}
	end_raw_tty();
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function print_sbuf()                                  */
/*                                                                    */
/*             struct Sbuf *sb : Transaction buffer for select command*/
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Show the selection result for debug.        */
/*                                                                    */
/**********************************************************************/
void print_sbuf( struct Sbuf *sb )
{
	int i;

	display( M_FORM, "select:kind:%s num:%d\n", StrSelKind[sb->selnum], sb->defnum );
	for( i = 0; i < sb->defnum; i++ ){
		display( M_FORM, "\t%d:(%d,%d)", i, sb->def[i].x, sb->def[i].y );
	}
	display( M_FORM, "\n" );
}

/**********************************************************************/
/*                                                                    */
/*             function com_rcv()                                     */
/*                                                                    */
/*             struct Message *mes : Transaction message              */
/*                                                                    */
/*             return  int : 0==success  -1==fail at transaction      */
/*                                                                    */
/*             function : Transaction(RECEIVE) low level function.    */
/*                                                                    */
/**********************************************************************/
int com_rcv( struct Message *mes )
{
	if( msgrcv( MesId, (struct msgbuf *)mes, sizeof(struct Message) - sizeof(long), ProcessId, 0 ) == -1 ){
#if DEBUG
		perror( "" );
#endif
		reset_raw_tty();
		return( -1 );
	}
        reset_raw_tty();

#if DEBUG
	printf( "rcv:%s:%s:%s:%s:%s\n", StrServer[mes->server], 
		StrCommand[mes->command], StrSubCom[mes->subcommand],
		StrStatus[mes->status], StrErr[mes->errorcode] );
	if( mes->command == CSELECT && mes->subcommand == SCEXEC ){
		print_sbuf( &mes->u.sb );
	} else if( mes->datasize ){
		printf( "%s\n", mes->u.data );
	}
#endif
	return( 0 );
}
