/**********************************************************************/
/*                                                                    */
/*        exec.c -- gedo interpreter routines                         */
/*                                                                    */
/*                by K.sugiura   Ver 1.0  for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*        11.21,1996  A0 add rewindjmp in exec_exp                    */
/*        11.30,1996  A0 add delete_Lisp in exec_statement            */
/*         1.24,1997  B0 push_node,clear_prevstack for save mem       */
/*         7.16,1998  D1 delete loop_GC() in for and while loop       */
/*         9. 7,1998  D2 why missed this! expand macro error          */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

static Lisp  Cp;	/* Top-node of copied CommandList */
static Lisp  Top;	/* Top-node of CommandList */
int Debug_flag = 0;	/* if #debug={1} then ==1 */
extern Lisp NowBody;


/* B0 add routine */
static Lisp *Nodestack;  /* Node stack for save memory */
static int  Nodesp = 0;	 /* stack pointer for Nodestack */
static int  NodestackNum = 0;  /* alloc number of Nodestack */

int DebugAt;
int Debug[100];
/**********************************************************************/
/*                                                                    */
/*             function push_node()                                   */
/*                                                                    */
/*             Lisp at    : a patial of command                       */
/*             char *name : names for distinct the node               */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: Register the node by name for free the cell. */
/*            add B0                                                  */
/**********************************************************************/
void push_node( Lisp at, char *name )
{
	void *tmp;

	if( Nodestack == NULL || Nodesp + 1 >= NodestackNum ){
		tmp = realloc( Nodestack, (100 + NodestackNum) * sizeof(int) );
		if( tmp == NULL ){
			dwarning( "NodeStack cannot realloc\n" );
			return;
		}
		Nodestack = tmp;
		NodestackNum += 100;
	}
	Nodestack[Nodesp++] = at;
	if( check_value( GLOBAL, "#chase" ) ){
		chase_Lisp( at, "push_node %d: %s\n", Nodesp-1, name );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function clear_nodestack()                             */
/*                                                                    */
/*             args    NONE                                           */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: Reset the node stack.                        */
/*            add B0                                                  */
/**********************************************************************/
void clear_nodestack( void )
{
	Nodesp = 0;
}

/**********************************************************************/
/*                                                                    */
/*             function clean_prevstack()                             */
/*                                                                    */
/*             Lisp at    : top of the stack to release               */
/*             char *name : names for distinct the node               */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: Release the temporary List for reuse.        */
/*            B0 add                                                  */
/**********************************************************************/
void clean_prevstack( Lisp at, char *name )
{
	int flag;

	if( check_value( GLOBAL, "#chase" ) ){
		flag = 1;
		chase_Lisp( at, "return of %s\n", name );
	} else {
		flag = 0;
	}
	for( --Nodesp; Nodestack[Nodesp]; Nodesp-- ){
		if( Nodesp <= 0 ){
			dwarning( "illegal clean_prevstack\n" );
			return;
		}
		if( flag ){
			chase_Lisp( Nodestack[Nodesp], 
			     "clear_prevstack now %d: %s\n", Nodesp, name );
		}
		delete_Lisp( TMP, Nodestack[Nodesp] );
	}
}
/* end B0 */

/**********************************************************************/
/*                                                                    */
/*             function expand_Lisp()                                 */
/*                                                                    */
/*             Lisp at: CommandList                                   */
/*                                                                    */
/*             return  Lisp: sharp-expaned CommandList                */
/*                                                                    */
/*             function: Expand # in CommandList.(recursive)          */
/*                                                                    */
/**********************************************************************/
Lisp expand_Lisp( Lisp at )
{
	Lisp mc;

	if( at == NIL ){
		return( NIL );
	}
	switch( List[at].kind ){
	case KSHARP:
		mc = expand_Lisp( expand_macro( List[at].v ) );
		if( op_kind(List[mc].kind)  == 8 || List[mc].kind == KCOM ){
			return( cons( List[mc].v, List[mc].kind, expand_Lisp(List[mc].next) ) );
		} else {
			return( cons( List[mc].v, List[mc].kind, expand_Lisp(List[at].next) ) );
		}
	}
	switch( op_kind(List[at].kind) ){
	case 8:
		return( cons( expand_Lisp(List[at].v), List[at].kind, expand_Lisp( List[at].next ) ) );
		break;
	case 1:
	case 0:
		return( cons( List[at].v, List[at].kind, expand_Lisp( List[at].next ) ) );
		break;
	default:
		warning( W_ILLKIND, List[at].kind, "expand_Lisp" );
		return( NIL );
		break;
	}
}


/**********************************************************************/
/*                                                                    */
/*             function exec_term()                                   */
/*                                                                    */
/*             Lisp com: CommandList                                  */
/*                                                                    */
/*             return  Lisp: Evaluated Symbol CommandList             */
/*                                                                    */
/*             function: Evaluate Symbol in CommandList.              */
/*                                                                    */
/**********************************************************************/
Lisp exec_term( Lisp com )
{
	Lisp at;

	if( com == NIL ){
		return( NIL );
	}
	if( List[com].v == NIL ){
		return( com );
	}
	switch( List[com].kind ){
	case KSHARP:
		return( expand_macro( List[com].v ) );
		break;
	case KSYMBOL:
	  	at = get_variable( List[com].v );
		if( List[at].kind == KMASK ){
			return( eval_doller( at ) );
		} else {
			return( at );
		}
		break;
	default:
		return( com );
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function compare_number()                              */
/*                                                                    */
/*             int i1: value of INT-List                              */
/*             int kind: Comparision kind like ==, != or etc.         */
/*             int i2: value of INT-List                              */
/*                                                                    */
/*             return  int: true or not                               */
/*                                                                    */
/*             function: Compare the integer obeying kind.            */
/*                                                                    */
/**********************************************************************/
int compare_number( int i1, int kind, int i2 )
{
	switch( kind ){
	case KEQ:
		return( (i1 == i2) );
		break;
	case KNE:
		return( (i1 != i2) );
		break;
	case KGE:
		return( (i1 >= i2) );
		break;
	case KLE:
		return( (i1 <= i2) );
		break;
	case KGT:
		return( (i1 > i2) );
		break;
	case KLT:
		return( (i1 < i2) );
		break;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function compare_literal()                             */
/*                                                                    */
/*             String i1: String index of TMP                         */
/*             int kind: Comparision kind                             */
/*             String i2: String index of TMP                         */
/*                                                                    */
/*             return  int: true or false                             */
/*                                                                    */
/*             function: Compare string using strcmp obeying kind.    */
/*                                                                    */
/**********************************************************************/
int compare_literal( String i1, int kind, String i2 )
{
	int ret;

	ret = strcmp( ts(i1), ts(i2) );
	switch( kind ){
	case KEQ:
		return( (ret)?(0):(1) );
		break;
	case KNE:
		return( (ret)?(1):(0) );
		break;
	case KGE:
		return( (ret>=0)?(1):(0) );
		break;
	case KLE:
		return( (ret<=0)?(1):(0) );
		break;
	case KGT:
		return( (ret>0)?(1):(0) );
		break;
	case KLT:
		return( (ret<0)?(1):(0) );
		break;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function compare_position()                            */
/*                                                                    */
/*             Lisp i1: List of POSITION                              */
/*             int kind: Comparision kind                             */
/*             Lisp i2: List of POSITION                              */
/*                                                                    */
/*             return  int: true or false                             */
/*                                                                    */
/*             function: Compare POSITION by absolute value           */
/*                                                                    */
/**********************************************************************/
int compare_position( Lisp i1, int kind, Lisp i2 )
{
	int r1, r2;

	r1 = List[List[i1].v].v + List[List[i1].next].v;
	r2 = List[List[i2].v].v + List[List[i2].next].v;
	switch( kind ){
	case KEQ:
		return( (r1 == r2) );
		break;
	case KNE:
		return( (r1 != r2) );
		break;
	case KGE:
		return( (r1 >= r2) );
		break;
	case KLE:
		return( (r1 <= r2) );
		break;
	case KGT:
		return( (r1 > r2) );
		break;
	case KLT:
		return( (r1 < r2) );
		break;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function compare_poslist()                             */
/*                                                                    */
/*             Lisp i1: List of POSLIST                               */
/*             int kind: Comparision kind                             */
/*             Lisp i2: List of POSLIST                               */
/*                                                                    */
/*             return  int: true or false                             */
/*                                                                    */
/*             function: Compare POSLIST                              */
/*                                                                    */
/**********************************************************************/
int compare_poslist( Lisp i1, int kind, Lisp i2 )
{
	Lisp a1, a2;
	int ret;

	a1 = i1; a2 = i2;
	while( 1 ){
		if( a1 == NIL && a2 == NIL ){
			return( compare_number( 0, kind, 0 ) );
		} else if( a1 == NIL ){
			return( compare_number( 0, kind, 1 ) );
		} else if( a2 == NIL ){
			return( compare_number( 1, kind, 0 ) );
		}
		ret = compare_position( List[a1].v, kind, List[a2].v );
		if( ret == 1 ){
			a1 = List[a1].next; a2 = List[a2].next;
		} else {
			return( 0 );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function compare()                                     */
/*                                                                    */
/*             Lisp l1: Value List                                    */
/*             int kind: Comparision kind                             */
/*             Lisp l2: Value List                                    */
/*                                                                    */
/*             return  Lisp: 1 or NIL                                 */
/*                                                                    */
/*             function: Compare List obeying kind for each ListKind. */
/*                                                                    */
/**********************************************************************/
Lisp compare( Lisp l1, int kind, Lisp l2 )
{
	int ret;

	ret = 0;
	if( List[l1].kind != List[l2].kind ){
		if( kind == KNE ){
			return( cons( 1, KNUMBER, NIL ) );
		} else {
			return( cons( 0, KNUMBER, NIL ) );
		}
	}
	switch( List[l1].kind ){
	case KNUMBER:
	case KNIL:
		ret = compare_number( List[l1].v, kind, List[l2].v );
		break;
	case KLITERAL:
		ret = compare_literal( List[l1].v, kind, List[l2].v );
		break;
	case KPOSITION:
		ret = compare_position( List[l1].v, kind, List[l2].v );
		break;
	case KPOSLIST:
		ret = compare_poslist( List[l1].v, kind, List[l2].v );
		break;
	default:
		warning( W_ILLCMP );
		ret = 0;
		break;
	}
	return( cons( ret, KNUMBER, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function exec_exp()                                    */
/*                                                                    */
/*             Lisp com: CommandList                                  */
/*                                                                    */
/*             return  Lisp: Evaluated CommandList                    */
/*                                                                    */
/*             function: Evaluate CommandList.                        */
/*                      (command execution and comparition)           */
/**********************************************************************/
Lisp exec_exp( Lisp com )
{
	Lisp ret, at;
	int flag;

	ret = NIL;
	if( Error_flag ){
		error_jump();
	}
	if( com == NIL ){
		return( NIL );
	}
	switch( Kind_spec[List[com].kind] & 0xf0 ){
	case KSBASIC:
		ret = exec_term( com );
		break;
	case KSFOP:
		ret = eval_exp( com );
		break;
	case KSCOP:
		ret = compare( eval_exp( List[com].v ), List[com].kind, eval_exp( List[com].next ) );
		break;
	case KSLIST:
		at = exec_exp( List[com].v );
		if( List[com].kind == KPOSLIST && List[com].next == NIL ){
			return( at );
		} else {
			ret = cons( at, List[com].kind, exec_exp( List[com].next ) );
		}
		break;
	case KSCOMPLEX:
		switch( List[com].kind ){
		case KCOM:
/* B2 expression */
			if( check_value( GLOBAL, "#noexp" ) ){
				ret = com;
				break;
			}
/* end B2 */
			flag = 0;
			switch( List[com].v ){
			case MLOADER:
			case MSAVER:
			case MSELECTER:
			case MEFFECTER:
				at = List[com].next;
				if( List[at].kind == KFN ){
					ret = cons( List[com].v, KCOM, cons( List[at].v, KFN, NIL ) );
					break;
				}
				at = List[at].next;
				if( at == NIL ){
					ret = NIL;
				} else if( List[at].kind == KFN ){
					ret = cons( NIL, KFN, NIL );
				} else if( List[at].kind == KVOID ){
					ret = cons( List[at].v, KVOID, exec_exp( List[at].next ) );
				} else {
					ret = exec_exp( at );
				}
				

				ret = cons( List[com].v, KCOM, cons( List[List[com].next].v, KSYMBOL, ret ) );
				break;
			case MLOAD:
			case MSAVE:
			case MSELECT:
				if( List[com].next == NIL ){
					ret = com;
					break;
				}
				if( List[List[com].next].kind == KVOID ){
					ret = exec_exp( List[List[com].next].next );
					ret = cons( List[com].v, List[com].kind, cons( List[List[com].next].v, KVOID, ret ) );
				} else {
 					ret = exec_exp( List[com].next );
					ret = cons( List[com].v, List[com].kind, ret );
				}
				break;
			case MPURGE:
			case MPRINT:
				ret = com;
				break;
			case MPREV:
			case MNEXT:
			case MGC:
				ret = cons( List[com].v, List[com].kind, NIL );
				flag = 1;
				break;
			case MGOTO:
			case MINCLUDE:
				flag = 1;
				ret = cons( List[com].v, List[com].kind, exec_exp( List[com].next ) );
				break;
			default:
				ret = cons( List[com].v, List[com].kind, exec_exp( List[com].next ) );
				break;
			}
			eval( ret );
			/* MGC, MPREV, MNEXT, MGOTO  destroy List */
/* A0 add rewindjmp */
			if( !File_flag && flag ){
				rewind_input();
			}
/* A0 end */
			if( !flag && List[com].v < MLIST ){ 
				List_flag++;
			}
			break;
		case KPOSITION:
			ret = eval_exp( com );
			break;
		case KRANGE:
			at = exec_term( List[com].v );
			ret = cons( at, KRANGE, exec_term( List[com].next ) );
			break;
		default:
			warning( W_ILLKIND, List[com].kind, "exec_exp" );
			return( NIL );
			break;
		}
		break;
	case KSTOP:
		if( List[com].kind == KRET ){
			ret = eval_exp( com );
		} else if( List[com].kind == KFUNC ){
			ret = eval_exp( com );
		} else {
			warning( W_ILLKIND, List[com].kind, "exec_exp" );
			return( NIL );
		}
		break;
	default:
		warning( W_ILLKIND, List[com].kind, "exec_exp" );
		return( NIL );
		break;
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function set_name()                                    */
/*                                                                    */
/*             Lisp name: Left value of = (variable)                  */
/*             Lisp body: Right value of =                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Execution of set-value to variable.          */
/*                                                                    */
/**********************************************************************/
void set_name( Lisp name, Lisp body )
{
	int mask;
	Lisp at, ls;
	char *vname;

	vname = tsv(name);
	if( List[name].kind == KSHARP ){
		if( List[body].kind == KLIST && List[body].next == NIL ){
			at = List[body].v;
		} else {
			at = body;
		}
		if( (ls = search_hash( GLOBAL, vname )) ){
			if( GlobalList[ls].kind == KPLUG ){
				warning( W_BATPLUG, vname );
			}
		}
		delete_Lisp( GLOBAL, ls );
		set_hash( GLOBAL, vname, dup_t2gLisp(at) );
	} else if( List[name].kind == KSYMBOL ){
		if( List[body].kind == KSHARP ){
			at = expand_Lisp( body );
			set_clever_name( vname, eval_exp(at) );
		} else {
			at = eval_exp( body );
			if( List[at].kind == KMASK ){
				mask = List[at].v;
				name_mask( mask );
				set_clever_name( vname, cons( mask, KMASK, NIL ) );
				set_area( at );
				show_mask( 0 );
			} else {
				set_clever_name( vname, at );
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function exec_statement()                              */
/*                                                                    */
/*             Lisp com: CommandList                                  */
/*                                                                    */
/*             return  Lisp: Evaluated CommandList                    */
/*                                                                    */
/*             function: Evaluate CommandList.                        */
/*                      (=,if,for,while)                              */
/**********************************************************************/
/* FIXME: exec_statement consumes too many Cells. It is a WRONG THING. */
/* I will FIX this after. */
/* if, for, while copy the List and execute. */
Lisp exec_statement( Lisp com )
{
	Lisp at;
	Lisp ret;

	ret = NIL;
	if( Error_flag ){
		error_jump();
	}
	if( com == NIL ){
		return( NIL );
	}
	switch( Kind_spec[List[com].kind] & 0xf0 ){
	case KSBASIC:
		if( List[com].kind == KSHARP ){
/*			at = expand_Lisp( com );  C1*/
/* D2			at = expand_macro( com );*/
/* D2 why missed this ! */
			at = expand_macro( List[com].v );
			ret = exec_statement( at );
		} else {
			ret = exec_exp( com );
		}
		break;
	case KSCOMPLEX:
		switch( List[com].kind ){
		case KWHILE:
			In_loop++;
			while( 1 ){
				at = dup_t2tLisp( List[com].v );
				ret = exec_statement( at );
				if( List[ret].kind != KNUMBER ){
					warning( W_LOOP );
					return( NIL );
				}
				if( List[ret].v == 0 ){
					break;
				}
/* add A0 Optimise */
				delete_Lisp( TMP, at );
/* end A0 */
				if( List[List[com].next].kind == KSHARP ){
					at = expand_Lisp( List[com].next );
				} else {
					at = dup_t2tLisp( List[com].next );
				}
				exec_statement( at );
/* add D1 */			delete_Lisp( TMP, at );
/* del D1			loop_GC();*/
			}
			In_loop--;
			break;
		case KFOR:
			exec_statement( List[List[com].v].v );
			In_loop++;
			while( 1 ){
				at = dup_t2tLisp( List[List[List[com].v].next].v );
				ret = exec_statement( at );
				if( List[ret].kind != KNUMBER ){
					warning( W_LOOP );
					return( NIL );
				}
				if( List[ret].v == 0 ){
					break;
				}
/* add A0 Optimise */
				delete_Lisp( TMP, at );
/* end A0 */
				if( List[List[com].next].kind == KSHARP ){
					at = expand_Lisp( List[com].next );
				} else {
					at = dup_t2tLisp( List[com].next );
				}
				exec_statement( at );
/* add D1 */			delete_Lisp( TMP, at );
				at = dup_t2tLisp( List[List[List[com].v].next].next );
				exec_statement( at );
/* add D1 */			delete_Lisp( TMP, at );
/* del D1			loop_GC();*/
			}
			In_loop--;
			break;
		case KIF:
			at = dup_t2tLisp( List[com].v );
			push_node( at, "if" );
			ret = exec_statement( at );
			if( List[ret].kind != KNUMBER && List[ret].kind != KNIL ){
				warning( W_LOOP );
				return( NIL );
			}
/* add A0 Optimise */
/*			delete_Lisp( TMP, at );*/
/* end A0 */
			if( List[ret].v ){
				if( List[List[List[com].next].v].kind == KSHARP ){
					at = expand_Lisp( List[List[com].next].v );
					push_node( at, "if ture(sharp)" );
				} else {
					at = dup_t2tLisp( List[List[com].next].v );
					push_node( at, "if true" );
				}
				exec_statement( at );
			} else if( List[List[com].next].next ){
				if( List[List[List[com].next].next].kind == KSHARP ){
					at = expand_Lisp( List[List[com].next].next );
					push_node( at, "if false(sharp)" );
				} else {
					at = dup_t2tLisp( List[List[com].next].next );
					push_node( at, "if false" );
				}
				exec_statement( at );
			}
			break;
		case KCOM:
			ret = exec_exp( com );
			break;
		case KPOSITION:
		case KRANGE:
			ret = exec_exp( com );
			break;
		default:
			warning( W_ILLKIND, List[com].kind, "exec_statement" );
			return( NIL );
			break;
		}
		break;
	case KSLIST:
		at = exec_statement( List[com].v );
		if( List[com].kind == KPOSLIST && List[com].next == NIL ){
			return( at );
		} else {
			ret = cons( at, List[com].kind, exec_statement( List[com].next ) );
		}
		break;
	case KSLET:
		set_name( List[com].v, List[com].next );
		ret = com;
		List_flag++;
		break;
	case KSFOP:
	case KSCOP:
	case KSTOP:
		ret = exec_exp( com );
		break;
	default:
		warning( W_ILLKIND, List[com].kind, "exec_statement" );
		return( NIL );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function exec_command()                                */
/*                                                                    */
/*             Lisp com: CommandList                                  */
/*                                                                    */
/*             return  Lisp: Evaluated CommandList                    */
/*                                                                    */
/*             function: Evaluate CommandList.                        */
/*                      (defun, demacro)                              */
/**********************************************************************/
Lisp exec_command( Lisp com )
{
	Lisp at;
	Lisp ret;

	if( com == NIL ){
		return( NIL );
	}
	if( List[com].kind == KLET && List[List[com].v].kind == KSHARP ){
		set_name( List[com].v, List[com].next );
		ret = com;
		List_flag++;
	} else if( List[com].kind == KSLIST ){
		at = com; ret = NIL;
		while( List[at].kind == KSLIST ){
			ret = cons( ret, List[at].kind, exec_statement( List[at].v ) );
			at = List[at].next;
			if( List[at].kind == KPOSITION ){
				ret = cons( ret, List[at].kind, exec_statement( List[at].v ) );
				break;
			}
		}
	} else if( List[com].kind == KDEFUNC || List[com].kind == KDEMACRO ){
		if( (at = search_hash( GLOBAL, tsvv(com) ) ) ){
			if( GlobalList[at].kind == KPLUG ){
				warning( W_BATPLUG, tsvv(com) );
			}
		}
		set_hash( GLOBAL, tsvv(com), dup_t2gLisp(List[com].next) );
		ret = com;
		List_flag++;
	} else {
		ret = exec_statement( com );
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function do_execf()                                    */
/*                                                                    */
/*             Lisp com: Local List of Command                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Top of the command execution.                */
/*                      (duplicate LOCAL com to TMP at)               */
/**********************************************************************/
void do_execf( Lisp com )
{
	Lisp at;

	if( check_value( GLOBAL, "#debug" ) ){
		Debug_flag = 1;
	} else {
		Debug_flag = 0;
	}
	In_loop = 0;
	Top = com;
	at = dup_doLisp( com );
	Cp = at;
/*B0*/	clear_nodestack();
	exec_command( at );
}

/**********************************************************************/
/*                                                                    */
/*             function get_top()                                     */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  Lisp: Top node of Now command(LOCAL)           */
/*                                                                    */
/*             function: Return Now command top.                      */
/*                                                                    */
/**********************************************************************/
Lisp get_top( void )
{
	return( Top );
}


/**********************************************************************/
/*                                                                    */
/*             function loop_GC()                                     */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: GabageCollection in loop.(automatically raise)*/
/*                                                                    */
/**********************************************************************/
void loop_GC( void )
{
	int flag;

	check_used( List, Cp );
	check_used( List, NowBody );
	markup_hash( TMP, List );
	if( check_value( GLOBAL, "#GCreport" ) ){
		flag = 1;
	} else {
		flag = 0;
	}
	collect_list( TMP, flag );
	if( check_string_used( TMP, flag) ){
		gc_string( flag, TMP );
	}
}
