/**********************************************************************/
/*                                                                    */
/*        funcs.c -- gedo function routines                           */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*        11.21,1996 A0 add searchpath in do_include()                */
/*         4.23,1997 B1 add command viewer in eval()                  */
/*         4.27,1997 B2 add next_file() for level 2 execution.        */
/*         5.13,1997 B2 change GLOBAL var masked by TMP var           */
/*        12.16,1998 C1 flush_operation() change directory if write   */
/*                       permittion is protected.                     */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <setjmp.h>
#include <stdarg.h>
#include <ctype.h>
#include <unistd.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

extern FILE *yyin;		/* lex scanner input FILE * (==stdin) */

#ifdef MINIMUM
#define MAXLINE 	10
#define MAXSTACK	5
#else
#define MAXLINE 	256	/* Line node alloc unit */
#define MAXSTACK	50	/* Edit stack alloc unit */
#endif

static Lisp *Stack;		/* Edit stack buffer */
static int Stackp = 0;		/* stack pointer for Edit stack */
static Lisp *Line;		/* Line node */
static int MaxLine;		/* alloced number of Line node */
static int MaxStack;		/* alloced number of Edit stack */ 
extern jmp_buf errjmp;		/* if error occured, rewind stack */

/**********************************************************************/
/*                                                                    */
/*             function init_Line()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : 0==success  -1==alloc error              */
/*                                                                    */
/*             function : Initialize Line node and Edit stack.        */
/*                                                                    */
/**********************************************************************/
int init_Line( void )
{
	if( (Line = (Lisp *)malloc( sizeof(Lisp) * MAXLINE )) == NULL ){
		fatal( F_MALLOC, "Line" );
		return( -1 );
	}
	MaxLine = MAXLINE;
	if( (Stack = (Lisp *)malloc( sizeof(Lisp) * MAXSTACK )) == NULL ){
		fatal( F_MALLOC, "Stack" );
		return( -1 );
	}
	MaxStack = MAXSTACK;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function free_Line()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Free the Line node and Edit stack.          */
/*                                                                    */
/**********************************************************************/
void free_Line( void )
{
	free( Line );
	free( Stack );
}

/**********************************************************************/
/*                                                                    */
/*             function do_help()                                     */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to help command.                  */
/*                                                                    */
/**********************************************************************/
void do_help( void )
{
	char *pager;
	Lisp at;
	char *path;
	char wk[256];

	if( Level == 1 ){
		return;
	}
/* B2 add search_global() */
/*	if( (at = search_hash( GLOBAL, "#pager" )) != NIL ){
		pager = gsv(at);*/
	if( (at = search_global( "#pager" )) != NIL ){
		pager = tsv(at);
	} else {
		if( (pager = getenv( PAGER )) == NULL ){
			warning( W_NOENV, "PAGER" );
		}
	}
/* B2 add search_global() */
/*	if( (at = search_hash( GLOBAL, "#gedopath" )) != NIL ){
		strcpy( wk, gsv(at) );*/
	if( (at = search_global( "#gedopath" )) != NIL ){
		strcpy( wk, tsv(at) );
		strcat( wk, "gedo.help" );
		path = wk;
	} else {
		path = HELPFILEPATH;
	}
	sprintf( Work, "%s %s", pager, path );
	end_raw_tty();
	if( system( Work ) != 0 ){
		reset_raw_tty();
		warning( W_SHELL );
	}
	reset_raw_tty();
}

/**********************************************************************/
/*                                                                    */
/*             function set_range()                                   */
/*                                                                    */
/*             Lisp com       : RANGE list                            */
/*             struct Pos *at : start and end                         */
/*                                                                    */
/*             return  int : 0==success  -1==illegal syntax           */
/*                                                                    */
/*             function : At RANGE use command, interpret the range   */
/*                       expression.                                  */
/**********************************************************************/
int set_range( Lisp com, struct Pos *at )
{
	Lisp lst, led;

	if( List[com].kind != KRANGE ){
		return( -1 );
	}
	lst = List[com].v; led = List[com].next;
	if( List[lst].kind != KNUMBER ){
		return( -1 );
	}
	if( led == NIL ){
		if( List[lst].v < 0 ){
			at->x = 1; at->y = List[lst].v * (-1);
		} else {
			at->x = at->y = List[lst].v;
		}
	} else {
		if( List[led].kind != KNUMBER ){
			return( -1 );
		}
		if( List[lst].v > 0 ){
			at->x = List[lst].v;
		} else {
			return( -1 );
		}
		if( List[led].v < 0 ){
			at->y = List[led].v * (-1);
		} else {
			at->y = List[led].v;
		}
	}
	return( 0 );
}	


/**********************************************************************/
/*                                                                    */
/*             function do_eval()                                     */
/*                                                                    */
/*             Lisp com : argment of eval command(RANGE)              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to eval command.                  */
/*                                                                    */
/**********************************************************************/
void do_eval( Lisp com )
{
	int s, e, i;
	struct Pos at;

	Level = 0;
	if( com == NIL ){
		s = 0; e = NowLine;
	} else {
		if( set_range( com, &at ) == -1 ){
			warning( W_EVAL );
			return;
		}
		s = at.x - 1; e = at.y;
	}
	for( i = s; i < e && i < NowLine; i++ ){
		do_execf( Line[i] );
		sleep( 1 );
	}
	Level = RealLevel;
	List_flag = 0;
}


/**********************************************************************/
/*                                                                    */
/*             function do_list()                                     */
/*                                                                    */
/*             Lisp com : argment of list command(RANGE)              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to list command.                  */
/*                                                                    */
/**********************************************************************/
void do_list( Lisp com )
{
	int s, e, i;
	int kind;
	struct Pos po;

	kind = 0;
	if( com == NIL ){
		s = 0; e = NowLine;
	} else if( List[com].kind == KCSTACK ){
		if( List[com].next ){
			if( set_range( List[com].next, &po ) == -1 ){
				warning( W_LIST );
				return;
			}
			s = po.x - 1; e = po.y;
		} else {
			s = 0; e = Stackp;
		}
		kind = 1;
	} else {
		if( set_range( com, &po ) == -1 ){
			warning( W_LIST );
			return;
		}
		s = po.x - 1; e = po.y;
	}
	if( kind == 0 ){
		for( i = s; i < e && i < NowLine; i++ ){
			display( M_LINE, i + 1 );
			display_command( dup_l2tLisp(Line[i]) );
		}
	} else {
		for( i = s; i < e && i < Stackp; i++ ){
			display( M_LINE, i + 1 );
			display_command( dup_l2tLisp(Stack[i]) );
		}
	}
	putchar( '\n' );
}

/**********************************************************************/
/*                                                                    */
/*             function do_clear()                                    */
/*                                                                    */
/*             Lisp com : argment of clear command(RANGE)             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to clear command.                 */
/*                                                                    */
/**********************************************************************/
void do_clear( Lisp com )
{
	int s, e;
	int kind;
	struct Pos po;

	kind = 0;
	if( com == NIL ){
		NowLine = 0;
		return;
	} else if( List[com].kind == KCSTACK ){
		if( List[com].next ){
			if( set_range( List[com].next, &po ) == -1 ){
				warning( W_CLEAR );
				return;
			}
			s = po.x - 1; e = po.y;
		} else {
			Stackp = 0;
			return;
		}
		kind = 1;
	} else {
		if( set_range( com, &po ) == -1 ){
			warning( W_CLEAR );
			return;
		}
		s = po.x - 1; e = po.y;
	}
	if( kind == 0 ){
		if( e >= NowLine ){
			e = NowLine;
		}
		memmove( &Line[s], &Line[e], (NowLine-e+1) * sizeof(Lisp) );
		NowLine = NowLine + s - e;
	} else {
		if( e >= Stackp ){
			e = Stackp;
		}
		memmove( &Stack[s], &Stack[e], (Stackp-e+1) * sizeof(Lisp) );
		Stackp = Stackp + s - e;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_pop()                                      */
/*                                                                    */
/*             Lisp com : argment of pop command                      */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to pop command.                   */
/*                                                                    */
/**********************************************************************/
void do_pop( Lisp com )
{
	int s, e, i;
	Lisp *tmp;
	struct Pos po;

	if( com == NIL ){
		if( Stackp + 1 >= MaxStack ){
			if( (tmp = (Lisp *)realloc( Stack, sizeof(Lisp) * (MaxStack + MAXSTACK) )) == NULL ){
				warning( W_STCOSM );
				return;
			}
			Stack = tmp;
			MaxStack += MAXSTACK;
		} else if( NowLine == 0 ){
			warning( W_NOLINE );
			return;
		} else {
			Stack[Stackp++] = Line[--NowLine];
			set_command_buf( Work, dup_l2tLisp(Line[NowLine]) );
			set_ring( Work, strlen(Work) );
		}
		return;
	}
	if( set_range( com, &po ) == -1 ){
		warning( W_POP );
		return;
	}
	s = po.x - 1; e = po.y;

	if( e >= NowLine ){
		e = NowLine;
	}
	for( i = e - 1; i >= s; i-- ){
		if( Stackp + 1 >= MaxStack ){
			if( (tmp = (Lisp *)realloc( Stack, sizeof(Lisp) * (MaxStack + MAXSTACK) )) == NULL ){
				s = i;
				display( M_STCOSM );
				break;
			}
			Stack = tmp;
			MaxStack += MAXSTACK;
		}
		Stack[Stackp++] = Line[i];
		Work[0] = '\0';
		set_command_buf( Work, dup_l2tLisp(Line[i]) );
		set_ring( Work, strlen(Work) );
	}
	memmove( &Line[s], &Line[e], (NowLine-e) * sizeof(Lisp) );
	NowLine = NowLine + s - e;
}

/**********************************************************************/
/*                                                                    */
/*             function do_get()                                      */
/*                                                                    */
/*             Lisp com : argment of get command.                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to get command.                   */
/*                                                                    */
/**********************************************************************/
void do_get( Lisp com )
{
	int s, e, i;
	Lisp *tmp;
	struct Pos po;

	if( com == NIL ){
		if( Stackp + 1 >= MaxStack ){
			if( (tmp = (Lisp *)realloc( Stack, sizeof(Lisp) * (MaxStack + MAXSTACK) )) == NULL ){
				warning( W_STCOSM );
			}
			Stack = tmp;
			MaxStack += MAXSTACK;
		} else if( NowLine == 0 ){
			warning( W_NOLINE );
		} else {
			Stack[Stackp++] = dup_l2lLisp(Line[NowLine-1]);
		}
		set_command_buf( Work, dup_l2tLisp(Line[NowLine-1]) );
		set_ring( Work, strlen(Work) );
		return;
	}
	if( set_range( com, &po ) == -1 ){
		warning( W_GET );
		return;
	}
	s = po.x - 1; e = po.y;

	if( e >= NowLine ){
		e = NowLine;
	}
	for( i = e - 1; i >= s; i-- ){
		if( Stackp + 1 >= MaxStack ){
			if( (tmp = (Lisp *)realloc( Stack, sizeof(Lisp) * (MaxStack + MAXSTACK) )) == NULL ){
				display( M_STCOSM );
				break;
			}
			Stack = tmp;
			MaxStack += MAXSTACK;
		}
		Stack[Stackp++] = dup_l2lLisp(Line[i]);
		Work[0] = '\0';
		set_command_buf( Work, dup_l2tLisp(Line[i]) );
		set_ring( Work, strlen(Work) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_push()                                     */
/*                                                                    */
/*             Lisp com : argment of push command.                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to push command.                  */
/*                                                                    */
/**********************************************************************/
void do_push( Lisp com )
{
	int s, i;
	Lisp *tmp;

	if( com == NIL ){
		while( Stackp > 0 ){
			if( NowLine + 1 >= MaxLine ){
				if( (tmp = (Lisp *)realloc( Line, sizeof(Lisp) * (MaxLine + MAXLINE) )) == NULL ){
					display( M_LNCOSM );
					return;
				}
				Line = tmp;
				MaxLine += MAXLINE;
			}
			Line[NowLine++] = Stack[--Stackp];
		}
		return;
	}
	if( List[com].kind != KNUMBER ){
		warning( W_PUSH );
		return;
	} else {
		s = List[com].v - 1;
		if( s < 0 ){
			warning( W_PUSH );
			return;
		}
	}
	while( NowLine + Stackp >= MaxLine ){
		if( (tmp = (Lisp *)realloc( Line, sizeof(Lisp) * (MaxLine + MAXLINE) )) == NULL ){
			warning( W_LNCOSM );
			return;
		}
		Line = tmp;
		MaxLine += MAXLINE;
	}
	memmove( &Line[s+Stackp], &Line[s], Stackp * sizeof(Lisp) );
	NowLine = NowLine + Stackp;
	for( i = s; Stackp; i++ ){
		Line[i] = Stack[--Stackp];
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_put()                                      */
/*                                                                    */
/*             Lisp com : argment of put command                      */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to put command.                   */
/*                                                                    */
/**********************************************************************/
void do_put( Lisp com )
{
	int s, i;
	Lisp *tmp;

	if( com == NIL ){
		s = Stackp;
		while( s > 0 ){
			if( NowLine + 1 >= MaxLine ){
				if( (tmp = (Lisp *)realloc( Line, sizeof(Lisp) * (MaxLine + MAXLINE) )) == NULL ){
					display( M_LNCOSM );
					return;
				}
				Line = tmp;
				MaxLine += MAXLINE;
			}
			Line[NowLine++] = dup_l2lLisp(Stack[--s]);
		}
		return;
	}
	if( List[com].kind != KNUMBER ){
		warning( W_PUT );
		return;
	} else {
		s = List[com].v - 1;
		if( s < 0 ){
			warning( W_PUT );
		}
	}
	while( NowLine + Stackp >= MaxLine ){
		if( (tmp = (Lisp *)realloc( Line, sizeof(Lisp) * (MaxLine + MAXLINE) )) == NULL ){
			warning( W_LNCOSM );
			return;
		}
		Line = tmp;
		MaxLine += MAXLINE;
	}
	memmove( &Line[s+Stackp], &Line[s], Stackp * sizeof(Lisp) );
	NowLine = NowLine + Stackp;
	for( i = Stackp - 1; i >= 0; i-- ){
		Line[s++] = dup_l2lLisp(Stack[i]);
	}
}

/**********************************************************************/
/*                                                                    */
/*             function do_undo()                                     */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to undo command.                  */
/*                                                                    */
/**********************************************************************/
void do_undo( void )
{
	int i;

	if( NowLine == 0 ){
		warning( W_NOLINE );
		return;
	}
	for( i = 0; i < NowLine - 1; i++ ){
		do_execf( Line[i] );
	}
	NowLine--;
}


/**********************************************************************/
/*                                                                    */
/*             function copy_commands()                               */
/*                                                                    */
/*             int src   : Orginal File number for copy.              */
/*             int dests : Destination File number to copy.           */
/*             int deste : End of the Destination File number.        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Do the copy routine called from do_copy().  */
/*                                                                    */
/**********************************************************************/
void copy_commands( int src, int dests, int deste )
{
	char filename[FILENAME_MAX];
	FILE *fpin, *fpout;
	char *buff;
	int i, size;

	strcpy( filename, gs(FileName[src]) );
 	make_cmdname( 0, filename );
	if( (fpin = fopen( filename, "rt" )) == NULL ){
		fclose( fpin );
		warning( W_ECOPY, filename );
	}
	size = filelength( fileno(fpin) );
	if( (buff = malloc( size )) == NULL ){
		warning( W_MALLOC );
	}
	if( fread( buff, size, 1, fpin ) != 1 ){
		fclose( fpin );
		free( buff );
		warning( W_RTMP, gs(FileName[src]) );
	}
	fclose( fpin );
	fpout = NULL;
	for( i = dests; i < deste; i++ ){
		strcpy( filename, gs(FileName[i]) );
		make_cmdname( 0, filename );
		if( (fpout = fopen( filename, "wt" )) == NULL ){
			fclose( fpout );
			free( buff );
			warning( W_ECOPY, filename );
		}
		if( fwrite( buff, size, 1, fpout ) != 1 ){
			fclose( fpout );
			free( buff );
			warning( W_WTMP, gs(FileName[i]) );
		}
		fclose( fpout );
	}
	free( buff );
}

/**********************************************************************/
/*                                                                    */
/*             function do_copy()                                     */
/*                                                                    */
/*             Lisp com : argment of copy command                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to copy command.                  */
/*                                                                    */
/**********************************************************************/
void do_copy( Lisp com )
{
	int src;
	int dests;
	int deste;
	Lisp at;

	if( List[com].kind == KNUMBER ){
		src = List[com].v - 1;
	} else if( List[com].kind == KFN ){
		switch( List[com].v ){
		case KCAT:
			src = FileNum;
			break;
		case KCPREV:
			src = FileNum - 1;
			break;
		case KCNEXT:
			src = FileNum + 1;
			break;
		default:
			warning( W_COPY );
			return;
			break;
		}
	} else {
		warning( W_COPY );
		return;
	}
	at = List[com].next;
	if( List[at].kind == KNUMBER ){
		dests = List[at].v - 1;
		deste = dests + 1;
	} else if( List[at].kind == KFN ){
		switch( List[at].v ){
		case KCAT:
			dests = FileNum;
			deste = dests + 1;
			break;
		case KCPREV:
			dests = FileNum - 1;
			deste = dests + 1;
			break;
		case KCNEXT:
			dests = FileNum + 1;
			deste = dests + 1;
			break;
		case KCBEFORE:
			dests = 0;
			deste = FileNum;
			break;
		case KCAFTER:
			dests = FileNum + 1;
			deste = TotalFileNum;
			break;
		default:
			warning( W_COPY );
			return;
			break;
		}
	} else {
		warning( W_COPY );
		return;
	}
	if( src == FileNum ){
		flush_operation( 0 );
	}
	copy_commands( src, dests, deste );
	if( dests == FileNum ){
		clear_localLisp();
		clear_maskinfo();
		NowLine = 0;
		change_file();
	}
}


/**********************************************************************/
/*                                                                    */
/*             function flush_operation()                             */
/*                                                                    */
/*             int flag : 1==write the operation command called this. */
/*                        0==not write                                */
/*             return NONE                                            */
/*                                                                    */
/*             function : Flush all Line node at the File to the tmp  */
/*                       file.                                        */
/*             C1: change directory if write permittion is protected. */
/**********************************************************************/
void flush_operation( int flag )
{
	FILE *fp;
	char name[FILENAME_MAX];
	int i;

	if( FileName[FileNum] == NIL || *(gs(FileName[FileNum])) == '\0' ){
		sprintf( name, "def%d", FileNum );
		FileName[FileNum] = global_strdup( name );
	} else {
		strcpy( name, gs(FileName[FileNum)] );
	}
	make_cmdname( 0, name );
	if( NowLine == 0 ){
		if( access( name, F_OK ) == -1 ){
			return;
		}
	}
	if( (fp = fopen( name, "w" )) == NULL ){
/* add C1 */
	     Lisp lwp;
	     if( (lwp = search_global( "#lwp" )) != NIL ){
		  if( List[lwp].kind != KLIST || (lwp = List[lwp].next) == NIL
		      || List[lwp].kind != KLIST || (lwp = List[lwp].v) == NIL
		      || List[lwp].kind != KLITERAL ){
		       display( W_ECOPY, "tmp file(not set #lwp)" );
		       return;
		  }
		  sprintf( Work, "%s/%s", tsv(lwp), name );
		  if( (fp = fopen( Work, "w" )) == NULL ){
		       display( W_ECOPY, "tmp file" );
		       return;
		  }
		  strcpy( name, Work );
	     } else {
/* add end */
		warning( W_ECOPY, name );
	     }
	}
	for( i = 0; i < NowLine; i++ ){
		write_command_file( fp, dup_l2tLisp(Line[i]) );
	}
	if( flag ){
		write_command_file( fp, dup_l2tLisp(get_top()) );
	}
	fclose( fp );
}

/**********************************************************************/
/*                                                                    */
/*             function level2_chgfile()                              */
/*                                                                    */
/*             String str : load file name                            */
/*                                                                    */
/*             return  int : 0==success  -1==fail                     */
/*                                                                    */
/*             function : At level==2, read output file and change    */
/*                       File.                                        */
/**********************************************************************/
int level2_chgfile( String str )
{
	char *p, *q;
	String *tmp;
	char *file;

	file = ts(str);
	if( Level != 2 ){
		warning( W_NOCOMLN );
		return( -1 );
	}
	if( FileNum >= SerialEnd  ){
		do_quit();
		return( 0 );
	}
	for( p = file; isspace(*p); p++ );
	for( q = Work; *p; p++ ){
		if( isspace(*p) ){
			*q = '\0';
			break;
		} else {
			*q++ = *p;
		}
	}
	FileName[FileNum] = global_strdup( Work );
	clear_localLisp();
	clear_maskinfo();
	NowLine = 0;
	if( FileNum + 1 >= FileMax ){
		if( (tmp = (String *)realloc( FileName, sizeof(String) * (FileMax + FILEMAX) )) == NULL ){
			warning( W_FNCOSM );
			return( -1 );
		}
		FileName = tmp;
		FileMax += FILEMAX;
	}
	FileNum++;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function do_quit()                                     */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to quit command.                  */
/*                                                                    */
/**********************************************************************/
void do_quit( void )
{
	int c;

	if( (c = check_save_outfile()) == 0 ){
		return;
	}
	flush_operation( 0 );
	clear_localLisp();
	clear_maskinfo();
	NowLine = 0;
	endp( c );
	exit( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function do_print()                                    */
/*                                                                    */
/*             Lisp com : argment of print command                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to print command.                 */
/*                                                                    */
/**********************************************************************/
void do_print( Lisp com )
{
	Lisp at;

	if( com == NIL ){
		warning( W_NOMACRO, "(null)" );
	}
	if( List[com].v == NIL ){
		warning( W_ILLLST );
		return;
	}

	if( List[com].kind == KSHARP ){
		at = expand_macro( List[com].v );
	} else if( List[com].kind == KSYMBOL ){
		if( (at = get_var( List[com].v )) == NIL ){
			at = expand_macro( List[com].v );
		}
	} else {
		at = eval_exp( com );
	}
	if( List[at].kind == KARGS ){
		at = cons( com, KDEFUNC, at );
	} else if( List[at].kind == KMARGS ){
		at = cons( com, KDEMACRO, at );
	}
	write_of_print( at );
}

/**********************************************************************/
/*                                                                    */
/*             function set_clever_name()                             */
/*                                                                    */
/*             char *name : variable name                             */
/*             Lisp at    : the value of the variable                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : At tmp or local varibale, set the variable  */
/*                       with deleting old value.                     */
/**********************************************************************/
void set_clever_name( char *name, Lisp at )
{
	Lisp ls;

	if( (ls = search_hash( TMP, name )) ){
		delete_Lisp( TMP, ls );
		set_hash( TMP, name, at );
	} else {
		if( (ls = search_hash( LOCAL, name )) ){
			delete_Lisp( LOCAL, ls );
		}
		set_hash( LOCAL, name, dup_t2lLisp(at) );
	}
}
/**********************************************************************/
/*                                                                    */
/*             function do_include()                                  */
/*                                                                    */
/*             Lisp at : argment of include command(FILENAME)         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Interface to include command.               */
/*                                                                    */
/**********************************************************************/
void do_include( Lisp at )
{
	int save_init_flag;
	FILE *save_yyin;

	if( List[at].kind != KLITERAL ){
		warning( W_INCLUDE );
		return;
	}
/* A0 add searchpath */
	if( search_path( tsv(at), MINCLUDE, Work ) == -1 ){
		warning( W_NOFILE );
	}
/* A0 end */
	save_init_flag = Init_flag;
	Init_flag = 1;
	save_yyin = yyin;
/* A0 
	load_init( tsv(at) );*/
	load_init( Work );
/* A0 end */
	Init_flag = save_init_flag;
	yyin = save_yyin;
	yyrestart( yyin );
	List_flag = 0;
}
/*
void do_query( void )	in lisp.c
*/

/*
void do_gc( void )	in lisp.c
*/

/**********************************************************************/
/*                                                                    */
/*             function next_file()                                   */
/*                                                                    */
/*             int flag : 0==if next is not, create new               */
/*                        1==if next is not, abort                    */
/*             return  int : 0==next do                               */
/*                           1==next is not                           */
/*                          -1==alloc error                           */
/*             function : Prepare the next file.                      */
/*            B2 add for level 2 proc                                 */
/**********************************************************************/
int next_file( int flag )
{
	String *tmp;

	flush_operation( 0 );
	clear_localLisp();
	clear_maskinfo();
	NowLine = 0; Stackp = 0;
	if( ++FileNum >= TotalFileNum ){
		if( flag ){
			return( 1 );
		}
		SerialEnd++; TotalFileNum++;
		sprintf( Work, LoadFileForm, FileNum );
		if( FileNum >= FileMax ){
			if( (tmp = (String *)realloc( FileName, sizeof(String) * (FileMax + FILEMAX) )) == NULL ){
				warning( W_FNCOSM );
				return( -1 );
			}
			FileName = tmp;
			FileMax += FILEMAX;
		}
		FileName[FileNum] = global_strdup( Work );
	}
	change_file();
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function do_del()                                      */
/*                                                                    */
/*             Lisp com : variable of sharp                           */
/*                                                                    */
/*             return  void                                           */
/*                                                                    */
/*             function : del command                                 */
/*                                                                    */
/**********************************************************************/
void do_del( Lisp com )
{

	fprintf( stderr, "sorry, not supported at ver 0.99\nnext version should be supported\n" );
#if 0
	Lisp at;

	if( List[com].kind == KSYMBOL ){
		if( (at = search_hash( LOCAL, tsv(com) )) == NIL ){
			return;
		}
		delete_Lisp( LOCAL, at );
		delete_hash( LOCAL, tsv(com) );
	} else if( List[com].kind == KSHARP ){
		if( (at = search_hash( GLOBAL, tsv(com) )) == NIL ){
			return;
		}
		delete_Lisp( GLOBAL, at );
		delete_hash( GLOBAL, tsv(com) );
	}
	do_gc();
#endif
}

/**********************************************************************/
/*                                                                    */
/*             function eval()                                        */
/*                                                                    */
/*             Lisp com : COMMAND list                                */
/*                                                                    */
/*             return  Lisp : return com                              */
/*                                                                    */
/*             function : Top level of executing the command.         */
/*            B2 : move the operation of MNEXT to next_file().        */
/**********************************************************************/
Lisp eval( Lisp com )
{
	Lisp second;
	int val;

	second = List[com].next;
	switch( List[com].v ){
	case MLOADER:
		do_configer( com );
		break;
	case MSAVER:
		do_configer( com );
		break;
	case MLEVEL:
		if( List[second].kind != KNUMBER ){
			warning( W_LEVEL );
		}
		Level = List[second].v;
		if( Init_flag ){
			if( Level > 2 ){
				warning( W_NOLEVEL, Level );
			}
		} else {
			if( Level > 1 ){
				warning( W_NOLEVEL, Level );
			}
		}
		RealLevel = Level;
		display( M_LEVEL, Level );
		break;
	case MLOAD:
		do_load( com );
		break;
	case MSAVE:
		do_save( com );
		break;
	case MSELECTER:
		do_configer( com );
		break;
	case MSELECT:
		do_select( com );
		break;
	case MEFFECT:
		do_effect( com );
		break;
	case MEFFECTER:
		do_configer( com );
		break;
	case MEVAL:
		do_eval( second );
		break;
	case MCOPY:
		do_copy( second );
		break;
	case MQUIT:
		do_quit();
		break;
	case MNEXT:
		/* B2 move the operation to next_file() */
		next_file( 0 );
		break;
	case MPREV:
		if( FileNum - 1 >= 0 ){
			flush_operation( 0 );
			clear_localLisp();
			clear_maskinfo();
			NowLine = 0; Stackp = 0;
			FileNum--;
			change_file();
		}
		break;
	case MGOTO:
		if( List[second].kind != KNUMBER ){
			warning( W_GOTO );
		}
		val = List[second].v - 1;
		if( val >= TotalFileNum ){
			warning( W_GOAFT, TotalFileNum );
		} else if( val < 0 ){
			warning( W_GOBFR );
		}
		flush_operation( 0 );
		clear_localLisp();
		clear_maskinfo();
		NowLine = 0; Stackp = 0;
		FileNum = val;
		change_file();
		break;
	case MLIST:
		do_list( second );
		break;
	case MUNDO:
		do_undo();
		break;
	case MCLEAR:
		do_clear( second );
		break;
	case MPUSH:
		do_push( second );
		break;
	case MPOP:
		do_pop( second );
		break;
	case MPUT:
		do_put( second );
		break;
	case MGET:
		do_get( second );
		break;
	case MSHOW:
		do_show( second );
		break;
	case MGUI:
		init_gui();
		break;
	case MPURGE:
		do_purge( second );
		break;
	case MHELP:
		do_help();
		break;
	case MINCLUDE:
		do_include( second );
		break;
	case MPRINT:
		do_print( second );
		break;
	case MQUERY:
		do_query( 1 );
		break;
	case MGC:
		do_gc();
		break;
	case MREHASH:
		NowLine = 0; Stackp = 0;
		change_file();
		List_flag = 0;
		break;
	case MCREATE:
		do_create( second );
		break;
/* B1 add */
	case MVIEWER:
		do_viewer( second );
		break;
	case MDEL:
		do_del( second );
		break;
	case MRESET:
		do_reset();
		break;
/* B1 end */
	default:
		warning( W_NOCOM );
		break;
	}
	return( com );
}

/**********************************************************************/
/*                                                                    */
/*             function execf()                                       */
/*                                                                    */
/*             Lisp com : command line                                */
/*                                                                    */
/*             return  int : 0==success  1==fail                      */
/*                                                                    */
/*             function : Top level of command-line execution.        */
/*                                                                    */
/**********************************************************************/
int execf( Lisp com )
{
	Lisp *tmp;

	if( com == NIL ){ /* .ini section */
		return( 0 );
	}
	if( setjmp( errjmp ) ){
		return( 0 );
	}
	if( File_flag ){
		display_command( dup_l2tLisp(com) );
	}
	List_flag = 0;
	do_execf( com );
	if( List_flag ){
		if( NowLine + 1 >= MaxLine ){
			if( (tmp = (Lisp *)realloc( Line, sizeof(Lisp) * (MaxLine + MAXLINE) )) == NULL ){
				warning( W_LNCOSM );
				return( 1 );
			}
			Line = tmp;
			MaxLine += MAXLINE;
		}
		Line[NowLine++] = com;
	}
	if( query( 0, LOCAL ) || query( 0, GLOBAL ) ){
		do_gc();
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function markup_line()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Mark-up the Cell in Line node and Edit stack*/
/*                       (alive Cell) for GC.                         */
/**********************************************************************/
void markup_line( void )
{
	int i;

	for( i = 0; i < NowLine; i++ ){
		check_used( LocalList, Line[i] );
	}
	check_used( LocalList, get_top() );
	for( i = 0; i < Stackp; i++ ){
		check_used( LocalList, Stack[i] );
	}
}
