%{
/**********************************************************************/
/*                                                                    */
/*        gedo.y -- gedo syntax parser generator                      */
/*                                                                    */
/*                by K.sugiura   Ver 1.0  for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*        1.16,1996: B0 add kakko2 and branch (calc) and (calc,calc,..) */
/*        4.23,1997: B2 add viewer and VIEWER token                   */
/**********************************************************************/
#include <stdio.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"
%}
%union {
	String   pt;
	int   val;
	Lisp  lp;
}
%token <pt> SHARP LITERAL NAME ITEM
%token <val> NUM
%token LOAD SAVE SELECTER SELECT EFFECT LIST UNDO CLEAR EVAL COPY
%token QUIT LEVEL NEXT PREV GOTO BEFORE AFTER AT PUSH POP STACK SHOW GUI
%token LOADER SAVER HELP CONFIG FILEMARK GET PUT
%token ENDSECTION SELECTSECTION EFFECTSECTION LOADSECTION EFFECTER
%token INCLUDE PURGE PRINT WHILE FOR
%token EOL FEND EQ NE GE LE SNIL
%token IF ELSE QUERY GC REHASH CREATE
%token LETPLUS LETMINUS LETMULT LETDIVIDE LETAND LETOR
%token DEFUN RETURN DEMAC
%token VIEWER DEL RESET
%nonassoc '(' ')'
%left ','
%left '+' '-'
%left '*' '/' '%'
%left '&' '|'
%token INC DEC
%nonassoc '~'
%nonassoc UMINUS
%type <lp> load save selecter select effect effecter list undo
%type <lp> eval clear copy quit level next prev goto push pop show gui
%type <lp> loader saver src dest
%type <lp> include purge help print get put
%type <lp> items loop statement_list statement command
%type <lp> expression poslist position calclist calc data forexp
%type <lp> num literal sharp name nil
%type <lp> block query gc rehash create
%type <lp> defun namelist range kakko kakko2
%type <lp> viewer del reset
%%
top:		statement_list EOL	{ return( execf( $1 ) );}
	|	sharp '=' block EOL	{ return( execf( l_cons( $1, KLET, $3 ) ) ); }
	|	FILEMARK LITERAL EOL	{ return( level2_chgfile( $2 ) ); }
	|	initial	EOL		{ return( 0 ); }
	|	FEND			{ return( 1 ); }
	|	EOL			{ return( 0 ); }
	|	defun EOL		{ return( execf( $1 ) ); }
	;

defun:		DEFUN name '(' namelist ')' '{' statement_list '}'
			{ $$ = l_cons( $2, KDEFUNC, l_cons( $4, KARGS, $7 ) ); }
	|	DEFUN name '(' ')' '{' statement_list '}'
			{ $$ = l_cons( $2, KDEFUNC, l_cons( NIL, KARGS, $6 ) ); }
	;
	|	DEMAC name '(' namelist ')' '{' statement_list '}'
			{ $$ = l_cons( $2, KDEMACRO, l_cons( $4, KMARGS, $7 ) ); }
	|	DEMAC name '(' ')' '{' statement_list '}'
			{ $$ = l_cons( $2, KDEMACRO, l_cons( NIL, KMARGS, $6 ) ); }
	;

namelist:	name			{ $$ = l_cons( $1, KLIST, NIL ); }
	|	name ',' namelist	{ $$ = l_cons( $1, KLIST, $3 ); }
	;

loop:		WHILE '(' expression ')' block	{ $$ = l_cons( $3, KWHILE, $5 ); }
	|	FOR forexp block  	{ $$ = l_cons( $2, KFOR, $3 ); }
	|	IF '(' expression ')' block	{ $$ = l_cons( $3, KIF, l_cons( $5, KLIST, NIL ) ); }
	|	IF '(' expression ')' block ELSE block	{ $$ = l_cons( $3, KIF, l_cons( $5, KLIST, $7 ) ); }
	;

block:		'{' statement_list '}'	{ $$ = $2; }
	|	sharp			{ $$ = $1; }
	;

forexp:		'(' expression ';' expression ';' expression ')'
			{ $$ = l_cons( $2, KLIST, l_cons( $4, KLIST, $6 ) ); }
	;

statement_list:	statement	{ $$ = l_cons( $1, KLIST, NIL ); }
	|	statement ',' statement_list	{ $$ = l_cons( $1, KLIST, $3 ); }
	;

statement:	command		{ $$ = $1; }
	|	expression	{ $$ = $1; }
	|	loop		{ $$ = $1; }
	;

command:	load		{ $$ = $1; }
	|	save		{ $$ = $1; }
	|	loader		{ $$ = $1; }
	|	saver		{ $$ = $1; }
	|	selecter	{ $$ = $1; }
	|	select		{ $$ = $1; }
	|	effect		{ $$ = $1; }
	|	effecter	{ $$ = $1; }
	|	list		{ $$ = $1; }
	|	undo		{ $$ = $1; }
	|	push		{ $$ = $1; }
	|	pop		{ $$ = $1; }
	|	clear		{ $$ = $1; }
	|	eval		{ $$ = $1; }
	|	copy		{ $$ = $1; }
	|	quit		{ $$ = $1; }
	|	level		{ $$ = $1; }
	|	next		{ $$ = $1; }
	|	prev		{ $$ = $1; }
	|	goto		{ $$ = $1; }
	|	show		{ $$ = $1; }
	|	gui		{ $$ = $1; }
	|	get		{ $$ = $1; }
	|	put		{ $$ = $1; }
	|	include		{ $$ = $1; }
	|	purge		{ $$ = $1; }
	|	help		{ $$ = $1; }
	|	print		{ $$ = $1; }
	|	query		{ $$ = $1; }
	|	gc		{ $$ = $1; }
	|	rehash		{ $$ = $1; }
	|	create		{ $$ = $1; }
	|	viewer		{ $$ = $1; }
	|	del		{ $$ = $1; }
	|	reset		{ $$ = $1; }
	;

initial:	section		{ }
	|	ENDSECTION	{ }
	|	items		{ set_child( $1 ); }
	;

section:	LOADSECTION	{ set_section( MLOADER ); }
	|	SELECTSECTION	{ set_section( MSELECTER ); }
	|	EFFECTSECTION	{ set_section( MEFFECTER ); }
	;

items:		ITEM 		{ $$ = l_cons( $1, KSYMBOL, NIL ); }
	|	ITEM items	{ $$ = l_cons( $1, KSYMBOL, $2 ); }
	;

gui:		GUI		{ $$ = l_cons( MGUI, KCOM, NIL ); }
	;

load:		LOAD 		{ $$ = l_cons( MLOAD, KCOM, l_cons( NIL, KVOID, NIL ) ); }
	|	LOAD calc	{ $$ = l_cons( MLOAD, KCOM, $2 ); }
	;

save:		SAVE        	{ $$ = l_cons( MSAVE, KCOM, l_cons( NIL, KVOID, NIL ) ); }
	|	SAVE calc	{ $$ = l_cons( MSAVE, KCOM, $2 ); }
	;

loader:		LOADER name
		{ $$ = l_cons( MLOADER, KCOM, $2 ); }
	|	LOADER name calc
		{ $$ = l_cons( MLOADER, KCOM, llist( $2, $3 ) ); }
	|	LOADER name CONFIG
		{ $$ = l_cons( MLOADER, KCOM, llist( $2, l_cons( NIL, KVOID, NIL ) ) ); }
	|	LOADER HELP	{ $$ = l_cons( MLOADER, KCOM, l_cons( NIL, KFN, NIL ) ); }
	|	LOADER name HELP
		{ $$ = l_cons( MLOADER, KCOM, llist( $2, l_cons( NIL, KFN, NIL ) ) ); }
	;

saver:		SAVER name
		{ $$ = l_cons( MSAVER, KCOM, $2 ); }
	|	SAVER name calc
		{ $$ = l_cons( MSAVER, KCOM, llist( $2, $3 ) ); }
	|	SAVER name CONFIG
		{ $$ = l_cons( MSAVER, KCOM, llist( $2, l_cons( NIL, KVOID, NIL ) ) ); }
	|	SAVER HELP	{ $$ = l_cons( MSAVER, KCOM, l_cons( NIL, KFN, NIL ) ); }
	|	SAVER name HELP
		{ $$ = l_cons( MSAVER, KCOM, llist( $2, l_cons( NIL, KFN, NIL ) ) ); }
	;

selecter:	SELECTER name
		{ $$ = l_cons( MSELECTER, KCOM, $2 ); }
	|	SELECTER name calc
		{ $$ = l_cons( MSELECTER, KCOM, llist( $2, $3 ) ); }
	|	SELECTER name CONFIG
		{ $$ = l_cons( MSELECTER, KCOM, llist( $2, l_cons( NIL, KVOID, NIL ) ) ); }
	|	SELECTER HELP
		{ $$ = l_cons( MSELECTER, KCOM, l_cons( NIL, KFN, NIL ) ); }
	|	SELECTER name HELP
		{ $$ = l_cons( MSELECTER, KCOM, llist( $2, l_cons( NIL, KFN, NIL ) ) ); }
	;

select:		SELECT		{ $$ = l_cons( MSELECT, KCOM, l_cons( NIL, KVOID, NIL ) ); }
	|	SELECT calc	{ $$ = l_cons( MSELECT, KCOM, l_cons( NIL, KVOID, $2 ) ); }
	;

effecter:	EFFECTER name
		{ $$ = l_cons( MEFFECTER, KCOM, $2 ); }
	|	EFFECTER name calc
		{ $$ = l_cons( MEFFECTER, KCOM, llist( $2, $3 ) ); }
	|	EFFECTER name CONFIG
		{ $$ = l_cons( MEFFECTER, KCOM, llist( $2, l_cons( NIL, KVOID, NIL ) ) ); }
	|	EFFECTER HELP	{ $$ = l_cons( MEFFECTER, KCOM, l_cons( NIL, KFN, NIL ) ); }
	|	EFFECTER name HELP
		{ $$ = l_cons( MEFFECTER, KCOM, llist( $2, l_cons( NIL, KFN, NIL ) ) ); }
	;

effect:		EFFECT		{ $$ = l_cons( MEFFECT, KCOM, NIL ); }
	|	EFFECT calc	{ $$ = l_cons( MEFFECT, KCOM, $2 ); }
	;

list:	 	LIST		{ $$ = l_cons( MLIST, KCOM, NIL ); }
	|	LIST range	{ $$ = l_cons( MLIST, KCOM, $2 ); }
	|	LIST STACK	{ $$ = l_cons( MLIST, KCOM, l_cons( NIL, KCSTACK, NIL ) ); }
	|	LIST STACK range { $$ = l_cons( MLIST, KCOM, l_cons( NIL, KCSTACK, $3 ) ); }
	;

undo:		UNDO		{ $$ = l_cons( MUNDO, KCOM, NIL ); }
	;

push:		PUSH 		{ $$ = l_cons( MPUSH, KCOM, NIL ); }
	|	PUSH data	{ $$ = l_cons( MPUSH, KCOM, $2  ); }
	;

pop:		POP		{ $$ = l_cons( MPOP, KCOM, NIL ); }
	|	POP range	{ $$ = l_cons( MPOP, KCOM, $2 ); }
	;

put:		PUT 		{ $$ = l_cons( MPUT, KCOM, NIL ); }
	|	PUT data	{ $$ = l_cons( MPUT, KCOM, $2 ); }
	;

get:		GET		{ $$ = l_cons( MGET, KCOM, NIL ); }
	|	GET range	{ $$ = l_cons( MGET, KCOM, $2 ); }
	;

clear:		CLEAR		{ $$ = l_cons( MCLEAR, KCOM, NIL ); }
	|	CLEAR range	{ $$ = l_cons( MCLEAR, KCOM, $2 ); }
	|	CLEAR STACK	{ $$ = l_cons( MCLEAR, KCOM, l_cons( NIL, KCSTACK, NIL ) ); }
	|	CLEAR STACK range { $$ = l_cons( MCLEAR, KCOM, l_cons( NIL, KCSTACK, $3 ) ); }
	;

eval:		EVAL		{ $$ = l_cons( MEVAL, KCOM, NIL ); }
	|	EVAL range	{ $$ = l_cons( MEVAL, KCOM, $2 ); }
	;

copy:		COPY data data	{ $$ = l_cons( MCOPY, KCOM, llist( $2, $3 ) ); }
	|	COPY src data	{ $$ = l_cons( MCOPY, KCOM, llist( $2, $3 ) ); }
	|	COPY data dest	{ $$ = l_cons( MCOPY, KCOM, llist( $2, $3 ) ); }
	|	COPY src dest	{ $$ = l_cons( MCOPY, KCOM, llist( $2, $3 ) ); }
	;

range:		data  data	{ $$ = l_cons( $1, KRANGE, $2 ); }
	|	data '-'	{ $$ = l_cons( $1, KRANGE, l_cons( 999, KNUMBER, NIL ) ); }
	|	data 		{ $$ = l_cons( $1, KRANGE, NIL ); }
	;

src:		PREV		{ $$ = l_cons( KCPREV, KFN, NIL ); }
	|	AT		{ $$ = l_cons( KCAT, KFN, NIL ); }
	|	NEXT		{ $$ = l_cons( KCNEXT, KFN, NIL ); }
	;

dest:		PREV		{ $$ = l_cons( KCPREV, KFN, NIL ); }
	|	AT		{ $$ = l_cons( KCAT, KFN, NIL ); }
	|	NEXT		{ $$ = l_cons( KCNEXT, KFN, NIL ); }
	|	BEFORE		{ $$ = l_cons( KCBEFORE, KFN, NIL ); }
	|	AFTER		{ $$ = l_cons( KCAFTER, KFN, NIL ); }
	;


quit:		QUIT		{ $$ = l_cons( MQUIT, KCOM, NIL ); }
	;

level:		LEVEL data	{ $$ = l_cons( MLEVEL, KCOM, $2 ); }
	;

next:		NEXT		{ $$ = l_cons( MNEXT, KCOM, NIL ); }
	;

prev:		PREV		{ $$ = l_cons( MPREV, KCOM, NIL ); }
	;

goto:		GOTO calc	{ $$ = l_cons( MGOTO, KCOM, $2 ); }
	;

show:		SHOW calc	{ $$ = l_cons( MSHOW, KCOM, $2 ); }
	|	SHOW		{ $$ = l_cons( MSHOW, KCOM, NIL ); }
	;

include:	INCLUDE calc	{ $$ = l_cons( MINCLUDE, KCOM, $2 ); }
	;

help:		HELP		{ $$ = l_cons( MHELP, KCOM, NIL ); }
	;

purge:	        PURGE name	{ $$ = l_cons( MPURGE, KCOM, $2 ); }
	|	PURGE literal	{ $$ = l_cons( MPURGE, KCOM, $2 ); }
	;

print:		PRINT calc		{ $$ = l_cons( MPRINT, KCOM, $2 ); }
	;

query:		QUERY			{ $$ = l_cons( MQUERY, KCOM, NIL ); }
	;

gc:		GC			{ $$ = l_cons( MGC, KCOM, NIL ); }
	;

rehash:		REHASH			{ $$ = l_cons( MREHASH, KCOM, NIL ); }
	;

create:		CREATE calc		{ $$ = l_cons( MCREATE, KCOM, $2 ); }
	;

viewer:		VIEWER calc		{ $$ = l_cons( MVIEWER, KCOM, $2 ); }
	|	VIEWER			{ $$ = l_cons( MVIEWER, KCOM, NIL ); }
	;

del:		DEL name		{ $$ = l_cons( MDEL, KCOM, $2 ); }
	|	DEL sharp		{ $$ = l_cons( MDEL, KCOM, $2 ); }
	;

reset:		RESET			{ $$ = l_cons( MRESET, KCOM, NIL ); }
	;

expression:	name '=' calc	{ $$ = l_cons( $1, KLET, $3 ); }
	|	name LETPLUS calc	{ $$ = l_cons( $1, KLET, l_cons( $1, KPLUS, $3 ) ); }
	|	name LETMINUS calc	{ $$ = l_cons( $1, KLET, l_cons( $1, KMINUS, $3 ) ); }
	|	name LETMULT calc	{ $$ = l_cons( $1, KLET, l_cons( $1, KMULT, $3 ) ); }
	|	name LETDIVIDE calc	{ $$ = l_cons( $1, KLET, l_cons( $1, KDIVIDE, $3 ) ); }
	|	name LETAND calc	{ $$ = l_cons( $1, KLET, l_cons( $1, KDAND, $3 ) ); }
	|	name LETOR calc		{ $$ = l_cons( $1, KLET, l_cons( $1, KDOR, $3 ) ); }
	|	calc EQ calc	{ $$ = l_cons( $1, KEQ, $3 ); }
	|	calc NE calc	{ $$ = l_cons( $1, KNE, $3 ); }
	|	calc GE calc	{ $$ = l_cons( $1, KGE, $3 ); }
	|	calc LE calc	{ $$ = l_cons( $1, KLE, $3 ); }
	|	calc '<' calc	{ $$ = l_cons( $1, KLT, $3 ); }
	|	calc '>' calc	{ $$ = l_cons( $1, KGT, $3 ); }
	|	calc		{ $$ = $1; }
	;


poslist:	position		{ $$ = $1; }
	|	position poslist	{ $$ = l_cons( $1, KPOSLIST, $2 ); }
	;

position:	'(' calc ',' calc ')'	{ $$ = l_cons( $2, KPOSITION, $4 ); }
	;


calclist:	calc			{ $$ = l_cons( $1, KLIST, NIL ); }
	|	calc ',' calclist	{ $$ = l_cons( $1, KLIST, $3 ); }
	;

kakko2:		'(' calclist ')'	{ $$ = $2; }
	|	'(' ')'			{ $$ = NIL; }
	;

kakko:		'(' calc ')'		{ $$ = $2; }
	;

calc:		data			{ $$ = $1; }
	|	'{' calclist '}'	{ $$ = $2; }
	|	poslist			{ $$ = $1; }
	|	name kakko2		{ $$ = l_cons( $1, KFUNC, $2 ); }
	|	RETURN kakko		{ $$ = l_cons( $2, KRET, NIL ); }
	|	kakko			{ $$ = $1; }
	|	calc '+' calc		{ $$ = l_cons( $1, KPLUS, $3 ); }
	|	calc '-' calc		{ $$ = l_cons( $1, KMINUS, $3 ); }
	|	calc '*' calc		{ $$ = l_cons( $1, KMULT, $3 ); }
	|	calc '/' calc		{ $$ = l_cons( $1, KDIVIDE, $3 ); }
	|	calc '%' calc		{ $$ = l_cons( $1, KMOD, $3 ); }
	|	calc '&' calc		{ $$ = l_cons( $1, KDAND, $3 ); }
	|	calc '|' calc		{ $$ = l_cons( $1, KDOR, $3 ); }
	|	INC name	{ $$ = l_cons( $2, KINCL, NIL ); }
	|	DEC name	{ $$ = l_cons( $2, KDECL, NIL ); }
	|	name INC	{ $$ = l_cons( $1, KINCR, NIL ); }
	|	name DEC	{ $$ = l_cons( $1, KDECR, NIL ); }
	|	'~' calc		{ $$ = l_cons( $2, KDNOT, NIL ); }
	;

data:		num		{ $$ = $1; }
	|	literal		{ $$ = $1; }
	|	name		{ $$ = $1; }
	|	sharp		{ $$ = $1; }
	|	nil		{ $$ = $1; }
	;

num:		NUM		{ $$ = l_cons( $1, KNUMBER, NIL ); }
	|       '-' NUM %prec UMINUS { $$ = l_cons( $2*(-1), KNUMBER, NIL ); }
	;

literal:	LITERAL		{ $$ = l_cons( $1, KLITERAL, NIL ); }
	;

name:		NAME		{ $$ = l_cons( $1, KSYMBOL, NIL ); }
	;

sharp:		SHARP		{ $$ = l_cons( $1, KSHARP, NIL ); }
	;

nil:		SNIL		{ $$ = l_cons( NIL, KNIL, NIL ); }
	;


%%
/* B0 original
kakko:		'(' calc ')'		{ $$ = $2; }
	|	'(' calclist ')'	{ $$ = $2; }
	|	'(' ')'			{ $$ = NIL; }
	;
add kakko2
*/
