/**********************************************************************/
/*                                                                    */
/*        hash.c -- gedo hash table operation                         */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*         1.17,1997 B0 func_funcs add function                       */
/*         5.13,1997 B2 change GLOBAL var masked by TMP var           */
/*         6. 3,1997 B2 add smart_set_global_literal()                */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

#define HASHSIZE 100		/* Hash Table Size */

static Lisp Hash[3][HASHSIZE];	/* Hash Table of LOCAL, GLOBAL, TMP */

struct HashStruct {
	struct List *lst;	/* Hash Table */
	char * (*getst)(Lisp);	/* get string function */
};

struct HashStruct2 {
	struct List **lst;	/* Hash Table */
	char *(*getst)(Lisp);	/* get string function */
	Lisp (*mklist)(char *,Lisp);	/* cons function */
};

/**********************************************************************/
/*                                                                    */
/*             function clear_hash()                                  */
/*                                                                    */
/*             int kind: LOCAL, GLOBAL or TMP                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Clear the Hash Table.                        */
/*                                                                    */
/**********************************************************************/
void clear_hash( int kind )
{
	memset( Hash[kind], 0, sizeof(Hash[kind]) );
}


/**********************************************************************/
/*                                                                    */
/*             function get_hash()                                    */
/*                                                                    */
/*             char *str: string                                      */
/*             int len  : length of str                               */
/*                                                                    */
/*             return  int: Hash Value                                */
/*                                                                    */
/*             function: Return Hash Value.                           */
/*                                                                    */
/**********************************************************************/
int get_hash( char *str, int len )
{
	int num;

	num = str[0] + str[len-1] + str[len/2];
	return( (num - 0x60) % 100 );
}

/**********************************************************************/
/*                                                                    */
/*             function *get_tmp()                                    */
/*                                                                    */
/*             Lisp at: LITERAL List                                  */
/*                                                                    */
/*             return  char *: string                                 */
/*                                                                    */
/*             function: char *(getst)() pointer for TMP Hash.        */
/*                                                                    */
/**********************************************************************/
static char *get_tmp( Lisp at )
{
	return( &TmpString[List[at].v] );
}

/**********************************************************************/
/*                                                                    */
/*             function *get_local()                                  */
/*                                                                    */
/*             Lisp at: LITERAL List                                  */
/*                                                                    */
/*             return  char *: string                                 */
/*                                                                    */
/*             function: char *(getst)() pointer for LOCAL Hash.      */
/*                                                                    */
/**********************************************************************/
static char *get_local( Lisp at )
{
	return( &LocalString[LocalList[at].v] );
}

/**********************************************************************/
/*                                                                    */
/*             function *get_global()                                 */
/*                                                                    */
/*             Lisp at: LITERAL List                                  */
/*                                                                    */
/*             return  char *: string                                 */
/*                                                                    */
/*             function: cahr *(getst)() pointer for GLOBAL Hash.     */
/*                                                                    */
/**********************************************************************/
static char *get_global( Lisp at )
{
	return( &GlobalString[GlobalList[at].v] );
}

/**********************************************************************/
/*                                                                    */
/*             function search_hash2()                                */
/*                                                                    */
/*             struct HashStruct *hs: Abstructed Hash Infomation      */
/*             Lisp now: start position:(fist Hash Value)             */
/*             char *key : search Key                                 */
/*                                                                    */
/*             return  Lisp: ==NIL: not found  else found Value       */
/*                                                                    */
/*             function: Search the Hash by Key.(real body)           */
/*                                                                    */
/**********************************************************************/
static Lisp search_hash2( struct HashStruct *hs, Lisp now, char *key )
{
	Lisp at;

	for( ; now != NIL; now = hs->lst[now].next ){
		if( hs->lst[now].kind != KLIST ){
fprintf( stderr, "ill1\n" );
			warning( W_ILLHASH );
			return( NIL );
		}
		at = hs->lst[now].v;
		if( at == NIL || hs->lst[at].kind != KHASH ){
fprintf( stderr, "ill2\n" );
			warning( W_ILLHASH );
			return( NIL );
		}
		if( strcmp( key, (*hs->getst)(at) ) == 0 ){
			return( hs->lst[at].next );
		}
	}
	return( NIL );
}

/**********************************************************************/
/*                                                                    */
/*             function search_hash()                                 */
/*                                                                    */
/*             int flag: LOCAL, GLOBAL or TMP                         */
/*             char *key: search Key                                  */
/*                                                                    */
/*             return  Lisp: found Value or NIL(not found)            */
/*                                                                    */
/*             function: Search the Hash front end.                   */
/*                                                                    */
/**********************************************************************/
Lisp search_hash( int flag, char *key )
{
	int len;
	int at;
	struct HashStruct hs;

	len = strlen( key );
	at = get_hash( key, len );
	if( flag == TMP ){
		hs.lst = List;
		hs.getst = get_tmp;
	} else if( flag == LOCAL ){
		hs.lst = LocalList;
		hs.getst = get_local;
	} else if( flag == GLOBAL ){
		hs.lst = GlobalList;
		hs.getst = get_global;
	} else {
fprintf( stderr, "ill3\n" );
		warning( W_ILLHASH );
		return( NIL );
	}
	return( search_hash2( &hs, Hash[flag][at], key ) );
}


/**********************************************************************/
/*                                                                    */
/*             function set_hash2()                                   */
/*                                                                    */
/*             struct HashStruct2 *hs: Abstruct Hash Infomation       */
/*             char *key : search Key                                 */
/*             Lisp now  : start position(first Hash Value)           */
/*             Lisp data : set Value                                  */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Set the Value(data) to the Key(key).         */
/*                                                                    */
/**********************************************************************/
void set_hash2( struct HashStruct2 *hs, char *key, Lisp now, Lisp data )
{
	Lisp at;
	volatile Lisp tmp;

	at = (*(hs->lst))[now].v;
	if( (*(hs->lst))[now].kind != KLIST ||
	   (*(hs->lst))[at].kind != KHASH ){
fprintf( stderr, "ill4 now=%d at=%d\n", (*(hs->lst))[now].kind,(*(hs->lst))[at].kind );
		warning( W_ILLHASH );
		return;
	}
	if( strcmp( key, (*hs->getst)(at) ) == 0 ){
		(*(hs->lst))[(*(hs->lst))[now].v].next = data;
		return;
	} else {
		if( (*(hs->lst))[now].next == NIL ){
			tmp = (*hs->mklist)(key, data);
			(*(hs->lst))[now].next = tmp;
			return;
		}
		for( now = (*(hs->lst))[now].next; now; now = (*(hs->lst))[now].next ){
			at = (*(hs->lst))[now].v;
			if( (*(hs->lst))[now].kind != KLIST ||
				(*(hs->lst))[at].kind != KHASH ){
fprintf( stderr, "ill5\n" );
				warning( W_ILLHASH );
				return;
			}
			if( strcmp( key, (*hs->getst)(at) ) == 0 ){
				(*(hs->lst))[at].next = data;
				return;
			}
			if( (*(hs->lst))[now].next == NIL ){
				tmp = (*hs->mklist)( key, data );
				(*(hs->lst))[now].next = tmp; 
				return;
			}
		}	
fprintf( stderr, "ill6\n" );
		warning( W_ILLHASH );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function mklist_tmp()                                  */
/*             function mklist_local()                                */
/*             function mklist_global()                               */
/*                                                                    */
/*             char *key: Hash Key                                    */
/*             Lisp data: Hash Value                                  */
/*                                                                    */
/*             return  Lisp: pair of Key and Value                    */
/*                                                                    */
/*             function: Make each Storage Hash Data.                 */
/*                       Lisp (*mklist)() pointer                     */
/**********************************************************************/
Lisp mklist_tmp( char *key, Lisp data )
{
	return( cons( cons( tmp_strdup(key), KHASH, data ), KLIST, NIL ) );
}
Lisp mklist_local( char *key, Lisp data )
{
	return( l_cons( l_cons( local_strdup(key), KHASH, data ), KLIST, NIL ) );
}
Lisp mklist_global( char *key, Lisp data )
{
	return( g_cons( g_cons( global_strdup(key), KHASH, data ), KLIST, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function set_hash()                                    */
/*                                                                    */
/*             int flag: LOCAL, GLOBAL or TMP                         */
/*             char *key: Hash key                                    */
/*             Lisp data: Hash value                                  */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Set the each Storage Hash.                   */
/*                                                                    */
/**********************************************************************/
void set_hash( int flag, char *key, Lisp data )
{
	int len;
	int at;
	struct HashStruct2 hs;

	len = strlen( key );
	at = get_hash( key, len );

	if( flag == LOCAL ){
		hs.lst = &LocalList;
		hs.getst = get_local;
		hs.mklist = mklist_local;
	} else if( flag == GLOBAL ){
		hs.lst = &GlobalList;
		hs.getst = get_global;
		hs.mklist = mklist_global;
	} else if( flag == TMP ){
		hs.lst = &List;
		hs.getst = get_tmp;
		hs.mklist = mklist_tmp;
	} else {	
fprintf( stderr, "ill7\n" );
                warning( W_ILLHASH );
		return;
	}
	if( Hash[flag][at] == NIL ){
		Hash[flag][at] = (*hs.mklist)( key, data );
		return;
	}
	set_hash2( &hs, key, Hash[flag][at], data );
}

#if 0
/**********************************************************************/
/*                                                                    */
/*             function delete_hash2()                                */
/*                                                                    */
/*             struct HashStruct2 *hs                                   */
/*             char *key                                              */
/*             Lisp now                                               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function                                               */
/*                                                                    */
/**********************************************************************/
void delete_hash2( struct HashStruct2 *hs, char *key, Lisp now )
{
	Lisp at;
	volatile Lisp tmp;

	at = (*(hs->lst))[now].v;
	if( (*(hs->lst))[now].kind != KLIST ||
	   (*(hs->lst))[at].kind != KHASH ){
fprintf( stderr, "ill8\n" );
		warning( W_ILLHASH );
		return;
	}
	if( strcmp( key, (*hs->getst)(at) ) == 0 ){
		(*(hs->lst))[(*(hs->lst))[now].v].next = data;
		return;
	} else {
		if( (*(hs->lst))[now].next == NIL ){
			tmp = (*hs->mklist)(key, data);
			(*(hs->lst))[now].next = tmp;
			return;
		}
		for( now = (*(hs->lst))[now].next; now; now = (*(hs->lst))[now].next ){
			at = (*(hs->lst))[now].v;
			if( (*(hs->lst))[now].kind != KLIST ||
				(*(hs->lst))[at].kind != KHASH ){
fprintf( stderr, "ill9\n" );
				warning( W_ILLHASH );
				return;
			}
			if( strcmp( key, (*hs->getst)(at) ) == 0 ){
				(*(hs->lst))[at].next = data;
				return;
			}
			if( (*(hs->lst))[now].next == NIL ){
				tmp = (*hs->mklist)( key, data );
				(*(hs->lst))[now].next = tmp; 
				return;
			}
		}
fprintf( stderr, "ill10\n" );
		warning( W_ILLHASH );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function delete_hash()                                 */
/*                                                                    */
/*             int flag                                               */
/*             char *key                                              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function                                               */
/*                                                                    */
/**********************************************************************/
void delete_hash( int flag, char *key )
{
	int len;
	int at;
	struct HashStruct2 hs;

	len = strlen( key );
	at = get_hash( key, len );

	if( flag == LOCAL ){
		hs.lst = &LocalList;
		hs.getst = get_local;
	} else if( flag == GLOBAL ){
		hs.lst = &GlobalList;
		hs.getst = get_global;
	} else if( flag == TMP ){
		hs.lst = &List;
		hs.getst = get_tmp;
	} else {
fprintf( stderr, "ill11\n" );
		warning( W_ILLHASH );
		return;
	}
	if( Hash[flag][at] == NIL ){
		return;
	}
	delete_hash2( &hs, key, Hash[flag][at] );
}
#endif

/**********************************************************************/
/*                                                                    */
/*             function get_mac()                                     */
/*                                                                    */
/*             String mac: index of string (.v of kind==KLITERAL List)*/
/*                                                                    */
/*             return  Lisp: found GLOBAL Hash Value or NIL(fail)     */
/*                           duplicated to TMP List data.             */
/*             function: Search Global Hash and duplicate to TMP data.*/
/*                       (internal use)                               */
/**********************************************************************/
Lisp get_mac( String mac )
{
	Lisp at;

	if( (at = search_hash( GLOBAL, ts(mac) )) ){
		return( dup_g2tLisp(at) );
	} else {
		return( NIL );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function expand_macro()                                */
/*                                                                    */
/*             String mac: index of string (.v of kind==KLITERAL List)*/
/*                                                                    */
/*             return  Lisp: found GLOBAL Hash Value or NIL(fail)     */
/*                           duplicated to TMP List data.             */
/*             function: Search Global Hash and duplicate to TMP data.*/
/*                       (frontend use)                               */
/**********************************************************************/
Lisp expand_macro( String mac )
{
	Lisp at;

	if( (at = search_hash( TMP, ts(mac) )) ){ /* add C1 */
		return( dup_t2tLisp(at) );
	}

	if( (at = search_hash( GLOBAL, ts(mac) )) ){
		return( dup_g2tLisp(at) );
	} else {
		if( check_value( GLOBAL, "#noerror" ) ){
			return( cons( NIL, KNIL, NIL ) );
		} else {
			warning( W_NOMACRO, ts(mac) );
			return( NIL );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function get_var()                                     */
/*                                                                    */
/*             String mac: index of string (.v of kind==KLITERAL List)*/
/*                                                                    */
/*             return  Lisp: found LOCAL or TMP Hash Value or NIL(fail)*/
/*                 if LOCAL, duplicated to TMP List data.             */
/*             function: Search LOCAL or TMP Hash and be TMP data.    */
/*                       (internal use)                               */
/**********************************************************************/
Lisp get_var( String mac )
{
	Lisp at;

	if( (at = search_hash( TMP, ts(mac) )) ){
		return( at );
	} else if( (at = search_hash( LOCAL, ts(mac) )) ){
		return( dup_l2tLisp(at) );
	} else {
		return( NIL );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function get_variable()                                */
/*                                                                    */
/*             String mac: index of string (.v of kind==KLITERAL List)*/
/*                                                                    */
/*             return  Lisp: found LOCAL or TMP Hash Value or NIL(fail)*/
/*                 if LOCAL, duplicated to TMP List data.             */
/*             function: Search LOCAL or TMP Hash and be TMP data.    */
/*                       (frontend use)                               */
/**********************************************************************/
Lisp get_variable( String mac )
{
	Lisp at;

	if( (at = get_var( mac )) == NIL ){
		if( check_value( GLOBAL, "#noerror" ) ){
			return( cons( NIL, KNIL, NIL ) );
		} else {
			warning( W_NOMACRO, ts(mac) );
			return( NIL );
		}
	} else {
		return( at );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function smart_set_number()                            */
/*                                                                    */
/*             char *name: Variable name                              */
/*             int val   : set value(INT)                             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Each TMP or LOCAL variable, change INT value.*/
/*                                                                    */
/**********************************************************************/
void smart_set_number( char *name, int val )
{
	Lisp at;

	if( (at = search_hash( TMP, name ))
	   && List[at].kind == KNUMBER ){
		List[at].v = val;
	} else if( (at = search_hash( LOCAL, name ))
	   && LocalList[at].kind == KNUMBER ){
		LocalList[at].v = val;
	} else {
		at = l_cons( val, KNUMBER, NIL );
		set_hash( LOCAL, name, at );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function smart_set_literal()                           */
/*                                                                    */
/*             char *name: variable name                              */
/*             char *val : set value(LITERAL)                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Each TMP or LOCAL variable, change LITERAL   */
/*                      cleverly.                                     */
/**********************************************************************/
void smart_set_literal( char *name, char *val )
{
	Lisp at;
	size_t len;

	if( (at = search_hash( TMP, name ))
	   && List[at].kind == KLITERAL ){
		len = strlen( tsv(at) );
		if( len >= strlen(val) ){
			strcpy( tsv(at), val );
		} else {
			List[at].v = tmp_strdup( val );
		}
	} else if( (at = search_hash( LOCAL, name ))
	   && LocalList[at].kind == KLITERAL 
	   && LocalList[at].v != NIL ){ /* B2 add */
		len = strlen( lsv(at) );
		if( len >= strlen(val) ){
			strcpy( lsv(at), val );
		} else {
			LocalList[at].v = local_strdup( val );
		}
	} else {
		at = l_cons( local_strdup(val), KLITERAL, NIL );
		set_hash( LOCAL, name, at );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function smart_set_global_literal()                    */
/*                                                                    */
/*             char *name: variable name                              */
/*             char *val : set value(LITERAL)                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: For GLOBAL LITERAL variable, change LITERAL  */
/*                      cleverly.                                     */
/**********************************************************************/
/* B2 add */
void smart_set_global_literal( char *name, char *val )
{
	Lisp at;
	size_t len;

	if( (at = search_hash( GLOBAL, name ))
	   && GlobalList[at].kind == KLITERAL 
	   && GlobalList[at].v != NIL ){
		len = strlen( gsv(at) );
		if( len >= strlen(val) ){
			strcpy( gsv(at), val );
		} else {
			GlobalList[at].v = global_strdup( val );
		}
	} else {
		at = g_cons( global_strdup(val), KLITERAL, NIL );
		set_hash( GLOBAL, name, at );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function check_value()                                 */
/*                                                                    */
/*             int kind: LOCAL or GLOBAL                              */
/*             char *name: variable name                              */
/*                                                                    */
/*             return  int: value of the variable                     */
/*                                                                    */
/*             function: Only checking value.(use in Special Symbol   */
/*                      value check)                                  */
/*             rewrite B2                                             */
/**********************************************************************/
int check_value( int kind, char *name )
{
	Lisp at;

	if( (at = search_hash( TMP, name )) == NIL ){
		if( (at = search_hash( kind, name )) == NIL ){
			return( 0 );
		}
		if( kind == LOCAL ){
			return( LocalList[at].v );
		} else {
			return( GlobalList[at].v );
		}
	} else {
		return( List[at].v );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function search_global()                               */
/*                                                                    */
/*             char *name: variable name                              */
/*                                                                    */
/*             return  int: value of the variable                     */
/*                                                                    */
/*             function: get GLOBAL var from TMP or GLOBAL.           */
/*            B2 create                                               */
/**********************************************************************/
Lisp search_global( char *name )
{
	Lisp at;

	if( (at = search_hash( TMP, name )) == NIL ){
		if( (at = search_hash( GLOBAL, name )) == NIL ){
			return( NIL );
		}
		return( dup_g2tLisp(at) );
	} else {
		return( at );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function markup_hash()                                 */
/*                                                                    */
/*             int kind: LOCAL, GLOBAL or TMP                         */
/*             struct List *lp: Local, Global or Tmp LIST             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: For GC, markup the LIST used in Hash Table,  */
/*                      and exhibit Collect the used List.            */
/**********************************************************************/
void markup_hash( int kind, struct List *lp )
{
	int i;

	for( i = 0; i < HASHSIZE; i++ ){
		check_used( lp, Hash[kind][i] );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function search_names()                                */
/*                                                                    */
/*             Lisp now: each Hash Table                              */
/*             int kind: search kind == 0: function(builtin&user)     */
/*                    ==1:variable  ==2:sharp  ==3:builtin function   */
/*                    ==4:user function ==5:usr macro                 */
/*             Lisp ret: result chain                                 */
/*                                                                    */
/*             return  Lisp: ==NIL: no more in the Hash Value         */
/*                           else: name of the search object          */
/*             function: Search a name by kind.                       */
/*                                                                    */
/**********************************************************************/
Lisp search_names( Lisp now, int kind, Lisp ret )
{
	Lisp at;
	struct List *lst;

/* B0	if( kind == 0 ){  search_funcs */
/*B0*/	if( kind == 0 || kind == 3 || kind == 4 || kind == 5 ){ /* search_funcs */
		lst = GlobalList;
	} else if( kind == 1 ){ /* search vars */
		lst = LocalList;
	} else if( kind == 2 ){ /* search sharps */
		lst = GlobalList;
	} else {
		return( NIL );
	}
        if( now == NIL ){
		return( ret );
	}
	if( lst[now].kind != KLIST ){
		return( ret );
	}
	for( ; now != NIL; now = lst[now].next ){
		if( lst[now].kind != KLIST ){
fprintf( stderr, "ill12\n" );
			warning( W_ILLHASH );
			return( NIL );
		}
		at = lst[now].v;
		if( at == NIL || lst[at].kind != KHASH ){
fprintf( stderr, "ill13\n" );
			warning( W_ILLHASH );
			return( NIL );
		}
		
		if( kind == 0 ){
			if( lst[lst[at].next].kind == KARGS 
			   || lst[lst[at].next].kind == KCFUNC
			   || lst[lst[at].next].kind == KMARGS ){
				ret = cons( cons( tmp_strdup(gsv(at)), KLITERAL, NIL ), KLIST, ret );
			}
		} else if( kind == 1 ){
			ret = cons( cons( tmp_strdup(lsv(at)), KLITERAL, NIL ), KLIST, ret );
		} else if( kind == 2 ){
			if( *gsv(at) == '#' ){
				ret = cons( cons( tmp_strdup(gsv(at)), KLITERAL, NIL ), KLIST, ret );
			}
/* add B0 */
		} else if( kind == 3 ){
			if( lst[lst[at].next].kind == KCFUNC ){
				ret = cons( cons( tmp_strdup(gsv(at)), KLITERAL, NIL ), KLIST, ret );
			}
		} else if( kind == 4 ){
			if( lst[lst[at].next].kind == KARGS ){
				ret = cons( cons( tmp_strdup(gsv(at)), KLITERAL, NIL ), KLIST, ret );
			}
		} else if( kind == 5 ){
			if( lst[lst[at].next].kind == KMARGS ){
				ret = cons( cons( tmp_strdup(gsv(at)), KLITERAL, NIL ), KLIST, ret );
			}
/* end B0 */
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function search_vars()                                 */
/*                                                                    */
/*             int flag: search kind == 0: function(builtin&user)     */
/*                    ==1:variable  ==2:sharp  ==3:builtin function   */
/*                    ==4:user function  ==5:usr macro                */
/*                                                                    */
/*             return  Lisp: LIST{LITERAL,...}                        */
/*                                                                    */
/*             function: Search the name of flag and return the LIST. */
/*                                                                    */
/**********************************************************************/
Lisp search_vars( int flag )
{
	int i;
	Lisp ret;

	ret = NIL;
	if( flag == 0 ){ /* search_funcs builtin & usrfunc & macro */
		for( i = 0; i < HASHSIZE; i++ ){
			ret = search_names( Hash[GLOBAL][i], flag, ret );
		}
	} else if( flag == 1 ){ /* search vars */
		for( i = 0; i < HASHSIZE; i++ ){
			ret = search_names( Hash[LOCAL][i], flag, ret );
		}
	} else if( flag == 2 ){ /* search sharps */
		for( i = 0; i < HASHSIZE; i++ ){
			ret = search_names( Hash[GLOBAL][i], flag, ret );
		}
/* add B0 */
	} else if( flag == 3 ){ /* search_funcs builtin only */
		for( i = 0; i < HASHSIZE; i++ ){
			ret = search_names( Hash[GLOBAL][i], flag, ret );
		}
	} else if( flag == 4 ){ /* search_funcs usr func only */
		for( i = 0; i < HASHSIZE; i++ ){
			ret = search_names( Hash[GLOBAL][i], flag, ret );
		}
	} else if( flag == 5 ){ /* search_funcs macro only */
		for( i = 0; i < HASHSIZE; i++ ){
			ret = search_names( Hash[GLOBAL][i], flag, ret );
		}
/* end B0 */
	} else {
		return( NIL );
	}
	return( ret );
}

