/**********************************************************************/
/*                                                                    */
/*        init.c -- gedo initialize routines                          */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*         B2 : add initilize symbol #viewer at set_initial_data()    */
/*         B2 : careless miss. cannot read -outfile at init()         */
/*         B2 : re-write init() for execution level2 moved to main()  */
/*         B2 : deleted level2_proc() not necessary.                  */
/*         C1 : set_initoption(), endp()  change directory if write   */
/*                 permittion is protected.                           */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <setjmp.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

#define ARGTMP		"argtmp.$$$"
char ArgTmp[128] = { ARGTMP };
extern FILE *yyout;		/* stdout defined at lex scanner */
extern FILE *yyin;		/* stdin defined at lex scanner */

/* argment code */
#define INIT    0
#define SERIALS	1
#define SERIALE	2
#define LEVEL   3
#define LDFF    4
#define SVFF    5
#define OUTFILE 6
#define SIZE	7

static char *Args[8];		/* each argment */
static char *InitStr[6] = { 
	"SerialStart", "SerialEnd", "level", "LoadFileForm", "SaveFileForm", 
	"OutFile" };

static int InitCount = 1;	/* number of initilize file */

extern jmp_buf rewindjmp;	/* rewind the input at command goto, next, prev */ 

/**********************************************************************/
/*                                                                    */
/*             function set_flag()                                    */
/*                                                                    */
/*             char **argv : argments for gedo                        */
/*                                                                    */
/*             return  int : processed number of argments             */
/*                          -1==abort the process.                    */
/*             function  : set command-line argment to the struct.    */
/*                                                                    */
/**********************************************************************/
int set_flag( char **argv )
{
	char *p;
	int ret = -1;

	p = *argv;
	if( strcmp( p, "-init" ) == 0 ){
		if( InitCount < INITFILEMAX ){
			InitFile[InitCount++] = *(argv+1);
		} else {
			display( M_MANYARG, "-init" );
		}
		ret = 2;
	} else if( strcmp( p, "-loadfile" ) == 0 ){
		Args[LDFF] = *(argv+1);
		ret = 2;
	} else if( strcmp( p, "-savefile" ) == 0 ){
		Args[SVFF] = *(argv+1);
		ret = 2;
	} else if( strcmp( p, "-serials" ) == 0 ){
		Args[SERIALS] = *(argv+1);
		ret = 2;
	} else if( strcmp( p, "-seriale" ) == 0 ){
		Args[SERIALE] = *(argv+1);
		ret = 2;
	} else if( strcmp( p, "-level" ) == 0 ){
		Args[LEVEL] = *(argv+1);
		ret = 2;
	} else if( strcmp( p, "-outfile" ) == 0 ){
		Args[OUTFILE] = *(argv+1);
		ret = 2;
	} else if( strcmp( p, "-help" ) == 0 ){
		ret = -1;
	} else if( strcmp( p, "-version" ) == 0 ){
		display( M_OMNI, CopyRight );
		ret = -1;
	} else if( strcmp( p, "-size" ) == 0 ){
		Args[SIZE] = *(argv+1);
		ret = 2;
	} else if( strcmp( p, "-maxsize" ) == 0 ){
		Args[SIZE] = "max";
		ret = 1;
	} else {
		display( M_ILLARG, p );
		ret = -1;
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function set_initoption()                              */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : 0==success -1==fail                      */
/*                                                                    */
/*             function : Flush the command-line argment for gedo     */
/*                       into the virtual init file.                  */
/*             C1: change directory if write permittion is protected. */
/**********************************************************************/
int set_initoption( void )
{
	int i;
	FILE *fp;

	if( (fp = fopen( ArgTmp, "w" )) == NULL ){
/* add C1 */
	     Lisp lwp;
	     if( (lwp = search_global( "#lwp" )) != NIL ){
		  if( List[lwp].kind != KLIST || (lwp = List[lwp].next) == NIL
		      || List[lwp].kind != KLIST || (lwp = List[lwp].v) == NIL
		      || List[lwp].kind != KLITERAL ){
		       display( M_OPENTMP, "argment(not set #lwp)" );
		       return( -1 );
		  }
		  sprintf( Work, "%s/%s", tsv(lwp), ArgTmp );
		  if( (fp = fopen( Work, "w" )) == NULL ){
		       display( M_OPENTMP, "argment" );
		       return( -1 );
		  }
		  strcpy( ArgTmp, Work );
	     } else {
/* add end */
		  display( M_OPENTMP, "argment" );
		  return( -1 );
	     }
	}
	for( i = 1; i < 7; i++ ){
		if( Args[i] ){
			switch( i ){
			case SERIALS:
			case SERIALE:
				fprintf( fp, "#%s={%s}\n", InitStr[i-1], Args[i] );
				break;
			case LEVEL:
				fprintf( fp, "%s %s\n", InitStr[i-1], Args[i] );
				break;
			case LDFF:
			case SVFF:
			case OUTFILE:
				fprintf( fp, "#%s={\"%s\"}\n", InitStr[i-1], Args[i] );
				break;
			}
		}
	}
	fclose( fp );
        return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function *get_name()                                   */
/*                                                                    */
/*             char *str : string in init file "file ***"             */
/*                                                                    */
/*             return  char * : filename only                         */
/*                                                                    */
/*             function  : Get filename only.                         */
/*                                                                    */
/**********************************************************************/
char *get_name( char *str )
{
	char *ret, *p;

	ret = str;
	while( *str ){
		if( *str == PATHCHAR ){
			ret = str + 1;
		}
		str++;
	}
	for( p = ret; *p; p++ ){
		if( *p == '\"' || *p == ' ' || *p == '\n' ){
			*p = '\0';
			break;
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function make_cmdname()                                */
/*                                                                    */
/*             int flag   : 0==extension must be ".$$$"               */
/*                        : 1==extension must be ".\$\$\$" for shell  */
/*             char *name : base name                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Create cmdfile name from filename.          */
/*                                                                    */
/**********************************************************************/
void make_cmdname( int flag, char *name )
{
	char *point, *add;

	point = NULL;
	while( *name ){
		if( *name == '.' ){
			point = name;
		}
		name++;
	}
	if( flag == 0 ){
		add = CMDEXT0;
	} else {
		add = CMDEXT1;
	}
	if( point == NULL ){
		strcat( name, add );
	} else {
		strcpy( point, add );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function cutout_outfile()                              */
/*                                                                    */
/*             char *file : output file                               */
/*                                                                    */
/*             return  int : 0==success  -1==error                    */
/*                                                                    */
/*             function : Cut-out the output file into cmd files.     */
/*                                                                    */
/**********************************************************************/
int cutout_outfile( char *file )
{
	FILE *fpin, *fpout;
	char *p;
	char filename[FILENAME_MAX];

        fpout = NULL;
	if( (fpin = fopen( file, "rt" )) == NULL ){
		display( M_OPENOUT, file );
		return( -1 );
	}
	while( fgets( Work, sizeof(Work)-1, fpin ) ){
		for( p = Work; *p == ' ' || *p == '\t'; p++ );
		if( *p == '\n' || *p == ';' ){
			continue;
		}
		if( strncmp( p, "file", 4 ) == 0 ){
			for( p += 4; *p == ' ' || *p == '\t' || *p == '\"'; p++ );
			p = get_name( p );
			if( fpout ){
				fclose( fpout );
			}
			strcpy( filename, p );
			make_cmdname( 0, filename );
			if( (fpout = fopen( filename, "wt" )) == NULL ){
				display( M_OPENTMP, p );
				fclose( fpin );
				fclose( fpout );
				return( -1 );
			}
		} else {
			fputs( Work, fpout );
		}
	}
	fclose( fpout );
	fclose( fpin );
        return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function filelength()                                  */
/*                                                                    */
/*             int fd  : file descripter                              */
/*                                                                    */
/*             return  long : length of the fd file                   */
/*                                                                    */
/*             function : Get the file-length of fd.                  */
/*                                                                    */
/**********************************************************************/
long filelength( int fd )
{
struct stat buf;
        fstat( fd, &buf );
        return( buf.st_size );
}

/**********************************************************************/
/*                                                                    */
/*             function load_init()                                   */
/*                                                                    */
/*             char *file : init file name or cmd file name           */
/*                                                                    */
/*             return  int : 0==success  -1==fail                     */
/*                                                                    */
/*             function : Automatically read the init file or cmd file*/
/*                       and execute the content.                     */
/**********************************************************************/
int load_init( char *file )
{
	char *p;
	char filename[FILENAME_MAX];
	FILE *fp;
	int save_file_flag;

	if( Init_flag == 0 ){
		p = get_name( file );
		strcpy( filename, p );
		make_cmdname( 0, filename );
	} else {
		strcpy( filename, file );
	}
	if( access( filename, 0 ) == 0 ){
		if( (fp = fopen( filename, "rt" )) == NULL ){
			display( M_OPENTMP, filename );
			return( -1 );
		}
		if( filelength( fileno(fp) ) == 0 ){
			return( 0 );
		}
		display( M_LOAD, filename );
		yyin = fp; save_file_flag = File_flag;
		File_flag = 1;
		yyrestart( yyin );
		while( yyparse() == 0 );
		File_flag = save_file_flag;
		fclose( fp );
		display( M_FILEND, filename );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function usage()                                       */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Display usage.                              */
/*                                                                    */
/**********************************************************************/
void usage( void )
{
	fprintf( stderr, 
			"options: -init\n"
			"         -name\n"
			"         -loadtype\n"
			"         -savetype\n"
			"         -type\n"
			"         -serial\n"
			"         -level\n"
                        "         -outfile\n"
		        "         -size\n"
		        "         -maxsize\n"
			"         -help\n"
			"         -version\n" );
}

/* B2 int level2_proc( void ) deleted */

/**********************************************************************/
/*                                                                    */
/*             function get_MesID()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : IPC message ID                           */
/*                                                                    */
/*             function : Get the IPC message ID from env or default. */
/*                                                                    */
/**********************************************************************/
int get_MesID( void )
{
	char *env;
	int id, ret;

	if( (env = getenv( "GEDO_IPCMES" )) == NULL ){
		id = IPCMSGID;
	} else {
		id = atoi( env );
		if( id <= 0 ){
			id = IPCMSGID;
		}
	}
        if( (ret = msgget( id, IPC_CREAT | 0666 )) == -1 ){
		return( -1 );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function set_initial_data()                            */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Set argment defintion to the variables.     */
/*            B2: add symbol #viewer                                  */
/**********************************************************************/
void set_initial_data( void )
{
	Lisp at;

	if( (at = search_hash( GLOBAL, "#LoadFileForm" )) != NIL ){
		if( GlobalList[at].kind == KLITERAL ){
			strcpy( LoadFileForm, gsv(at) );
		}
	}
	if( (at = search_hash( GLOBAL, "#SaveFileForm" )) != NIL ){
		if( GlobalList[at].kind == KLITERAL ){
			strcpy( SaveFileForm, gsv(at) );
		}
	}
	if( (at = search_hash( GLOBAL, "#OutFile" )) != NIL ){
		if( GlobalList[at].kind == KLITERAL ){
			strcpy( OutFile, gsv(at) );
		}
	}
	if( (at = search_hash( GLOBAL, "#SerialStart" )) != NIL ){
		if( GlobalList[at].kind == KNUMBER ){
			SerialStart = GlobalList[at].v;
		}
	}
	if( (at = search_hash( GLOBAL, "#SerialEnd" )) != NIL ){
		if( GlobalList[at].kind == KNUMBER ){
			SerialEnd = GlobalList[at].v;
		}
	}
/* B2 add */
	set_hash( GLOBAL, "#viewer", g_cons( NIL, KLITERAL, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function init()                                        */
/*                                                                    */
/*             int argc    : command line argc                        */
/*             char **argv : command line argv                        */
/*                                                                    */
/*             return  int : 0==success  -1=fail                      */
/*                                                                    */
/*             function : Main routine at the init process.           */
/*           B2: careless miss. cannot read -outfile                  */
/**********************************************************************/
int init( int argc, char **argv )
{
	int ret;
	int i;
	char filename[FILENAME_MAX];
	String *tmp;

	Init_flag = 1;
	argc--; argv++;
	while( argc > 0 &&  **argv == '-' ){
		ret = set_flag( argv );
		if( ret == -1 ){
			usage();
			return( -1 );
		} else if( ret == 0 ){
			break;
		} else {
			argv += ret;
			argc -= ret;
		}
	}
	InitFile[InitCount] = NULL;

	init_Line();
	init_List();
	init_Childs();
	regist_cfunc();
	init_calcstack();

	ProcessId = getpid();
        if( (MesId = get_MesID()) == -1 ){
		fatal( F_IPCMSG );
		return( -1 );
	}
        signal( SIGCLD, SIG_IGN );
        signal( SIGFPE, fpe_error );
	init_raw_tty();
	signal( SIGINT, ctrl_c );
	if( (NowShmid = get_shmid( Args[SIZE] )) == -1 ){
		return( -1 );
	}
	if( (FileName = (String *)malloc( sizeof(String) * FILEMAX )) == NULL ){
		fatal( F_MALLOC, "FileName" );
		return( -1 );
	}
	FileMax = FILEMAX;
	FileNum = 0;

/* moved the position C3 */
	init_viewer();

	if( load_init( GLOBALINITFILE ) == -1 ){
		return( -1 );
	}
	NowLine = 0;

	sprintf( Work, "%s/%s", getenv( "HOME" ), LOCALINITFILE );
	if( load_init( Work ) == -1 ){
		return( -1 );
	}
	NowLine = 0;

	getcwd( Work, FILENAME_MAX );
	if( strcmp( Work, getenv( "HOME" ) ) != 0 
		&& strcmp( Work, GEDOPATH ) != 0 ){
		if( load_init( LOCALINITFILE ) == -1 ){
			return( -1 );
		}
	}
	for( i = 1; i < InitCount; i++ ){
		if( load_init( InitFile[i] ) == -1 ){
			return( -1 );
		}
	}
	NowLine = 0;

	if( set_initoption() == -1 ){
		return( -1 );
	}
	if( load_init( ArgTmp ) == -1 ){
		return( -1 );
	}
	remove( ArgTmp );

	set_initial_data();

	set_NowChild();
/*	init_viewer();   C3 move before read *.ini file */
	clear_localLisp();
	NowLine = 0;

	if( argc > 0 && *argv ){
		while( argc-- > 0 ){
			if( FileNum + 1 >= FileMax ){
				if( (tmp = (String *)realloc( FileName, sizeof(String) * (FileMax + FILEMAX) )) == NULL ){
					fatal( F_MALLOC, "FileName" );
					return( -1 );
				}
				FileName = tmp;
				FileMax += FILEMAX;
			}
			FileName[FileNum++] =  global_strdup(*argv++);
		}
	}
        for( i = SerialStart; i <= SerialEnd; i++ ){
         	if( LoadFileForm[0] != '\0' ){
		        sprintf( Work, LoadFileForm, i );
		} else {
			sprintf( Work, "def%d", i );
		}
		if( FileNum + 1 >= FileMax ){
			if( (tmp = (String *)realloc( FileName, sizeof(String) * (FileMax + FILEMAX) )) == NULL ){
				fatal( F_MALLOC, "FileName" );
				return( -1 );
			}
			FileName = tmp;
			FileMax += FILEMAX;
		}
		FileName[FileNum++] =  global_strdup(Work);
	}
        if( SaveFileForm[0] == '\0' && LoadFileForm[0] != '\0' ){
		strcpy( SaveFileForm, LoadFileForm );
		strcat( SaveFileForm, ".out" );
	}
	Init_flag = 0;
        for( i = 0; i < FileNum; i++ ){
		strcpy( filename, gs(FileName[i]) );
		make_cmdname( 0, filename );
		remove( filename );
	}
	if( access( OutFile, 0 ) == 0 ){
		cutout_outfile( OutFile );
	}
	TotalFileNum = FileNum;
	FileNum = 0;
	NowLine = 0;
	init_ring();
	return( 0 );
 }

/**********************************************************************/
/*                                                                    */
/*             function check_save_outfile()                          */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : 'Y'==save and quit                       */
/*                           'N'==not save and quit                   */
/*                            0 ==abort and continue                  */
/*             function : Confirm save or quit to user.               */
/*                        Called from do_quit() in funcs.c.           */
/**********************************************************************/
int check_save_outfile( void )
{
	int c;

	end_raw_tty();
	if( OutFile[0] == '\0' ){
		strcpy( OutFile, "outfile.out" );
	}
	display( M_SAVEOUT, OutFile );
	c = getchar();
	c = toupper( c );
	if( c != 'Y' && c != 'N' ){
		reset_raw_tty();
		return( 0 );
	} else {
		return( c );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function check_childPID()                              */
/*                                                                    */
/*             pid_t pid :  forked child process PID                  */
/*                          if pid == 0, kill all childs              */
/*             return NONE                                            */
/*                                                                    */
/*             function : check the newly forked child process PID    */
/*                     for killing at endp()                          */
/*             D2 : created                                           */
/**********************************************************************/
void check_childPID( pid_t pid )
{
     static pid_t *pids = NULL;
     static int  alloced = 0;
     static int  now = 0;

     if( (int)pid == 0 ){
	  int i;
	  for( i = 0; i < now; i++ ){
	       kill( pids[i], SIGKILL );
	  }
     } else if( (int)pid < 0 ){
	  return;
     } else {
	  if( now + 1 >= alloced ){
	       void *tmp;
	       tmp = realloc( pids, sizeof(pid_t) * (alloced + 10) );
	       if( tmp == NULL ){
		    return;
	       }
	       pids = tmp; alloced += 10;
	  }
	  pids[now++] = pid;
     }
}

/**********************************************************************/
/*                                                                    */
/*             function endp()                                        */
/*                                                                    */
/*             int c :  return of check_save_outfile()                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Main routine at the End process.            */
/*             C1: change directory if write permittion is protected. */
/**********************************************************************/
void endp( int c )
{
	FILE *fpin, *fpout;
	int i;
	char filename[FILENAME_MAX];
	Lisp lwp;

	fpout = NULL; /* dummy, to trick cc -W */
	kill_child();
        msgctl( MesId, IPC_RMID, NULL );
	lwp = NIL;
        if( c == 'Y' ){
		if( (fpout = fopen( OutFile, "wt" )) == NULL ){
/* add C1 */
		     if( (lwp = search_global( "#lwp" )) != NIL ){
			  if( List[lwp].kind != KLIST 
			      || (lwp = List[lwp].next) == NIL
			      || List[lwp].kind != KLIST 
			      || (lwp = List[lwp].v) == NIL
			      || List[lwp].kind != KLITERAL ){
			       display( M_OPENOUT, "OutFile(not set #lwp)" );
			       return;
			  }
			  sprintf( Work, "%s/%s", tsv(lwp), OutFile );
			  if( (fpout = fopen( Work, "w" )) == NULL ){
			       display( M_OPENOUT, OutFile );
			       return;
			  }
			  strcpy( OutFile, Work );
		     } else {
/* add end */

			display( M_OPENOUT, OutFile );
			return;
		     }
		}
	}
        for( i = 0; i < TotalFileNum; i++ ){
		strcpy( filename, gs(FileName[i]) );
		make_cmdname( 0, filename );
		if( c == 'Y' ){
/* add C1 */
			if( lwp != NIL ){
			     sprintf( Work, "%s/%s", tsv(lwp), filename );
			     strcpy( filename, Work );
			}
/* add end */	
			if( (fpin = fopen( filename, "rt" )) == NULL ){
				continue;
			}
			fprintf( fpout, "file \"%s\"\n", gs(FileName[i]) );
			while( fgets( Work, sizeof(Work)-1, fpin ) ){
				fputs( Work, fpout );
			}
			fclose( fpin );
		}
		remove( filename );
	}
        if( c == 'Y' ){
		fclose( fpout );
	}
	free( FileName );
	free_shmid();
	free_Childs();
	free_calcstack();
	free_Line();
	free_List();
/* D2 kill ALL child process */
	check_childPID( 0 );
/* end D2 */
}
