/**********************************************************************/
/*                                                                    */
/*           lisp.c -- gedo List struct basic DATA operation          */
/*                                                                    */
/*                by K.sugiura   Ver 1.0  for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*           11.22,1996 A0  memory BUG Fixed: have a copy of argment  */
/*                          in gedo_strdup                            */
/*            1.24,1997 B0  If TMP to TMP copy at SYMBOL, not copy String.*/
/*            1.24,1997 B0  chase_chain, chase_Lisp for query         */
/*            6.15,1997 B2  if realloc error, clean up TMP string&list*/
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <stdarg.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

#define LISTLOG 0

#if 0
#define LISTMAX		 64u
#define STRINGMAX	128u
#else
#define LISTMAX	 	1024u
#define STRINGMAX	2048u
#endif


struct ListStruct {		/* List Infomation struct */
	struct List **lisp;	/* pointer for the List, changed in realloc */
	unsigned int listnum;	/* Now used List number(maximum) */
	unsigned int listmax;	/* Now alloced List number */
	char **string;		/* pointer for the String buffer, changed in realloc */
	unsigned int stringnum;	/* Now used String number(maximum) */
	unsigned int stringmax;	/* Now alloced String number */
	unsigned int minimum;	/* KNOUSE first cell, fast search of blank */
	char *name;		/* name string like "TMP", "LOCAL", "GLOBAL" */
};

static struct ListStruct Tmp = { NULL, 1, 1, NULL, 1, 1, 1, "TMP" };
static struct ListStruct Local = { NULL, 1, 1, NULL, 1, 1, 1, "LOCAL" };
static struct ListStruct Global = { NULL, 1, 1, NULL, 1, 1, 1, "GLOBAL" };

struct dup_s {		/* dup_Lisp List Infomation */
	struct List **Lat;	/* List Infomation(changed by realloc) */
	Lisp (*Conat)(Lisp,int,Lisp);	/* cons, l_cons, g_cons */
	String (*Dupat)(char *);	/* tmp_strdup, local(global)_strdup */
	char * (*GetStr)(String);	/* get_[tlg]str */
	int Exp_flag;			/* used in t2tLisp, inhibit copy String */
	int Do_flag;			/* used in dup_doLisp, KVOID special */
};

static Lisp dup_Lisp( Lisp at, struct dup_s *ds );
#if LISTLOG
FILE *Tfp;
#endif

/**********************************************************************/
/*                                                                    */
/*             function init_List()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int: ==0:success   ==-1:fail                    */
/*                                                                    */
/*             function: intialize all Lists and String Buffer.       */
/*                                                                    */
/**********************************************************************/
int init_List( void )
{
	if( (List = (struct List *)malloc( sizeof(struct List) * LISTMAX )) == NULL ){
		fatal( F_MALLOC, "List" );
		return( -1 );
	}
	if( (LocalList = (struct List *)malloc( sizeof(struct List) * LISTMAX )) == NULL ){
		fatal( F_MALLOC, "Local List" );
		return( -1 );
	}
	if( (GlobalList = (struct List *)malloc( sizeof(struct List) * LISTMAX )) == NULL ){
		fatal( F_MALLOC, "Global List" );
		return( -1 );
	}
	Tmp.listmax = LISTMAX;
	Local.listmax = LISTMAX;
	Global.listmax = LISTMAX;

	if( (TmpString = (char *)malloc( sizeof(char) * STRINGMAX )) == NULL ){
		fatal( F_MALLOC, "Local String" );
		return( -1 );
	}
	if( (LocalString = (char *)malloc( sizeof(char) * STRINGMAX )) == NULL ){
		fatal( F_MALLOC, "Local String" );
		return( -1 );
	}
	if( (GlobalString = (char *)malloc( sizeof(char) * STRINGMAX )) == NULL ){
		fatal( F_MALLOC, "Global String" );
		return( -1 );
	}
	Tmp.stringmax = STRINGMAX;
	Local.stringmax = STRINGMAX;
	Global.stringmax = STRINGMAX;
#if LISTLOG
	if( (Tfp = fopen( "listlog", "w" )) == NULL ){
		printf( "listlog cannot open\n" );
	}
#endif
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function free_List()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: free all Lists and String Buffer.            */
/*                                                                    */
/**********************************************************************/
void free_List()
{
	free( List );
	free( LocalList );
	free( GlobalList );
	free( TmpString );
	free( LocalString );
	free( GlobalString );
#if LISTLOG
	fclose( Tfp );
#endif
}

/**********************************************************************/
/*                                                                    */
/*             function clear_localLisp()                             */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Erase the content of LocalList and String.   */
/*                                                                    */
/**********************************************************************/
void clear_localLisp( void )
{
	memset( List, 0, sizeof(List) * Local.listmax );
	Local.listnum = 1; Local.minimum = 1;
	memset( LocalString, 0, sizeof(char) * Local.stringmax );
	Local.stringnum = 1;
	clear_hash( LOCAL );
}

/**********************************************************************/
/*                                                                    */
/*             function check_used_tmp()                              */
/*                                                                    */
/*             int flag: ==0: set the previously used Cell number     */
/*                       !=0: display the previously use Cell number  */
/*             return NONE                                            */
/*                                                                    */
/*             function: query of the previously used Cell.           */
/*                                                                    */
/**********************************************************************/
void check_used_tmp( int flag )
{
	int i;
	static int lcount, lnum;

	if( flag ){
		display( M_DQUERY, "previous", lcount, lnum, Tmp.stringnum, Tmp.stringmax );
	} else {
		for( lcount = 0, i = 1; i < Tmp.listnum; i++ ){
			if( List[i].kind != KNOUSE ){
				lcount++;
			}
		}
		lnum = Tmp.listnum - 1;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function clear_tmpLisp()                               */
/*                                                                    */
/*             args NONE                                              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Erase the content of the TMPList and String. */
/*                                                                    */
/**********************************************************************/
void clear_tmpLisp( void )
{
	check_used_tmp( 0 );
	memset( List, 0, sizeof(List) * Tmp.listmax );
	Tmp.listnum = 1; Tmp.minimum = 1;
	memset( TmpString, 0, sizeof(char) * Tmp.stringmax );
	Tmp.stringnum = 1;
	clear_hash( TMP );
}


/**********************************************************************/
/*                                                                    */
/*             function clean_up()                                    */
/*                                                                    */
/*             struct ListStruct *ls: List Infomation                  */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function: If realloc error occured, free and re-malloc */
/*                      the TMP list and TMP string.                  */
/*            add B2                                                  */
/**********************************************************************/
static void clean_up( struct ListStruct *ls, int kind )
{
	if( strcmp( ls->name, "TMP" ) != 0 ){
		return;
	}

	free( List );
	if( (List = (struct List *)malloc( sizeof(struct List) * LISTMAX )) == NULL ){
		fatal( F_MALLOC, "List" );
	}
	Tmp.listmax = LISTMAX;

	free( TmpString );
	if( (TmpString = (char *)malloc( sizeof(char) * STRINGMAX )) == NULL ){
		fatal( F_MALLOC, "Local String" );
	}
	Tmp.stringmax = STRINGMAX;

	memset( List, 0, sizeof(List) * Tmp.listmax );
	Tmp.listnum = 1; Tmp.minimum = 1;
	memset( TmpString, 0, sizeof(char) * Tmp.stringmax );
	Tmp.stringnum = 1;
	clear_hash( TMP );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_strdup()                                 */
/*                                                                    */
/*             struct ListStruct *ls: List Infomation                  */
/*             char *str  : copy String(src)                          */
/*                                                                    */
/*             return  String: copied String(dst)                     */
/*                                                                    */
/*             function: Copy the String(Abstruct Level).             */
/*                                                                    */
/**********************************************************************/
String gedo_strdup( struct ListStruct *ls, char *str )
{
	char *tmp;
	int len;
	int ret;
/* A0 add */
	char strbuff[1024];
/* A0 end */

	len = strlen( str ) + 1;
/* A0 add */
	if( len < 1024 ){
		strcpy( strbuff, str );
	}
/* A0 end */
	if( ls->stringnum + len >= ls->stringmax ){
		if( (tmp = (char *)realloc( *(ls->string), sizeof(char) * (ls->stringmax + STRINGMAX) )) == NULL ){
/* B2 add */
			clean_up( ls, 1 );
/* B2 end */
			warning( W_LSCOSM, ls->name );
		}
		*(ls->string) = tmp;
		ls->stringmax += STRINGMAX;
		if( check_value( GLOBAL, "#GCreport" ) ){
			display( M_STRDUP, ls->name, ls->stringmax );
		}
	}
/* A0	strcpy( &(*ls->string)[ls->stringnum], str ); */
	strcpy( &(*ls->string)[ls->stringnum], strbuff );
/* A0 end */
	ret = ls->stringnum;
	ls->stringnum += len;
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function tmp_strdup()                                  */
/*             function local_strdup()                                */
/*             function global_strdup()                               */
/*                                                                    */
/*             char *str: copy String(src)                            */
/*                                                                    */
/*             return  String: copied String(dst)                     */
/*                                                                    */
/*             function: Each interface of String duplication.        */
/*                                                                    */
/**********************************************************************/
String tmp_strdup( char *str )
{
	Tmp.string = &TmpString;
	return( gedo_strdup( &Tmp, str ) );
}


String local_strdup( char *str )
{
	Local.string = &LocalString;
	return( gedo_strdup( &Local, str ) );
}

String global_strdup( char *str )
{
	Global.string = &GlobalString;
	return( gedo_strdup( &Global, str ) );
}

/**********************************************************************/
/*                                                                    */
/*             function local_qstrdup()                               */
/*                                                                    */
/*             char *str: copy String(src)                            */
/*                                                                    */
/*             return  String: copied String(dst)                     */
/*                                                                    */
/*             function: Interface for "string" duplication.          */
/*                       (neglect "", called from gedo.y)             */
/**********************************************************************/
String local_qstrdup( char *str )
{
	char work[FILENAME_MAX]; /* 1024 in linux */
	int i;

	for( i = 0; str[i+1] != '\"'; i++ ){
		work[i] = str[i+1];
	}
	work[i] = '\0';
	return( local_strdup( work ) );
}


/**********************************************************************/
/*                                                                    */
/*             function llist()                                       */
/*                                                                    */
/*             Lisp l1: first List                                    */
/*             Lisp l2: second List                                   */
/*                                                                    */
/*             return  Lisp: chained List.                            */
/*                                                                    */
/*             function: Chain the Lists(not LIST concat).            */
/*                       called from gedo.y                           */
/**********************************************************************/
Lisp llist( Lisp l1, Lisp l2 )
{
	Lisp at;

	for( at = l1; LocalList[at].next; at = LocalList[at].next );
	LocalList[at].next = l2;
	return( l1 );
}

/**********************************************************************/
/*                                                                    */
/*             function gedo_cons()                                   */
/*                                                                    */
/*             struct ListStruct *ls: List Infomation                  */
/*             int atom: values etc.                                  */
/*             int type: type of the List(see gedo.h)                 */
/*             Lisp at : cdr part of the List                         */
/*                                                                    */
/*             return  Lisp: created List                             */
/*                                                                    */
/*             function: create the cons List.(Abstruct Level)        */
/*                                                                    */
/**********************************************************************/
Lisp gedo_cons( struct ListStruct *ls, int atom, int type, Lisp at )
{
	struct List *tmp;

	while( ls->minimum < ls->listnum ){
		if( (*(ls->lisp))[ls->minimum].kind == KNOUSE ){
			break;
		}
		ls->minimum++;
	}
	if( ls->listnum >= ls->listmax ){
		if( (tmp = (struct List *)realloc( *(ls->lisp), sizeof(struct List) * (ls->listmax + LISTMAX) )) == NULL ){
/* B2 add */
			clean_up( ls, 0 );
/* B2 end */
			warning( W_LLCOSM, ls->name );
		}
		*(ls->lisp) = tmp;	/* change original adress of List */
		ls->listmax += LISTMAX;
		if( check_value( GLOBAL, "#GCreport" ) ){
			display( M_CONS, ls->name, ls->listmax );
		}
	}

	(*(ls->lisp))[ls->minimum].v = atom;
	(*(ls->lisp))[ls->minimum].kind = type;
	(*(ls->lisp))[ls->minimum++].next = at;
	if( ls->minimum > ls->listnum ){
		ls->listnum++;
	}
#if LISTLOG
	fprintf( Tfp, "%d used(%d)\n", ls->minimum-1, ls->listnum );
#endif	
	return( ls->minimum-1 );
}

/**********************************************************************/
/*                                                                    */
/*             function cons()                                        */
/*             function l_cons()                                      */
/*             function g_cons()                                      */
/*                                                                    */
/*             int atom: values etc.                                  */
/*             int type: type of the List(see gedo.h)                 */
/*             Lisp at : cdr part of the List                         */
/*                                                                    */
/*             return  Lisp: created cons List                        */
/*                                                                    */
/*             function: create each kind of the cons List.           */
/*                                                                    */
/**********************************************************************/
Lisp cons( int atom, int type, Lisp at )
{
	Lisp lp;

	Tmp.lisp = &List;
	lp = gedo_cons( &Tmp, atom, type, at );
	return( lp );
}

Lisp l_cons( int atom, int type, Lisp at )
{
	Local.lisp = &LocalList;
	return( gedo_cons( &Local, atom, type, at ) );
}

Lisp g_cons( int atom, int type, Lisp at )
{
	Global.lisp = &GlobalList;
	return( gedo_cons( &Global, atom, type, at ) );
}

/**********************************************************************/
/*                                                                    */
/*             function op_kind()                                     */
/*                                                                    */
/*             int kind:  List kind( cons(,kind,) )                   */
/*                                                                    */
/*             return  int: ==0: numeric kind(KNUMBER etc.)           */
/*                          ==1: literal kind(KSYMBOL, KLITERAAL etc.)*/
/*                          ==8: list kind(KLIST,KPOSLIST etc.)       */
/*             function: Return the conjunction kind of the List.     */
/*                                                                    */
/**********************************************************************/
int op_kind( int kind )
{
int r;
	if( kind >= KIND_MAX ){
		return( -1 );
	}
	r = Kind_spec[kind] & 0x0f;
	return( r );
}




/**********************************************************************/
/*                                                                    */
/*             function *get_tstr()                                   */
/*             function *get_lstr()                                   */
/*             function *get_gstr()                                   */
/*                                                                    */
/*             Lisp at: op_kind()==1 kind List(LITERAL kind)          */
/*                                                                    */
/*             return  char *: adress at String Buffer                */
/*                                                                    */
/*             function: Return each kind real String adress.         */
/*                                                                    */
/**********************************************************************/
static char *get_tstr( Lisp at )
{
	return( &TmpString[at] );
}

static char *get_lstr( Lisp at )
{
	return( &LocalString[at] );
}

static char *get_gstr( Lisp at )
{
	return( &GlobalString[at] );
}

/**********************************************************************/
/*                                                                    */
/*             function check_void()                                  */
/*                                                                    */
/*             Lisp at  : List of KCOM kind                           */
/*             struct dup_s *ds : List Infomation                     */
/*                                                                    */
/*             return  Lisp: KVOID changed List                       */
/*                                                                    */
/*             function: Special handling of first copy of the Command*/
/*                      List. Result reflection in commandline.       */
/**********************************************************************/
static int check_void( Lisp at, struct dup_s *ds )
{
	Lisp cdr;
	Lisp plug, after;

	/* monkey magic begins here */
	switch( (*(ds->Lat))[at].v ){
	case MLOAD:
	case MSAVE:
		after = (*(ds->Lat))[at].next;
		if( (*(ds->Lat))[after].kind == KVOID ){
			/* load */
			if( ds->Do_flag ){
				cdr = (*ds->Conat)( at, KVOID, NIL );
			} else {
				cdr = NIL;
			}
		} else {
			/* load calc */
			cdr = dup_Lisp( after, ds );
		}
		break;
	case MSELECT:
		after = (*(ds->Lat))[at].next;
		if( (*(ds->Lat))[after].kind == KVOID ){
			/* select, select calc */
			if( ds->Do_flag ){
				cdr = (*ds->Conat)( at, KVOID, dup_Lisp( (*(ds->Lat))[after].next, ds ) );
			} else {
				cdr = dup_Lisp( (*(ds->Lat))[after].next, ds );
			}
		} else {
			/* select calc */
			cdr = dup_Lisp( after, ds );
		}
		break;
	case MLOADER :
	case MSAVER:
	case MSELECTER:
	case MEFFECTER:
		if( (*(ds->Lat))[at].next == NIL ){
			cdr = NIL;
		} else if( (*(ds->Lat))[(*(ds->Lat))[at].next].kind == KFN ){
			/* selecter help */
			if( ds->Do_flag ){
				cdr = (*ds->Conat)( at, KFN, NIL );
			} else {
				cdr = (*ds->Conat)( NIL, KFN, NIL );
			}
		} else {
			plug = (*(ds->Lat))[at].next;
			after = (*(ds->Lat))[plug].next;
			if( (*(ds->Lat))[after].kind == KVOID ){
				/* selecter plug config */
				if( ds->Do_flag ){
					cdr = (*ds->Conat)( plug, KVOID, NIL );
				} else {
					cdr = (*ds->Conat)( NIL, KVOID, NIL );
				}
			} else if( (*(ds->Lat))[after].kind == NIL ){
				/* selecter plug */
				cdr = NIL;
			} else if( (*(ds->Lat))[after].kind == KFN ){ 
				/* selecter plug help */
				cdr = (*ds->Conat)( NIL, KFN, NIL );
			} else {
				/* selecter plug calc */
				cdr = dup_Lisp( after, ds );
			}
			cdr = (*ds->Conat)( (*ds->Dupat)( (*ds->GetStr)((*(ds->Lat))[plug].v)), (*(ds->Lat))[plug].kind, cdr );
		}
		break;
	default:
		return( NIL );
		break;
	}
	return( (*ds->Conat)( (*(ds->Lat))[at].v, (*(ds->Lat))[at].kind, cdr ) );
}

/**********************************************************************/
/*                                                                    */
/*             function dup_Lisp()                                    */
/*                                                                    */
/*             Lisp at : duplicate List(src)                          */
/*             struct dup_s *ds: List Infomation                      */
/*                                                                    */
/*             return  Lisp: duplicated List(dst)                     */
/*                                                                    */
/*             function: Copy the List(Abstruct Level).               */
/*                                                                    */
/**********************************************************************/
static Lisp dup_Lisp( Lisp at, struct dup_s *ds )
{

	if( at == NIL ){
		return( NIL );
	}
	switch( op_kind((*(ds->Lat))[at].kind) ){
	case 8:
		return( (*ds->Conat)( dup_Lisp((*(ds->Lat))[at].v, ds), (*(ds->Lat))[at].kind, dup_Lisp( (*(ds->Lat))[at].next, ds ) ) );
		break;
	case 1:
/* add B0 */
		if( ds->Exp_flag ){
			if( (*(ds->Lat))[at].kind == KSYMBOL ){
				return( (*ds->Conat)( (*(ds->Lat))[at].v, (*(ds->Lat))[at].kind, dup_Lisp( (*(ds->Lat))[at].next, ds ) ) );
			}
		}
/* end B0 */
		if( (*(ds->Lat))[at].v ){
			return( (*ds->Conat)( (*ds->Dupat)( (*ds->GetStr)((*(ds->Lat))[at].v) ), (*(ds->Lat))[at].kind, dup_Lisp( (*(ds->Lat))[at].next, ds ) ) );
		} else {
			return( (*ds->Conat)( NIL, (*(ds->Lat))[at].kind, dup_Lisp( (*(ds->Lat))[at].next, ds ) ) );
		}
		break;
	case 0:
		if( (*(ds->Lat))[at].kind == KCOM ){
			switch( (*(ds->Lat))[at].v ){
			case MLOAD:
			case MSAVE:
			case MSELECT:
			case MLOADER :
			case MSAVER:
			case MSELECTER:
			case MEFFECTER:
				return( check_void( at, ds ) );
				break;
			default:
				break;
			}
		}
		return( (*ds->Conat)( (*(ds->Lat))[at].v, (*(ds->Lat))[at].kind, dup_Lisp( (*(ds->Lat))[at].next, ds ) ) );
		break;
	default:
		warning( W_ILLKIND,  (*(ds->Lat))[at].kind, "dup_Lisp" );
		return( NIL );
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function dup_t2tLisp()                                 */
/*             function dup_l2lLisp()                                 */
/*             function dup_l2tLisp()                                 */
/*             function dup_t2lLisp()                                 */
/*             function dup_t2gLisp()                                 */
/*             function dup_g2tLisp()                                 */
/*                                                                    */
/*             Lisp at: duplicate List(src)                           */
/*                                                                    */
/*             return  Lisp: duplicated List(dst)                     */
/*                                                                    */
/*             function: Duplicate List by each kind.(interface)      */
/*                                                                    */
/**********************************************************************/
Lisp dup_t2tLisp( Lisp at )
{
	struct dup_s ds;

	ds.Lat = &List;
	ds.Conat = cons;
	ds.Dupat = tmp_strdup;
	ds.GetStr = get_tstr;
/*B0*/	ds.Exp_flag = 1;
/*	ds.Exp_flag = 0;*/
	ds.Do_flag = 0;
	return( dup_Lisp( at, &ds ) );
}

Lisp dup_l2lLisp( Lisp at )
{
	struct dup_s ds;

	ds.Lat = &LocalList;
	ds.Conat = l_cons;
	ds.Dupat = local_strdup;
	ds.GetStr = get_lstr;
	ds.Exp_flag = 0;
	ds.Do_flag = 0;
	return( dup_Lisp( at, &ds ) );
}

Lisp dup_l2tLisp( Lisp at )
{
	struct dup_s ds;

	ds.Lat = &LocalList;
	ds.Conat = cons;
	ds.Dupat = tmp_strdup;
	ds.GetStr = get_lstr;
	ds.Exp_flag = 0;
	ds.Do_flag = 0;
	return( dup_Lisp( at, &ds ) );
}

Lisp dup_t2lLisp( Lisp at )
{
	struct dup_s ds;

	ds.Lat = &List;
	ds.Conat = l_cons;
	ds.Dupat = local_strdup;
	ds.GetStr = get_tstr;
	ds.Exp_flag = 0;
	ds.Do_flag = 0;
	return( dup_Lisp( at, &ds ) );
}


Lisp dup_t2gLisp( Lisp at )
{
	struct dup_s ds;

	ds.Lat = &List;
	ds.Conat = g_cons;
	ds.Dupat = global_strdup;
	ds.GetStr = get_tstr;
	ds.Exp_flag = 0;
	ds.Do_flag = 0;
	return( dup_Lisp( at, &ds ) );
}

Lisp dup_g2tLisp( Lisp at )
{
	struct dup_s ds;

	ds.Lat = &GlobalList;
	ds.Conat = cons;
	ds.Dupat = tmp_strdup;
	ds.GetStr = get_gstr;
	ds.Exp_flag = 0;
	ds.Do_flag = 0;
	return( dup_Lisp( at, &ds ) );
}

/**********************************************************************/
/*                                                                    */
/*             function dup_doLisp()                                  */
/*                                                                    */
/*             Lisp at: duplicate List(src)                           */
/*                                                                    */
/*             return  Lisp: duplicated List(dst)                     */
/*                                                                    */
/*             function: Duplicate List but enable KVOID complete.    */
/*                                                                    */
/**********************************************************************/
Lisp dup_doLisp( Lisp at )
{
	struct dup_s ds;

	ds.Lat = &LocalList;
	ds.Conat = cons;
	ds.Dupat = tmp_strdup;
	ds.GetStr = get_lstr;
	ds.Exp_flag = 0;
	ds.Do_flag = 1;
	return( dup_Lisp( at, &ds ) );
}

/**********************************************************************/
/*                                                                    */
/*             function query()                                       */
/*                                                                    */
/*             int flag: ==0: not display the result                  */
/*                       !=0: display the result                      */
/*             int kind: storage kind(TMP,LOCAL,GLOBAL)               */
/*                                                                    */
/*             return  int: ==1: need GabageCollection                */
/*                          ==0: sucesss(not need GC)                 */
/*                          <0 : fail                                 */
/*             function: Query the Memory Usage                       */
/*                                                                    */
/**********************************************************************/
int query( int flag, int kind )
{
	unsigned int i;
	int fst;
	int min;
	int used, free;
	long perc;
	struct ListStruct *ls;
	struct List *lst;

	if( kind == GLOBAL ){
		ls = &Global;
		lst = GlobalList;
	} else if( kind == LOCAL ){
		ls = &Local;
		lst = LocalList;
	} else {
		warning( W_QUERY, kind );
		return( -1 );
	}
	fst = 1; used = free = 0; min = ls->minimum;
	for( i = 1; i < ls->listnum; i++ ){
		if( lst[i].kind != KNOUSE ){
			used++;
		} else {
			if( fst == 1 ){
				min = i;
				fst = 0;
			}
			free++;
		}
	}
	if( flag ){
		perc = 100L * (long)used / (long)ls->listmax;
                display( M_PERC, ls->name, used, free, ls->listmax, perc );
	}
	if( 100 * used / ls->listmax >= 90 ){
		return( 1 );
	} else {
		return( 0 );
	}
}



/**********************************************************************/
/*                                                                    */
/*             function do_query()                                    */
/*                                                                    */
/*             int flag: ==0: do not display the result               */
/*                       !=0: display the result                      */
/*             return NONE                                            */
/*                                                                    */
/*             function: Query all Memory Usage.                      */
/*                                                                    */
/**********************************************************************/
void do_query( int flag )
{
	query( flag, LOCAL );
	query( flag, GLOBAL );
	if( flag ){
		check_used_tmp( 1 );
		display( M_DQUERY, Tmp.name, Tmp.listnum, Tmp.listmax, Tmp.stringnum, Tmp.stringmax );
		display( M_DQUERY, Local.name, Local.listnum, Local.listmax, Local.stringnum, Local.stringmax );
		display( M_DQUERY, Global.name, Global.listnum, Global.listmax, Global.stringnum, Global.stringmax );
	}
#if LISTLOG
	fprintf( Tfp, "TMP   : List %7d/%7d  String %7d/%7d\n", Tmp.listnum, Tmp.listmax, Tmp.stringnum, Tmp.stringmax );
	fprintf( Tfp, "LOCAL : List %7d/%7d  String %7d/%7d\n", Local.listnum, Local.listmax, Local.stringnum, Local.stringmax );
	fprintf( Tfp, "GLOBAL: List %7d/%7d  String %7d/%7d\n", Global.listnum, Global.listmax, Global.stringnum, Global.stringmax );
#endif
}

/* add B0 */
/**********************************************************************/
/*                                                                    */
/*             function chase_chain()                                 */
/*                                                                    */
/*             Lisp at: start List                                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Show detail of the List for debug TMP List.  */
/*                                                                    */
/**********************************************************************/
static char *kname[KIND_MAX] = {
"KNOUSE", "KNUMBER", "KSYMBOL", "KSHARP", "KLITERAL", "KFN", "KCSTACK", "KNIL", "KMASK", "KPLUG", "KCOM", "KPOSITION", "KRANGE", "KFOR", "KWHILE", "KIF", "KLIST", "KPOSLIST", "KLET", "KPLUS", "KMINUS", "KMULT", "KDIVIDE", "KINCR", "KINCL", "KDECR", "KDECL", "KMOD", "KDNOT", "KDAND", "KDOR", "KEQ", "KNE", "KGE", "KGT", "KLE", "KLT", "KDEFUNC", "KFUNC", "KARGS", "KRET", "KCFUNC", "KDEMACRO", "KMARGS", "KHASH", "KVOID" };

static void chase_chain( Lisp at )
{
	if( at == NIL ){
		return;
	}
	if( List[at].kind >= KIND_MAX ){
		printf( "chase:%d:ILLKIND(%x)\n", at, List[at].kind );
		return;
	}
	switch( op_kind( List[at].kind ) ){
	case 8:
		printf( "chase:%d:%s\n", at, kname[List[at].kind] );
		chase_chain( List[at].v );
		chase_chain( List[at].next );
		break;
	case 1:
		printf( "chase:%d:%s(%s)\n", at, kname[List[at].kind], tsv(at) );
		chase_chain( List[at].next );
		break;
	case 0:
		printf( "chase:%d:%s(%d)\n", at, kname[List[at].kind], List[at].v );
		chase_chain( List[at].next );
		break;
	default:
		printf( "chase:%d:%s\n", at, kname[List[at].kind] );
		switch( List[at].kind ){
		case KHASH:
			chase_chain( List[at].next );
			break;
		case KVOID:
			chase_chain( List[at].next );
			break;
		case KNOUSE:
			return;
			break;
		default:
			printf( "chase:%d:ILLKIND(%x)\n", at, List[at].kind );
			break;
		}
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function chase_Lisp()                                  */
/*                                                                    */
/*             Lisp at: start List                                    */
/*             char *fmt: showing strings                             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Show detail of the List for debug.           */
/*                                                                    */
/**********************************************************************/
void chase_Lisp( Lisp at, char *fmt, ... )
{
	va_list argp;

	va_start( argp, fmt );
	vfprintf( stdout, fmt, argp );
	chase_chain( at );
}
/* end B0 */
/**********************************************************************/
/*                                                                    */
/*             function delete_chain()                                */
/*                                                                    */
/*             struct ListStruct *ls: List Infomation                  */
/*             struct List *lp: List Adress                           */
/*             Lisp at        : delete List                           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Delete the List(Abstruct Level).             */
/*                                                                    */
/**********************************************************************/
void delete_chain( struct ListStruct *ls, struct List *lp, Lisp at )
{
	if( at == NIL ){
		return;
	}
	if( lp[at].kind == KNOUSE ){
		return;
	}
	if(lp[at].kind >= KIND_MAX ){
		printf( "illegal access of delete_chain %02x\n", lp[at].kind );
		return;
	}
	switch( op_kind( lp[at].kind ) ){
	case 8:
		lp[at].kind = KNOUSE;
		if( at < ls->minimum ){
			ls->minimum = at;
		}
		delete_chain( ls, lp, lp[at].v );
		delete_chain( ls, lp, lp[at].next );
		lp[at].v = lp[at].next = NIL;
		break;
	case 1:
	case 0:
		lp[at].kind = KNOUSE;
		if( at < ls->minimum ){
			ls->minimum = at;
		}
		delete_chain( ls, lp, lp[at].next );
		lp[at].v = lp[at].next = NIL;
		break;
	default:
		if( lp[at].kind == KHASH || lp[at].kind == KVOID ){
			if( at < ls->minimum ){
				ls->minimum = at;
			}
			lp[at].kind = KNOUSE;
			delete_chain( ls, lp, lp[at].next );
			lp[at].v = lp[at].next = NIL;
		} else {
			warning( W_ILLKIND, lp[at].kind, "delete_chain" );
		}
		break;
	}
}


/**********************************************************************/
/*                                                                    */
/*             function delete_Lisp()                                 */
/*                                                                    */
/*             int kind: storage kind(TMP,LOCAL,GLOBAL)               */
/*             Lisp at : delete List                                  */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Delete the List.                             */
/*                                                                    */
/**********************************************************************/
void delete_Lisp( int kind, Lisp at )
{
	struct ListStruct *ls;
	struct List *lp;

	if( kind == LOCAL ){
		ls = &Local;
		lp = LocalList;
	} else if( kind == GLOBAL ){
		ls = &Global;
		lp = GlobalList;
	} else if( kind == TMP ){
		ls = &Tmp;
		lp = List;
	} else {
		warning( W_COLLECT );
		return;
	}
	delete_chain( ls, lp, at );
}
	

/**********************************************************************/
/*                                                                    */
/*             function check_used()                                  */
/*                                                                    */
/*             struct List *lp: List adress                           */
/*             Lisp at        : start List                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Check the each used Cell.                    */
/*                                                                    */
/**********************************************************************/
void check_used( struct List *lp, Lisp at )
{
	if( at == NIL ){
		return;
	}
	if( lp[at].kind & 0x80 ){
		return;
	}
	switch( op_kind( lp[at].kind ) ){
	case 8:
		lp[at].kind |= 0x80;
		check_used( lp, lp[at].v );
		check_used( lp, lp[at].next );
		break;
	case 1:
	case 0:
		lp[at].kind |= 0x80;
		check_used( lp, lp[at].next );
		break;
	default:
		if( lp[at].kind == KHASH || lp[at].kind == KVOID ){
			lp[at].kind |= 0x80;
			check_used( lp, lp[at].next );
		} else {
			warning( W_ILLKIND, lp[at].kind, "check_used" );
		}
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function collect_list()                                */
/*                                                                    */
/*             int kind: storage kind(TMP,LOCAL,GLOBAL)               */
/*             int flag: ==0:display result                           */
/*                       !=0:do not display result                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: GabageCollect the Non-check List.            */
/*                       (use check_used() before collect_list())     */
/**********************************************************************/
void collect_list( int kind, int flag )
{
	unsigned int i;
	int fst;
	struct ListStruct *ls;
	struct List *lp;
	int num;

	if( kind == LOCAL ){
		ls = &Local;
		lp = LocalList;
	} else if( kind == GLOBAL ){
		ls = &Global;
		lp = GlobalList;
	} else if( kind == TMP ){
		ls = &Tmp;
		lp = List;
	} else {
		warning( W_COLLECT );
		return;
	}
	fst = 1; num = 0;
	for( i = 1; i < ls->listnum; i++ ){
		if( lp[i].kind & 0x80 ){
			lp[i].kind &= 0xff7f;
		} else {
			if( fst == 1 ){
				ls->minimum = i;
				fst = 0;
			}
			lp[i].kind = KNOUSE;
			lp[i].v = lp[i].next = NIL;
#if LISTLOG
			fprintf( Tfp, "%d freed\n", i );
#endif
			num++;
		}
	}
	if( flag ){
		display( M_FREE, ls->name, ls->listnum, ls->listmax, num );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function gc_string()                                   */
/*                                                                    */
/*             int flag: ==0:display result                           */
/*                       !=0:do not display result                    */
/*             int kind: storage kind(TMP,LOCAL,GLOBAL)               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: GabageCollect the String Buffer.             */
/*                                                                    */
/**********************************************************************/
void gc_string( int flag, int kind )
{
	struct ListStruct *ls, new;
	char *work;
	unsigned int i;
	int num;
	int len;
	Lisp at;

	if( kind == LOCAL ){
		ls = &Local;
		ls->lisp = &LocalList;
		ls->string = &LocalString;
	} else if( kind == GLOBAL ){
		ls = &Global;
		ls->lisp = &GlobalList;
		ls->string = &GlobalString;
	} else if( kind == TMP ){
		ls = &Tmp;
		ls->lisp = &List;
		ls->string = &TmpString;
	} else {
		return;
	}
	if( (work = (char *)malloc( sizeof(char) * ls->stringmax )) == NULL ){
		clean_up( ls, 1 );
		warning( W_LSCOSM, ls->name );
		return;
	}

	new.string = &work;
	new.stringnum = 1;
	new.stringmax = ls->stringmax;
	
	for( i = 1; i < ls->listnum; i++ ){
		if( op_kind((*(ls->lisp))[i].kind) == 1 || 
		    (*(ls->lisp))[i].kind == KHASH ){
			    at = (*(ls->lisp))[i].v;
			    len = strlen( &(*(ls->string))[at] );
			    strcpy( &work[new.stringnum], &(*(ls->string))[at] );
			    ((*ls->lisp))[i].v = new.stringnum;
			    new.stringnum += (len+1);
		    }
	}
	if( kind == GLOBAL ){
		num = get_childsnum();
		for( i = 0; i < (unsigned int)num; i++ ){
			if( Childs[i].used == NOUSE ){
				Childs[i].name = NIL;
				Childs[i].path = NIL;
				Childs[i].option = NIL;
				Childs[i].help = NIL;
			} else {
				if( Childs[i].name ){
					len = strlen( &(*(ls->string))[Childs[i].name] );
					strcpy( &work[new.stringnum], &(*(ls->string))[Childs[i].name] );
					Childs[i].name = new.stringnum;
					new.stringnum += (len+1);
				}
				if( Childs[i].path ){
					len = strlen( &(*(ls->string))[Childs[i].path] );
					strcpy( &work[new.stringnum], &(*(ls->string))[Childs[i].path] );
					Childs[i].path = new.stringnum;
					new.stringnum += (len+1);
				}
				if( Childs[i].option ){
					len = strlen( &(*(ls->string))[Childs[i].option] );
					strcpy( &work[new.stringnum], &(*(ls->string))[Childs[i].option] );
					Childs[i].option = new.stringnum;
					new.stringnum += (len+1);
				}
				if( Childs[i].help ){
					len = strlen( &(*(ls->string))[Childs[i].help] );
					strcpy( &work[new.stringnum], &(*(ls->string))[Childs[i].help] );
					Childs[i].help = new.stringnum;
					new.stringnum += (len+1);
				}
			}
		}
		for( i = 0; i <	(unsigned int)TotalFileNum; i++ ){
			if( FileName[i] ){
				len = strlen( &(*(ls->string))[FileName[i]] );
				strcpy( &work[new.stringnum], &(*(ls->string))[FileName[i]] );
				FileName[i] = new.stringnum;
				new.stringnum += (len+1);
			}
		}
	}
	if( flag ){
		display( M_COLLECT, ls->name, ls->stringnum, new.stringnum, ls->stringmax );
	}
	 free( *(ls->string) );
	*(ls->string) = work;
	ls->stringnum = new.stringnum;
}

/**********************************************************************/
/*                                                                    */
/*             function check_string_used()                           */
/*                                                                    */
/*             int kind: storage kind(TMP,LOCAL,GLOBAL)               */
/*             int flag: ==0:display result                           */
/*                       !=0:do not display result                    */
/*                                                                    */
/*             return  int: ==1: need GabageCollection                */
/*                          ==0: not need GC                          */
/*             function: Do need the string GC?                       */
/*                                                                    */
/**********************************************************************/
int check_string_used( int kind, int flag )
{
	struct ListStruct *ls;

	if( kind == GLOBAL ){
		ls = &Global;
	} else if( kind == LOCAL ){
		ls = &Local;
	} else if( kind == TMP ){
		ls = &Tmp;
	} else {
		return( 0 );
	}
	if( flag ){
		display( M_STRNOW, ls->name, ls->stringnum, ls->stringmax );
	}
	if( ls->stringnum * 100 / ls->stringmax > 80 ){
		return( 1 );
	} else {
		return( 0 );
	}
}
					     
/**********************************************************************/
/*                                                                    */
/*             function do_gc()                                       */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: GabageCollect the List and String Buffer.    */
/*                                                                    */
/**********************************************************************/
void do_gc( void )
{
	int flag;

	if( check_value( GLOBAL, "#GCreport" ) ){
		flag = 1;
	} else {
		flag = 0;
	}
	do_query( flag );
	markup_line();
	markup_hash( LOCAL, LocalList );
	collect_list( LOCAL, flag );
	markup_hash( GLOBAL, GlobalList );
	collect_list( GLOBAL, flag );
	gc_string( flag, LOCAL );
	gc_string( flag, GLOBAL );
	do_query( flag );
}
