/**********************************************************************/
/*                                                                    */
/*           main.c -- gedo main routine and I/O                      */
/*                                                                    */
/*                by K.sugiura   Ver 0.9  for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*           11.21,1996 A0 add rewind_jmp for include,gc,prev,next,goto */
/*            4.27,1997 B2 level 2 execution re-write at main()       */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <setjmp.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>

#define MAIN	1
#include "gedo.h"
#include "../loader.h"
#include "prot.h"
#include "error.h"

extern FILE *yyout;	/* lex scanner stdout */
extern FILE *yyin;	/* lex scanner stdin */
extern int Parse;	/* 0==command is complete,  0>command line continue */
jmp_buf errjmp;		/* jump buffer at error */
/* A0 add */
jmp_buf rewindjmp;	/* rewind the input at command goto, next, prev */ 

/**********************************************************************/
/*                                                                    */
/*             function change_file()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : At change file, reset specail variable and  */
/*                       restart lex scanner.                         */
/**********************************************************************/
void change_file( void )
{
	char filename[FILENAME_MAX];

	if( FileName[FileNum] ){
		strcpy( filename, gs(FileName[FileNum]) );
/* B2 		load_init( filename );*/
		set_hash( LOCAL, "$filenum", l_cons( FileNum, KNUMBER, NIL ) );
		set_hash( LOCAL, "$filename", l_cons( local_strdup(filename), KLITERAL, NIL ) );
		
		make_cmdname( 0, filename );
		set_hash( LOCAL, "$cmdname", l_cons( local_strdup(filename), KLITERAL, NIL ) );
		make_cmdname( 1, filename );
		set_hash( LOCAL, "$$cmdname", l_cons( local_strdup(filename), KLITERAL, NIL ) );
		set_hash( LOCAL, "$all", l_cons( 0, KMASK, NIL ) );
/* B2*/		load_init( filename );
	}
	yyin = stdin;
	yyrestart( yyin );
}

/**********************************************************************/
/*                                                                    */
/*             function fpe_error()                                   */
/*                                                                    */
/*             int signo : singal no.                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Signal handler for FPE error interrupt.     */
/*                                                                    */
/**********************************************************************/
void fpe_error( int signo )
{
        signal( SIGFPE, fpe_error );
	display( M_FPE );
	In_loop = 0; Func_count = 0;
	Error_flag = 0;
	longjmp( errjmp, 1 );
}


/**********************************************************************/
/*                                                                    */
/*             function ctrl_c()                                      */
/*                                                                    */
/*             int signo : signal no.                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Signal handler for ^C interrput.            */
/*                                                                    */
/**********************************************************************/
void ctrl_c( int signo )
{
	int c;

	signal( SIGINT, ctrl_c );
	display( M_ABORT );
	while( (c = getchar()) == 0x0a );
	c = toupper(c);
	if( c == 'S' ){
		if( (c = check_save_outfile()) == 0 ){
			return;
		}
		endp( c );
		exit( 0 );
	} else if( c == 'A' ){
		set_org_termio();
		kill_child();
		exit( 3 );
	}
	Error_flag = 1;
}

/**********************************************************************/
/*                                                                    */
/*             function fatal_signal_trap()                           */
/*                                                                    */
/*             int signo : signal no.                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Signal handler for Segmentation Fault.      */
/*                                                                    */
/**********************************************************************/
void fatal_signal_trap( int signo )
{
	signal( SIGSEGV, fatal_signal_trap );
	display( M_SEGV );
	endp( 'N' );
	exit( 2 );
}

/**********************************************************************/
/*                                                                    */
/*             function error_jump()                                  */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : If user error occured, rewind the state.    */
/*                                                                    */
/**********************************************************************/
void error_jump( void )
{
	In_loop = 0; Func_count = 0;
	Error_flag = 0;
	longjmp( errjmp, 1 );
}

/**********************************************************************/
/*                                                                    */
/*             function rewind_input()                                */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : If command prev, next or goto, rewind the   */
/*                       state.                                       */
/**********************************************************************/
/* A0 add rewindjmp */
void rewind_input( void )
{
	In_loop = 0; Func_count = 0;
	Error_flag = 0;
	longjmp( rewindjmp, 1 );
}
/* A0 end */
/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             int argc    : command-line                             */
/*             char **argv : command-line                             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : The Main Routine.                           */
/*                                                                    */
/**********************************************************************/
void main( int argc, char **argv )
{
	if( init( argc, argv ) == -1 ){
		exit( 1 );
	}
	signal( SIGSEGV, fatal_signal_trap );
	fprintf( stderr, CopyRight );
	change_file();
	while( 1 ){
		setjmp( rewindjmp );
/* B2		yyparse();*/
/* add B2 */
		if( Level == 2 ){
			if( next_file( 1 ) == 1 ){
				endp( 'N' );
				break;
			}
		} else {
			yyparse();
		}
/* end B2 */
	}
}

/**********************************************************************/
/*                                                                    */
/*             function my_input()                                    */
/*                                                                    */
/*             char *buf : input buffer                               */
/*             int max   : input maximum                              */
/*                                                                    */
/*             return  int : read length                              */
/*                                                                    */
/*             function : Read the string from each device.           */
/*                        Replace the default lex scanner.            */
/**********************************************************************/
int my_input( char *buf, int max )
{
	int meslen;
	char mes[64];

	if( !Init_flag ){
		if( FileName[FileNum] == NIL ){
			strcpy( mes, "noname>" );
		} else if( !Parse ){
			strcpy( mes, gs(FileName[FileNum]) );
			strcat( mes, ">" );
		} else {
			mes[0] = '\0';
		}
	} else {
		mes[0] = '\0';
	}
	if( isatty(fileno(yyin)) == 0 ){
                printf( mes );
		if( fgets( buf, max, yyin ) == NULL ){
			return( 0 );
		}
	} else {
		printf( mes );
		meslen = strlen( mes );
		get_str( buf, max - 1, meslen );
	}
	In_loop = 0; Func_count = 0;
	Error_flag = 0;
	clear_tmpLisp();
	return( strlen( buf ) );
}


/**********************************************************************/
/*                                                                    */
/*             function fatal()                                       */
/*                                                                    */
/*             int no : fatal error no.                               */
/*             ...    : other argments.                               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : If fatal error occured, show message and    */
/*                       quit the process.                            */
/**********************************************************************/
void fatal( int no, ... )
{
	va_list argp;
	char work[128];
	int c;

        perror( "gedo:" );
	va_start( argp, no );
	sprintf( work, "fatal:%s\n", FatalMes[no] );
	vfprintf( stdout, work, argp );
	if( isatty(fileno(stdout)) == 0 ){
		vfprintf( stderr, work, argp );
	}
	c = check_save_outfile();
	endp( c );
	exit( 1 );
}

/**********************************************************************/
/*                                                                    */
/*             function warning()                                     */
/*                                                                    */
/*             int no : warning no.                                   */
/*             ...    : other argments                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : If user error occured, show the message     */
/*                       and rewind the state.                        */
/**********************************************************************/
void warning( int no, ... )
{
	va_list argp;
	char work[128];

	argp = NULL;
	Level = RealLevel;
	if( no != W_NONE ){
		sprintf( work, "warning:%s\n", WarningMes[no] );
		va_start( argp, no );
		vfprintf( stdout, work, argp );
	}
	if( Level == 2 ){
		if( isatty(fileno(stdout)) == 0 ){
			vfprintf( stderr, work, argp );
		}
		fprintf( stderr, "Level 2 abort\n" );
		endp( 'N' );
		exit( 1 );
	} else {
		In_loop = 0; Func_count = 0;
		Error_flag = 0;
		longjmp( errjmp, 1 );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function display()                                     */
/*                                                                    */
/*             int no : message no.                                   */
/*             ...    : other argments                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Show the message.                           */
/*                                                                    */
/**********************************************************************/
void display( int no, ... )
{
	va_list argp;
	char *arg1;

	va_start( argp, no );
	if( no == M_FORM ){
		arg1 = va_arg(argp, char *);
		vfprintf( stdout, arg1, argp );
	} else {
		vfprintf( stdout, DisplayMes[no], argp );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function dwarning()                                    */
/*                                                                    */
/*             char *form : display format                            */
/*             ...        : other argments                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : If usr error occured, show message and      */
/*                       rewind the state.                            */
/*                       At debug phase, use equally warning().       */
/**********************************************************************/
void dwarning( char *form, ... )
{
	va_list argp;

	Level = RealLevel;
	va_start( argp, form );
	vfprintf( stdout, form, argp );
	if( Level == 2 ){
		if( isatty(fileno(stdout)) == 0 ){
			vfprintf( stderr, form, argp );
		}
		fprintf( stderr, "Level 2 abort\n" );
		endp( 'N' );
		exit( 1 );
	} else {
		longjmp( errjmp, 1 );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function yywrap()                                      */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : true                                     */
/*                                                                    */
/*             function : Need at yacc.                               */
/*                                                                    */
/**********************************************************************/
int yywrap( void )
{
	return( 1 );
}

#ifdef MINIMUM
#undef realloc

/**********************************************************************/
/*                                                                    */
/*             function *my_realloc()                                 */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Debug tool for memory allocation.           */
/*                                                                    */
/**********************************************************************/
void *my_realloc( void * ptr, size_t size )
{
	printf( "realloc occured\n" );
	return( realloc( ptr, size ) );
}
#endif

