/**********************************************************************/
/*                                                                    */
/*           mask.c -- gedo mask operation routines                   */
/*                                                                    */
/*                by K.sugiura   Ver 1.0  for GEDO of LINUX           */
/*                                                                    */
/*         4.23,1997 B2 rewrite func get_shmid, free_shmid            */
/*                      create new equal_name, search_Mem, new_Mem    */
/*                                 reset_shmid   for multiple viewer. */
/*         7.5, 1997 B2 multi size mask can handle.                   */
/**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#ifndef SHMMAX
#define SHMMAX  0x2000000
#endif

#include "gedo.h"
#include "error.h"
#include "prot.h"


static unsigned char BMask[8]	/* Bit Mask for Bit bashing */
 = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };
static struct Shm *Shm;		/* Now selected Shared Memory */
/* B2 add */
struct Memory {			/* Shared Memory control struct */
	int shmid;		/* IPC shm ID */
	Lisp name;		/* registerd name */
	struct Shm *shm;	/* alloced shm adress */
};

static struct Memory Mem[MAXVIEWER];

static int MaxShmID;		/* number of now alloced IPC shm buffers */
/* B2 end */

#ifdef MINIMUM
#define TMPMASKMAX 3
#else
#define TMPMASKMAX 30		/* mask variable alloc unit */
#endif


struct TmpMask {		/* mask variable control struct */
	unsigned char *buff;	/* buffer of mask */
	int flag;		/* 0==not used  1==using 2==used but not free*/
	int mask;		/* mask kind */
/* B2 add */
	int width;		/* mask width */
	int height;		/* mask height */
};

struct TmpMask *TM;

static int TmpMaskNum;		/* now used TmpMask */
static int TmpMaskMax;		/* now alloced TmpMask */

static int ShmBuffSize;		/* size of IPC shm buffer size */
static int ShmPixelSize;	/* pixels maximum at a IPC shm */
static int SimpleMaskSize;	/* maximum size of SIMPLEMASK */
static int DepthMaskSize;	/* maximum size of DEPTHMASK */

/**********************************************************************/
/*                                                                    */
/*             function get_shmid()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : first alloced anonymous shm ID           */
/*                                                                    */
/*             function : get MAXVIEWER IPC shm ID, alloc 1st anonymous*/
/*                       shm to address, and alloc TmpMask.           */
/*           B2 rewrited for multi viewer.                            */
/**********************************************************************/
int get_shmid( char *ini )
{
	int shmid;
	int shmkey;
	int count;
	int shmsize;
	int num;

	if( ini == NULL ){
		shmsize = SHMBUFFSIZE;
	} else if( strcmp( ini, "max" ) == 0 ){
		shmsize = SHMMAX - 16;
	} else if( (num = atoi( ini )) > 0 ){
		shmsize = num * num * 3;
		if( shmsize > SHMMAX -16 ){
			shmsize = SHMMAX - 16;
		}
	} else {
		shmsize = SHMBUFFSIZE;
	}
	num = (shmsize - sizeof(struct Shm)) / sizeof(int);
	set_hash( GLOBAL, "#maxpixel", g_cons( num, KNUMBER, NIL ) );

	ShmBuffSize = shmsize;
	ShmPixelSize = shmsize / 4;
	SimpleMaskSize = ShmPixelSize / 8;
	DepthMaskSize = ShmPixelSize;
	count = 0;
	for( shmkey = IPCSHMID; shmkey < IPCSHMID + 20; shmkey++ ){
		if( (shmid = shmget( shmkey, ShmBuffSize, 
				    IPC_CREAT | IPC_EXCL | 0600 ) ) == -1 ){
			shmid = shmget( shmkey, ShmBuffSize, 0600 ) ;
		}
		if( shmid > 0 ){
			Mem[count++].shmid = shmid;
			if( count >= MAXVIEWER ){
				break;
			}
		}
	}
	if( count == 0 ){
		fatal( F_IPCSHM );
		return( -1 );
	} else if( count < MAXVIEWER ){
		printf( "IPC shm alloced only %d\n", count );
	}
	MaxShmID = count;

	Shm = (struct Shm *)shmat( Mem[0].shmid, NULL, 0 );
	if( Shm == (struct Shm *)-1 ){
		fatal( F_IPCSHM );
		return( -1 );
	}
	Mem[0].shm = Shm;
	Mem[0].name = NIL;

	if( (TM = (struct TmpMask *)malloc( sizeof(struct TmpMask) * TMPMASKMAX )) == NULL ){
		fatal( F_MALLOC, "tmp mask " );
		return( -1 );
	}
        if( (TM[0].buff = malloc( SimpleMaskSize )) == NULL ){
		fatal( F_MALLOC, "tmp mask" );
		return( -1 );
	}
        memset( TM[0].buff, 0, SimpleMaskSize );
        TM[0].flag = 1;
        TM[0].mask = SIMPLEMASK;
        TmpMaskNum = 1;
	TM[0].width = TM[0].height = 0;

        return( Mem[0].shmid );
}

/**********************************************************************/
/*                                                                    */
/*             function equal_name()                                  */
/*                                                                    */
/*             args   Lisp name1 : LITERAL global List(in Mem)        */
/*                    Lisp name2 : LITERAL tmp List(from commandline) */
/*                                                                    */
/*             return  int : 1==equal 0==different                    */
/*                                                                    */
/*             function : name comparision for search Mem by name.    */
/*                                                                    */
/*           B2 for multi viewers                                     */
/**********************************************************************/
static int equal_name( Lisp name1, Lisp name2 )
{
	unsigned char *s1, *s2;

	if( name1 == NIL ){
		return( 0 );
	}
	if( GlobalList[name1].kind == KLITERAL ){
		s1 = gsv(name1);
	} else {
		warning( W_VIEWER );
		return( 0 );
	}
	if( List[name2].kind == KLITERAL ){
		s2 = tsv(name2);
	} else {
		warning( W_VIEWER );
		return( 0 );
	}
	if( strcmp( s1, s2 ) == 0 ){
		return( 1 );
	} else {
		return( 0 );
	}
}

/* add B2 for multi viewers */
/**********************************************************************/
/*                                                                    */
/*             function search_Mem()                                  */
/*                                                                    */
/*             args   Lisp name : search Mem name                     */
/*                                                                    */
/*             return  int : >=0 found index of name, -1==not found   */
/*                                                                    */
/*             function : search Mem by name.                         */
/*                                                                    */
/*           B2 for multi viewers                                     */
/**********************************************************************/
static int search_Mem( Lisp name )
{
	int i;

	if( name == NIL ){
		return( 0 );
	} else if( List[name].kind == KNIL ){
		return( 0 );
	}
	for( i = 1; i < MaxShmID; i++ ){
		if( equal_name( Mem[i].name, name ) ){
			return( i );
		}
	}
	return( -1 );
}

/**********************************************************************/
/*                                                                    */
/*             function new_Mem()                                     */
/*                                                                    */
/*             args   Lisp name : create Mem name                     */
/*                                                                    */
/*             return  int : >=0 created Mem index  -1=canot create   */
/*                                                                    */
/*             function : create new Mem my name.                     */
/*                                                                    */
/*           B2 for multi viewers                                     */
/**********************************************************************/
/* add B2 for multi viewers */
static int new_Mem( Lisp name )
{
	int i;
	struct Shm *s;

	if( List[name].kind != KLITERAL ){
		printf( "name must be LITERAL\n" );
		return( -1 );
	}
	for( i = 1; i < MaxShmID; i++ ){
		if( Mem[i].shm == NULL ){
			break;
		}
	}
	if( i == MaxShmID ){
		warning( W_MAXVIEWER );
		return( -1 );
	}
	s = (struct Shm *)shmat( Mem[i].shmid, NULL, 0 );
	if( s == (struct Shm *)-1 ){
		warning( W_IPCSHM );
		return( -1 );
	}
	Mem[i].shm = s;
	Mem[i].name = dup_t2gLisp( name );

	return( i );
}


/**********************************************************************/
/*                                                                    */
/*             function reset_shmid()                                 */
/*                                                                    */
/*             args   Lisp name : get Mem name                        */
/*                                                                    */
/*             return  int :  0==success errorcode==error             */
/*                                                                    */
/*             function : change current Mem by name.                 */
/*                                                                    */
/*           B2 for multi viewers                                     */
/**********************************************************************/
/* add B2 for multi viewers */
int reset_shmid( Lisp name )
{
	int index;

	if( (index = search_Mem( name )) == -1 ){
		if( (index = new_Mem( name )) == -1 ){
			return( W_IPCSHM );
		}
/* add D2 for change $width and $height */
		Shm = Mem[index].shm;
		NowShmid = Mem[index].shmid;
		Mem[index].shm->width = Mem[index].shm->height = 0;
		smart_set_number( "$width", Mem[index].shm->width );
		smart_set_number( "$height", Mem[index].shm->height );
		return 0;
/* end D2 */
	}
	if( index < MaxShmID ){
		if( Mem[index].shm == NULL ){
			return( W_IPCSHM );
		}
		Shm = Mem[index].shm;
		NowShmid = Mem[index].shmid;
/* add D2 for change $width and $height */
		smart_set_number( "$width", Mem[index].shm->width );
		smart_set_number( "$height", Mem[index].shm->height );
/* end D2 */
		return( 0 );
	} else {
		return( W_MAXVIEWER );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function get_shmid_by_name()                           */
/*                                                                    */
/*             args   Lisp name : get Mem name                        */
/*                                                                    */
/*             return  int :  -1==cannot find,  else == shmid         */
/*                                                                    */
/*             function : Get shmid by name.                          */
/*                                                                    */
/*           B2 for multi viewers                                     */
/**********************************************************************/
int get_shmid_by_name( Lisp name )
{
	int index;

	if( (index = search_Mem( name )) == -1 ){
		return( -1 );
	}
	return( Mem[index].shmid );
}

/**********************************************************************/
/*                                                                    */
/*             function delete_Mem_by_name()                          */
/*                                                                    */
/*             args   Lisp name : delete Mem name                     */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function : Make UNSED the Mem by name.                 */
/*                                                                    */
/*           B2 for multi viewers                                     */
/**********************************************************************/
void delete_Mem_by_name( Lisp name )
{
	int index;

	if( (index = search_Mem( name )) == -1 ){
		return;
	}
	if( Shm == Mem[index].shm ){
		Shm = Mem[0].shm;
		NowShmid = Mem[0].shmid;
		set_hash( GLOBAL, "#viewer", g_cons( NIL, KLITERAL, NIL ) );
	}

	shmdt( (char *)Mem[index].shm );
	Mem[index].shm = NULL;
	Mem[index].name = NIL;
}

/**********************************************************************/
/*                                                                    */
/*             function do_viewer()                                   */
/*                                                                    */
/*             args   Lisp name : set current Mem by name             */
/*                                                                    */
/*             return  NONE                                           */
/*                                                                    */
/*             function : Set current Shm and NowShmid by name.       */
/*                        Interface for eval() in funcs.c.            */
/*           B2 for multi viewers                                     */
/**********************************************************************/
void do_viewer( Lisp com )
{
	int ret;
	Lisp g;

	if( Level == 1 ){
		return;
	}
/* D2 multi viewer selection showing */
	if( reconfig_viewer( 0 ) != 0 ){
	     warning( W_OMNI, "cannot down now viewer" );
	}
/* end D2 */
	if( (ret = reset_shmid( com )) != 0 ){
		warning( ret );
	}
/* D2 multi viewer selection showing */
	if( reconfig_viewer( 1 ) != 0 ){
	     warning( W_OMNI, "cannot up new viewer" );
	}
/* end D2 */
	if( com == NIL ){
		g = g_cons( NIL, KLITERAL, NIL );
		set_hash( GLOBAL, "#viewer", g );
	} else {
		smart_set_global_literal( "#viewer", tsv(com) );
	}
}
/* add B2 end */

/**********************************************************************/
/*                                                                    */
/*             function free_shmid()                                  */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Destruct all alloced IPC shm and alloced    */
/*                       Mask buffer.                                 */
/*           B2 rewrited for multi viewer.                            */
/**********************************************************************/
void free_shmid( void )
{
	int i;

	for( i = 0; i < MaxShmID; i++ ){
		if( Mem[i].shm ){
			shmdt( (char *)Mem[i].shm );
		}
		if( shmctl( Mem[i].shmid, IPC_RMID, NULL ) == -1 ){
			perror("");
		}
	}
	for( i = 0; i < TmpMaskNum; i++ ){
		if( TM[i].buff ){
			free( TM[i].buff );
		}
	}
	free( TM );
}

/**********************************************************************/
/*                                                                    */
/*             function clear_maskinfo()                              */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Reset all Mask not used.                    */
/*                        Called if changed current File(next,prev,goto) */
/**********************************************************************/
void clear_maskinfo( void )
{
	int i;

	for( i = 1; i < TmpMaskNum; i++ ){
		TM[i].flag = 0;
	}
	TmpMaskNum = 1;
}

/**********************************************************************/
/*                                                                    */
/*             function get_newmask()                                 */
/*                                                                    */
/*             int kind : mask kind                                   */
/*                                                                    */
/*             return  int : 0> Tmp Mask index  -1==fail              */
/*                                                                    */
/*             function : Get new available Mask.(reuse or new alloc) */
/*                                                                    */
/**********************************************************************/
int get_newmask( int kind )
{
	int i;
	int size;
	struct TmpMask *tmp;

	for( i = 0; i < TmpMaskNum; i++ ){
		if( TM[i].flag == 0 && TM[i].mask == kind ){
			return( i );
		}
	}
	if( TmpMaskNum >= TmpMaskMax ){
		if( (tmp = (struct TmpMask *)realloc( TM, sizeof(struct TmpMask) * (TmpMaskMax + TMPMASKMAX) )) == NULL ){
			warning( W_TMCOSM );
		}
		TM = tmp;
		TmpMaskMax += TMPMASKMAX;
	}

	if( kind == SIMPLEMASK ){
		size = SimpleMaskSize;
	} else if( kind == DEPTHMASK ){
		size = DepthMaskSize;
	} else {
		warning( W_MASKTYPE );
		return( -1 );
	}
	if( (TM[TmpMaskNum].buff = malloc( size )) == NULL ){
		warning( W_TMCOSM );
	}
	TM[TmpMaskNum].mask = kind;
	TmpMaskNum++;
	return( TmpMaskNum - 1 );
}

/**********************************************************************/
/*                                                                    */
/*             function name_mask()                                   */
/*                                                                    */
/*             int id : TmpMask index                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Assoc Mask to symbol and clear used flag    */
/*                       at tmporary evaluated Mask.                  */
/**********************************************************************/
void name_mask( int id )
{
	int i;

	TM[id].flag = 1;
	for( i = 0; i < TmpMaskNum; i++ ){
		if( TM[i].flag == 2 ){
			TM[i].flag = 0;
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function make_tmpmask()                                */
/*                                                                    */
/*             int id : alloced Mask index(not used)                  */
/*                                                                    */
/*             return  int : set Mask Index                           */
/*                                                                    */
/*             function : Get new mask and set current Data to the    */
/*                       mask.                                        */
/*          B2: rewrite for size of TM                                */
/**********************************************************************/
int make_tmpmask( int id )
{
	int i, j;
	unsigned int *p;
	unsigned char *q;
	unsigned char b;
	int count, new;
	int pmask;

	new = get_newmask( Shm->mask_kind );
	if( Shm->mask_kind == SIMPLEMASK ){
		pmask = MASKBIT;
		for( i = 0; i < Shm->height; i++ ){
			p = Shm->data + i * Shm->width;
			q = TM[new].buff + i * (Shm->width + 7 ) / 8;
			count = 0; b = 0;
			for( j = 0; j < Shm->width; j++ ){
				if( p[j] & pmask){
					b |= BMask[count];
				}
				if( ++count >= 8 ){
					q[j / 8] = b;
					count = 0; b = 0;
				}
			}
			if( count % 8 ){
				q[j / 8] = b;
			}
		}
	} else if( Shm->mask_kind == DEPTHMASK ){
		for( i = 0; i < Shm->height; i++ ){
			p = Shm->data + i * Shm->width;
			q = TM[new].buff + i * Shm->width;
			for( j = 0; j < Shm->width; j++ ){
				*q++ = (unsigned char)((*p++) >> 24);
			}
		}
	}
	TM[new].width = Shm->width;
	TM[new].height = Shm->height;
	return( new );
}

/**********************************************************************/
/*                                                                    */
/*             function do_not()                                      */
/*                                                                    */
/*             Lisp at : MASK List                                    */
/*                                                                    */
/*             return  Lisp : new reverse Mask List                   */
/*                                                                    */
/*             function : get reverse Mask  (new_reverse=~mask())     */
/*          B2: rewrite for size of TM                                */
/*                                                                    */
/**********************************************************************/
Lisp do_not( Lisp at )
{
	int i, size;
	unsigned char *p;
	unsigned char *q;
	int id, new;

	if( !at && List[at].kind != KMASK ){
		warning( W_NOTMASK );
		return( NIL );
	}
	id = List[at].v;
	p = TM[id].buff;
        new = get_newmask( TM[id].mask );
        q = TM[new].buff;
        TM[new].flag = 2;
	TM[new].width = TM[id].width;
	TM[new].height = TM[id].height;
	if( TM[id].mask == SIMPLEMASK ){
		size = TM[id].height * (TM[id].width + 7) / 8;
	} else if( TM[id].mask == DEPTHMASK ){
		size = TM[id].height * TM[id].width;
	} else {
		warning( W_MASKTYPE );
		return( NIL );
	}
	for( i = 0; i < size; i++ ){
		*q++ = (unsigned char)(~*p++);
	}
	return( cons( new, KMASK, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function do_and()                                      */
/*                                                                    */
/*             Lisp at1 : mask List                                   */
/*             Lisp at2 : mask List                                   */
/*                                                                    */
/*             return  Lisp : and mask of at1 and at2                 */
/*                                                                    */
/*             function : Get new and-mask of at1 and at2.            */
/*          B2: rewrite for size of TM                                */
/*                                                                    */
/**********************************************************************/
Lisp do_and( Lisp at1, Lisp at2 )
{
	int i, j;
	unsigned char *p, *q, *r;
	int new;
	int d1, d2;
	unsigned int tmp;
	int minx, miny;

	if( !at1 && List[at1].kind != KMASK ){
		warning( W_NOTMASK );
		return( NIL );
	}
	if( !at2 && List[at2].kind != KMASK ){
		warning( W_NOTMASK );
		return( NIL );
	}
	d1 = List[at1].v; d2 = List[at2].v;
	p = TM[d1].buff;
	q = TM[d2].buff;
	if( TM[d1].mask == SIMPLEMASK && TM[d2].mask == SIMPLEMASK ){
		new = get_newmask( SIMPLEMASK );
		TM[new].mask = SIMPLEMASK;
	} else {
		new = get_newmask( DEPTHMASK );
		TM[new].mask = DEPTHMASK;
	}
        r = TM[new].buff;
        TM[new].flag = 2;
	TM[new].width =  min( TM[d1].width, TM[d2].width );
	TM[new].height = min( TM[d1].height, TM[d2].height );

	minx = min( TM[d1].width, TM[d2].width );
	miny = min( TM[d1].height, TM[d2].height );

	if( TM[d1].mask == SIMPLEMASK && TM[d2].mask == SIMPLEMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[d1].buff + i * (TM[d1].width + 7) / 8;
			q = TM[d2].buff + i * (TM[d2].width + 7) / 8;
			r = TM[new].buff + i * (TM[new].width + 7) / 8;
			for( j = 0; j < (minx + 7) / 8; j++ ){
				*r++ = (*p++ | *q++);
			}
		}
	} else if( TM[d1].mask == DEPTHMASK && TM[d2].mask == DEPTHMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[d1].buff + i * TM[d1].width;
			q = TM[d2].buff + i * TM[d2].width;
			r = TM[new].buff + i * TM[new].width;
			for( j = 0; j < minx; j++ ){
				tmp = (unsigned int)(*p++) + (unsigned int)(*q++);
				if( tmp > 0xff ){
					tmp = 0xff;
				}
				*r++ = (unsigned char)tmp;
			}
		}
	} else if( TM[d1].mask == SIMPLEMASK && TM[d2].mask == DEPTHMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[d1].buff + i * (TM[d1].width + 7) / 8;
			q = TM[d2].buff + i * TM[d2].width;
			r = TM[new].buff + i * TM[new].width;
			for( j = 0; j < minx; j++ ){
				if( BMask[j%8] & p[j / 8] ){
					r[j] = 0;
				} else {
					r[j] = q[j];
				}
			}
		}
	} else if( TM[d1].mask == DEPTHMASK && TM[d2].mask == SIMPLEMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[d1].buff + i * TM[d1].width;
			q = TM[d2].buff + i * (TM[d2].width + 7) / 8;
			r = TM[new].buff + i * TM[new].width;
			for( j = 0; j < minx; j++ ){
				if( BMask[j%8] & q[j / 8] ){
					r[j] = 0;
				} else {
					r[j] = p[j];
				}
			}
		}
	} else {
		warning( W_MASKTYPE );
	}
	return( cons( new, KMASK, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function do_or()                                       */
/*                                                                    */
/*             Lisp at1 : MASK List                                   */
/*             Lisp at2 : MASK List                                   */
/*                                                                    */
/*             return  Lisp : new or-mask with at1 and at2            */
/*                                                                    */
/*             function : Get or-mask with at1 and at2.               */
/*          B2: rewrite for size of TM                                */
/*                                                                    */
/**********************************************************************/
Lisp do_or( Lisp at1, Lisp at2 )
{
	int i, j, new;
	int d1, d2;
	unsigned char *p, *q, *r;
	int minx, miny;

	if( !at1 && List[at1].kind != KMASK ){
		warning( W_NOTMASK );
		return( NIL );
	}
	if( !at2 && List[at2].kind != KMASK ){
		warning( W_NOTMASK );
		return( NIL );
	}
	d1 = List[at1].v; d2 = List[at2].v;
	p = TM[d1].buff;
	q = TM[d2].buff;
	if( TM[d1].mask == SIMPLEMASK && TM[d2].mask == SIMPLEMASK ){
		new = get_newmask( SIMPLEMASK );
		TM[new].mask = SIMPLEMASK;
	} else {
		new = get_newmask( DEPTHMASK );
		TM[new].mask = DEPTHMASK;
	}
        r = TM[new].buff;
        TM[new].flag = 2;
	TM[new].width =  min( TM[d1].width, TM[d2].width );
	TM[new].height = min( TM[d1].height, TM[d2].height );

	minx = min( TM[d1].width, TM[d2].width );
	miny = min( TM[d1].height, TM[d2].height );

	if( TM[d1].mask == SIMPLEMASK && TM[d2].mask == SIMPLEMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[d1].buff + i * (TM[d1].width + 7) / 8;
			q = TM[d2].buff + i * (TM[d2].width + 7) / 8;
			r = TM[new].buff + i * (TM[new].width + 7) / 8;
			for( j = 0; j < (minx + 7) / 8; j++ ){
				*r++ = (*p++ & *q++);
			}
		}
	} else if( TM[d1].mask == DEPTHMASK && TM[d2].mask == DEPTHMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[d1].buff + i * TM[d1].width;
			q = TM[d2].buff + i * TM[d2].width;
			r = TM[new].buff + i * TM[new].width;
			for( j = 0; j < minx; j++ ){
				*r++ = max(*p,*q);
				p++; q++;
			}
		}
	} else if( TM[d1].mask == SIMPLEMASK && TM[d2].mask == DEPTHMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[d1].buff + i * (TM[d1].width + 7) / 8;
			q = TM[d2].buff + i * TM[d2].width;
			r = TM[new].buff + i * TM[new].width;
			for( j = 0; j < minx; j++ ){
				if( BMask[j%8] & p[j / 8] ){
					r[j] = q[j];
				} else {
					if( q[j] >= 0x80 ){
						r[j] = 0xff;
					} else {
						r[j] = 0x00;
					}
				}
			}
		}
	} else if( TM[d1].mask == DEPTHMASK && TM[d2].mask == SIMPLEMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[d1].buff + i * TM[d1].width;
			q = TM[d2].buff + i * (TM[d2].width + 7) / 8;
			r = TM[new].buff + i * TM[new].width;
			for( j = 0; j < minx; j++ ){
				if( BMask[j%8] & q[j / 8] ){
					r[j] = p[j];
				} else {
					if( p[j] >= 0x80 ){
						r[j] = 0xff;
					} else {
						r[j] = 0x00;
					}
				}
			}
		}
	} else {
		warning( W_MASKTYPE );
	}
	return( cons( new, KMASK, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function set_area()                                    */
/*                                                                    */
/*             Lisp mask : MASK List                                  */
/*                                                                    */
/*             return  int : 0==success -1==fail                      */
/*                                                                    */
/*             function : Set mask to current image buffer.           */
/*          B2: rewrite for size of TM                                */
/*                                                                    */
/**********************************************************************/
int set_area( Lisp mask )
{
	unsigned int *q;
	unsigned char *p;
	unsigned int maskd;
	int i, j, num;
	int minx, miny;

	if( List[mask].kind != KMASK ){
		warning( W_ILLHASH );
		return( -1 );
	}
	num = List[mask].v;
	maskd = MASKBIT;
	q = Shm->data;
	p = TM[num].buff;
	minx = min( Shm->width, TM[num].width );
	miny = min( Shm->height, TM[num].height );
	if( TM[num].mask == SIMPLEMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[num].buff + i * (TM[num].width + 7) / 8;
			q = Shm->data + i * Shm->width;
			for( j = 0; j < minx; j++ ){
				if( BMask[j%8] & p[j / 8] ){
					q[j] = (q[j] | maskd);
				} else {
					q[j] = (q[j] & (~maskd));
				}
			}
		}
	} else if( TM[num].mask == DEPTHMASK ){
		for( i = 0; i < miny; i++ ){
			p = TM[num].buff + i * TM[num].width;
			q = Shm->data + i * Shm->width;
			for( j = 0; j < minx; j++ ){
				q[j] = ((q[j] & 0x00ffffff) | (p[j] << 24));
			}
		}
	} else {
		warning( W_MASKTYPE );
	}
	Shm->mask_kind = TM[num].mask;
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function do_show()                                     */
/*                                                                    */
/*             Lisp com : MASK List(set mask) or                      */
/*                        LITERAL List(configure viewer)              */
/*             return NONE                                            */
/*                                                                    */
/*             function : Show mask or configure viewer.              */
/*                        Interface for eval() in funcs.c.            */
/**********************************************************************/
void do_show( Lisp com )
{
	if( Level == 1 ){
		return;
	}
	if( com ){

/* del B1      	if( List[com].kind != KMASK ){
			warning( W_SHOW );
		}
		set_area( com );
*/
/* add B1 */
		if( List[com].kind == KMASK ){
			set_area( com );
		} else if( List[com].kind == KLITERAL ){
			if( strlen(tsv(com) ) < 20 ){
				viewer_config( tsv(com) );
				return;
			} else {
				warning( W_SHOW );
			}
		} else {
			warning( W_SHOW );
		}
/* add end */
	}
	if( Level == 0 ){
		show_mask( 0 );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function eval_doller()                                 */
/*                                                                    */
/*             Lisp ls : Mask operation formula                       */
/*                                                                    */
/*             return  Lisp : evaluated new Mask                      */
/*                                                                    */
/*             function : Evaluate formula and operate Mask.          */
/*                                                                    */
/**********************************************************************/
Lisp eval_doller( Lisp ls )
{
	int num;
	Lisp ret;

	if( List[ls].kind == KLIST || List[ls].kind == KCOM ){
		/* from command */
		exec_exp( ls );
		num = make_tmpmask( 0 );
	} else if( List[ls].kind == KMASK ){
		num = List[ls].v;
	} else {
		warning( W_ILLKIND, List[ls].kind, "eval_doller" );
		return( NIL );
	}
	ret = cons( num, KMASK, NIL );
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function set_wh()                                      */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : Set the width and height if new image       */
/*                        loaded.                                     */
/**********************************************************************/
void set_wh( void )
{
/* B2 add */
	TM[0].width = Shm->width;
	TM[0].height = Shm->height;
/* B2 end */
	smart_set_number( "$width", Shm->width );
	smart_set_number( "$height", Shm->height );
}

/**********************************************************************/
/*                                                                    */
/*             function create_shm()                                  */
/*                                                                    */
/*             int width : width of new image                         */
/*             int height : height of new image                       */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function : if create, set new image to current shm.    */
/*                                                                    */
/**********************************************************************/
void create_shm( int width, int height )
{
	Shm->width = width;
	Shm->height = height;
	Shm->mask_kind = SIMPLEMASK;
	memset( Shm->data, 0, width * height * sizeof(unsigned int) );
}
