/**********************************************************************/
/*                                                                    */
/*        sharp.c -- gedo variable routines                           */
/*                                                                    */
/*                by K.sugiura   Ver 1.0  for GEDO of LINUX           */
/*                                                                    */
/*        A0 11.19,1996  eval_position, calc had a BUG, Fixed         */
/*        B0  1.16,1997   demacro add the function                     */
/*        B0  1.17,1997   function argment must be LIST set_function_arg()*/
/*        B0  1.20,1997   macro can execute KRET, in calc_value()     */
/*        B0  1.24,1997   superflous copy collecter push_node()       */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <setjmp.h>
#include <ctype.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

static Lisp *CalcStack;		/* Calculation Stack */
static int CalcSp = 0;		/* Stack Pointer of CalcStack */
#define CALCSTACKMAX 100	/* alloc unit of CalcStack */
static int CalcStackMax;	/* now alloced CalcStack */

static Lisp do_calc( Lisp ls );

extern struct CFuncTable CFuncTable[];
extern int Debug_flag;
Lisp NowBody;

struct StackStruct {		/* Local Variable Stack */
	Lisp ls;		/* Mask the before AUTO variable */
	int kind;		/* =0:not used =TMP:used before function */
};


/**********************************************************************/
/*                                                                    */
/*             function init_calcstack()                              */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int =0:success =-1:fail                        */
/*                                                                    */
/*             function: Alloc the CalcStack(realloc in use).         */
/*                                                                    */
/**********************************************************************/
int init_calcstack( void )
{
	if( (CalcStack = (int *)malloc( sizeof(Lisp)* CALCSTACKMAX )) == NULL ){
		fatal( F_MALLOC, "calc stack" );
		return( -1 );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function free_calcstack()                              */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Free the CalcStack at the end_proc.          */
/*                                                                    */
/**********************************************************************/
void free_calcstack( void )
{
	free( CalcStack );
}

/**********************************************************************/
/*                                                                    */
/*             function push_calcstack()                              */
/*                                                                    */
/*             Lisp ls : Operand of the calculation                   */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Push the Operand to CalcStack and Check the MAX */
/*                                                                    */
/**********************************************************************/
void push_calcstack( Lisp ls )
{
	int *tmp;

	CalcStack[CalcSp++] = ls;
	if( CalcSp >= CalcStackMax ){
		if( (tmp = (int *)realloc( CalcStack, sizeof(Lisp) * (CalcStackMax + CALCSTACKMAX) )) == NULL ){
			warning( W_MSCOSM );
		}
		CalcStack = tmp;
		CalcStackMax += CALCSTACKMAX;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function append_string()                               */
/*                                                                    */
/*             Lisp v1: List of LITERAL 1                             */
/*             Lisp v2: List of LITERAL 2                             */
/*                                                                    */
/*             return  Lisp: Appended new LTERAL-List of v1 and v2    */
/*                                                                    */
/*             function: Append 2 string-List.                        */
/*                                                                    */
/**********************************************************************/
Lisp append_string( Lisp v1, Lisp v2 )
{
	sprintf( Work, "%s%s", tsv(v1), tsv(v2) );
	return( cons( tmp_strdup( Work ), KLITERAL, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function make_string()                                 */
/*                                                                    */
/*             Lisp v1: List of INT                                   */
/*                                                                    */
/*             return  Lisp: List of LITERAL                          */
/*                                                                    */
/*             function: Convert INT-List to LITERAL-List.            */
/*                                                                    */
/**********************************************************************/
Lisp make_string( Lisp v1 )
{
	sprintf( Work, "%d", List[v1].v );
	return( cons( tmp_strdup( Work ), KLITERAL, NIL ) );
}

/**********************************************************************/
/*                                                                    */
/*             function eval_position()                               */
/*                                                                    */
/*             Lisp pos: List of POSITION                             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Eval each element of POSITION-LIST.          */
/*                                                                    */
/**********************************************************************/
/* A0 (n,n)+(m,m) cannot work */
/* void eval_position( Lisp pos )
{
	int ret;

	if( List[pos].kind == KSYMBOL ){
		ret = get_variable( List[pos].v );
		if( List[ret].kind != KPOSITION ){
			warning( W_ILLPOSL );
		}
		List[pos].kind = List[ret].kind;
		List[pos].v = List[ret].v;
		List[pos].next = List[ret].next;
	} else {
		ret = do_calc( List[pos].v );
		List[pos].v = ret;
		ret = do_calc( List[pos].next );
		List[pos].next = ret;
	}
}*/
Lisp eval_position( Lisp pos )
{
	int ret;

	if( List[pos].kind == KSYMBOL ){
		ret = get_variable( List[pos].v );
		if( List[ret].kind != KPOSITION ){
			warning( W_ILLPOSL );
		}
		return( ret );
	} else {
		return( cons( do_calc( List[pos].v ), KPOSITION, do_calc( List[pos].next ) ) );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function dup_poslist()                                 */
/*                                                                    */
/*             Lisp v: List of POSLIST                                */
/*                                                                    */
/*             return  Lisp: List of POSLIST(New)                     */
/*                                                                    */
/*             function: Duplicate the POSLIST-List.                  */
/*                                                                    */
/**********************************************************************/
Lisp dup_poslist( Lisp v )
{
	Lisp at;

	if( List[v].kind == KPOSITION ){
/* A0 */
/*		eval_position( v );
		return( v );*/
		return( eval_position( v ) );
	} else if( List[v].kind == KPOSLIST ){
/* A0 */
/*		eval_position( List[v].v );
		return( cons( List[v].v, List[v].kind, dup_poslist(List[v].next) ) );*/
		return( cons( eval_position( List[v].v ), KPOSLIST, dup_poslist(List[v].next) ) );

	} else if( List[v].kind == KSYMBOL ){
		at = get_variable( List[v].v );
		if( List[at].kind == KPOSITION ){
			return( at );
		} else if( List[at].kind == KPOSLIST ){
			return( at );
		} else {
			warning( W_ILLPOSL );
			return( NIL );
		}
	} else {
		warning( W_ILLPOSL );
		return( NIL );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function append_position()                             */
/*                                                                    */
/*             Lisp v1: List of POSITION or POSLIST                   */
/*             Lisp v2: List of POSITION or POSLIST                   */
/*                                                                    */
/*             return  Lisp: Appended POSLIST of v1 and v2(new)       */
/*                                                                    */
/*             function                                               */
/*                                                                    */
/**********************************************************************/
Lisp append_position( Lisp v1, Lisp v2 )
{
	Lisp v, at, ret, r2;

	v = ret = dup_poslist( v1 );
	r2 = dup_poslist( v2 );
/* add A0 */
	if( List[v].kind == KPOSITION ){
		return( cons( v, KPOSLIST, r2 ) );
	}
/* add A0 end */
	while( v ){
		if( List[List[v].next].kind == KPOSITION ){
			at = cons( List[v].next, KPOSLIST, r2 );
			List[v].next = at;
			return( ret );
		} else if( List[List[v].next].kind == KPOSLIST ){
			v = List[v].next;
		} else {
			warning( W_ILLPOSL );
			return( NIL );
		}
	}
	return( NIL );
}

/**********************************************************************/
/*                                                                    */
/*             function calc()                                        */
/*                                                                    */
/*             int command: Calculation Command like KPLUS and etc.   */
/*                                                                    */
/*             return  Lisp: Calculated Result List                   */
/*                                                                    */
/*             function: Calculate value from CalcStack obeying command. */
/*                                                                    */
/**********************************************************************/
static Lisp calc( int command )
{
	Lisp v1, v2;
	Lisp ret;

	ret = NIL;
	v1 = CalcStack[CalcSp-2]; v2 = CalcStack[CalcSp-1];
	if( List[v1].kind == KNIL ){
		if( List[v2].kind == KPOSITION || 
				List[v2].kind == KPOSLIST ){
			return( dup_poslist(v2) );
		} else {
			return( v2 );
		}
	} else if( List[v2].kind == KNIL ){
		if( List[v1].kind == KPOSITION ||
				List[v1].kind == KPOSLIST ){
			return( dup_poslist(v1) );
		} else {
			return( v1 );
		}
	}
	switch( command ){
	case KPLUS:
		switch( List[v1].kind ){
		case KNUMBER:
			switch( List[v2].kind ){
			case KNUMBER:
				ret = cons( List[v1].v + List[v2].v, KNUMBER, NIL );
				break;
			case KLITERAL:
				ret = append_string( make_string( v1 ), v2 );
				break;
			default:
				warning( W_ILLTYPE );
				break;
			}
			break;
		case KLITERAL:
			switch( List[v2].kind ){
			case KNUMBER:
				ret = append_string( v1, make_string( v2 ) );
				break;
			case KLITERAL:
				ret = append_string( v1, v2 );
				break;
			default:
				warning( W_ILLTYPE );
				break;
			}
			break;
		case KPOSITION:
			switch( List[v2].kind ){
			case KPOSLIST:
			case KPOSITION:
/* A0 */
/*				eval_position( v1 );
				ret = cons( v1, KPOSLIST, dup_poslist(v2) );*/
				ret = cons( eval_position( v1 ), KPOSLIST, dup_poslist(v2) );
				break;
			default:
				warning( W_ILLTYPE );
				break;
			}
			break;
		case KPOSLIST:
			switch( List[v2].kind ){
			case KPOSLIST:
			case KPOSITION:
				ret = append_position( v1, v2 );
				break;
			default:
				warning( W_ILLTYPE );
				break;
			}
			break;
		}
		break;
	case KMINUS:
	        if( List[v1].kind != KNUMBER || List[v2].kind != KNUMBER ){
			warning( W_ILLTYPE );
			return( NIL );
		}
		ret = cons( List[v1].v - List[v2].v, KNUMBER, NIL );
		break;
	case KMULT:
	        if( List[v1].kind != KNUMBER || List[v2].kind != KNUMBER ){
			warning( W_ILLTYPE );
			return( NIL );
		}
		ret = cons( List[v1].v * List[v2].v, KNUMBER, NIL );
		break;
	case KDIVIDE:
	        if( List[v1].kind != KNUMBER || List[v2].kind != KNUMBER ){
			warning( W_ILLTYPE );
			return( NIL );
		}
		if(  List[v2].v == 0 ){
			warning( W_DIV0 );
		}
		ret = cons( List[v1].v / List[v2].v, KNUMBER, NIL );
		break;
	case KMOD:
	        if( List[v1].kind != KNUMBER || List[v2].kind != KNUMBER ){
			warning( W_ILLTYPE );
			return( NIL );
		}
		if(  List[v2].v == 0 ){
			warning( W_DIV0 );
		}
		ret = cons( List[v1].v % List[v2].v, KNUMBER, NIL );
		break;
	case KDAND:
		if( List[v1].kind == KMASK && List[v2].kind == KMASK ){
			ret = do_and( v1, v2 );
		} else if( List[v1].kind == KNUMBER && List[v2].kind == KNUMBER ){
			ret = cons( List[v1].v & List[v2].v, KNUMBER, NIL );
		} else {
			return( NIL );
		}
		break;
	case KDOR:
		if( List[v1].kind == KMASK && List[v2].kind == KMASK ){
			ret = do_or( v1, v2 );
		} else if( List[v1].kind == KNUMBER && List[v2].kind == KNUMBER ){
			ret = cons( List[v1].v | List[v2].v, KNUMBER, NIL );
		} else {
			return( NIL );
		}
		break;
	default:
		return( NIL );
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function mono_op()                                     */
/*                                                                    */
/*             Lisp ls: Calculate Operation                           */
/*                                                                    */
/*             return  Lisp: Calculated Result                        */
/*                                                                    */
/*             function: Calculate the mono-operator.                 */
/*                                                                    */
/**********************************************************************/
static Lisp mono_op( Lisp ls )
{
	Lisp var;
	Lisp lp;
	Lisp ret;

	var = List[ls].v;
	switch( List[var].kind ){
	case KSYMBOL:
		lp = get_variable( List[var].v );
		break;
	default:
		lp = do_calc( var );
		break;
	}
	if( List[lp].kind == KNUMBER ){
		switch( List[ls].kind ){	
		case KINCR:
			ret = cons( List[lp].v, KNUMBER, NIL );
			smart_set_number( tsv(var), List[lp].v+1 );
			return( ret );
			break;
		case KINCL:
			ret = cons( List[lp].v+1, KNUMBER, NIL );
			smart_set_number( tsv(var), List[lp].v+1 );
			return( ret );
			break;
		case KDECR:
			ret = cons( List[lp].v, KNUMBER, NIL );
			smart_set_number( tsv(var), List[lp].v-1 );
			return( ret );
			break;
		case KDECL:
			ret = cons( List[lp].v-1, KNUMBER, NIL );
			smart_set_number( tsv(var), List[lp].v-1 );
			return( ret );
			break;
		case KDNOT:
			ret = cons( ~List[lp].v, KNUMBER, NIL );
			return( ret );
			break;
		default:
			warning( W_ILLTYPE );
			break;
		}
	} else if( List[lp].kind == KMASK ){
		if( List[ls].kind != KDNOT ){
			warning( W_ILLTYPE );
			return( NIL );
		}
		return( do_not( lp ) );
	} else if( List[lp].kind == KLIST || List[lp].kind == KCOM ){
		if( List[ls].kind != KDNOT ){
			warning( W_ILLTYPE );
			return( NIL );
		}
		lp = eval_doller( var );
		return( do_not( lp ) );
	}
	return( NIL );
}

/**********************************************************************/
/*                                                                    */
/*             function calc_value()                                  */
/*                                                                    */
/*             Lisp ls: Calculate List                                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Caluculate the formula.                      */
/*                       Push value to CalcStack and calculate.       */
/*             rev B0: return in macro can behavor                    */
/**********************************************************************/
static void calc_value( Lisp ls )
{
	Lisp at;

	switch( List[ls].kind ){
	case KNUMBER:
	case KLITERAL:
	case KNIL:
		push_calcstack( ls );
		break;
	case KPOSITION:
/* A0 */
/*		CalcStack[CalcSp-1] = cons( do_calc( List[ls].v ), KPOSITION, do_calc( List[ls].next ) );*/
		push_calcstack( cons( do_calc( List[ls].v ), KPOSITION, do_calc( List[ls].next ) ) );
		break;
	case KPOSLIST:
		push_calcstack( ls );
		break;
	case KSYMBOL:
		push_calcstack( get_variable( List[ls].v ) );
		break;
	case KSHARP:
		push_calcstack( expand_macro( List[ls].v ) );
		break;
	case KMASK:
		push_calcstack( eval_doller( ls ) );
		break;
	case KCOM:
		CalcStack[CalcSp++] = ls;
		break;
	case KFUNC:
		push_calcstack( eval_exp( ls ) );
		break;
	case KINCL:
	case KDECL:
	case KINCR:
	case KDECR:
	case KDNOT:
		push_calcstack( mono_op( ls ) );
		break;
	case KPLUS:
	case KMINUS:
	case KMULT:
	case KDIVIDE:
	case KMOD:
	case KDAND:
	case KDOR:
		at = List[ls].v;
		calc_value( at );
		at = List[ls].next;
		calc_value( at );
		CalcStack[CalcSp-2] = calc( List[ls].kind );
		CalcSp--;
		break;
	case KLIST:
		do_calc( ls );
		break;
	case KLET:
		set_name( List[ls].v, List[ls].next );
		calc_value( List[ls].v );
		break;
/* add B0 */
	case KRET:
		eval_exp( ls );
		break;
/* end B0 */
	default:
		warning( W_ILLKIND, List[ls].kind, "calc_value" );
		break;
	}
}


/**********************************************************************/
/*                                                                    */
/*             function do_calc()                                     */
/*                                                                    */
/*             Lisp ls: Calculate List                                */
/*                                                                    */
/*             return  Lisp: Calculated List                          */
/*                                                                    */
/*             function: Evaluate and Calculate the List.             */
/*                       (only LIST processing.)                      */
/**********************************************************************/
static Lisp do_calc( Lisp ls )
{
	Lisp ret, last;
	Lisp at;
	int sp_sv;

	if( ls == NIL ){
		return( NIL );
	}
	if( List[ls].kind == KLIST ){
		last = NIL; ret = NIL;
		for( at = ls; at; at = List[at].next ){
			if( last ){
				List[last].next = cons( NIL, KLIST, NIL );
				last = List[last].next;
			} else {
				ret = last = cons( NIL, KLIST, NIL );
			}
			if( List[List[at].v].kind == KFUNC ){
				List[at].v = eval_exp( List[at].v );
			}
			sp_sv = CalcSp;
			push_calcstack( List[at].v );
			calc_value( List[at].v );
			List[last].v = CalcStack[--CalcSp];
			if( CalcSp != sp_sv ){
				CalcSp = sp_sv;
			}
		}
	} else {
		if( List[ls].kind == KFUNC ){
			ls = eval_exp( ls );
		}
		sp_sv = CalcSp;
		calc_value( ls );
		ret = CalcStack[--CalcSp];
		if( CalcSp != sp_sv ){
			CalcSp = sp_sv;
		}
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function get_varkind()                                 */
/*                                                                    */
/*             int *kind: <OUT> =TMP: mac is the TMP-Variable.        */
/*                              =LOCAL: mac is the LOCAL-Variable.    */
/*             Lisp mac:  the SYMBOL List                             */
/*                                                                    */
/*             return  Lisp: the Value of mac                         */
/*                                                                    */
/*             function: expand variable and return *kind             */
/*                                                                    */
/**********************************************************************/
Lisp get_varkind( int *kind, Lisp mac )
{
	Lisp at;
	if( (at = search_hash( TMP, ts(mac) )) ){
		*kind = TMP;
		return( at );
	} else if( (at = search_hash( LOCAL, ts(mac) )) ){
		*kind = LOCAL;
		return( dup_l2tLisp(at) );
	} else {
		*kind = -1;
		return( NIL );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function restore_arg()                                 */
/*                                                                    */
/*             Lisp dst: Virtual-Argment of function                 */
/*             struct StackStruct *stack: Before value of V-Arg       */
/*             Lisp *ev: not used.(reserved for release List)         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Restore the original value to V-Arg after    */
/*                      returning from function.                      */
/**********************************************************************/
void restore_arg( Lisp dst, struct StackStruct *stack, Lisp *ev )
{
	int i;

	for( i= 0; List[dst].kind == KLIST; i++ ){
		if( stack[i].kind == TMP ){
			set_hash( TMP, tsvv(dst), stack[i].ls );
		} else {
			set_hash( TMP, tsvv(dst), NIL );
		}
		dst = List[dst].next;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function debug_routine()                               */
/*                                                                    */
/*             int flag: =0:Called at the top of function.            */
/*                       =1:Called at the argment of function.        */
/*                       =-1:Called at the RETURN of function.        */
/*             char *name: Display string                             */
/*             Lisp at:    Display value                              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: if #debug is true, this routine called.      */
/*                      Display infomation using nest.                */
/**********************************************************************/
void debug_routine( int flag, char *name, Lisp at )
{
	static int nest = 1;

	if( flag == 0 ){
		printf( "%*ceval %s:\n", (nest++)*4, ' ', name );
	} else if( flag == -1 ){
		printf( "%*cret  %s ", (--nest)*4, ' ', name );
		write_of_print( at );
	} else {
		printf( "%*carg %d:%s ", nest*4, ' ', flag, name );
		write_of_print( at ); 
	}
}

/**********************************************************************/
/*                                                                    */
/*             function set_function_arg()                            */
/*                                                                    */
/*             Lisp dst: Virtual-Argments of the function             */
/*             Lisp src: Real-Argments of the function                */
/*             struct StackStruct *stack: Save the original V-Arg value*/
/*             Lisp *ev: Evaluated R-Arg                              */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Apply the R-Arg to V-Arg, masking old value  */
/*                      at the calling of function.                   */
/*             rev B0: not appear args !KLIST                         */
/**********************************************************************/
void set_function_arg( Lisp dst, Lisp src, struct StackStruct *stack, Lisp *ev )
{
	Lisp s, d;
	int i;
	int argnum;

	for( d = dst, i= 0; List[d].kind == KLIST; i++, d = List[d].next ){
		if( i >= ARGMENT_MAX ){
			restore_arg( dst, stack, ev );
			warning( W_MANYARG );
		}
		if( List[List[d].v].kind != KSYMBOL ){
			restore_arg( dst, stack, ev );
			warning( W_ILLARG );
		}
		if( (stack[i].ls = search_hash( TMP, tsvv(d) )) ){
			stack[i].kind = TMP;
		} else {
			stack[i].kind = 0;
		}
	}
	argnum = i;
	for( s = src, i= 0; i < argnum; i++ ){
		if( s ){
			if( List[s].kind == KLIST ){
				ev[i] = eval_exp(List[s].v);
				s = List[s].next;
			} else {
/* B0				ev[i] = eval_exp( s );*/
				s = NIL;
			}
		} else {
			ev[i] = cons( NIL, KNIL, NIL );
		}
	}
	for( d = dst, i= 0; i < argnum; i++, d = List[d].next ){
		if( Debug_flag ){
			debug_routine( i+1, tsvv(d), ev[i] );
		}
		set_hash( TMP, tsvv(d), ev[i] );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function search_macro()                                */
/*                                                                    */
/*             char *name: Macro argment                              */
/*             Lisp *ev: the pair of Macro-Arg and Value              */
/*                                                                    */
/*             return Lisp: !=NIL:found value  ==NIL:not found        */
/*                                                                    */
/*             function: Search the Macro-arg by name.                */
/*             added B0                                               */
/**********************************************************************/
Lisp search_macro( char *name, Lisp *ev )
{
	int i;
	Lisp ret;

	for( i = 0; i < ARGMENT_MAX; i++ ){
		if( ev[i] == NIL ){
			break;
		}
		if( strcmp( name, tsv(ev[i]) ) == 0 ){
			ret = dup_t2tLisp( List[ev[i]].next );
			push_node( ret, "search_macro" );
			return( ret );
		}
	}
	return( NIL );
}


/**********************************************************************/
/*                                                                    */
/*             function replace_argment()                             */
/*                                                                    */
/*             Lisp dst: Macro Body                                   */
/*             Lisp *ev: Pairs of Macro arg and value.                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Convert Macro-Body by translating V-arg to   */
/*                      R-Arg. (recursive)                            */
/*             added B0                                               */
/**********************************************************************/
void replace_argment( Lisp at, Lisp *ev )
{
	Lisp ret;

	if( at == NIL ){
		return;
	}
	if( List[at].kind == KSYMBOL ){
		if( (ret = search_macro( tsv(at), ev )) != NIL ){
			List[at].v = List[ret].v;
			List[at].kind = List[ret].kind;
			List[at].next = List[ret].next;
			return;
		}
	}
	switch( op_kind( List[at].kind) ){
	case 8:
		replace_argment( List[at].v, ev );
		replace_argment( List[at].next, ev );
		break;
	case 1:
		replace_argment( List[at].next, ev );
		break;
	case 0:
		break;
	default:
		warning( W_ILLKIND,  List[at].kind, "dup_Lisp" );
		break;
	}
	return;
}

/**********************************************************************/
/*                                                                    */
/*             function set_macro_arg()                               */
/*                                                                    */
/*             Lisp body: Macro Body                                  */
/*             Lisp kari: Macro Virtual Argment                       */
/*             Lisp jitu: Macro Real Argment                          */
/*             Lisp *ev:  Pairs of kari and jitu.                     */
/*                                                                    */
/*             return Lisp: expanded Macro-Body                       */
/*                                                                    */
/*             function: Do the Macro-expandion.                      */
/*             added B0                                               */
/**********************************************************************/
Lisp set_macro_arg( Lisp body, Lisp kari, Lisp jitu, Lisp *ev )
{
	Lisp ret;
	Lisp k, j;
	int i;

	ret = dup_t2tLisp( body );
	push_node( ret, "set_macro(body)" );
	for( i = 0, k = kari, j = jitu; ; k = List[k].next, j = List[j].next, i++ ){
		if( i >= ARGMENT_MAX ){
			warning( W_MANYARG );
		}
		if( k == NIL && j == NIL ){
			break;
		}
		if( k == NIL || j == NIL ){
			dwarning( "kari/jitu macro arg not match\n" );
		}
		if( List[k].kind != KLIST || List[j].kind != KLIST ){
			dwarning( "macro arg not LIST\n" );
		}
		if( List[List[k].v].kind != KSYMBOL ){
			dwarning( "kari not SYMBOL\n" );
		}
		if( List[j].kind == KSHARP ){
			j = expand_macro( List[j].v );
		}
		ev[i] = cons( List[List[k].v].v, KSYMBOL, List[j].v );
	}
	ev[i] = NIL;
	replace_argment( ret, ev );
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function get_function_body()                           */
/*                                                                    */
/*             Lisp at: Function name                                 */
/*                                                                    */
/*             return  Lisp: Function definition                      */
/*                                                                    */
/*             function: Cash the function definition.                */
/*                                                                    */
/**********************************************************************/
Lisp get_function_body( Lisp at )
{
	Lisp ls;

	if( (ls = search_hash( TMP, ts(at) )) ){
		return( ls );
	} else if( (ls = search_hash( GLOBAL, ts(at) )) ){
		NowBody = dup_g2tLisp(ls);
		set_hash( TMP, ts(at), NowBody );
		return( NowBody );
	} else {
		warning( W_NOMACRO, ts(at) );
		return( NIL );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function eval_exp()                                    */
/*                                                                    */
/*             Lisp com: Evaluate List                                */
/*                                                                    */
/*             return  Lisp: Evaluated List                           */
/*                                                                    */
/*             function: Evaluate the List.                           */
/*                       (function execution etc.)                    */
/*             rev B0: push_node(), clear_prevstack() for save mem    */
/**********************************************************************/
Lisp eval_exp( Lisp com )
{
	Lisp ret;
	Lisp val, at;
	static struct List RV;
	static jmp_buf fjump;
	jmp_buf prev_jump;
	struct StackStruct stack[ARGMENT_MAX];
	Lisp ev[ARGMENT_MAX];

	if( Error_flag ){
		error_jump();
	}
	ret = NIL;
	switch( List[com].kind ){
	case KFUNC:
		val = List[com].v;
		if( List[val].kind != KSYMBOL ){
			warning( W_NOFUNC );
		}
		ret = get_function_body( List[val].v );
		if( List[ret].kind == KARGS 
		   || List[ret].kind == KMARGS ){
			if( Debug_flag ){
				debug_routine( 0, tsv(val), NIL );
			}
			if( List[ret].kind == KARGS ){
/* B0 */	       		push_node( NIL, tsv(val) );
				set_function_arg( List[ret].v, List[com].next, stack, ev );
				RV.v = RV.next = NIL; RV.kind = KNOUSE;
				memcpy( &prev_jump, &fjump, sizeof(jmp_buf) );
				Func_count++;
				if( setjmp( fjump ) == 0 ){
					exec_command( List[ret].next );
				}
				Func_count--;
				memcpy( &fjump, &prev_jump, sizeof(jmp_buf) );
				restore_arg( List[ret].v, stack, ev );
				if( RV.kind == KNOUSE ){
					ret = cons( NIL, KNIL, NIL );
				} else {
					ret = cons( RV.v, RV.kind, RV.next );
				}
/* B0 */	       		clean_prevstack( ret, tsv(val) );
			} else {
				ret = set_macro_arg( List[ret].next, 
					List[ret].v, List[com].next, ev );
				exec_command( ret );
			}
			if( Debug_flag ){
				debug_routine( -1, tsv(val), ret );
			}
		} else if( List[ret].kind == KCFUNC ){
/* B0 			at = eval_exp( List[com].next );
			ret = (*CFuncTable[List[ret].v].func)(at);*/
/* B0 */		ret = (*CFuncTable[List[ret].v].func)(List[com].next);
/* B0 */		ret = dup_t2tLisp( ret );
		} else {
			warning( W_NOFUNC );
		}
		break;
	case KRET:
		if( Func_count == 0 ){
			warning( W_OUTRET );
			return( NIL );
		}
		ret = eval_exp( List[com].v );
/* B0 add */	ret = dup_t2tLisp( ret );
		RV.v = List[ret].v; RV.kind = List[ret].kind;
		RV.next = List[ret].next;
		longjmp( fjump, 1 );
		break;
	case KCOM:
		ret = com;
		break;
	case KPOSITION:
		at = eval_exp( List[com].v );
		ret = cons( at, KPOSITION, eval_exp( List[com].next ) );
		break;
	case KFN:
	case KCSTACK:
	case KPLUG:
	case KMASK:
	case KRANGE:
	case KFOR:
	case KWHILE:
	case KIF:
	case KLET:
	case KDEFUNC:
	case KARGS:
	case KCFUNC:
	case KHASH:
		warning( W_ILLKIND, List[com].kind, "eval_exp" );
		ret = NIL;
		break;
	default:
		if( List[com].kind < KIND_MAX ){
			ret = do_calc( com );
		} else {
			warning( W_ILLKIND, List[com].kind, "eval_exp" );
			ret = NIL;
		}
		break;
	}
	return( ret );
}


