/**********************************************************************/
/*                                                                    */
/*        comline.c -- gedo display LIST routines                     */
/*                                                                    */
/*                by K.sugiura   Ver 1.0  for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*                                                                    */
/*    void write_command( void *fp, int kind, Lisp com, int flag )    */
/*        function:  display command-line from List                   */
/*        void *fp:  FILE * or char *                                 */
/*        int kind:  == 0 fp is FILE *, == 1 fp is char *             */
/*        Lisp com:  display List                                     */
/*        int flag:  == 0 add NONE, == 1 add \n, ==2 add :            */
/*    rev B0: 1.16,1997 : double LIST {} add                          */ 
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "gedo.h"
#include "prot.h"
#include "error.h"

static char *Operator[]		/* display operator */
 = { " + ", " - ", " * ", " / ", "++", "++", "--", "--", " % ",
	 "~",  " & ", " | ", " == ", " != ", " >= ", " > ", " <= ", " < " };
static void *Fp;	/* (FILE *) or (char *), used nprintf */
int Out_kind = 0;	/*  ==0:write to file, !=0: write to buff */

static void write_exp( Lisp com );
static void write_command_arg( Lisp com );

/**********************************************************************/
/*                                                                    */
/*             function nprintf()                                     */
/*                                                                    */
/*             char *form: format string                              */
/*             ...       : va_arg output argments                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: output to file(Out_kind==0) or buffer(!=0)   */
/*                      lowest level of write.c                       */
/**********************************************************************/
static void nprintf( char *form, ... )
{
	va_list argp;
	char tmp[64];

	va_start( argp, form );
	if( Out_kind == 0 ){
		vfprintf( (FILE *)Fp, form, argp );
	} else {
		vsprintf( tmp, form, argp );
		strcat( (char *)Fp, tmp );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function kakko()                                       */
/*                                                                    */
/*             Lisp com: output List                                  */
/*             int *prev: previous output kind                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Memory previous output paren and re-output.  */
/*                                                                    */
/**********************************************************************/
void kakko( Lisp com, int *prev )
{
	static char *kak[2][3]
		= { { "", "(" }, { "", ")" } };

	if( *prev ){
		nprintf( kak[1][*prev - 1] );
		*prev = 0;
		return;
	}
	switch( Kind_spec[List[com].kind] & 0xf0){
	case KSBASIC:
		*prev = 1;
		break;
	case KSFOP:
	case KSCOP:
		*prev = 2;
		break;
	default:
		switch( List[com].kind ){
		case KPOSLIST:
		case KFUNC:
			*prev = 1;
			break;
		default:
			return;
			break;
		}
		break;
	}
	nprintf( kak[0][*prev - 1] );
}

/**********************************************************************/
/*                                                                    */
/*             function write_term()                                  */
/*                                                                    */
/*             Lisp com: output List(terminal elements)               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Output terminal elements.                    */
/*                                                                    */
/**********************************************************************/
void write_term( Lisp com )
{
	int prev;

	if( com == NIL ){
		return;
	}
#if DEBUG
	printf( "wt%02x:",List[com].kind );
#endif
	if( (Kind_spec[List[com].kind] & 0x0f) == 1 ){
		if( List[com].v == NIL ){
			return;
		}
		if( List[com].kind == KLITERAL ){
			nprintf( "\"%s\"", tsv(com) );
		} else {
			nprintf( "%s", tsv(com) );
		}
	} else {
		switch( List[com].kind ){
		case KNUMBER:
			nprintf( "%d", List[com].v );
			break;
		case KCOM:
			write_command_arg( com );
			break;
		case KFN:
			nprintf( "%s", FnStr[List[com].v] );
			break;
		case KCSTACK:
			nprintf( "stack" );
			break;
		case KNIL:
			nprintf( "NIL" );
			break;
		case KMASK:
			nprintf( "mask%d", List[com].v );
			break;
		default:
			warning( W_ILLKIND, List[com].kind, "write_term" );
			break;
		}
	}
	if( List[com].next ){
		nprintf( " " );
		prev = 0;
		kakko( List[com].next, &prev );
		write_exp( List[com].next );
		kakko( List[com].next, &prev );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function write_command_arg()                           */
/*                                                                    */
/*             Lisp com: Output List(COMMAND)                         */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Output symbol after COMMAND.(special patterm)*/
/*                                                                    */
/**********************************************************************/
void write_command_arg( Lisp com )
{
	Lisp next, aft;

	nprintf( "%s ",  ComStr[List[com].v - 1] );
	next = List[com].next;
	if( next == NIL ){
		return;
	}
	switch( List[com].v ){
	case MPRINT:
		if( List[next].kind == KLIST ){
			nprintf( "{" );
			write_exp( next );
			nprintf( "}" );
			return;
		}
		break;
	case MLOADER:
	case MSAVER:
	case MSELECTER:
	case MEFFECTER:
		if( List[next].kind == KFN ){
			nprintf( "help" );
			return;
		}
		nprintf( "%s ", tsv(next) );
		aft = List[next].next;
		if( aft == NIL ){
			return;
		}
		if( List[aft].kind == KFN ){
			nprintf( "help" );
		} else if( List[aft].kind == KVOID ){
			if( List[aft].next == NIL ){
				nprintf( "config" );
				return;
			}
			write_exp( List[aft].next );
		} else {
			write_exp( aft );
		}
		return;
		break;
	case MLOAD:
	case MSAVE:
		if( List[next].kind == KVOID ){
			return;
		}
		break;
	case MSELECT:
		if( List[next].kind == KVOID ){
			write_exp( List[next].next );
			return;
		}
		break;
	}
	write_exp( next );
}

/**********************************************************************/
/*                                                                    */
/*             function write_exp()                                   */
/*                                                                    */
/*             Lisp com: Output List(calculation)                     */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Output formula.                              */
/*                                                                    */
/**********************************************************************/
void write_exp( Lisp com )
{
	int prev;

	if( com == NIL ){
		return;
	}
#if DEBUG
	printf( "we%02x:",List[com].kind );
#endif
	prev = 0;
	switch( Kind_spec[List[com].kind] & 0xf0 ){
	case KSBASIC:
		write_term( com );
		break;
	case KSFOP:
		switch( List[com].kind ){
		case KINCL:
		case KDECL:
			nprintf( "%s", Operator[List[com].kind - KPLUS] );
			write_term( List[com].v );
			break;
		case KINCR:
		case KDECR:
			write_term( List[com].v );
			nprintf( "%s", Operator[List[com].kind - KPLUS] );
			break;
		case KDNOT:
			nprintf( "%s", Operator[List[com].kind -KPLUS] );
			kakko( List[com].v, &prev );
			write_exp( List[com].v );
			kakko( List[com].v, &prev );
			break;
		default:
			kakko( List[com].v, &prev );
			write_exp( List[com].v );
			kakko( List[com].v, &prev );
			nprintf( "%s", Operator[List[com].kind - KPLUS] );
			kakko( List[com].next, &prev );
			write_exp( List[com].next );
			kakko( List[com].next, &prev );
			break;
		}
		break;
	case KSCOP:
		kakko( List[com].v, &prev );
		write_exp( List[com].v );
		kakko( List[com].v, &prev );
		nprintf( "%s", Operator[List[com].kind - KPLUS] );
		kakko( List[com].next, &prev );
		write_exp( List[com].next );
		kakko( List[com].next, &prev );
		break;
	case KSLIST:
/* B0 add */
		if( List[List[com].v].kind == KLIST ){
			nprintf( "{" );
		}
/* B0 end */
		write_exp( List[com].v );
/* B0 add */
		if( List[List[com].v].kind == KLIST ){
			nprintf( "}" );
		}
/* end B0 */
		if( List[com].kind != KPOSLIST && List[com].next ){
			nprintf( ", " );
		}
		write_exp( List[com].next );
		break;
	case KSCOMPLEX:
		switch( List[com].kind ){
		case KCOM:
			write_command_arg( com );
			break;
		case KPOSITION:
			nprintf( "(" );
			write_exp( List[com].v );
			nprintf( ", " );
			write_exp( List[com].next );
			nprintf( ")" );
			break;
		case KRANGE:
			write_term( List[com].v );
			nprintf( " " );
			write_term( List[com].next );
			break;
		default:
			warning( W_ILLKIND, List[com].kind, "write_exp" );
			break;
		}
		break;
	case KSTOP:
		if( List[com].kind == KRET ){
			nprintf( "return(" );
			write_exp( List[com].v );
			nprintf( ")" );
		} else if( List[com].kind == KFUNC ){
			write_term( List[com].v );
			nprintf( "(" );
			write_exp( List[com].next );
			nprintf( ")" );
		} else {
			warning( W_ILLKIND, List[com].kind, "write_exp" );
		}
		break;
	default:
		warning( W_ILLKIND, List[com].kind, "write_exp" );
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function write_statement()                             */
/*                                                                    */
/*             Lisp com: output List(statement)                       */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Output statements                            */
/*                                                                    */
/**********************************************************************/
void write_statement( Lisp com )
{
	if( com == NIL ){
		return;
	}
#if DEBUG
	printf( "ws%02x:",List[com].kind );
#endif
	switch( Kind_spec[List[com].kind] & 0xf0 ){
	case KSBASIC:
		write_term( com );
		break;
	case KSCOMPLEX:
		switch( List[com].kind ){
		case KWHILE:
			nprintf( "while(" );
			write_statement( List[com].v );
			nprintf( ")" );
			if( List[List[com].next].kind == KSHARP ){
				write_term( List[com].next );
			} else {
				nprintf( "{" );
				write_command( List[com].next );
				nprintf( "}" );
			}
			break;
		case KFOR:
			nprintf( "for( " );
			write_statement( List[List[com].v].v );
			nprintf( "; " );
			write_statement( List[List[List[com].v].next].v );
			nprintf( "; " );
			write_statement( List[List[List[com].v].next].next );
			nprintf( ")" );
			if( List[List[com].next].kind == KSHARP ){
				write_term( List[com].next );
			} else {
				nprintf( "{" );
				write_command( List[com].next );
				nprintf( "}" );
			}
			break;
		case KIF:
			nprintf( "if( " );
			write_statement( List[com].v );
			nprintf( ")" );
			if( List[List[List[com].next].v].kind == KSHARP ){
				write_term( List[List[com].next].v );
			} else {
				nprintf( "{" );
				write_command( List[List[com].next].v );
				nprintf( "}" );
			}
			if( List[List[com].next].next == NIL ){
				break;
			}
			nprintf( " else " );
			if( List[List[List[com].next].next].kind == KSHARP ){
				write_term( List[List[com].next].next );
			} else {
				nprintf( "{" );
				write_command( List[List[com].next].next );
				nprintf( "}" );
			}
			break;
		case KCOM:
		case KPOSITION:
		case KRANGE:
			write_exp( com );
			break;
		default:
			warning( W_ILLKIND, List[com].kind, "write_statement" );
			break;
		}
		break;
	case KSLIST:
		if( List[com].kind == KPOSLIST ){
			write_exp( com );
			break;
		}
		for( ;com != NIL; com = List[com].next ){
			if( List[com].kind != KLIST ){
				write_statement( com );
				break;
			} else if( List[List[com].v].kind == KLIST ){
				nprintf( "{" );
				write_statement( List[com].v );
				nprintf( "}" );
			} else {
				write_statement( List[com].v );
			}
			if( List[com].next != NIL ){
				nprintf( ", " );
			}
		}
		break;
	case KSLET:
		write_term( List[com].v );
		if( List[List[com].next].kind == KLIST ){
			nprintf( " ={" );
			write_statement( List[com].next );
			nprintf( "}" );
		} else {
			nprintf( " = " );
			write_statement( List[com].next );
		}
		break;
	case KSFOP:
	case KSCOP:
	case KSTOP:
		write_exp( com );
		break;
	default:
		warning( W_ILLKIND, List[com].kind, "write_statement" );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function write_command()                               */
/*                                                                    */
/*             Lisp com: output List(toplevel)                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Output Toplevel List.                        */
/*                                                                    */
/**********************************************************************/
void write_command( Lisp com )
{
	Lisp at;

	if( com == NIL ){
		return;
	}
#if DEBUG
	printf( "wc:%02x",List[com].kind );
#endif
	if( List[com].kind == KLET && List[List[com].v].kind == KSHARP ){
		write_term( List[com].v );
		nprintf( " ={" );
		write_command( List[com].next );
		nprintf( "}" );
	} else if( List[com].kind == KLIST ){
		write_statement( com );
	} else if( List[com].kind == KDEFUNC || List[com].kind == KDEMACRO ){
		if( List[com].kind == KDEFUNC ){
			nprintf( "defun " );
		} else {
			nprintf( "demacro " );
		}
		write_term( List[com].v );
		nprintf( "(" );
		at = List[com].next;
		write_statement( List[at].v );
		if( !Out_kind ){
			nprintf( "){\n" );
			write_statement( List[at].next );
			nprintf( "\n}\n" );
		} else {
			nprintf( "){" );
			write_statement( List[at].next );
			nprintf( "}" );
		}
	} else {
		write_statement( com );
	}
	return;
}

/**********************************************************************/
/*                                                                    */
/*             function display_command()                             */
/*                                                                    */
/*             Lisp line: output List                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Output to stdout.(wrapper for write_command) */
/*                                                                    */
/**********************************************************************/
void display_command( Lisp line )
{
	if( File_flag && check_value( GLOBAL, "#silent" ) ){
		return;
	}
	Fp = stdout; Out_kind = 0;
	write_command( line );
	nprintf( "\n" );
}

/**********************************************************************/
/*                                                                    */
/*             function write_command_file()                          */
/*                                                                    */
/*             FILE *fp: File pointer                                 */
/*             Lisp line: output List                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Output to file(*fp).                         */
/*                      wrapper for write_command()                   */
/**********************************************************************/
void write_command_file( FILE *fp, Lisp line )
{
	Fp = fp; Out_kind = 0;
	write_command( line );
	nprintf( "\n" );
}

/**********************************************************************/
/*                                                                    */
/*             function set_command_buf()                             */
/*                                                                    */
/*             char *work: output buffer(dst)                         */
/*             Lisp line:  output List(src)                           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Output to buffer. Used to set Ringbuffer(hist.c)*/
/*                      wrapper for write_command()                   */
/**********************************************************************/
void set_command_buf( char *work, Lisp line )
{
	Fp = work; Out_kind = 1;
	write_command( line );
}

/**********************************************************************/
/*                                                                    */
/*             function write_of_print()                              */
/*                                                                    */
/*             Lisp com: output List                                  */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: PRINT command calles this.(adds spec)        */
/*                                                                    */
/**********************************************************************/
void write_of_print( Lisp com )
{
	Fp = stdout; Out_kind = 0;
	if( com == NIL ){
		nprintf( "not bounded\n" );
		return;
	}
	switch( List[com].kind ){
	case KNUMBER:
		nprintf( "variable:integer %d\n", List[com].v );
		break;
	case KLITERAL:
		nprintf( "variable:literal \"%s\"\n", tsv(com) );
		break;
	case KPOSITION:
	case KPOSLIST:
		nprintf( "variable:position " );
		write_exp( dup_poslist(com) );
		nprintf( "\n" );
		break;
	case KLIST:
		nprintf( "list: {" );
		write_statement( com );
		nprintf( "}\n" );
		break;
	case KCOM:
		nprintf( "command: " );
		write_statement( com );
		nprintf( "\n" );
		break;
	case KCFUNC:
		nprintf( "builtin function\n" );
/* B2 */
		nprintf( "%s\n", get_cfunc_comment(List[com].v) );
		break;
	case KDEFUNC:
		nprintf( "user function\n" );
		write_command( com );
		break;
	case KDEMACRO:
		nprintf( "user macro\n" );
		write_command( com );
		break;
	case KPLUG:
		nprintf( "plugin\n" );
		nprintf( "%s\n", plug_info( List[com].v ) );
		break;
	case KNIL:
		nprintf( "NIL\n" );
		break;
	case KSYMBOL:
	case KSHARP:
		nprintf( "symbol :" );
		write_command( com );
		nprintf( "\n" );
		break;
	case KMASK:
		nprintf( "mask : id=%d\n", List[com].v );
		break;
	default:
		write_statement( com );
		nprintf( "\n" );
		break;
	}
}
