/**********************************************************************/
/*                                                                    */
/*           fspline.c  fpline function library                       */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*    External functions:
      	struct fPosition *spline_line( struct fPosition *pos, int dim, 
		int num, int pat );
	struct fPosition *spline_poly( struct fPosition *pos, int dim, 
		int num, int pat );                                   */
/*                                                                    */
/**********************************************************************/
/******** KOMAKAI KOTO HA YOKU WAKARANAI !  **************************/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <memory.h>

#define IDM	100
#define JDM	3
#define NPM	10

#define IPM	NPM*IDM+1
#define MDM	IDM+JDM*2

struct fPosition {
	float x;
	float y;
};

static float B[IDM][JDM+1];
static float Bc[IDM];
static float Q[MDM];
static struct fPosition XY[IPM];	/* result */

#define LINE	0
#define POLY	1

struct fPosition *spline_line( struct fPosition *pos, int dim, int num, int pat );
struct fPosition *spline_poly( struct fPosition *pos, int dim, int num, int pat );
/**********************************************************************/
/*                                                                    */
/*             function alloc_works()                                 */
/*                                                                    */
/*             int dim : Dimension of spline                          */
/*             int num : Number of points                             */
/*             int pat : Partition of line                            */
/*                                                                    */
/*             return  int : 0==OK   -1==too large                    */
/*                                                                    */
/*             function: check the parameter with static array.       */
/*                                                                    */
/**********************************************************************/
static int alloc_works( int dim, int num, int pat )
{
	if( dim > JDM ){
		return( -1 );
	}
	if( num > IDM ){
		return( -1 );
	}
	if( pat > NPM ){
		return( -1 );
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function setten_line()                                 */
/*                                                                    */
/*             int dim  : Dimemsion                                   */
/*             int num  : Number of points                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: At spline_line(), get 1st setten.            */
/*                                                                    */
/**********************************************************************/
static void setten_line( int dim, int num )
{
	int i;

	for( i = 0; i < dim; i++ ){
		Q[i] = 0.0;
	}
	for( i = 0; i < num - dim; i++ ){
		Q[i+dim] = (float)i + (float)dim / 2.0;
	}
	for( i = 0; i < dim; i++ ){
		Q[i+num] = (float)(num - 1);
	}
}

/**********************************************************************/
/*                                                                    */
/*             function setten_poly()                                 */
/*                                                                    */
/*             int dim  : Dimemsion                                   */
/*             int num  : Number of points                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: At spline_poly(), get 1st setten.            */
/*                                                                    */
/**********************************************************************/
static void setten_poly( int dim, int num )
{
	int i;

	for( i = 0; i < num + 2 * dim; i++ ){
		Q[i] = (float)(i - dim);
	}
}

/**********************************************************************/
/*                                                                    */
/*             function deboor_line()                                 */
/*                                                                    */
/*             int dim : Dimension                                    */
/*             int num : Number of points                             */
/*             float t : Deboor-t                                     */
/*                                                                    */
/*             return  int : Deboor function value                    */
/*                                                                    */
/*             function: At spline_line(), get Deboor function.       */
/*                                                                    */
/**********************************************************************/
static int deboor_line( int dim, int num, float t )
{
	int i, k;
	int ret;
	float a1, a2;

	memset( B, 0, sizeof(B) );

	ret = 0;
	for( i = 0; i < num; i++ ){
		if( Q[i] <= t && t < Q[i+1]){
			B[i][0] = 1;
			ret = i;
		}
	}
	if( Q[num-1] <= t && t <= Q[num] + 0.00001 ){
		B[num-1][0] = 1;
		ret = num - 1;
	}
	for( k = 1; k < dim + 1; k++ ){
		for( i = ret - k; i <= ret; i++ ){
			a1 = a2 = 0.0;
			if( Q[i+1] != Q[i+k+1] ){
				a1 = (Q[i+k+1] - t) * B[i+1][k-1] / (Q[i+k+1] - Q[i+1]);
			}
			if( Q[i] != Q[i+k] ){
				a2 = (t - Q[i]) * B[i][k-1] / (Q[i+k] - Q[i]);
			}
			B[i][k] = a1 + a2;
		}
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function deboor_poly()                                 */
/*                                                                    */
/*             int dim : Dimension                                    */
/*             int num : Number of points                             */
/*             float t : Deboor-t                                     */
/*                                                                    */
/*             return  int : Deboor function value                    */
/*                                                                    */
/*             function: At spline_poly(), get Deboor function.       */
/*                                                                    */
/**********************************************************************/
static int deboor_poly( int dim, int num, float t )
{
	int i, k;
	int ret;
	float a1, a2;

	memset( B, 0, sizeof(B) );

	ret = 0;
	for( i = 0; i < num+dim; i++ ){
		if( Q[i] <= t && t < Q[i+1]){
			B[i][0] = 1;
			ret = i;
		}
	}
	if( Q[num+dim] <= t && t <= Q[num+dim+1] + 0.00001 ){
		B[num+dim][0] = 1;
		ret = num+dim;
	}
	for( k = 1; k < dim + 1; k++ ){
		for( i = ret - k; i <= ret; i++ ){
			a1 = a2 = 0.0;
			if( Q[i+1] != Q[i+k+1] ){
				a1 = (Q[i+k+1] - t) * B[i+1][k-1] / (Q[i+k+1] - Q[i+1]);
			}
			if( Q[i] != Q[i+k] ){
				a2 = (t - Q[i]) * B[i][k-1] / (Q[i+k] - Q[i]);
			}
			B[i][k] = a1 + a2;
		}
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*             function shuki_spline()                                */
/*                                                                    */
/*             int dim : Dimension                                    */
/*             int num : Number of points                             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: At spline_poly(), set close data.            */
/*                                                                    */
/**********************************************************************/
static void shuki_spline( int dim, int num )
{
	int i;
	int mjis;

	mjis = (dim + 1) / 2;
	for( i = 0; i < mjis; i++ ){
		Bc[i] = B[i+mjis-1][dim] + B[i+num+mjis-1][dim];
	}
	for( i = 0; i < num - dim; i++ ){
		Bc[i+mjis] = B[i+dim][dim];
	}
	for( i = 0; i < mjis - 1; i++ ){
		Bc[i+num-mjis+1] = B[i+num][dim] + B[i][dim];
	}
}

/**********************************************************************/
/*                                                                    */
/*             function *kansuti_line()                               */
/*                                                                    */
/*             struct fPosition *org: original data                    */
/*             int dim : Dimension                                    */
/*             int num : Number of points                             */
/*             int pat : Partition                                    */
/*                                                                    */
/*             return  struct fPosition* :result of spline transfar   */
/*                                                                    */
/*             function: Main process of spline_line().               */
/*                                                                    */
/**********************************************************************/
static struct fPosition *kansuti_line( struct fPosition *org, int dim, int num, int pat )
{
	int i, j;
	int ret;
	float t;

	for( i = 0; i < pat * (num-1) + 1; i++ ){
		t = (float)(i+1) / (float)pat;
		ret = deboor_line( dim, num, t );
		XY[i].x = XY[i].y = 0.0;
		for( j = ret - dim; j <= ret; j++ ){
			XY[i].x += (org[j].x * B[j][dim]);
			XY[i].y += (org[j].y * B[j][dim]);
		}
	}
	return( XY );
}

/**********************************************************************/
/*                                                                    */
/*             function *kansuti_poly()                               */
/*                                                                    */
/*             struct fPosition *org: original data                    */
/*             int dim : Dimension                                    */
/*             int num : Number of points                             */
/*             int pat : Partition                                    */
/*                                                                    */
/*             return  struct fPosition *:result of spline transfar   */
/*                                                                    */
/*             function: Main process of spline_poly().               */
/*                                                                    */
/**********************************************************************/
static struct fPosition *kansuti_poly( struct fPosition *org, int dim, int num, int pat )
{
	int i, j;
	int ret, ist, isp;
	float t;
	int mjis;

	mjis = (dim + 1) / 2;
	for( i = 0; i < pat * num + 1; i++ ){
		t = (float)i / (float)pat;
		ret = deboor_poly( dim, num, t );
		shuki_spline( dim, num );
		isp = ret - mjis + 1;
		ist = isp - dim;
		if( ist < 0 || isp > num - 1 ){
			ist = 0;
			isp = num - 1;
		}
		XY[i].x = XY[i].y = 0.0;
		for( j = ist; j <= isp; j++ ){
			XY[i].x += (org[j].x * Bc[j]);
			XY[i].y += (org[j].y * Bc[j]);
		}
	}
	return( XY );
}

/**********************************************************************/
/*                                                                    */
/*             function  spline_line()                                */
/*                                                                    */
/*             struct fPosition *org: original data                    */
/*             int dim : Dimension                                    */
/*             int num : Number of points                             */
/*             int pat : Partition                                    */
/*                                                                    */
/*             return  struct fPosition*: result of spline transfar   */
/*                                                                    */
/*             function: Interface of LINE spline.                    */
/*                                                                    */
/**********************************************************************/
struct fPosition *spline_line( struct fPosition *pos, int dim, int num, int pat )
{
	if( alloc_works( dim, num, pat ) != 0 ){
		return( NULL );
	}
	setten_line( dim+1, num );
	return( kansuti_line( pos, dim, num, pat ) );
}

/**********************************************************************/
/*                                                                    */
/*             function  spline_poly()                                */
/*                                                                    */
/*             struct fPosition *org: original data                    */
/*             int dim : Dimension                                    */
/*             int num : Number of points                             */
/*             int pat : Partition                                    */
/*                                                                    */
/*             return  struct fPosition*: result of spline transfar   */
/*                                                                    */
/*             function: Interface of POLY spline.                    */
/*                                                                    */
/**********************************************************************/
struct fPosition *spline_poly( struct fPosition *pos, int dim, int num, int pat )
{
	if( alloc_works( dim, num, pat ) != 0 ){
		return( NULL );
	}
	setten_poly( dim, num );
	return( kansuti_poly( pos, dim, num, pat ) );
}


#ifdef  TESTLINE

#include <graph.h>
#define circle(x,y) _ellipse( _GBORDER, x-3, y-3, x+3, y+3 );

/**********************************************************************/
/*                                                                    */
/*             function plot_line()                                   */
/*                                                                    */
/*             struct fPosition *org:                                  */
/*             struct fPosition *xy:                                  */
/*             int num             :                                  */
/*             int pat             :                                  */
/*                                                                    */
/*             return NONE                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
void plot_line( struct fPosition *org, struct fPosition *xy, int num, int pat )
{
	int i;
	short cursor;

	cursor = _gettextcursor();
	_setvideomode( _DEFAULTMODE );
	_setcolor( 16 );
	_setbkcolor( 0 );
	_clearscreen( _GCLEARSCREEN );

	_moveto( (int)(xy[0].x*40.0), (int)(xy[0].y*40.0) );
	for( i = 1; i < pat * (num-1); i++ ){
		_lineto( (int)(xy[i].x*40.0), (int)(xy[i].y*40.0) );
		_moveto( (int)(xy[i].x*40.0), (int)(xy[i].y*40.0) );
	}
	for( i = 0; i < num; i++ ){
		circle( (int)(org[i].x*40.0), (int)(org[i].y*40.0) );
	}
	getchar();
	_clearscreen( _GCLEARSCREEN );
	_setvideomode( _TEXT );
	_settextcursor( cursor );
}


/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             int argc                                               */
/*             char **argv                                            */
/*                                                                    */
/*             return NONE                :                           */
/*                                                                    */
/*             function:                                              */
/*                                                                    */
/**********************************************************************/
void main( int argc, char **argv )
{
	struct fPosition *xy;
	struct fPosition P[7] = { {0,0}, {3,0}, {3,1}, {0.5,2},
				   {0.5,6}, {5,6}, {5,9} };
	int kosu = 7;
	int kubun = 10;
	int jisu = 6;

	if( argc >= 2 ){
		jisu = atoi( argv[1] );
		if( jisu == 0 ){
			printf( "spline Jisu\n" );
			return;
		}
	}
	if( argc >= 3 ){
		kosu = atoi( argv[2] );
		if( kosu == 0 ){
			printf( "spline Jisu Kosu\n" );
			return;
		}
	}
	if( argc >= 4 ){
		kubun = atoi( argv[3] );
		if( kubun == 0 ){
			printf( "spline Jisu Kosu Kubun\n" );
			return;
		}
	}
	xy = spline_line( P, jisu, kosu, kubun );
	plot_line( P, XY, kosu, kubun );
}
#endif

#ifdef TESTPOLY
#include <graph.h>
#define circle(x,y) _ellipse( _GBORDER, x-3, y-3, x+3, y+3 );

void plot_poly( struct fPosition *org, struct fPosition *xy, int num, int pat )
{
	int i;
	short cursor;

	cursor = _gettextcursor();
	_setvideomode( _DEFAULTMODE );
	_setcolor( 16 );
	_setbkcolor( 0 );
	_clearscreen( _GCLEARSCREEN );

	_moveto( (int)(xy[0].x*40.0), (int)(xy[0].y*40.0) );
	for( i = 1; i < pat * num; i++ ){
		_lineto( (int)(xy[i].x*40.0), (int)(xy[i].y*40.0) );
		_moveto( (int)(xy[i].x*40.0), (int)(xy[i].y*40.0) );
	}
	_lineto( (int)(xy[0].x*40.0), (int)(xy[0].y*40.0) );
	for( i = 0; i < num; i++ ){
		circle( (int)(org[i].x*40.0), (int)(org[i].y*40.0) );
	}
	getchar();
	_clearscreen( _GCLEARSCREEN );
	_setvideomode( _TEXT );
	_settextcursor( cursor );
}


void main( int argc, char **argv )
{
	struct fPosition *xy;
	struct fPosition P[8] = { {1,5}, {2,2}, {5,1}, {8,2},
				 {9,5}, {8,8}, {5,9}, {2,8} };

	int kosu = 8;
	int kubun = 10;
	int jisu = 3;

	if( argc >= 2 ){
		jisu = atoi( argv[1] );
		if( jisu == 0 ){
			printf( "spline Jisu\n" );
			return;
		}
	}
	if( jisu % 2 == 0 ){
		printf( "Jisu must be odd\n" );
		return;
	}
	if( argc >= 3 ){
		kosu = atoi( argv[2] );
		if( kosu == 0 ){
			printf( "spline Jisu Kosu\n" );
			return;
		}
	}
	if( argc >= 4 ){
		kubun = atoi( argv[3] );
		if( kubun == 0 ){
			printf( "spline Jisu Kosu Kubun\n" );
			return;
		}
	}
	xy = spline_poly( P, jisu, kosu, kubun );
	plot_poly( P, XY, kosu, kubun );
}
#endif
