/**********************************************************************/
/*                                                                    */
/*           image.c    viewer display image sections.                */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*           created Ver 0.92  from loaderx.c                         */
/*                                                                    */
/*           External Interfaces:
	int init_16bit( void );
	int init_24bit( void );
	void init_scale( struct WinSize *win );
	void put_image( struct Shm *buff, int kind );
	void free_masks( void );
	int  set_new_winsize( int kind, struct WinSize *win, int num1, 
				int num2 );
	void move_display( int kind, int num );
	void change_view( int kind, unsigned char *s );
	void position_r2w( struct Positions *p );
	void position_w2r( struct Positions *p );
	int get_scale_rate( int kind );				      */
/*      C1: *2 /2 size change lost the translation                    */
/*	C2: add get_scale(), for get scale rate from shower           */
/*      C3: gray area for not effective window area                   */
/*      C4: changed for memory allocate economy                       */
/**********************************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "viewer.h"

extern Display *D;		/* original Display from loaderx.c */
extern Window   W;		/* main Window from loaderx.c */
extern GC	 gc;		/* GC of strings from loaderx.c */
extern GC	 gcy;		/* GC of selection from loaderx.c */
extern XImage  *Image;		/* XImage of the image from loaderx.c */
extern Visual  *Vis;		/* default Visual from loaderx.c */
extern char *strsts[2];		/* string of the Mode Button */

extern int SelectStatus;	/* kind of Selection, defined in ../loader.h
					#define ADD       0
					#define CHANGE    1
					#define EDIT      1	*/

extern int EventStatus;		/* kind of the Send Event, in ../loader.h
					#define ESDISPLAY   0
					#define ESMASK      1
					#define ESSELECT    2
					#define ESKILL      4
					#define ESCONFIG    5 */
extern int NowMask;		/* Mask Sequence No.(not used) */

struct Scale {			/* description of the current image */
	int style;		/* current style of WinSize.style */
	int scale1;		/* ratio of Loope, scale1/scale2 */
	int scale2;
	int xstart;		/* Now display area No. of X-axis */
	int ystart;		/* Now display area No. of Y-axis */

/* delete this entry */
	int image_width;	/* ==window_org_width */
	int image_height;

	int window_org_width;	/* Image original Size(Width) == buf->width */
	int window_org_height;  /* Image original Size(Height)==buf->height */
	int window_width;	/* Now showing Window(W) width */
	int window_height;	/* Now showing Window(W) height */
	int window_xstart;	/* Now showing start point in buf->data */
	int window_ystart;	/* Now showing start point in buf->data */
};
static struct Scale Scale;

static int	 FullColor;	/* ==1:3byte Color, ==0:2byte Color */
static int 	 Rshift, 	/* In 2byte Color, shift count for Red pixel */
		 Gshift, 	/* Shift count for Green pixel */
		 Bshift;	/* Shift count for Blue pixel */
static unsigned int *Shadow;	/* Shadow Image buffer for display */
static int ShadowSize;		/* size of Shadow */
static int ClearColor;		/* Color of clear window add C3 */
#define BACKGROUND  0x00808080
extern unsigned int MaskColor;	/* Selection Color(default=0xffff00) */

/**********************************************************************/
/*                                                                    */
/*            function set_maskshift()                                */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: From Visual, get each Shift-Count of RGB for  */
/*                                     2 byte color class.            */
/**********************************************************************/
static void set_maskshift( void )
{
	int i;
	signed int tmp;

        tmp = (signed int)Vis->red_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Rshift = i - 8;
	tmp = (signed int)Vis->green_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Gshift = i - 16;
	tmp = (signed int)Vis->blue_mask;
	for( i = 0; i < 32; i++ ){
		if( (tmp << i) < 0 ){
			break;
		}
	}
	Bshift = i - 24;
}

/**********************************************************************/
/*                                                                    */
/*            function init_16bit()                                   */
/*                                                                    */
/*            args  int size  : Shadow buffer pixel size              */
/*                                                                    */
/*            return int   :0=success  -1=malloc error                */
/*                                                                    */
/*            function: In 2byte Color Class, set ShiftCount and      */
/*                         malloc Shadow Image Buffer.                */
/**********************************************************************/
int init_16bit( int size )
{
	static unsigned int change_24to16( unsigned int v );

	FullColor = 0;
	set_maskshift();
	ShadowSize = 10;
/* C4 change size * sizeof(unsigned int); */ 
	if( (Shadow = (unsigned int *)malloc( ShadowSize )) == NULL ){
		return( -1 );
	}
	ClearColor = change_24to16(BACKGROUND); /* add C3 */
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function init_24bit()                                   */
/*                                                                    */
/*            args int size  : Shadow buffer pixel size               */
/*                                                                    */
/*            return int   :  0=success  -1=malloc error              */
/*                                                                    */
/*            function:    In 3byte Color Class, malloc Shadow Image  */
/*                           Buffer.                                  */
/**********************************************************************/
int init_24bit( int size )
{
	FullColor = 1;
	ShadowSize = 10;
/* C4 change size * sizeof(unsigned int); */ 
	if( (Shadow = (unsigned int *)malloc( ShadowSize )) == NULL ){
		return( -1 );
	}
	ClearColor = BACKGROUND;	/* add C3 */
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function change_24to16()                                */
/*                                                                    */
/*            args  unsigned int v   : 3byte Color Class Color Pixel  */
/*                                                                    */
/*            return  unsigned int   : 2byte Color Class Color Pixel  */
/*                                                                    */
/*            function: Convert the expression of Color Pixel         */
/*                      from 3byte to 2byte.                          */
/*                      Interface for realsize(),zoomup(),zoomback()  */
/**********************************************************************/
static unsigned int change_24to16( unsigned int v )
{
	return( (((v >> Rshift) & Vis->red_mask )
	      | ((v >> Gshift) & Vis->green_mask )
	      | ((v >> Bshift) & Vis->blue_mask )) );
}

/**********************************************************************/
/*                                                                    */
/*            function no_change()                                    */
/*                                                                    */
/*            args  unsigned int v   : Color Pixel                    */
/*                                                                    */
/*            return  unsigned int   : Color Pixel equal v            */
/*                                                                    */
/*            function: Dummy Interface of realsize(),zoomup(),zoomback()*/
/*                                                                    */
/**********************************************************************/
static unsigned int no_change( unsigned int v )
{
	return( v );
}

/**********************************************************************/
/*                                                                    */
/*            function depthmask_change()                             */
/*                                                                    */
/*            args  unsigned int v   : Color Pixel                    */
/*                                                                    */
/*            return  unsigned int   : DepthMask Color expression     */
/*                                                                    */
/*            function: In DepthMask Display Mode, convert Pixel      */
/*                      to DepthMask Density.                         */
/*                      Interface for realsize(),zoomup(),zoomback()  */
/**********************************************************************/
static unsigned int depthmask_change( unsigned int v )
{
	v = (v >> 24);
	return( ((v << 16) | (v << 8) | v) & MaskColor );
}

/**********************************************************************/
/*                                                                    */
/*            function simplemask_change()                            */
/*                                                                    */
/*            args  unsigned int v   : Color Pixel                    */
/*                                                                    */
/*            return  unsigned int   : SimpleMask Color expression    */
/*                                                                    */
/*            function: In SimpleMask Display Mode, convert Pixel     */
/*                      to XOR of MaskColor if it is ON-MASK.         */
/*                      Interface for realsize(),zoomup(),zoomback()  */
/**********************************************************************/
static unsigned int simplemask_change( unsigned int v )
{
	static int maskp[] = { 0x80000000, 0x40000000, 0x20000000, 0x10000000,
			      0x08000000, 0x04000000, 0x02000000, 0x01000000 };

	if( (v & maskp[NowMask]) == 0 ){
		return( v ^ MaskColor );
	} else {
		return( v );
	}
}


/**********************************************************************/
/*                                                                    */
/*            function init_scale()                                   */
/*                                                                    */
/*            args  struct WinSize *win   :  Now Display Style        */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: In WIN_INIT and WIN_LOAD, initialize Scale.   */
/*                      Else reset Scale by WinSize *win.             */
/**********************************************************************/
void init_scale( struct WinSize *win )
{
	Scale.style = win->style;
	switch( win->command ){
	case WIN_INIT:
	case WIN_LOAD:
		Scale.scale1 = Scale.scale2 = 1;
		Scale.xstart = Scale.ystart = 0;
		Scale.window_xstart = Scale.window_ystart = 0;
		Scale.image_width = win->image.x;
		Scale.image_height = win->image.y;
		Scale.window_org_width = win->image.x;
		Scale.window_org_height = win->image.y;
		Scale.window_width = win->image.x;
		Scale.window_height = win->image.y;
		break;
	}
}

static unsigned int (*change1)( unsigned int );
static unsigned int (*change2)( unsigned int );


/**********************************************************************/
/*                                                                    */
/*            function real_size()                                    */
/*                                                                    */
/*            args  struct Shm *buf   : The Image Data                */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: If scale1==scale2==1, it is realsize display. */
/*                      The Display routine copies buf->data to Shadow.*/
/**********************************************************************/
static void real_size( struct Shm *buf )
{
	int i, j;
	unsigned int *p, *q;
	int sx, sy, ex, ey;

	sx = (Scale.window_width / 2) * Scale.xstart;
	sy = (Scale.window_height / 2) * Scale.ystart;
	ex = sx + Scale.window_width;
	ey = sy + Scale.window_height;
	if( ex >= buf->width ){
		ex = buf->width;
		sx = ex - Scale.window_width;
		if( sx < 0 ){	/* add C1 */
			sx = 0;
			ex = sx + buf->width;
		}
	}
	if( ex + Scale.window_width - buf->width < 10 ){
		ex = buf->width - Scale.window_width;
	}
	if( ey >= buf->height ){
		ey = buf->height;
		sy = ey - Scale.window_height;
		if( sy < 0 ){	/* add C1 */
			sy = 0;
			ey = sx + buf->height;
		}
	}
	if( ey + Scale.window_height - buf->height < 10 ){
		ey = buf->height - Scale.window_height;
	}
/* add C3 */
	q = Shadow;
	for( i = 0; i < ShadowSize / sizeof(int); i++ ){
		*q++ = ClearColor;
	}
	p = buf->data + buf->width * sy;
	q = Shadow;
	Scale.window_xstart = sx;
	Scale.window_ystart = sy;

	for( i = sy; i < ey; i++ ){
		if( i - sy >= Scale.window_height ){
			break;
		}
		for( j = sx; j < ex; j++ ){
			if( j - sx >= Scale.window_width ){
				break;
			}
			q[j-sx] = (*change1)( (*change2)( p[j] ) );
		}
		q += Scale.window_width;
		p += buf->width;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function zoom_up()                                      */
/*                                                                    */
/*            args  struct Shm *buf   :  The Image Data               */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: If scale1 > scale2, it is zoom_up Display.    */
/*                      The Display routine copies buf->data to Shadow.*/
/**********************************************************************/
static void zoom_up( struct Shm *buf )
{
	int i, j, k;
	unsigned int *p, *q;
	int sx, sy, ex, ey;
	int l, m;
	int mag;

	mag = Scale.scale1;
	sx = (Scale.window_width / mag / 2) * Scale.xstart;
	sy = (Scale.window_height / mag / 2) * Scale.ystart;
	ex = sx + Scale.window_width / mag;
	ey = sy + Scale.window_height / mag;
	if( ex >= buf->width ){
		ex = buf->width;
		sx = ex - Scale.window_width / mag;
		if( sx < 0 ){
			sx = 0;
		}
	}
	if( ex + Scale.window_width / mag - buf->width < 10 ){
		ex = buf->width - Scale.window_width / mag;
	}
	if( ey >= buf->height ){
		ey = buf->height;
		sy = ey - Scale.window_height / mag;
		if( sy < 0 ){
			sy = 0;
		}
	}
	if( ey + Scale.window_height / mag - buf->height < 10 ){
		ey = buf->height - Scale.window_height / mag;
	}
/* add C3 */
	q = Shadow;
	for( i = 0; i < ShadowSize / sizeof(int); i++ ){
		*q++ = ClearColor;
	}
	p = buf->data + buf->width * sy;
	q = Shadow;
	Scale.window_xstart = sx;
	Scale.window_ystart = sy;
	for( i = sy; i < ey; i++ ){
		for( k = 0, j = sx; j < ex; j++ ){
			for( m = 0; m < mag; m++ ){
				if( k >= Scale.window_width ){
					break;
				}
				q[k++] = (*change1)( (*change2)( p[j] ) );
			}
		}
		q += Scale.window_width;
		for( l = 1; l < mag; l++ ){	/* optimized */
			if( (i - sy) * mag + l >= Scale.window_height ){
				return;
			}
			memcpy( q, q - Scale.window_width, Scale.window_width * 4 );
			q += Scale.window_width;
		}
		p += buf->width;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function zoom_back()                                    */
/*                                                                    */
/*            args  struct Shm *buf   : The Image Data                */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: If scale1 < scale2, it is zoom_back Display.  */
/*                      The Display routine copies buf->data to Shadow.*/
/**********************************************************************/
static void zoom_back( struct Shm *buf )
{
	int i, j, k;
	unsigned int *p, *q;
	int sx, sy, ex, ey;
	int mag;

	mag = Scale.scale2;
	sx = (Scale.window_width * mag / 2) * Scale.xstart;
	sy = (Scale.window_height * mag / 2) * Scale.ystart;
	ex = sx + Scale.window_width * mag;
	ey = sy + Scale.window_height * mag;
	if( ex >= buf->width ){
		ex = buf->width;
		sx = ex - Scale.window_width * mag;
		if( sx < 0 ){
			sx = 0;
		}
	}
	if( ex + Scale.window_width * mag - buf->width < 10 ){
		ex = buf->width - Scale.window_width * mag;
	}
	if( ey >= buf->height ){
		ey = buf->height;
		sy = ey - Scale.window_height * mag;
		if( sy < 0 ){
			sy = 0;
		}
	}
	if( ey + Scale.window_height * mag - buf->height < 10 ){
		ey = buf->height - Scale.window_height * mag;
	}
	p = buf->data + buf->width * sy;
	q = Shadow;
	Scale.window_xstart = sx;
	Scale.window_ystart = sy;
/* add C3 */
	for( i = 0; i < ShadowSize / sizeof(int); i++ ){
		*q++ = ClearColor;
	}
	q = Shadow;
	for( i = sy; i < ey; i++ ){
		if( i % mag == 0 ){
			for( k = 0, j = sx; j < ex; j++ ){
				if( j % mag ){
					continue;
				}
				q[k++] = (*change1)( (*change2)( p[j] ) );
			}
			q += Scale.window_width;
		}
		p += buf->width;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function put_image2()                                   */
/*                                                                    */
/*            args  int kind  : 0=SHOW_CLEAR  1=SHOW_MASK             */
/*                  struct Shm *buff   : The Image Data               */
/*                                                                    */
/*            return  NONE                                            */
/*                                                                    */
/*            function: 2nd Display Routine determins convert funcs   */
/*                      and call buff->data to Shadow copy routine    */
/*                      and call XPutImage().                         */
/**********************************************************************/
static void put_image2( int kind, struct Shm *buff )
{
	int size;
	unsigned int *tmp;

	if( !FullColor ){
		change1 = change_24to16;
	} else {
		change1 = no_change;
	}
	if( EventStatus == ESMASK && kind == SHOW_MASK ){
		if( buff->mask_kind == DEPTHMASK ){
			change2 = depthmask_change;
		} else if( buff->mask_kind == SIMPLEMASK ){
			change2 = simplemask_change;
		} else {
			change2 = no_change;
		}
	} else {
		change2 = no_change;
	}

/* C4 changed for memory allocate economy */
	size= Scale.window_width * Scale.window_height * 4 ;
	if( ShadowSize != size ){
		tmp = (unsigned int *)realloc( Shadow, size );
		if( tmp == NULL ){
			fprintf( stderr, "cannot realloc shadow buff\n" );
			return;
		}
		Shadow = tmp;
		ShadowSize = size;
	}
	if( Scale.scale1 == 1 ){
		if( Scale.scale2 == 1 ){
			real_size( buff );
		} else {
			zoom_back( buff );
		}
	} else {
		if( Scale.scale2 == 1 ){
			zoom_up( buff );
		} else {
			return;
		}
	}

	Image->data = (char *)Shadow;
	Image->width = Scale.window_width;
	Image->height = Scale.window_height;
	Image->bytes_per_line = Image->width * 4;
	XPutImage( D, W, gc, Image, 0, 0, 0, 0, Image->width, Image->height );
}


/**********************************************************************/
/*                                                                    */
/*            function put_image()                                    */
/*                                                                    */
/*            args  struct Shm *buff  : The Image Data                */
/*                  int kind   :        0==SHOW_CLEAR 1==SHOW_MASK    */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: External Interface for Display Image.         */
/*                      Optimized if real buff can use.               */
/**********************************************************************/
void put_image( struct Shm *buff, int kind )
{
	if( kind == SHOW_MASK ){
		put_image2( kind, buff );
	} else if( kind == SHOW_CLEAR ){
		if( FullColor && Scale.scale1 == 1 && Scale.scale2 == 1 
		   && Scale.xstart == 0 && Scale.ystart == 0 ){ /* KLUGE */
			Image->data = (char *)buff->data;
			Image->width = buff->width;
			Image->height = buff->height;
			Image->bytes_per_line = Image->width * 4;
			XPutImage( D, W, gc, Image, 0, 0, 0, 0, 
				  Image->width, Image->height );
		} else {
			put_image2( kind, buff );
		}
	} else {
		fprintf( stderr, "cannot happen\n" );
		return;
	}
	XFlush( D );
}

/**********************************************************************/
/*                                                                    */
/*            function free_masks()                                   */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Free the Shadow Image Buffer.                 */
/*                                                                    */
/**********************************************************************/
void free_masks( void )
{
	free( Shadow );
}


/**********************************************************************/
/*                                                                    */
/*            function set_button_style()                             */
/*                                                                    */
/*            args  int kind  : 0==Width op, 1==Height op             */
/*                  struct WinSize *win  : style struct               */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Each Width and Height, adjust the size by     */
/*                      sidebar width.                                */
/**********************************************************************/
static void set_button_style( int kind, struct WinSize *win )
{
	int a, b;
	int *imagelen, *orglen, *start, mask;

	if( kind == 0 ){
		imagelen = &win->image.x;
		orglen = &Scale.window_org_width;
		start = &Scale.xstart;
		mask = WIN_XMIN;
	} else {
		imagelen = &win->image.y;
		orglen = &Scale.window_org_height;
		start = &Scale.ystart;
		mask = WIN_YMIN;
	}
	a = *imagelen;
	b = *orglen * Scale.scale1 / Scale.scale2;
	if( a < b ){
		if( b - a < 10 ){
			win->style &= ~mask;
			*imagelen += b - a;
		} else if( win->style & mask ){
		} else {
			win->style |= mask;
		}
	} else {
		if( win->style & mask ){
			win->style &= ~mask;
		} else {
		}
		*start = 0;
	}
	if( *start < 0 ){
		*start = 0;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function set_scale()                                    */
/*                                                                    */
/*            args  int kind  :  1=CONFIG_LOOPE  0=CONFIG_SIZE etc.   */
/*                  int num1  :  ratio for loope or size(num1/num2)   */
/*                  int num2  :  ratio for loope or size(num1/num2)   */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: change windowsize and loope ratio from        */
/*                      commandline 'show "**=<num1>/<num2>"'         */
/*                      Change procedure of Scale struct.             */
/**********************************************************************/
static void set_scale( int kind, int num1, int num2  )
{
	switch( kind ){
	case CONFIG_LOOPE_GT:
		Scale.scale1 = num1 / num2;
		Scale.scale2 = 1;
		Scale.xstart = Scale.xstart * num1 / num2 + 1;
		Scale.ystart = Scale.ystart * num1 / num2 + 1;
		break;
	case CONFIG_LOOPE_LT:
		Scale.scale1 = 1;
		Scale.scale2 = num2 / num1;
		Scale.xstart = Scale.xstart * num2 / num1;
		Scale.ystart = Scale.ystart * num2 / num1;
		break;
	case CONFIG_LOOPE_EQ:
		Scale.scale1 = Scale.scale2 = 1;
		Scale.xstart = Scale.ystart = 0;
		break;
	case CONFIG_SIZE:
		Scale.xstart = Scale.xstart * num2 / num1;
		Scale.ystart = Scale.ystart * num2 / num1;
		Scale.window_width = Scale.window_org_width * num1 / num2;
		Scale.window_height = Scale.window_org_height * num1 / num2;
		break;
	case WINDOW_LOOPE_DOUBLE:
		if( Scale.scale2 > 1 ){
			Scale.scale2 /= 2;
		} else {
			Scale.scale1 *= 2;
		}
		Scale.xstart = Scale.xstart * 2 + 1;
		Scale.ystart = Scale.ystart * 2 + 1;
		break;
	case WINDOW_LOOPE_HALF:
		if( Scale.scale1 > 1 ){
			Scale.scale1 /= 2;
		} else {
			Scale.scale2 *= 2;
		}
		Scale.xstart = Scale.xstart / 2;
		Scale.ystart = Scale.ystart / 2;
		break;
	case WINDOW_SIZE_DOUBLE:
		Scale.window_width *= 2;
		Scale.window_height *= 2;
		Scale.xstart /= 2;
		Scale.ystart /= 2;
		break;
	case WINDOW_SIZE_HALF:
		Scale.window_width /= 2;
		Scale.window_height /= 2;
		Scale.xstart *= 2;
		Scale.ystart *= 2;
		break;
	case WINDOW_SIZE_EQ:
		Scale.xstart = 0;
		Scale.ystart = 0;
		Scale.window_width = Scale.window_org_width * Scale.scale1 / Scale.scale2;
		Scale.window_height = Scale.window_org_height * Scale.scale1 / Scale.scale2;
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function set_scale()                                    */
/*                                                                    */
/*            args  int kind  :  1=CONFIG_LOOPE  0=CONFIG_SIZE etc.   */
/*                  struct WinSize *win : style struct                */
/*                  int num1  :  ratio for loope or size(num1/num2)   */
/*                  int num2  :  ratio for loope or size(num1/num2)   */
/*                                                                    */
/*            return int : 0==success  -1=too large                   */
/*                                                                    */
/*            function: change windowsize and loope ratio from        */
/*                      commandline 'show "**=<num1>/<num2>"'         */
/*                      Change procedure of win by set_scale() and    */
/*                      follows resize_window().                      */
/**********************************************************************/
int set_new_winsize( int kind, struct WinSize *win, int num1, int num2 )
{
	struct Scale org;
	unsigned int *tmp;
	int size;

	memcpy( &org, &Scale, sizeof(struct Scale) );
	if( kind == CONFIG_LOOPE ){
		if( num1 > num2 ){
			kind = CONFIG_LOOPE_GT;
		} else if( num1 < num2 ){
			kind = CONFIG_LOOPE_LT;
		} else {
			kind = CONFIG_LOOPE_EQ;
		}
	}
	win->style = 0;
	set_scale( kind, num1, num2 );
	switch( kind ){
	case CONFIG_LOOPE_GT:
	case CONFIG_LOOPE_LT:
	case CONFIG_LOOPE_EQ:
	case WINDOW_LOOPE_DOUBLE:
	case WINDOW_LOOPE_HALF:
		win->command = WIN_LOOPE;
		break;
	case CONFIG_SIZE:
	case WINDOW_SIZE_DOUBLE:
	case WINDOW_SIZE_HALF:
	case WINDOW_SIZE_EQ:
		win->command = WIN_RESIZE;
		break;
	}
	win->image.x = Scale.window_width;
	win->image.y = Scale.window_height;
	set_button_style( 0, win );
	set_button_style( 1, win );
/* C4 changed for memory allocate economy */
	size = Scale.window_width * Scale.window_height * 4;
	if( size != ShadowSize ){
		tmp = (unsigned int *)realloc( Shadow, size );
		if( tmp == NULL ){
			printf( "too large -- ignore\n" );
			memcpy( &Scale, &org, sizeof(struct Scale) );
			return( -1 );
		}
		ShadowSize = size;
		Shadow = tmp;
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function move_display()                                 */
/*                                                                    */
/*            args  int kind  : 2==CONFIG_UP 3==CONFIG_DOWN           */
/*                              4==CONFIG_LEFT 5==CONFIG_RIGHT        */
/*                  int num  :  How much move?                        */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Move the Display Area in Button Press at Bar. */
/*                     And in commandline 'show' called from change_sizes()*/
/**********************************************************************/
void move_display( int kind, int num )
{
	int xy;
	int xratio, yratio;

	xratio = Scale.scale1 * Scale.image_width * 2 
		/ Scale.scale2 / Scale.window_width - 1;
	yratio = Scale.scale1 * Scale.image_height * 2 
		/ Scale.scale2 / Scale.window_height - 1;

	switch( kind ){
	case CONFIG_UP:
		xy = Scale.ystart;
		xy -= num;
		if( xy < 0 ){
			xy = 0;
		}
		Scale.ystart = xy;
		break;
	case CONFIG_DOWN:
		xy = Scale.ystart;
		xy += num;
		if( xy > yratio ){
			xy = yratio;
		}
		Scale.ystart = xy;
		break;
	case CONFIG_LEFT:
		xy = Scale.xstart;
		xy -= num;
		if( xy < 0 ){
			xy = 0;
		}
		Scale.xstart = xy;
		break;
	case CONFIG_RIGHT:
		xy = Scale.xstart;
		xy += num;
		if( xy > xratio ){
			xy = xratio;
		}
		Scale.xstart = xy;
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function change_view()                                  */
/*                                                                    */
/*            args  int kind  : 2==CONFIG_UP 3==CONFIG_DOWN           */
/*                              4==CONFIG_LEFT 5==CONFIG_RIGHT        */
/*                  unsigned char *s : Configure string from gedo.    */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Move Display Area from command line of gedo   */
/*                      'show "up=<num>"' etc.                        */
/**********************************************************************/
void change_view( int kind, unsigned char *s )
{
	int num;

	num = atoi( s );
	if( num == 0 ){
		num = 1;
	}
	move_display( kind, num );
}


/**********************************************************************/
/*                                                                    */
/*            function position_r2w()                                 */
/*                                                                    */
/*            args  struct Positions *p : convert Positions           */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Convert the geometry from real to window.     */
/*                                                                    */
/**********************************************************************/
void position_r2w( struct Positions *p )
{
	p->window.x = (p->real.x - Scale.window_xstart) * Scale.scale1 / Scale.scale2;
	p->window.y = (p->real.y - Scale.window_ystart) * Scale.scale1 / Scale.scale2;
}

/**********************************************************************/
/*                                                                    */
/*            function position_w2r()                                 */
/*                                                                    */
/*            args  struct Positions *p  : convert Positions          */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Convert the geometry from window to real.     */
/*                                                                    */
/**********************************************************************/
void position_w2r( struct Positions *p )
{
	p->real.x = Scale.window_xstart + p->window.x * Scale.scale2 / Scale.scale1;
	p->real.y = Scale.window_ystart + p->window.y * Scale.scale2 / Scale.scale1;
}

/**********************************************************************/
/*                                                                    */
/*            function get_scale_rate()                               */
/*                                                                    */
/*            args  int kind : 0==Width(Lower) Bar, 1==Height(Right) Bar */
/*                                                                    */
/*            return int  : Display Thum position.                    */
/*                                                                    */
/*            function: Calcurate Thum position for Bars.             */
/*                                                                    */
/**********************************************************************/
int get_scale_rate( int kind )
{
	float ret;
	int ratio;

	if( kind == 1 ){
		ratio = Scale.scale1 * Scale.image_width * 2 
			/ Scale.scale2 / Scale.window_width - 1;
		if( ratio == 0 ){
			return( 0 );
		}
		ret = (float)Scale.window_width * (float)Scale.xstart / (float)ratio;
	} else {
		ratio = Scale.scale1 * Scale.image_height * 2 
			/ Scale.scale2 / Scale.window_height - 1;
		if( ratio == 0 ){
			return( 0 );
		}
		ret = (float)Scale.window_height * (float)Scale.ystart / (float)ratio;
	}
	return( (int)ret );
}


/**********************************************************************/
/*                                                                    */
/*            function get_scale()                                    */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return int  : Return Scale: >0 == zoom <0 == back       */
/*                                                                    */
/*            function: return the Scale for show process.            */
/*            add C2                                                  */
/**********************************************************************/
int get_scale( void )
{
	if( Scale.scale1 / Scale.scale2 > 1 ){
		return( Scale.scale1 / Scale.scale2 );
	} else {
		return( (-1) * Scale.scale2 / Scale.scale1 );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function show_scale()                                   */
/*                                                                    */
/*            args   NONE                                             */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Scale Struct printer(for debug).              */
/*                                                                    */
/**********************************************************************/
void show_scale( void )
{
	printf( "Scale:scale1=%d/%d start=%d:%d\n", Scale.scale1, Scale.scale2, Scale.xstart, Scale.ystart );
	printf( "      image=%d:%d ", Scale.image_width, Scale.image_height );
	printf( "org=%d:%d window=%d:%d\n", Scale.window_org_width, Scale.window_org_height, Scale.window_width, Scale.window_height );
}

/**********************************************************************/
/*                                                                    */
/*            function show_winsize()                                 */
/*                                                                    */
/*            args  struct WinSize *win  : Style Struct               */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Style struct printer(for debug).              */
/*                                                                    */
/**********************************************************************/
void show_winsize( struct WinSize *win )
{
	printf( "WinSize:command=%d style=%08x imagetop=%d\n", win->command, win->style, win->imagetop );
	printf( "        org=%d:%d  all=%d:%d  image=%d:%d\n", win->org.x, win->org.y, win->all.x, win->all.y, win->image.x, win->image.y );
}

