/**********************************************************************/
/*                                                                    */
/*           loaderx.c  viewer 2nd GUI process routine                */
/*                                                                    */
/*                by K.sugiura   Ver 0.93 for GEDO of LINUX           */
/*                                                                    */
/*           External Interfaces:
	int realloc_sel( int num );
	int  create_image(char *name, struct  Shm *buff);
	void resize_window( struct WinSize *win );
	void  disp_image(struct Shm *buf);			      */
/*      C1: select edit -> ESC occure SEGV, fixed.                    */
/*      C2: select edit increase Now multiply                         */
/*      C3: if unexpected error occured, can resume error.            */ 
/*      D1: arrow key FIX                                             */
/*                                                                    */
/**********************************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <setjmp.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>

#include "viewer.h"

#define MAXPOSITION 30

Display *D;		/* child process Display */
Window   W0;		/* top level Window */
Window   W;		/* Image display Window */

static Window	 Wmode;	/* Mode Button */
static Window	 Wok;	/* OK Button */
static Window	 Wesc;	/* ESC Button */
static Window	 Wdel;	/* DEL Button */
static Window	 Ws1;	/* *2 Button */
static Window	 Ws2;	/* /2 Button */
static Window	 Wl1;	/* ZOOM Button */
static Window	 Wl2;	/* BACK Button */
static Window	 Wwbar;	/* Bottom Bar(Width) */
static Window	 Whbar;	/* Right Bar(Height) */
static Window	 Winit; /* init size Button (Right-Bottom)*/

GC	 gc;		/* GC for image display */
GC	 gcy;		/* GC for selection and labber band */
static GC	 gcb;	/* GC for string displayed BLACK(can select) */
static GC	 gcg;	/* GC for string displayed GRAY(cannot select) */

XImage  *Image;		/* Image interface to X11 */
Visual  *Vis;		/* Visual information of X11 */
static int      Dep;	/* Visual Depth of X11 */

extern int MesId;	/* Message ID of GEDO transaction */

struct Pos *Sel;	/* Selected Points (real geometory) */
int MaxPos = MAXPOSITION;	/* select Points Maximum */
static int AllocedPosNum;	/* Now Alloced Sel num */

int SKind;		/* Kind of the File/Position select. 
			  defined in ../loader.h
				#define SKPOS		0
				#define SKFILE		1 */

int Snum;		/* Now selected Sel number */
int SelectStatus;	/* kind of Selection, defined in ../loader.h
					#define ADD       0
					#define CHANGE    1
					#define EDIT      1	*/

int EventStatus;	/* kind of the Send Event, in ../loader.h
					#define ESDISPLAY   0
					#define ESMASK      1
					#define ESSELECT    2
					#define ESKILL      4
					#define ESCONFIG    5 */

static int Mode;	/* Kind of the Selection defined in ../loader.h
				#define SKNONE    0
				#define SKPOINT   1
				#define SKBOX     2
				#define SKPOLY    3
				#define SKLINES   4
				#define SKFREE	  5
				#define SKSPPOLY  6
				#define SKSPLINE  7  */
static int *Viewer_Status;	/* pointer to Shm->status, BUSY, IDOL, or ERROR */
static int ShiftStatus;		/* KeyBoad Shift Status */
static int AltStatus;		/* KeyBoad Alt Status */

int MaskColor = 0x00ffff00;	/* Mask Selection Color */

struct Function {		/* Selection Database by Mode */
	void (*function)(int, struct Pos *);	/* handler address */
	int maxselect;		/* selection Maximum */
	int delable;		/* DEL process enable or disable */
};

#define DEL_DISABLE	0
#define DEL_ABLE	1

#define KINDNUM	8

struct Function Func[KINDNUM] = {
	{ selnone, 0, DEL_DISABLE },	{ selpoint, 1, DEL_DISABLE },
	{ selbox,  2, DEL_DISABLE },	{ selpoly,  INFINIT, DEL_ABLE },
	{ selline, INFINIT, DEL_ABLE },	{ selfree, INFINIT, DEL_DISABLE },
	{ selsppoly, MAXSP, DEL_ABLE },	{ selspline, MAXSP, DEL_ABLE }
};
static int DefaultMax[KINDNUM] 	/* Func default Maximum */
= { 0, 1, 2, INFINIT, INFINIT, INFINIT, MAXSP, MAXSP };

int Now;		/* Now selecting Sel */
int NowMask = 0;	/* not used */
static char *strsts[2] 	/* strings for MODE Button */
= { " Add  ", "Change" };

static char SFFilename[240];	/* if SKind==SKFILE, read and write filename */
static jmp_buf jmpbuf;		/* Ctrl+C STOP  at select */
static int CtrlcFlag = 0;	/* Ctrl+C STOP  at showing */

/* add D2 */
unsigned long GrayPixel;
/* end D2 */

/**********************************************************************/
/*                                                                    */
/*            function realloc_sel()                                  */
/*                                                                    */
/*            args  int num : want more <num> Sel                     */
/*                                                                    */
/*            return int   : 0==success  -1==malloc error             */
/*                                                                    */
/*            function:  realloc Sel and get needed more Sel.         */
/*                                                                    */
/**********************************************************************/
int realloc_sel( int num )
{
	struct Pos *tmp;

	if( MaxPos + num <= AllocedPosNum ){
		return( 0 );
	}

	tmp = (struct Pos *)malloc( (AllocedPosNum + num) * sizeof(struct Pos) );
	if( tmp  == NULL ){
		return( -1 );
	}
	if( tmp != Sel ){
		memcpy( tmp, Sel, AllocedPosNum * sizeof(struct Pos) );
/*		free( Sel );	Why, I cannot understand! (segv creater)*/
		Sel = tmp;
	}
	AllocedPosNum += num;
	MaxPos  = AllocedPosNum;
	return( 0 );
}



/**********************************************************************/
/*                                                                    */
/*            function create_image()                                 */
/*                                                                    */
/*            args  struct Shm *buff : The Image Data                 */
/*                                                                    */
/*            return int   : 0==success  -1==malloc error             */
/*                                                                    */
/*            function: Called after fork(), create New Image window. */
/*                                                                    */
/**********************************************************************/
int create_image( int size, char *name, struct Shm *buff )
{
	Colormap cmap;
	XColor yellow, c;
	XTextProperty win_name;	
	XClassHint  class;
	char *argv[1];
	int ret;

	
	if( (Sel = malloc( MAXPOSITION * sizeof(struct Pos) * 50 )) == NULL ){
		return( -1 );
	}
	AllocedPosNum = MAXPOSITION * 50;
/* KLUGE   -- FIXME */

	Vis = DefaultVisual( D, 0 );
	Dep = DefaultDepth( D, 0 );

	if( Vis->red_mask != 0x00ff0000 || Vis->green_mask != 0x0000ff00
		|| Vis->blue_mask != 0x000000ff ){
		ret = init_16bit( size );
	} else {
		ret = init_24bit( size );
	}
	if( ret == -1 ){
		fprintf( stderr, "cannot malloc shadow buffer\n" );
		return( 1 );
	}

	if( name[0] == '\0' ){
		win_name.value = (unsigned char *)"gedo";
	} else {
	        win_name.value = (unsigned char *)name;
	}
	win_name.encoding = XA_STRING;
	win_name.format = 8;
	win_name.nitems = strlen( win_name.value );
	class.res_name = NULL;
	class.res_class = "Gedo";
	argv[0] = NULL;
	XSetWMProperties( D, W0, &win_name, NULL, argv, 0, 
			 NULL, NULL, &class );

	XFlush( D );

	gc = XCreateGC( D, W, 0, 0 );
	gcy = XCreateGC( D, W, 0, 0 );
	gcb = XCreateGC( D, W0, 0, 0 );
	gcg = XCreateGC( D, W0, 0, 0 );
	XFlush( D );

	cmap = DefaultColormap( D, 0 );
	XAllocNamedColor( D, cmap, "yellow", &yellow, &c );
	XSetForeground( D, gcy, yellow.pixel );
	XAllocNamedColor( D, cmap, "gray", &yellow, &c );
	XSetForeground( D, gcg, yellow.pixel );
/* add D2 */
	GrayPixel = yellow.pixel;
/* end D2 */
	XSetLineAttributes( D, gcy, 1, LineSolid, CapButt, JoinMiter );
	XSetFunction( D, gcy, GXxor );

	XSelectInput( D, W0, ExposureMask | KeyPressMask | KeyReleaseMask );
	XSelectInput( D, W, ExposureMask | ButtonPressMask 
		          | ButtonMotionMask | ButtonReleaseMask );
	XSelectInput( D, Wmode, ButtonPressMask );
	XSelectInput( D, Wok, ButtonPressMask );
	XSelectInput( D, Wesc, ButtonPressMask );
	XSelectInput( D, Wdel, ButtonPressMask );
	XSelectInput( D, Ws1, ButtonPressMask );
	XSelectInput( D, Ws2, ButtonPressMask );
	XSelectInput( D, Wl1, ButtonPressMask );
	XSelectInput( D, Wl2, ButtonPressMask );
	XSelectInput( D, Wwbar, ButtonPressMask );
	XSelectInput( D, Whbar, ButtonPressMask );
	XSelectInput( D, Winit, ButtonPressMask );
	XFlush( D );

	Image = XCreateImage( D, Vis, Dep, ZPixmap, 0, 0,
			     0, 0, 32, 0 );
	Image->data    = (char *)buff->data;
	Image->xoffset = 0;
	Image->depth   = Dep;
	Image->byte_order = LSBFirst;
	Image->bitmap_unit = 32;
	Image->bitmap_bit_order = LSBFirst;
	Image->bytes_per_line = buff->width * 4;
	Image->bits_per_pixel = 32;
	Image->red_mask = Vis->red_mask;
	Image->green_mask = Vis->green_mask;
	Image->blue_mask = Vis->blue_mask;
	Image->width = buff->width;
	Image->height = buff->height;

	XMapSubwindows( D, W0 );
	XMapWindow( D, W0 );
	XFlush( D );

	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function print_sbuf()                                   */
/*                                                                    */
/*            args  struct Sbuf *sb : showed Sbuf                     */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Show Sbuf(transaction) for debug.             */
/*                                                                    */
/**********************************************************************/
static void  print_sbuf( struct Sbuf *sb )
{
	int i;

	printf( "selnum:%d\n", sb->selnum );
	printf( "defnum:%d\n", sb->defnum );
	for( i = 0; i < sb->defnum; i++ ){
		printf( "%d:(%d,%d)\n", i, sb->def[i].x, sb->def[i].y );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function read_pos()                                     */
/*                                                                    */
/*            args  struct Sfile *sf : transaction Sfile              */
/*                                                                    */
/*            return int : 0==sucess  errorcode==fail                 */
/*                                                                    */
/*            function: if SKind==SKFILE, read from transaction file  */
/*                      and set data available.                       */
/**********************************************************************/
static int  read_pos( struct Sfile *sf )
{
	struct Shead sh;
	FILE *fp;

	strcpy( SFFilename, sf->file );
	if( (fp = fopen( sf->file, "r" )) == NULL ){
		sf->defnum = 0;
		return( 0 );
	}
	if( fread( &sh, sizeof(struct Shead), 1, fp ) != 1 ){
		fclose( fp );
		return( EREADFILE );
	}
	sf->defnum = sh.defnum;
	if( sf->defnum <= 0 ){
		fclose( fp );
		return( 0 );
	}
	if( sf->defnum >= MaxPos ){
		if( realloc_sel( sf->defnum ) != 0 ){
			return( EMALLOC );
		}
	}
	if( fread( Sel, sizeof(struct Pos), (size_t)sf->defnum, fp ) != (size_t)sf->defnum ){
		fclose( fp );
		return( EREADFILE );
	}
	fclose( fp );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function draw_bar()                                     */
/*                                                                    */
/*            args  int kind : 0==Bottom Bar(Width)                   */
/*                             1==Right Bar(Height)                   */
/*            return NONE                                             */
/*                                                                    */
/*            function: Draw the Bar and Thum.                        */
/*                                                                    */
/**********************************************************************/
void draw_bar( int kind )
{
	int ret;

	ret = get_scale_rate( kind );
	if( kind == 1 ){
		XClearWindow( D, Wwbar );
		XFillRectangle( D, Wwbar, gcb, ret-20, 0, 40, 20 );
	} else {
		XClearWindow( D, Whbar );
		XFillRectangle( D, Whbar, gcb, 0, ret-20, 20, 40 );
	}
}


/**********************************************************************/
/*                                                                    */
/*            function create_buttons()                               */
/*                                                                    */
/*            args  int flag : struct WinSize style                   */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Create Buttons.                               */
/*                                                                    */
/**********************************************************************/
void create_buttons( int flag )
{
	int x, y;

	if( flag & WIN_2ROWBUTTON ){
		x = 0;
		y = 35;
	} else {
		x = 200;
		y = 5;
	}
	Wmode = XCreateSimpleWindow( D, W0, 5, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wok = XCreateSimpleWindow( D, W0, 55, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wesc = XCreateSimpleWindow( D, W0, 105, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wdel = XCreateSimpleWindow( D, W0, 155, 5,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Ws1 = XCreateSimpleWindow( D, W0, x + 5, y,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Ws2 = XCreateSimpleWindow( D, W0, x + 55, y,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wl1 = XCreateSimpleWindow( D, W0, x + 105, y,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wl2 = XCreateSimpleWindow( D, W0, x + 155, y,
		40, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Wwbar = XCreateSimpleWindow( D, W0, 0, 0,
		1, 1, 1, BlackPixel(D,0), WhitePixel(D,0) );	
	Whbar = XCreateSimpleWindow( D, W0, 0, 0,
		1, 1, 1, BlackPixel(D,0), WhitePixel(D,0) );
	Winit = XCreateSimpleWindow( D, W0, 0, 0,
		20, 20, 1, BlackPixel(D,0), WhitePixel(D,0) );
}


/**********************************************************************/
/*                                                                    */
/*            function rewrite_mode_button()                          */
/*                                                                    */
/*            args  GC gcp : Draw string GC(gcb or gcg)               */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Refresh MODE button Window string.            */
/*                                                                    */
/**********************************************************************/
static void rewrite_mode_button( GC gcp )
{
	XClearWindow( D, Wmode );
	XDrawString( D, Wmode, gcp, 2, 15, strsts[SelectStatus], 6 );
}

/**********************************************************************/
/*                                                                    */
/*            function selected()                                     */
/*                                                                    */
/*            args  int flag : ESUP == 1  ESDOWN == 0 for viewer new  */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: viewer "new" command changes window border line */
/*            D2: created                                             */
/**********************************************************************/
static void selected( int flag )
{
     XSetWindowAttributes xwa;

     if( !flag ){
	  xwa.border_pixel = GrayPixel;
     } else {
	  xwa.border_pixel = BlackPixel(D,0);
     }
     XChangeWindowAttributes( D, Wmode, CWBorderPixel, &xwa ); 
     XChangeWindowAttributes( D, Wok, CWBorderPixel, &xwa ); 
     XChangeWindowAttributes( D, Wesc, CWBorderPixel, &xwa ); 
     XChangeWindowAttributes( D, Wdel, CWBorderPixel, &xwa ); 
     XChangeWindowAttributes( D, Ws1, CWBorderPixel, &xwa ); 
     XChangeWindowAttributes( D, Ws2, CWBorderPixel, &xwa ); 
     XChangeWindowAttributes( D, Wl1, CWBorderPixel, &xwa ); 
     XChangeWindowAttributes( D, Wl2, CWBorderPixel, &xwa ); 
     XFlush( D );
}

/**********************************************************************/
/*                                                                    */
/*            function rewrite_button()                               */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Refresh each button Window string.            */
/*                                                                    */
/**********************************************************************/
static void rewrite_button( void )
{
	GC gcp;

	if( EventStatus == ESSELECT ){
		gcp = gcb;
	} else {
		gcp = gcg;
	}
	rewrite_mode_button( gcp );
	XDrawString( D, Wok, gcp, 2, 15, " OK", 3 );
	XDrawString( D, Wesc, gcp, 2, 15, " ESC", 4 );
	XDrawString( D, Wdel, gcp, 2, 15, " DEL", 4 );
	XDrawString( D, Ws1, gcb, 2, 15, " *2 ", 4 );
	XDrawString( D, Ws2, gcb, 2, 15, " /2 ", 4 );
	XDrawString( D, Wl1, gcb, 2, 15, " ZOOM", 5 );
	XDrawString( D, Wl2, gcb, 2, 15, " BACK", 5 );
}

/**********************************************************************/
/*                                                                    */
/*            function resize_window()                                */
/*                                                                    */
/*            args  struct WinSize *win : style struct                */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Resize each Window by style.                  */
/*                                                                    */
/**********************************************************************/
void resize_window( struct WinSize *win )
{
	int barx, bary;

	if( win->image.x >= 200 ){
		win->all.x = win->image.x;
	} else {
		win->all.x = 200;
	}
	if( win->image.x >= 400 ){
		win->all.y = win->image.y + 30;
		win->imagetop = 30;
	} else {
		win->all.y = win->image.y + 60;
		win->style |= WIN_2ROWBUTTON;
		win->imagetop = 60;
	}
	barx = win->all.x;
	bary = win->all.y;
	if( win->style & WIN_XMIN ){
		win->all.x += 20;
	}
	if( win->style & WIN_YMIN ){
		win->all.y += 20;
	}
	if( win->command == WIN_INIT ){
		create_buttons( win->style );
	}
	XResizeWindow( D, W0, win->all.x, win->all.y );
	XResizeWindow( D, W,  win->image.x, win->image.y );
	XMoveWindow( D, W, 0, win->imagetop );
	XResizeWindow( D, Wwbar, win->image.x, 20 );
	XMoveWindow( D, Wwbar, 0, bary );
	XResizeWindow( D, Whbar, 20, win->image.y );
	XMoveWindow( D, Whbar, barx, win->imagetop );
	XMoveWindow( D, Winit, barx, bary );
	if( win->style & WIN_2ROWBUTTON ){
		XMoveWindow( D, Ws1, 5, 35 );
		XMoveWindow( D, Ws2, 55, 35 );
		XMoveWindow( D, Wl1, 105, 35 );
		XMoveWindow( D, Wl2, 155, 35 );
		XMoveWindow( D, W, 0, 60 );
	} else {
		XMoveWindow( D, Ws1, 205, 5 );
		XMoveWindow( D, Ws2, 255, 5 );
		XMoveWindow( D, Wl1, 305, 5 );
		XMoveWindow( D, Wl2, 355, 5 );
		XMoveWindow( D, W, 0, 30 );
	}
	if( win->style & WIN_XMIN ){
		draw_bar( 0 );
	}
	if( win->style & WIN_YMIN ){
		draw_bar( 1 );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function flush_image()                                  */
/*                                                                    */
/*            args  struct Shm *buff : the image data                 */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Refresh the Image Window(W).     .            */
/*                                                                    */
/**********************************************************************/
static void flush_image( struct Shm *buff )
{
	if( EventStatus == ESSELECT ){
		put_image( buff, SHOW_CLEAR );
		(*Func[Mode].function)( DRAWALL, NULL );
	} else if( EventStatus == ESMASK ){
		put_image( buff, SHOW_MASK );
	} else {
		put_image( buff, SHOW_CLEAR );
	}
	XFlush( D );
}

/**********************************************************************/
/*                                                                    */
/*            function change_button_size()                           */
/*                                                                    */
/*            args  int kind  : Pushed Button                         */
/*                                                                    */
/*            return int : 0==success  -1=too large                   */
/*                                                                    */
/*            function: Called from commandpress(), and resize and    */
/*                      loope the window.                             */
/**********************************************************************/
static int change_button_size( int kind )
{
	struct WinSize win;

	if( set_new_winsize( kind, &win, 0, 0 ) != 0 ){
		return( -1 );
	}
	resize_window( &win );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function xtoi()                                         */
/*                                                                    */
/*            args  char *s : HEX string                              */
/*                                                                    */
/*            return  unsigned char : numbers                         */
/*                                                                    */
/*            function: convert HEX string to number.                 */
/*                                                                    */
/**********************************************************************/
static unsigned char xtoi( char *s )
{
	unsigned char ret;

	if( isdigit(*s) ){
		ret = *s - '0';
	} else if( (ret = toupper(*s)) ){
		ret -= ('A' - 10);
	}
	ret *= 16;
	if( isdigit(*++s) ){
		ret += (*s - '0');
	} else if( (ret += toupper(*s)) ){
		ret -= ('A' - 10);
	}
	return( ret );
}

/**********************************************************************/
/*                                                                    */
/*            function set_sizes()                                    */
/*                                                                    */
/*            args  int kind : 0==CONFIG_SIZE  1==CONFIG_LOOPE        */
/*                  unsigned char *s : configure string               */
/*                                                                    */
/*            return int : 0==success  -1==too large                  */
/*                                                                    */
/*            function: Set size and loope ratio by the string        */
/*                      from GEDO 'show "string"'                     */
/**********************************************************************/
int set_sizes( int kind, unsigned char *s  )
{
	int num1, num2;
	struct WinSize win;

	for( num1 = num2 = 0; *s && isdigit(*s); s++ ){
		num1 = num1 * 10 + *s - '0';
	}
	if( *s == '/' ){
		for( s++; *s && isdigit(*s); s++ ){
			num2 = num2 * 10 + *s - '0';
		}
	}
	if( num2 <= 0 ){
		num2 = 1;
	}
	if( num1 <= 0 ){
		num1 = 1;
	}
	if( set_new_winsize( kind, &win, num1, num2 ) != 0 ){
		return( -1 );
	}
	resize_window( &win );
	return( 0 );
}



/**********************************************************************/
/*                                                                    */
/*            function config_options()                               */
/*                                                                    */
/*            args  unsigned char *str : transaction string from gedo */
/*                                                                    */
/*            return  int : 0==success   -1==error                    */
/*                                                                    */
/*            function: From gedo 'show "**"', viewer get configure   */
/*                     string and handle the configuration.           */
/**********************************************************************/
static int config_options( unsigned char *str )
{
	Colormap cmap;
	XColor col, c;

	if( strncmp( str, "color=", 6 ) == 0 ){
		cmap = DefaultColormap( D, 0 );
		if( str[8] == ':' && str[11] == ':' ){
			col.red = xtoi( &str[6] ) * 256;
			col.green = xtoi( &str[9] ) * 256;
			col.blue = xtoi( &str[12] ) * 256;
			if( XAllocColor( D, cmap, &col ) == 0 ){
				printf( "cannot alloc new color %s\n", &str[6] );
				return( -1 );
			}
		} else if( XAllocNamedColor( D, cmap, &str[6], &col, &c ) == 0 ){
			printf( "cannot alloc new color %s\n", &str[6] );
			return( -1 );
		}
		XSetForeground( D, gcy, col.pixel );
		MaskColor = ((col.blue >> 8) & 0x000000ff);
		MaskColor |= (col.green & 0x0000ff00);
		MaskColor |= ((col.red << 8) & 0x00ff0000);
	} else if( strncmp( str, "loope=", 6 ) == 0 ){
		if( set_sizes( CONFIG_LOOPE, &str[6] ) != 0 ){
			return( -1 );
		}
	} else if( strncmp( str, "size=", 5 ) == 0 ){
		if( set_sizes( CONFIG_SIZE, &str[5] ) != 0 ){
			return( -1 );
		}
	} else if( strncmp( str, "up", 2 ) == 0 ){
		change_view( CONFIG_UP, &str[2] );
		draw_bar( 0 );
	} else if( strncmp( str, "down", 4 ) == 0 ){
		change_view( CONFIG_DOWN, &str[4] );
		draw_bar( 0 );
	} else if( strncmp( str, "right", 5 ) == 0 ){
		change_view( CONFIG_RIGHT, &str[5] );
		draw_bar( 1 );
	} else if( strncmp( str, "left", 4 ) == 0 ){
		change_view( CONFIG_LEFT, &str[4] );
		draw_bar( 1 );
	} else {
		printf( "undefined config string \"%s\"\n", str );
		return( -1 );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function event_handler()                                */
/*                                                                    */
/*            args  XEvent *event  : Send Event                       */
/*                  struct Shm *buff : The Image Data                 */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Handler of the SendEvent.                     */
/*                                                                    */
/**********************************************************************/
static void event_handler( XEvent *event, struct Shm *buff )
{
	struct sbufmes {
		long mtype;
		struct Sbuf sb;
	} ms;
	struct WinSize win;

	if( event->xclient.message_type == ESSELECT ){
		ms.sb.selnum = event->xclient.data.l[0];
		ms.sb.defnum = event->xclient.data.l[1];
		ms.sb.skind = event->xclient.data.l[2];
		ms.sb.maxnum = event->xclient.data.l[3];
		EventStatus = ESSELECT;
		Snum = ms.sb.defnum;
		Mode = ms.sb.selnum;
		SKind = ms.sb.skind;
		if( ms.sb.maxnum == 0 ){
			Func[Mode].maxselect = DefaultMax[Mode];
		} else {
			Func[Mode].maxselect = ms.sb.maxnum;
		}
		if( Snum || ms.sb.skind == SKFILE ){
			if( msgrcv( MesId, (struct msgbuf *)&ms, sizeof(struct sbufmes) - sizeof(long), getpid(), 0 ) == -1 ){
				return;
			}
			if( ms.sb.skind == SKPOS ){
				memcpy( Sel, ms.sb.def, sizeof(struct Pos) * Snum );
			} else {
				MaxPos = AllocedPosNum;
				if( read_pos( (struct Sfile *)&ms.sb ) != 0 ){
					return;
				}
				Snum = ms.sb.defnum;
			}
		}
		if( ms.sb.skind == SKPOS ){
			MaxPos = MAXPOSITION;
		}
		Now = 0;
		if( Func[Mode].maxselect > Snum ){
			SelectStatus = ADD;
		} else {
			SelectStatus = CHANGE;
		}

		switch( Mode ){
		case SKNONE:
			ms.mtype = getppid();
			if( msgsnd( MesId, (struct msgbuf *)&ms, sizeof(struct sbufmes) - sizeof(long), 0 ) == -1 ){
				return;
			}
			break;
		case SKPOLY:
		case SKLINES:
		case SKSPPOLY:
		case SKSPLINE:
			if( Snum > 0 ){
				SelectStatus = CHANGE;
			}
			break;
		case SKFREE:
			Now = Snum;
			put_image( buff, SHOW_CLEAR );
			break;
		}
		(*Func[Mode].function)( DRAWALL, NULL );
		rewrite_button();
		XFlush( D );
	} else if( event->xclient.message_type == ESDISPLAY ){
		if( event->xclient.data.l[0] == LOADFILE ){
			EventStatus = ESDISPLAY;
			NowMask = 0;

                        Image->data = NULL;
			XDestroyImage( Image );
		        Image = XCreateImage( D, Vis, Dep, ZPixmap, 0, 0,
					     0, 0, 32, 0 );
			Image->data    = (char *)buff->data;
			Image->xoffset = 0;
			Image->depth   = Dep;
			Image->byte_order = LSBFirst;
			Image->bitmap_unit = 32;
			Image->bitmap_bit_order = LSBFirst;
			Image->bytes_per_line = buff->width * 4;
			Image->bits_per_pixel = 32;
			Image->red_mask = Vis->red_mask;
			Image->green_mask = Vis->green_mask;
			Image->blue_mask = Vis->blue_mask;
			Image->width = buff->width;
			Image->height = buff->height;

			win.command = WIN_LOAD;
			win.style = 0;
			win.image.x = buff->width;
			win.image.y = buff->height;
			init_scale( &win );
			resize_window( &win );
		} else if( event->xclient.data.l[0] == REFLESH ){	
			EventStatus = ESDISPLAY;
			NowMask = 0;
		} else if( event->xclient.data.l[0] == SHOWMASK ){
			EventStatus = ESMASK;
			NowMask = event->xclient.data.l[1];
		} else {
		}
		flush_image( buff );
	} else if( event->xclient.message_type == ESKILL ){
		shmdt( (char *)buff );
		free( Sel );
		free_masks();
		XFreeGC( D, gc );
		XFreeGC( D, gcy );
                Image->data = NULL;
		XDestroyImage( Image );
		XDestroyWindow( D, W0 );
		XCloseDisplay( D );
		D = NULL; W0 = W = 0;
		exit( 0 );
	} else if( event->xclient.message_type == ESCONFIG ){
		if( config_options( event->xclient.data.b ) == 0 ){
			flush_image( buff );
		}
/* D2 multi viewer selection showing */
/* D2  ESUP and ESDOWN proc */
	} else if( event->xclient.message_type == ESUP ){
		XRaiseWindow( D, W0 );
		selected( 1 );
	} else if( event->xclient.message_type == ESDOWN ){		
		selected( 0 );
/* end D2 */
	} else {
		printf( "NAZO command=%ld\n", event->xclient.message_type );
	}
}

/**********************************************************************/
/*                                                                    */
/*            function write_file()                                   */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return int : 0==success  -1==fail                       */
/*                                                                    */
/*            function: At the end of transaction, if SKind==SKFILE,  */
/*                      write back the selections to file.            */
/**********************************************************************/
static int write_file( void )
{
	FILE *fp;
	struct Shead sh;

	if( (fp = fopen( SFFilename, "w" )) == NULL ){
		printf( "file %s cannot open\n", SFFilename );
		return( -1 );
	}
	if( Snum < 0 ){
		Snum = 0;
	}
 	sh.defnum = Snum;
	sh.selnum = Mode;
	sh.skind = SKind;
	if( fwrite( &sh, sizeof(struct Shead), 1, fp ) != 1 ){
		printf( "file %s write error\n", SFFilename );
		return( -1 );
	}
	if( Snum > 0 ){
		if( fwrite( Sel, sizeof(struct Pos), (size_t)Snum, fp ) != (size_t)Snum ){
			printf( "file %s write error\n", SFFilename );
			return( -1 );
		}
	}
	fclose( fp );
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*            function reply_selection()                              */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Communicate  result of selection to GEDO.     */
/*                                                                    */
/**********************************************************************/
static void reply_selection( void )
{
	struct sbufmes {
		long mtype;
		union {
			struct Sbuf sb;
			struct Sfile sf;
		} u;
	} ms;

	memset( &ms, 0, sizeof(struct sbufmes) );
	EventStatus = ESDISPLAY;
	if( Mode == SKFREE ){
/*		Snum--; */		/* C2 delete */
		if( Snum < 0 ){		/* C1 add */
			Snum = 0;
		}
	}
	ms.u.sb.defnum = Snum;
	ms.u.sb.selnum = Mode;
	ms.u.sb.skind  = SKind;

	if( SKind == SKPOS ){
		if( Snum > 0 ){  /* C1 add */
			memcpy( ms.u.sb.def, Sel, sizeof(struct Pos) * Snum );
		}
	} else {
		if( write_file() == -1 ){
			ms.u.sb.defnum = 0;
		} else {
			strcpy( ms.u.sf.file, SFFilename );
		}
	}
	ms.mtype = getppid();

	if( msgsnd( MesId, (struct msgbuf *)&ms, sizeof(struct sbufmes) - sizeof(long), 0 ) == -1 ){
		return;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function del_proc()                                     */
/*                                                                    */
/*            args  NONE  :                                           */
/*                                                                    */
/*            return int  : 0==deleted  -1=cannot                     */
/*                                                                    */
/*            function: Delete process at each Mode.                  */
/*                                                                    */
/**********************************************************************/
static int del_proc( void )
{
	if( SelectStatus == CHANGE ){
		if( Snum == 0 ){
			return( -1 );
		}
		if( Now == Snum - 1 ){
			Snum--;
			Now = Snum - 1;
		} else if( Snum > 0 ){
			memmove( &Sel[Now], &Sel[Now+1], (Snum-Now) * sizeof(struct Pos) );
			Snum--;
		}
	} else {
		if( Snum > 0 ){
			Snum--;
			Now = Snum;
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function keypress()                                     */
/*                                                                    */
/*            args  XEvent *event  :  KeyPress Event                  */
/*                  struct Shm *buff : The Image Data                 */
/*                                                                    */
/*            return int   :  0==success                              */
/*                                                                    */
/*            function: KeyPress Event handler.                       */
/*                                                                    */
/**********************************************************************/
static int keypress( XEvent *event, struct Shm *buff )
{

	KeySym keysym;
	XComposeStatus compose;
	char buf[10];
/* add D1 arrow key FIX */
	struct Positions pos;

	XLookupString( (XKeyEvent *)event, buf, 10, &keysym, &compose );
	switch( keysym ){
	case XK_Return:
	case XK_KP_Enter:
		if( EventStatus != ESSELECT ){
			break;
		}
		reply_selection();
		rewrite_button();
		break;
	case XK_Escape:
		if( EventStatus != ESSELECT ){
			break;
		}
		Snum = 0;
		put_image( buff, SHOW_CLEAR );
		reply_selection();
		rewrite_button();
		break;
	case XK_space:
		if( EventStatus != ESSELECT ){
			break;
		}
		if( Func[Mode].delable == DEL_DISABLE ){
			if( Snum <= DefaultMax[Mode] ){
				put_image( buff, SHOW_CLEAR );
				(*Func[Mode].function)( DRAWALL, NULL );
				XFlush( D );
				break;
			}
		}
	   	if( SelectStatus == ADD ){
			if( Snum == 0 ){
				break;
			}
			Now = Snum - 1;
	        	SelectStatus = CHANGE;
	   	} else if( SelectStatus == CHANGE ){
			Now = Snum;
	   		SelectStatus = ADD;
		}
		rewrite_mode_button( gcb );
		flush_image( buff );
		break;
	case XK_Delete:
		if( EventStatus != ESSELECT ){
			break;
		}
		(*Func[Mode].function)( DRAWDEL, NULL );
		if( Func[Mode].delable == DEL_ABLE ){
			if( del_proc() ){
				break;
			}
		}
		flush_image( buff );
		break;
	case XK_Left:
		if( AltStatus ){
			if( ShiftStatus ){
				move_display( CONFIG_LEFT, 100 );
			} else {
				move_display( CONFIG_LEFT, 1 );
			}
			draw_bar( 1 );
			flush_image( buff );
		} else if( EventStatus == ESSELECT ){
/* add D1 arrow key FIX */
			(*Func[Mode].function)( DRAWPRESS, &Sel[Now] );
			pos.real.y = Sel[Now].y;
			pos.real.x = Sel[Now].x;
			position_r2w( &pos );
			if( ShiftStatus ){
/*				Sel[Now].x = 0;*/
				pos.window.x = 0;
			} else { /*if( Sel[Now].x ){*/
/*				Sel[Now].x--;*/
				pos.window.x--;
			}
			position_w2r( &pos );
			Sel[Now].x = pos.real.x;
			if( Sel[Now].x < 0 ){
				Sel[Now].x = 0;
			}
/*			(*Func[Mode].function)( DRAWMOTION, &Sel[Now] );*/
			(*Func[Mode].function)( DRAWRELEASE, &Sel[Now] );
		}
		break;
	case XK_Right:
		if( AltStatus ){
			if( ShiftStatus ){
				move_display( CONFIG_RIGHT, 100 );
			} else {
				move_display( CONFIG_RIGHT, 1 );
			}
			draw_bar( 1 );
			flush_image( buff );
		} else if( EventStatus == ESSELECT ){
/* add D1 arrow key FIX */
			(*Func[Mode].function)( DRAWPRESS, &Sel[Now] );
			pos.real.y = Sel[Now].y;
			pos.real.x = Sel[Now].x;
			position_r2w( &pos );
			if( ShiftStatus ){
/*				Sel[Now].x = Image->width - 1;*/
				pos.window.x = Image->width - 1;;
			} else { /*if( Sel[Now].x < Image->width - 1 ){*/
/*				Sel[Now].x++;*/
				pos.window.x++;
			}
			position_w2r( &pos );
			Sel[Now].x = pos.real.x;
			if( Sel[Now].x >= buff->width ){
				Sel[Now].x = buff->width;
			}
/*			(*Func[Mode].function)( DRAWMOTION, &Sel[Now] );*/
			(*Func[Mode].function)( DRAWRELEASE, &Sel[Now] );
		}
		break;
	case XK_Up:
		if( AltStatus ){
			if( ShiftStatus ){
				move_display( CONFIG_UP, 100 );
			} else {
				move_display( CONFIG_UP, 1 );
			}
			draw_bar( 0 );
			flush_image( buff );
		} else if( EventStatus == ESSELECT ){
/* add D1 arrow key FIX */
			(*Func[Mode].function)( DRAWPRESS, &Sel[Now] );
			pos.real.y = Sel[Now].y;
			pos.real.x = Sel[Now].x;
			position_r2w( &pos );
			if( ShiftStatus ){
/*				Sel[Now].y = 0;*/
/* add D1 arrow key FIX */
				pos.window.y = 0;
			} else { /*if( Sel[Now].y ){*/
/*				Sel[Now].y--;*/
				pos.window.y--;
			}
			position_w2r( &pos );
			Sel[Now].y = pos.real.y;
			if( Sel[Now].y < 0 ){
				Sel[Now].y = 0;
			}
/*			(*Func[Mode].function)( DRAWMOTION, &Sel[Now] );*/
			(*Func[Mode].function)( DRAWRELEASE, &Sel[Now] );
		}
		break;
	case XK_Down:
		if( AltStatus ){
			if( ShiftStatus ){
				move_display( CONFIG_DOWN, 100 );
			} else {
				move_display( CONFIG_DOWN, 1 );
			}
			draw_bar( 0 );
			flush_image( buff );
		} else if( EventStatus == ESSELECT ){
/* add D1 arrow key FIX */
			(*Func[Mode].function)( DRAWPRESS, &Sel[Now] );
			pos.real.y = Sel[Now].y;
			pos.real.x = Sel[Now].x;
			position_r2w( &pos );
			if( ShiftStatus ){
/*				Sel[Now].y = Image->height - 1;*/
				pos.window.y = Image->height - 1;
			} else {
/*				Sel[Now].y++;*/
				pos.window.y++;
			}
			position_w2r( &pos );
			Sel[Now].y = pos.real.y;
			if( Sel[Now].y >= buff->height ){
				Sel[Now].y = buff->height;
			}
/*			(*Func[Mode].function)( DRAWMOTION, &Sel[Now] );*/
			(*Func[Mode].function)( DRAWRELEASE, &Sel[Now] );
		}
		break;
	case XK_Shift_L:
		ShiftStatus = 1;
/* add D1 arrow key FIX */
		break;
	case XK_Shift_R:
		AltStatus = 1;
		break;
	case XK_Tab:
		AltStatus = 1;
		break;
	default:
		break;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function keyrelease()                                   */
/*                                                                    */
/*            args  XEvent *event  :  KeyRelease Event                */
/*                  struct Shm *buff : The Image Data                 */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: KeyRelease Event handler.                     */
/*                                                                    */
/**********************************************************************/
static void keyrelease( XEvent *event, struct Shm *buff )
{
	KeySym keysym;
	XComposeStatus compose;
	char buf[10];

	XLookupString( (XKeyEvent *)event, buf, 10, &keysym, &compose );
	switch( keysym ){
	case XK_Shift_L:
		ShiftStatus = 0;
		break;
/* add D1 arrow key FIX */
	case XK_Shift_R:
		AltStatus = 0;	
	        break;
	case XK_Tab:
		AltStatus = 0;	
		break;
	default:
		break;
	}
}

/**********************************************************************/
/*                                                                    */
/*            function buttonpress()                                  */
/*                                                                    */
/*            args  struct Positions *p : Positions                   */
/*                  struct Shm *buff   :  The Image Data              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: KeyPress Event handler at Image Window(W).    */
/*                                                                    */
/**********************************************************************/
static void buttonpress( struct Positions *p, struct Shm *buff )
{
	struct Pos pos;

	pos.x = p->real.x;
	pos.y = p->real.y;
	(*Func[Mode].function)( DRAWPRESS, &pos );
}


/**********************************************************************/
/*                                                                    */
/*            function buttonrelease()                                */
/*                                                                    */
/*            args  struct Positions *p : Positions                   */
/*                  struct Shm *buff    : The Image Data              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: ButtonRelease Event handler at Image Window(W)*/
/*                                                                    */
/**********************************************************************/
static void buttonrelease( struct Positions *p, struct Shm *buff )
{
	int prev;	/* C2 add */

	Sel[Now].x = p->real.x;
	Sel[Now].y = p->real.y;
	prev = Now;	/* C2 add */
	(*Func[Mode].function)( DRAWRELEASE, &Sel[Now] );
	if( Mode == SKFREE && Now == prev ){	/* C2 add to ignore same point */
		return;
	}
	if( SelectStatus == ADD ){
		Snum++;
		if( Snum >= Func[Mode].maxselect ){
			SelectStatus = CHANGE;
			rewrite_mode_button( gcb );
			flush_image( buff );
		} else if( SKind == SKPOS && Snum >= MaxPos ){
			if( Mode != SKFREE ){
				SelectStatus = CHANGE;
				rewrite_mode_button( gcb );
			} else {
				Snum--;
				printf( "ignored\n" );
			}
			flush_image( buff );
		} else if( SKind == SKFILE && Snum >= MaxPos ){
			if( realloc_sel( MAXPOSITION ) == -1 ){
				SelectStatus = CHANGE;
				rewrite_mode_button( gcb );
				flush_image( buff );
			} else {
				Now++;
			}	
		} else {
			Now++;
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*            function command_press()                                */
/*                                                                    */
/*            args  XEvent *event  : ButtonPress Event                */
/*                  struct Shm *buff : The Image Data                 */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: ButtonPress Event handler at Button Windows.  */
/*                                                                    */
/**********************************************************************/
static void command_press( XEvent *event, struct Shm *buff )
{
	if( event->xbutton.window == Wok ){
		reply_selection();
		rewrite_button();
		return;
	} else if( event->xbutton.window == Wesc ){
		Snum = 0;
		put_image( buff, SHOW_CLEAR );
		reply_selection();
		rewrite_button();
		return;
	} else if( event->xbutton.window == Wmode ){
		if( Func[Mode].delable != DEL_DISABLE 
		   || Snum > DefaultMax[Mode]){
			if( SelectStatus == ADD ){
				if( Snum == 0 ){
					return;
				}
				Now = Snum - 1;
				SelectStatus = CHANGE;
			} else if( SelectStatus == CHANGE ){
				Now = Snum;
				SelectStatus = ADD;
			}
			rewrite_mode_button( gcb );
		}
	} else if( event->xbutton.window == Wdel ){
		(*Func[Mode].function)( DRAWDEL, NULL );
		if( Func[Mode].delable == DEL_ABLE ){
			if( del_proc() ){
				return;
			}
		}
	} else if( event->xbutton.window == Ws1 ){
 		if( change_button_size( WINDOW_SIZE_DOUBLE ) != 0 ){
			return;
		}
	} else if( event->xbutton.window == Ws2 ){
		if( change_button_size( WINDOW_SIZE_HALF ) != 0 ){
			return;
		}
	} else if( event->xbutton.window == Wl1 ){
		if( change_button_size( WINDOW_LOOPE_DOUBLE ) != 0 ){
			return;
		}
	} else if( event->xbutton.window == Wl2 ){
		if( change_button_size( WINDOW_LOOPE_HALF ) != 0 ){
			return;
		}
	} else if( event->xbutton.window == Wwbar ){
		if( event->xbutton.button == Button3 ){
			move_display( CONFIG_RIGHT, 1 );
		} else if( event->xbutton.button == Button1 ){
			move_display( CONFIG_LEFT, 1 );
		}
		draw_bar( 1 );
	} else if( event->xbutton.window == Whbar ){
		if( event->xbutton.button == Button3 ){
			move_display( CONFIG_DOWN, 1 );
		} else if( event->xbutton.button == Button1 ){
			move_display( CONFIG_UP, 1 );
		}
		draw_bar( 0 );
	} else if( event->xbutton.window == Winit ){
		if( change_button_size( WINDOW_SIZE_EQ ) != 0 ){
			return;
		}
	}
	flush_image( buff );
}


/**********************************************************************/
/*                                                                    */
/*            function gedo_ErrorHandler()                            */
/*                                                                    */
/*            args  Display *d  :  Display of X11                     */
/*                  XErrorEvent *err   : Error Event occured          */
/*                                                                    */
/*            return int   :                                          */
/*                                                                    */
/*            function: error hancler for X11.                        */
/*                                                                    */
/**********************************************************************/
static int gedo_ErrorHandler( Display *d, XErrorEvent *err )
{
	char msg[80];

	*Viewer_Status = -1;
	XGetErrorText( d, err->error_code, msg, 80 );
	fprintf( stderr, "X Error of failed request:  %s\n", msg );
	fprintf( stderr, "  Major opcode of failed request: %d\n", err->error_code );
	fprintf( stderr, "  Resource id in failed request: 0x%lx\n", err->resourceid );
	fprintf( stderr, "  Serial number of failed request:  %ld\n", err->serial );
	fprintf( stderr, "  Current serial number in output stream: %d\n", err->request_code );
	fprintf( stderr, "Clean and exit\n" );
	exit( 1 );
}

/**********************************************************************/
/*                                                                    */
/*            function gedo_FatalIOErrorHandler()                     */
/*                                                                    */
/*            args  Display *d   :  Display of X11                    */
/*                                                                    */
/*            return int   :                                          */
/*                                                                    */
/*            function: Error Handler of X11(Fatal I/O)               */
/*                                                                    */
/**********************************************************************/
static int gedo_FatalIOErrorHandler( Display *d )
{
	*Viewer_Status = -1;
	fprintf( stderr, "X Fatal IO Detected\n" );
	fprintf( stderr, "Clean and exit\n" );
	exit( 1 );
}

/**********************************************************************/
/*                                                                    */
/*            function report_expose()                                */
/*                                                                    */
/*            args  XEvent *ev : Expousure Event                      */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Report the detail of Exposure Event           */
/*                      for debug.                                    */
/**********************************************************************/
void report_expose( XEvent *ev )
{
	static char *str[2] = { "W", "W0" };
	char *p;
	int sx, sy, ex, ey, h, w;

	if( ev->xexpose.window == W ){
		p = str[0];
	} else {
		return;
		p = str[1];
	}
	sx = ev->xexpose.x;
	sy = ev->xexpose.y;
	w = ev->xexpose.width;
	h = ev->xexpose.height;
	ex = sx + w;
	ey = sy + h;
	printf( "expose(%s) %d (%d,%d)-(%d,%d) %d:%d\n", p, ev->xexpose.count,
	       sx, sy, ex, ey, w, h );
}

/**********************************************************************/
/*                                                                    */
/*            function loader_ctr_c()                                 */
/*                                                                    */
/*            args  int signo : SIGNAL number                         */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: To stop if SIGINT, signal handler.            */
/*                                                                    */
/**********************************************************************/
void loader_ctrl_c( int signo )
{
	signal( SIGINT, loader_ctrl_c );
	if( EventStatus == ESSELECT ){
		longjmp( jmpbuf, 1 );
	} else {
		CtrlcFlag = 1;
	}
}

void loader_signal( int signo )
{
	fprintf( stderr, "segmentation fault\n" );
/*	reply_selection();  delete C3 */
	XDestroyWindow( D, W0 );
	XCloseDisplay( D );
	D = 0;		/* add C3 */
	exit( 2 );
}

/**********************************************************************/
/*                                                                    */
/*            function flush_queues()                                 */
/*                                                                    */
/*            args  NONE                                              */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: To stop if SIGINT, if showing is, read all    */
/*                      queued events and flush.                      */
/**********************************************************************/
void flush_queues( void )
{
	int num, i;
	XEvent event;

	num = XEventsQueued( D, QueuedAfterReading );
	for( i = 0; i < num; i++ ){
		XNextEvent( D, &event );
	}
}

/* add C3 */
void destroy_resource( void )
{
	if( D ){
		XDestroyWindow( D, W );
		XDestroyWindow( D, W0 );
		XCloseDisplay( D );
	}
	kill( getppid(), SIGUSR1 );
}

/**********************************************************************/
/*                                                                    */
/*            function disp_image()                                   */
/*                                                                    */
/*            args  struct Shm *buff : The Image Data                 */
/*                                                                    */
/*            return NONE                                             */
/*                                                                    */
/*            function: Event Loop of viewer(2nd).                    */
/*                                                                    */
/**********************************************************************/
void disp_image( struct Shm *buff )
{
	XEvent event;
	struct Positions pos;
	struct Pos p;
	int flag_w, flag_w0;

	ShiftStatus = 0;
	Viewer_Status = &buff->status;

	XSetErrorHandler( gedo_ErrorHandler );
	XSetIOErrorHandler( gedo_FatalIOErrorHandler );
	kill( getppid(), SIGUSR1 );
	flag_w = flag_w0 = 1;
	signal( SIGSEGV, loader_signal );
	signal( SIGINT, loader_ctrl_c );
	signal( SIGUSR1, SIG_IGN );
	atexit( destroy_resource );	/* add C3 */
	while( 1 ){
		if( setjmp( jmpbuf ) ){
			if( EventStatus == ESSELECT ){
				Snum = 0;
				put_image( buff, SHOW_CLEAR );
				reply_selection();
				rewrite_button();
			}
		}
		if( CtrlcFlag ){
			flush_queues();
		}
		CtrlcFlag = 0;
		buff->status = 0;
		XNextEvent( D, &event );
		buff->status = 1;
		if( event.xexpose.send_event == True ){
			event_handler( &event, buff );
			event.xexpose.send_event = False;
			continue;
		} else {
		}
                switch( event.type ){
		case MappingNotify:
			XRefreshKeyboardMapping( (XMappingEvent *)&event );
			break;
		case Expose:
			if( event.xexpose.window == W ){
		  		if( flag_w ){
					flush_image( buff );
					if( event.xexpose.count == 0 ){
						flag_w = 1;
					} else {
						flag_w = 0;
					}
				} else {
					if( event.xexpose.count == 0 ){
						flag_w = 1;
					}
				}
			} else if( event.xexpose.window == W0 ){
		  		if( flag_w0 ){
					rewrite_button();
					if( event.xexpose.count == 0 ){
						flag_w0 = 1;
					} else {
						flag_w0 = 0;
					}
				} else {
					if( event.xexpose.count == 0 ){
						flag_w0 = 1;
					}
				}
			}
			break;
		case KeyPress:
			keypress( &event, buff );
			break;
		case KeyRelease:
			keyrelease( &event, buff );
			break;
		case ButtonPress:
			if( event.xbutton.window != W ){
				command_press( &event, buff );
				break;
			}
			pos.window.x = event.xbutton.x;
			pos.window.y = event.xbutton.y;
			position_w2r( &pos );
			if( pos.real.x >= buff->width || 
			   pos.real.x < 0 ||
			   pos.real.y >= buff->height ||
			   pos.real.y < 0 ){
				break;
			}
			if( EventStatus == ESSELECT ){
				buttonpress( &pos, buff );
			} else if( EventStatus == ESDISPLAY ){
				EventStatus = ESMASK;
				put_image( buff, SHOW_MASK );
			} else {
				put_image( buff, SHOW_CLEAR );
				EventStatus = ESDISPLAY;
			}
			break;
		case ButtonRelease:
			pos.window.x = event.xbutton.x;
			pos.window.y = event.xbutton.y;
			position_w2r( &pos );
			if( pos.real.x >= buff->width || 
			   pos.real.x < 0 ||
			   pos.real.y >= buff->height ||
			   pos.real.y < 0 ){
				break;
			}	
			if( EventStatus == ESSELECT ){
				buttonrelease( &pos, buff );
			}
			break;
		case MotionNotify:
			pos.window.x = event.xmotion.x;
			pos.window.y = event.xmotion.y;
			position_w2r( &pos );
			if( pos.real.x >= buff->width || 
			   pos.real.x < 0 ||
			   pos.real.y >= buff->height ||
			   pos.real.y < 0 ){
				break;
			}
			p.x = pos.real.x;
			p.y = pos.real.y;
			if( EventStatus == ESSELECT ){
				(*Func[Mode].function)( DRAWMOTION, &p );
			}
			break;
		default:
			break;
		}
		XFlush( D );
	}
}
