/**********************************************************************/
/*                                                                    */
/*               ngui.c   ngui, a command rancher of GEDO             */
/*                                                                    */
/*                by K.sugiura   Ver 0.995 for GEDO of LINUX          */
/*                                                                    */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <tcl.h>
#include <tk.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include "../common.h"

static int Ppid;		/* parent process(GEDO) ID */
static int Pid;			/* self process ID */
static int MesId;		/* IPC message ID */
static int Gui;

struct TYPE {
     char *name;
     char *loader;
};

struct TYPE *Type;
int TypeNum = 0;
int TypeAlloc = 0;

/**********************************************************************/
/*                                                                    */
/*             function send_str()                                    */
/*                                                                    */
/*             char *str : message to be send                         */
/*                                                                    */
/*             return  int : 0==success  -1==fail                     */
/*                                                                    */
/*             function: Send the selection message to  GEDO.         */
/*                                                                    */
/**********************************************************************/
int send_str( char *str )
{
	struct smes {
		long mtype;
		char str[256];
	} mes;

	strcpy( mes.str, str );
	strcat( mes.str, "\n" );
	if( Gui ){
	     mes.mtype = Gui;
	} else {
	     mes.mtype = Pid;
	}
	if( msgsnd( MesId, (struct msgbuf *)&mes, strlen(mes.str), 0 ) == -1 ){
	        perror("send_str");
		return( -1 );
	}
	kill( Ppid, SIGUSR2 );
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*            function Trans_Proc()                                   */
/*                                                                    */
/*            args  ClientData cd  : call back data, not used.        */
/*                  Tcl_Interp *interp : TclTk interpreter handle     */
/*                  int argc  :  argment number                       */
/*                  char **argv : argments                            */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: trans  command on Tcl/Tk loadgui.tcl          */
/*                                                                    */
/**********************************************************************/
int Trans_Proc( ClientData cd, Tcl_Interp *interp, int argc, char **argv )
{
	if( argc < 2 ){
		return( TCL_ERROR );
	}
	if( send_str( argv[1] ) == 0 ){
		return( TCL_OK );
	} else {
		return( TCL_ERROR );
	}
}


/**********************************************************************/
/*                                                                    */
/*            function Trans_Init()                                   */
/*                                                                    */
/*            args  Tcl_Interp *interp : TclTk interpreter handle     */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: set the command trans  to the interp.         */
/*                                                                    */
/**********************************************************************/
int Trans_Init( Tcl_Interp *interp )
{
	Tcl_CreateCommand( interp, "gsend", Trans_Proc, (ClientData *)NULL,
			  (Tcl_CmdDeleteProc *)NULL );
	return( TCL_OK );
}

/**********************************************************************/
/*                                                                    */
/*             function get_MesID()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : IPC message ID                           */
/*                                                                    */
/*             function: Get IPC message ID.                          */
/*                                                                    */
/**********************************************************************/
int get_MesID( void )
{
	char *env;
	int id;

	Ppid = getppid();
	Pid  = getpid();
	signal( SIGINT, SIG_IGN );

	if( (env = getenv( "GEDO_IPCMES" )) == NULL ){
		id = IPCMSGID;
	} else {
		id = atoi( env );
		if( id <= 0 ){
			id = IPCMSGID;
		}
	}
        if( (MesId = msgget( id, IPC_CREAT | 0666 )) == -1 ){
		return( -1 );
	}
	return( MesId );
}


/**********************************************************************/
/*                                                                    */
/*            function main()                                         */
/*                                                                    */
/*            args   NONE                                             */
/*                                                                    */
/*            return int : 0==success  -1==error                      */
/*                                                                    */
/*            function: main routine                                  */
/*                                                                    */
/**********************************************************************/
int main( int argc, char **argv )
{
	Tcl_Interp *interp;
	char buff[256];
	char *e, *path;

        path = NULL;
	++argv;
	while( *argv ){
	     if( strncmp( *argv, "gui=", 4 ) == 0 ){
		  Gui = atoi( *argv + 4 );
	     } else {
		  path = *argv;
		  break;
	     }
	     argv++;
	}

        if( get_MesID() == -1 ){
		printf( "gui:cannto make IPC message\n" );
		exit( 1 );
	}

	interp = Tcl_CreateInterp();
	if( Tcl_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "%s\n", interp->result );
		exit( 1 );
	}
	if( Tk_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "Tk_Init failed:%s\n", interp->result );
		exit( 1 );
	}

	if( Trans_Init( interp ) == TCL_ERROR ){
		fprintf( stderr, "Dith_Init failed:%s\n", interp->result );
		exit( 1 );
	}

	Tcl_SetVar( interp, "FromC", "1", 0 );
	if( path ){
		Tcl_SetVar( interp, "path", path, 0 );
	} else {
		Tcl_SetVar( interp, "path", "", 0 );
	}
	if( (e = getenv( "GEDO_LWP" )) == NULL ){
		Tcl_SetVar( interp, "lwp", "", 0 );
	} else {
		Tcl_SetVar( interp, "lwp", e, 0 );
	}
	if( (e = getenv("GEDO_PATH")) == NULL ){
	     sprintf( buff, "./ngui.tcl" );
	} else {
	     sprintf( buff, "%s../ngui.tcl", e );
	}
	if( Tcl_EvalFile( interp, buff ) == TCL_ERROR ){
		fprintf( stderr, "%s TCL error\n", buff );
		perror("");
		exit( 1 );
	}

	Tk_MainLoop();
	return( 0 );
}
