/**********************************************************************/
/*                                                                    */
/*           viewer.c   viewer main routine                           */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                                                                    */
/*           External Interfaces  NONE                                */
/*                                                                    */
/**********************************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <memory.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>

#include "viewer.h"
#define MAXVIEWER	10		/* maximum of viewer(2) process */

static struct proc Proc;		/* process struct */

static struct Message Mes;		/* message struct from GEDO */
static struct File File[MAXVIEWER];	/* plug original struct */
static int FileNum = 0;			/* viewer(2) number */

static char *FatalForm 			/* fatal error message */
= "viewer fatal error -- %s: end\n";

extern Display 	 *D;			/* viewer(2) X11 Display */
Display 	 *D0;			/* viewer(1) X11 Display */
extern Window W;			/* viewer(2) image Window */
extern Window W0;			/* viewer(2) top Window */
int MesId;				/* Message ID of IPC message from GEDO */
static int CtrlCFlag = 0;		/* ^C detection */
static int CatchFlag = 0;		/* add C3, receive SIGUSR1 from viewer2 */
volatile int Abort_flag = 0;		/* add D0, Ctrl+C and Abort flag */
/* D2 add .. atexit() of viewer2 destroy all childs, it is BAD.*/
int Parent = 1;				/* is parent(viewer1) or child */

/**********************************************************************/
/*                                                                    */
/*             function set_shm()                                     */
/*                                                                    */
/*             struct File *file  : process infomation                */
/*                                                                    */
/*             return  int : 0==success errorcode==fail               */
/*                                                                    */
/*             function: IPCshm alloc the core.                       */
/*                                                                    */
/**********************************************************************/
int set_shm( struct File *file )
{
	file->buf = (struct Shm *)shmat( file->mes->shmid, NULL, 0 );
	if( file->buf == (struct Shm *)-1 ){
		return( ESHMAT );
	}
	return( 0 );
}
  
/**********************************************************************/
/*                                                                    */
/*             function close_shm()                                   */
/*                                                                    */
/*             struct File *file : process infomation                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: free IPCshm                                  */
/*                                                                    */
/**********************************************************************/
void close_shm( struct File *file )
{
	shmdt( (char *)file->buf );
}


/**********************************************************************/
/*                                                                    */
/*             function make_viewer()                                 */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int : 0==success errorcode=fail                */
/*                                                                    */
/*             function: fork() and create viewer(2).                 */
/*                                                                    */
/**********************************************************************/
int make_viewer( struct proc *proc )
{
	struct Message *mes;
	Visual *vis;
	struct WinSize win;
	struct shmid_ds dat;
	int len;
	int ret;

	mes = proc->file->mes;

	if( (D0 = XOpenDisplay( NULL )) == NULL ){
		return( ENOAVAILX );
	}
	if( (D = XOpenDisplay( NULL )) == NULL ){
		return( ENOAVAILX );
	}
	vis = DefaultVisual( D, 0 );
	if( vis->class != TrueColor && vis->class != DirectColor ){
		return( ENOAVAILVC );
	}

	W0 = XCreateSimpleWindow( D, RootWindow( D, 0 ), 100, 100,
		200, 200, 1, 0, WhitePixel( D, 0 ) );

	W = XCreateSimpleWindow( D, W0, 0, 0,
			200, 200, 0, 1, 0 );

	if( (proc->file->pid = fork()) == 0 ){
	        Parent = 0;
		if( shmctl( proc->file->mes->shmid, IPC_STAT, &dat ) == -1 ){
			return( ESHMAT );
		}
		proc->shm_max = (dat.shm_segsz - sizeof(struct Shm)) / sizeof(unsigned int);
		proc->file->buf = (struct Shm *)shmat( proc->file->mes->shmid, NULL, 0 );
		if( proc->file->buf == (struct Shm *)-1 ){
			return( ESHMAT );
		}

		win.command = WIN_INIT;
		win.style = 0;
		win.image.x = proc->file->buf->width;
		win.image.y = proc->file->buf->height;
		init_scale( &win );
		resize_window( &win );
		len = strlen( mes->u.data ) + 1;
		ret = create_image( proc->shm_max, &mes->u.data[len], proc->file->buf );
		if( ret != 0 ){
			kill( getppid(), SIGUSR1 ); 
			return( 0 );
		}
		disp_image( proc->file->buf );
		/* NOTREACHED */
	} else {
		proc->file->W0 = W0;
		proc->file->W = W;
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function send_disp_mes()                               */
/*                                                                    */
/*             int spec  : 0==REFRESH  1==LOADFILE 2==SHOWMASK        */
/*             int mask  : mask No.                                   */
/*             struct proc *proc  :                                   */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: XSendEvent to viewer(2) for display image.   */
/*                                                                    */
/**********************************************************************/
void send_disp_mes( int spec, int mask, struct proc *proc )
{
	int ret;
	XEvent event;

	event.xclient.type = ClientMessage;
	event.xclient.serial = 0L;
	event.xclient.send_event = 1;
	event.xclient.display  = D0;
	event.xclient.window = proc->file->W;
	event.xclient.message_type = ESDISPLAY;
	event.xclient.format = 32;
	event.xclient.data.l[0] = spec;
	event.xclient.data.l[1] = mask;
	ret = XSendEvent( D0, proc->file->W, True, ButtonPressMask, &event );
	XFlush( D0 );
}

/**********************************************************************/
/*                                                                    */
/*             function send_viewer_mes()                             */
/*                                                                    */
/*             struct proc *proc   : process infomation               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: XSendEvent to viewer(2) for UP and DOWN.     */
/*             D2: created                                            */
/**********************************************************************/
/* D2 multi viewer selection showing */
void send_viewer_mes( struct proc *proc, int flag )
{
	int ret;
	XEvent event;
	struct Message *mes;

	mes = proc->file->mes;
	memset( &event, 0, sizeof(event) );
	event.xclient.type = ClientMessage;
	event.xclient.serial = 0L;
	event.xclient.send_event = 1;
	event.xclient.display  = D0;
	event.xclient.window = proc->file->W;
	event.xclient.message_type = flag;
	event.xclient.format = 8;
	ret = XSendEvent( D0, proc->file->W, True, ButtonPressMask, &event );
	XFlush( D0 );
}
/* end D2 */

/**********************************************************************/
/*                                                                    */
/*             function send_config_mes()                             */
/*                                                                    */
/*             struct proc *proc   : process infomation               */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: XSendEvent to viewer(2) for configuration.   */
/*                                                                    */
/**********************************************************************/
void send_config_mes( struct proc *proc )
{
	int ret;
	XEvent event;
	struct Message *mes;

	mes = proc->file->mes;
	event.xclient.type = ClientMessage;
	event.xclient.serial = 0L;
	event.xclient.send_event = 1;
	event.xclient.display  = D0;
	event.xclient.window = proc->file->W;
	event.xclient.message_type = ESCONFIG;
	event.xclient.format = 8;
	strcpy( event.xclient.data.b, mes->u.data );
	ret = XSendEvent( D0, proc->file->W, True, ButtonPressMask, &event );
	XFlush( D0 );
}

/**********************************************************************/
/*                                                                    */
/*             function send_kill_mes()                               */
/*                                                                    */
/*             args   struct proc *proc : process infomation          */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: XSendEvent to viewer(2) for KILL process.    */
/*                                                                    */
/**********************************************************************/
void send_kill_mes( struct File *file )
{
	int ret;
	XEvent event;

	event.xclient.type = ClientMessage;
	event.xclient.serial = 0L;
	event.xclient.send_event = 1;
	event.xclient.display  = D0;
	event.xclient.window = file->W;
	event.xclient.message_type = ESKILL;
	event.xclient.format = 32;
	event.xclient.data.l[0] = 0;
	ret = XSendEvent( D0, file->W, True, ButtonPressMask, &event );
	XFlush( D0 );
}

/**********************************************************************/
/*                                                                    */
/*             function send_select_mes()                             */
/*                                                                    */
/*             struct Sbuf *sb : Selection data                       */
/*             int pid         : process ID of viewer(2)              */
/*                                                                    */
/*             return 0==success  -1==fail                            */
/*                                                                    */
/*             function: XSendEvent to viewer(2) for selection.       */
/*                                                                    */
/**********************************************************************/
int send_select_mes( struct proc *proc )
{
	int ret;
	XEvent event;
	struct Sbuf *sb;
	int pid;
	struct sbufmes {
		long mtype;
		struct Sbuf sb;
	} ms;

	sb = &proc->file->mes->u.sb;
	pid = proc->file->pid;
	event.xclient.type = ClientMessage;
	event.xclient.serial = 0L;
	event.xclient.send_event = 1;
	event.xclient.display  = D0;
	event.xclient.window = proc->file->W;
	event.xclient.message_type = ESSELECT;

	event.xclient.format = 32;
	event.xclient.data.l[0] = sb->selnum;
	event.xclient.data.l[1] = sb->defnum;
	event.xclient.data.l[2] = sb->skind;
	event.xclient.data.l[3] = sb->maxnum;
	
	ret = XSendEvent( D0, proc->file->W, True, ButtonPressMask, &event );
	XFlush( D0 );
	if( sb->defnum || sb->skind == SKFILE ){
		memcpy( &ms.sb, sb, sizeof(struct Sbuf) ); 
		ms.mtype = pid;
		if( msgsnd( MesId, (struct msgbuf *)&ms, sizeof(struct sbufmes) - sizeof(long), 0 ) == -1 ){
			return( -1 );
		}
	}
	if( msgrcv( MesId, (struct msgbuf *)&ms, sizeof(struct sbufmes) - sizeof(long), getpid(), 0 ) == -1 ){
		return( -1 );
	}
	sb->defnum = ms.sb.defnum;
	if( sb->skind == SKPOS ){
		memcpy( sb->def, ms.sb.def, sizeof(struct Pos) * ms.sb.defnum );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function end_proc()                                    */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*             int eno           : return errorcode to GEDO           */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Close viewer(1) process.                     */
/*                                                                    */
/**********************************************************************/
/* void end_proc( struct proc *proc, int errno ) */
/* Batting with system errno, errno changed int to (*__errno_localtion()) 
   for REENTRANT module, so... */ 
void end_proc( struct proc *proc, int eno )
{
	struct Message *mes;
	int i;

	mes = proc->file->mes;
	if( eno ){
		sprintf( mes->u.data, "fatal error: viewer suicided" );
		mes->status  = STFATAL;
	} else {
		sprintf( mes->u.data, "viewer killed" );
		mes->status  = STDONE;
	}
	for( i = 0; i < FileNum; i++ ){
		if( File[i].buf ){
			shmdt( (char *)File[i].buf );
		}
		if( File[i].pid ){
			send_kill_mes( &File[i] );
		}
	}
	FileNum = 0;
	mes->server = SVIEWER;
	mes->errorcode  = eno;
	mes->datasize = strlen( mes->u.data );
	mes->mtype = (long)mes->pid[SMAIN];
	msgsnd( MesId, (struct msgbuf *)mes, sizeof(struct Message) - sizeof(long), 0 );
	sleep( 1 );
	exit( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function reply()                                       */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*             int dest          : destination of reply               */
/*                                 0==GEDO 1==LOADER 2==SELECTER      */
/*                                 3==EFFECTER                        */
/*             return  int  : 0==success or abort -1==error           */
/*                                                                    */
/*             function: Send message to another process.             */
/*                                                                    */
/**********************************************************************/
int reply( struct proc *proc, int dest )
{
	struct Message *mes;

	if( CtrlCFlag ){
		CtrlCFlag = 0;
		return( 0 );
	}
	mes = proc->file->mes;
	if( mes->pid[dest] <= 1 ){
		return( -1 );
	}
	mes->mtype = (long)mes->pid[dest];
	if( msgsnd( MesId, (struct msgbuf *)mes, sizeof(struct Message) - sizeof(long), 0 ) == -1 ){
		end_proc( proc, EPIPEWRITE );
	}

#ifdef DEBUG
	printf( "viewer:reply:kill dest=%d pid=%d\n", dest, mes->pid[dest] );
#endif
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function show_command()                                */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int  : 0==success                              */
/*                                                                    */
/*             function: show procedure from another process          */
/*                       and transport to viewer(2).                  */
/**********************************************************************/
int show_command( struct proc *proc )
{
	struct Message *mes;

	mes = proc->file->mes;
	if( mes->subcommand == SCCONFIG && mes->command == CSHOW ){
		send_config_mes( proc );
/* D2 multi viewer selection showing */
	} else if( mes->subcommand == SCDOWN && mes->command == CSHOW ){
	     send_viewer_mes( proc, ESDOWN );
	     mes->errorcode = ENOERROR;
	} else if( mes->subcommand == SCUP && mes->command == CSHOW ){
	     send_viewer_mes( proc, ESUP );
	     mes->errorcode = ENOERROR;
/* end D2 */
	} else if( mes->subcommand == SCEXEC ){
		switch( mes->command ){
		case CLOAD:
			if( mes->server != SLOADER && mes->server != SMAIN ){
				mes->errorcode = EILLSERVER;
				break;
			}
			send_disp_mes( LOADFILE, 0, proc );
			break;
		case CEFFECT:
			if( mes->server != SEFFECTER ){
				mes->errorcode = EILLSERVER;
				break;
			}
			if( (mes->comatr & AINT) == 0 ){
				if( proc->file->buf->status == 0 ){
					send_disp_mes( REFLESH, 0, proc );
				}
			} else {
				send_disp_mes( REFLESH, 0, proc );
			}
			break;
		case CSHOW:
			if( mes->server != SMAIN ){
				mes->errorcode = EILLSERVER;
				break;
			}
			if( (mes->comatr & AINT) == 0 ){
				if( proc->file->buf->status == 0 ){
					send_disp_mes( SHOWMASK, mes->u.data[0] - '0', proc );
				}
			} else {
				send_disp_mes( SHOWMASK, mes->u.data[0] - '0', proc );
			}
			break;
		default:
			mes->errorcode = ENOCOMMAND;
			break;
		}
	} else {
		mes->errorcode = ENOAVAILSUBCOMMAND;
	}

	mes->receiver = SMAIN;
	if( mes->errorcode != ENOERROR ){
		mes->status = STWARNING;
	} else {
		mes->status = STDONE;
	}
	mes->datasize = strlen( mes->u.data );
	reply( proc, SMAIN );
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function select_command()                              */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int : 0==success -1==illegal data              */
/*                                                                    */
/*             function: select command from GEDO(selection) and      */
/*                       from selecter(show) proc.                    */
/**********************************************************************/
int select_command( struct proc *proc )
{
	struct Message *mes;

	mes = proc->file->mes;
	if( mes->subcommand != SCEXEC ){
		mes->status = STWARNING;
		mes->errorcode = ENOAVAILSUBCOMMAND;
		mes->datasize = 0;
		reply( proc, mes->server );
		return( -1 );
	}
	if( mes->server == SMAIN ){
		if( (mes->comatr & AINT) && (mes->u.sb.selnum != SKNONE) ){
			if( send_select_mes( proc ) == -1 ){
				if( CtrlCFlag ){
					mes->status = STWARNING;
					mes->errorcode = EUSRBREAK;
				} else {
					mes->status = STFATAL;
					mes->errorcode = ENOTCREATE;
					mes->datasize = 0;
				}
				mes->datasize = 0;
				reply( proc, SMAIN );
				return( 0 );
			}
		}
		if( mes->u.sb.defnum == 0 && mes->u.sb.selnum != SKNONE ){
			mes->status = STWARNING;
			mes->errorcode = EUSRBREAK;
			mes->datasize = 0;
			reply( proc, SMAIN );
		} else {
			reply( proc, SSELECTER );
		}
	} else if( mes->server == SSELECTER ){
		send_disp_mes( SHOWMASK, 0, proc );
		reply( proc, SMAIN );
	} else {
		mes->status = STWARNING;
		mes->errorcode = EILLSERVER;
		mes->datasize = 0;
		reply( proc, SMAIN );
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function purge_command()                              */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: purge command handler                        */
/*                                                                    */
/**********************************************************************/
void purge_command( struct proc *proc )
{
	struct Message *mes;
	int i;
	int shmid;

	mes = proc->file->mes;

	shmid = *(int *)mes->u.data;
	for( i = 0; i < FileNum; i++ ){
		if( File[i].shmid == shmid ){
			break;
		}
	}
	if( i == FileNum || File[i].buf == NULL || File[i].pid == 0 ){
		mes->errorcode = ENOTREADY;
		mes->u.data[0] =  '\0'; 
	} else {
		shmdt( (char *)File[i].buf );
/* D2 add .. atexit() of viewer2 destroy all childs, it is BAD.*/
/*           refine purge command, not affect SIGUSR1 handler */
		CatchFlag = 1;
		send_kill_mes( &File[i] );
		pause();
		CatchFlag = 0;
/* end D2 */
		mes->errorcode = ENOERROR;
		File[i].buf = NULL;
		File[i].shmid = 0;
		File[i].pid = 0;
	}

	mes->receiver = SMAIN;
	if( mes->errorcode != ENOERROR ){
		mes->status = STWARNING;
	} else {
		mes->status = STDONE;
	}
	mes->datasize = strlen( mes->u.data );
	reply( proc, SMAIN );
}

/**********************************************************************/
/*                                                                    */
/*             function get_MesID()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return  int : Message ID of IPC                        */
/*                                                                    */
/*             function: get message ID                               */
/*                                                                    */
/**********************************************************************/
int get_MesID( void )
{
	char *env;
	int id, ret;

	if( (env = getenv( "GEDO_IPCMES" )) == NULL ){
		id = IPCMSGID;
	} else {
		id = atoi( env );
		if( id <= 0 ){
			id = IPCMSGID;
		}
	}
        if( (ret = msgget( id, IPC_CREAT | 0666 )) == -1 ){
		return( -1 );
	}
	return( ret );
}


/**********************************************************************/
/*                                                                    */
/*             function ctrl_c()                                      */
/*                                                                    */
/*             int signo : signal No.                                 */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Ctrl+C handler                               */
/*                                                                    */
/**********************************************************************/
void ctrl_c( int signo )
{
/*	struct Message dum;*/
	struct Message mes;

	sleep(3);
	signal( SIGINT, ctrl_c );
/* D0: Ctrl+C and Abort process */
/*	while( msgrcv( MesId, (struct msgbuf *)&dum, 
                      sizeof(struct Message) - sizeof(long), 
		       getpid(), IPC_NOWAIT ) != -1 );*/
	while( msgrcv( MesId, (struct msgbuf *)&mes, 
                      sizeof(struct Message) - sizeof(long), 
		       getpid(), IPC_NOWAIT ) != -1 ){
	     if( mes.command == CKILL ){
		  Abort_flag = 1;
	     }
	}
	CtrlCFlag = 1;
}


/**********************************************************************/
/*                                                                    */
/*             function init_file()                                   */
/*                                                                    */
/*             struct proc *proc : process struct                     */
/*                                                                    */
/*             return int  : File struct found shmID number           */
/*                                                                    */
/*             function: initialize File struct (IPCshm)              */
/*                                                                    */
/**********************************************************************/
int init_file( struct proc *proc )
{
	int i;

	struct Message *mes;

	mes = proc->file->mes;
	if( proc->file->shmid == mes->shmid ){
		return( 0 );
	}
	for( i = 0; i < FileNum; i++ ){
		if( mes->shmid == File[i].shmid ){
			File[i].mes = mes;
			proc->file = &File[i];
			return( i );
		}
	}
	if( i == FileNum ){
/* D2 multi viewer selection showing */
/* D2  add for SCUP(not prepared new Window) */
		if( proc->file->mes->command == CSHOW
		  && proc->file->mes->subcommand == SCDOWN ){
		       mes->status = STDONE;
		       mes->errorcode = ENOERROR;
		       mes->datasize = 0;
		       reply( proc, SMAIN );
		       return -1;
	     }
/* end D2 */
		for( i = 0; i < FileNum; i++ ){
			if( File[i].shmid == 0 ){
				break;
			}
		}
	}

	if( i == MAXVIEWER ){
		mes->status = STWARNING;
		mes->errorcode = EVIEWERMAX;
		mes->datasize = 0;
		reply( proc, SMAIN );
		return( -1 );
	}

	File[i].mes = mes;
	File[i].shmid = mes->shmid;
	proc->file = &File[i];
	if( i == FileNum ){
		FileNum++;
	}

	if( set_shm( proc->file ) == -1 ){
		mes->status = STWARNING;
		mes->errorcode = ESHMAT;
		mes->datasize = 0;
		reply( proc, SMAIN );
		return( -1 );
	}
/* D2 multi viewer selection showing */
/* D2  add for SCUP(not prepared new Window) */
	if( proc->file->mes->command == CSHOW ){
	    if( proc->file->mes->subcommand == SCUP ){
		mes->status = STDONE;
		mes->errorcode = ENOERROR;
		mes->datasize = 0;
		reply( proc, SMAIN );
		return -1;
	    }
	}
/* end D2 */
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function free_child()                                  */
/*                                                                    */
/*             args  NONE                                             */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: atexit() proc for parent viewer.             */
/*                                                                    */
/**********************************************************************/
void free_child( void )
{
	int i;

/* D2 add .. atexit() of viewer2 destroy all childs, it is BAD.*/
	if( Parent == 0 ){
	     return;
	}
/* end D2 */
	for( i = 0; i < FileNum; i++ ){
		if( File[i].buf ){
			shmdt( (char *)File[i].buf );
		}
		if( File[i].pid ){
			send_kill_mes( &File[i] );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function process()                                     */
/*                                                                    */
/*             struct proc *proc : process infomation                 */
/*                                                                    */
/*             return  int: 0==success -1==IPC error                  */
/*                                                                    */
/*             function: Event loop of viewer(1).                     */
/*                                                                    */
/**********************************************************************/
int process( struct proc *proc )
{
	struct Message *mes;
	int ret;
	
        if( (MesId = get_MesID()) == -1 ){
		printf( "viewer:cannto make IPC message\n" );
		return( -1 );
	}

	mes = proc->file->mes;
	signal( SIGINT, ctrl_c );
	atexit( free_child );
	while( 1 ){
		CtrlCFlag = 0;
/* D0: Ctrl+C and Abort */
		if( Abort_flag ){
		     end_proc( proc, 0 );
		}
/* D0 end */
		if( msgrcv( MesId, (struct msgbuf *)mes, sizeof(struct Message) - sizeof(long), proc->pid, 0 ) == -1 ){
			mes->errorcode = EPIPEREAD;
			mes->status = STFATAL;
			sprintf( mes->u.data, FatalForm, "read pipe error" );
			mes->datasize = strlen( mes->u.data );
			reply( proc, SMAIN );
			continue;
		}
		if( mes->status != STDONE ){
			reply( proc, SMAIN );
                        continue;
		}
		if( mes->command != CKILL ){
			if( init_file( proc ) == -1 ){
				continue;
			}
			if( proc->file->buf && proc->file->buf->status == -1 ){
				printf( "viewer:Status error\nNow re-voke\n" );
				proc->file->pid = 0;
			}
		}
		if( proc->file->pid == 0 ){
/* 			if( mes->command == CLOAD ){ */
			if( mes->command != CKILL ){ 
				if( (ret = make_viewer( proc )) != 0 ){
					mes->status = STWARNING;
					mes->errorcode = ret;
					mes->datasize = 0;
					reply( proc, SMAIN );
					continue;
				}
				CatchFlag = 1;
				pause();
				sleep( 1 );
				CatchFlag = 0;
				if( kill( proc->file->pid, SIGUSR1 ) == -1 ){
					mes->status = STFATAL;
					mes->errorcode = ENOTCREATE;
					mes->datasize = 0;
					reply( proc, SMAIN );
					continue;
				}
			}
		}
		if( proc->file->pid == 0 ){
			if( mes->command != CKILL ){
				mes->status = STWARNING;
				mes->errorcode = ENOTREADY;
				mes->datasize = 0;
				reply( proc, SMAIN );
				continue;
			}
		}
		switch( mes->command ){
		case CLOAD:
		case CEFFECT:
		case CSHOW:
			show_command( proc );
		        break;
		case CSELECT:
			select_command( proc );
		        break;
		case CKILL:
			end_proc( proc, 0 );
			return( 0 );
		        break;
		case CPURGE:
			purge_command( proc );
			break;
		default:
		        break;
		}
	}
	return( 0 );
}

/**********************************************************************/
/*                                                                    */
/*             function catch()                                       */
/*                                                                    */
/*             int signo  : signal No.                                */
/*                                                                   */
/*             return NONE                                            */
/*                                                                    */
/*             function: SIGUSR1 signal handler.(used communication   */
/*                       with GEDO).                                  */
/**********************************************************************/
void catch( int signo )
{
	signal( SIGUSR1, catch );
	if( CatchFlag == 0 ){		/* add C3 */
		kill( Proc.file->pid, SIGTERM );
		Proc.file->pid = 0;
	}
}

/**********************************************************************/
/*                                                                    */
/*             function main()                                        */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Main routine of viewer(1)                    */
/*                                                                    */
/**********************************************************************/
void main()
{
	Proc.file = &File[FileNum];
	File[FileNum].mes  = &Mes;
	signal( SIGUSR1, catch );
	Proc.pid = getpid();
	process( &Proc );
}
