/**********************************************************************/
/*                                                                    */
/*           viewsel.c  viewer selection pattern routine              */
/*                                                                    */
/*                by K.sugiura   Ver 0.92 for GEDO of LINUX           */
/*                               freezed at 1 Nov. 1996               */
/*     External interface
		void selnone( int mode, struct Pos *pos );
		void selpoint( int mode, struct Pos *pos );
		void selbox( int mode, struct Pos *pos );
		void selpoly( int mode, struct Pos *pos );
		void selline( int mode, struct Pos *pos );
		void selfree( int mode, struct Pos *pos );
		void selsppoly( int mode, struct Pos *pos );
		void selspline( int mode, struct Pos *pos ); */
/*                                                                    */
/*     Need this object the spline functions in libspline.            */
/*      D1: arrow key FIX                                             */
/**********************************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

#include "viewer.h"
#define MAXPOSITION 30		/* select position maximum */

extern Display *D;		/* Display struct for X11 */
extern Window   W;		/* Image Window */
extern GC	 gcy;		/* GC of Image Window, XOR and MaskColor draw */

extern struct Pos *Sel;		/* Selected positions. real geometory  loaderx.c*/
extern int Snum;		/* Number of selected positions loaderx.c*/
extern int SelectStatus;	/* kind of selection. in loaderx.c 
					#define ADD       0
					#define CHANGE    1
					#define EDIT      1	*/
extern int Now;			/* Now pointed Sel  in loaderx.c */
extern int MaxPos;		/* Maximum can select Sel. loaderx.c */

extern int SKind;		/* Selection save kind.
				#define SKPOS		0
				#define SKFILE		1 */

/* Operations for prev_stack() */
#define PUSH    0		/* PUSH the position */
#define POPARC  1		/* Draw ARC by stack and POP */
#define DOARC   2		/* Draw ARC by stack and not POP */
#define POPRECT 3		/* Draw RECTangle by stack and POP */
#define DORECT  4		/* Draw RECTangle by stack and not POP */
#define POPLINE 5		/* Draw LINE by stack and POP */
#define DOLINE  6		/* Draw LINE by stakc and not POP */
#define POPSP   7		/* Draw SPLINE by stack and POP */
#define CLEAR  10		/* CLEAR the prev_stack() */

int get_scale( void );

/**********************************************************************/
/*                                                                    */
/*             function search_nearlest()                             */
/*                                                                    */
/*             int x   : X of the source point                        */
/*             int y   : Y of the source point                        */
/*                                                                    */
/*             return  int : The nearest position in Sel              */
/*                                                                    */
/*             function: Search the nearest Sel from argments.        */
/*                                                                    */
/**********************************************************************/
static int search_nearlest( int x, int y )
{
	int i;
	int distx, disty, dist, mini, min;

	min = 256*256; mini = -1;
	for( i = 0; i < Snum; i++ ){
		distx = x - Sel[i].x;
		distx = distx * distx;
		disty = y - Sel[i].y;
		disty = disty * disty;
		dist = distx + disty;
		if( min > dist ){
			min = dist;
			mini = i;
		}
	}
	if( mini == -1 ){
		return( -1 );
	} else {
		return( mini );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function DrawRect()                                    */
/*                                                                    */
/*             struct Pos p1 : Rectangle position1(not need LeftTop)  */
/*             struct Pos p2 : Rectangle position2(not need RightBottom)*/
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw the Rectangle from real geometry to     */
/*                      the Window.                                   */
/**********************************************************************/
static void DrawRect( struct Pos p1, struct Pos p2 )
{
	struct Positions min, max;

	min.real.x = min( p1.x, p2.x );
	max.real.x = max( p1.x, p2.x );
	min.real.y = min( p1.y, p2.y );
	max.real.y = max( p1.y, p2.y );
	position_r2w( &min );
	position_r2w( &max );
	
	XDrawRectangle( D, W, gcy, min.window.x, min.window.y, max.window.x-min.window.x, max.window.y-min.window.y );
	XFlush( D );
}

/**********************************************************************/
/*                                                                    */
/*             function DrawLine()                                    */
/*                                                                    */
/*             struct Pos p1 : Line position1(not need LeftTop)       */
/*             struct Pos p2 : Line position2(not need RightBottom)   */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw the Line from real geometry to          */
/*                      the Window.                                   */
/**********************************************************************/
static void DrawLine( struct Pos p1, struct Pos p2 )
{
	struct Positions min, max;

	min.real.x = p1.x;
	max.real.x = p2.x;
	min.real.y = p1.y;
	max.real.y = p2.y;
	position_r2w( &min );
	position_r2w( &max );
	
	XDrawLine( D, W, gcy, min.window.x, min.window.y, max.window.x, max.window.y);
}

/**********************************************************************/
/*                                                                    */
/*             function DrawArc()                                     */
/*                                                                    */
/*             struct Pos p1 : ARC center position                    */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw the Circle of radius 10 from real       */
/*                      geometory to the Window.                      */
/**********************************************************************/
static void DrawArc( struct Pos p1 )
{
	struct Positions min;

	min.real.x = p1.x;
	min.real.y = p1.y;
	position_r2w( &min );
	
	XDrawArc( D, W, gcy, min.window.x-5, min.window.y-5,10,10,0,360*64);
}

/**********************************************************************/
/*                                                                    */
/*             function DrawCross()                                   */
/*                                                                    */
/*             struct Pos p1 : Cross center position                  */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw the Cross length 10 from real           */
/*                      geometory to the Window.                      */
/**********************************************************************/
static void DrawCross( struct Pos p1 )
{
	struct Positions min;

	min.real.x = p1.x;
	min.real.y = p1.y;
	position_r2w( &min );
	
	XDrawLine( D, W, gcy, min.window.x - 5, min.window.y,
		  min.window.x + 5, min.window.y );
	XDrawLine( D, W, gcy, min.window.x, min.window.y - 5,
		  min.window.x, min.window.y + 5 );
}

/**********************************************************************/
/*                                                                    */
/*             function DrawPoint()                                   */
/*                                                                    */
/*             struct Pos p1 : A Point of the position                */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw the Point from real geometory to the    */
/*                      Window.                                       */
/**********************************************************************/
static void DrawPoint( struct Pos p1 )
{
	struct Positions min;
	int scale, i, j;

	min.real.x = p1.x;
	min.real.y = p1.y;
	position_r2w( &min );

	XDrawPoint( D, W, gcy, min.window.x, min.window.y );
	/* add C2 for display SKFREE selection clearly */
	if( (scale = get_scale()) > 1 ){
		for( i = 1; i < scale; i++ ){
			for( j = 1; j < scale; j++ ){
				XDrawPoint( D, W, gcy, min.window.x+i, min.window.y+j );
			}
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function prev_stack()                                  */
/*                                                                    */
/*             int flag        : command                              */
/*             struct Pos *pos : The Positon or NULL                  */
/*                                                                    */
/*             return  int : 0==success   -1==alloc error             */
/*                                                                    */
/*             function: Stack operation of Draw abstruct process.    */
/*                                                                    */
/**********************************************************************/
static int prev_stack( int flag, struct Pos *pos )
{
	static struct Pos *stack = NULL;
	static int  sp = 0;
	static int  stacksize = 0;
	struct Pos *tmp;

	if( stack == NULL ){
		if( (stack = malloc( sizeof(struct Pos) * 100 )) == NULL ){
			printf( "prev_stack:malloc error\n" );
			return( -1 );
		}
		stacksize = 100;
	}

	switch( flag ){
	case PUSH:
		stack[sp].x = pos->x;
		stack[sp].y = pos->y;
		if( ++sp >= stacksize - 1 ){
			if( (tmp = realloc( stack, sizeof(struct Pos) * (stacksize+100) )) == NULL ){
				printf( "prev_stack:realloc error\n" );
				return( -1 );
			}
			stack = tmp;
			stacksize += 100;
		}
		break;
	case POPARC:
	case DOARC:
		if( sp ){
			DrawCross( stack[0] );
		}
		if( flag == POPARC ){
			sp = 0;
		}
		break;
	case POPLINE:
	case DOLINE:
		if( sp  >= 3 ){
			DrawLine( stack[0], stack[2] );
			DrawLine( stack[0], stack[1] );
		} else if( sp == 2 ){
			DrawLine( stack[0], stack[1] );
		} else if( sp == 1 ){
			DrawArc( stack[0] );
		}
		if( flag == POPLINE ){
			sp = 0;
		}
		break;
	case POPRECT:
	case DORECT:
		if( sp >= 2 ){
			DrawRect( stack[0], stack[1] );
		} else if( sp == 1 ){
			DrawArc( stack[0] );
		}
		if( flag == POPRECT ){
			sp = 0;
		}
		break;
	case POPSP:
		if( sp == 1 ){
			DrawArc( stack[0] );
		} else {
			while( stack[--sp].x != -1 ){
				DrawArc( stack[sp] );
			}
			while( --sp > 0 ){
				DrawLine( stack[sp], stack[sp-1] );
			}
		}
		sp = 0;
		break;
	case CLEAR:
		sp = 0;
		break;
	}
	return( 0 );
}


/**********************************************************************/
/*                                                                    */
/*             function selnone()                                     */
/*                                                                    */
/*             int mode         : Draw Mode                           */
/*             struct Pos *pos  : new Position                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Not Draw dummy Interface for SKNONE.         */
/*                                                                    */
/**********************************************************************/
void selnone( int mode, struct Pos *pos )
{
	return;
}


/**********************************************************************/
/*                                                                    */
/*             function selpoint()                                    */
/*                                                                    */
/*             int mode         : Draw Mode                           */
/*             struct Pos *pos  : new Position                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw Interface to SKPOINT.                   */
/*                                                                    */
/**********************************************************************/
void selpoint( int mode, struct Pos *pos )
{
	int tmp, i;

	if( mode == DRAWALL ){
		prev_stack( CLEAR, NULL );
		for( i = 0; i < Snum; i++ ){
			DrawCross( Sel[i] );
		}
	} else if( mode == DRAWPRESS ){
		prev_stack( CLEAR, NULL );
		if( SelectStatus == CHANGE ){
			if( (tmp = search_nearlest( pos->x, pos->y )) == -1 ){
				return;
			}
			Now = tmp;
			prev_stack( PUSH, &Sel[Now] );
		} else if( SelectStatus == ADD ){
			prev_stack( POPARC, NULL );
			Sel[Now].x = pos->x;
			Sel[Now].y = pos->y;
			prev_stack( PUSH, &Sel[Now] );
			prev_stack( DOARC, NULL );
		}
	} else if( mode == DRAWMOTION || mode == DRAWRELEASE ){
		prev_stack( POPARC, NULL );
		Sel[Now].x = pos->x;
		Sel[Now].y = pos->y;
		prev_stack( PUSH, &Sel[Now] );
		prev_stack( DOARC, NULL );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function selbox()                                      */
/*                                                                    */
/*             int mode         : Draw Mode                           */
/*             struct Pos *pos  : new Position                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw Interface to SKBOX.                     */
/*                                                                    */
/**********************************************************************/
void selbox( int mode, struct Pos *pos )
{
	int tmp, i;

	if( mode == DRAWALL ){
		prev_stack( CLEAR, NULL );
		for( i = 0; i < Snum / 2; i++ ){
			DrawRect( Sel[i*2], Sel[i*2+1] );
		}
		if( Snum % 2 ){
			DrawArc( Sel[i*2] );
		}
	} else if( mode == DRAWPRESS ){
		if( SelectStatus == CHANGE ){
			prev_stack( CLEAR, NULL );
			if( (tmp = search_nearlest( pos->x, pos->y )) == -1 ){
				return;
			}
			Now = tmp;
			if( Now % 2 ){
				prev_stack( PUSH, &Sel[Now-1] );
			} else {
				prev_stack( PUSH, &Sel[Now+1] );
			}
			prev_stack( PUSH, &Sel[Now] );
		} else if( SelectStatus == ADD ){
			prev_stack( POPRECT, NULL );
			Sel[Now].x = pos->x;
			Sel[Now].y = pos->y;
			if( Now % 2 == 1 ){
				prev_stack( PUSH, &Sel[Now-1] );
			}
			prev_stack( PUSH, &Sel[Now] );
			prev_stack( DORECT, NULL );
		}
	} else if( mode == DRAWMOTION || mode == DRAWRELEASE ){
		prev_stack( POPRECT, NULL );
		Sel[Now].x = pos->x;
		Sel[Now].y = pos->y;
		if( Now % 2 == 1 ){
			prev_stack( PUSH, &Sel[Now-1] );
		} else {
			if( SelectStatus == CHANGE ){
				prev_stack( PUSH, &Sel[Now+1] );
			}
		}
		prev_stack( PUSH, &Sel[Now] );
		prev_stack( DORECT, NULL );
		if( SelectStatus == ADD && mode == DRAWRELEASE ){
			prev_stack( CLEAR, NULL );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function poly_line()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Sub-function for selpoly().                  */
/*                                                                    */
/**********************************************************************/
static void poly_line( void )
{
	if( Now == 0 ){
		prev_stack( PUSH, &Sel[1] );
		prev_stack( PUSH, &Sel[Snum-1] );
	} else if( Now == Snum -1 ){
		prev_stack( PUSH, &Sel[0] );
		prev_stack( PUSH, &Sel[Snum-2] );
	} else {
		prev_stack( PUSH, &Sel[Now+1] );
		prev_stack( PUSH, &Sel[Now-1] );
	}
}


/**********************************************************************/
/*                                                                    */
/*             function selpoly()                                     */
/*                                                                    */
/*             int mode         : Draw Mode                           */
/*             struct Pos *pos  : new Position                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw Interface for SKPOLY.                   */
/*                                                                    */
/**********************************************************************/
void selpoly( int mode, struct Pos *pos )
{
	int i;
	int tmp;

	if( mode == DRAWALL ){	
		prev_stack( CLEAR, NULL );
		if( Snum > 1 ){
 			for( i = 0; i < Snum - 1; i++ ){
				DrawLine( Sel[i], Sel[i+1] );
			}
			if( SelectStatus == CHANGE ){
				DrawLine( Sel[i], Sel[0] );
			}
		} else if( Snum == 1 ){
			DrawArc( Sel[0] );
		}
	} else if( mode == DRAWPRESS ){
		if( SelectStatus == CHANGE ){
			if( (tmp = search_nearlest( pos->x, pos->y )) == -1 ){
				return;
			}
			Now = tmp;
			prev_stack( PUSH, &Sel[Now] );
			poly_line();
		} else if( SelectStatus == ADD ){
			prev_stack( POPLINE, NULL );
			Sel[Now].x = pos->x;
			Sel[Now].y = pos->y;
			prev_stack( PUSH, &Sel[Now] );
			if( Snum > 0 ){
				prev_stack( PUSH, &Sel[Now-1] );
			}
			prev_stack( DOLINE, NULL );
		}
	} else if( mode == DRAWMOTION || mode == DRAWRELEASE ){
		prev_stack( POPLINE, NULL );
		Sel[Now].x = pos->x;
		Sel[Now].y = pos->y;
		prev_stack( PUSH, &Sel[Now] );
		if( SelectStatus == ADD ){
			if( Snum > 0 ){
				prev_stack( PUSH, &Sel[Now-1] ); 
			}
			prev_stack( DOLINE, NULL );
		} else if( SelectStatus == CHANGE ){
			poly_line();
			prev_stack( DOLINE, NULL );
		}
		if( mode == DRAWRELEASE ){
			prev_stack( CLEAR, NULL );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function line_line()                                   */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Sub-function for selline()                   */
/*                                                                    */
/**********************************************************************/
static void line_line( void )
{
	if( Now == 0 ){
		prev_stack( PUSH, &Sel[1] );
	} else if( Now == Snum -1 ){
		prev_stack( PUSH, &Sel[Now-1] );
	} else {
		prev_stack( PUSH, &Sel[Now+1] );
		prev_stack( PUSH, &Sel[Now-1] );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function selline()                                     */
/*                                                                    */
/*             int mode         : Draw Mode                           */
/*             struct Pos *pos  : new Position                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw Interface for SKLINE.                   */
/*                                                                    */
/**********************************************************************/
void selline( int mode, struct Pos *pos )
{
	int i;
	int tmp;

	if( mode == DRAWALL ){
		prev_stack( CLEAR, NULL );
		if( Snum > 1 ){
 			for( i = 0; i < Snum - 1; i++ ){
				DrawLine( Sel[i], Sel[i+1] );
			}
		} else if( Snum == 1 ){
			DrawArc( Sel[0] );
		}
	} else if( mode == DRAWPRESS ){
		if( SelectStatus == CHANGE ){
			if( (tmp = search_nearlest( pos->x, pos->y )) == -1 ){
				return;
			}
			Now = tmp;
			prev_stack( PUSH, &Sel[Now] );
			line_line();
		} else if( SelectStatus == ADD ){
			prev_stack( POPLINE, NULL );
			Sel[Now].x = pos->x;
			Sel[Now].y = pos->y;
			prev_stack( PUSH, &Sel[Now] );
			if( Snum > 0 ){
				prev_stack( PUSH, &Sel[Now-1] );
			}
			prev_stack( DOLINE, NULL );
		}
	} else if( mode == DRAWMOTION || mode == DRAWRELEASE ){
		prev_stack( POPLINE, NULL );
		Sel[Now].x = pos->x;
		Sel[Now].y = pos->y;
		prev_stack( PUSH, &Sel[Now] );
		if( SelectStatus == ADD ){
			if( Snum > 0 ){
				prev_stack( PUSH, &Sel[Now-1] ); 
			}
			prev_stack( DOLINE, NULL );
		} else if( SelectStatus == CHANGE ){
			line_line();
			prev_stack( DOLINE, NULL );
		}
		if( mode == DRAWRELEASE ){
			prev_stack( CLEAR, NULL );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function selfree()                                     */
/*                                                                    */
/*             int mode         : Draw Mode                           */
/*             struct Pos *pos  : new Position                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw Interface for SKFREE.                   */
/*                                                                    */
/**********************************************************************/
void selfree( int mode, struct Pos *pos )
{
	int i;

	if( mode == DRAWALL ){
		for( i = 0; i < Snum; i++ ){
			DrawPoint( Sel[i] );
		}
	} else if( mode == DRAWMOTION || mode == DRAWPRESS || mode == DRAWRELEASE ){
		if( Now > 0 && Sel[Now-1].x == pos->x 
		   && Sel[Now-1].y == pos->y ){		/* add C1 */
			return;
		}
		Sel[Now].x = pos->x;
		Sel[Now].y = pos->y;
		if( SKind == SKPOS && Snum >= MaxPos ){
			printf( "ignored\n" );
			return;
		} else if( SKind == SKFILE && Snum >= MaxPos ){
			if( realloc_sel( MAXPOSITION ) == -1 ){
				printf( "realloc fail:ignore\n" );
				return;
			}
		}
		DrawPoint( Sel[Now] );
		Snum++; Now++;
	} else if( mode == DRAWDEL ){
		if( --Snum < 0 ){
			Snum = 0;
		}
		if( --Now < 0 ){
			Now = 0;
		} else {
			DrawPoint( Sel[Now] );
		}
	}
}

/**********************************************************************/
/*                                                                    */
/*             function draw_allspline_poly()                          */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Sub-fucntion for selsppoly()                 */
/*                                                                    */
/**********************************************************************/
static void draw_allspline_poly( void )
{
	int total;
	struct Pos x0, x1;
	struct fPosition fpos[MAXSP];
	struct fPosition *ret;
	int i;

	if( Snum > 2 ){
		for( i = 0; i < Snum; i++ ){
			fpos[i].x = (float)Sel[i].x;
			fpos[i].y = (float)Sel[i].y;
		}
		total = Snum * 10;
		
		if( (ret = spline_poly( fpos, 3, Snum, 10 )) == NULL ){
			return;
		}
		x0.x = (int)ret[0].x; x0.y = (int)ret[0].y;
		prev_stack( PUSH, &x0 );
		for( i = 1; i < total; i++ ){
			x1.x = (int)ret[i].x; x1.y = (int)ret[i].y;
			DrawLine( x0, x1 );
			x0.x = x1.x; x0.y = x1.y;
			prev_stack( PUSH, &x1 );
		}
		x1.x = (int)ret[0].x; x1.y = (int)ret[0].y;
		DrawLine( x0, x1 );
		prev_stack( PUSH, &x1 );
		x1.x = x1.y = -1;
		prev_stack( PUSH, &x1 );
		for( i = 0; i < Snum; i++ ){
			DrawArc( Sel[i] );
			prev_stack( PUSH, &Sel[i] );
		}
	} else if( Snum == 1 ){
		DrawArc( Sel[0] );
		prev_stack( PUSH, &Sel[0] );
	} else if( Snum == 2 ){
		DrawLine( Sel[0], Sel[1] );
		prev_stack( PUSH, &Sel[0] );
		prev_stack( PUSH, &Sel[1] );
		x1.x = x1.y = -1;
		prev_stack( PUSH, &x1 );
		DrawArc( Sel[0] );
		prev_stack( PUSH, &Sel[0] );
		DrawArc( Sel[1] );
		prev_stack( PUSH, &Sel[1] );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function selsppoly()                                   */
/*                                                                    */
/*             int mode         : Draw Mode                           */
/*             struct Pos *pos  : new Position                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Not Draw dummy Interface for SKSPPOLY        */
/*                                                                    */
/**********************************************************************/
void selsppoly( int mode, struct Pos *pos )
{
	int tmp;

	if( mode == DRAWALL ){
		prev_stack( CLEAR, NULL );
		draw_allspline_poly();
	} else if( mode == DRAWPRESS ){
		if( SelectStatus == CHANGE ){
			if( (tmp = search_nearlest( pos->x, pos->y )) == -1 ){
				return;
			}
			Now = tmp;
		} else if( SelectStatus == ADD ){
			prev_stack( POPLINE, NULL );
			Sel[Now].x = pos->x;
			Sel[Now].y = pos->y;
			prev_stack( PUSH, &Sel[Now] );
			if( Snum > 0 ){
				prev_stack( PUSH, &Sel[Now-1] );
			}
			prev_stack( DOLINE, NULL );
		}
	} else if( mode == DRAWMOTION || mode == DRAWRELEASE ){
		if( SelectStatus == ADD ){
			prev_stack( POPLINE, NULL );
		} else {
			prev_stack( POPSP, NULL );
		}
		Sel[Now].x = pos->x;
		Sel[Now].y = pos->y;
		if( SelectStatus == ADD ){
			prev_stack( PUSH, &Sel[Now] );
			if( Snum > 0 ){
				prev_stack( PUSH, &Sel[Now-1] ); 
			}
			prev_stack( DOLINE, NULL );
			if( mode == DRAWRELEASE ){
				prev_stack( CLEAR, NULL );
			}
		} else if( SelectStatus == CHANGE ){
			draw_allspline_poly();
		}
	}
}


/**********************************************************************/
/*                                                                    */
/*             function draw_allspline_line()                          */
/*                                                                    */
/*             args   NONE                                            */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Sub-function for selspline()                 */
/*                                                                    */
/**********************************************************************/
static void draw_allspline_line()
{
	int total;
	struct Pos x0, x1;
	struct fPosition fpos[MAXSP];
	struct fPosition *ret;
	int i;

	if( Snum > 1 ){
		for( i = 0; i < Snum; i++ ){
			fpos[i].x = (float)Sel[i].x;
			fpos[i].y = (float)Sel[i].y;
		}
		total = (Snum - 1) * 10;
		
		if( (ret = spline_line( fpos, 3, Snum, 10 )) == NULL ){
			return;
		}
		x0.x = (int)fpos[0].x; x0.y = (int)fpos[0].y;
		prev_stack( PUSH, &x0 );
		for( i = 0; i < total; i++ ){
			x1.x = (int)ret[i].x; x1.y = (int)ret[i].y;
			DrawLine( x0 , x1 );
			x0.x = x1.x; x0.y = x1.y;
			prev_stack( PUSH, &x1 );
		}
		x1.x = x1.y = -1;
		prev_stack( PUSH, &x1 );
		for( i = 0; i < Snum; i++ ){
			DrawArc( Sel[i] );
			prev_stack( PUSH, &Sel[i] );
		}
	} else if( Snum == 1 ){
		DrawArc( Sel[0] );
		prev_stack( PUSH, &Sel[0] );
	}
}

/**********************************************************************/
/*                                                                    */
/*             function selspline()                                   */
/*                                                                    */
/*             int mode         : Draw Mode                           */
/*             struct Pos *pos  : new Position                        */
/*                                                                    */
/*             return NONE                                            */
/*                                                                    */
/*             function: Draw Interface for SKSPLINE.                 */
/*                                                                    */
/**********************************************************************/
void selspline( int mode, struct Pos *pos )
{
	int tmp;

	if( mode == DRAWALL ){
		prev_stack( CLEAR, NULL );
		draw_allspline_line();
	} else if( mode == DRAWPRESS ){
		if( SelectStatus == CHANGE ){
			if( (tmp = search_nearlest( pos->x, pos->y )) == -1 ){
				return;
			}
			Now = tmp;
		} else if( SelectStatus == ADD ){
			prev_stack( POPLINE, NULL );
			Sel[Now].x = pos->x;
			Sel[Now].y = pos->y;
			prev_stack( PUSH, &Sel[Now] );
			if( Snum > 0 ){
				prev_stack( PUSH, &Sel[Now-1] );
			}
			prev_stack( DOLINE, NULL );
		}
	} else if( mode == DRAWMOTION || mode == DRAWRELEASE ){
		if( SelectStatus == ADD ){
			prev_stack( POPLINE, NULL );
		} else {
			prev_stack( POPSP, NULL );
		}
		Sel[Now].x = pos->x;
		Sel[Now].y = pos->y;
		if( SelectStatus == ADD ){
			prev_stack( PUSH, &Sel[Now] );
			if( Snum > 0 ){
				prev_stack( PUSH, &Sel[Now-1] ); 
			}
			prev_stack( DOLINE, NULL );
			if( mode == DRAWRELEASE ){
				prev_stack( CLEAR, NULL );
			}
		} else if( SelectStatus == CHANGE ){
			draw_allspline_line();
		}
	}
}
